Package: WFC
Title: Workflow-Oriented Survey Weight Calibration
Version: 2.0.1
Authors@R: c(
    person(given = "Kunxiang", family = "Ma", email = "makunxiang@weiandata.com", role = c("aut", "cre")),
    person("WEIAN DATA TECH (Beijing) Co., Ltd.", role = c("cph", "fnd"),
           email = "contact@weiandata.com"))
Author: Kunxiang Ma [aut, cre],
  WEIAN DATA TECH (Beijing) Co., Ltd. [cph, fnd]
Maintainer: Kunxiang Ma <makunxiang@weiandata.com>
Description: Provides a disciplined precheck, execution, and diagnostics workflow
    for survey weighting and raking. Weight construction requires design-only data,
    a verified external target, outcome-blind planning, and human approval before
    weight locking, with bilingual reports for decision and statistical audiences.
    Converts calibrated and replicate weights into standard survey designs,
    provides optional broom-style result projections, and records serializable
    production pipeline provenance. Supports fixed, predeclared soft calibration
    tolerances and categorical entropy balancing from verified margins, plus panel
    attrition weighting, high-influence unit diagnostics, Fay's balanced repeated
    replication, and opt-in parallel execution for long runs.
    Calibration methods follow Deville and Saerndal (1992)
    <doi:10.1080/01621459.1992.10475217>, and entropy balancing follows
    Hainmueller (2012) <doi:10.1093/pan/mpr025>.
License: GPL (>= 2)
Copyright: See file inst/COPYRIGHTS.
URL: https://github.com/weiandata/WFC
BugReports: https://github.com/weiandata/WFC/issues
Encoding: UTF-8
Depends: R (>= 3.6.0)
Imports: digest, parallel, stats, utils
LazyData: true
Suggests: cli, covr, generics, knitr, openxlsx, rmarkdown, survey,
        testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-28 21:39:56 UTC; makunxiang
Repository: CRAN
Date/Publication: 2026-08-06 13:50:06 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,wf_quality_report)
S3method(as.data.frame,wf_variance_result)
S3method(generics::augment,wf_weights)
S3method(generics::glance,wf_blend_result)
S3method(generics::glance,wf_diagnostics)
S3method(generics::glance,wf_variance_result)
S3method(generics::glance,wf_weights)
S3method(generics::tidy,wf_blend_result)
S3method(generics::tidy,wf_diagnostics)
S3method(generics::tidy,wf_variance_result)
S3method(generics::tidy,wf_weights)
S3method(plot,wf_auto_trim)
S3method(plot,wf_blend_result)
S3method(plot,wf_diagnostics)
S3method(plot,wf_propensity_weights)
S3method(plot,wf_weights)
S3method(print,wf_attrition_weights)
S3method(print,wf_auto_trim)
S3method(print,wf_autoweigh_result)
S3method(print,wf_blend_result)
S3method(print,wf_cell_merge_plan)
S3method(print,wf_collapse_plan)
S3method(print,wf_design_data)
S3method(print,wf_diagnostics)
S3method(print,wf_impact)
S3method(print,wf_influence)
S3method(print,wf_ladder_draft)
S3method(print,wf_pipeline)
S3method(print,wf_poststrat_plan)
S3method(print,wf_precheck)
S3method(print,wf_quality_report)
S3method(print,wf_replicate_weights)
S3method(print,wf_validation)
S3method(print,wf_variance_result)
S3method(print,wf_weight_plan)
S3method(print,wf_weights)
S3method(summary,wf_poststrat_plan)
export(as_svrepdesign)
export(as_svydesign)
export(wf_apply_collapse)
export(wf_approve_plan)
export(wf_assess_impact)
export(wf_attach_weights)
export(wf_attrition)
export(wf_audit_export)
export(wf_auto_trim)
export(wf_autoweigh)
export(wf_blend)
export(wf_calibrate)
export(wf_collapse_ladder)
export(wf_compose)
export(wf_diagnose)
export(wf_dims)
export(wf_execute_plan)
export(wf_guided_execute)
export(wf_guided_plan)
export(wf_import_reference)
export(wf_import_target)
export(wf_influence)
export(wf_pipeline)
export(wf_plan_cells)
export(wf_plan_poststrat)
export(wf_plan_weights)
export(wf_poststrat)
export(wf_precheck)
export(wf_prepare_design)
export(wf_propensity)
export(wf_rake)
export(wf_replicates)
export(wf_report)
export(wf_run)
export(wf_suggest_collapse)
export(wf_suggest_ladder)
export(wf_target_population)
export(wf_target_propensity)
export(wf_target_reference)
export(wf_target_template)
export(wf_validate)
export(wf_variance)

# WFC

WFC is an R package for reviewable survey weighting. Version 2.0 makes one
rule non-optional: weights may be built only from declared design variables and
an independently sourced, verified target. Study outcomes enter only after the
weights are locked.

This design serves two kinds of users:

- survey researchers get a guided sequence with plain-language stops and next
  actions; and
- statistical teams and AI agents get stable objects, identities, condition
  codes, and complete audit fields.

The same result can be shown as a short decision view or a detailed statistical
view. The two views come from one object and must agree.

## What WFC 2.0 prevents

Public weighting functions reject raw data frames, ordinary target objects,
demo targets, changed identities, self-approval by an agent, and runtime
changes to ID or base-weight roles. Manual targets, target shrinkage, inline
target moments, manual pipeline modes, and runtime margin injection are not
supported.

These controls reduce accidental and ordinary misuse. They cannot prove that a
source document is truthful or stop a determined person from editing open
source code. Accountable human review remains necessary.

## Installation

Install the development version from GitHub:

```r
remotes::install_github("weiandata/WFC")
```

## Files used by the safe workflow

A production run normally starts with four files:

1. a design-only survey table containing IDs, calibration fields, and declared
   sampling-design fields;
2. a separate analysis table containing study outcomes;
3. a CSV or Excel target table from an external authority; and
4. a companion `.source.dcf` evidence record for that target table.

Use `wf_target_template()` if the target-table layout is unfamiliar:

```r
library(WFC)

dims <- wf_dims(
  age_group = c("18-34", "35-54", "55+"),
  region = c("north", "south")
)

wf_target_template(
  "population-margins-template.csv",
  dims = dims
)
```

The template creates the data file and its companion DCF form. Complete the
source fields, update the checksum after the data file is final, and obtain the
file before inspecting study outcomes.

WFC also installs synthetic format examples named `safe-target-example.csv`,
`safe-target-example.xlsx`, and their separate `.source.dcf` files. They are
marked demo-only and cannot enter production planning.

<!-- SAFE_WORKFLOW_START -->

## Complete controlled workflow

### 1. Prepare outcome-free design data

```r
library(WFC)

dims <- wf_dims(
  age_group = c("18-34", "35-54", "55+"),
  region = c("north", "south")
)

design_only <- read.csv("survey-design.csv")
analysis_data <- read.csv("survey-outcomes.csv")

design <- wf_prepare_design(
  design_only,
  id = "person_id",
  calibration = c("age_group", "region"),
  base_weight = "base_weight"
)
```

Every column in `design_only` must have a declared design role. Keep outcomes
such as satisfaction, vote choice, approval, score, or pass/fail in
`analysis_data`.

### 2. Import an external target and its evidence

Population-count target:

```r
target <- wf_import_target(
  data_file = "population-margins.csv",
  source_file = "population-margins.csv.source.dcf",
  dims = dims,
  key_map = c(age_group = "age_group", region = "region"),
  count = "population_count",
  production = TRUE
)
```

Independent reference-sample target:

```r
reference_target <- wf_import_reference(
  data_file = "reference-sample.csv",
  source_file = "reference-sample.csv.source.dcf",
  dims = dims,
  feature = "reference_weight",
  production = TRUE
)
```

Import verifies source completeness, declared selection timing, demo status,
and the SHA-256 checksum. It does not decide whether the source is scientifically
appropriate; that remains part of human review.

### 3. Plan without seeing outcomes

```r
cell_plan <- wf_plan_cells(
  design,
  target,
  dims,
  min_cell = 5,
  max_weight_ratio = 4
)

plan <- wf_plan_weights(
  design,
  target,
  dims,
  method = "raking",
  bounds = c(0.3, 3),
  min_cell = 5,
  cell_plan = cell_plan
)

plan$ready
plan$issues
is.null(plan$weights)
```

Planning does not calculate weights. It records the exact inputs, checks,
method, limits, and any deterministic category merge for review.

### 4. Obtain separate human approval

```r
approval <- wf_approve_plan(
  plan,
  approver = "Qualified reviewer full name",
  role = "Statistician",
  note = "Reviewed source, support, method, limits, and intended use"
)
```

An AI agent may prepare the plan but may not create this attestation for itself.
The name and role must identify the actual human reviewer.

### 5. Execute the unchanged plan once

```r
locked <- wf_execute_plan(
  plan,
  approval,
  design,
  target
)
```

Changing the design, target, plan, or approval breaks the identity chain and
stops execution.

### 6. Attach outcomes after weight locking

```r
analysis_ready <- wf_attach_weights(
  analysis_data,
  locked,
  id = "person_id",
  weight_name = ".weight"
)

impact <- wf_assess_impact(
  locked,
  analysis_data,
  id = "person_id",
  outcomes = c("satisfaction", "approved")
)
```

Impact assessment describes what the already locked weights change. It cannot
re-plan, reapprove, or overwrite the weights.

### 7. Produce views for different readers

```r
decision_view <- wf_report(
  locked,
  audience = "decision"
)

statistical_view <- wf_report(
  locked,
  audience = "statistician"
)

impact_detail <- wf_report(
  impact,
  audience = "statistician"
)

wf_audit_export(locked, "weighting-audit.json")
```

The decision view emphasizes status, main risks, and next actions. The
statistical view exposes full tables, convergence information, identities, and
provenance for further analysis.

<!-- SAFE_WORKFLOW_END -->

## Direct functions for statistical users

Advanced users may call `wf_calibrate()`, `wf_rake()`, `wf_poststrat()`,
`wf_auto_trim()`, or `wf_autoweigh()` directly, but their first inputs must
still be the unchanged `design` and `target` objects:

```r
raked <- wf_rake(design, target, tol = 1e-8)

bounded <- wf_calibrate(
  design,
  target,
  method = "logit",
  bounds = c(0.3, 3)
)

trim_review <- wf_auto_trim(
  design,
  target,
  caps = c(2, 4, 6, 8)
)
```

ID and base-weight columns come from `wf_prepare_design()` and cannot be
overridden at this stage.

## AI agent integration

An agent should preserve whole WFC objects and their identities. A minimal
handoff object can be prepared like this:

```r
handoff <- list(
  plan = plan,
  plan_identity = plan$identity,
  design_identity = design$identity,
  target_identity = target$identity,
  required_human_action = "Review and approve the unchanged plan"
)
```

If an agent attempts approval, WFC returns `wf_error_safety`. Agents should read
the stable payload and stop:

```r
refusal <- tryCatch(
  wf_approve_plan(
    plan,
    approver = "Automated agent",
    role = "assistant",
    actor_type = "agent"
  ),
  wf_error_safety = function(condition) condition$data
)

refusal[c("code", "severity", "field", "next_actions")]
```

Do not edit identities, manufacture a human name, widen limits, switch targets,
or retry after viewing outcomes.

## Migration from WFC 1.x

See [Migrating from WFC 1.x to WFC 2.0](https://github.com/weiandata/WFC/blob/main/docs/migration/wfc-1-to-2.md). Behavior
that selected desired results has no compatibility switch and no supported
replacement.

## License

WFC is licensed under GPL (>= 2). Copyright and contributor information are in
`inst/COPYRIGHTS` and `DESCRIPTION`.

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----influence, eval=FALSE----------------------------------------------------
# influence <- wf_influence(
#   locked,
#   target = target,
#   sample = design$data,
#   id = design$roles$id,
#   top = 20
# )
# 
# influence$table

## ----impact, eval=FALSE-------------------------------------------------------
# impact <- wf_assess_impact(
#   locked,
#   analysis_data,
#   id = "person_id",
#   outcomes = c("score", "approved")
# )
# 
# wf_report(impact, audience = "statistician")

#' Draft a post-stratification collapse ladder
#'
#' Examines worst-group category support and proposes adjacent category merges
#' in the explicit level order declared by [wf_dims()]. The result is a draft
#' for review; no input is modified and no merge is applied automatically.
#'
#' @param sample Sample data frame.
#' @param target A `wf_target` object.
#' @param dims A `wf_dims` object with explicit ordered levels for every
#'   calibration dimension.
#' @param min_cell Minimum required unweighted sample support in every group.
#'
#' @return A `wf_ladder_draft` containing reviewable levels and a validated
#'   `wf_collapse_ladder` in `$ladder`.
#' @export
#' @examples
#' data(wfc_example)
#' target <- wf_target_population(
#'   wfc_example$population,
#'   c(gender = "gender", age = "age"),
#'   "count",
#'   wfc_example$dims,
#'   by = "province"
#' )
#' wf_suggest_ladder(wfc_example$sample, target, wfc_example$dims, min_cell = 25)
wf_suggest_ladder <- function(sample, target, dims, min_cell = 5) {
  if (!is.data.frame(sample) || nrow(sample) == 0) {
    wf_abort("`sample` must be a non-empty data frame.", "wf_error_input")
  }
  if (!inherits(target, "wf_target")) {
    wf_abort("`target` must be a wf_target object.", "wf_error_input")
  }
  if (!inherits(dims, "wf_dims")) {
    wf_abort("`dims` must be a wf_dims object.", "wf_error_input")
  }
  dvars <- names(dims$vars)
  if (!identical(target$dims, dvars)) {
    wf_abort(
      "`dims` must declare the target dimensions in the same order.",
      "wf_error_schema",
      list(target_dims = target$dims, dims = dvars)
    )
  }
  if (any(vapply(dims$vars, is.null, logical(1)))) {
    wf_abort(
      paste(
        "wf_suggest_ladder() requires explicit ordered levels for every",
        "dimension; inferred levels cannot define defensible adjacency."
      ),
      "wf_error_input"
    )
  }
  if (length(min_cell) != 1 || !is.finite(min_cell) || min_cell < 1 ||
      min_cell != as.integer(min_cell)) {
    wf_abort("`min_cell` must be a positive integer.", "wf_error_input")
  }

  .require_cols(sample, c(target$by, dvars), "sample")
  gkey <- .wf_group_keys(sample, target$by)
  if (any(is.na(gkey))) {
    wf_abort(
      "Sample grouping keys must not contain NA when drafting a ladder.",
      "wf_error_schema"
    )
  }
  groups <- names(target$groups)
  missing_groups <- setdiff(groups, unique(gkey))
  if (length(missing_groups) > 0) {
    wf_abort(
      sprintf(
        "Cannot draft a ladder: target group(s) have no sample rows: %s.",
        paste(missing_groups, collapse = ", ")
      ),
      "wf_error_feasibility",
      list(groups = missing_groups)
    )
  }

  t0 <- Sys.time()
  drafts <- list()
  affected <- numeric()
  support_before <- list()

  for (d in dvars) {
    values <- .chr(sample[[d]])
    levels <- dims$vars[[d]]
    unknown <- setdiff(unique(values[!is.na(values)]), levels)
    if (length(unknown) > 0) {
      wf_abort(
        sprintf(
          "Sample dimension '%s' contains undeclared level(s): %s.",
          d,
          paste(unknown, collapse = ", ")
        ),
        "wf_error_schema",
        list(dim = d, levels = unknown)
      )
    }

    support <- matrix(
      0L,
      nrow = length(levels),
      ncol = length(groups),
      dimnames = list(levels, groups)
    )
    for (g in groups) {
      for (level in levels) {
        support[level, g] <- sum(gkey == g & !is.na(values) & values == level)
      }
    }
    support_before[[d]] <- support

    partition <- as.list(levels)
    partition_worst <- function(parts) {
      vapply(parts, function(categories) {
        min(colSums(support[categories, , drop = FALSE]))
      }, numeric(1))
    }
    worst <- partition_worst(partition)
    if (all(worst >= min_cell)) {
      next
    }

    while (any(worst < min_cell)) {
      if (length(partition) == 1) {
        wf_abort(
          sprintf(
            "Dimension '%s' cannot reach min_cell = %d even after full collapse.",
            d,
            as.integer(min_cell)
          ),
          "wf_error_feasibility",
          list(dim = d, min_cell = min_cell, support = worst)
        )
      }
      i <- which.min(worst)[1]
      neighbour <- if (i == 1) {
        2L
      } else if (i == length(partition)) {
        i - 1L
      } else if (worst[i - 1L] <= worst[i + 1L]) {
        i - 1L
      } else {
        i + 1L
      }
      left <- min(i, neighbour)
      right <- max(i, neighbour)
      partition[[left]] <- c(partition[[left]], partition[[right]])
      partition[[right]] <- NULL
      worst <- partition_worst(partition)
    }

    merged <- partition[lengths(partition) > 1]
    if (length(merged) == 0) {
      next
    }
    merged_names <- vapply(merged, paste, character(1), collapse = "+")
    map <- stats::setNames(
      rep(merged_names, lengths(merged)),
      unlist(merged, use.names = FALSE)
    )
    drafts[[d]] <- map
    affected[[d]] <- sum(!is.na(values) & values %in% names(map)) /
      sum(!is.na(values))
  }

  order_dims <- if (length(affected) > 0) {
    names(sort(affected, method = "radix"))
  } else {
    character()
  }
  levels <- lapply(order_dims, function(d) {
    stats::setNames(list(drafts[[d]]), d)
  })
  if (length(levels) > 0) {
    names(levels) <- paste0("level", seq_along(levels))
  }
  affected <- affected[order_dims]
  ladder <- do.call(wf_collapse_ladder, c(list(dims), levels))

  structure(
    list(
      levels = levels,
      affected_share = affected,
      support_before = support_before,
      min_cell = as.integer(min_cell),
      ladder = ladder,
      provenance = list(
        method = "suggest_ladder",
        by = target$by,
        dims = dvars,
        created = t0,
        package_version = .wf_package_version()
      )
    ),
    class = "wf_ladder_draft"
  )
}

#' Print a collapse-ladder draft
#'
#' @param x A `wf_ladder_draft` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_ladder_draft <- function(x, ...) {
  cat(sprintf(
    "<wf_ladder_draft> %d level(s); min_cell = %d (review before use)\n",
    length(x$levels),
    x$min_cell
  ))
  for (level in names(x$levels)) {
    d <- names(x$levels[[level]])
    map <- x$levels[[level]][[d]]
    cat(sprintf(
      "  %s [%s, %.1f%% affected]: %s\n",
      level,
      d,
      100 * x$affected_share[[d]],
      paste(sprintf("%s->%s", names(map), map), collapse = ", ")
    ))
  }
  invisible(x)
}

#' Return the loaded package version for blend provenance.
#'
#' @keywords internal
#' @noRd
.wf_blend_package_version <- function() {
  tryCatch(
    as.character(utils::packageVersion("WFC")),
    error = function(e) "0.5.0"
  )
}

#' Validate one argument against allowed values without partial matching.
#'
#' @param value User-supplied scalar character value.
#' @param choices Supported values.
#' @param arg Argument name for error messages.
#' @keywords internal
#' @noRd
.wf_blend_match <- function(value, choices, arg) {
  if (length(value) == length(choices) && identical(value, choices)) {
    value <- value[[1]]
  }
  if (length(value) != 1 || is.na(value) || !value %in% choices) {
    shown <- if (length(value) == 0) "<empty>" else as.character(value[[1]])
    wf_abort(
      sprintf(
        "Unsupported %s '%s'. Supported values: %s.",
        arg,
        shown,
        paste(choices, collapse = ", ")
      ),
      "wf_error_input",
      stats::setNames(list(value), arg)
    )
  }
  value
}

#' Validate a blend cell variable vector.
#'
#' @param by_cell Requested cell columns.
#' @keywords internal
#' @noRd
.wf_blend_by_cell <- function(by_cell) {
  if (!is.character(by_cell) || length(by_cell) < 1 || any(is.na(by_cell)) || any(by_cell == "")) {
    wf_abort("`by_cell` must contain at least one non-empty column name.", "wf_error_input")
  }
  if (anyDuplicated(by_cell)) {
    wf_abort("`by_cell` must not contain duplicated column names.", "wf_error_input")
  }
  by_cell
}

#' Validate trim bounds for data-driven lambdas.
#'
#' @param trim_lambda Two finite bounds in increasing order.
#' @keywords internal
#' @noRd
.wf_blend_trim <- function(trim_lambda) {
  if (!is.numeric(trim_lambda) || length(trim_lambda) != 2 ||
    any(is.na(trim_lambda)) || any(!is.finite(trim_lambda)) ||
    trim_lambda[[1]] < 0 || trim_lambda[[2]] > 1 ||
    trim_lambda[[1]] > trim_lambda[[2]]) {
    wf_abort("`trim_lambda` must be two finite increasing values inside [0, 1].", "wf_error_input")
  }
  as.numeric(trim_lambda)
}

#' Validate one wf_weights source for blending.
#'
#' @param source Input object.
#' @param label Source label.
#' @param by_cell Cell columns.
#' @param outcome Optional outcome column.
#' @keywords internal
#' @noRd
.wf_blend_check_source <- function(source, label, by_cell, outcome) {
  if (!inherits(source, "wf_weights")) {
    wf_abort(
      sprintf("`%s` must be a wf_weights object.", label),
      "wf_error_input",
      list(source = label)
    )
  }
  if (!is.data.frame(source$data)) {
    wf_abort(
      sprintf("`%s$data` must be a data frame.", label),
      "wf_error_schema",
      list(source = label)
    )
  }

  required <- c("group", "weight", by_cell)
  if (!is.null(outcome)) {
    required <- c(required, outcome)
  }
  .require_cols(source$data, required, sprintf("%s$data", label))

  weight <- source$data$weight
  if (!is.numeric(weight) || any(is.na(weight)) || any(!is.finite(weight)) || any(weight < 0)) {
    wf_abort(
      sprintf("`%s$data$weight` must be finite, non-missing, and non-negative.", label),
      "wf_error_input",
      list(source = label)
    )
  }

  group <- .chr(source$data$group)
  if (any(is.na(group)) || any(group == "")) {
    wf_abort(
      sprintf("`%s$data$group` must not contain missing or empty values.", label),
      "wf_error_input",
      list(source = label)
    )
  }

  for (cell_col in by_cell) {
    cell_value <- .chr(source$data[[cell_col]])
    if (any(is.na(cell_value)) || any(cell_value == "")) {
      wf_abort(
        sprintf("`%s$data$%s` must not contain missing or empty values.", label, cell_col),
        "wf_error_input",
        list(source = label, column = cell_col)
      )
    }
  }

  if (!is.null(outcome)) {
    y <- source$data[[outcome]]
    if (!is.numeric(y) || all(is.na(y))) {
      wf_abort(
        sprintf("`%s$data$%s` must be numeric and not entirely missing.", label, outcome),
        "wf_error_input",
        list(source = label, outcome = outcome)
      )
    }
  }

  invisible(source)
}

#' Validate fixed lambda settings.
#'
#' @param lambda Strategy.
#' @param lambda_fixed User-supplied fixed lambda.
#' @keywords internal
#' @noRd
.wf_blend_check_fixed_required <- function(lambda, lambda_fixed) {
  if (lambda == "fixed" && is.null(lambda_fixed)) {
    wf_abort("`lambda_fixed` is required when `lambda = \"fixed\"`.", "wf_error_input")
  }
  invisible(lambda_fixed)
}

#' Build grouping keys for blend summaries.
#'
#' @param data Source data.
#' @param by_cell Cell columns.
#' @keywords internal
#' @noRd
.wf_blend_key <- function(data, by_cell) {
  key_data <- data.frame(group = .chr(data$group), stringsAsFactors = FALSE)
  for (cell_col in by_cell) {
    key_data[[cell_col]] <- .chr(data[[cell_col]])
  }
  .wf_cell_key(as.matrix(key_data), names(key_data))
}

#' Summarize one source by cell.
#'
#' @param source Source `wf_weights`.
#' @param label Source label.
#' @param by_cell Cell columns.
#' @param outcome Optional outcome column.
#' @keywords internal
#' @noRd
.wf_blend_source_cells <- function(source, label, by_cell, outcome) {
  data <- source$data
  key <- .wf_blend_key(data, by_cell)
  split_rows <- split(seq_len(nrow(data)), key)
  rows <- lapply(split_rows, function(idx) {
    part <- data[idx, , drop = FALSE]
    w_all <- as.numeric(part$weight)
    key_values <- data.frame(
      group = .chr(part$group[[1]]),
      stringsAsFactors = FALSE
    )
    for (cell_col in by_cell) {
      key_values[[cell_col]] <- .chr(part[[cell_col]][[1]])
    }

    if (is.null(outcome)) {
      estimate <- NA_real_
      variance <- NA_real_
      missing_outcome <- NA_integer_
      contributing <- w_all > 0
      w <- w_all[contributing]
    } else {
      y_all <- part[[outcome]]
      contributing <- w_all > 0 & !is.na(y_all)
      w <- w_all[contributing]
      y <- y_all[contributing]
      missing_outcome <- sum(is.na(y_all))
      if (length(w) > 0 && sum(w) > 0) {
        estimate <- sum(w * y) / sum(w)
        variance <- sum((w^2) * ((y - estimate)^2)) / (sum(w)^2)
      } else {
        estimate <- NA_real_
        variance <- NA_real_
      }
    }

    weight_sum <- sum(w)
    neff <- if (length(w) > 0 && sum(w^2) > 0) {
      (sum(w)^2) / sum(w^2)
    } else {
      0
    }

    cbind(
      key_values,
      data.frame(
        row_count = length(idx),
        weight_sum = weight_sum,
        neff = neff,
        estimate = estimate,
        variance = variance,
        missing_outcome = missing_outcome,
        estimable = if (is.null(outcome)) weight_sum > 0 else is.finite(estimate),
        stringsAsFactors = FALSE
      )
    )
  })

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  stat_cols <- c(
    "row_count", "weight_sum", "neff", "estimate",
    "variance", "missing_outcome", "estimable"
  )
  names(out)[names(out) %in% stat_cols] <- paste0(
    names(out)[names(out) %in% stat_cols],
    "_",
    label
  )
  out
}

#' Summarize one source by group for group-level lambda.
#'
#' @param source Source `wf_weights`.
#' @param label Source label.
#' @param outcome Optional outcome column.
#' @keywords internal
#' @noRd
.wf_blend_source_groups <- function(source, label, outcome) {
  data <- source$data
  split_rows <- split(seq_len(nrow(data)), .chr(data$group))
  rows <- lapply(names(split_rows), function(group_name) {
    idx <- split_rows[[group_name]]
    part <- data[idx, , drop = FALSE]
    w_all <- as.numeric(part$weight)

    if (is.null(outcome)) {
      estimate <- NA_real_
      variance <- NA_real_
      missing_outcome <- NA_integer_
      contributing <- w_all > 0
      w <- w_all[contributing]
    } else {
      y_all <- part[[outcome]]
      contributing <- w_all > 0 & !is.na(y_all)
      w <- w_all[contributing]
      y <- y_all[contributing]
      missing_outcome <- sum(is.na(y_all))
      if (length(w) > 0 && sum(w) > 0) {
        estimate <- sum(w * y) / sum(w)
        variance <- sum((w^2) * ((y - estimate)^2)) / (sum(w)^2)
      } else {
        estimate <- NA_real_
        variance <- NA_real_
      }
    }

    weight_sum <- sum(w)
    neff <- if (length(w) > 0 && sum(w^2) > 0) {
      (sum(w)^2) / sum(w^2)
    } else {
      0
    }

    data.frame(
      group = group_name,
      row_count = length(idx),
      weight_sum = weight_sum,
      neff = neff,
      estimate = estimate,
      variance = variance,
      missing_outcome = missing_outcome,
      estimable = if (is.null(outcome)) weight_sum > 0 else is.finite(estimate),
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  stat_cols <- c(
    "row_count", "weight_sum", "neff", "estimate",
    "variance", "missing_outcome", "estimable"
  )
  names(out)[names(out) %in% stat_cols] <- paste0(
    names(out)[names(out) %in% stat_cols],
    "_",
    label
  )
  out
}

#' Merge online and offline cell summaries.
#'
#' @param online_cells Online cell summary.
#' @param offline_cells Offline cell summary.
#' @param by_cell Cell columns.
#' @keywords internal
#' @noRd
.wf_blend_merge_cells <- function(online_cells, offline_cells, by_cell) {
  keys <- c("group", by_cell)
  merged <- merge(
    online_cells,
    offline_cells,
    by = keys,
    all = TRUE,
    sort = FALSE
  )
  numeric_zero <- c(
    "row_count_online", "weight_sum_online", "neff_online",
    "missing_outcome_online", "row_count_offline",
    "weight_sum_offline", "neff_offline", "missing_outcome_offline"
  )
  for (col in intersect(numeric_zero, names(merged))) {
    merged[[col]][is.na(merged[[col]])] <- 0
  }
  logical_false <- c("estimable_online", "estimable_offline")
  for (col in intersect(logical_false, names(merged))) {
    merged[[col]][is.na(merged[[col]])] <- FALSE
  }
  merged
}

#' Compute data-driven lambda from effective sample sizes.
#'
#' @param cells Merged source cells.
#' @keywords internal
#' @noRd
.wf_blend_lambda_neff <- function(cells) {
  denom <- cells$neff_online + cells$neff_offline
  ifelse(denom > 0, cells$neff_online / denom, NA_real_)
}

#' Compute inverse-variance lambda.
#'
#' @param cells Merged source cells or groups.
#' @keywords internal
#' @noRd
.wf_blend_lambda_inverse_variance <- function(cells) {
  vo <- cells$variance_online
  vf <- cells$variance_offline
  denom <- vo + vf
  lambda <- ifelse(is.finite(denom) & denom > 0, vf / denom, NA_real_)
  zero_zero <- is.finite(vo) & is.finite(vf) & vo == 0 & vf == 0
  if (any(zero_zero)) {
    lambda[zero_zero] <- .wf_blend_lambda_neff(cells[zero_zero, , drop = FALSE])
  }
  lambda
}

#' Apply lambda trimming and one-source overrides.
#'
#' @param cells Merged source cells.
#' @param lambda_raw Raw lambda values.
#' @param lambda_strategy Strategy name.
#' @param trim_lambda Trim bounds.
#' @keywords internal
#' @noRd
.wf_blend_finalize_lambda <- function(cells, lambda_raw, lambda_strategy, trim_lambda) {
  online_ok <- cells$estimable_online
  offline_ok <- cells$estimable_offline

  lambda <- lambda_raw
  reason <- rep(lambda_strategy, length(lambda))
  trimmed <- rep(FALSE, length(lambda))

  only_online <- online_ok & !offline_ok
  only_offline <- !online_ok & offline_ok
  no_source <- !online_ok & !offline_ok
  both <- online_ok & offline_ok

  lambda[only_online] <- 1
  reason[only_online] <- "online_only"
  lambda[only_offline] <- 0
  reason[only_offline] <- "offline_only"

  trim_hit <- both & is.finite(lambda) &
    (lambda < trim_lambda[[1]] | lambda > trim_lambda[[2]])
  lambda[trim_hit] <- pmin(
    pmax(lambda[trim_hit], trim_lambda[[1]]),
    trim_lambda[[2]]
  )
  reason[trim_hit] <- "trimmed"
  trimmed[trim_hit] <- TRUE

  if (any(no_source)) {
    wf_abort(
      "At least one fusion cell has no estimable source.",
      "wf_error_feasibility",
      list(rows = which(no_source))
    )
  }

  data.frame(
    lambda = lambda,
    lambda_reason = reason,
    lambda_trimmed = trimmed,
    stringsAsFactors = FALSE
  )
}

#' Build scalar fixed lambda values.
#'
#' @param lambda_fixed User-supplied fixed lambda.
#' @param n Number of rows.
#' @keywords internal
#' @noRd
.wf_blend_fixed_scalar <- function(lambda_fixed, n) {
  if (!is.numeric(lambda_fixed) || length(lambda_fixed) != 1 ||
    is.na(lambda_fixed) || !is.finite(lambda_fixed) ||
    lambda_fixed < 0 || lambda_fixed > 1) {
    wf_abort("Scalar `lambda_fixed` must be one finite value inside [0, 1].", "wf_error_input")
  }
  rep(as.numeric(lambda_fixed), n)
}

#' Resolve fixed lambda values for cells.
#'
#' @param lambda_fixed Scalar or data frame.
#' @param cells Merged source cells.
#' @param level Lambda level.
#' @param by_cell Cell columns.
#' @keywords internal
#' @noRd
.wf_blend_fixed_lambda <- function(lambda_fixed, cells, level, by_cell) {
  if (is.numeric(lambda_fixed) && length(lambda_fixed) == 1) {
    return(.wf_blend_fixed_scalar(lambda_fixed, nrow(cells)))
  }
  if (!is.data.frame(lambda_fixed)) {
    wf_abort("`lambda_fixed` must be a scalar or data frame.", "wf_error_input")
  }

  keys <- if (level == "group") "group" else c("group", by_cell)
  .require_cols(lambda_fixed, c(keys, "lambda"), "`lambda_fixed`")
  if (!is.numeric(lambda_fixed$lambda) || any(is.na(lambda_fixed$lambda)) ||
    any(!is.finite(lambda_fixed$lambda)) || any(lambda_fixed$lambda < 0) ||
    any(lambda_fixed$lambda > 1)) {
    wf_abort("`lambda_fixed$lambda` must contain finite values inside [0, 1].", "wf_error_input")
  }

  lambda_key <- .wf_cell_key(as.matrix(lambda_fixed[, keys, drop = FALSE]), keys)
  if (anyDuplicated(lambda_key)) {
    wf_abort("`lambda_fixed` contains duplicate keys.", "wf_error_input")
  }

  cell_key <- .wf_cell_key(as.matrix(cells[, keys, drop = FALSE]), keys)
  idx <- match(cell_key, lambda_key)
  if (any(is.na(idx))) {
    wf_abort(
      "`lambda_fixed` does not provide lambda values for every requested output cell.",
      "wf_error_input"
    )
  }
  lambda_fixed$lambda[idx]
}

#' Merge source group summaries.
#'
#' @param online_groups Online group summary.
#' @param offline_groups Offline group summary.
#' @keywords internal
#' @noRd
.wf_blend_merge_groups <- function(online_groups, offline_groups) {
  merged <- merge(
    online_groups,
    offline_groups,
    by = "group",
    all = TRUE,
    sort = FALSE
  )
  numeric_zero <- c(
    "row_count_online", "weight_sum_online", "neff_online",
    "missing_outcome_online", "row_count_offline",
    "weight_sum_offline", "neff_offline", "missing_outcome_offline"
  )
  for (col in intersect(numeric_zero, names(merged))) {
    merged[[col]][is.na(merged[[col]])] <- 0
  }
  logical_false <- c("estimable_online", "estimable_offline")
  for (col in intersect(logical_false, names(merged))) {
    merged[[col]][is.na(merged[[col]])] <- FALSE
  }
  merged
}

#' Build lambda at the requested level and merge it to cells.
#'
#' @param online Online source.
#' @param offline Offline source.
#' @param cells Cell-level source summary.
#' @param by_cell Cell columns.
#' @param outcome Optional outcome.
#' @param lambda Strategy.
#' @param lambda_fixed Fixed lambda.
#' @param level Lambda level.
#' @param trim_lambda Trim bounds.
#' @keywords internal
#' @noRd
.wf_blend_lambda_for_level <- function(online, offline, cells, by_cell, outcome,
                                       lambda, lambda_fixed, level, trim_lambda) {
  keys <- c("group", by_cell)
  if (level == "group" && lambda != "fixed") {
    online_groups <- .wf_blend_source_groups(online, "online", outcome)
    offline_groups <- .wf_blend_source_groups(offline, "offline", outcome)
    base <- .wf_blend_merge_groups(online_groups, offline_groups)
    lambda_raw <- switch(
      lambda,
      neff = .wf_blend_lambda_neff(base),
      inverse_variance = .wf_blend_lambda_inverse_variance(base)
    )
    lambda_info <- .wf_blend_finalize_lambda(base, lambda_raw, lambda, trim_lambda)
    lambda_table <- cbind(base["group"], lambda_info)
    idx <- match(cells$group, lambda_table$group)
    out <- cbind(cells[keys], lambda_table[idx, c("lambda", "lambda_reason", "lambda_trimmed")])
  } else {
    lambda_raw <- switch(
      lambda,
      neff = .wf_blend_lambda_neff(cells),
      inverse_variance = .wf_blend_lambda_inverse_variance(cells),
      fixed = .wf_blend_fixed_lambda(lambda_fixed, cells, level, by_cell)
    )
    lambda_info <- .wf_blend_finalize_lambda(cells, lambda_raw, lambda, trim_lambda)
    out <- cbind(cells[keys], lambda_info)
  }

  if (any(out$lambda_trimmed)) {
    wf_warn(
      sprintf("Trimmed %d data-driven lambda value(s).", sum(out$lambda_trimmed)),
      "wf_warning_quality",
      list(n = sum(out$lambda_trimmed))
    )
  }
  out
}

#' Warn for one-source fusion cells.
#'
#' @param cells Cell result table.
#' @keywords internal
#' @noRd
.wf_blend_warn_one_source <- function(cells) {
  n <- sum(cells$lambda_reason %in% c("online_only", "offline_only"))
  if (n > 0) {
    wf_warn(
      sprintf("%d fusion cell(s) are estimated from one source only.", n),
      "wf_warning_quality",
      list(n = n)
    )
  }
  invisible(n)
}

#' Build summary rows from fused cell estimates.
#'
#' @param cells Fused estimate cells.
#' @keywords internal
#' @noRd
.wf_blend_summary <- function(cells) {
  group_rows <- lapply(split(cells, cells$group), function(part) {
    valid <- is.finite(part$estimate) &
      is.finite(part$cell_weight) &
      part$cell_weight > 0
    estimate <- if (any(valid)) {
      sum(part$estimate[valid] * part$cell_weight[valid]) / sum(part$cell_weight[valid])
    } else {
      NA_real_
    }
    data.frame(
      group = part$group[[1]],
      estimate = estimate,
      cell_weight = sum(part$cell_weight[valid]),
      n_cells = nrow(part),
      stringsAsFactors = FALSE
    )
  })
  group_summary <- do.call(rbind, group_rows)
  rownames(group_summary) <- NULL

  valid <- is.finite(cells$estimate) &
    is.finite(cells$cell_weight) &
    cells$cell_weight > 0
  overall <- data.frame(
    group = "__overall__",
    estimate = if (any(valid)) {
      sum(cells$estimate[valid] * cells$cell_weight[valid]) / sum(cells$cell_weight[valid])
    } else {
      NA_real_
    },
    cell_weight = sum(cells$cell_weight[valid]),
    n_cells = nrow(cells),
    stringsAsFactors = FALSE
  )
  rbind(group_summary, overall)
}

#' Build global fixed-lambda sensitivity output.
#'
#' @param cells Fused estimate cells.
#' @keywords internal
#' @noRd
.wf_blend_sensitivity <- function(cells) {
  sweep_values <- seq(0.3, 0.9, by = 0.1)
  rows <- lapply(sweep_values, function(lambda_value) {
    tmp <- cells
    tmp$estimate <- lambda_value * tmp$estimate_online +
      (1 - lambda_value) * tmp$estimate_offline
    tmp$cell_weight <- lambda_value * tmp$weight_sum_online +
      (1 - lambda_value) * tmp$weight_sum_offline
    summary <- .wf_blend_summary(tmp)
    data.frame(
      lambda = lambda_value,
      summary,
      stringsAsFactors = FALSE
    )
  })
  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

#' Print a blend result.
#'
#' @param x A `wf_blend_result` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_blend_result <- function(x, ...) {
  n_cells <- if (!is.null(x$estimates) && nrow(x$estimates) > 0) {
    nrow(x$estimates)
  } else if (!is.null(x$cell_weights)) {
    nrow(x$cell_weights)
  } else {
    0L
  }
  lambda_range <- if (!is.null(x$lambda) && nrow(x$lambda) > 0) {
    sprintf("[%.4g, %.4g]", min(x$lambda$lambda), max(x$lambda$lambda))
  } else {
    "[NA, NA]"
  }
  warning_count <- sum(
    c(
      x$diagnostics$trimmed_lambda_count,
      x$diagnostics$one_source_cell_count
    ),
    na.rm = TRUE
  )
  cat(sprintf(
    "<wf_blend_result> %d cell(s); lambda %s; warnings: %d\n",
    n_cells,
    lambda_range,
    warning_count
  ))
  invisible(x)
}

#' Blend online and offline calibrated estimates
#'
#' Combines two `wf_weights` sources at the estimator level. Each source is
#' estimated within each fusion cell first; the source estimates are then
#' combined using the effective lambda recorded in the result.
#'
#' @param online Online-source `wf_weights`.
#' @param offline Offline-source `wf_weights`.
#' @param by_cell Character vector of cell columns.
#' @param lambda Lambda strategy: `"neff"`, `"inverse_variance"`, or `"fixed"`.
#' @param lambda_fixed Fixed lambda scalar or key table when `lambda = "fixed"`.
#' @param outcome Optional numeric outcome column.
#' @param level Lambda level: `"cell"` or `"group"`.
#' @param trim_lambda Two bounds used to clamp data-driven lambdas.
#' @param sensitivity Whether to compute a global-lambda sensitivity sweep.
#'
#' @return A `wf_blend_result` object.
#' @export
wf_blend <- function(online, offline, by_cell,
                     lambda = c("neff", "inverse_variance", "fixed"),
                     lambda_fixed = NULL,
                     outcome = NULL,
                     level = c("cell", "group"),
                     trim_lambda = c(0.05, 0.95),
                     sensitivity = TRUE) {
  t0 <- Sys.time()
  lambda <- .wf_blend_match(lambda, c("neff", "inverse_variance", "fixed"), "lambda")
  level <- .wf_blend_match(level, c("cell", "group"), "level")
  by_cell <- .wf_blend_by_cell(by_cell)
  trim_lambda <- .wf_blend_trim(trim_lambda)
  if (!is.null(outcome) && (length(outcome) != 1 || is.na(outcome) || !nzchar(outcome))) {
    wf_abort("`outcome` must be `NULL` or a single non-empty column name.", "wf_error_input")
  }
  sensitivity <- isTRUE(sensitivity)

  .wf_blend_check_source(online, "online", by_cell, outcome)
  .wf_blend_check_source(offline, "offline", by_cell, outcome)
  .wf_blend_check_fixed_required(lambda, lambda_fixed)

  online_cells <- .wf_blend_source_cells(online, "online", by_cell, outcome)
  offline_cells <- .wf_blend_source_cells(offline, "offline", by_cell, outcome)
  cells <- .wf_blend_merge_cells(online_cells, offline_cells, by_cell)

  lambda_table <- .wf_blend_lambda_for_level(
    online,
    offline,
    cells,
    by_cell,
    outcome,
    lambda,
    lambda_fixed,
    level,
    trim_lambda
  )
  cell_keys <- .wf_cell_key(as.matrix(cells[, c("group", by_cell), drop = FALSE]), c("group", by_cell))
  lambda_keys <- .wf_cell_key(as.matrix(lambda_table[, c("group", by_cell), drop = FALSE]), c("group", by_cell))
  cells <- cbind(cells, lambda_table[match(cell_keys, lambda_keys), c("lambda", "lambda_reason", "lambda_trimmed")])

  if (!is.null(outcome)) {
    cells$estimate <- cells$lambda * cells$estimate_online +
      (1 - cells$lambda) * cells$estimate_offline
    cells$variance <- (cells$lambda^2) * cells$variance_online +
      ((1 - cells$lambda)^2) * cells$variance_offline
    cells$cell_weight <- cells$lambda * cells$weight_sum_online +
      (1 - cells$lambda) * cells$weight_sum_offline
  }

  .wf_blend_warn_one_source(cells)

  if (!is.null(outcome)) {
    summary <- .wf_blend_summary(cells)
    sensitivity_out <- if (sensitivity) .wf_blend_sensitivity(cells) else NULL
    cell_weights <- NULL
  } else {
    cells$fused_cell_total <- cells$lambda * cells$weight_sum_online +
      (1 - cells$lambda) * cells$weight_sum_offline
    summary <- data.frame()
    sensitivity_out <- NULL
    cell_weights <- cells
  }

  out <- list(
    estimates = if (is.null(outcome)) data.frame() else cells,
    summary = summary,
    lambda = cells[c("group", by_cell, "lambda", "lambda_reason", "lambda_trimmed")],
    diagnostics = list(
      source_support = cells,
      trimmed_lambda_count = sum(cells$lambda_trimmed),
      one_source_cell_count = sum(cells$lambda_reason %in% c("online_only", "offline_only"))
    ),
    sensitivity = sensitivity_out,
    provenance = list(
      method = "blend",
      by_cell = by_cell,
      outcome = outcome,
      lambda = lambda,
      level = level,
      trim_lambda = trim_lambda,
      sources = list(
        online = online$provenance,
        offline = offline$provenance
      ),
      assumptions = c(
        "Convex fusion assumes both source estimates are approximately unbiased for the cell quantity.",
        "Online-source unbiasedness depends on calibration variables explaining the selection mechanism.",
        "Sensitivity output exposes dependence on lambda choices."
      ),
      created = t0,
      elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = .wf_blend_package_version()
    )
  )
  if (!is.null(cell_weights)) {
    out$cell_weights <- cell_weights
  }
  structure(out, class = "wf_blend_result")
}

make_blend_weights <- function(source = "online",
                               group = c("A", "A", "A", "A"),
                               cell = c("urban", "urban", "rural", "rural"),
                               weight = c(1, 1, 1, 1),
                               outcome = c(1, 0, 1, 0)) {
  data <- data.frame(
    id = paste0(source, "_", seq_along(group)),
    group = group,
    cell = cell,
    weight = weight,
    outcome = outcome,
    stringsAsFactors = FALSE
  )
  structure(
    list(
      data = data,
      log = data.frame(
        group = unique(group),
        iterations = NA_integer_,
        converged = TRUE,
        trimmed = 0L,
        stringsAsFactors = FALSE
      ),
      achieved = NULL,
      provenance = list(method = source, created = Sys.time())
    ),
    class = "wf_weights"
  )
}

test_that("wf_blend is exported", {
  expect_true("wf_blend" %in% getNamespaceExports("WFC"))
  expect_true(is.function(wf_blend))
})

test_that("wf_blend validates source objects and required columns", {
  online <- make_blend_weights("online")
  offline <- make_blend_weights("offline")

  expect_error(
    wf_blend(list(), offline, by_cell = "cell", outcome = "outcome"),
    class = "wf_error_input"
  )
  expect_error(
    wf_blend(online, list(), by_cell = "cell", outcome = "outcome"),
    class = "wf_error_input"
  )

  missing_group <- online
  missing_group$data$group <- NULL
  expect_error(
    wf_blend(missing_group, offline, by_cell = "cell", outcome = "outcome"),
    class = "wf_error_schema"
  )

  missing_cell <- offline
  missing_cell$data$cell <- NULL
  expect_error(
    wf_blend(online, missing_cell, by_cell = "cell", outcome = "outcome"),
    class = "wf_error_schema"
  )

  missing_weight <- offline
  missing_weight$data$weight <- NULL
  expect_error(
    wf_blend(online, missing_weight, by_cell = "cell", outcome = "outcome"),
    class = "wf_error_schema"
  )

  missing_outcome <- online
  missing_outcome$data$outcome <- NULL
  expect_error(
    wf_blend(missing_outcome, offline, by_cell = "cell", outcome = "outcome"),
    class = "wf_error_schema"
  )
})

test_that("wf_blend validates weights, outcome, lambda, level, and trim settings", {
  online <- make_blend_weights("online")
  offline <- make_blend_weights("offline")

  bad_weight <- online
  bad_weight$data$weight[[1]] <- -1
  expect_error(
    wf_blend(bad_weight, offline, by_cell = "cell", outcome = "outcome"),
    class = "wf_error_input"
  )

  bad_outcome <- online
  bad_outcome$data$outcome <- as.character(bad_outcome$data$outcome)
  expect_error(
    wf_blend(bad_outcome, offline, by_cell = "cell", outcome = "outcome"),
    class = "wf_error_input"
  )

  expect_error(
    wf_blend(online, offline, by_cell = "cell", outcome = "outcome", lambda = "median"),
    class = "wf_error_input"
  )
  expect_error(
    wf_blend(online, offline, by_cell = "cell", outcome = "outcome", level = "province"),
    class = "wf_error_input"
  )
  expect_error(
    wf_blend(
      online,
      offline,
      by_cell = "cell",
      outcome = "outcome",
      trim_lambda = c(0.9, 0.1)
    ),
    class = "wf_error_input"
  )
  expect_error(
    wf_blend(online, offline, by_cell = "cell", outcome = "outcome", lambda = "fixed"),
    class = "wf_error_input"
  )
})

test_that("wf_blend computes source estimates before fusing", {
  online <- make_blend_weights(
    "online",
    group = c("A", "A"),
    cell = c("urban", "urban"),
    weight = c(100, 100),
    outcome = c(1, 0)
  )
  offline <- make_blend_weights(
    "offline",
    group = c("A", "A"),
    cell = c("urban", "urban"),
    weight = c(1, 1),
    outcome = c(0, 0)
  )

  out <- wf_blend(
    online,
    offline,
    by_cell = "cell",
    outcome = "outcome",
    lambda = "fixed",
    lambda_fixed = 0.5,
    sensitivity = FALSE
  )

  stacked_weight <- c(online$data$weight, offline$data$weight)
  stacked_outcome <- c(online$data$outcome, offline$data$outcome)
  stacked <- sum(stacked_weight * stacked_outcome) / sum(stacked_weight)

  expect_s3_class(out, "wf_blend_result")
  expect_equal(out$estimates$estimate_online, 0.5, tolerance = 1e-10)
  expect_equal(out$estimates$estimate_offline, 0, tolerance = 1e-10)
  expect_equal(out$estimates$lambda, 0.5, tolerance = 1e-10)
  expect_equal(out$estimates$estimate, 0.25, tolerance = 1e-10)
  expect_false(isTRUE(all.equal(out$estimates$estimate, stacked)))
})

test_that("wf_blend uses effective sample size for neff lambda", {
  online <- make_blend_weights(
    "online",
    group = c("A", "A", "A"),
    cell = c("urban", "urban", "urban"),
    weight = c(1, 1, 1),
    outcome = c(1, 1, 0)
  )
  offline <- make_blend_weights(
    "offline",
    group = c("A", "A"),
    cell = c("urban", "urban"),
    weight = c(2, 2),
    outcome = c(0, 0)
  )

  out <- wf_blend(
    online,
    offline,
    by_cell = "cell",
    outcome = "outcome",
    lambda = "neff",
    trim_lambda = c(0, 1),
    sensitivity = FALSE
  )

  expect_equal(out$estimates$neff_online, 3, tolerance = 1e-10)
  expect_equal(out$estimates$neff_offline, 2, tolerance = 1e-10)
  expect_equal(out$estimates$lambda, 3 / 5, tolerance = 1e-10)
  expect_equal(out$estimates$estimate, (3 / 5) * (2 / 3), tolerance = 1e-10)
})

test_that("wf_blend uses inverse-variance lambda", {
  online <- make_blend_weights(
    "online",
    group = c("A", "A"),
    cell = c("urban", "urban"),
    weight = c(1, 1),
    outcome = c(1, 0)
  )
  offline <- make_blend_weights(
    "offline",
    group = c("A", "A"),
    cell = c("urban", "urban"),
    weight = c(1, 1),
    outcome = c(1, 1)
  )

  out <- wf_blend(
    online,
    offline,
    by_cell = "cell",
    outcome = "outcome",
    lambda = "inverse_variance",
    trim_lambda = c(0, 1),
    sensitivity = FALSE
  )

  expect_equal(out$estimates$variance_online, 0.125, tolerance = 1e-10)
  expect_equal(out$estimates$variance_offline, 0, tolerance = 1e-10)
  expect_equal(out$estimates$lambda, 0, tolerance = 1e-10)
  expect_equal(out$estimates$estimate, 1, tolerance = 1e-10)
})

test_that("wf_blend applies fixed lambda tables by cell and group", {
  online <- make_blend_weights(
    "online",
    group = c("A", "A", "B", "B"),
    cell = c("urban", "rural", "urban", "rural"),
    outcome = c(1, 1, 0, 0)
  )
  offline <- make_blend_weights(
    "offline",
    group = c("A", "A", "B", "B"),
    cell = c("urban", "rural", "urban", "rural"),
    outcome = c(0, 0, 1, 1)
  )

  by_cell_lambda <- data.frame(
    group = c("A", "A", "B", "B"),
    cell = c("urban", "rural", "urban", "rural"),
    lambda = c(0.2, 0.4, 0.6, 0.8),
    stringsAsFactors = FALSE
  )
  out_cell <- wf_blend(
    online,
    offline,
    by_cell = "cell",
    outcome = "outcome",
    lambda = "fixed",
    lambda_fixed = by_cell_lambda,
    level = "cell",
    sensitivity = FALSE
  )
  out_cell <- out_cell$estimates[order(out_cell$estimates$group, out_cell$estimates$cell), ]

  expect_equal(out_cell$lambda, c(0.4, 0.2, 0.8, 0.6), tolerance = 1e-10)

  by_group_lambda <- data.frame(
    group = c("A", "B"),
    lambda = c(0.25, 0.75),
    stringsAsFactors = FALSE
  )
  out_group <- wf_blend(
    online,
    offline,
    by_cell = "cell",
    outcome = "outcome",
    lambda = "fixed",
    lambda_fixed = by_group_lambda,
    level = "group",
    sensitivity = FALSE
  )
  out_group <- out_group$estimates[order(out_group$estimates$group, out_group$estimates$cell), ]

  expect_equal(out_group$lambda, c(0.25, 0.25, 0.75, 0.75), tolerance = 1e-10)
})

test_that("wf_blend computes group-level data-driven lambda and trims two-source cells", {
  online <- make_blend_weights(
    "online",
    group = c("A", "A", "A", "B"),
    cell = c("urban", "urban", "rural", "urban"),
    weight = c(1, 1, 1, 100),
    outcome = c(1, 1, 1, 1)
  )
  offline <- make_blend_weights(
    "offline",
    group = c("A", "A", rep("B", 10)),
    cell = c("urban", "rural", rep("urban", 10)),
    weight = rep(1, 12),
    outcome = rep(0, 12)
  )

  expect_warning(
    out <- wf_blend(
      online,
      offline,
      by_cell = "cell",
      outcome = "outcome",
      lambda = "neff",
      level = "group",
      trim_lambda = c(0.2, 0.8),
      sensitivity = FALSE
    ),
    class = "wf_warning_quality"
  )

  expect_equal(out$estimates$lambda[out$estimates$group == "A"], c(0.6, 0.6), tolerance = 1e-10)
  expect_equal(out$estimates$lambda[out$estimates$group == "B"], 0.2, tolerance = 1e-10)
  expect_true(any(out$estimates$lambda_trimmed))
})

test_that("wf_blend handles one-source cells and rejects no-source outcome cells", {
  online <- make_blend_weights(
    "online",
    group = c("A", "A"),
    cell = c("online_only", "shared"),
    weight = c(1, 1),
    outcome = c(1, 1)
  )
  offline <- make_blend_weights(
    "offline",
    group = c("A", "A"),
    cell = c("offline_only", "shared"),
    weight = c(1, 1),
    outcome = c(0, 0)
  )

  expect_warning(
    out <- wf_blend(
      online,
      offline,
      by_cell = "cell",
      outcome = "outcome",
      lambda = "neff",
      sensitivity = FALSE
    ),
    class = "wf_warning_quality"
  )

  one_source <- out$estimates[out$estimates$cell %in% c("online_only", "offline_only"), ]
  expect_equal(one_source$lambda[one_source$cell == "online_only"], 1, tolerance = 1e-10)
  expect_equal(one_source$lambda[one_source$cell == "offline_only"], 0, tolerance = 1e-10)
  expect_equal(out$diagnostics$one_source_cell_count, 2)

  bad_online <- online
  bad_offline <- offline
  bad_online$data$outcome[bad_online$data$cell == "shared"] <- NA_real_
  bad_offline$data$outcome[bad_offline$data$cell == "shared"] <- NA_real_
  expect_error(
    suppressWarnings(
      wf_blend(bad_online, bad_offline, by_cell = "cell", outcome = "outcome", sensitivity = FALSE)
    ),
    class = "wf_error_feasibility"
  )
})

test_that("wf_blend outcome-null mode returns a cell ledger instead of wf_weights", {
  online <- make_blend_weights("online")
  offline <- make_blend_weights("offline")

  out <- wf_blend(
    online,
    offline,
    by_cell = "cell",
    outcome = NULL,
    lambda = "fixed",
    lambda_fixed = 0.7,
    sensitivity = FALSE
  )

  expect_s3_class(out, "wf_blend_result")
  expect_false(inherits(out, "wf_weights"))
  expect_equal(nrow(out$estimates), 0)
  expect_true(all(c("group", "cell", "lambda", "fused_cell_total") %in% names(out$cell_weights)))
  expect_equal(out$cell_weights$lambda, rep(0.7, nrow(out$cell_weights)), tolerance = 1e-10)
})

test_that("wf_blend builds summaries, sensitivity output, print output, and provenance", {
  online <- make_blend_weights(
    "online",
    group = c("A", "A", "B", "B"),
    cell = c("urban", "rural", "urban", "rural"),
    weight = c(1, 1, 1, 1),
    outcome = c(1, 1, 0, 0)
  )
  offline <- make_blend_weights(
    "offline",
    group = c("A", "A", "B", "B"),
    cell = c("urban", "rural", "urban", "rural"),
    weight = c(1, 1, 1, 1),
    outcome = c(0, 0, 1, 1)
  )

  out <- wf_blend(
    online,
    offline,
    by_cell = "cell",
    outcome = "outcome",
    lambda = "fixed",
    lambda_fixed = 0.25,
    sensitivity = TRUE
  )

  expect_true(all(c("group", "estimate", "cell_weight") %in% names(out$summary)))
  expect_true(all(c("lambda", "group", "estimate") %in% names(out$sensitivity)))
  expect_equal(sort(unique(out$sensitivity$lambda)), seq(0.3, 0.9, by = 0.1), tolerance = 1e-10)
  expect_equal(out$provenance$method, "blend")
  expect_equal(out$provenance$sources$online$method, "online")
  expect_equal(out$provenance$sources$offline$method, "offline")
  expect_true(length(out$provenance$assumptions) >= 3)
  expect_output(print(out), "<wf_blend_result>")
  expect_output(print(out), "lambda")
})

#' Confirm one guided workflow decision.
#'
#' @param prompt Localized prompt text.
#' @keywords internal
#' @noRd
.wf_autoweigh_confirm <- function(prompt) {
  answer <- tolower(trimws(readline(prompt)))
  answer %in% c("y", "yes")
}

#' Run a guided workflow from verified design and target objects
#'
#' Provides the guided WFC experience while enforcing the same WFC 2.0 input
#' boundary used by direct calibration and approved-plan execution.
#'
#' @param design An unchanged `wf_design_data`.
#' @param target An unchanged, non-demo `wf_verified_target`.
#' @param dims A `wf_dims` declaration matching `target`.
#' @param method Calibration method. `"auto"` selects between supported safe
#'   engines from declared settings only.
#' @param ladder Reviewed collapse ladder for post-stratification.
#' @param min_cell Positive post-stratification cell-size threshold.
#' @param bounds Two-element bounds for logit calibration.
#' @param trim Raking trim control.
#' @param max_deff Maximum design effect accepted by automatic trimming.
#' @param max_residual Maximum margin residual accepted by automatic trimming.
#' @param interactive Whether review prompts may be shown.
#' @param lang Output language.
#' @param ... Method settings. ID and base-weight roles cannot be overridden.
#'
#' @return A `wf_autoweigh_result` carrying verified input identities.
#' @export
wf_autoweigh <- function(design, target, dims,
                         method = c("auto", "raking", "poststrat", "logit"),
                         ladder = NULL, min_cell = NULL,
                         bounds = c(0.3, 3), trim = "auto",
                         max_deff = 6, max_residual = 0.02,
                         interactive = base::interactive(),
                         lang = NULL, ...) {
  settings <- c(
    list(
      dims = dims,
      method = method,
      ladder = ladder,
      min_cell = min_cell,
      bounds = bounds,
      trim = trim,
      max_deff = max_deff,
      max_residual = max_residual,
      interactive = interactive,
      lang = lang
    ),
    list(...)
  )
  .wf_execute_verified_engine(design, target, "autoweigh", settings)
}

#' Run a guided and auditable weighting workflow
#'
#' Builds or accepts a canonical target, enforces the precheck discipline,
#' applies only declared collapse remediations, optionally recommends and
#' applies raking trim bounds, then returns weights, diagnostics, a manager
#' report, and a machine-readable decision ledger.
#'
#' @param sample Non-empty sample data frame.
#' @param population A raw population data frame or ready `wf_target`.
#' @param dims A `wf_dims` declaration.
#' @param key_map Named population-column mapping required for raw population
#'   data.
#' @param count Population count column required for raw population data.
#' @param by Optional grouping variable used when building a raw target.
#' @param id Optional unique unit identifier column.
#' @param method Calibration method. `"auto"` selects post-stratification only
#'   when both `ladder` and `min_cell` are supplied; otherwise it selects
#'   raking.
#' @param ladder Reviewed `wf_collapse_ladder`, or a `wf_ladder_draft` whose
#'   validated `$ladder` should be used for post-stratification.
#' @param min_cell Positive post-stratification cell-size threshold.
#' @param bounds Two-element bounds for logit calibration.
#' @param trim Raking trim control: `"auto"`, `NULL`, or two numeric bounds.
#' @param max_deff Maximum design effect accepted by automatic trimming.
#' @param max_residual Maximum margin residual accepted by automatic trimming.
#' @param interactive Whether declared collapse and finite trim recommendations
#'   require confirmation.
#' @param lang Output language for narration and the manager report.
#' @param ... Method controls. Raking accepts `caps`, `lo`, `na`,
#'   `trim_cycles`, `tol`, `max_iter`, and `init_weight`; logit accepts `na`,
#'   `tol`, `max_iter`, and `init_weight`; post-stratification accepts
#'   `granularity`, `empty_cell`, `tol`, and `init_weight`.
#'
#' @return A `wf_autoweigh_result` containing weights, diagnostics, report,
#'   ordered ledger, aligned artifacts, final sample and target, selected
#'   method, and normalized language.
#' @keywords internal
#' @noRd
.wf_autoweigh_engine <- function(sample, population, dims,
                                 key_map = NULL, count = NULL,
                                 by = NULL, id = NULL,
                                 method = c("auto", "raking", "poststrat", "logit"),
                                 ladder = NULL, min_cell = NULL,
                                 bounds = c(0.3, 3),
                                 trim = "auto",
                                 max_deff = 6, max_residual = 0.02,
                                 interactive = base::interactive(),
                                 lang = NULL, ...) {
  method <- tryCatch(
    match.arg(method),
    error = function(e) {
      wf_abort(
        "`method` must be one of auto, raking, poststrat, or logit.",
        "wf_error_input"
      )
    }
  )
  language <- .wf_lang(lang)

  if (!is.data.frame(sample) || nrow(sample) == 0) {
    wf_abort("`sample` must be a non-empty data frame.", "wf_error_input")
  }
  if (!inherits(dims, "wf_dims")) {
    wf_abort("`dims` must be a wf_dims object.", "wf_error_input")
  }
  if (length(interactive) != 1 || !is.logical(interactive) ||
      is.na(interactive)) {
    wf_abort("`interactive` must be TRUE or FALSE.", "wf_error_input")
  }
  if (length(max_deff) != 1 || !is.numeric(max_deff) ||
      !is.finite(max_deff) || max_deff <= 0) {
    wf_abort("`max_deff` must be a positive finite number.", "wf_error_input")
  }
  if (length(max_residual) != 1 || !is.numeric(max_residual) ||
      !is.finite(max_residual) || max_residual < 0) {
    wf_abort(
      "`max_residual` must be a non-negative finite number.",
      "wf_error_input"
    )
  }

  dots <- list(...)
  if (length(dots) > 0 &&
      (is.null(names(dots)) || any(!nzchar(names(dots))))) {
    wf_abort("All arguments in `...` must be named.", "wf_error_input")
  }
  if ("precheck" %in% names(dots)) {
    wf_abort(
      "`precheck` is owned by wf_autoweigh() and cannot be overridden.",
      "wf_error_input"
    )
  }

  selected_method <- if (method == "auto") {
    if (!is.null(ladder) && !is.null(min_cell)) "poststrat" else "raking"
  } else {
    method
  }
  allowed_dots <- switch(
    selected_method,
    raking = c(
      "caps", "lo", "na", "trim_cycles", "tol", "max_iter",
      "init_weight"
    ),
    logit = c("na", "tol", "max_iter", "init_weight"),
    poststrat = c("granularity", "empty_cell", "tol", "init_weight")
  )
  unknown_dots <- setdiff(names(dots), allowed_dots)
  if (length(unknown_dots) > 0) {
    wf_abort(
      sprintf(
        "Unsupported guided-workflow argument(s) for %s: %s.",
        selected_method,
        paste(unknown_dots, collapse = ", ")
      ),
      "wf_error_input",
      list(arguments = unknown_dots, method = selected_method)
    )
  }

  trim_is_auto <- is.character(trim) && length(trim) == 1 &&
    !is.na(trim) && identical(trim, "auto")
  trim_is_numeric <- is.numeric(trim) && length(trim) == 2 &&
    !anyNA(trim) && all(is.finite(trim)) && all(trim > 0) &&
    trim[[1]] < trim[[2]]
  if (!is.null(trim) && !trim_is_auto && !trim_is_numeric) {
    wf_abort(
      "`trim` must be 'auto', NULL, or two increasing positive bounds.",
      "wf_error_input"
    )
  }
  if (trim_is_numeric && selected_method != "raking") {
    wf_abort("Numeric `trim` bounds apply only to raking.", "wf_error_input")
  }

  target_is_ready <- inherits(population, "wf_target")
  if (!target_is_ready) {
    if (!is.data.frame(population)) {
      wf_abort(
        "`population` must be a data frame or wf_target object.",
        "wf_error_input"
      )
    }
    if (is.null(key_map) || is.null(count)) {
      wf_abort(
        "Raw population data requires both `key_map` and `count`.",
        "wf_error_schema"
      )
    }
  }
  target <- if (target_is_ready) {
    population
  } else {
    wf_target_population(
      population,
      key_map = key_map,
      count = count,
      dims = dims,
      by = by,
      keep_joint = selected_method == "poststrat"
    )
  }
  if (!identical(target$dims, names(dims$vars))) {
    wf_abort(
      "`dims` must declare the target dimensions in the same order.",
      "wf_error_schema",
      list(target_dims = target$dims, dims = names(dims$vars))
    )
  }
  .require_cols(sample, c(id, target$by, target$dims), "sample")

  if (inherits(ladder, "wf_ladder_draft")) {
    ladder <- ladder$ladder
  }
  if (selected_method == "poststrat") {
    valid_min_cell <- length(min_cell) == 1 && is.numeric(min_cell) &&
      is.finite(min_cell) && min_cell >= 1 &&
      min_cell == as.integer(min_cell)
    if (!inherits(ladder, "wf_collapse_ladder") || !valid_min_cell) {
      wf_abort(
        paste(
          "Post-stratification requires a reviewed wf_collapse_ladder and",
          "a positive integer `min_cell`."
        ),
        "wf_error_feasibility"
      )
    }
    if (is.null(target$joint) || length(target$joint) == 0) {
      wf_abort(
        paste(
          "Post-stratification requires a target built with",
          "keep_joint = TRUE."
        ),
        "wf_error_feasibility"
      )
    }
  }

  ledger_rows <- list()
  artifacts <- list()
  note <- function(action, detail_key, ..., artifact = NULL) {
    detail <- .wf_tr(detail_key, ..., lang = language)
    artifact_class <- if (is.null(artifact)) {
      NA_character_
    } else {
      class(artifact)[[1]]
    }
    ledger_rows[[length(ledger_rows) + 1L]] <<- data.frame(
      step = length(ledger_rows) + 1L,
      action = action,
      detail_key = detail_key,
      detail = detail,
      artifact_class = artifact_class,
      time = Sys.time(),
      stringsAsFactors = FALSE
    )
    artifacts <<- c(artifacts, list(artifact))
    message(detail)
    invisible(detail)
  }

  note(
    "start",
    "autoweigh_start",
    nrow(sample),
    length(target$groups),
    length(target$dims)
  )
  note(
    "target",
    if (target_is_ready) "autoweigh_target_ready" else "autoweigh_target_built",
    artifact = target
  )
  note("calibrate", "autoweigh_method", selected_method)

  current_sample <- sample
  current_target <- target
  eligible_collapse_checks <- c(
    "cat_infeasible", "support_thin", "risk_extreme_ratio"
  )
  precheck_na <- if (!is.null(dots$na)) {
    dots$na
  } else if (selected_method == "raking") {
    "fractional"
  } else if (selected_method == "logit") {
    "drop"
  } else {
    "error"
  }
  collapse_round <- 0L
  max_collapse_rounds <- max(
    1L,
    sum(vapply(dims$collapse, length, integer(1))) + 1L
  )

  repeat {
    precheck <- wf_precheck(
      current_sample,
      current_target,
      id = id,
      na = precheck_na
    )
    if (precheck$ok) {
      note(
        "precheck",
        "autoweigh_precheck_ok",
        artifact = precheck
      )
      break
    }

    blocking <- precheck$issues[
      precheck$issues$severity == "error",
      ,
      drop = FALSE
    ]
    note(
      "precheck",
      "autoweigh_precheck_blocked",
      nrow(blocking),
      artifact = precheck
    )
    plan <- wf_suggest_collapse(precheck, dims)
    unresolved_blocking <- plan$unresolved[
      plan$unresolved$severity == "error",
      ,
      drop = FALSE
    ]
    nonremediable <- !blocking$check %in% eligible_collapse_checks
    if (length(plan$maps) == 0 || any(nonremediable) ||
        nrow(unresolved_blocking) > 0 ||
        collapse_round >= max_collapse_rounds) {
      wf_abort(
        paste(
          "Precheck has blocking issues with no applicable declared collapse",
          "remediation. Update wf_dims() or the input data and retry."
        ),
        "wf_error_feasibility",
        list(precheck = precheck, plan = plan)
      )
    }
    if (interactive) {
      print(plan)
      if (!.wf_autoweigh_confirm(.wf_tr(
        "autoweigh_prompt_collapse",
        lang = language
      ))) {
        wf_abort(
          "Declared collapse remediation was declined; weighting stopped.",
          "wf_error_feasibility",
          list(precheck = precheck, plan = plan)
        )
      }
    }

    applied <- wf_apply_collapse(current_sample, current_target, plan)
    current_sample <- applied$sample
    current_target <- applied$target
    collapse_round <- collapse_round + 1L
    for (i in seq_len(nrow(plan$actions))) {
      note(
        "collapse",
        "autoweigh_collapse",
        plan$actions$dim[[i]],
        plan$actions$check[[i]],
        plan$actions$group[[i]],
        artifact = plan
      )
    }
  }

  trim_arg <- NULL
  if (selected_method == "raking" && trim_is_auto) {
    trim_search_args <- c(
      list(
        sample = current_sample,
        target = current_target,
        id = id,
        max_deff = max_deff,
        max_residual = max_residual,
        precheck = FALSE
      ),
      dots
    )
    trim_search <- do.call(.wf_auto_trim_engine, trim_search_args)
    recommendation <- trim_search$recommended_cap
    if (is.na(recommendation)) {
      note(
        "trim",
        "autoweigh_trim_no_solution",
        artifact = trim_search
      )
    } else if (!is.finite(recommendation)) {
      note(
        "trim",
        "autoweigh_trim_none",
        artifact = trim_search
      )
    } else {
      if (interactive && !.wf_autoweigh_confirm(.wf_tr(
        "autoweigh_prompt_trim",
        lang = language
      ))) {
        wf_abort(
          "Recommended trim was declined; weighting stopped.",
          "wf_error_feasibility",
          list(auto_trim = trim_search)
        )
      }
      trim_arg <- c(trim_search$criteria$lo, recommendation)
      note(
        "trim",
        "autoweigh_trim_applied",
        recommendation,
        artifact = trim_search
      )
    }
  } else if (selected_method == "raking" && trim_is_numeric) {
    trim_arg <- trim
    note("trim", "autoweigh_trim_manual", trim[[1]], trim[[2]])
  }

  note("calibrate", "autoweigh_calibrating", selected_method)
  engine_dots <- dots[!names(dots) %in% c("caps", "lo")]
  engine_args <- c(
    list(
      sample = current_sample,
      target = current_target,
      method = selected_method,
      id = id
    ),
    switch(
      selected_method,
      raking = list(trim = trim_arg, precheck = FALSE),
      logit = list(bounds = bounds, precheck = FALSE),
      poststrat = list(
        min_cell = as.integer(min_cell),
        ladder = ladder,
        precheck = FALSE
      )
    ),
    engine_dots
  )
  weights <- do.call(.wf_calibrate_engine, engine_args)
  diagnostics <- wf_diagnose(weights, target = current_target)
  report <- wf_report(
    weights,
    target = current_target,
    audience = "manager",
    lang = language
  )

  overall_light <- if (any(report$table$light == "red")) {
    "red"
  } else if (any(report$table$light == "yellow")) {
    "yellow"
  } else {
    "green"
  }
  overall_verdict <- .wf_tr(
    paste0("verdict_", overall_light),
    lang = language
  )
  note(
    "done",
    "autoweigh_done",
    nrow(weights$data),
    overall_verdict,
    artifact = report
  )

  ledger <- do.call(rbind, ledger_rows)
  names(artifacts) <- sprintf(
    "step_%03d_%s",
    ledger$step,
    ledger$action
  )
  structure(
    list(
      weights = weights,
      diagnostics = diagnostics,
      report = report,
      ledger = ledger,
      artifacts = artifacts,
      sample = current_sample,
      target = current_target,
      method = selected_method,
      language = language
    ),
    class = "wf_autoweigh_result"
  )
}

#' Print a guided weighting result
#'
#' @param x A `wf_autoweigh_result` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_autoweigh_result <- function(x, ...) {
  cat(sprintf(
    "<wf_autoweigh_result> %s\n",
    .wf_tr("autoweigh_print_header", lang = x$language)
  ))
  cat(sprintf(
    "  %d weight(s); method: %s; %d ledger step(s)\n",
    nrow(x$weights$data),
    x$method,
    nrow(x$ledger)
  ))
  for (i in seq_len(nrow(x$ledger))) {
    cat(sprintf(
      "  [%d] %-9s %s\n",
      x$ledger$step[[i]],
      x$ledger$action[[i]],
      x$ledger$detail[[i]]
    ))
  }
  invisible(x)
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----conversion, eval=FALSE---------------------------------------------------
# analysis_ready <- wf_attach_weights(
#   analysis_data,
#   locked,
#   id = "person_id",
#   weight_name = ".weight"
# )
# 
# design_survey <- as_svydesign(
#   locked,
#   analysis_ready,
#   id = "person_id"
# )

#' Cached localized output catalogs.
#'
#' @keywords internal
#' @noRd
.wf_i18n_cache <- new.env(parent = emptyenv())

#' Normalize an explicit or configured language value.
#'
#' @param value Language value.
#' @param strict Whether unsupported values should raise an input error.
#' @keywords internal
#' @noRd
.wf_normalize_lang <- function(value, strict = TRUE) {
  if (length(value) != 1 || is.na(value) || !is.character(value) ||
      !nzchar(trimws(value))) {
    if (strict) {
      wf_abort("`lang` must be one supported language name.", "wf_error_input")
    }
    return("en")
  }
  normalized <- tolower(gsub("-", "_", trimws(value), fixed = TRUE))
  if (grepl("^en($|_)", normalized)) {
    return("en")
  }
  if (grepl("^zh($|_)", normalized)) {
    return("zh_CN")
  }
  if (strict) {
    wf_abort(
      sprintf("Unsupported output language '%s'. Supported languages: en, zh_CN.", value),
      "wf_error_input",
      list(lang = value)
    )
  }
  "en"
}

#' Resolve the current output language.
#'
#' @param lang Optional explicit language.
#' @keywords internal
#' @noRd
.wf_lang <- function(lang = NULL) {
  if (!is.null(lang)) {
    return(.wf_normalize_lang(lang, strict = TRUE))
  }
  option <- getOption("wfc.lang", NULL)
  if (!is.null(option)) {
    return(.wf_normalize_lang(option, strict = TRUE))
  }
  locale <- tolower(paste(
    Sys.getenv("LANGUAGE"),
    Sys.getenv("LC_ALL"),
    Sys.getenv("LC_MESSAGES"),
    Sys.getenv("LANG"),
    Sys.getlocale("LC_CTYPE")
  ))
  if (grepl("zh", locale, fixed = TRUE)) "zh_CN" else "en"
}

#' Locate one installed or development output catalog.
#'
#' @param lang Normalized language.
#' @keywords internal
#' @noRd
.wf_i18n_path <- function(lang) {
  filename <- paste0(lang, ".dcf")
  path <- system.file("i18n", filename, package = "WFC")
  if (!nzchar(path)) {
    development <- file.path("inst", "i18n", filename)
    if (file.exists(development)) path <- development
  }
  path
}

#' Load a localized output catalog.
#'
#' @param lang Normalized language.
#' @keywords internal
#' @noRd
.wf_i18n_catalog <- function(lang) {
  lang <- .wf_normalize_lang(lang, strict = TRUE)
  if (exists(lang, envir = .wf_i18n_cache, inherits = FALSE)) {
    return(get(lang, envir = .wf_i18n_cache, inherits = FALSE))
  }
  path <- .wf_i18n_path(lang)
  if (!nzchar(path) || !file.exists(path)) {
    wf_abort(
      sprintf("Output catalog '%s' is missing from the installed package.", lang),
      "wf_error_internal",
      list(lang = lang)
    )
  }
  raw <- tryCatch(read.dcf(path), error = function(e) e)
  if (inherits(raw, "error") || nrow(raw) != 1) {
    wf_abort(
      sprintf("Output catalog '%s' is malformed.", lang),
      "wf_error_internal",
      list(lang = lang)
    )
  }
  catalog <- as.character(raw[1, ])
  names(catalog) <- colnames(raw)
  assign(lang, catalog, envir = .wf_i18n_cache)
  catalog
}

#' Translate one human-facing output string.
#'
#' @param key Stable catalog key.
#' @param ... Values interpolated with `sprintf()`.
#' @param lang Optional language selector.
#' @keywords internal
#' @noRd
.wf_tr <- function(key, ..., lang = NULL) {
  language <- .wf_lang(lang)
  catalog <- .wf_i18n_catalog(language)
  english <- .wf_i18n_catalog("en")
  template <- unname(catalog[key])
  if (length(template) == 0 || is.na(template)) {
    template <- unname(english[key])
  }
  if (length(template) == 0 || is.na(template)) {
    wf_abort(
      sprintf("Unknown output translation key '%s'.", key),
      "wf_error_internal",
      list(key = key)
    )
  }
  sprintf(template, ...)
}

#' Localize a known report column label.
#'
#' @param name Stable column name.
#' @param lang Optional language selector.
#' @keywords internal
#' @noRd
.wf_i18n_label <- function(name, lang = NULL) {
  language <- .wf_lang(lang)
  key <- paste0("column_", name)
  catalog <- .wf_i18n_catalog(language)
  english <- .wf_i18n_catalog("en")
  if (key %in% union(names(catalog), names(english))) {
    return(.wf_tr(key, lang = language))
  }
  gsub("_", " ", name, fixed = TRUE)
}

#' Localize a known report section label.
#'
#' @param name Stable section name.
#' @param lang Optional language selector.
#' @keywords internal
#' @noRd
.wf_i18n_section <- function(name, lang = NULL) {
  language <- .wf_lang(lang)
  key <- paste0("section_", name)
  catalog <- .wf_i18n_catalog(language)
  english <- .wf_i18n_catalog("en")
  if (key %in% union(names(catalog), names(english))) {
    return(.wf_tr(key, lang = language))
  }
  gsub("_", " ", name, fixed = TRUE)
}

make_report_weights <- function() {
  fixture <- make_weightflow_fixture()
  list(
    weights = .wf_rake_engine(fixture$sample, fixture$target, id = "id", tol = 1e-8),
    target = fixture$target
  )
}

make_report_propensity <- function() {
  structure(
    list(
      data = data.frame(
        id = paste0("p", 1:4),
        group = ".all",
        weight = c(0.7, 0.9, 1.1, 1.3),
        feature = 1 / c(0.7, 0.9, 1.1, 1.3)
      ),
      log = data.frame(group = ".all", n = 4),
      achieved = NULL,
      overlap = list(
        threshold = 0.99,
        online = c(`0%` = 0.2, `50%` = 0.6, `100%` = 0.9),
        reference = c(`0%` = 0.1, `50%` = 0.4, `100%` = 0.8),
        n_boundary = 0L,
        n_online = 4L
      ),
      balance = data.frame(
        variable = "age",
        level = NA_character_,
        smd_unweighted = 0.4,
        smd_weighted = 0.1
      ),
      provenance = list(method = "propensity")
    ),
    class = c("wf_propensity_weights", "wf_weights")
  )
}

make_report_blend <- function() {
  structure(
    list(
      summary = data.frame(
        group = c("A", "__overall__"),
        estimate = c(0.55, 0.55),
        cell_weight = c(100, 100),
        n_cells = c(2L, 2L)
      ),
      lambda = data.frame(
        group = c("A", "A"),
        cell = c("urban", "rural"),
        lambda = c(0.6, 0.7),
        lambda_reason = c("neff", "neff"),
        lambda_trimmed = c(FALSE, TRUE)
      ),
      diagnostics = list(
        source_support = data.frame(
          group = c("A", "A"),
          cell = c("urban", "rural"),
          lambda_reason = c("neff", "neff")
        ),
        trimmed_lambda_count = 1L,
        one_source_cell_count = 0L
      ),
      sensitivity = data.frame(
        lambda = c(0.3, 0.4),
        group = c("__overall__", "__overall__"),
        estimate = c(0.50, 0.52),
        cell_weight = c(100, 100),
        n_cells = c(2L, 2L)
      ),
      provenance = list(method = "blend")
    ),
    class = "wf_blend_result"
  )
}

test_that("wf_report builds manager and analyst projections for weights", {
  fixture <- make_report_weights()

  manager <- wf_report(
    fixture$weights,
    target = fixture$target,
    audience = "manager"
  )
  analyst <- wf_report(
    fixture$weights,
    target = fixture$target,
    audience = "analyst"
  )

  expect_s3_class(manager, "wf_quality_report")
  expect_true(all(c(
    "group", "n", "ess", "deff", "verdict", "light",
    "publish_separately", "action"
  ) %in% names(manager$table)))
  expect_true("margin_maxerr" %in% names(analyst$table))
  expect_identical(manager$audience, "manager")
  expect_identical(analyst$audience, "analyst")
  expect_identical(manager$language, "en")
  expect_equal(as.data.frame(manager), manager$table)
})

test_that("wf_report carries propensity-specific overlap and balance sections", {
  report <- wf_report(make_report_propensity(), audience = "analyst")

  expect_identical(report$source_type, "weights")
  expect_true(all(c(
    "propensity_overlap", "propensity_quantiles", "propensity_balance"
  ) %in% names(report$sections)))
  expect_equal(report$sections$propensity_overlap$n_boundary, 0)
  expect_equal(report$sections$propensity_balance$variable, "age")
})

test_that("wf_report supports blend summaries and sensitivity", {
  report <- wf_report(make_report_blend(), audience = "manager")

  expect_identical(report$source_type, "blend")
  expect_true(all(c(
    "blend_lambda", "blend_sensitivity", "blend_source_support"
  ) %in% names(report$sections)))
  expect_true(all(c("light", "action") %in% names(report$table)))
  expect_equal(report$table$light[report$table$group == "A"], "yellow")
})

test_that("wf_report renders Markdown and escaped base HTML", {
  fixture <- make_report_weights()
  markdown <- wf_report(
    fixture$weights,
    target = fixture$target,
    output = "markdown"
  )
  expect_type(markdown, "character")
  expect_length(markdown, 1)
  expect_match(markdown, "# Weighting quality report", fixed = TRUE)
  expect_match(markdown, "| Group |", fixed = TRUE)

  unsafe <- structure(
    list(
      data = data.frame(
        id = 1:3,
        group = "<unsafe&>",
        weight = c(1, 2, 3),
        feature = 1 / c(1, 2, 3)
      ),
      log = data.frame(group = "<unsafe&>", n = 3),
      achieved = NULL,
      provenance = list(method = "test")
    ),
    class = "wf_weights"
  )
  html <- wf_report(unsafe, output = "html")
  expect_match(html, "<!DOCTYPE html>", fixed = TRUE)
  expect_match(html, "&lt;unsafe&amp;&gt;", fixed = TRUE)
  expect_false(grepl("<unsafe&>", html, fixed = TRUE))
})

test_that("wf_report localizes human text while preserving stable schemas", {
  fixture <- make_report_weights()
  chinese_catalog <- .wf_i18n_catalog("zh_CN")
  english <- wf_report(
    fixture$weights,
    target = fixture$target,
    audience = "manager",
    lang = "en"
  )
  chinese <- wf_report(
    fixture$weights,
    target = fixture$target,
    audience = "manager",
    lang = "zh_CN"
  )

  expect_identical(chinese$language, "zh_CN")
  expect_identical(names(chinese$table), names(english$table))
  expect_identical(chinese$table$verdict, english$table$verdict)
  expect_true("verdict_label" %in% names(chinese$table))
  action_keys <- c(
    "action_weights_green",
    "action_weights_yellow",
    "action_weights_red"
  )
  expect_true(all(chinese$table$action %in% unname(chinese_catalog[action_keys])))

  markdown <- wf_report(
    fixture$weights,
    target = fixture$target,
    lang = "zh_CN",
    output = "markdown"
  )
  html <- wf_report(
    fixture$weights,
    target = fixture$target,
    lang = "zh_CN",
    output = "html"
  )
  expect_match(
    markdown,
    paste0("# ", unname(chinese_catalog["report_title"])),
    fixed = TRUE
  )
  expect_match(
    markdown,
    paste0("| ", unname(chinese_catalog["column_group"]), " |"),
    fixed = TRUE
  )
  expect_match(
    html,
    paste0("<title>", unname(chinese_catalog["report_title"]), "</title>"),
    fixed = TRUE
  )
})

test_that("wf_report honors the configured default language", {
  fixture <- make_report_weights()
  old <- options(wfc.lang = "zh_CN")
  on.exit(options(old), add = TRUE)

  report <- wf_report(fixture$weights, target = fixture$target)
  expect_identical(report$language, "zh_CN")
  expect_identical(
    report$title,
    unname(.wf_i18n_catalog("zh_CN")["report_title"])
  )
})

test_that("wf_report writes rendered output and invisibly returns the object", {
  fixture <- make_report_weights()
  path <- tempfile(fileext = ".md")

  result <- withVisible(wf_report(
    fixture$weights,
    target = fixture$target,
    output = "markdown",
    file = path
  ))

  expect_false(result$visible)
  expect_s3_class(result$value, "wf_quality_report")
  expect_true(file.exists(path))
  expect_match(paste(readLines(path), collapse = "\n"), "Weighting quality report")
})

test_that("wf_report validates sources, language, output and file combinations", {
  fixture <- make_report_weights()

  expect_error(wf_report(list()), class = "wf_error_input")
  expect_error(
    wf_report(fixture$weights, lang = "fr"),
    class = "wf_error_input"
  )
  expect_error(
    wf_report(fixture$weights, output = "object", file = tempfile()),
    class = "wf_error_input"
  )
})

test_that("wf_quality_report print renders the report", {
  fixture <- make_report_weights()
  report <- wf_report(fixture$weights, target = fixture$target)

  expect_output(print(report), "Weighting quality report")
  expect_output(print(report), "publishable separately")
})

with_language_environment <- function(code, language = "", lang = "") {
  vars <- c("LANGUAGE", "LC_ALL", "LC_MESSAGES", "LANG")
  old_options <- options(wfc.lang = NULL)
  old_env <- Sys.getenv(vars, unset = NA_character_)
  old_ctype <- Sys.getlocale("LC_CTYPE")
  on.exit({
    options(old_options)
    for (var in vars) {
      if (is.na(old_env[[var]])) {
        Sys.unsetenv(var)
      } else {
        do.call(Sys.setenv, stats::setNames(list(old_env[[var]]), var))
      }
    }
    suppressWarnings(Sys.setlocale("LC_CTYPE", old_ctype))
  }, add = TRUE)
  # .wf_lang() also inspects LC_ALL, LC_MESSAGES and the C-level LC_CTYPE.
  # Neutralize all three so the resolved language depends only on the
  # LANGUAGE/LANG values this helper simulates, not on the host locale.
  Sys.unsetenv(c("LC_ALL", "LC_MESSAGES"))
  suppressWarnings(Sys.setlocale("LC_CTYPE", "C"))
  Sys.setenv(LANGUAGE = language, LANG = lang)
  force(code)
}

test_that("English and Simplified Chinese catalogs have identical keys", {
  english <- .wf_i18n_catalog("en")
  chinese <- .wf_i18n_catalog("zh_CN")

  expect_type(english, "character")
  expect_type(chinese, "character")
  expect_setequal(names(english), names(chinese))
  expect_true(all(nzchar(english)))
  expect_true(all(nzchar(chinese)))
})

test_that("language aliases normalize to stable catalog names", {
  expect_identical(.wf_lang("en"), "en")
  expect_identical(.wf_lang("en_US"), "en")
  expect_identical(.wf_lang("en-GB"), "en")
  expect_identical(.wf_lang("zh"), "zh_CN")
  expect_identical(.wf_lang("zh_CN"), "zh_CN")
  expect_identical(.wf_lang("zh-CN"), "zh_CN")

  expect_error(.wf_lang("fr"), class = "wf_error_input")
  expect_error(.wf_lang(c("en", "zh")), class = "wf_error_input")
})

test_that("explicit language overrides options and options override locale", {
  old <- options(wfc.lang = "zh_CN")
  on.exit(options(old), add = TRUE)

  expect_identical(.wf_lang(), "zh_CN")
  expect_identical(.wf_lang("en"), "en")

  options(wfc.lang = "unsupported")
  expect_error(.wf_lang(), class = "wf_error_input")
})

test_that("session locale resolves Chinese and otherwise falls back to English", {
  with_language_environment({
    expect_identical(.wf_lang(), "zh_CN")
  }, language = "zh_CN", lang = "C")

  with_language_environment({
    expect_identical(.wf_lang(), "en")
  }, language = "", lang = "fr_FR.UTF-8")
})

test_that("translations interpolate and preserve stable keys", {
  chinese_catalog <- .wf_i18n_catalog("zh_CN")

  expect_identical(.wf_tr("report_title", lang = "en"), "Weighting quality report")
  expect_identical(
    .wf_tr("report_title", lang = "zh_CN"),
    unname(chinese_catalog["report_title"])
  )

  english <- .wf_tr("autoweigh_start", 80, 2, 2, lang = "en")
  chinese <- .wf_tr("autoweigh_start", 80, 2, 2, lang = "zh_CN")
  expect_match(english, "80 rows", fixed = TRUE)
  expect_match(chinese, "80", fixed = TRUE)
  expect_false(identical(english, chinese))

  expect_error(.wf_tr("missing_translation_key", lang = "en"),
               class = "wf_error_internal")
})

test_that("known column and section labels localize without changing names", {
  chinese_catalog <- .wf_i18n_catalog("zh_CN")

  expect_identical(.wf_i18n_label("group", "en"), "Group")
  expect_identical(
    .wf_i18n_label("group", "zh_CN"),
    unname(chinese_catalog["column_group"])
  )
  expect_identical(
    .wf_i18n_section("propensity_balance", "zh_CN"),
    unname(chinese_catalog["section_propensity_balance"])
  )
  expect_identical(.wf_i18n_label("unknown_field", "zh_CN"), "unknown field")
})

make_design_data <- function() {
  data.frame(
    id = paste0("u", 1:8),
    stratum = c("A", "A", "A", "A", "B", "B", "B", "B"),
    psu = c("a1", "a1", "a2", "a2", "b1", "b1", "b2", "b2"),
    y = c(1, 0, 1, 1, 0, 0, 1, 0),
    stringsAsFactors = FALSE
  )
}

make_weightflow_fixture <- function() {
  sample <- data.frame(
    id = sprintf("r%02d", 1:16),
    province = rep(c("A", "B"), each = 8),
    gender = rep(c("female", "male", "female", "male"), times = 4),
    age = rep(c("young", "young", "old", "old"), times = 4),
    stringsAsFactors = FALSE
  )

  pop <- data.frame(
    province = rep(c("A", "B"), each = 4),
    gender = rep(c("female", "male", "female", "male"), times = 2),
    age = rep(c("young", "young", "old", "old"), times = 2),
    count = c(40, 60, 60, 40, 30, 70, 50, 50),
    stringsAsFactors = FALSE
  )

  dims <- wf_dims(
    gender = c("female", "male"),
    age = c("young", "old")
  )

  target <- wf_target_population(
    pop = pop,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    dims = dims,
    by = "province"
  )

  list(sample = sample, pop = pop, dims = dims, target = target)
}

make_poststrat_fixture <- function() {
  dims <- wf_dims(gender = c("female", "male"), age = c("young", "old"))
  pop <- data.frame(
    province = "A",
    gender = c("female", "male", "female", "male"),
    age = c("young", "young", "old", "old"),
    count = c(50, 50, 50, 50),
    stringsAsFactors = FALSE
  )
  sample <- data.frame(
    id = paste0("r", 1:10),
    province = "A",
    gender = c(rep("female", 5), rep("male", 5)),
    age = c(rep("young", 4), "old", rep("young", 4), "old"),
    stringsAsFactors = FALSE
  )
  target <- wf_target_population(
    pop,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    dims = dims,
    by = "province",
    keep_joint = TRUE
  )
  ladder <- wf_collapse_ladder(
    dims,
    level1 = list(age = c("young" = "all", "old" = "all"))
  )
  list(
    sample = sample,
    pop = pop,
    target = target,
    dims = dims,
    ladder = ladder
  )
}

make_safe_workflow_fixture <- function(with_outcomes = FALSE) {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$base_weight <- 1
  design <- wf_prepare_design(
    sample,
    id = "id",
    calibration = c("province", "gender", "age"),
    base_weight = "base_weight"
  )
  target <- wf_target_population(
    fixture$pop,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    dims = fixture$dims,
    by = "province",
    by_key = "province",
    keep_joint = TRUE
  )
  target <- .wf_verified_target(
    target,
    evidence = list(demo_only = FALSE, data_checksum = strrep("c", 64)),
    source_type = "population"
  )
  analysis <- data.frame(
    id = sample$id,
    score = seq_len(nrow(sample)),
    approved = rep(c(0, 1), length.out = nrow(sample)),
    stringsAsFactors = FALSE
  )

  list(
    design = design,
    target = target,
    dims = fixture$dims,
    analysis = if (with_outcomes) analysis else NULL
  )
}

make_safe_population_files <- function(population) {
  path <- tempfile("wf-safe-population-")
  dir.create(path)
  data_file <- file.path(path, "population.csv")
  source_file <- paste0(data_file, ".source.dcf")
  utils::write.csv(population, data_file, row.names = FALSE)
  source <- c(
    publisher = "Synthetic fixture publisher",
    dataset_title = "Synthetic population fixture",
    citation = "Synthetic test fixture; not an external authority",
    reference_period = "2026",
    population_scope = "Synthetic adults",
    retrieved_at = "2026-07-15",
    license = "CC0-1.0",
    checksum_algorithm = "sha256",
    checksum = .wf_sha256_file(data_file),
    transformation = "No transformation",
    selected_before_outcomes = "true",
    demo_only = "false"
  )
  writeLines(sprintf("%s: %s", names(source), source), source_file)
  list(data_file = data_file, source_file = source_file)
}

make_locked_safe_weights <- function(fixture = make_safe_workflow_fixture()) {
  plan <- wf_plan_weights(fixture$design, fixture$target, fixture$dims)
  approval <- wf_approve_plan(plan, "Test reviewer", "statistician")
  wf_execute_plan(plan, approval, fixture$design, fixture$target)
}

make_safe_review_workflow <- function(fixture = make_safe_workflow_fixture()) {
  cell_plan <- wf_plan_cells(
    fixture$design,
    fixture$target,
    fixture$dims,
    min_cell = 1,
    max_weight_ratio = 10
  )
  plan <- wf_plan_weights(
    fixture$design,
    fixture$target,
    fixture$dims,
    min_cell = 1,
    cell_plan = cell_plan
  )
  workflow <- list(
    design = fixture$design,
    target = fixture$target,
    cell_plan = cell_plan,
    plan = plan,
    weights = NULL,
    created = .wf_iso_time(),
    package_version = .wf_package_version()
  )
  workflow$identity <- .wf_safe_workflow_identity(workflow)
  class(workflow) <- "wf_safe_workflow"
  workflow
}

test_that("wf_rake agrees with survey::rake on grouped margins", {
  skip_if_not_installed("survey")

  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$design_weight <- seq(0.75, 1.5, length.out = nrow(sample))

  actual <- .wf_rake_engine(
    sample,
    fixture$target,
    id = "id",
    init_weight = "design_weight",
    trim = NULL,
    tol = 1e-10,
    max_iter = 1000
  )

  for (group in names(fixture$target$groups)) {
    group_sample <- sample[sample$province == group, , drop = FALSE]
    group_target <- fixture$target$groups[[group]]
    design <- survey::svydesign(
      ids = ~1,
      weights = ~design_weight,
      data = group_sample
    )
    oracle <- survey::rake(
      design,
      sample.margins = list(~gender, ~age),
      population.margins = list(
        data.frame(
          gender = names(group_target$margins$gender),
          Freq = unname(group_target$margins$gender)
        ),
        data.frame(
          age = names(group_target$margins$age),
          Freq = unname(group_target$margins$age)
        )
      ),
      control = list(
        maxit = 1000,
        epsilon = 1e-10,
        verbose = FALSE
      )
    )

    actual_weight <- actual$data$weight[
      match(group_sample$id, actual$data$id)
    ]
    expect_equal(
      unname(actual_weight),
      unname(stats::weights(oracle)),
      tolerance = 1e-7,
      info = paste("group", group)
    )
  }
})

test_that("precheck reports unknown sample categories", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$gender[1] <- "other"

  pc <- wf_precheck(sample, fixture$target, id = "id")

  expect_false(pc$ok)
  expect_true("cat_unknown_in_sample" %in% pc$issues$check)
})

test_that("precheck reports infeasible positive target cells", {
  fixture <- make_weightflow_fixture()
  sample <- subset(fixture$sample, !(province == "B" & gender == "male"))

  pc <- wf_precheck(sample, fixture$target, id = "id", na = "drop")

  expect_false(pc$ok)
  expect_true("cat_infeasible" %in% pc$issues$check)
})

test_that("precheck reports duplicate ids", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$id[2] <- sample$id[1]

  pc <- wf_precheck(sample, fixture$target, id = "id")

  expect_false(pc$ok)
  expect_true("dup_id" %in% pc$issues$check)
})

test_that("precheck reports overloaded missing dimensions", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$gender[1] <- NA
  sample$age[1] <- NA

  pc <- wf_precheck(sample, fixture$target, id = "id", max_na_dims = 1)

  expect_false(pc$ok)
  expect_true("na_overload" %in% pc$issues$check)
})

test_that("precheck returns a structured schema failure before deeper checks", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$age <- NULL

  pc <- wf_precheck(sample, fixture$target, id = "id")

  expect_s3_class(pc, "wf_precheck")
  expect_false(pc$ok)
  expect_identical(pc$issues$check, "schema_missing_var")
  expect_match(pc$issues$detail, "age")
})

test_that("precheck reports unmatched and missing group keys", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$province[1] <- "C"
  sample$province[2] <- NA

  pc <- wf_precheck(sample, fixture$target, id = "id")

  expect_false(pc$ok)
  expect_true("group_unmatched" %in% pc$issues$check)
  expect_true("na_group" %in% pc$issues$check)
})

test_that("precheck distinguishes missing-value policies", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$age[1] <- NA

  fractional <- wf_precheck(sample, fixture$target, id = "id", na = "fractional")
  strict <- wf_precheck(sample, fixture$target, id = "id", na = "error")

  expect_true(fractional$ok)
  expect_true(any(
    fractional$issues$check == "na_load" &
      fractional$issues$severity == "note"
  ))
  expect_false(strict$ok)
  expect_true(any(
    strict$issues$check == "na_load" &
      strict$issues$severity == "error"
  ))
})

test_that("precheck records thin-support and extreme-ratio quality warnings", {
  fixture <- make_weightflow_fixture()

  pc <- wf_precheck(
    fixture$sample,
    fixture$target,
    id = "id",
    thin_min = 5,
    risk_ratio = 1
  )

  expect_true(pc$ok)
  expect_true("support_thin" %in% pc$issues$check)
  expect_true("risk_extreme_ratio" %in% pc$issues$check)
  expect_true(all(pc$issues$severity != "error"))
})

test_that("precheck print covers clean and blocked summaries", {
  fixture <- make_weightflow_fixture()
  clean <- wf_precheck(
    fixture$sample,
    fixture$target,
    id = "id",
    thin_min = 1,
    risk_ratio = 100
  )
  expect_output(print(clean), "no issues found")

  blocked_sample <- fixture$sample
  blocked_sample$gender[1] <- "other"
  blocked <- wf_precheck(blocked_sample, fixture$target, id = "id")
  expect_output(print(blocked), "BLOCKED")
  expect_output(print(blocked), "Categories in sample")
})

run_autoweigh <- function(...) {
  suppressMessages(.wf_autoweigh_engine(...))
}

make_autoweigh_collapse_fixture <- function() {
  sample <- data.frame(
    id = paste0("r", 1:6),
    province = "A",
    edu = c("low", "mid", "mid", "high", "high", "high"),
    stringsAsFactors = FALSE
  )
  population <- data.frame(
    province = "A",
    edu = c("low", "mid", "high"),
    count = c(40, 30, 30),
    stringsAsFactors = FALSE
  )
  dims <- wf_dims(
    edu = c("low", "mid", "high"),
    .collapse = list(
      edu = list(
        step1 = c(low = "low_mid", mid = "low_mid", high = "high")
      )
    )
  )
  target <- wf_target_population(
    population,
    key_map = c(edu = "edu"),
    count = "count",
    dims = dims,
    by = "province"
  )
  list(sample = sample, target = target, dims = dims)
}

test_that("wf_autoweigh builds a target and returns an auditable raking result", {
  fixture <- make_weightflow_fixture()
  original_sample <- fixture$sample
  original_population <- fixture$pop

  result <- run_autoweigh(
    fixture$sample,
    fixture$pop,
    fixture$dims,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    by = "province",
    id = "id",
    interactive = FALSE
  )

  expect_s3_class(result, "wf_autoweigh_result")
  expect_s3_class(result$weights, "wf_weights")
  expect_s3_class(result$diagnostics, "wf_diagnostics")
  expect_s3_class(result$report, "wf_quality_report")
  expect_identical(result$method, "raking")
  expect_identical(result$language, "en")
  expect_identical(
    names(result$ledger),
    c("step", "action", "detail_key", "detail", "artifact_class", "time")
  )
  expect_identical(result$ledger$step, seq_len(nrow(result$ledger)))
  expect_true(all(c(
    "start", "target", "precheck", "trim", "calibrate", "done"
  ) %in% result$ledger$action))
  expect_true(all(nzchar(result$ledger$detail_key)))
  expect_length(result$artifacts, nrow(result$ledger))
  expect_identical(fixture$sample, original_sample)
  expect_identical(fixture$pop, original_population)
})

test_that("wf_autoweigh honors ready targets and explicit logit calibration", {
  fixture <- make_weightflow_fixture()

  result <- run_autoweigh(
    fixture$sample,
    fixture$target,
    fixture$dims,
    id = "id",
    method = "logit",
    bounds = c(0.2, 5),
    trim = NULL,
    interactive = FALSE
  )

  expect_identical(result$method, "logit")
  expect_identical(result$weights$provenance$method, "logit")
  expect_equal(result$weights$provenance$bounds, c(0.2, 5))
  expect_true("autoweigh_target_ready" %in% result$ledger$detail_key)
  expect_false("trim" %in% result$ledger$action)
})

test_that("auto method selects reviewed post-stratification when configured", {
  fixture <- make_poststrat_fixture()

  result <- run_autoweigh(
    fixture$sample,
    fixture$pop,
    fixture$dims,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    by = "province",
    id = "id",
    ladder = fixture$ladder,
    min_cell = 1,
    interactive = FALSE
  )

  expect_identical(result$method, "poststrat")
  expect_identical(result$weights$provenance$method, "poststrat")
  expect_true(length(result$target$joint) > 0)
  expect_true(all(c("cell_report", "collapse_map") %in% names(result$weights)))
})

test_that("wf_autoweigh applies only declared collapse remediations", {
  fixture <- make_autoweigh_collapse_fixture()
  sample <- subset(fixture$sample, edu != "low")
  original_target <- fixture$target

  result <- run_autoweigh(
    sample,
    fixture$target,
    fixture$dims,
    id = "id",
    trim = NULL,
    interactive = FALSE
  )

  expect_true("collapse" %in% result$ledger$action)
  expect_true("low_mid" %in% result$sample$edu)
  expect_true(
    "low_mid" %in% names(result$target$groups$A$margins$edu)
  )
  expect_identical(fixture$target, original_target)

  fixture$dims$collapse <- list()
  expect_error(
    run_autoweigh(
      sample,
      fixture$target,
      fixture$dims,
      id = "id",
      trim = NULL,
      interactive = FALSE
    ),
    class = "wf_error_feasibility"
  )
})

test_that("wf_autoweigh applies finite trim recommendations", {
  fixture <- make_weightflow_fixture()
  fixture$sample$base_weight <- 1
  fixture$sample$base_weight[1] <- 50

  result <- run_autoweigh(
    fixture$sample,
    fixture$target,
    fixture$dims,
    id = "id",
    max_deff = 2.1,
    max_residual = 0.04,
    caps = c(2, 3, 4),
    init_weight = "base_weight",
    tol = 1e-8,
    interactive = FALSE
  )

  expect_equal(result$weights$provenance$trim, c(0.05, 3))
  trim_step <- result$ledger[result$ledger$action == "trim", , drop = FALSE]
  expect_identical(trim_step$detail_key, "autoweigh_trim_applied")
  expect_true(any(result$ledger$artifact_class == "wf_auto_trim"))
})

test_that("wf_autoweigh records an unsuccessful trim search and continues", {
  fixture <- make_weightflow_fixture()
  fixture$sample$base_weight <- 1
  fixture$sample$base_weight[1] <- 50

  result <- run_autoweigh(
    fixture$sample,
    fixture$target,
    fixture$dims,
    id = "id",
    max_deff = 1,
    max_residual = 0,
    caps = c(2, 3),
    init_weight = "base_weight",
    interactive = FALSE
  )

  expect_null(result$weights$provenance$trim)
  expect_true(
    "autoweigh_trim_no_solution" %in% result$ledger$detail_key
  )
})

test_that("guided narration localizes without changing stable ledger keys", {
  fixture <- make_weightflow_fixture()

  english <- run_autoweigh(
    fixture$sample,
    fixture$target,
    fixture$dims,
    id = "id",
    trim = NULL,
    interactive = FALSE,
    lang = "en"
  )
  chinese <- run_autoweigh(
    fixture$sample,
    fixture$target,
    fixture$dims,
    id = "id",
    trim = NULL,
    interactive = FALSE,
    lang = "zh_CN"
  )

  expect_identical(chinese$language, "zh_CN")
  expect_identical(chinese$ledger$action, english$ledger$action)
  expect_identical(chinese$ledger$detail_key, english$ledger$detail_key)
  expect_false(identical(chinese$ledger$detail, english$ledger$detail))
  expect_identical(chinese$report$language, "zh_CN")
})

test_that("wf_autoweigh validates workflow-owned decisions", {
  fixture <- make_weightflow_fixture()

  expect_error(
    run_autoweigh(fixture$sample, fixture$pop, fixture$dims),
    class = "wf_error_schema"
  )
  expect_error(
    run_autoweigh(
      fixture$sample,
      fixture$target,
      fixture$dims,
      method = "poststrat",
      interactive = FALSE
    ),
    class = "wf_error_feasibility"
  )
  expect_error(
    run_autoweigh(
      fixture$sample,
      fixture$target,
      fixture$dims,
      interactive = FALSE,
      lang = "fr"
    ),
    class = "wf_error_input"
  )
})

test_that("interactive guided decisions can stop before calibration", {
  fixture <- make_autoweigh_collapse_fixture()
  sample <- subset(fixture$sample, edu != "low")
  testthat::local_mocked_bindings(
    .wf_autoweigh_confirm = function(prompt) FALSE,
    .package = "WFC"
  )

  expect_error(
    capture.output(
      run_autoweigh(
        sample,
        fixture$target,
        fixture$dims,
        id = "id",
        trim = NULL,
        interactive = TRUE
      )
    ),
    class = "wf_error_feasibility"
  )
})

test_that("wf_autoweigh_result prints localized ledger details", {
  fixture <- make_weightflow_fixture()
  result <- run_autoweigh(
    fixture$sample,
    fixture$target,
    fixture$dims,
    id = "id",
    trim = NULL,
    interactive = FALSE
  )

  expect_output(
    print(result),
    .wf_tr("autoweigh_print_header", lang = "en"),
    fixed = TRUE
  )
  expect_output(print(result), "precheck", fixed = TRUE)
})

test_that("decision and statistician aliases preserve legacy report tables", {
  fixture <- make_weightflow_fixture()
  f <- list(
    weights = .wf_rake_engine(fixture$sample, fixture$target, id = "id"),
    target = fixture$target
  )

  decision <- wf_report(f$weights, f$target, audience = "decision")
  manager <- wf_report(f$weights, f$target, audience = "manager")
  statistician <- wf_report(f$weights, f$target, audience = "statistician")
  analyst <- wf_report(f$weights, f$target, audience = "analyst")

  expect_identical(decision$table, manager$table)
  expect_identical(statistician$table, analyst$table)
  expect_identical(decision$requested_audience, "decision")
  expect_identical(decision$audience, "manager")
})

test_that("safe workflow reports separate decision and statistician detail", {
  workflow <- make_safe_review_workflow()

  decision <- wf_report(workflow, audience = "decision")
  statistician <- wf_report(workflow, audience = "statistician")

  expect_identical(decision$source_type, "safe_workflow")
  expect_true(all(c("status", "method", "affected_share", "next_action") %in%
    names(decision$table)))
  expect_true(all(c("source_evidence", "precheck_issues", "cell_map") %in%
    names(statistician$sections)))
  expect_gt(ncol(statistician$sections$cell_map), 0)
})

test_that("locked and impact reports expose evidence without planning controls", {
  f <- make_safe_workflow_fixture(with_outcomes = TRUE)
  locked <- make_locked_safe_weights(f)
  impact <- wf_assess_impact(locked, f$analysis, "id", c("score", "approved"))

  locked_report <- wf_report(locked, audience = "statistician")
  impact_report <- wf_report(impact, audience = "decision")

  expect_identical(locked_report$source_type, "locked_weights")
  expect_true("lock_evidence" %in% names(locked_report$sections))
  expect_identical(impact_report$source_type, "impact")
  expect_true(all(c("outcome", "unweighted", "weighted", "difference") %in%
    names(impact_report$table)))
  expect_false(any(c("bounds", "target", "method") %in%
    names(impact_report$table)))
})

test_that("safe audit v2 carries the complete available identity chain", {
  f <- make_safe_workflow_fixture(with_outcomes = TRUE)
  locked <- make_locked_safe_weights(f)
  impact <- wf_assess_impact(locked, f$analysis, "id", "score")

  locked_payload <- .wf_audit_payload(locked)
  impact_payload <- .wf_audit_payload(impact)

  expect_identical(locked_payload$schema, "wfc_audit_v2")
  expect_identical(
    locked_payload$identities$locked_weight,
    locked$identity
  )
  expect_identical(locked_payload$identities$plan, locked$plan_identity)
  expect_identical(locked_payload$identities$approval, locked$approval_identity)
  expect_identical(impact_payload$identities$impact, impact$identity)
  expect_identical(
    impact_payload$identities$locked_weight,
    impact$weight_identity
  )

  path <- tempfile(fileext = ".json")
  expect_identical(wf_audit_export(impact, path), path)
  expect_match(
    paste(readLines(path, warn = FALSE), collapse = "\n"),
    "\"schema\":\"wfc_audit_v2\"",
    fixed = TRUE
  )
})

test_that("wf_rake init_weight = NULL reproduces the default result exactly", {
  fixture <- make_weightflow_fixture()
  a <- .wf_rake_engine(fixture$sample, fixture$target, id = "id")
  b <- .wf_rake_engine(fixture$sample, fixture$target, id = "id", init_weight = NULL)
  expect_equal(a$data$weight, b$data$weight)
})

test_that("wf_rake honors non-uniform init weights while matching margins", {
  fixture <- make_weightflow_fixture()
  s <- fixture$sample
  # Vary init WITHIN margin categories (young females only), so it is not
  # absorbed by the marginal gender/age calibration and actually shifts the
  # within-cell association.
  s$bw <- ifelse(s$gender == "female" & s$age == "young", 3, 1)

  uniform <- .wf_rake_engine(s, fixture$target, id = "id")
  weighted <- .wf_rake_engine(s, fixture$target, id = "id", init_weight = "bw")

  # init weights change the within-solution distribution
  expect_false(isTRUE(all.equal(uniform$data$weight, weighted$data$weight)))

  # but achieved gender totals still match (raking still hits its margins)
  female_uni <- sum(uniform$data$weight[s$gender == "female"])
  female_wtd <- sum(weighted$data$weight[s$gender == "female"])
  expect_equal(female_uni, female_wtd, tolerance = 1e-6)
})

test_that("wf_rake errors when init_weight column is missing", {
  fixture <- make_weightflow_fixture()
  expect_error(
    .wf_rake_engine(fixture$sample, fixture$target, id = "id", init_weight = "nope"),
    class = "wf_error_schema"
  )
})

test_that("wf_rake rejects invalid initial weights", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$base_weight <- 1
  sample$base_weight[1] <- -1

  expect_error(
    .wf_rake_engine(sample, fixture$target, id = "id", init_weight = "base_weight"),
    class = "wf_error_input"
  )

  sample$base_weight[1] <- NA_real_
  expect_error(
    .wf_rake_engine(sample, fixture$target, id = "id", init_weight = "base_weight"),
    class = "wf_error_input"
  )
})

test_that("wf_rake applies explicit missing-value policies", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$age[1] <- NA

  expect_error(
    .wf_rake_engine(sample, fixture$target, id = "id", na = "error"),
    class = "wf_error_feasibility"
  )

  expect_warning(
    dropped <- .wf_rake_engine(sample, fixture$target, id = "id", na = "drop"),
    class = "wf_warning_data"
  )
  expect_equal(nrow(dropped$data), nrow(sample) - 1)

  fractional <- .wf_rake_engine(sample, fixture$target, id = "id", na = "fractional")
  expect_equal(nrow(fractional$data), nrow(sample))
  expect_true(all(fractional$data$weight > 0))
})

test_that("poststrat planner requires a joint target", {
  fixture <- make_weightflow_fixture()
  ladder <- wf_collapse_ladder(fixture$dims)

  expect_error(
    wf_plan_poststrat(fixture$sample, fixture$target, min_cell = 2, ladder = ladder),
    class = "wf_error_schema"
  )
})

test_that("poststrat planner resolves sparse cells using the ladder", {
  fixture <- make_poststrat_fixture()

  plan <- wf_plan_poststrat(
    fixture$sample,
    fixture$target,
    min_cell = 2,
    ladder = fixture$ladder,
    granularity = "adaptive"
  )

  expect_s3_class(plan, "wf_poststrat_plan")
  expect_equal(nrow(plan$plan), 4)
  expect_true(all(c("group", "gender", "age", "pop", "n_sample", "ladder_level", "resolved_cell", "orphan") %in% names(plan$plan)))
  expect_true(any(plan$plan$ladder_level == 1))
  expect_equal(plan$diagnostics$n_cells_raw, 4)
  expect_equal(plan$diagnostics$n_orphan, 0)
})

test_that("poststrat planner flags unsupported cells when requested", {
  fixture <- make_poststrat_fixture()
  fixture$sample <- fixture$sample[fixture$sample$gender == "female", ]

  plan <- wf_plan_poststrat(
    fixture$sample,
    fixture$target,
    min_cell = 1,
    ladder = fixture$ladder,
    empty_cell = "flag"
  )

  expect_true(any(plan$plan$orphan))
  expect_gt(plan$diagnostics$pop_orphan, 0)
})

test_that("wf_calibrate dispatches to raking", {
  fixture <- make_weightflow_fixture()

  direct <- .wf_rake_engine(fixture$sample, fixture$target, id = "id", tol = 1e-8)
  via <- .wf_calibrate_engine(fixture$sample, fixture$target, method = "raking", id = "id", tol = 1e-8)

  expect_s3_class(via, "wf_weights")
  expect_equal(via$data$weight, direct$data$weight, tolerance = 1e-8)
  expect_equal(via$provenance$method, "raking")
})

test_that("wf_calibrate defaults to raking", {
  fixture <- make_weightflow_fixture()

  via <- .wf_calibrate_engine(fixture$sample, fixture$target, id = "id", tol = 1e-8)

  expect_s3_class(via, "wf_weights")
  expect_equal(via$provenance$method, "raking")
})

test_that("wf_calibrate dispatches to poststrat", {
  fixture <- make_poststrat_fixture()

  direct <- .wf_poststrat_engine(fixture$sample, fixture$target, min_cell = 1, ladder = fixture$ladder, id = "id")
  via <- .wf_calibrate_engine(
    fixture$sample,
    fixture$target,
    method = "poststrat",
    min_cell = 1,
    ladder = fixture$ladder,
    id = "id"
  )

  expect_s3_class(via, "wf_weights")
  expect_equal(via$data$weight, direct$data$weight, tolerance = 1e-8)
  expect_equal(via$provenance$method, "poststrat")
})

test_that("wf_calibrate rejects unknown methods", {
  fixture <- make_weightflow_fixture()

  expect_error(
    .wf_calibrate_engine(fixture$sample, fixture$target, method = "linear"),
    class = "wf_error_input"
  )
})

test_that("foundation API exports are available", {
  expect_false("wf_target_manual" %in% getNamespaceExports("WFC"))
  expect_false("wf_target_shrink" %in% getNamespaceExports("WFC"))
  expect_true(is.function(wf_suggest_collapse))
  expect_true(is.function(wf_calibrate))
})

expect_safety_code <- function(object, code) {
  expect_s3_class(object, "wf_error_safety")
  expect_identical(object$data$code, code)
}

test_that("raw samples and arbitrary targets do not reach weighting engines", {
  fixture <- make_safe_workflow_fixture()
  raw <- fixture$design$data

  expect_safety_code(tryCatch(
    wf_calibrate(raw, fixture$target, method = "raking"),
    error = identity
  ), "verified_weighting_inputs_required")
  expect_safety_code(tryCatch(
    wf_rake(raw, fixture$target),
    error = identity
  ), "verified_weighting_inputs_required")
  expect_safety_code(tryCatch(
    wf_autoweigh(raw, fixture$target, fixture$dims, interactive = FALSE),
    error = identity
  ), "verified_weighting_inputs_required")
  expect_safety_code(tryCatch(
    wf_auto_trim(raw, fixture$target, caps = c(2, 4)),
    error = identity
  ), "verified_weighting_inputs_required")
  expect_error(
    wf_plan_weights(fixture$design, unclass(fixture$target), fixture$dims),
    class = "wf_error_safety"
  )
})

test_that("verified design and target objects reach supported public methods", {
  fixture <- make_safe_workflow_fixture()

  calibrated <- wf_calibrate(
    fixture$design,
    fixture$target,
    method = "raking"
  )
  raked <- wf_rake(fixture$design, fixture$target)
  guided <- wf_autoweigh(
    fixture$design,
    fixture$target,
    fixture$dims,
    method = "raking",
    trim = NULL,
    interactive = FALSE
  )
  trim_review <- wf_auto_trim(
    fixture$design,
    fixture$target,
    caps = c(2, 4),
    max_deff = 20
  )

  expect_s3_class(calibrated, "wf_weights")
  expect_s3_class(raked, "wf_weights")
  expect_s3_class(guided, "wf_autoweigh_result")
  expect_s3_class(trim_review, "wf_auto_trim")
  expect_identical(calibrated$provenance$design_identity, fixture$design$identity)
  expect_identical(calibrated$provenance$target_identity, fixture$target$identity)
})

test_that("manual pipelines and unverified object targets are blocked", {
  manual <- tryCatch(
    wf_pipeline(
      list(mode = "manual"),
      list(calibrate = list(method = "raking"))
    ),
    error = identity
  )
  expect_safety_code(manual, "manual_pipeline_unsupported")

  ordinary_target <- make_weightflow_fixture()$target
  object_mode <- tryCatch(
    wf_pipeline(
      ordinary_target,
      list(calibrate = list(method = "raking"))
    ),
    error = identity
  )
  expect_safety_code(object_mode, "verified_target_required")

  expect_false("margins" %in% names(formals(wf_run)))
})

test_that("the 1.1 contract preserves the completed deprecation record", {
  freeze <- system.file("stability/api-freeze.md", package = "WFC")
  expect_true(nzchar(freeze))
  contract <- readLines(freeze, warn = FALSE)

  expect_true(any(grepl("scheduled for removal in 2.0.0", contract, fixed = TRUE)))
  expect_true(any(grepl("manual target margins", contract, fixed = TRUE)))
  expect_true(any(grepl("target shrinkage", contract, fixed = TRUE)))
  expect_true(any(grepl("inline entropy-balancing moment targets", contract, fixed = TRUE)))
})

test_that("WFC 2.0 enforces rather than repeats 1.1 warnings", {
  exports <- getNamespaceExports("WFC")

  expect_false("wf_target_manual" %in% exports)
  expect_false("wf_target_shrink" %in% exports)

  fixture <- make_safe_workflow_fixture()
  err <- tryCatch(
    wf_calibrate(
      fixture$design$data,
      fixture$target,
      method = "ebal",
      moments = c(outcome = 0.5)
    ),
    error = identity
  )
  expect_s3_class(err, "wf_error_safety")
  expect_identical(err$data$code, "inline_moments_unsupported")
})

test_that("parallel worker resolution never exceeds the CRAN two-core limit", {
  old <- options(wfc.parallel.cores = 64L)
  on.exit(options(old), add = TRUE)

  expected <- if (identical(.Platform$OS.type, "windows")) 1L else 2L
  expect_equal(.wf_parallel_workers(100L, use_parallel = TRUE), expected)
})

test_that("wf_rake parallel execution matches serial output", {
  old <- options(wfc.parallel.cores = 2L)
  on.exit(options(old), add = TRUE)

  fixture <- make_weightflow_fixture()
  serial <- .wf_rake_engine(fixture$sample, fixture$target, id = "id", tol = 1e-10)
  forked <- .wf_rake_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    tol = 1e-10,
    parallel = TRUE,
    progress = TRUE
  )

  serial_weight <- stats::setNames(serial$data$weight, serial$data$id)
  forked_weight <- stats::setNames(forked$data$weight, forked$data$id)

  expect_equal(forked_weight[names(serial_weight)], serial_weight, tolerance = 1e-10)
  expect_true(forked$provenance$parallel)
  expect_true(forked$provenance$progress)
  expect_gte(forked$provenance$parallel_workers, 1)
})

test_that("wf_poststrat records parallel and progress provenance", {
  old <- options(wfc.parallel.cores = 2L)
  on.exit(options(old), add = TRUE)

  fixture <- make_poststrat_fixture()
  weights <- .wf_poststrat_engine(
    fixture$sample,
    fixture$target,
    min_cell = 1,
    ladder = fixture$ladder,
    id = "id",
    parallel = TRUE,
    progress = TRUE
  )

  expect_s3_class(weights, "wf_weights")
  expect_true(weights$provenance$parallel)
  expect_true(weights$provenance$progress)
  expect_gte(weights$provenance$parallel_workers, 1)
})

test_that("wf_replicates parallel execution matches serial output", {
  old <- options(wfc.parallel.cores = 2L)
  on.exit(options(old), add = TRUE)

  d <- make_design_data()
  refit <- function(data, weights) {
    structure(list(
      data = data.frame(
        id = data$id,
        group = "all",
        weight = weights,
        feature = 1 / weights,
        stringsAsFactors = FALSE
      )
    ), class = "wf_weights")
  }
  serial <- wf_replicates(
    d,
    refit,
    method = "bootstrap",
    R = 12,
    strata = "stratum",
    clusters = "psu",
    id = "id",
    seed = 10
  )
  forked <- wf_replicates(
    d,
    refit,
    method = "bootstrap",
    R = 12,
    strata = "stratum",
    clusters = "psu",
    id = "id",
    seed = 10,
    parallel = TRUE
  )

  expect_equal(forked$base, serial$base)
  expect_equal(forked$replicates, serial$replicates)
  expect_true(forked$provenance$parallel)
  expect_gte(forked$provenance$parallel_workers, 1)
})

test_that(".wf_parallel_map replays classed warnings and errors", {
  old <- options(wfc.parallel.cores = 2L)
  on.exit(options(old), add = TRUE)

  expect_warning(
    .wf_parallel_map(
      as.list(1:2),
      function(i) {
        if (i == 2) {
          wf_warn("parallel warning", "wf_warning_data", list(i = i))
        }
        i
      },
      use_parallel = TRUE
    ),
    class = "wf_warning_data"
  )

  expect_error(
    .wf_parallel_map(
      as.list(1:2),
      function(i) {
        if (i == 2) {
          wf_abort("parallel error", "wf_error_input", list(i = i))
        }
        i
      },
      use_parallel = TRUE
    ),
    class = "wf_error_input"
  )
})

test_that("subjective target APIs are absent", {
  exports <- getNamespaceExports("WFC")

  expect_false("wf_target_manual" %in% exports)
  expect_false("wf_target_shrink" %in% exports)
})

test_that("inline moments are blocked before entropy calibration", {
  fixture <- make_safe_workflow_fixture()
  sample <- fixture$design$data
  sample$x <- seq_len(nrow(sample))

  err <- tryCatch(
    wf_calibrate(
      sample,
      fixture$target,
      method = "ebal",
      moments = c(x = 1)
    ),
    error = identity
  )

  expect_s3_class(err, "wf_error_safety")
  expect_identical(err$data$code, "inline_moments_unsupported")
  expect_identical(err$data$field, "moments")
  expect_identical(
    err$data$next_actions,
    "import_verified_external_margins"
  )
})

interoperability_refit <- function(data, weights) {
  structure(
    list(
      data = data.frame(
        id = data$id,
        group = "all",
        weight = weights,
        feature = 1 / weights,
        stringsAsFactors = FALSE
      )
    ),
    class = "wf_weights"
  )
}

interoperability_mean <- function(weights, data) {
  sum(weights * data$y) / sum(weights)
}

make_interoperability_replicates <- function(method, strata = "stratum") {
  data <- make_design_data()
  replicates <- suppressWarnings(wf_replicates(
    data,
    interoperability_refit,
    method = method,
    R = 40,
    strata = strata,
    clusters = "psu",
    id = "id",
    seed = 11
  ))
  list(data = data, replicates = replicates)
}

test_that("as_svrepdesign is exported", {
  expect_true("as_svrepdesign" %in% getNamespaceExports("WFC"))
  expect_true(is.function(as_svrepdesign))
})

test_that("as_svrepdesign maps WFC replication methods and stored scales", {
  testthat::skip_if_not_installed("survey")
  cases <- list(
    list(method = "bootstrap", strata = "stratum", type = "bootstrap"),
    list(method = "jackknife", strata = NULL, type = "JK1"),
    list(method = "jackknife", strata = "stratum", type = "JKn"),
    list(method = "brr", strata = "stratum", type = "BRR")
  )

  for (case in cases) {
    setup <- make_interoperability_replicates(case$method, case$strata)
    data <- setup$data[rev(seq_len(nrow(setup$data))), ]
    original <- setup$replicates
    design <- as_svrepdesign(setup$replicates, data, id = "id")
    order <- match(data$id, setup$replicates$base$id)

    expect_s3_class(design, "svyrep.design")
    expect_identical(design$type, case$type)
    expect_equal(design$scale, setup$replicates$scale)
    expect_equal(design$rscales, setup$replicates$rscales)
    expect_true(isTRUE(design$mse))
    expect_equal(
      as.numeric(stats::weights(design, type = "sampling")),
      setup$replicates$base$weight[order]
    )
    expect_equal(
      unname(stats::weights(design, type = "replication")),
      unname(setup$replicates$replicates[order, , drop = FALSE])
    )
    expect_equal(
      design$variables$.wf_weight,
      setup$replicates$base$weight[order]
    )
    expect_identical(
      attr(design, "wfc_provenance"),
      setup$replicates$provenance
    )
    expect_identical(setup$replicates, original)
  }
})

test_that("survey replicate estimates reproduce wf_variance", {
  testthat::skip_if_not_installed("survey")

  for (method in c("bootstrap", "jackknife", "brr")) {
    setup <- make_interoperability_replicates(method)
    data <- setup$data[rev(seq_len(nrow(setup$data))), ]
    design <- as_svrepdesign(setup$replicates, data, id = "id")
    survey_result <- survey::svymean(~y, design)
    wfc_result <- wf_variance(
      setup$replicates,
      interoperability_mean,
      setup$data
    )

    expect_equal(
      as.numeric(survey_result),
      wfc_result$table$estimate,
      tolerance = 1e-12
    )
    expect_equal(
      as.numeric(survey::SE(survey_result)),
      wfc_result$table$se,
      tolerance = 1e-12
    )
  }
})

test_that("as_svrepdesign forwards explicit degrees of freedom", {
  testthat::skip_if_not_installed("survey")
  setup <- make_interoperability_replicates("bootstrap")

  design <- as_svrepdesign(
    setup$replicates,
    setup$data,
    id = "id",
    degf = 3
  )

  expect_equal(as.numeric(survey::degf(design)), 3)
})

test_that("as_svrepdesign validates IDs, metadata, and reserved controls", {
  testthat::skip_if_not_installed("survey")
  setup <- make_interoperability_replicates("bootstrap")
  data <- setup$data

  expect_error(as_svrepdesign(list(), data), class = "wf_error_input")
  expect_error(
    as_svrepdesign(setup$replicates, list()),
    class = "wf_error_input"
  )
  expect_error(
    as_svrepdesign(setup$replicates, data[-1, ]),
    class = "wf_error_schema"
  )

  duplicate <- setup$replicates
  duplicate$base$id[[2]] <- duplicate$base$id[[1]]
  expect_error(
    as_svrepdesign(duplicate, data),
    class = "wf_error_schema"
  )

  collision <- data
  collision$.wf_weight <- 1
  expect_error(
    as_svrepdesign(setup$replicates, collision),
    class = "wf_error_schema"
  )

  malformed <- setup$replicates
  malformed$replicates <- malformed$replicates[-1, , drop = FALSE]
  expect_error(
    as_svrepdesign(malformed, data),
    class = "wf_error_internal"
  )

  malformed_method <- setup$replicates
  malformed_method$method <- "unknown"
  expect_error(
    as_svrepdesign(malformed_method, data),
    class = "wf_error_internal"
  )

  for (argument in c(
    "variables", "repweights", "weights", "type", "scale", "rscales",
    "combined.weights", "mse", "rho"
  )) {
    args <- list(r = setup$replicates, data = data)
    args[[argument]] <- 1
    expect_error(
      do.call(as_svrepdesign, args),
      class = "wf_error_input"
    )
  }
})

test_that("planning is deterministic and computes no weights", {
  f <- make_safe_workflow_fixture()

  a <- wf_plan_weights(f$design, f$target, f$dims)
  b <- wf_plan_weights(f$design, f$target, f$dims)

  expect_s3_class(a, "wf_weight_plan")
  expect_identical(a$identity, b$identity)
  expect_null(a$weights)
  expect_s3_class(a$precheck, "wf_precheck")
})

test_that("only verified non-demo targets can enter a weight plan", {
  f <- make_safe_workflow_fixture()
  class(f$target) <- "wf_target"

  expect_error(
    wf_plan_weights(f$design, f$target, f$dims),
    class = "wf_error_safety"
  )
})

test_that("a supplied cell plan must match immutable inputs", {
  f <- make_safe_workflow_fixture()
  cells <- wf_plan_cells(
    f$design,
    f$target,
    f$dims,
    min_cell = 1,
    max_weight_ratio = 10
  )

  plan <- wf_plan_weights(
    f$design,
    f$target,
    f$dims,
    min_cell = 1,
    cell_plan = cells
  )

  expect_identical(plan$cell_plan$identity, cells$identity)
  expect_identical(plan$input_identities$design, f$design$identity)
  expect_identical(plan$input_identities$target, f$target$identity)
})

test_that("non-substantive cell-plan timestamps do not change plan identity", {
  f <- make_safe_workflow_fixture()
  cells_a <- wf_plan_cells(
    f$design,
    f$target,
    f$dims,
    min_cell = 1,
    max_weight_ratio = 10
  )
  cells_b <- cells_a
  cells_b$created <- "2099-01-01T00:00:00+0000"

  plan_a <- wf_plan_weights(
    f$design,
    f$target,
    f$dims,
    min_cell = 1,
    cell_plan = cells_a
  )
  plan_b <- wf_plan_weights(
    f$design,
    f$target,
    f$dims,
    min_cell = 1,
    cell_plan = cells_b
  )

  expect_identical(plan_a$identity, plan_b$identity)
})

test_that("wf_poststrat matches joint cell population totals", {
  fixture <- make_poststrat_fixture()

  weights <- .wf_poststrat_engine(
    fixture$sample,
    fixture$target,
    min_cell = 1,
    ladder = fixture$ladder,
    id = "id"
  )

  expect_s3_class(weights, "wf_weights")
  expect_equal(weights$provenance$method, "poststrat")
  expect_true(all(weights$data$weight >= 0))
  expect_equal(sum(weights$data$weight), fixture$target$groups$A$total, tolerance = 1e-8)
  expect_true(all(c("cell_report", "collapse_map") %in% names(weights)))
  expect_equal(max(weights$log$total_dev), 0, tolerance = 1e-8)
})

test_that("wf_poststrat rejects a missing init_weight column with a classed error", {
  fixture <- make_poststrat_fixture()

  expect_error(
    .wf_poststrat_engine(
      fixture$sample,
      fixture$target,
      min_cell = 1,
      ladder = fixture$ladder,
      init_weight = "no_such_column",
      id = "id"
    ),
    class = "wf_error_schema"
  )
})

test_that("wf_poststrat preserves initial weight ratios within resolved cells", {
  fixture <- make_poststrat_fixture()
  fixture$sample$base_w <- seq_len(nrow(fixture$sample))

  weights <- .wf_poststrat_engine(
    fixture$sample,
    fixture$target,
    min_cell = 2,
    ladder = fixture$ladder,
    init_weight = "base_w",
    id = "id"
  )

  merged <- merge(weights$data, fixture$sample[, c("id", "base_w")], by = "id")
  by_cell <- split(merged, merged$resolved_cell)
  checked <- FALSE
  for (d in by_cell) {
    if (nrow(d) >= 2 && all(d$weight > 0)) {
      expect_equal(d$weight[1] / d$weight[2], d$base_w[1] / d$base_w[2], tolerance = 1e-8)
      checked <- TRUE
      break
    }
  }
  expect_true(checked)
})

test_that("wf_poststrat can flag or reject empty cells", {
  fixture <- make_poststrat_fixture()
  fixture$sample <- fixture$sample[fixture$sample$gender == "female", ]

  expect_error(
    suppressWarnings(.wf_poststrat_engine(
      fixture$sample,
      fixture$target,
      min_cell = 1,
      ladder = fixture$ladder,
      empty_cell = "error"
    )),
    class = "wf_error_feasibility"
  )

  flagged <- suppressWarnings(.wf_poststrat_engine(
    fixture$sample,
    fixture$target,
    min_cell = 1,
    ladder = fixture$ladder,
    empty_cell = "flag"
  ))
  expect_lt(sum(flagged$data$weight), fixture$target$groups$A$total)
  expect_gt(flagged$log$total_dev, 0)
})

test_that("wf_diagnose supports poststrat weights", {
  fixture <- make_poststrat_fixture()

  weights <- .wf_poststrat_engine(
    fixture$sample,
    fixture$target,
    min_cell = 1,
    ladder = fixture$ladder,
    id = "id"
  )
  diag <- wf_diagnose(weights, target = fixture$target)

  expect_s3_class(diag, "wf_diagnostics")
  expect_true(all(c("ess", "deff", "verdict") %in% names(diag$table)))
  expect_true(all(diag$table$converged))
  expect_false("margin_maxerr" %in% names(diag$table))
})

make_tidier_weights <- function() {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    tol = 1e-8
  )
  list(
    fixture = fixture,
    weights = weights,
    diagnostics = wf_diagnose(weights, fixture$target)
  )
}

make_tidier_blend_weights <- function(source, outcome) {
  structure(
    list(
      data = data.frame(
        id = paste0(source, "_", 1:4),
        group = "A",
        cell = rep(c("urban", "rural"), each = 2),
        weight = c(1, 2, 1, 2),
        outcome = outcome,
        stringsAsFactors = FALSE
      ),
      log = data.frame(group = "A", n = 4),
      achieved = NULL,
      provenance = list(method = source)
    ),
    class = "wf_weights"
  )
}

make_tidier_blend <- function(outcome = TRUE) {
  online <- make_tidier_blend_weights("online", c(1, 0, 1, 1))
  offline <- make_tidier_blend_weights("offline", c(0, 0, 1, 0))
  wf_blend(
    online,
    offline,
    by_cell = "cell",
    outcome = if (outcome) "outcome" else NULL,
    lambda = "fixed",
    lambda_fixed = 0.5,
    sensitivity = FALSE
  )
}

tidier_refit <- function(data, weights) {
  structure(
    list(
      data = data.frame(
        id = data$id,
        group = "all",
        weight = weights,
        feature = 1 / weights,
        stringsAsFactors = FALSE
      )
    ),
    class = "wf_weights"
  )
}

make_tidier_variance <- function() {
  data <- make_design_data()
  replicates <- wf_replicates(
    data,
    tidier_refit,
    method = "bootstrap",
    R = 20,
    strata = "stratum",
    clusters = "psu",
    id = "id",
    seed = 4
  )
  wf_variance(
    replicates,
    function(weights, data) sum(weights * data$y) / sum(weights),
    data
  )
}

test_that("generics dispatches tidy methods for WFC result classes", {
  testthat::skip_if_not_installed("generics")
  setup <- make_tidier_weights()
  blend <- make_tidier_blend()
  variance <- make_tidier_variance()

  expect_identical(generics::tidy(setup$weights), setup$weights$data)
  expect_identical(
    generics::tidy(setup$diagnostics),
    setup$diagnostics$table
  )
  expect_identical(generics::tidy(blend), blend$estimates)
  expect_identical(generics::tidy(variance), variance$table)

  cell_blend <- make_tidier_blend(outcome = FALSE)
  expect_identical(generics::tidy(cell_blend), cell_blend$cell_weights)
})

test_that("glance.wf_weights returns an overall weighting summary", {
  testthat::skip_if_not_installed("generics")
  setup <- make_tidier_weights()
  glance <- generics::glance(setup$weights)
  weight <- setup$weights$data$weight

  expect_s3_class(glance, "data.frame")
  expect_equal(nrow(glance), 1)
  expect_identical(
    names(glance),
    c("n", "groups", "total_weight", "ess", "deff", "method")
  )
  expect_equal(glance$n, nrow(setup$weights$data))
  expect_equal(glance$groups, length(unique(setup$weights$data$group)))
  expect_equal(glance$total_weight, sum(weight))
  expect_equal(glance$ess, sum(weight)^2 / sum(weight^2))
  expect_equal(glance$deff, 1 + (stats::sd(weight) / mean(weight))^2)
  expect_identical(glance$method, "raking")
})

test_that("diagnostic, blend, and variance glance methods are stable", {
  testthat::skip_if_not_installed("generics")
  setup <- make_tidier_weights()
  blend <- make_tidier_blend()
  variance <- make_tidier_variance()

  diagnostic_glance <- generics::glance(setup$diagnostics)
  expect_identical(
    names(diagnostic_glance),
    c(
      "groups", "ok", "caveat", "failed", "worst_deff", "minimum_ess"
    )
  )
  expect_equal(diagnostic_glance$groups, nrow(setup$diagnostics$table))
  expect_equal(
    diagnostic_glance$worst_deff,
    max(setup$diagnostics$table$deff)
  )

  blend_glance <- generics::glance(blend)
  expect_identical(
    names(blend_glance),
    c(
      "cells", "groups", "lambda_min", "lambda_mean", "lambda_max",
      "trimmed_lambda_count", "one_source_cell_count"
    )
  )
  expect_equal(blend_glance$cells, nrow(blend$lambda))
  expect_equal(blend_glance$lambda_mean, mean(blend$lambda$lambda))

  variance_glance <- generics::glance(variance)
  expect_identical(
    names(variance_glance),
    c("method", "replicates", "level", "ci", "quantities", "maximum_se")
  )
  expect_equal(variance_glance$replicates, variance$provenance$R)
  expect_equal(variance_glance$maximum_se, max(variance$table$se))
})

test_that("augment.wf_weights joins by ID without mutating inputs", {
  testthat::skip_if_not_installed("generics")
  setup <- make_tidier_weights()
  data <- setup$fixture$sample[rev(seq_len(nrow(setup$fixture$sample))), ]
  original_data <- data
  original_weights <- setup$weights

  augmented <- generics::augment(setup$weights, data = data, id = "id")
  order <- match(data$id, setup$weights$data$id)

  expect_identical(
    names(augmented),
    c(names(data), ".weight", ".feature")
  )
  expect_equal(augmented$.weight, setup$weights$data$weight[order])
  expect_equal(augmented$.feature, setup$weights$data$feature[order])
  expect_identical(augmented$id, data$id)
  expect_identical(data, original_data)
  expect_identical(setup$weights, original_weights)
})

test_that("augment.wf_weights enforces exact IDs and output columns", {
  testthat::skip_if_not_installed("generics")
  setup <- make_tidier_weights()
  data <- setup$fixture$sample

  expect_error(
    generics::augment(setup$weights, data = list()),
    class = "wf_error_input"
  )
  expect_error(
    generics::augment(setup$weights, data = data, id = "missing"),
    class = "wf_error_schema"
  )
  expect_error(
    generics::augment(setup$weights, data = data[-1, ]),
    class = "wf_error_schema"
  )

  duplicate <- data
  duplicate$id[[2]] <- duplicate$id[[1]]
  expect_error(
    generics::augment(setup$weights, data = duplicate),
    class = "wf_error_schema"
  )

  collision <- data
  collision$.weight <- 1
  expect_error(
    generics::augment(setup$weights, data = collision),
    class = "wf_error_schema"
  )
  collision <- data
  collision$.feature <- 1
  expect_error(
    generics::augment(setup$weights, data = collision),
    class = "wf_error_schema"
  )
})

make_auto_trim_fixture <- function() {
  fixture <- make_weightflow_fixture()
  fixture$sample$base_weight <- 1
  fixture$sample$base_weight[1] <- 50
  fixture
}

test_that("wf_auto_trim recommends no trimming when the baseline meets criteria", {
  fixture <- make_auto_trim_fixture()

  out <- .wf_auto_trim_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    caps = c(2, 4),
    max_deff = 10,
    max_residual = 1,
    init_weight = "base_weight"
  )

  expect_s3_class(out, "wf_auto_trim")
  expect_identical(out$recommended_cap, Inf)
  expect_equal(out$frontier$cap, c(2, 4, Inf))
  expect_true(all(c(
    "feasible", "worst_deff", "worst_residual", "warning_count",
    "error_class", "error_message"
  ) %in% names(out$frontier)))
})

test_that("wf_auto_trim selects the loosest finite cap meeting both criteria", {
  fixture <- make_auto_trim_fixture()

  out <- .wf_auto_trim_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    caps = c(2, 3, 4),
    max_deff = 2.1,
    max_residual = 0.04,
    init_weight = "base_weight",
    tol = 1e-8
  )

  expect_equal(out$recommended_cap, 3)
  expect_true(out$frontier$meets_criteria[out$frontier$cap == 3])
  expect_false(out$frontier$meets_criteria[out$frontier$cap == 4])
  expect_false(out$frontier$meets_criteria[is.infinite(out$frontier$cap)])
})

test_that("wf_auto_trim returns NA when no candidate meets the criteria", {
  fixture <- make_auto_trim_fixture()

  out <- .wf_auto_trim_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    caps = c(2, 3),
    max_deff = 1,
    max_residual = 0,
    init_weight = "base_weight"
  )

  expect_true(is.na(out$recommended_cap))
  expect_false(any(out$frontier$meets_criteria))
})

test_that("wf_auto_trim records candidate failures without aborting the sweep", {
  fixture <- make_auto_trim_fixture()

  out <- .wf_auto_trim_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    caps = c(0.5, 2),
    max_deff = 10,
    max_residual = 1,
    init_weight = "base_weight"
  )

  failed <- out$frontier[out$frontier$cap == 0.5, ]
  expect_false(failed$feasible)
  expect_match(failed$error_class, "wf_error_feasibility")
  expect_true(nzchar(failed$error_message))
  expect_true(out$frontier$feasible[out$frontier$cap == 2])
})

test_that("wf_auto_trim validates its controls and owns the trim argument", {
  fixture <- make_weightflow_fixture()

  expect_error(
    .wf_auto_trim_engine(fixture$sample, fixture$target, caps = c(2, NA)),
    class = "wf_error_input"
  )
  expect_error(
    .wf_auto_trim_engine(fixture$sample, fixture$target, caps = c(2, 2)),
    class = "wf_error_input"
  )
  expect_error(
    .wf_auto_trim_engine(fixture$sample, fixture$target, lo = 0),
    class = "wf_error_input"
  )
  expect_error(
    .wf_auto_trim_engine(fixture$sample, fixture$target, max_deff = 0),
    class = "wf_error_input"
  )
  expect_error(
    .wf_auto_trim_engine(fixture$sample, fixture$target, trim = c(0.1, 4)),
    class = "wf_error_input"
  )
})

test_that("wf_auto_trim prints its recommendation and frontier", {
  fixture <- make_weightflow_fixture()
  out <- .wf_auto_trim_engine(
    fixture$sample,
    fixture$target,
    caps = c(2, 4),
    max_deff = 10,
    max_residual = 1
  )

  expect_output(print(out), "recommended cap")
  expect_output(print(out), "worst_deff")
})

test_that("wf_rake refuses blocked prechecks", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$gender[1] <- "other"

  expect_error(
    .wf_rake_engine(sample, fixture$target, id = "id"),
    class = "wf_error_feasibility"
  )
})

test_that("wf_rake returns positive weights matching target margins", {
  fixture <- make_weightflow_fixture()

  weights <- .wf_rake_engine(fixture$sample, fixture$target, id = "id", tol = 1e-8)

  expect_s3_class(weights, "wf_weights")
  expect_true(all(weights$data$weight > 0))
  expect_equal(sum(weights$data$weight[weights$data$group == "A"]), 200, tolerance = 1e-6)
  expect_equal(sum(weights$data$weight[weights$data$group == "B"]), 200, tolerance = 1e-6)

  for (group in names(fixture$target$groups)) {
    for (dim_name in fixture$target$dims) {
      expect_equal(
        weights$achieved[[group]][[dim_name]],
        fixture$target$groups[[group]]$margins[[dim_name]],
        tolerance = 1e-6
      )
    }
  }
})

test_that("wf_rake raises a classed convergence error when IPF cannot finish", {
  fixture <- make_weightflow_fixture()

  err <- expect_error(
    .wf_rake_engine(fixture$sample, fixture$target, id = "id", max_iter = 0),
    class = "wf_error_convergence"
  )
  expect_true(!is.null(err$data$group))
  expect_true(!is.null(err$data$worst_dim))
})

test_that("wf_rake records the installed package version in provenance", {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(fixture$sample, fixture$target, id = "id")

  expect_identical(
    weights$provenance$package_version,
    as.character(utils::packageVersion("WFC"))
  )
})

test_that("wf_diagnose reports diagnostics and margin error", {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(fixture$sample, fixture$target, id = "id", tol = 1e-8)

  diag <- wf_diagnose(weights, target = fixture$target)

  expect_s3_class(diag, "wf_diagnostics")
  expect_true(all(c("ess", "deff", "verdict", "margin_maxerr") %in% names(diag$table)))
  expect_true(all(diag$table$margin_maxerr <= 1e-4))
})

test_that("raking print and diagnose remain compatible with poststrat changes", {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(fixture$sample, fixture$target, id = "id", tol = 1e-8)

  expect_output(print(weights), "<wf_weights>")
  diag <- wf_diagnose(weights, target = fixture$target)
  expect_true(all(c("iterations", "converged", "trimmed") %in% names(diag$table)))
  expect_true(all(diag$table$converged))
})

trivial_refit <- function(data, weights) {
  structure(list(
    data = data.frame(id = data$id, group = "all",
                      weight = weights, feature = 1 / weights,
                      stringsAsFactors = FALSE)
  ), class = "wf_weights")
}

wmean_y <- function(weights, data) sum(weights * data$y) / sum(weights)

test_that("wf_variance reproduces the unified formula on a bootstrap fixture", {
  d <- make_design_data()
  rep_w <- wf_replicates(d, trivial_refit, method = "bootstrap", R = 100,
                         strata = "stratum", clusters = "psu", id = "id",
                         seed = 3)
  out <- wf_variance(rep_w, wmean_y, d)

  theta <- wmean_y(rep_w$base$weight, d)
  tr <- vapply(seq_len(ncol(rep_w$replicates)),
               function(r) wmean_y(rep_w$replicates[, r], d), numeric(1))
  expected_var <- rep_w$scale * sum(rep_w$rscales * (tr - theta)^2)

  expect_equal(out$table$estimate, theta)
  expect_equal(out$table$variance, expected_var)
  expect_equal(out$table$se, sqrt(expected_var))
})

test_that("wf_variance matches a hand-computed jackknife variance", {
  d <- make_design_data()
  rep_w <- wf_replicates(d, trivial_refit, method = "jackknife",
                         strata = "stratum", clusters = "psu", id = "id")
  out <- wf_variance(rep_w, wmean_y, d)

  theta <- wmean_y(rep_w$base$weight, d)
  tr <- vapply(seq_len(ncol(rep_w$replicates)),
               function(r) wmean_y(rep_w$replicates[, r], d), numeric(1))
  expected_var <- sum(rep_w$rscales * (tr - theta)^2)  # scale = 1
  expect_equal(out$table$variance, expected_var)
})

test_that("wf_variance supports vector (subgroup) estimators", {
  d <- make_design_data()
  rep_w <- wf_replicates(d, trivial_refit, method = "bootstrap", R = 50,
                         strata = "stratum", clusters = "psu", id = "id",
                         seed = 5)
  by_stratum <- function(weights, data) {
    c(A = sum(weights[data$stratum == "A"] * data$y[data$stratum == "A"]) /
          sum(weights[data$stratum == "A"]),
      B = sum(weights[data$stratum == "B"] * data$y[data$stratum == "B"]) /
          sum(weights[data$stratum == "B"]))
  }
  out <- wf_variance(rep_w, by_stratum, d)
  expect_equal(nrow(out$table), 2)
  expect_equal(out$table$quantity, c("A", "B"))
})

test_that("wf_variance builds normal and percentile CIs", {
  d <- make_design_data()
  rep_w <- wf_replicates(d, trivial_refit, method = "bootstrap", R = 200,
                         strata = "stratum", clusters = "psu", id = "id",
                         seed = 9)
  normal <- wf_variance(rep_w, wmean_y, d, ci = "normal")
  pctile <- wf_variance(rep_w, wmean_y, d, ci = "percentile")

  expect_true(normal$table$ci_lower < normal$table$estimate)
  expect_true(pctile$table$ci_upper > pctile$table$estimate)
  expect_equal(as.data.frame(normal), normal$table)
})

test_that("wf_variance rejects percentile CI for non-bootstrap replicates", {
  d <- make_design_data()
  rep_w <- wf_replicates(d, trivial_refit, method = "jackknife",
                         strata = "stratum", clusters = "psu", id = "id")
  expect_error(wf_variance(rep_w, wmean_y, d, ci = "percentile"),
               class = "wf_error_input")
})

test_that("wf_variance validates its inputs", {
  d <- make_design_data()
  rep_w <- wf_replicates(d, trivial_refit, method = "bootstrap", R = 10,
                         strata = "stratum", clusters = "psu", id = "id",
                         seed = 1)
  expect_error(wf_variance(list(), wmean_y, d), class = "wf_error_input")
  expect_error(wf_variance(rep_w, "notfun", d), class = "wf_error_input")
  expect_error(wf_variance(rep_w, wmean_y, d, level = 1.5),
               class = "wf_error_input")
})

test_that("wf_replicates + wf_variance run through a real raking refit", {
  fixture <- make_weightflow_fixture()
  d <- fixture$sample
  d$y <- as.numeric(d$age == "young")

  refit <- function(data, weights) {
    data$.bw <- weights
    .wf_rake_engine(data, fixture$target, id = "id", init_weight = ".bw",
            precheck = FALSE)
  }

  # Delete-one jackknife on this small fixture: every marginal category keeps
  # >= 3 units per replicate, so each re-raking stays feasible. (A bootstrap on
  # a fixture this sparse can empty a calibration cell and is infeasible by
  # design.)
  rep_w <- wf_replicates(d, refit, method = "jackknife", id = "id")
  out <- wf_variance(rep_w, function(w, data) sum(w * data$y) / sum(w), d)

  expect_s3_class(out, "wf_variance_result")
  expect_true(is.finite(out$table$se))
  expect_true(out$table$se >= 0)
  expect_true(out$table$ci_lower <= out$table$estimate)
})

test_that("impact compares estimates without changing locked weights", {
  f <- make_safe_workflow_fixture(with_outcomes = TRUE)
  w <- make_locked_safe_weights(f)
  before <- w

  x <- wf_assess_impact(w, f$analysis, "id", c("score", "approved"))

  expect_s3_class(x, "wf_impact")
  expect_true(all(
    c("outcome", "level", "unweighted", "weighted", "difference") %in%
      names(x$summary)
  ))
  expect_true(all(c("kish_ess", "se") %in% names(x$summary)))
  expect_identical(w, before)
  expect_identical(x$weight_identity, w$identity)
})

test_that("categorical impact uses declared factor levels", {
  f <- make_safe_workflow_fixture(with_outcomes = TRUE)
  f$analysis$binary <- factor(
    rep(c("no", "yes"), length.out = nrow(f$analysis)),
    levels = c("no", "yes")
  )
  f$analysis$region <- factor(
    rep(c("north", "south", "west"), length.out = nrow(f$analysis)),
    levels = c("north", "south", "west")
  )
  w <- make_locked_safe_weights(f)

  x <- wf_assess_impact(w, f$analysis, "id", c("binary", "region"))

  expect_identical(x$summary$level[x$summary$outcome == "binary"], "yes")
  expect_identical(
    x$summary$level[x$summary$outcome == "region"],
    c("north", "south", "west")
  )
})

test_that("impact rejects ambiguous or structured outcome columns", {
  f <- make_safe_workflow_fixture(with_outcomes = TRUE)
  f$analysis$when <- as.Date("2026-01-01") + seq_len(nrow(f$analysis))
  f$analysis$items <- I(rep(list(1:2), nrow(f$analysis)))
  w <- make_locked_safe_weights(f)

  expect_error(
    wf_assess_impact(w, f$analysis, "id", "when"),
    class = "wf_error_safety"
  )
  expect_error(
    wf_assess_impact(w, f$analysis, "id", "items"),
    class = "wf_error_safety"
  )
})

test_that("changed outcomes alter impact but never locked weights", {
  f <- make_safe_workflow_fixture(with_outcomes = TRUE)
  w <- make_locked_safe_weights(f)
  before <- w
  first <- wf_assess_impact(w, f$analysis, "id", "score")
  f$analysis$score <- rev(f$analysis$score)

  second <- wf_assess_impact(w, f$analysis, "id", "score")

  expect_false(identical(first$summary, second$summary))
  expect_identical(first$weight_identity, second$weight_identity)
  expect_identical(w, before)
})

test_that("population target can retain joint cells without changing margins", {
  fixture <- make_weightflow_fixture()

  target <- wf_target_population(
    pop = fixture$pop,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    dims = fixture$dims,
    by = "province",
    keep_joint = TRUE
  )

  expect_s3_class(target, "wf_target")
  expect_named(target$joint, c("A", "B"))
  expect_equal(names(target$joint$A), c("gender", "age", "pop"))
  expect_equal(sum(target$joint$A$pop), target$groups$A$total)
  expect_equal(sum(target$joint$B$pop), target$groups$B$total)
  expect_equal(
    unname(tapply(target$joint$A$pop, target$joint$A$gender, sum)["female"]),
    unname(target$groups$A$margins$gender["female"])
  )
})

test_that("population target omits joint cells by default", {
  fixture <- make_weightflow_fixture()

  target <- wf_target_population(
    pop = fixture$pop,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    dims = fixture$dims,
    by = "province"
  )

  expect_null(target$joint)
})

test_that("joint cells follow target scaling", {
  fixture <- make_weightflow_fixture()

  target <- wf_target_population(
    pop = fixture$pop,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    dims = fixture$dims,
    by = "province",
    scale = "custom",
    totals = c(A = 20, B = 30),
    keep_joint = TRUE
  )

  expect_equal(sum(target$joint$A$pop), 20)
  expect_equal(sum(target$joint$B$pop), 30)
  expect_equal(target$groups$A$total, 20)
  expect_equal(target$groups$B$total, 30)
})

# ============================================================================
# WFC Extension 3 -- reference implementation, part 1 of 3 (releases 0.10-0.11)
# Usability foundations and the guided workflow.
#
# Prototype code: source AFTER library(WFC). It calls the installed public API
# (wf_precheck, wf_rake, wf_diagnose, wf_suggest_collapse, wf_apply_collapse,
# wf_target_population) and is NOT itself part of the installed API.
#
# Contents:
#   1. Output localization layer (.wf_lang / .wf_tr)          [design §4.2]
#   2. wf_auto_trim()  -- bias-variance frontier trim sweep    [design §3.3]
#   3. wf_suggest_ladder() -- collapse-ladder drafting         [design §3.3]
#   4. wf_report()     -- dual-audience, bilingual dossier     [design §3.1]
#   5. wf_autoweigh()  -- guided workflow with decision ledger [design §4.1]
#   6. plot() methods                                          [design §3.2]
# ============================================================================

# ----------------------------------------------------------------------------
# 1. Output localization -----------------------------------------------------
# Messages catalogs; production version reads inst/i18n/*.dcf. Rendering only:
# condition classes and payloads are never localized (design P9).
# ----------------------------------------------------------------------------

.wf_i18n <- new.env(parent = emptyenv())
.wf_i18n$catalog <- list(
  en = c(
    rpt_title        = "Weighting quality report",
    rpt_group        = "Group",
    rpt_verdict_g    = "publishable separately",
    rpt_verdict_y    = "usable with caveats; prefer aggregated publication",
    rpt_verdict_r    = "do not publish separately; merge with a neighbour",
    rpt_ess          = "%s effective respondents out of %s (design effect %.1f)",
    rpt_trimmed      = "%d respondent weight(s) were capped to keep results stable",
    rpt_margin       = "largest remaining margin deviation: %.2f%%",
    aw_start         = "Starting guided weighting: %d rows, %d group(s), %d dimension(s).",
    aw_precheck_ok   = "Pre-check passed: sample and target are compatible.",
    aw_precheck_bad  = "Pre-check found %d blocking issue(s); trying declared remediations.",
    aw_collapse      = "Applied declared merge for dimension '%s' (issue: %s in group %s).",
    aw_no_fix        = "No pre-declared ladder can fix: %s. Declare a collapse ladder for this dimension in wf_dims() and re-run.",
    aw_trim          = "Recommended weight cap: %.1fx the average weight (keeps design effect below %.1f).",
    aw_done          = "Weighting finished: %d weights produced; overall verdict: %s.",
    aw_ask           = "Apply this remediation? (yes/no): "
  ),
  zh_CN = c(
    rpt_title        = "加权质量报告",
    rpt_group        = "群组",
    rpt_verdict_g    = "可单独发布",
    rpt_verdict_y    = "可用但需谨慎，建议合并发布",
    rpt_verdict_r    = "不建议单独发布，请与相邻群组合并",
    rpt_ess          = "有效样本量 %s / %s（设计效应 %.1f）",
    rpt_trimmed      = "为保持结果稳定，%d 个权重被封顶",
    rpt_margin       = "剩余最大边际偏差：%.2f%%",
    aw_start         = "开始引导式加权：%d 行，%d 个群组，%d 个维度。",
    aw_precheck_ok   = "预检查通过：样本与目标兼容。",
    aw_precheck_bad  = "预检查发现 %d 个阻塞性问题，尝试已声明的整改方案。",
    aw_collapse      = "已对维度 '%s' 应用声明的合并（问题：%s，群组 %s）。",
    aw_no_fix        = "没有预声明的阶梯能解决：%s。请在 wf_dims() 中为该维度声明折叠阶梯后重试。",
    aw_trim          = "建议权重上限：平均权重的 %.1f 倍（使设计效应低于 %.1f）。",
    aw_done          = "加权完成：产出 %d 个权重；总体结论：%s。",
    aw_ask           = "是否应用该整改？(yes/no)："
  )
)

.wf_lang <- function(lang = NULL) {
  if (!is.null(lang)) return(lang)
  opt <- getOption("wfc.lang", NULL)
  if (!is.null(opt)) return(opt)
  loc <- tolower(paste(Sys.getenv("LANGUAGE"), Sys.getenv("LANG"), Sys.getlocale("LC_CTYPE")))
  if (grepl("zh", loc, fixed = TRUE)) "zh_CN" else "en"
}

.wf_tr <- function(key, ..., lang = NULL) {
  lg <- .wf_lang(lang)
  cat <- .wf_i18n$catalog[[lg]]
  if (is.null(cat) || is.na(cat[key])) cat <- .wf_i18n$catalog$en
  sprintf(cat[[key]], ...)
}

# ----------------------------------------------------------------------------
# 2. wf_auto_trim() -----------------------------------------------------------
# Sweeps candidate weight caps; for each cap reports the worst-group design
# effect and worst residual margin error, then recommends the LOOSEST cap that
# meets both stated criteria. Recommends only -- the analyst sets the bound.
# ----------------------------------------------------------------------------

wf_auto_trim <- function(sample, target, id = NULL,
                         caps = c(2, 3, 4, 5, 6, 8, 10, 12),
                         lo = 0.05, max_deff = 6, max_residual = 0.02, ...) {
  one <- function(cap) {
    trim <- if (is.finite(cap)) c(lo, cap) else NULL
    w <- tryCatch(wf_rake(sample, target, id = id, trim = trim, ...),
                  error = function(e) NULL)
    if (is.null(w)) {
      return(data.frame(cap = cap, feasible = FALSE,
                        worst_deff = NA_real_, worst_residual = NA_real_))
    }
    d <- wf_diagnose(w, target = target)$table
    data.frame(
      cap = cap, feasible = TRUE,
      worst_deff = max(d$deff),
      worst_residual = if ("margin_maxerr" %in% names(d)) max(d$margin_maxerr) else NA_real_
    )
  }
  frontier <- do.call(rbind, lapply(c(sort(caps), Inf), one))
  ok <- frontier$feasible &
    frontier$worst_deff <= max_deff &
    (is.na(frontier$worst_residual) | frontier$worst_residual <= max_residual)
  finite_ok <- ok & is.finite(frontier$cap)
  recommended <- if (any(ok & !is.finite(frontier$cap))) {
    Inf                                   # no trimming needed at all
  } else if (any(finite_ok)) {
    max(frontier$cap[finite_ok])          # loosest cap meeting the criteria
  } else {
    NA_real_                              # nothing meets the criteria
  }
  structure(
    list(frontier = frontier, recommended_cap = recommended,
         criteria = list(max_deff = max_deff, max_residual = max_residual, lo = lo)),
    class = "wf_auto_trim"
  )
}

print.wf_auto_trim <- function(x, ...) {
  cat("<wf_auto_trim>  recommended cap:",
      if (is.na(x$recommended_cap)) "none meets the criteria"
      else if (!is.finite(x$recommended_cap)) "no trimming needed"
      else sprintf("%.1fx mean weight", x$recommended_cap), "\n")
  print(x$frontier, row.names = FALSE)
  invisible(x)
}

# ----------------------------------------------------------------------------
# 3. wf_suggest_ladder() ------------------------------------------------------
# Drafts a collapse ladder from observed sparsity: within each dimension,
# repeatedly merges the thinnest category (worst group counts) into its
# neighbour in declared level order until every merged category reaches
# min_cell. One ladder level per dimension, ordered by ascending affected
# sample share (least information loss first). Output is a DRAFT for review.
# ----------------------------------------------------------------------------

wf_suggest_ladder <- function(sample, target, dims, min_cell = 5) {
  dvars <- names(dims$vars)
  gkey <- if (is.null(target$by)) rep("_all_", nrow(sample))
          else trimws(as.character(sample[[target$by]]))
  drafts <- list()
  affected <- numeric(0)

  for (d in dvars) {
    v <- trimws(as.character(sample[[d]]))
    lev <- dims$vars[[d]]
    if (is.null(lev)) lev <- sort(unique(v[!is.na(v)]))
    worst <- vapply(lev, function(l) {
      per_group <- tapply(!is.na(v) & v == l, gkey, sum)
      min(as.numeric(per_group))
    }, numeric(1))
    if (all(worst >= min_cell)) next

    parts <- as.list(lev)                 # ordered partition of adjacent levels
    counts <- as.numeric(worst)
    while (length(counts) > 1 && any(counts < min_cell)) {
      i <- which.min(counts)
      j <- if (i == 1) 2
           else if (i == length(counts)) i - 1
           else if (counts[i - 1] <= counts[i + 1]) i - 1 else i + 1
      a <- min(i, j); b <- max(i, j)
      parts[[a]] <- c(parts[[a]], parts[[b]])
      counts[a] <- counts[a] + counts[b]
      parts[[b]] <- NULL
      counts <- counts[-b]
    }
    new_name <- vapply(parts, paste, "", collapse = "+")
    map <- stats::setNames(rep(new_name, lengths(parts)), unlist(parts))
    map <- map[map != names(map)]         # only genuine merges
    if (length(map) == 0) next
    drafts[[d]] <- map
    affected[d] <- mean(v %in% names(map), na.rm = TRUE)
  }

  ord <- names(sort(affected))            # least affected share first
  levels <- lapply(ord, function(d) stats::setNames(list(drafts[[d]]), d))
  names(levels) <- paste0("level", seq_along(ord))
  structure(
    list(levels = levels, affected_share = affected[ord], min_cell = min_cell),
    class = "wf_ladder_draft"
  )
}

#' Materialize a reviewed draft into a validated wf_collapse_ladder.
as_ladder <- function(draft, dims) {
  stopifnot(inherits(draft, "wf_ladder_draft"))
  do.call(wf_collapse_ladder, c(list(dims), draft$levels))
}

print.wf_ladder_draft <- function(x, ...) {
  cat(sprintf("<wf_ladder_draft>  %d level(s); min_cell = %d  (review before use)\n",
              length(x$levels), x$min_cell))
  for (nm in names(x$levels)) {
    d <- names(x$levels[[nm]])
    map <- x$levels[[nm]][[1]]
    cat(sprintf("  %s [%s, %.0f%% of rows affected]: %s\n", nm, d,
                100 * x$affected_share[[d]],
                paste(sprintf("%s->%s", names(map), map), collapse = ", ")))
  }
  invisible(x)
}

# ----------------------------------------------------------------------------
# 4. wf_report() --------------------------------------------------------------
# One structured object, two renderings. audience = "manager" projects the
# analyst dossier onto a traffic-light summary in plain language (P7).
# ----------------------------------------------------------------------------

wf_report <- function(w, target = NULL,
                      audience = c("manager", "analyst"),
                      lang = NULL,
                      output = c("object", "markdown", "html"),
                      file = NULL) {
  audience <- match.arg(audience)
  output <- match.arg(output)
  diag <- wf_diagnose(w, target = target)$table

  light <- ifelse(diag$verdict == "OK", "green",
           ifelse(diag$verdict == "usable with caveats", "yellow", "red"))
  body <- data.frame(diag, light = light, stringsAsFactors = FALSE)

  lines <- c(paste0("# ", .wf_tr("rpt_title", lang = lang)), "")
  for (i in seq_len(nrow(body))) {
    key <- c(green = "rpt_verdict_g", yellow = "rpt_verdict_y", red = "rpt_verdict_r")[body$light[i]]
    lines <- c(lines, sprintf("## %s %s -- %s",
                              .wf_tr("rpt_group", lang = lang),
                              body$group[i], .wf_tr(key, lang = lang)))
    lines <- c(lines, paste0("- ", .wf_tr("rpt_ess",
                             format(body$ess[i], big.mark = ","),
                             format(body$n[i], big.mark = ","),
                             body$deff[i], lang = lang)))
    if (audience == "analyst") {
      lines <- c(lines, sprintf("- min/mean/max weight: %.3g / %.3g / %.3g; max/mean: %.1f",
                                body$min_w[i], body$mean_w[i], body$max_w[i],
                                body$max_over_mean[i]))
    }
    if (!is.na(body$trimmed[i]) && body$trimmed[i] > 0) {
      lines <- c(lines, paste0("- ", .wf_tr("rpt_trimmed", body$trimmed[i], lang = lang)))
    }
    if ("margin_maxerr" %in% names(body) && !is.na(body$margin_maxerr[i])) {
      lines <- c(lines, paste0("- ", .wf_tr("rpt_margin", 100 * body$margin_maxerr[i], lang = lang)))
    }
    lines <- c(lines, "")
  }

  res <- structure(
    list(table = body, markdown = lines, audience = audience,
         lang = .wf_lang(lang), provenance = w$provenance),
    class = "wf_quality_report"
  )
  if (output == "object") return(res)
  text <- if (output == "markdown") paste(lines, collapse = "\n")
          else paste0("<!DOCTYPE html><html><head><meta charset='utf-8'></head><body><pre>\n",
                      paste(lines, collapse = "\n"), "\n</pre></body></html>")
  if (!is.null(file)) { writeLines(text, file, useBytes = TRUE); return(invisible(res)) }
  cat(text, "\n")
  invisible(res)
}

print.wf_quality_report <- function(x, ...) {
  cat(paste(x$markdown, collapse = "\n"), "\n")
  invisible(x)
}

# ----------------------------------------------------------------------------
# 5. wf_autoweigh() -----------------------------------------------------------
# The guided path. One engine, two paths (P8): this function only orchestrates
# the public API and records every automated choice in a decision ledger.
# It can never bypass a blocking precheck (P1): remediation is limited to
# ladders/maps the user declared in wf_dims(); anything else stops with a
# plain-language explanation of what to declare.
# ----------------------------------------------------------------------------

wf_autoweigh <- function(sample, population, dims,
                         key_map = NULL, count = NULL, by = NULL, id = NULL,
                         method = c("auto", "raking", "logit"),
                         bounds = NULL,
                         trim = "auto", max_deff = 6,
                         interactive = base::interactive(),
                         lang = NULL, ...) {
  method <- match.arg(method)
  ledger <- list()
  note <- function(action, detail, artifact = NULL) {
    ledger[[length(ledger) + 1]] <<- list(
      step = length(ledger) + 1L, action = action,
      detail = detail, artifact = artifact, time = Sys.time()
    )
    message(detail)
  }

  # -- target ----------------------------------------------------------------
  target <- if (inherits(population, "wf_target")) {
    population
  } else {
    if (is.null(key_map) || is.null(count)) {
      stop("Provide `key_map` and `count` when `population` is a raw data frame.")
    }
    wf_target_population(population, key_map, count, dims, by = by)
  }
  note("start", .wf_tr("aw_start", nrow(sample), length(target$groups),
                       length(target$dims), lang = lang))

  # -- precheck / remediation loop --------------------------------------------
  cur_sample <- sample
  cur_target <- target
  for (round in 1:2) {
    pc <- wf_precheck(cur_sample, cur_target, id = id)
    if (pc$ok) {
      note("precheck", .wf_tr("aw_precheck_ok", lang = lang), artifact = pc)
      break
    }
    n_err <- sum(pc$issues$severity == "error")
    note("precheck", .wf_tr("aw_precheck_bad", n_err, lang = lang), artifact = pc)
    if (round == 2) {
      bad <- pc$issues[pc$issues$severity == "error", , drop = FALSE]
      stop(.wf_tr("aw_no_fix",
                  paste(unique(stats::na.omit(bad$dim)), collapse = ", "),
                  lang = lang))
    }
    plan <- wf_suggest_collapse(pc, dims)
    if (length(plan$maps) == 0) {
      bad <- pc$issues[pc$issues$severity == "error", , drop = FALSE]
      stop(.wf_tr("aw_no_fix",
                  paste(unique(stats::na.omit(bad$dim)), collapse = ", "),
                  lang = lang))
    }
    if (interactive) {
      print(plan)
      ans <- tolower(trimws(readline(.wf_tr("aw_ask", lang = lang))))
      if (!ans %in% c("y", "yes")) stop("Remediation declined; weighting stopped.")
    }
    out <- wf_apply_collapse(cur_sample, cur_target, plan)
    cur_sample <- out$sample
    cur_target <- out$target
    for (i in seq_len(nrow(plan$actions))) {
      note("collapse",
           .wf_tr("aw_collapse", plan$actions$dim[i], plan$actions$check[i],
                  plan$actions$group[i], lang = lang),
           artifact = plan)
    }
  }

  # -- trimming recommendation -------------------------------------------------
  trim_arg <- NULL
  if (identical(trim, "auto") && method %in% c("auto", "raking")) {
    at <- wf_auto_trim(cur_sample, cur_target, id = id, max_deff = max_deff, ...)
    if (is.finite(at$recommended_cap)) {
      trim_arg <- c(at$criteria$lo, at$recommended_cap)
      note("trim", .wf_tr("aw_trim", at$recommended_cap, max_deff, lang = lang),
           artifact = at)
    }
  } else if (is.numeric(trim) && length(trim) == 2) {
    trim_arg <- trim
  }

  # -- calibrate ---------------------------------------------------------------
  w <- if (method == "logit") {
    wf_calibrate(cur_sample, cur_target, method = "logit",
                 bounds = if (is.null(bounds)) c(0.3, 3) else bounds, id = id, ...)
  } else {
    wf_rake(cur_sample, cur_target, id = id, trim = trim_arg, ...)
  }

  diag <- wf_diagnose(w, target = cur_target)
  report <- wf_report(w, target = cur_target, audience = "manager", lang = lang)
  overall <- if (all(diag$table$verdict == "OK")) .wf_tr("rpt_verdict_g", lang = lang)
             else if (any(diag$table$verdict %in% c("FAILED", "do not publish separately")))
               .wf_tr("rpt_verdict_r", lang = lang)
             else .wf_tr("rpt_verdict_y", lang = lang)
  note("done", .wf_tr("aw_done", nrow(w$data), overall, lang = lang))

  structure(
    list(weights = w, diagnostics = diag, report = report,
         ledger = ledger, sample = cur_sample, target = cur_target),
    class = "wf_autoweigh_result"
  )
}

print.wf_autoweigh_result <- function(x, ...) {
  cat(sprintf("<wf_autoweigh_result>  %d weight(s); %d ledger step(s)\n",
              nrow(x$weights$data), length(x$ledger)))
  for (e in x$ledger) cat(sprintf("  [%d] %-9s %s\n", e$step, e$action, e$detail))
  invisible(x)
}

# ----------------------------------------------------------------------------
# 6. plot() methods (base graphics, zero dependencies) ------------------------
# ----------------------------------------------------------------------------

plot.wf_weights <- function(x, max_groups = 9, ...) {
  groups <- unique(x$data$group)[seq_len(min(9, length(unique(x$data$group))))]
  oldpar <- graphics::par(mfrow = grDevices::n2mfrow(length(groups)))
  for (g in groups) {
    wt <- x$data$weight[x$data$group == g]
    graphics::hist(wt, main = paste("Group", g), xlab = "weight",
                   col = "grey85", border = "white", ...)
    graphics::abline(v = mean(wt), lty = 2)
  }
  graphics::par(oldpar)
  invisible(x)
}

plot.wf_diagnostics <- function(x, ...) {
  t <- x$table[order(x$table$deff), ]
  graphics::dotchart(t$deff, labels = t$group, xlab = "design effect", ...)
  graphics::abline(v = c(3, 10), lty = 3)
  invisible(x)
}

plot.wf_auto_trim <- function(x, ...) {
  f <- x$frontier[x$frontier$feasible & is.finite(x$frontier$cap), ]
  graphics::plot(f$cap, f$worst_deff, type = "b", xlab = "trim cap (x mean)",
                 ylab = "worst design effect", ...)
  graphics::abline(h = x$criteria$max_deff, lty = 3)
  if (is.finite(x$recommended_cap)) graphics::abline(v = x$recommended_cap, lty = 2)
  invisible(x)
}

#' Broom-style projections for WFC results
#'
#' These S3 methods return base data frames and are registered for the generic
#' names supplied by the suggested `generics` package. They do not require or
#' return tibbles, and all programmatic column names remain stable English keys.
#'
#' @param x A supported WFC result object.
#' @param data Analysis data to augment with unit weights.
#' @param id Unit-ID column in `data`, matched to `x$data$id`.
#' @param ... Additional arguments. `augment.wf_weights()` does not accept
#'   additional arguments.
#'
#' @return `tidy()` returns an object's primary detail table. `glance()` returns
#'   a one-row summary. `augment()` returns `data` with `.weight` and `.feature`
#'   appended in data-row order.
#' @name wfc-tidiers
NULL

#' @rdname wfc-tidiers
#' @exportS3Method generics::tidy
tidy.wf_weights <- function(x, ...) {
  x$data
}

#' @rdname wfc-tidiers
#' @exportS3Method generics::tidy
tidy.wf_diagnostics <- function(x, ...) {
  x$table
}

#' @rdname wfc-tidiers
#' @exportS3Method generics::tidy
tidy.wf_blend_result <- function(x, ...) {
  if (is.data.frame(x$estimates) && nrow(x$estimates) > 0) {
    return(x$estimates)
  }
  if (is.data.frame(x$cell_weights)) {
    return(x$cell_weights)
  }
  data.frame()
}

#' @rdname wfc-tidiers
#' @exportS3Method generics::tidy
tidy.wf_variance_result <- function(x, ...) {
  x$table
}

#' @rdname wfc-tidiers
#' @exportS3Method generics::glance
glance.wf_weights <- function(x, ...) {
  weight <- as.numeric(x$data$weight)
  total <- sum(weight)
  sum_squares <- sum(weight^2)
  ess <- if (sum_squares > 0) total^2 / sum_squares else NA_real_
  deff <- if (length(weight) > 1 && mean(weight) > 0) {
    1 + (stats::sd(weight) / mean(weight))^2
  } else {
    NA_real_
  }
  method <- x$provenance$method
  if (is.null(method) || length(method) == 0 || is.na(method[[1]])) {
    method <- "raking"
  }
  data.frame(
    n = nrow(x$data),
    groups = length(unique(x$data$group)),
    total_weight = total,
    ess = ess,
    deff = deff,
    method = as.character(method[[1]]),
    stringsAsFactors = FALSE
  )
}

#' @rdname wfc-tidiers
#' @exportS3Method generics::glance
glance.wf_diagnostics <- function(x, ...) {
  table <- x$table
  verdict <- as.character(table$verdict)
  ok <- sum(verdict == "OK", na.rm = TRUE)
  caveat <- sum(verdict == "usable with caveats", na.rm = TRUE)
  failed <- nrow(table) - ok - caveat
  worst_deff <- if (nrow(table) > 0 && any(is.finite(table$deff))) {
    max(table$deff[is.finite(table$deff)])
  } else {
    NA_real_
  }
  minimum_ess <- if (nrow(table) > 0 && any(is.finite(table$ess))) {
    min(table$ess[is.finite(table$ess)])
  } else {
    NA_real_
  }
  data.frame(
    groups = nrow(table),
    ok = ok,
    caveat = caveat,
    failed = failed,
    worst_deff = worst_deff,
    minimum_ess = minimum_ess,
    stringsAsFactors = FALSE
  )
}

#' @rdname wfc-tidiers
#' @exportS3Method generics::glance
glance.wf_blend_result <- function(x, ...) {
  lambda <- x$lambda
  values <- if (is.data.frame(lambda) && "lambda" %in% names(lambda)) {
    as.numeric(lambda$lambda)
  } else {
    numeric()
  }
  groups <- if (is.data.frame(lambda) && "group" %in% names(lambda)) {
    length(unique(lambda$group))
  } else {
    0L
  }
  lambda_min <- if (length(values) > 0) min(values) else NA_real_
  lambda_mean <- if (length(values) > 0) mean(values) else NA_real_
  lambda_max <- if (length(values) > 0) max(values) else NA_real_
  trimmed <- x$diagnostics$trimmed_lambda_count
  one_source <- x$diagnostics$one_source_cell_count
  data.frame(
    cells = if (is.data.frame(lambda)) nrow(lambda) else 0L,
    groups = groups,
    lambda_min = lambda_min,
    lambda_mean = lambda_mean,
    lambda_max = lambda_max,
    trimmed_lambda_count = if (is.null(trimmed)) NA_integer_ else trimmed,
    one_source_cell_count = if (is.null(one_source)) NA_integer_ else one_source,
    stringsAsFactors = FALSE
  )
}

#' @rdname wfc-tidiers
#' @exportS3Method generics::glance
glance.wf_variance_result <- function(x, ...) {
  table <- x$table
  maximum_se <- if (is.data.frame(table) && nrow(table) > 0 &&
      any(is.finite(table$se))) {
    max(table$se[is.finite(table$se)])
  } else {
    NA_real_
  }
  data.frame(
    method = as.character(x$provenance$method),
    replicates = as.integer(x$provenance$R),
    level = as.numeric(x$provenance$level),
    ci = as.character(x$provenance$ci),
    quantities = if (is.data.frame(table)) nrow(table) else 0L,
    maximum_se = maximum_se,
    stringsAsFactors = FALSE
  )
}

#' @rdname wfc-tidiers
#' @exportS3Method generics::augment
augment.wf_weights <- function(x, data, id = "id", ...) {
  if (!is.data.frame(data) || nrow(data) == 0) {
    wf_abort("`data` must be a non-empty data frame.", "wf_error_input")
  }
  dots <- list(...)
  if (length(dots) > 0) {
    wf_abort(
      "augment.wf_weights() does not accept additional arguments.",
      "wf_error_input",
      list(arguments = names(dots))
    )
  }
  collision <- intersect(c(".weight", ".feature"), names(data))
  if (length(collision) > 0) {
    wf_abort(
      sprintf(
        "`data` already contains reserved augment column(s): %s.",
        paste(collision, collapse = ", ")
      ),
      "wf_error_schema",
      list(columns = collision)
    )
  }
  if (!is.data.frame(x$data) ||
      !all(c("id", "weight", "feature") %in% names(x$data))) {
    wf_abort(
      "`x` is missing required unit ID, weight, or feature data.",
      "wf_error_internal"
    )
  }
  order <- .wf_match_unit_ids(
    data,
    id,
    x$data$id,
    "`x$data$id`",
    "augment.wf_weights()"
  )
  out <- data
  out$.weight <- x$data$weight[order]
  out$.feature <- x$data$feature[order]
  out
}

#' Return the loaded package version for provenance.
#'
#' @keywords internal
#' @noRd
.wf_propensity_package_version <- function() {
  tryCatch(
    as.character(utils::packageVersion("WFC")),
    error = function(e) "0.6.0"
  )
}

#' Standardized mean difference between online and reference for one covariate.
#'
#' @param x_on Numeric online values.
#' @param x_ref Numeric reference values.
#' @param w_on Online pseudo-weights (same length/order as `x_on`).
#' @keywords internal
#' @noRd
.wf_propensity_smd <- function(x_on, x_ref, w_on) {
  sd_pool <- sqrt((stats::var(x_on) + stats::var(x_ref)) / 2)
  if (!is.finite(sd_pool) || sd_pool == 0) sd_pool <- NA_real_
  c(
    unweighted = (mean(x_on) - mean(x_ref)) / sd_pool,
    weighted = (stats::weighted.mean(x_on, w_on) - mean(x_ref)) / sd_pool
  )
}

#' Build the online-vs-reference covariate balance table.
#'
#' @param stacked The stacked online+reference frame.
#' @param is_online Logical index of online rows in `stacked`.
#' @param predictors Predictor names.
#' @param w_on Online pseudo-weights (online-row order).
#' @keywords internal
#' @noRd
.wf_propensity_balance <- function(stacked, is_online, predictors, w_on) {
  online_rows <- stacked[is_online, , drop = FALSE]
  ref_rows <- stacked[!is_online, , drop = FALSE]
  rows <- list()
  for (p in predictors) {
    xo <- online_rows[[p]]
    xr <- ref_rows[[p]]
    if (is.numeric(xo)) {
      s <- .wf_propensity_smd(xo, xr, w_on)
      rows[[length(rows) + 1]] <- data.frame(
        variable = p, level = NA_character_,
        smd_unweighted = unname(s["unweighted"]),
        smd_weighted = unname(s["weighted"]),
        stringsAsFactors = FALSE
      )
    } else {
      levs <- sort(unique(.chr(c(xo, xr))))
      for (lv in levs[-1]) {
        s <- .wf_propensity_smd(
          as.numeric(.chr(xo) == lv),
          as.numeric(.chr(xr) == lv),
          w_on
        )
        rows[[length(rows) + 1]] <- data.frame(
          variable = p, level = lv,
          smd_unweighted = unname(s["unweighted"]),
          smd_weighted = unname(s["weighted"]),
          stringsAsFactors = FALSE
        )
      }
    }
  }
  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

#' Build a propensity target: stacked reference frame and membership model spec.
#'
#' Stacks a self-selected `online` sample and a probability `reference` sample
#' into one frame with a membership indicator, so the online sample's selection
#' propensity can be modelled. No model is fit here; execution happens in
#' [wf_propensity()].
#'
#' @param online Data frame: the self-selected (non-probability) sample.
#' @param reference Data frame: the probability reference sample.
#' @param formula Two-sided membership formula, e.g. `member ~ age + edu`. The
#'   right-hand side names the model predictors; the left-hand side names the
#'   membership indicator the constructor creates (`1` online, `0` reference).
#' @param method Fit backend. Only `"logit"` is executable in this release;
#'   `"rf"` / `"gbm"` are reserved and abort in [wf_propensity()].
#' @param by Optional grouping column present in both frames; the propensity
#'   model is fit within each group.
#' @param id Optional id column in `online`; when `NULL`, online units are
#'   identified by row order.
#' @return A `wf_target_propensity` object.
#' @export
wf_target_propensity <- function(online, reference, formula,
                                 method = c("logit", "rf", "gbm"),
                                 by = NULL, id = NULL) {
  method <- match.arg(method)
  if (!is.data.frame(online) || nrow(online) == 0) {
    wf_abort("`online` must be a non-empty data frame.", "wf_error_input")
  }
  if (!is.data.frame(reference) || nrow(reference) == 0) {
    wf_abort("`reference` must be a non-empty data frame.", "wf_error_input")
  }
  if (!inherits(formula, "formula") || length(formula) != 3) {
    wf_abort(
      "`formula` must be a two-sided formula, e.g. member ~ age + edu.",
      "wf_error_input"
    )
  }
  membership <- all.vars(formula[[2]])
  if (length(membership) != 1) {
    wf_abort(
      "The left-hand side of `formula` must be a single membership name.",
      "wf_error_input"
    )
  }
  predictors <- all.vars(formula[[3]])
  if (length(predictors) == 0) {
    wf_abort(
      "`formula` must name at least one predictor on the right-hand side.",
      "wf_error_input"
    )
  }
  if (membership %in% predictors) {
    wf_abort(
      sprintf("Membership name '%s' collides with a predictor.", membership),
      "wf_error_input", list(membership = membership)
    )
  }
  miss_online <- setdiff(predictors, names(online))
  if (length(miss_online) > 0) {
    wf_abort(
      sprintf("`online` is missing predictor(s): %s",
              paste(miss_online, collapse = ", ")),
      "wf_error_input", list(missing = miss_online)
    )
  }
  miss_ref <- setdiff(predictors, names(reference))
  if (length(miss_ref) > 0) {
    wf_abort(
      sprintf("`reference` is missing predictor(s): %s",
              paste(miss_ref, collapse = ", ")),
      "wf_error_input", list(missing = miss_ref)
    )
  }
  if (!is.null(by)) {
    if (length(by) != 1 || !is.character(by)) {
      wf_abort("`by` must be a single column name.", "wf_error_input")
    }
    if (!by %in% names(online)) {
      wf_abort(sprintf("`online` is missing `by` column '%s'.", by),
               "wf_error_input", list(by = by))
    }
    if (!by %in% names(reference)) {
      wf_abort(sprintf("`reference` is missing `by` column '%s'.", by),
               "wf_error_input", list(by = by))
    }
  }
  if (!is.null(id)) {
    if (length(id) != 1 || !is.character(id)) {
      wf_abort("`id` must be a single column name.", "wf_error_input")
    }
    if (!id %in% names(online)) {
      wf_abort(sprintf("`online` is missing `id` column '%s'.", id),
               "wf_error_input", list(id = id))
    }
  }

  keep <- unique(c(predictors, by))
  online_part <- online[, keep, drop = FALSE]
  online_part[[membership]] <- 1L
  online_part$.wf_source <- "online"
  ref_part <- reference[, keep, drop = FALSE]
  ref_part[[membership]] <- 0L
  ref_part$.wf_source <- "reference"
  stacked <- rbind(online_part, ref_part)

  online_ids <- if (is.null(id)) {
    as.character(seq_len(nrow(online)))
  } else {
    .chr(online[[id]])
  }

  structure(list(
    online = online,
    reference = reference,
    stacked = stacked,
    membership = membership,
    predictors = predictors,
    formula = formula,
    method = method,
    by = by,
    id = id,
    online_ids = online_ids,
    n_online = nrow(online),
    n_reference = nrow(reference),
    provenance = list(
      created = Sys.time(),
      package_version = .wf_propensity_package_version()
    )
  ), class = "wf_target_propensity")
}

#' Correct a non-probability sample by inverse-propensity pseudo-weighting.
#'
#' Fits the membership model declared in a [wf_target_propensity()] object and
#' converts each online unit's fitted membership probability into a pseudo-design
#' weight. The result is a `wf_weights` object suitable as an `init_weight` for
#' [wf_rake()] / [wf_poststrat()] and as a stage in [wf_compose()].
#'
#' @param target A `wf_target_propensity` object.
#' @param weight Pseudo-weight form. Only `"ipw"` is executable in this release;
#'   `"kernel"` / `"matching"` are reserved.
#' @param stabilize Use stabilized IPW (`pi_bar / phat`) to tame extreme weights.
#' @param trim Optional positive scalar: clamp weights above `trim * median(w)`.
#' @return A `wf_propensity_weights` object inheriting from `wf_weights`, with
#'   `$overlap` and `$balance` diagnostics.
#' @export
wf_propensity <- function(target,
                          weight = c("ipw", "kernel", "matching"),
                          stabilize = TRUE, trim = NULL) {
  if (!inherits(target, "wf_target_propensity")) {
    wf_abort("`target` must be a wf_target_propensity object.", "wf_error_input")
  }
  weight <- match.arg(weight)
  if (weight != "ipw") {
    wf_abort(
      sprintf("weight = '%s' is not yet supported; only 'ipw' is implemented in this release.",
              weight),
      "wf_error_input", list(weight = weight)
    )
  }
  if (target$method != "logit") {
    wf_abort(
      sprintf("method = '%s' is not yet supported; only 'logit' is implemented in this release.",
              target$method),
      "wf_error_input", list(method = target$method)
    )
  }
  if (!is.null(trim) &&
      (length(trim) != 1 || !is.finite(trim) || trim <= 0)) {
    wf_abort("`trim` must be a single positive number or NULL.",
             "wf_error_input", list(trim = trim))
  }
  t0 <- Sys.time()

  stacked <- target$stacked
  membership <- target$membership
  by <- target$by
  fml <- stats::reformulate(target$predictors, response = membership)

  na_rows <- Reduce(
    `|`,
    lapply(target$predictors, function(p) is.na(stacked[[p]]))
  )
  if (any(na_rows)) {
    wf_abort(sprintf(
      "%d of %d stacked row(s) (%.1f%%) have NA in membership-model predictor(s). Clean or impute these values before fitting the propensity model.",
      sum(na_rows),
      nrow(stacked),
      100 * sum(na_rows) / nrow(stacked)
    ), "wf_error_input", list(n = sum(na_rows)))
  }

  grp <- if (is.null(by)) rep(".all", nrow(stacked)) else .chr(stacked[[by]])
  is_online <- stacked$.wf_source == "online"

  phat <- rep(NA_real_, nrow(stacked))
  for (g in unique(grp)) {
    sel <- grp == g
    n_on <- sum(sel & is_online)
    n_ref <- sum(sel & !is_online)
    if (n_on == 0 || n_ref == 0) {
      wf_abort(
        sprintf("Group '%s' is missing an entire source (online: %d, reference: %d).",
                g, n_on, n_ref),
        "wf_error_overlap",
        list(group = g, n_online = n_on, n_reference = n_ref)
      )
    }
    fit <- stats::glm(fml, family = stats::binomial(),
                      data = stacked[sel, , drop = FALSE])
    phat[sel] <- stats::fitted(fit)
  }

  p_on <- phat[is_online]
  grp_on <- grp[is_online]

  raw <- 1 / p_on
  if (stabilize) {
    pibar <- tapply(stacked[[membership]], grp, mean)
    raw <- as.numeric(pibar[grp_on]) / p_on
  }

  trimmed <- 0L
  if (!is.null(trim)) {
    cap <- trim * stats::median(raw)
    hits <- raw > cap
    trimmed <- sum(hits)
    raw[hits] <- cap
  }

  w <- raw
  for (g in unique(grp_on)) {
    sel <- grp_on == g
    w[sel] <- w[sel] / mean(w[sel])
  }

  data <- data.frame(
    id = target$online_ids,
    group = grp_on,
    weight = w,
    feature = 1 / w,
    stringsAsFactors = FALSE
  )

  log <- data.frame(
    group = unique(grp_on),
    n = as.integer(table(grp_on)[unique(grp_on)]),
    stringsAsFactors = FALSE
  )

  boundary <- 0.99
  probs <- c(0, 0.01, 0.25, 0.5, 0.75, 0.99, 1)
  p_ref <- phat[!is_online]
  n_boundary <- sum(p_on > boundary)
  overlap <- list(
    threshold = boundary,
    online = stats::quantile(p_on, probs, names = TRUE),
    reference = stats::quantile(p_ref, probs, names = TRUE),
    online_values = unname(p_on),
    reference_values = unname(p_ref),
    n_boundary = n_boundary,
    n_online = length(p_on)
  )
  if (n_boundary > 0) {
    wf_warn(
      sprintf(
        "%d online unit(s) have propensity > %.2f (poor common support; extreme pseudo-weights).",
        n_boundary, boundary
      ),
      "wf_warning_quality",
      list(n_boundary = n_boundary, threshold = boundary)
    )
  }

  balance <- .wf_propensity_balance(stacked, is_online, target$predictors, w)

  structure(list(
    data = data,
    log = log,
    achieved = NULL,
    overlap = overlap,
    balance = balance,
    provenance = list(
      method = "propensity",
      fit_method = target$method,
      weight = weight,
      stabilize = stabilize,
      trim = trim,
      trimmed = trimmed,
      by = by,
      id = target$id,
      predictors = target$predictors,
      assumption = paste(
        "Inverse-propensity correction is unbiased only if the model",
        "covariates capture the full online selection mechanism."
      ),
      created = t0,
      elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = .wf_propensity_package_version()
    )
  ), class = c("wf_propensity_weights", "wf_weights"))
}

open_test_device <- function() {
  path <- tempfile(fileext = ".pdf")
  grDevices::pdf(path)
  path
}

test_plot_language <- function() {
  "en"
}

test_that("plot methods expose explicit language selection", {
  methods <- c(
    "plot.wf_weights",
    "plot.wf_diagnostics",
    "plot.wf_auto_trim",
    "plot.wf_blend_result",
    "plot.wf_propensity_weights"
  )

  for (method in methods) {
    expect_true("lang" %in% names(formals(get(method, envir = asNamespace("WFC")))))
  }
})

make_plot_blend <- function(sensitivity = TRUE) {
  structure(
    list(
      sensitivity = if (sensitivity) data.frame(
        lambda = rep(c(0.3, 0.5, 0.7), 2),
        group = rep(c("A", "__overall__"), each = 3),
        estimate = c(0.4, 0.5, 0.6, 0.45, 0.52, 0.61)
      ) else NULL,
      provenance = list(method = "blend")
    ),
    class = "wf_blend_result"
  )
}

make_plot_propensity <- function() {
  online <- data.frame(x = c(1, 1.4, 1.8, 2.2, 2.6, 3, 3.4, -0.5))
  reference <- data.frame(x = c(-2, -1.6, -1.2, -0.8, -0.4, 0, 0.4, 2.8))
  target <- wf_target_propensity(online, reference, member ~ x)
  suppressWarnings(wf_propensity(target))
}

test_that("plot.wf_weights draws groups and restores graphics parameters", {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(fixture$sample, fixture$target, id = "id")
  weights$provenance$trim <- c(0.5, 3)

  path <- open_test_device()
  on.exit({
    grDevices::dev.off()
    unlink(path)
  }, add = TRUE)
  before <- graphics::par("mfrow")
  expect_invisible(plot(weights, max_groups = 1, lang = test_plot_language()))
  expect_identical(graphics::par("mfrow"), before)
  expect_error(plot(weights, max_groups = 0), class = "wf_error_input")
  expect_error(plot(weights, lang = "fr"), class = "wf_error_input")
})

test_that("plot.wf_diagnostics draws deff and ESS panels", {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(fixture$sample, fixture$target, id = "id")
  diagnostics <- wf_diagnose(weights, fixture$target)

  path <- open_test_device()
  on.exit({
    grDevices::dev.off()
    unlink(path)
  }, add = TRUE)
  before <- graphics::par("mfrow")
  expect_invisible(plot(diagnostics, lang = test_plot_language()))
  expect_identical(graphics::par("mfrow"), before)
})

test_that("plot.wf_auto_trim draws the two-axis recommendation frontier", {
  auto <- structure(
    list(
      frontier = data.frame(
        cap = c(2, 3, Inf),
        feasible = TRUE,
        worst_deff = c(1.5, 2, 3),
        worst_residual = c(0.05, 0.02, 0),
        meets_criteria = c(TRUE, TRUE, FALSE)
      ),
      recommended_cap = 3,
      criteria = list(max_deff = 2.5, max_residual = 0.05, lo = 0.05)
    ),
    class = "wf_auto_trim"
  )

  path <- open_test_device()
  on.exit({
    grDevices::dev.off()
    unlink(path)
  }, add = TRUE)
  expect_invisible(plot(auto, lang = test_plot_language()))

  auto$frontier$feasible <- FALSE
  expect_error(plot(auto), class = "wf_error_input")
})

test_that("plot.wf_blend_result draws sensitivity and rejects absent sweeps", {
  path <- open_test_device()
  on.exit({
    grDevices::dev.off()
    unlink(path)
  }, add = TRUE)
  expect_invisible(plot(make_plot_blend(), lang = test_plot_language()))
  expect_error(
    plot(make_plot_blend(sensitivity = FALSE)),
    class = "wf_error_input"
  )
})

test_that("plot.wf_propensity_weights draws overlap and balance panels", {
  weights <- make_plot_propensity()

  path <- open_test_device()
  on.exit({
    grDevices::dev.off()
    unlink(path)
  }, add = TRUE)
  before <- graphics::par("mfrow")
  expect_invisible(plot(weights, lang = test_plot_language()))
  expect_identical(graphics::par("mfrow"), before)
})

make_prop_frames <- function() {
  online <- data.frame(
    pid = paste0("o", 1:6),
    age = c(20, 25, 30, 35, 40, 45),
    edu = c("hs", "hs", "col", "col", "hs", "col"),
    stringsAsFactors = FALSE
  )
  reference <- data.frame(
    pid = paste0("r", 1:6),
    age = c(22, 33, 44, 55, 60, 28),
    edu = c("col", "col", "hs", "hs", "col", "hs"),
    stringsAsFactors = FALSE
  )
  list(online = online, reference = reference)
}

test_that("wf_target_propensity rejects a one-sided formula", {
  f <- make_prop_frames()
  expect_error(
    wf_target_propensity(f$online, f$reference, ~ age + edu),
    class = "wf_error_input"
  )
})

test_that("wf_target_propensity rejects an empty right-hand side", {
  f <- make_prop_frames()
  expect_error(
    wf_target_propensity(f$online, f$reference, member ~ 1),
    class = "wf_error_input"
  )
})

test_that("wf_target_propensity errors when a predictor is missing", {
  f <- make_prop_frames()
  expect_error(
    wf_target_propensity(f$online, f$reference, member ~ age + income),
    class = "wf_error_input"
  )
})

test_that("wf_target_propensity errors when membership name collides", {
  f <- make_prop_frames()
  expect_error(
    wf_target_propensity(f$online, f$reference, age ~ age + edu),
    class = "wf_error_input"
  )
})

test_that("wf_target_propensity errors on empty frames", {
  f <- make_prop_frames()
  expect_error(
    wf_target_propensity(f$online[0, ], f$reference, member ~ age),
    class = "wf_error_input"
  )
})

test_that("wf_target_propensity builds a membership indicator and keeps online order", {
  f <- make_prop_frames()
  tgt <- wf_target_propensity(f$online, f$reference, member ~ age + edu)

  expect_s3_class(tgt, "wf_target_propensity")
  expect_equal(tgt$membership, "member")
  expect_setequal(tgt$predictors, c("age", "edu"))
  # online rows first, in order, all member == 1
  expect_equal(tgt$stacked$member, c(rep(1L, 6), rep(0L, 6)))
  expect_equal(tgt$stacked$.wf_source, c(rep("online", 6), rep("reference", 6)))
  expect_equal(tgt$stacked$age[1:6], f$online$age)
})

test_that("wf_target_propensity uses row order ids by default and the id column when given", {
  f <- make_prop_frames()
  tgt_default <- wf_target_propensity(f$online, f$reference, member ~ age)
  expect_equal(tgt_default$online_ids, as.character(1:6))

  tgt_id <- wf_target_propensity(f$online, f$reference, member ~ age, id = "pid")
  expect_equal(tgt_id$online_ids, f$online$pid)
})

test_that("wf_target_propensity does not fit a model at construction", {
  f <- make_prop_frames()
  tgt <- wf_target_propensity(f$online, f$reference, member ~ age)
  expect_false("model" %in% names(tgt))
})

# Online sample skews toward high x; reference is centered lower but overlaps,
# so glm fits cleanly (no perfect separation).
make_prop_target <- function() {
  online <- data.frame(
    x = c(1.0, 1.4, 1.8, 2.2, 2.6, 3.0, 3.4, -0.5),
    stringsAsFactors = FALSE
  )
  reference <- data.frame(
    x = c(-2.0, -1.6, -1.2, -0.8, -0.4, 0.0, 0.4, 2.8),
    stringsAsFactors = FALSE
  )
  wf_target_propensity(online, reference, member ~ x)
}

test_that("wf_propensity rejects NA in membership-model predictors", {
  online <- data.frame(x = c(1.0, 1.4, NA, 2.2), stringsAsFactors = FALSE)
  reference <- data.frame(x = c(-2.0, -1.6, -1.2, -0.8), stringsAsFactors = FALSE)
  tgt <- wf_target_propensity(online, reference, member ~ x)

  expect_error(wf_propensity(tgt), class = "wf_error_input")
})

test_that("wf_propensity returns a wf_weights object with the id/group/weight/feature contract", {
  tgt <- make_prop_target()
  w <- suppressWarnings(wf_propensity(tgt))

  expect_s3_class(w, "wf_propensity_weights")
  expect_s3_class(w, "wf_weights")
  expect_named(w$data, c("id", "group", "weight", "feature"))
  expect_equal(nrow(w$data), tgt$n_online)
  expect_equal(w$data$id, as.character(1:8))
  expect_true(all(w$data$weight > 0))
  expect_equal(w$data$feature, 1 / w$data$weight)
  expect_equal(w$provenance$method, "propensity")
})

test_that("wf_propensity ipw weights are proportional to 1/phat and normalized to mean 1", {
  tgt <- make_prop_target()
  w <- suppressWarnings(wf_propensity(tgt, stabilize = FALSE))

  # Independently refit the same model to recover the fitted online propensities.
  fit <- stats::glm(member ~ x, family = stats::binomial(), data = tgt$stacked)
  phat_online <- stats::fitted(fit)[tgt$stacked$.wf_source == "online"]
  expected <- 1 / phat_online
  expected <- unname(expected / mean(expected))

  expect_equal(mean(w$data$weight), 1)
  expect_equal(w$data$weight, expected, tolerance = 1e-8)
})

test_that("wf_propensity stabilized weights differ from raw ipw but stay mean 1", {
  tgt <- make_prop_target()
  raw <- suppressWarnings(wf_propensity(tgt, stabilize = FALSE))
  stab <- suppressWarnings(wf_propensity(tgt, stabilize = TRUE))
  # Single group: stabilization is a constant factor, so mean-1 normalization
  # makes the final vectors equal. Assert both are valid mean-1 vectors.
  expect_equal(mean(stab$data$weight), 1)
  expect_equal(stab$provenance$stabilize, TRUE)
  expect_equal(raw$provenance$stabilize, FALSE)
})

test_that("wf_propensity trims extreme weights and records the count", {
  tgt <- make_prop_target()
  untrimmed <- suppressWarnings(wf_propensity(tgt, stabilize = FALSE))
  trimmed <- suppressWarnings(wf_propensity(tgt, stabilize = FALSE, trim = 1.5))

  expect_gte(trimmed$provenance$trimmed, 1)
  expect_lte(max(trimmed$data$weight), max(untrimmed$data$weight) + 1e-9)
  expect_equal(mean(trimmed$data$weight), 1)
})

test_that("wf_propensity rejects reserved methods, weights and bad trim", {
  tgt <- make_prop_target()
  tgt_rf <- tgt; tgt_rf$method <- "rf"
  expect_error(wf_propensity(tgt_rf), class = "wf_error_input")
  expect_error(suppressWarnings(wf_propensity(tgt, weight = "kernel")),
               class = "wf_error_input")
  expect_error(suppressWarnings(wf_propensity(tgt, trim = -1)),
               class = "wf_error_input")
  expect_error(wf_propensity(list()), class = "wf_error_input")
})

test_that("wf_propensity fits per by-group and normalizes within each group", {
  online <- data.frame(
    x = c(1.0, 1.4, 1.8, 2.2, -0.5, 0.9, 1.3, 3.0),
    region = c("n", "n", "n", "n", "s", "s", "s", "s"),
    stringsAsFactors = FALSE
  )
  reference <- data.frame(
    x = c(-2.0, -1.0, 0.0, 2.8, -1.5, -0.4, 0.4, 2.6),
    region = c("n", "n", "n", "n", "s", "s", "s", "s"),
    stringsAsFactors = FALSE
  )
  tgt <- wf_target_propensity(online, reference, member ~ x, by = "region")
  w <- suppressWarnings(wf_propensity(tgt))

  expect_setequal(unique(w$data$group), c("n", "s"))
  expect_equal(mean(w$data$weight[w$data$group == "n"]), 1)
  expect_equal(mean(w$data$weight[w$data$group == "s"]), 1)
})

test_that("wf_propensity errors when a by-group is missing a source", {
  online <- data.frame(
    x = c(1, 2, 3, 4),
    region = c("n", "n", "s", "s"),
    stringsAsFactors = FALSE
  )
  reference <- data.frame(
    x = c(-1, 0, 1, 2),
    region = c("n", "n", "n", "n"),  # no 's' reference rows
    stringsAsFactors = FALSE
  )
  tgt <- wf_target_propensity(online, reference, member ~ x, by = "region")
  expect_error(suppressWarnings(wf_propensity(tgt)), class = "wf_error_overlap")
})

test_that("wf_propensity attaches an overlap report", {
  tgt <- make_prop_target()
  w <- suppressWarnings(wf_propensity(tgt))

  expect_type(w$overlap, "list")
  expect_type(w$overlap$online_values, "double")
  expect_type(w$overlap$reference_values, "double")
  expect_length(w$overlap$online_values, tgt$n_online)
  expect_true(all(c("threshold", "online", "reference", "n_boundary", "n_online")
                  %in% names(w$overlap)))
  expect_equal(w$overlap$n_online, tgt$n_online)
})

test_that("wf_propensity warns on poor common support", {
  # A strongly separating (but still convergent) predictor drives some online
  # propensities above the 0.99 boundary.
  online <- data.frame(x = c(3, 5, 7, 9, 11, -1), stringsAsFactors = FALSE)
  reference <- data.frame(x = c(-11, -9, -7, -5, -3, 1), stringsAsFactors = FALSE)
  tgt <- wf_target_propensity(online, reference, member ~ x)

  expect_warning(wf_propensity(tgt), class = "wf_warning_quality")

  w <- suppressWarnings(wf_propensity(tgt))
  expect_gte(w$overlap$n_boundary, 1)
})

test_that("wf_propensity balance table reports unweighted and weighted SMDs", {
  tgt <- make_prop_target()
  w <- suppressWarnings(wf_propensity(tgt))

  expect_s3_class(w$balance, "data.frame")
  expect_named(w$balance, c("variable", "level", "smd_unweighted", "smd_weighted"))
  expect_true("x" %in% w$balance$variable)
})

test_that("wf_propensity weighting shrinks the covariate gap", {
  # Online over-represents high x; pseudo-weighting should pull its mean toward
  # the reference, shrinking the standardized mean difference.
  set.seed(1)
  online <- data.frame(x = c(rnorm(40, 1.2, 1), rnorm(10, -1, 1)))
  reference <- data.frame(x = c(rnorm(25, 1, 1), rnorm(25, -1, 1)))
  tgt <- wf_target_propensity(online, reference, member ~ x)
  w <- suppressWarnings(wf_propensity(tgt))

  row <- w$balance[w$balance$variable == "x", ]
  expect_lt(abs(row$smd_weighted), abs(row$smd_unweighted))
})

test_that("wf_propensity expands a factor predictor into per-level balance rows", {
  online <- data.frame(
    x = c(1.0, 1.5, 2.0, 2.5, -0.5),
    g = c("a", "a", "b", "b", "a"),
    stringsAsFactors = FALSE
  )
  reference <- data.frame(
    x = c(-1.0, -0.5, 0.0, 0.5, 1.8),
    g = c("b", "b", "a", "b", "a"),
    stringsAsFactors = FALSE
  )
  tgt <- wf_target_propensity(online, reference, member ~ x + g)
  w <- suppressWarnings(wf_propensity(tgt))

  expect_true("g" %in% w$balance$variable)
  expect_true(any(!is.na(w$balance$level)))
})

test_that("print.wf_weights reports the propensity method", {
  tgt <- make_prop_target()
  w <- suppressWarnings(wf_propensity(tgt))
  expect_output(print(w), "method: propensity")
  expect_output(print(w), "overlap:")
})

test_that("wf_propensity output composes via wf_compose", {
  tgt <- make_prop_target()
  stage1 <- suppressWarnings(wf_propensity(tgt))
  stage2 <- stage1
  stage2$data$weight <- rep(2, nrow(stage2$data))
  stage2$data$feature <- 1 / stage2$data$weight

  composed <- wf_compose(propensity = stage1, adj = stage2)
  expect_s3_class(composed, "wf_weights")
  expect_equal(composed$data$weight, stage1$data$weight * 2, tolerance = 1e-8)
})

test_that("wf_propensity weights feed wf_poststrat as init_weight", {
  # wf_poststrat is the calibration stage that exposes the init_weight seam.
  fixture <- make_poststrat_fixture()
  online <- fixture$sample[, c("gender", "age")]
  reference <- data.frame(
    gender = rep(c("female", "male"), each = 4),
    age = rep(c("young", "old"), times = 4),
    stringsAsFactors = FALSE
  )
  tgt <- wf_target_propensity(online, reference, member ~ gender + age)
  pw <- suppressWarnings(wf_propensity(tgt))

  # propensity weights are in online (== sample) row order
  fixture$sample$pw <- pw$data$weight
  weights <- .wf_poststrat_engine(
    fixture$sample,
    fixture$target,
    min_cell = 2,
    ladder = fixture$ladder,
    init_weight = "pw",
    id = "id"
  )
  expect_s3_class(weights, "wf_weights")
})

#' Plan post-stratification cell resolution
#'
#' Resolves each population joint cell to the finest supported collapse-ladder
#' level without computing weights. Review this plan before running
#' `wf_poststrat()`.
#'
#' @param sample Sample data frame.
#' @param target A `wf_target` object built with `keep_joint = TRUE`.
#' @param min_cell Minimum sample count per resolved cell.
#' @param ladder A `wf_collapse_ladder` object.
#' @param granularity Resolution strategy, either `"adaptive"` or `"province"`.
#' @param empty_cell Empty-cell policy, one of `"redistribute"`, `"flag"`, or
#'   `"error"`.
#' @param id Reserved for future plan-level row diagnostics.
#'
#' @return A `wf_poststrat_plan` object.
#' @export
wf_plan_poststrat <- function(sample, target, min_cell, ladder,
                              granularity = c("adaptive", "province"),
                              empty_cell = c("redistribute", "flag", "error"),
                              id = NULL) {
  granularity <- match.arg(granularity)
  empty_cell <- match.arg(empty_cell)
  .wf_check_poststrat_args(sample, target, min_cell, ladder)

  dvars <- target$dims
  gkey <- .wf_group_keys(sample, target$by)
  plan_rows <- list()
  diag_rows <- list()

  for (g in intersect(names(target$joint), unique(gkey))) {
    sel <- gkey == g
    smat <- as.matrix(sapply(dvars, function(d) .chr(sample[[d]][sel])))
    if (is.null(dim(smat))) {
      smat <- matrix(smat, ncol = length(dvars), dimnames = list(NULL, dvars))
    }
    jdf <- target$joint[[g]]
    res <- .wf_resolve_group(smat, jdf, dvars, ladder, min_cell, granularity, g)
    if (any(res$orphan) && empty_cell == "redistribute") {
      res <- .wf_redistribute(res, min_cell, g)
    }

    plan_rows[[g]] <- data.frame(
      group = g,
      jdf[, dvars, drop = FALSE],
      pop = jdf$pop,
      n_sample = res$supp_final,
      ladder_level = res$level_used,
      resolved_cell = res$resolved,
      orphan = res$orphan,
      stringsAsFactors = FALSE
    )
    diag_rows[[g]] <- data.frame(
      group = g,
      n_cells_raw = nrow(jdf),
      n_cells_resolved = length(unique(res$resolved)),
      granularity_used = res$granularity_used,
      n_at_level0 = sum(res$level_used == 0),
      max_level_used = max(res$level_used),
      n_orphan = sum(res$orphan),
      pop_orphan = sum(jdf$pop[res$orphan]),
      stringsAsFactors = FALSE
    )
  }

  structure(
    list(
      plan = do.call(rbind, plan_rows),
      diagnostics = do.call(rbind, diag_rows),
      settings = list(
        min_cell = min_cell,
        granularity = granularity,
        empty_cell = empty_cell,
        ladder = ladder
      )
    ),
    class = "wf_poststrat_plan"
  )
}

#' Print a post-stratification plan
#'
#' @param x A `wf_poststrat_plan` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_poststrat_plan <- function(x, ...) {
  d <- x$diagnostics
  cat(sprintf(
    "<wf_poststrat_plan>  %d group(s); min_cell = %d\n",
    nrow(d),
    x$settings$min_cell
  ))
  cat(sprintf(
    "  granularity: %d adaptive, %d province\n",
    sum(d$granularity_used == "adaptive"),
    sum(d$granularity_used == "province")
  ))
  cat(sprintf(
    "  orphan cells: %d; orphan population: %s\n",
    sum(d$n_orphan),
    format(round(sum(d$pop_orphan)), big.mark = ",")
  ))
  print(utils::head(d, 12), row.names = FALSE)
  if (nrow(d) > 12) {
    cat(sprintf("  ... %d more groups (see $diagnostics)\n", nrow(d) - 12))
  }
  invisible(x)
}

#' Summarize a post-stratification plan
#'
#' @param object A `wf_poststrat_plan` object.
#' @param ... Unused.
#' @return The plan diagnostics data frame.
#' @export
summary.wf_poststrat_plan <- function(object, ...) {
  object$diagnostics
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----template, eval=FALSE-----------------------------------------------------
# dims <- wf_dims(
#   age_group = c("18-34", "35-54", "55+"),
#   region = c("north", "south")
# )
# 
# wf_target_template("population-margins.csv", dims)

## ----import, eval=FALSE-------------------------------------------------------
# target <- wf_import_target(
#   "population-margins.csv",
#   "population-margins.csv.source.dcf",
#   dims,
#   key_map = c(age_group = "age_group", region = "region"),
#   count = "population_count",
#   production = TRUE
# )

#' Merge plot defaults with user arguments.
#'
#' @param defaults Default argument list.
#' @param dots User argument list.
#' @keywords internal
#' @noRd
.wf_plot_args <- function(defaults, dots) {
  utils::modifyList(defaults, dots)
}

#' Plot calibrated weight distributions
#'
#' Draws one histogram per group, including the group mean and recorded raking
#' trim bounds when available.
#'
#' @param x A `wf_weights` object.
#' @param max_groups Maximum number of groups to plot.
#' @param lang Output language.
#' @param ... Additional arguments passed to [graphics::hist()].
#' @return Invisibly returns `x`.
#' @export
plot.wf_weights <- function(x, max_groups = 9, lang = NULL, ...) {
  language <- .wf_lang(lang)
  if (length(max_groups) != 1 || !is.finite(max_groups) || max_groups < 1 ||
      max_groups != as.integer(max_groups)) {
    wf_abort("`max_groups` must be a positive integer.", "wf_error_input")
  }
  if (!is.data.frame(x$data) || nrow(x$data) == 0 ||
      !all(c("group", "weight") %in% names(x$data))) {
    wf_abort("`x` has no plottable weight data.", "wf_error_input")
  }
  groups <- unique(.chr(x$data$group))
  groups <- groups[seq_len(min(length(groups), as.integer(max_groups)))]
  old <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old), add = TRUE)
  layout <- if (length(groups) <= 3) {
    c(1, length(groups))
  } else {
    grDevices::n2mfrow(length(groups))
  }
  graphics::par(mfrow = layout)
  dots <- list(...)

  for (group in groups) {
    weight <- x$data$weight[.chr(x$data$group) == group]
    args <- .wf_plot_args(
      list(
        x = weight,
        main = .wf_tr("plot_group_title", group, lang = language),
        xlab = .wf_tr("plot_weight", lang = language),
        col = "grey85",
        border = "white"
      ),
      dots
    )
    do.call(graphics::hist, args)
    graphics::abline(v = mean(weight), lty = 2, lwd = 1.5)
    trim <- x$provenance$trim
    if (is.numeric(trim) && length(trim) == 2 && all(is.finite(trim))) {
      graphics::abline(v = mean(weight) * trim, lty = 3, col = "firebrick")
    }
  }
  invisible(x)
}

#' Plot weight diagnostics
#'
#' Draws design effects and effective-sample-size shares by group.
#'
#' @param x A `wf_diagnostics` object.
#' @param lang Output language.
#' @param ... Additional arguments passed to [graphics::dotchart()].
#' @return Invisibly returns `x`.
#' @export
plot.wf_diagnostics <- function(x, lang = NULL, ...) {
  language <- .wf_lang(lang)
  if (!is.data.frame(x$table) || nrow(x$table) == 0 ||
      !all(c("group", "n", "ess", "deff") %in% names(x$table))) {
    wf_abort("`x` has no plottable diagnostic table.", "wf_error_input")
  }
  table <- x$table[is.finite(x$table$deff) & is.finite(x$table$ess), , drop = FALSE]
  if (nrow(table) == 0) {
    wf_abort("No finite diagnostics are available to plot.", "wf_error_input")
  }
  table <- table[order(table$deff), , drop = FALSE]
  old <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old), add = TRUE)
  graphics::par(mfrow = c(1, 2))
  dots <- list(...)

  do.call(
    graphics::dotchart,
    .wf_plot_args(
      list(
        x = table$deff,
        labels = table$group,
        xlab = .wf_tr("plot_design_effect", lang = language)
      ),
      dots
    )
  )
  graphics::abline(v = c(3, 10), lty = 3, col = "grey45")

  do.call(
    graphics::dotchart,
    .wf_plot_args(
      list(
        x = table$ess / table$n,
        labels = table$group,
        xlab = .wf_tr("plot_effective_sample_share", lang = language),
        xlim = c(0, 1)
      ),
      dots
    )
  )
  invisible(x)
}

#' Plot an automatic trim frontier
#'
#' Draws the worst design effect and residual margin error for every feasible
#' finite cap.
#'
#' @param x A `wf_auto_trim` object.
#' @param lang Output language.
#' @param ... Additional arguments passed to the initial plot calls.
#' @return Invisibly returns `x`.
#' @export
plot.wf_auto_trim <- function(x, lang = NULL, ...) {
  language <- .wf_lang(lang)
  frontier <- x$frontier[
    x$frontier$feasible & is.finite(x$frontier$cap),
    ,
    drop = FALSE
  ]
  if (nrow(frontier) == 0) {
    wf_abort("No feasible finite trim candidates are available to plot.", "wf_error_input")
  }
  frontier <- frontier[order(frontier$cap), , drop = FALSE]
  old <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old), add = TRUE)
  graphics::par(mfrow = c(1, 2))
  dots <- list(...)

  do.call(
    graphics::plot,
    .wf_plot_args(
      list(
        x = frontier$cap,
        y = frontier$worst_deff,
        type = "b",
        xlab = .wf_tr("plot_trim_cap", lang = language),
        ylab = .wf_tr("plot_worst_deff", lang = language)
      ),
      dots
    )
  )
  graphics::abline(h = x$criteria$max_deff, lty = 3, col = "grey45")
  if (is.finite(x$recommended_cap)) {
    graphics::abline(v = x$recommended_cap, lty = 2, col = "firebrick")
  }

  do.call(
    graphics::plot,
    .wf_plot_args(
      list(
        x = frontier$cap,
        y = frontier$worst_residual,
        type = "b",
        xlab = .wf_tr("plot_trim_cap", lang = language),
        ylab = .wf_tr("plot_worst_residual", lang = language)
      ),
      dots
    )
  )
  graphics::abline(h = x$criteria$max_residual, lty = 3, col = "grey45")
  if (is.finite(x$recommended_cap)) {
    graphics::abline(v = x$recommended_cap, lty = 2, col = "firebrick")
  }
  invisible(x)
}

#' Plot blend lambda sensitivity
#'
#' @param x A `wf_blend_result` containing sensitivity output.
#' @param lang Output language.
#' @param ... Additional arguments passed to the initial plot.
#' @return Invisibly returns `x`.
#' @export
plot.wf_blend_result <- function(x, lang = NULL, ...) {
  language <- .wf_lang(lang)
  sensitivity <- x$sensitivity
  if (!is.data.frame(sensitivity) || nrow(sensitivity) == 0 ||
      !all(c("lambda", "group", "estimate") %in% names(sensitivity))) {
    wf_abort(
      "No lambda sensitivity is available; create the blend with sensitivity = TRUE.",
      "wf_error_input"
    )
  }
  keep <- is.finite(sensitivity$lambda) & is.finite(sensitivity$estimate)
  sensitivity <- sensitivity[keep, , drop = FALSE]
  if (nrow(sensitivity) == 0) {
    wf_abort("No finite lambda sensitivity values are available.", "wf_error_input")
  }
  groups <- unique(.chr(sensitivity$group))
  colours <- stats::setNames(grDevices::hcl.colors(length(groups), "Dark 3"), groups)
  dots <- list(...)
  do.call(
    graphics::plot,
    .wf_plot_args(
      list(
        x = range(sensitivity$lambda),
        y = range(sensitivity$estimate),
        type = "n",
        xlab = .wf_tr("plot_online_lambda", lang = language),
        ylab = .wf_tr("plot_fused_estimate", lang = language)
      ),
      dots
    )
  )
  for (group in groups) {
    part <- sensitivity[.chr(sensitivity$group) == group, , drop = FALSE]
    part <- part[order(part$lambda), , drop = FALSE]
    graphics::lines(
      part$lambda,
      part$estimate,
      col = colours[[group]],
      lwd = if (group == "__overall__") 3 else 1.5
    )
  }
  if (length(groups) > 1) {
    graphics::legend(
      "topleft",
      legend = groups,
      col = colours[groups],
      lty = 1,
      lwd = ifelse(groups == "__overall__", 3, 1.5),
      bty = "n"
    )
  }
  invisible(x)
}

#' Plot propensity overlap and covariate balance
#'
#' @param x A `wf_propensity_weights` object.
#' @param lang Output language.
#' @param ... Additional arguments passed to the initial plot calls.
#' @return Invisibly returns `x`.
#' @export
plot.wf_propensity_weights <- function(x, lang = NULL, ...) {
  language <- .wf_lang(lang)
  online <- x$overlap$online_values
  reference <- x$overlap$reference_values
  balance <- x$balance
  if (!is.numeric(online) || !is.numeric(reference) ||
      length(online) < 2 || length(reference) < 2) {
    wf_abort("Raw propensity values are unavailable for overlap plotting.", "wf_error_input")
  }
  if (!is.data.frame(balance) || nrow(balance) == 0 ||
      !all(c("variable", "level", "smd_unweighted", "smd_weighted") %in%
           names(balance))) {
    wf_abort("Covariate balance values are unavailable for plotting.", "wf_error_input")
  }
  old <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old), add = TRUE)
  graphics::par(mfrow = c(1, 2))
  dots <- list(...)

  if (length(unique(online)) > 1 && length(unique(reference)) > 1) {
    density_online <- stats::density(online, from = 0, to = 1)
    density_reference <- stats::density(reference, from = 0, to = 1)
    do.call(
      graphics::plot,
      .wf_plot_args(
        list(
          x = density_online,
          xlim = c(0, 1),
          ylim = c(0, max(density_online$y, density_reference$y)),
          main = .wf_tr("plot_propensity_overlap", lang = language),
          xlab = .wf_tr("plot_online_propensity", lang = language),
          ylab = .wf_tr("plot_density", lang = language),
          col = "#0072B2",
          lwd = 2
        ),
        dots
      )
    )
    graphics::lines(density_reference, col = "#D55E00", lwd = 2)
  } else {
    do.call(
      graphics::plot,
      .wf_plot_args(
        list(
          x = c(0, 1),
          y = c(0, 1),
          type = "n",
          main = .wf_tr("plot_propensity_overlap", lang = language),
          xlab = .wf_tr("plot_online_propensity", lang = language),
          ylab = ""
        ),
        dots
      )
    )
    graphics::rug(online, col = "#0072B2")
    graphics::rug(reference, side = 3, col = "#D55E00")
  }
  graphics::legend(
    "topleft",
    legend = c(
      .wf_tr("plot_online", lang = language),
      .wf_tr("plot_reference", lang = language)
    ),
    col = c("#0072B2", "#D55E00"),
    lty = 1,
    lwd = 2,
    bty = "n"
  )

  labels <- ifelse(
    is.na(balance$level),
    balance$variable,
    paste(balance$variable, balance$level, sep = ":")
  )
  unweighted <- abs(balance$smd_unweighted)
  weighted <- abs(balance$smd_weighted)
  limit <- max(c(unweighted, weighted, 0.1), na.rm = TRUE)
  y <- seq_along(labels)
  do.call(
    graphics::plot,
    .wf_plot_args(
      list(
        x = c(0, limit * 1.05),
        y = range(y),
        type = "n",
        yaxt = "n",
        xlab = .wf_tr("plot_absolute_smd", lang = language),
        ylab = "",
        main = .wf_tr("plot_covariate_balance", lang = language)
      ),
      dots
    )
  )
  graphics::axis(2, at = y, labels = labels, las = 1)
  graphics::segments(unweighted, y, weighted, y, col = "grey70")
  graphics::points(unweighted, y, pch = 1, col = "#D55E00")
  graphics::points(weighted, y, pch = 16, col = "#0072B2")
  graphics::abline(v = 0.1, lty = 3, col = "grey45")
  graphics::legend(
    "topright",
    legend = c(
      .wf_tr("plot_unweighted", lang = language),
      .wf_tr("plot_weighted", lang = language)
    ),
    col = c("#D55E00", "#0072B2"),
    pch = c(1, 16),
    bty = "n"
  )
  invisible(x)
}

#' Calculate a SHA-256 identity for a file.
#'
#' @param path Path to an existing file.
#' @return A lowercase SHA-256 string.
#' @keywords internal
#' @noRd
.wf_sha256_file <- function(path) {
  if (!.wf_is_string(path) || !file.exists(path)) {
    .wf_safety_abort(
      "source_file_missing",
      "Source file does not exist.",
      "path"
    )
  }

  digest::digest(file = path, algo = "sha256", serialize = FALSE)
}

#' Calculate a stable SHA-256 identity for an R object.
#'
#' @param x Object to identify.
#' @return A lowercase SHA-256 string.
#' @keywords internal
#' @noRd
.wf_sha256_object <- function(x) {
  digest::digest(
    .wf_sanitize_for_hash(x),
    algo = "sha256",
    serialize = TRUE
  )
}

test_that("safety errors expose stable machine payloads", {
  err <- tryCatch(.wf_safety_abort(
    "target_period_missing", "Reference period is required.",
    field = "reference_period", next_actions = "supply_source_metadata"
  ), error = identity)

  expect_s3_class(err, "wf_error_safety")
  expect_identical(err$data$code, "target_period_missing")
  expect_identical(err$data$severity, "blocking")
  expect_identical(err$data$next_actions, "supply_source_metadata")
})

test_that("identities are stable SHA-256 strings", {
  p <- tempfile()
  writeLines("x", p)

  expect_match(.wf_sha256_file(p), "^[0-9a-f]{64}$")
  expect_identical(
    .wf_sha256_object(list(a = 1)),
    .wf_sha256_object(list(a = 1))
  )
})

# ============================================================================
# weightflow 0.2 -- post-stratification extension (R/poststrat.R)
# Source AFTER weightflow_core.R. Zero hard dependencies.
# Adds: wf_collapse_ladder(), wf_plan_poststrat(), wf_poststrat()
# Discipline: wf_plan_poststrat() -> wf_poststrat() -> wf_diagnose()
# ============================================================================

# ----------------------------------------------------------------------------
# 0. Cell-key helpers --------------------------------------------------------
# ----------------------------------------------------------------------------

# Build a single string key per row from the character dimension columns.
.wf_cell_key <- function(mat, dvars) {
  do.call(paste, c(lapply(dvars, function(d) mat[, d]), sep = "\r"))
}
.wf_split_key <- function(keys, dvars) {
  parts <- do.call(rbind, strsplit(keys, "\r", fixed = TRUE))
  df <- as.data.frame(parts, stringsAsFactors = FALSE)
  names(df) <- dvars
  df
}

# ----------------------------------------------------------------------------
# 1. Collapse ladder ---------------------------------------------------------
# ----------------------------------------------------------------------------

#' Declare an ordered, cumulative, validated collapse ladder.
#' Level 0 is the raw cells (implicit). Each `levelK = list(dim = c(old=new,...))`
#' argument adds one more merge, applied on top of all previous levels.
wf_collapse_ladder <- function(dims, ...) {
  if (!inherits(dims, "wf_dims"))
    wf_abort("`dims` must be a wf_dims object.", "wf_error_input")
  steps <- list(...)
  if (length(steps) > 0 && (is.null(names(steps)) || any(names(steps) == "")))
    wf_abort("Ladder levels must be named (e.g. level1 = list(...)).",
             "wf_error_input")
  dvars <- names(dims$vars)

  # Track the evolving category universe per dimension to validate cumulatively.
  # When dims declares explicit levels we validate against them from level 1;
  # when levels are NULL (unknown), we cannot check level-1 source categories,
  # but we DO seed the universe from level-1 maps so that later levels are
  # validated against categories the ladder itself has produced.
  universe <- lapply(dvars, function(d) dims$vars[[d]])
  names(universe) <- dvars
  for (lv in seq_along(steps)) {
    step <- steps[[lv]]
    bad_d <- setdiff(names(step), dvars)
    if (length(bad_d) > 0)
      wf_abort(sprintf("Ladder level '%s' references unknown dimension(s): %s",
                       names(steps)[lv], paste(bad_d, collapse = ", ")),
               "wf_error_input", list(level = names(steps)[lv], dims = bad_d))
    for (d in names(step)) {
      map <- step[[d]]
      if (is.null(names(map)))
        wf_abort(sprintf("Ladder level '%s', dim '%s': map must be a named vector (old=new).",
                         names(steps)[lv], d), "wf_error_input")
      if (!is.null(universe[[d]])) {
        unknown <- setdiff(names(map), universe[[d]])
        if (length(unknown) > 0)
          wf_abort(sprintf(
            "Ladder level '%s', dim '%s': source categories not present at this level: %s",
            names(steps)[lv], d, paste(unknown, collapse = ", ")),
            "wf_error_input", list(level = names(steps)[lv], dim = d,
                                   unknown = unknown))
        kept <- setdiff(universe[[d]], names(map))
        universe[[d]] <- unique(c(kept, unname(map)))
      } else {
        # unknown starting universe: seed it so later levels are checkable
        universe[[d]] <- unique(unname(map))
      }
    }
  }
  structure(list(dims = dvars, steps = steps, n_levels = length(steps)),
            class = "wf_collapse_ladder")
}

# Apply the cumulative maps for ladder levels 1..lv to a character matrix,
# returning the coarsened matrix. lv = 0 returns the input unchanged.
.wf_apply_ladder <- function(mat, ladder, lv) {
  if (lv == 0) return(mat)
  for (k in seq_len(lv)) {
    step <- ladder$steps[[k]]
    for (d in names(step)) {
      map <- step[[d]]; col <- mat[, d]
      hit <- !is.na(col) & col %in% names(map)
      col[hit] <- map[col[hit]]
      mat[, d] <- col
    }
  }
  mat
}

# ----------------------------------------------------------------------------
# 2. Resolution engine (shared by planner and executor) ----------------------
# ----------------------------------------------------------------------------

# For one group: decide, per raw cell, the lowest ladder level meeting min_cell.
# Returns per-raw-cell resolution + chosen granularity + orphan info.
.wf_resolve_group <- function(smat, jdf, dvars, ladder, min_cell,
                              granularity, group) {
  # raw (level-0) cell keys for sample and for population cells
  jmat <- as.matrix(jdf[, dvars, drop = FALSE])
  raw_pop_key <- .wf_cell_key(jmat, dvars)

  # precompute sample counts at each ladder level, keyed by coarsened cell
  max_lv <- ladder$n_levels
  # coarsened sample keys per level
  s_keys <- lapply(0:max_lv, function(lv)
    .wf_cell_key(.wf_apply_ladder(smat, ladder, lv), dvars))
  s_counts <- lapply(s_keys, function(k) table(k))
  # coarsened population-cell keys per level (maps each raw pop cell -> level key)
  p_keys <- lapply(0:max_lv, function(lv)
    .wf_cell_key(.wf_apply_ladder(jmat, ladder, lv), dvars))

  n_raw <- nrow(jdf)
  level_used <- integer(n_raw)
  resolved   <- character(n_raw)
  gran_used  <- granularity

  support_at <- function(lv, cellkeys) {
    tb <- s_counts[[lv + 1]]
    as.integer(ifelse(cellkeys %in% names(tb), tb[cellkeys], 0L))
  }

  # The resolved partition must be VALID: if two raw cells share an ancestor at
  # level k, they must resolve at the SAME level (otherwise population and sample
  # are aggregated on incompatible boundaries, leaking population). We therefore
  # decide the level of each MERGE GROUP, not each raw cell. Start everyone at
  # level 0; a coarsened cell at its current level is acceptable iff its combined
  # sample support >= min_cell. Under-supported cells push ALL their members up
  # one level; because pushing up can merge previously-separate cells, iterate to
  # a fixed point. Members of any coarsened key always move together, so the
  # result is a genuine partition.
  if (granularity == "adaptive") {
    # Correct adaptive partition via monotone level-raising to a fixed point.
    # Invariant we enforce: for the FINAL levels, every coarsened key contains
    # members all at the same level, and every kept cell has >= min_cell support.
    #
    # Rule 1 (support): a coarsened cell with < min_cell support raises all its
    #   members one level.
    # Rule 2 (partition validity): if raw cells A and B collapse to the same key
    #   at level L, and A has been raised to level L (or higher), then B — which
    #   would merge with A at L — must also be at level >= L. We enforce this by
    #   lifting, at each level L from coarse to fine, any cell whose level-L key
    #   is shared by a cell already at level >= L.
    settle <- rep(0L, n_raw)
    repeat {
      before <- settle
      # Rule 1
      keys <- vapply(seq_len(n_raw),
                     function(i) p_keys[[settle[i] + 1]][i], character(1))
      supp <- vapply(seq_len(n_raw),
                     function(i) support_at(settle[i], keys[i]), integer(1))
      under <- supp < min_cell & settle < max_lv
      if (any(under)) {
        bad <- unique(keys[under])
        settle[keys %in% bad & settle < max_lv] <-
          settle[keys %in% bad & settle < max_lv] + 1L
      }
      # Rule 2: propagate merges upward. For each level L, cells sharing an
      # L-key with a member already at level >= L must be lifted to L.
      for (L in 1:max_lv) {
        kL   <- p_keys[[L + 1]]
        hasL <- tapply(settle >= L, kL, any)      # some merger already at >= L?
        lift <- as.logical(hasL[kL]) & settle < L
        if (any(lift)) settle[lift] <- L
      }
      if (identical(settle, before)) break
    }
    final_keys <- vapply(seq_len(n_raw),
                         function(i) p_keys[[settle[i] + 1]][i], character(1))
    final_supp <- vapply(seq_len(n_raw),
                         function(i) support_at(settle[i], final_keys[i]),
                         integer(1))
    if (any(final_supp < min_cell)) {
      gran_used <- "province"
    } else {
      level_used <- settle
      resolved   <- final_keys
    }
  }

  if (gran_used == "province") {
    chosen <- NA_integer_
    for (lv in 0:max_lv) {
      supp <- support_at(lv, p_keys[[lv + 1]])
      if (all(supp >= min_cell)) { chosen <- lv; break }
    }
    if (is.na(chosen)) chosen <- max_lv
    level_used <- rep(chosen, n_raw)
    resolved   <- p_keys[[chosen + 1]]
  }

  supp_final <- vapply(seq_len(n_raw),
                       function(i) support_at(level_used[i], resolved[i]),
                       integer(1))
  orphan <- supp_final == 0

  list(raw_key = raw_pop_key, level_used = level_used, resolved = resolved,
       supp_final = supp_final, orphan = orphan, granularity_used = gran_used,
       s_keys = s_keys, p_keys = p_keys, jmat = jmat)
}

# For orphan cells, find nearest supported ancestor (one or more levels up)
# and reassign their resolved key to that ancestor. Returns updated resolved +
# a redistribution log.
.wf_redistribute <- function(res, min_cell, group) {
  n_raw <- length(res$resolved)
  redist <- list()
  max_lv <- length(res$s_keys) - 1
  # support of raw cell k at its currently-assigned level
  support_at_final <- function(res, k) {
    tb <- table(res$s_keys[[res$level_used[k] + 1]])
    key <- res$resolved[k]
    if (key %in% names(tb)) as.integer(tb[key]) else 0L
  }
  for (i in which(res$orphan)) {
    moved <- FALSE
    lv0 <- res$level_used[i]
    if (lv0 < max_lv) {
      for (lv in (lv0 + 1):max_lv) {
        anc <- res$p_keys[[lv + 1]][i]
        tb  <- table(res$s_keys[[lv + 1]])
        if (anc %in% names(tb) && tb[anc] > 0) {
          redist[[length(redist) + 1]] <- data.frame(
            group = group, orphan_cell = res$resolved[i],
            receiving_cell = anc, ancestor_level = lv,
            stringsAsFactors = FALSE)
          res$resolved[i]   <- anc
          res$level_used[i] <- lv
          res$orphan[i]     <- FALSE
          moved <- TRUE; break
        }
      }
    }
    if (!moved) {
      # Ladder exhausted: no coarsened ancestor of this cell has support.
      # Province-level fallback (requirement 5, redistribute): fold the orphan
      # population into the province's most-supported resolved cell, so no
      # population is lost. This is the coarsest possible reallocation and is
      # always logged for audit. It implies the strongest exchangeability
      # assumption, flagged to the analyst.
      supported <- which(!res$orphan)
      if (length(supported) == 0)
        wf_abort(sprintf("Group '%s': no resolved cell has any sample support; cannot post-stratify this group at all.",
                         group), "wf_error_feasibility", list(group = group))
      counts <- vapply(supported,
        function(k) support_at_final(res, k), integer(1))
      recv <- res$resolved[supported[which.max(counts)]]
      redist[[length(redist) + 1]] <- data.frame(
        group = group, orphan_cell = res$resolved[i],
        receiving_cell = recv, ancestor_level = -1L,   # -1 = province fallback
        stringsAsFactors = FALSE)
      res$resolved[i] <- recv
      res$orphan[i]   <- FALSE
    }
  }
  res$redist_log <- if (length(redist)) do.call(rbind, redist) else NULL
  res
}

# ----------------------------------------------------------------------------
# 3. Planner (precheck side): wf_plan_poststrat() ----------------------------
# ----------------------------------------------------------------------------

wf_plan_poststrat <- function(sample, target, min_cell, ladder,
                              granularity = c("adaptive", "province"),
                              empty_cell = c("redistribute", "flag", "error"),
                              id = NULL) {
  granularity <- match.arg(granularity); empty_cell <- match.arg(empty_cell)
  .wf_check_poststrat_args(sample, target, min_cell, ladder)
  dvars <- target$dims
  gkey  <- .wf_group_keys(sample, target$by)

  plan_rows <- list(); diag_rows <- list()
  for (g in intersect(names(target$joint), unique(gkey))) {
    sel  <- gkey == g
    smat <- as.matrix(sapply(dvars, function(d) .chr(sample[[d]][sel])))
    if (is.null(dim(smat))) smat <- matrix(smat, ncol = length(dvars),
                                           dimnames = list(NULL, dvars))
    jdf  <- target$joint[[g]]
    res  <- .wf_resolve_group(smat, jdf, dvars, ladder, min_cell, granularity, g)
    if (any(res$orphan) && empty_cell == "redistribute")
      res <- .wf_redistribute(res, min_cell, g)

    plan_rows[[g]] <- data.frame(
      group = g, jdf[, dvars, drop = FALSE], pop = jdf$pop,
      n_sample = res$supp_final, ladder_level = res$level_used,
      resolved_cell = res$resolved, orphan = res$orphan,
      stringsAsFactors = FALSE)
    diag_rows[[g]] <- data.frame(
      group = g, n_cells_raw = nrow(jdf),
      n_cells_resolved = length(unique(res$resolved)),
      granularity_used = res$granularity_used,
      n_at_level0 = sum(res$level_used == 0),
      max_level_used = max(res$level_used),
      n_orphan = sum(res$orphan),
      pop_orphan = sum(jdf$pop[res$orphan]),
      stringsAsFactors = FALSE)
  }
  structure(list(
    plan = do.call(rbind, plan_rows),
    diagnostics = do.call(rbind, diag_rows),
    settings = list(min_cell = min_cell, granularity = granularity,
                    empty_cell = empty_cell, ladder = ladder)),
    class = "wf_poststrat_plan")
}

print.wf_poststrat_plan <- function(x, ...) {
  d <- x$diagnostics
  cat(sprintf("<wf_poststrat_plan>  %d group(s); min_cell = %d\n",
              nrow(d), x$settings$min_cell))
  degr <- sum(d$granularity_used == "province")
  cat(sprintf("  granularity: %d adaptive, %d degraded to province-uniform\n",
              sum(d$granularity_used == "adaptive"), degr))
  cat(sprintf("  orphan cells (redistributed/flagged): %d across groups; pop moved: %s\n",
              sum(d$n_orphan), format(round(sum(d$pop_orphan)), big.mark = ",")))
  print(utils::head(d, 12), row.names = FALSE)
  if (nrow(d) > 12) cat(sprintf("  ... %d more groups (see $diagnostics)\n",
                                nrow(d) - 12))
  invisible(x)
}

# ----------------------------------------------------------------------------
# 4. Argument validation -----------------------------------------------------
# ----------------------------------------------------------------------------

.wf_check_poststrat_args <- function(sample, target, min_cell, ladder) {
  if (is.null(target$joint))
    wf_abort("Target has no joint cell table. Rebuild with wf_target_population(..., keep_joint = TRUE).",
             "wf_error_schema")
  if (!inherits(ladder, "wf_collapse_ladder"))
    wf_abort("`ladder` must be a wf_collapse_ladder object.", "wf_error_input")
  if (missing(min_cell) || length(min_cell) != 1 || !is.finite(min_cell) ||
      min_cell < 1 || min_cell != as.integer(min_cell))
    wf_abort("`min_cell` must be a single integer >= 1.", "wf_error_input")
  .require_cols(sample, c(target$dims, target$by), "sample")
  if (!identical(sort(ladder$dims), sort(target$dims)))
    wf_abort("Ladder dimensions do not match target dimensions.",
             "wf_error_schema",
             list(ladder = ladder$dims, target = target$dims))
  invisible(TRUE)
}

# ----------------------------------------------------------------------------
# 5. Executor: wf_poststrat() ------------------------------------------------
# ----------------------------------------------------------------------------

wf_poststrat <- function(sample, target, min_cell, ladder,
                         init_weight = NULL,
                         granularity = c("adaptive", "province"),
                         empty_cell = c("redistribute", "flag", "error"),
                         id = NULL, precheck = TRUE, tol = 1e-8) {
  granularity <- match.arg(granularity); empty_cell <- match.arg(empty_cell)
  t0 <- Sys.time()
  .wf_check_poststrat_args(sample, target, min_cell, ladder)

  dvars <- target$dims
  gkey  <- .wf_group_keys(sample, target$by)
  ids   <- if (is.null(id)) seq_len(nrow(sample)) else sample[[id]]
  iw    <- if (is.null(init_weight)) rep(1, nrow(sample)) else sample[[init_weight]]
  if (any(!is.finite(iw)) || any(iw <= 0))
    wf_abort("Initial weights must be finite and > 0.", "wf_error_input")

  res_rows <- list(); logs <- list(); reports <- list(); maps <- list()

  for (g in intersect(names(target$joint), unique(gkey))) {
    sel  <- which(gkey == g)
    smat <- as.matrix(sapply(dvars, function(d) .chr(sample[[d]][sel])))
    if (is.null(dim(smat))) smat <- matrix(smat, ncol = length(dvars),
                                           dimnames = list(NULL, dvars))
    jdf  <- target$joint[[g]]
    res  <- .wf_resolve_group(smat, jdf, dvars, ladder, min_cell, granularity, g)

    if (res$granularity_used == "province" && granularity == "adaptive")
      wf_warn(sprintf("Group '%s': adaptive resolution insufficient; degraded to province-uniform (level %d).",
                      g, max(res$level_used)), "wf_warning_quality",
              list(group = g))

    pop_flagged <- 0
    if (any(res$orphan)) {
      if (empty_cell == "error")
        wf_abort(sprintf("Group '%s': %d population cell(s) have zero sample support and empty_cell='error'.",
                         g, sum(res$orphan)), "wf_error_feasibility",
                 list(group = g, n_orphan = sum(res$orphan)))
      if (empty_cell == "redistribute") {
        res <- .wf_redistribute(res, min_cell, g)
        if (!is.null(res$redist_log))
          wf_warn(sprintf("Group '%s': redistributed %d orphan cell(s) into supported ancestors.",
                          g, nrow(res$redist_log)), "wf_warning_data",
                  list(group = g))
      } else { # flag: orphan population left unassigned
        pop_flagged <- sum(jdf$pop[res$orphan])
        wf_warn(sprintf("Group '%s': %d orphan cell(s) flagged; %s population left unassigned (province total will be short).",
                        g, sum(res$orphan),
                        format(round(pop_flagged), big.mark = ",")),
                "wf_warning_data", list(group = g))
      }
    }

    # ---- population per resolved cell (post-redistribution, orphans folded) --
    pop_by_res <- tapply(jdf$pop[!res$orphan], res$resolved[!res$orphan], sum)

    # ---- sample side: map each respondent to its resolved cell -----------------
    # A respondent's resolved cell is determined by applying, to the respondent,
    # the SAME level that its raw population cell was resolved at. We build, for
    # every level, the respondent's coarsened key, then pick the level assigned
    # to its raw population cell. Respondents whose raw cell is absent from the
    # population table (structural-zero population but nonzero sample) are folded
    # into the resolved cell of their nearest populated ancestor.
    lvl_of_raw <- stats::setNames(res$level_used, res$raw_key)
    s_raw      <- .wf_cell_key(smat, dvars)
    s_res      <- rep(NA_character_, length(sel))

    in_pop <- s_raw %in% res$raw_key
    if (any(in_pop)) {
      wr   <- which(in_pop)
      lv_i <- as.integer(lvl_of_raw[s_raw[in_pop]])
      s_res[in_pop] <- vapply(seq_along(wr), function(j)
        .wf_cell_key(.wf_apply_ladder(smat[wr[j], , drop = FALSE],
                                      ladder, lv_i[j]), dvars),
        character(1))
    }
    # respondents in cells absent from population: climb until their coarsened
    # key matches some resolved cell; if none, drop with warning.
    if (any(!in_pop)) {
      valid_res <- unique(res$resolved)
      for (idxrow in which(!in_pop)) {
        assigned <- NA_character_
        for (lv in 0:(length(res$s_keys) - 1)) {
          k <- .wf_cell_key(.wf_apply_ladder(smat[idxrow, , drop = FALSE],
                                             ladder, lv), dvars)
          if (k %in% valid_res) { assigned <- k; break }
        }
        s_res[idxrow] <- assigned
      }
      n_drop <- sum(is.na(s_res))
      if (n_drop > 0)
        wf_warn(sprintf("Group '%s': %d respondent(s) fall in cells with no populated resolved match; excluded.",
                        g, n_drop), "wf_warning_data", list(group = g))
    }
    known <- !is.na(s_res)

    # ---- calibration factors: factor_c = pop_c / sum(init_weight in c) --------
    iw_g <- iw[sel]
    denom <- tapply(iw_g[known], s_res[known], sum)
    common <- intersect(names(pop_by_res), names(denom))
    factor <- stats::setNames(rep(NA_real_, length(common)), common)
    factor[common] <- pop_by_res[common] / denom[common]

    w <- rep(NA_real_, length(sel))
    w[known] <- iw_g[known] * factor[s_res[known]]
    # respondents in a resolved cell that received no population (only via flag
    # path where its pop was withheld) get factor NA -> weight 0
    w[is.na(w)] <- 0

    # ---- province-total constraint (requirement 7) ---------------------------
    target_total <- target$groups[[g]]$total - pop_flagged
    realized     <- sum(w)
    dev          <- abs(realized - target_total) / target_total
    if (dev > tol)
      wf_abort(sprintf("Group '%s': province-total constraint violated (realized %.6g vs target %.6g, dev %.2e). This indicates an internal bug.",
                       g, realized, target_total, dev),
               "wf_error_internal", list(group = g, dev = dev))

    cv <- stats::sd(w[known]) / mean(w[known])
    res_rows[[g]] <- data.frame(
      id = ids[sel], group = g, resolved_cell = s_res,
      weight = w, feature = 1 / w, stringsAsFactors = FALSE)

    # ---- audit report: one row per ORIGINAL population cell ------------------
    cellw_mean <- tapply(w[known], s_res[known], mean)
    rep_g <- data.frame(
      group = g, jdf[, dvars, drop = FALSE], pop = jdf$pop,
      n_sample = res$supp_final, ladder_level_used = res$level_used,
      resolved_cell = res$resolved,
      granularity_used = res$granularity_used,
      orphan = res$orphan,
      redistributed_to = NA_character_, pop_moved = 0,
      final_cell_weight_mean = cellw_mean[res$resolved],
      stringsAsFactors = FALSE)
    if (!is.null(res$redist_log)) {
      for (r in seq_len(nrow(res$redist_log))) {
        hit <- rep_g$resolved_cell == res$redist_log$receiving_cell[r] &
               rep_g$n_sample == 0
        # mark the specific orphan rows that moved
      }
      # simpler: annotate orphan rows directly from the log by matching original
      om <- match(res$redist_log$orphan_cell,
                  vapply(seq_len(nrow(rep_g)),
                         function(i) paste(rep_g[i, dvars], collapse = "\r"),
                         character(1)))
    }
    reports[[g]] <- rep_g
    maps[[g]] <- data.frame(
      group = g, raw_cell = res$raw_key, resolved_cell = res$resolved,
      level = res$level_used, stringsAsFactors = FALSE)

    logs[[g]] <- data.frame(
      group = g, n = length(sel),
      n_cells_raw = nrow(jdf),
      n_cells_resolved = length(unique(res$resolved)),
      granularity_used = res$granularity_used,
      n_orphan_cells = if (is.null(res$redist_log)) 0L else nrow(res$redist_log),
      pop_redistributed = if (is.null(res$redist_log)) 0 else
        sum(jdf$pop[match(res$redist_log$orphan_cell,
            vapply(seq_len(nrow(jdf)),
                   function(i) paste(jdf[i, dvars], collapse = "\r"),
                   character(1)))], na.rm = TRUE),
      total_target = target_total, total_realized = realized,
      total_dev = dev, deff = round(1 + cv^2, 3),
      stringsAsFactors = FALSE)
  }

  structure(list(
    data = do.call(rbind, res_rows),
    log  = do.call(rbind, logs),
    cell_report = do.call(rbind, reports),
    collapse_map = structure(
      list(map = do.call(rbind, maps),
           settings = list(min_cell = min_cell, granularity = granularity)),
      class = "wf_poststrat_plan"),
    provenance = list(
      method = "poststrat", dims = dvars, by = target$by,
      min_cell = min_cell, granularity = granularity, empty_cell = empty_cell,
      init_weight = init_weight, ladder_levels = ladder$n_levels,
      created = t0, elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = "0.2.0")
  ), class = "wf_weights")
}

# augment the 0.1 print method with a post-strat line (dispatch by method field)
print.wf_weights <- function(x, ...) {
  cat(sprintf("<wf_weights>  %d unit(s) in %d group(s); method: %s\n",
              nrow(x$data), nrow(x$log),
              if (is.null(x$provenance$method)) "raking" else x$provenance$method))
  cat(sprintf("  weight range [%.4g, %.4g]; elapsed %.2fs\n",
              min(x$data$weight), max(x$data$weight), x$provenance$elapsed))
  if (!is.null(x$provenance$method) && x$provenance$method == "poststrat") {
    lg <- x$log
    cat(sprintf("  cells: %d raw -> %d resolved (median); groups degraded to province: %d\n",
                stats::median(lg$n_cells_raw),
                stats::median(lg$n_cells_resolved),
                sum(lg$granularity_used == "province")))
    cat(sprintf("  max province-total deviation: %.2e\n", max(lg$total_dev)))
  }
  invisible(x)
}

test_that("collapse ladder validates dimensions and cumulative source categories", {
  dims <- wf_dims(age = c("1", "2", "3", "4", "5"), edu = c("L", "M", "H"))

  ladder <- wf_collapse_ladder(
    dims,
    level1 = list(age = c("4" = "45", "5" = "45")),
    level2 = list(age = c("1" = "123", "2" = "123", "3" = "123", "45" = "45"))
  )

  expect_s3_class(ladder, "wf_collapse_ladder")
  expect_equal(ladder$dims, c("age", "edu"))
  expect_equal(ladder$n_levels, 2)

  mat <- matrix(
    c("4", "L", "5", "H", "2", "M"),
    ncol = 2,
    byrow = TRUE,
    dimnames = list(NULL, c("age", "edu"))
  )
  expect_equal(WFC:::.wf_apply_ladder(mat, ladder, 1)[, "age"], c("45", "45", "2"))
  expect_equal(WFC:::.wf_apply_ladder(mat, ladder, 2)[, "age"], c("45", "45", "123"))
})

test_that("collapse ladder rejects unknown dimensions and typo categories", {
  dims <- wf_dims(age = c("1", "2"), edu = c("L", "H"))

  expect_error(
    wf_collapse_ladder(dims, level1 = list(region = c("A" = "B"))),
    class = "wf_error_input"
  )
  expect_error(
    wf_collapse_ladder(dims, level1 = list(age = c("3" = "2"))),
    class = "wf_error_input"
  )
})

test_that("cell keys split back into dimension columns", {
  mat <- matrix(c("female", "young", "male", "old"), ncol = 2, byrow = TRUE)
  colnames(mat) <- c("gender", "age")

  key <- WFC:::.wf_cell_key(mat, c("gender", "age"))
  out <- WFC:::.wf_split_key(key, c("gender", "age"))

  expect_equal(out$gender, c("female", "male"))
  expect_equal(out$age, c("young", "old"))
})

#' Return the loaded package version for provenance.
#'
#' @keywords internal
#' @noRd
.wf_compose_package_version <- function() {
  tryCatch(
    as.character(utils::packageVersion("WFC")),
    error = function(e) "0.4.0"
  )
}

#' Validate a compose normalization mode.
#'
#' @param normalize Requested normalization mode.
#' @keywords internal
#' @noRd
.wf_compose_normalize <- function(normalize) {
  supported <- c("none", "mean1", "sum")
  if (length(normalize) != 1 || is.na(normalize) || !normalize %in% supported) {
    shown <- if (length(normalize) == 0) "<empty>" else as.character(normalize[[1]])
    wf_abort(
      sprintf(
        "Unsupported compose normalization '%s'. Supported modes: none, mean1, sum.",
        shown
      ),
      "wf_error_input",
      list(normalize = normalize)
    )
  }
  normalize
}

#' Label a composition stage.
#'
#' @param index Stage index.
#' @param stage_names Names from the `...` list.
#' @keywords internal
#' @noRd
.wf_compose_stage_label <- function(index, stage_names) {
  if (!is.null(stage_names) && length(stage_names) >= index && nzchar(stage_names[[index]])) {
    return(stage_names[[index]])
  }
  paste0("stage_", index)
}

#' Resolve the ID column for composition.
#'
#' @param stages List of `wf_weights` objects.
#' @param id Optional user-supplied ID column.
#' @keywords internal
#' @noRd
.wf_compose_id_col <- function(stages, id) {
  if (!is.null(id)) {
    if (length(id) != 1 || is.na(id) || !nzchar(id)) {
      wf_abort("`id` must be a single non-empty column name.", "wf_error_input")
    }
    return(id)
  }

  has_id <- vapply(stages, function(stage) {
    is.data.frame(stage$data) && "id" %in% names(stage$data)
  }, logical(1))
  if (all(has_id)) {
    return("id")
  }
  if (!any(has_id)) {
    return(NULL)
  }

  wf_abort(
    "Either every stage must have an id column or no stage may have one. Supply `id` when using a different shared ID column.",
    "wf_error_input",
    list(has_id = has_id)
  )
}

#' Validate one stage and extract compose columns.
#'
#' @param stage A `wf_weights` object.
#' @param index Stage index.
#' @param label Stage label.
#' @param id_col ID column name or `NULL`.
#' @keywords internal
#' @noRd
.wf_compose_stage_data <- function(stage, index, label, id_col) {
  if (!inherits(stage, "wf_weights")) {
    wf_abort(
      sprintf("Compose stage %d ('%s') must be a wf_weights object.", index, label),
      "wf_error_input",
      list(stage = index)
    )
  }
  if (!is.data.frame(stage$data)) {
    wf_abort(
      sprintf("Compose stage %d ('%s') must contain a data frame in `$data`.", index, label),
      "wf_error_input",
      list(stage = index)
    )
  }

  required <- c("group", "weight")
  if (!is.null(id_col)) {
    required <- c(id_col, required)
  }
  .require_cols(stage$data, required, sprintf("compose stage %d data", index))

  weight <- stage$data$weight
  if (any(is.na(weight)) || any(!is.finite(weight)) || any(weight <= 0)) {
    wf_abort(
      sprintf("Compose stage %d ('%s') has non-finite or non-positive weights.", index, label),
      "wf_error_input",
      list(stage = index)
    )
  }

  group <- .chr(stage$data$group)
  if (any(is.na(group)) || any(group == "")) {
    wf_abort(
      sprintf("Compose stage %d ('%s') has missing group values.", index, label),
      "wf_error_input",
      list(stage = index)
    )
  }

  ids <- NULL
  if (!is.null(id_col)) {
    ids <- .chr(stage$data[[id_col]])
    if (any(is.na(ids)) || any(ids == "")) {
      wf_abort(
        sprintf("Compose stage %d ('%s') has missing ID values.", index, label),
        "wf_error_input",
        list(stage = index)
      )
    }
    dup <- unique(ids[duplicated(ids)])
    if (length(dup) > 0) {
      wf_abort(
        sprintf(
          "Compose stage %d ('%s') has duplicated ID value(s): %s.",
          index,
          label,
          paste(utils::head(dup, 5), collapse = ", ")
        ),
        "wf_error_input",
        list(stage = index, ids = dup)
      )
    }
  }

  list(
    id = ids,
    group = group,
    weight = as.numeric(weight),
    stage = stage
  )
}

#' Build a per-stage provenance summary.
#'
#' @param part Validated stage data.
#' @param index Stage index.
#' @param label Stage label.
#' @keywords internal
#' @noRd
.wf_compose_stage_summary <- function(part, index, label) {
  list(
    index = index,
    name = label,
    row_count = length(part$weight),
    group_count = length(unique(part$group)),
    total_weight = sum(part$weight),
    mean_weight = mean(part$weight),
    min_weight = min(part$weight),
    max_weight = max(part$weight),
    provenance = part$stage$provenance
  )
}

#' Build a composition log compatible with wf_diagnose().
#'
#' @param data Composed weight data.
#' @param stage_count Number of composed stages.
#' @keywords internal
#' @noRd
.wf_compose_log <- function(data, stage_count) {
  rows <- lapply(split(data, data$group), function(group_data) {
    data.frame(
      group = group_data$group[[1]],
      n = nrow(group_data),
      total_weight = sum(group_data$weight),
      mean_weight = mean(group_data$weight),
      min_weight = min(group_data$weight),
      max_weight = max(group_data$weight),
      stage_count = stage_count,
      iterations = NA_integer_,
      converged = TRUE,
      trimmed = 0L,
      stringsAsFactors = FALSE
    )
  })
  do.call(rbind, rows)
}

#' Compose multiple weighting stages
#'
#' Multiplies compatible `wf_weights` objects into one auditable pipeline result.
#'
#' @param ... Two or more `wf_weights` objects.
#' @param id Optional shared ID column in each stage's `$data`. If `NULL`,
#'   `wf_compose()` uses `data$id` when every stage has it, or row order when no
#'   stage has an ID column.
#' @param normalize Weight normalization mode: `"none"`, `"mean1"`, or `"sum"`.
#'
#' @return A composed `wf_weights` object.
#' @export
#'
#' @examples
#' stage1 <- structure(
#'   list(
#'     data = data.frame(id = c("r1", "r2"), group = "A", weight = c(2, 3)),
#'     log = data.frame(group = "A"),
#'     achieved = NULL,
#'     provenance = list(method = "stage1")
#'   ),
#'   class = "wf_weights"
#' )
#' stage2 <- stage1
#' stage2$data$weight <- c(0.5, 2)
#' stage2$provenance$method <- "stage2"
#' wf_compose(stage1, stage2)
wf_compose <- function(..., id = NULL, normalize = c("none", "mean1", "sum")) {
  t0 <- Sys.time()
  normalize <- if (missing(normalize)) "none" else .wf_compose_normalize(normalize)
  stages <- list(...)
  stage_names <- names(stages)

  if (length(stages) < 2) {
    wf_abort("`wf_compose()` requires at least two wf_weights stages.", "wf_error_input")
  }
  for (i in seq_along(stages)) {
    label <- .wf_compose_stage_label(i, stage_names)
    if (!inherits(stages[[i]], "wf_weights")) {
      wf_abort(
        sprintf("Compose stage %d ('%s') must be a wf_weights object.", i, label),
        "wf_error_input",
        list(stage = i)
      )
    }
  }

  id_col <- .wf_compose_id_col(stages, id)
  parts <- lapply(seq_along(stages), function(i) {
    .wf_compose_stage_data(
      stages[[i]],
      i,
      .wf_compose_stage_label(i, stage_names),
      id_col
    )
  })

  base <- parts[[1]]
  if (is.null(id_col)) {
    row_counts <- vapply(parts, function(part) length(part$weight), integer(1))
    if (length(unique(row_counts)) != 1) {
      wf_abort(
        "Row-order composition requires every stage to have the same row count.",
        "wf_error_input",
        list(row_counts = row_counts)
      )
    }
    out_id <- .chr(seq_len(row_counts[[1]]))
    out_group <- base$group
    product <- base$weight
    for (i in 2:length(parts)) {
      mismatch <- which(parts[[i]]$group != out_group)
      if (length(mismatch) > 0) {
        wf_abort(
          sprintf("Compose stage %d has group assignments that differ from stage 1.", i),
          "wf_error_input",
          list(stage = i, ids = out_id[utils::head(mismatch, 5)])
        )
      }
      product <- product * parts[[i]]$weight
    }
    id_source <- "row_order"
  } else {
    out_id <- base$id
    out_group <- base$group
    product <- base$weight
    for (i in 2:length(parts)) {
      missing <- setdiff(out_id, parts[[i]]$id)
      extra <- setdiff(parts[[i]]$id, out_id)
      if (length(missing) > 0 || length(extra) > 0) {
        wf_abort(
          sprintf("Compose stage %d has a different ID set from stage 1.", i),
          "wf_error_input",
          list(stage = i, missing = missing, extra = extra)
        )
      }
      idx <- match(out_id, parts[[i]]$id)
      group_i <- parts[[i]]$group[idx]
      mismatch <- which(group_i != out_group)
      if (length(mismatch) > 0) {
        wf_abort(
          sprintf("Compose stage %d has group assignments that differ from stage 1.", i),
          "wf_error_input",
          list(stage = i, ids = out_id[utils::head(mismatch, 5)])
        )
      }
      product <- product * parts[[i]]$weight[idx]
    }
    id_source <- id_col
  }

  normalization_factor <- switch(
    normalize,
    none = 1,
    mean1 = 1 / mean(product),
    sum = sum(base$weight) / sum(product)
  )
  final_weight <- product * normalization_factor
  out_data <- data.frame(
    id = out_id,
    group = out_group,
    weight = final_weight,
    feature = 1 / final_weight,
    stringsAsFactors = FALSE
  )

  structure(
    list(
      data = out_data,
      log = .wf_compose_log(out_data, length(parts)),
      achieved = NULL,
      provenance = list(
        method = "compose",
        mode = "compose",
        stages = lapply(seq_along(parts), function(i) {
          .wf_compose_stage_summary(
            parts[[i]],
            i,
            .wf_compose_stage_label(i, stage_names)
          )
        }),
        compose = list(
          normalize = normalize,
          normalization_factor = normalization_factor,
          id = id_source,
          stage_count = length(parts),
          created = t0,
          elapsed = as.numeric(Sys.time() - t0, units = "secs"),
          package_version = .wf_compose_package_version()
        )
      )
    ),
    class = "wf_weights"
  )
}

make_compose_stage <- function(ids = paste0("r", 1:4),
                               groups = c("A", "A", "B", "B"),
                               weights = c(1, 2, 3, 4),
                               method = "stage",
                               row_order = NULL,
                               drop_id = FALSE) {
  data <- data.frame(
    id = ids,
    group = groups,
    weight = weights,
    feature = 1 / weights,
    stringsAsFactors = FALSE
  )
  if (!is.null(row_order)) {
    data <- data[row_order, , drop = FALSE]
  }
  if (drop_id) {
    data$id <- NULL
  }
  structure(
    list(
      data = data,
      log = data.frame(
        group = unique(data$group),
        iterations = NA_integer_,
        converged = TRUE,
        trimmed = 0L,
        stringsAsFactors = FALSE
      ),
      achieved = NULL,
      provenance = list(method = method, created = Sys.time())
    ),
    class = "wf_weights"
  )
}

test_that("wf_compose multiplies stages by id and records provenance", {
  stage1 <- make_compose_stage(weights = c(2, 3, 4, 5), method = "design")
  stage2 <- make_compose_stage(weights = c(0.5, 2, 1.5, 1), method = "adjustment")

  out <- wf_compose(stage1, stage2)

  expect_s3_class(out, "wf_weights")
  expect_equal(out$data$id, paste0("r", 1:4))
  expect_equal(out$data$group, c("A", "A", "B", "B"))
  expect_equal(out$data$weight, c(1, 6, 6, 5), tolerance = 1e-10)
  expect_equal(out$data$feature, 1 / out$data$weight, tolerance = 1e-10)
  expect_equal(out$achieved, NULL)
  expect_equal(out$provenance$method, "compose")
  expect_equal(out$provenance$mode, "compose")
  expect_equal(length(out$provenance$stages), 2)
  expect_equal(out$provenance$stages[[1]]$provenance$method, "design")
  expect_equal(out$provenance$stages[[2]]$provenance$method, "adjustment")
  expect_equal(out$provenance$compose$stage_count, 2)
  expect_equal(out$provenance$compose$normalize, "none")
  expect_true(all(c("iterations", "converged", "trimmed") %in% names(out$log)))
})

test_that("wf_compose is invariant to input row order when ids match", {
  stage1 <- make_compose_stage(weights = c(2, 3, 4, 5))
  stage2 <- make_compose_stage(
    weights = c(0.5, 2, 1.5, 1),
    row_order = c(4, 2, 1, 3)
  )

  out <- wf_compose(stage1, stage2)

  expect_equal(out$data$id, paste0("r", 1:4))
  expect_equal(out$data$weight, c(1, 6, 6, 5), tolerance = 1e-10)
})

test_that("wf_compose supports row-order composition when no stage has ids", {
  stage1 <- make_compose_stage(
    ids = paste0("r", 1:3),
    weights = c(2, 3, 4),
    groups = c("A", "A", "B"),
    drop_id = TRUE
  )
  stage2 <- make_compose_stage(
    ids = paste0("r", 1:3),
    weights = c(0.5, 2, 1.5),
    groups = c("A", "A", "B"),
    drop_id = TRUE
  )

  out <- wf_compose(stage1, stage2)

  expect_equal(out$data$id, as.character(1:3))
  expect_equal(out$data$weight, c(1, 6, 6), tolerance = 1e-10)
})

test_that("wf_compose normalizes composed weights", {
  stage1 <- make_compose_stage(weights = c(2, 3, 4, 5))
  stage2 <- make_compose_stage(weights = c(0.5, 2, 1.5, 1))

  mean1 <- wf_compose(stage1, stage2, normalize = "mean1")
  sum_norm <- wf_compose(stage1, stage2, normalize = "sum")

  expect_equal(mean(mean1$data$weight), 1, tolerance = 1e-10)
  expect_equal(sum(sum_norm$data$weight), sum(stage1$data$weight), tolerance = 1e-10)
  expect_equal(mean1$provenance$compose$normalize, "mean1")
  expect_equal(sum_norm$provenance$compose$normalize, "sum")
})

test_that("wf_compose rejects unsafe inputs", {
  stage1 <- make_compose_stage()
  stage2 <- make_compose_stage()

  expect_error(wf_compose(stage1), class = "wf_error_input")
  expect_error(wf_compose(stage1, list()), class = "wf_error_input")
  expect_error(wf_compose(stage1, stage2, normalize = "median"), class = "wf_error_input")

  missing_group <- stage2
  missing_group$data$group <- NULL
  expect_error(wf_compose(stage1, missing_group), class = "wf_error_schema")

  duplicate_ids <- make_compose_stage(ids = c("r1", "r1", "r3", "r4"))
  expect_error(wf_compose(stage1, duplicate_ids), class = "wf_error_input")

  missing_ids <- make_compose_stage(ids = c("r1", NA, "r3", "r4"))
  expect_error(wf_compose(stage1, missing_ids), class = "wf_error_input")

  different_ids <- make_compose_stage(ids = c("r1", "r2", "r3", "rx"))
  expect_error(wf_compose(stage1, different_ids), class = "wf_error_input")

  group_mismatch <- make_compose_stage(groups = c("A", "A", "A", "B"))
  expect_error(wf_compose(stage1, group_mismatch), class = "wf_error_input")

  bad_weights <- make_compose_stage(weights = c(1, Inf, 1, 1))
  expect_error(wf_compose(stage1, bad_weights), class = "wf_error_input")

  non_positive_weights <- make_compose_stage(weights = c(1, 0, 1, 1))
  expect_error(wf_compose(stage1, non_positive_weights), class = "wf_error_input")
})

test_that("wf_compose requires consistent id-column availability", {
  with_id <- make_compose_stage()
  without_id <- make_compose_stage(drop_id = TRUE)

  expect_error(wf_compose(with_id, without_id), class = "wf_error_input")
  expect_error(wf_compose(without_id, with_id), class = "wf_error_input")
  expect_error(wf_compose(without_id, without_id, id = "id"), class = "wf_error_schema")
})

test_that("composed weights remain compatible with print and diagnose", {
  stage1 <- make_compose_stage(weights = c(2, 3, 4, 5))
  stage2 <- make_compose_stage(weights = c(0.5, 2, 1.5, 1))

  out <- wf_compose(stage1, stage2)
  diag <- wf_diagnose(out)

  expect_output(print(out), "<wf_weights>")
  expect_s3_class(diag, "wf_diagnostics")
  expect_true(all(c("ess", "deff", "verdict") %in% names(diag$table)))
  expect_false("margin_maxerr" %in% names(diag$table))
  expect_true(all(diag$table$converged))
})

test_that("wf_compose is available as a foundation pipeline API", {
  expect_true(is.function(wf_compose))
})

#' Declare calibration dimensions
#'
#' @param ... Named dimension-level pairs. Use `NULL` to infer levels from the target.
#' @param .collapse Named list of collapse ladders.
#'
#' @return A `wf_dims` object.
#' @export
#'
#' @examples
#' dims <- wf_dims(gender = c("female", "male"), age = c("young", "old"))
#' dims
wf_dims <- function(..., .collapse = list()) {
  vars <- list(...)
  if (length(vars) == 0 || is.null(names(vars)) || any(names(vars) == "")) {
    wf_abort(
      "wf_dims() requires named arguments: dimension = levels (or NULL).",
      "wf_error_input"
    )
  }
  vars <- lapply(vars, function(v) if (is.null(v)) NULL else .chr(v))
  bad <- setdiff(names(.collapse), names(vars))
  if (length(bad) > 0) {
    wf_abort(sprintf(
      "Collapse ladder(s) for unknown dimension(s): %s",
      paste(bad, collapse = ", ")
    ), "wf_error_input", list(unknown = bad))
  }
  structure(list(vars = vars, collapse = .collapse), class = "wf_dims")
}

#' Return the loaded package version for attrition provenance.
#'
#' @keywords internal
#' @noRd
.wf_attrition_package_version <- function() {
  tryCatch(
    as.character(utils::packageVersion("WFC")),
    error = function(e) "0.15.0"
  )
}

#' Extract predictors from an attrition model formula.
#'
#' @param formula Formula supplied by the user.
#' @keywords internal
#' @noRd
.wf_attrition_predictors <- function(formula) {
  if (!inherits(formula, "formula") || !length(formula) %in% c(2, 3)) {
    wf_abort(
      "`formula` must be a one-sided or two-sided formula naming retention predictors.",
      "wf_error_input"
    )
  }
  rhs <- if (length(formula) == 3) formula[[3]] else formula[[2]]
  predictors <- all.vars(rhs)
  if (length(predictors) == 0) {
    wf_abort(
      "`formula` must name at least one retention predictor.",
      "wf_error_input"
    )
  }
  predictors
}

#' Resolve a logical retention flag.
#'
#' @param x Retention indicator vector.
#' @param retained Column name.
#' @keywords internal
#' @noRd
.wf_attrition_flag <- function(x, retained) {
  if (anyNA(x)) {
    wf_abort(
      sprintf("`%s` must not contain missing retention values.", retained),
      "wf_error_input",
      list(retained = retained)
    )
  }
  if (is.logical(x)) {
    keep <- x
  } else if (is.numeric(x) || is.integer(x)) {
    keep <- as.numeric(x) == 1
  } else {
    value <- tolower(.chr(x))
    if (!all(value %in% c("1", "0", "true", "false", "yes", "no", "retained", "lost", NA))) {
      wf_abort(
        sprintf(
          "`%s` must be logical, 0/1, or a recognized retained/lost label.",
          retained
        ),
        "wf_error_input",
        list(retained = retained)
      )
    }
    keep <- value %in% c("1", "true", "yes", "retained")
  }
  keep
}

#' Standardized mean difference for attrition balance.
#'
#' @param retained Retained-unit values.
#' @param full Full prior-wave values.
#' @param weights Retained-unit weights.
#' @keywords internal
#' @noRd
.wf_attrition_smd <- function(retained, full, weights) {
  sd_pool <- stats::sd(full)
  if (!is.finite(sd_pool) || sd_pool == 0) {
    sd_pool <- NA_real_
  }
  c(
    unweighted = (mean(retained) - mean(full)) / sd_pool,
    weighted = (stats::weighted.mean(retained, weights) - mean(full)) / sd_pool
  )
}

#' Build attrition balance diagnostics.
#'
#' @param panel Full panel frame.
#' @param keep Retention flag.
#' @param predictors Model predictors.
#' @param weights Retained-unit attrition weights.
#' @keywords internal
#' @noRd
.wf_attrition_balance <- function(panel, keep, predictors, weights) {
  rows <- list()
  retained <- panel[keep, , drop = FALSE]
  for (predictor in predictors) {
    x_retained <- retained[[predictor]]
    x_full <- panel[[predictor]]
    if (is.numeric(x_full)) {
      smd <- .wf_attrition_smd(x_retained, x_full, weights)
      rows[[length(rows) + 1L]] <- data.frame(
        variable = predictor,
        level = NA_character_,
        full_mean = mean(x_full),
        retained_mean_unweighted = mean(x_retained),
        retained_mean_weighted = stats::weighted.mean(x_retained, weights),
        smd_unweighted = unname(smd[["unweighted"]]),
        smd_weighted = unname(smd[["weighted"]]),
        stringsAsFactors = FALSE
      )
    } else {
      levels <- sort(unique(.chr(x_full)))
      for (level in levels[-1]) {
        full_dummy <- as.numeric(.chr(x_full) == level)
        retained_dummy <- as.numeric(.chr(x_retained) == level)
        smd <- .wf_attrition_smd(retained_dummy, full_dummy, weights)
        rows[[length(rows) + 1L]] <- data.frame(
          variable = predictor,
          level = level,
          full_mean = mean(full_dummy),
          retained_mean_unweighted = mean(retained_dummy),
          retained_mean_weighted = stats::weighted.mean(retained_dummy, weights),
          smd_unweighted = unname(smd[["unweighted"]]),
          smd_weighted = unname(smd[["weighted"]]),
          stringsAsFactors = FALSE
        )
      }
    }
  }
  out <- do.call(rbind, rows)
  row.names(out) <- NULL
  out
}

#' Fit one attrition model group.
#'
#' @param panel Full group panel.
#' @param keep Retention flag for the group.
#' @param predictors Model predictors.
#' @param group Group label.
#' @keywords internal
#' @noRd
.wf_attrition_fit_group <- function(panel, keep, predictors, group) {
  if (sum(keep) == 0 || sum(!keep) == 0) {
    wf_abort(
      sprintf(
        "Group '%s' must contain both retained and lost units to fit attrition weights.",
        group
      ),
      "wf_error_overlap",
      list(group = group, retained = sum(keep), lost = sum(!keep))
    )
  }
  data <- panel
  data$.wf_retained <- as.integer(keep)
  formula <- stats::reformulate(predictors, response = ".wf_retained")
  fit <- stats::glm(formula, family = stats::binomial(), data = data)
  phat <- stats::fitted(fit)
  if (any(!is.finite(phat)) || any(phat <= 0 | phat >= 1)) {
    wf_warn(
      sprintf(
        "Group '%s': fitted retention probabilities touch 0 or 1; attrition weights may be unstable.",
        group
      ),
      "wf_warning_quality",
      list(group = group)
    )
  }
  pmax(pmin(phat, 1 - 1e-8), 1e-8)
}

#' Estimate panel attrition weights
#'
#' Models wave-to-wave retention in a panel and returns inverse-retention
#' weights for retained units. The result inherits from `wf_weights`, so it can
#' be composed with later calibration stages through [wf_compose()].
#'
#' @param panel Full prior-wave panel data.
#' @param retained Column indicating whether each prior-wave unit was retained
#'   into the next wave.
#' @param formula One-sided or two-sided formula naming retention predictors.
#' @param id Optional unique unit identifier column.
#' @param by Optional grouping column; separate retention models are fit by
#'   group.
#' @param stabilize Whether to multiply inverse probabilities by the group
#'   retention rate before normalization.
#' @param trim Optional positive scalar cap, applied as `trim * median(weight)`
#'   within each group before mean-one normalization.
#'
#' @return A `wf_attrition_weights` object inheriting from `wf_weights`, with a
#'   `$balance` diagnostic table and per-group retention log.
#' @export
wf_attrition <- function(panel, retained, formula, id = NULL, by = NULL,
                         stabilize = TRUE, trim = NULL) {
  t0 <- Sys.time()
  if (!is.data.frame(panel) || nrow(panel) == 0) {
    wf_abort("`panel` must be a non-empty data frame.", "wf_error_input")
  }
  if (!.wf_is_string(retained)) {
    wf_abort("`retained` must be one non-empty column name.", "wf_error_input")
  }
  .require_cols(panel, c(retained, id, by), "panel")
  predictors <- .wf_attrition_predictors(formula)
  .require_cols(panel, predictors, "panel")
  if (any(vapply(panel[predictors], function(x) anyNA(x), logical(1)))) {
    wf_abort(
      "Retention-model predictors must not contain missing values.",
      "wf_error_input"
    )
  }
  if (length(stabilize) != 1 || !is.logical(stabilize) || is.na(stabilize)) {
    wf_abort("`stabilize` must be TRUE or FALSE.", "wf_error_input")
  }
  if (!is.null(trim) &&
      (length(trim) != 1 || !is.numeric(trim) || !is.finite(trim) ||
       trim <= 0)) {
    wf_abort("`trim` must be NULL or one positive finite number.", "wf_error_input")
  }

  keep <- .wf_attrition_flag(panel[[retained]], retained)
  group <- if (is.null(by)) rep("_panel_", nrow(panel)) else .chr(panel[[by]])
  ids <- if (is.null(id)) as.character(seq_len(nrow(panel))) else .chr(panel[[id]])
  if (anyNA(ids) || any(!nzchar(ids)) || anyDuplicated(ids)) {
    wf_abort(
      "Attrition weighting requires non-missing unique unit IDs.",
      "wf_error_schema"
    )
  }

  rows <- list()
  logs <- list()
  balance <- list()
  probability <- rep(NA_real_, nrow(panel))
  for (group_name in unique(group)) {
    idx <- which(group == group_name)
    keep_g <- keep[idx]
    phat <- .wf_attrition_fit_group(
      panel[idx, , drop = FALSE],
      keep_g,
      predictors,
      group_name
    )
    probability[idx] <- phat
    raw <- 1 / phat[keep_g]
    if (stabilize) {
      raw <- mean(keep_g) * raw
    }
    trimmed <- 0L
    if (!is.null(trim)) {
      cap <- trim * stats::median(raw)
      hit <- raw > cap
      trimmed <- sum(hit)
      raw[hit] <- cap
    }
    weight <- raw / mean(raw)
    retained_idx <- idx[keep_g]
    rows[[group_name]] <- data.frame(
      id = ids[retained_idx],
      group = group_name,
      weight = weight,
      feature = 1 / weight,
      retention_probability = phat[keep_g],
      stringsAsFactors = FALSE
    )
    logs[[group_name]] <- data.frame(
      group = group_name,
      n = length(idx),
      retained = sum(keep_g),
      lost = sum(!keep_g),
      retention_rate = mean(keep_g),
      trimmed = trimmed,
      weight_min = min(weight),
      weight_max = max(weight),
      stringsAsFactors = FALSE
    )
    bal <- .wf_attrition_balance(
      panel[idx, , drop = FALSE],
      keep_g,
      predictors,
      weight
    )
    bal$group <- group_name
    balance[[group_name]] <- bal[c(
      "group", "variable", "level", "full_mean",
      "retained_mean_unweighted", "retained_mean_weighted",
      "smd_unweighted", "smd_weighted"
    )]
  }

  out <- structure(list(
    data = do.call(rbind, rows),
    log = do.call(rbind, logs),
    achieved = NULL,
    balance = do.call(rbind, balance),
    retention_probability = data.frame(
      id = ids,
      group = group,
      retained = keep,
      probability = probability,
      stringsAsFactors = FALSE
    ),
    provenance = list(
      method = "attrition",
      retained = retained,
      predictors = predictors,
      id = id,
      by = by,
      stabilize = stabilize,
      trim = trim,
      assumption = paste(
        "Inverse-retention weighting assumes retention is ignorable after",
        "conditioning on the model predictors."
      ),
      created = t0,
      elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = .wf_attrition_package_version()
    )
  ), class = c("wf_attrition_weights", "wf_weights"))
  row.names(out$data) <- NULL
  row.names(out$log) <- NULL
  row.names(out$balance) <- NULL
  out
}

#' Print attrition weights
#'
#' @param x A `wf_attrition_weights` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_attrition_weights <- function(x, ...) {
  cat(sprintf(
    "<wf_attrition_weights> %d retained unit(s) in %d group(s)\n",
    nrow(x$data),
    nrow(x$log)
  ))
  cat(sprintf(
    "  retention range [%.3f, %.3f]; weight range [%.4g, %.4g]\n",
    min(x$retention_probability$probability),
    max(x$retention_probability$probability),
    min(x$data$weight),
    max(x$data$weight)
  ))
  invisible(x)
}

#' Recommend a weight-trimming cap
#'
#' Sweeps candidate upper caps by rerunning raking and reports the trade-off
#' between design effect and residual margin error. The function recommends a
#' cap but never applies it to a production result.
#'
#' @param design An unchanged `wf_design_data`.
#' @param target An unchanged, non-demo `wf_verified_target`.
#' @param caps Unique positive finite upper-cap multipliers.
#' @param lo Positive lower-cap multiplier used for every finite candidate.
#' @param max_deff Maximum acceptable worst-group design effect.
#' @param max_residual Maximum acceptable worst relative margin residual.
#' @param ... Additional raking settings. ID and base-weight roles are owned by
#'   `design`; `trim` is not allowed because this function owns candidates.
#'
#' @return A `wf_auto_trim` object containing the candidate frontier and the
#'   recommended cap. `Inf` means no trimming is needed; `NA` means no candidate
#'   meets both criteria.
#' @export
wf_auto_trim <- function(design, target,
                         caps = c(2, 3, 4, 5, 6, 8, 10, 12),
                         lo = 0.05, max_deff = 6,
                         max_residual = 0.02, ...) {
  .wf_execute_verified_engine(
    design,
    target,
    "auto_trim",
    c(
      list(
        caps = caps,
        lo = lo,
        max_deff = max_deff,
        max_residual = max_residual
      ),
      list(...)
    )
  )
}

#' Internal automatic-trimming engine.
#'
#' @keywords internal
#' @noRd
.wf_auto_trim_engine <- function(sample, target, id = NULL,
                                 caps = c(2, 3, 4, 5, 6, 8, 10, 12),
                                 lo = 0.05, max_deff = 6,
                                 max_residual = 0.02,
                                 init_weight = NULL, ...) {
  if (!is.data.frame(sample) || nrow(sample) == 0) {
    wf_abort("`sample` must be a non-empty data frame.", "wf_error_input")
  }
  if (!inherits(target, "wf_target")) {
    wf_abort("`target` must be a wf_target object.", "wf_error_input")
  }
  if (!is.numeric(caps) || length(caps) == 0 ||
      any(!is.finite(caps)) || any(caps <= 0) || anyDuplicated(caps)) {
    wf_abort(
      "`caps` must contain unique positive finite numbers.",
      "wf_error_input",
      list(caps = caps)
    )
  }
  if (length(lo) != 1 || !is.finite(lo) || lo <= 0 || any(caps <= lo)) {
    wf_abort(
      "`lo` must be positive and smaller than every candidate cap.",
      "wf_error_input",
      list(lo = lo)
    )
  }
  if (length(max_deff) != 1 || !is.finite(max_deff) || max_deff <= 0) {
    wf_abort("`max_deff` must be a positive finite number.", "wf_error_input")
  }
  if (length(max_residual) != 1 || !is.finite(max_residual) ||
      max_residual < 0) {
    wf_abort(
      "`max_residual` must be a non-negative finite number.",
      "wf_error_input"
    )
  }

  dots <- list(...)
  if (!is.null(init_weight)) {
    dots$init_weight <- init_weight
  }
  if ("trim" %in% names(dots)) {
    wf_abort(
      "`trim` cannot be supplied to wf_auto_trim(); use `caps` and `lo`.",
      "wf_error_input"
    )
  }

  t0 <- Sys.time()
  run_candidate <- function(cap) {
    warnings <- character()
    args <- c(
      list(
        sample = sample,
        target = target,
        id = id,
        trim = if (is.finite(cap)) c(lo, cap) else NULL
      ),
      dots
    )
    fit <- tryCatch(
      withCallingHandlers(
        do.call(.wf_rake_engine, args),
        warning = function(w) {
          warnings <<- c(warnings, conditionMessage(w))
          invokeRestart("muffleWarning")
        }
      ),
      error = function(e) e
    )

    if (inherits(fit, "error")) {
      return(data.frame(
        cap = cap,
        feasible = FALSE,
        worst_deff = NA_real_,
        worst_residual = NA_real_,
        warning_count = length(warnings),
        error_class = paste(class(fit), collapse = "/"),
        error_message = conditionMessage(fit),
        stringsAsFactors = FALSE
      ))
    }

    diagnostics <- wf_diagnose(fit, target = target)$table
    residual <- if ("margin_maxerr" %in% names(diagnostics)) {
      max(diagnostics$margin_maxerr, na.rm = TRUE)
    } else {
      NA_real_
    }
    data.frame(
      cap = cap,
      feasible = TRUE,
      worst_deff = max(diagnostics$deff, na.rm = TRUE),
      worst_residual = residual,
      warning_count = length(warnings),
      error_class = NA_character_,
      error_message = NA_character_,
      stringsAsFactors = FALSE
    )
  }

  frontier <- do.call(rbind, lapply(c(sort(caps), Inf), run_candidate))
  frontier$meets_criteria <- frontier$feasible &
    is.finite(frontier$worst_deff) &
    frontier$worst_deff <= max_deff &
    is.finite(frontier$worst_residual) &
    frontier$worst_residual <= max_residual

  baseline_ok <- frontier$meets_criteria[is.infinite(frontier$cap)]
  finite_ok <- frontier$meets_criteria & is.finite(frontier$cap)
  recommended <- if (isTRUE(baseline_ok)) {
    Inf
  } else if (any(finite_ok)) {
    max(frontier$cap[finite_ok])
  } else {
    NA_real_
  }

  structure(
    list(
      frontier = frontier,
      recommended_cap = recommended,
      criteria = list(
        max_deff = max_deff,
        max_residual = max_residual,
        lo = lo
      ),
      provenance = list(
        method = "auto_trim",
        caps = sort(caps),
        id = id,
        created = t0,
        elapsed = as.numeric(Sys.time() - t0, units = "secs"),
        package_version = .wf_package_version()
      )
    ),
    class = "wf_auto_trim"
  )
}

#' Print an automatic trim recommendation
#'
#' @param x A `wf_auto_trim` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_auto_trim <- function(x, ...) {
  recommendation <- if (is.na(x$recommended_cap)) {
    "none meets the criteria"
  } else if (!is.finite(x$recommended_cap)) {
    "no trimming needed"
  } else {
    sprintf("%.3gx mean weight", x$recommended_cap)
  }
  cat("<wf_auto_trim> recommended cap:", recommendation, "\n")
  print(x$frontier, row.names = FALSE)
  invisible(x)
}

#' Post-stratify verified design data to a verified joint target
#'
#' Enforces the WFC 2.0 verified-input boundary before cell-level
#' post-stratification.
#'
#' @param design An unchanged `wf_design_data`.
#' @param target An unchanged, non-demo `wf_verified_target` with joint cells.
#' @param min_cell Minimum sample count per resolved cell.
#' @param ladder A `wf_collapse_ladder` object.
#' @param ... Post-stratification settings other than ID and base-weight roles,
#'   which are owned by `design`.
#'
#' @return A `wf_weights` object with `cell_report` and `collapse_map`.
#' @export
wf_poststrat <- function(design, target, min_cell, ladder, ...) {
  .wf_execute_verified_engine(
    design,
    target,
    "poststrat",
    c(list(min_cell = min_cell, ladder = ladder), list(...))
  )
}

#' Internal cell-level post-stratification engine.
#'
#' @keywords internal
#' @noRd
.wf_poststrat_engine <- function(sample, target, min_cell, ladder,
                                 init_weight = NULL,
                                 granularity = c("adaptive", "province"),
                                 empty_cell = c("redistribute", "flag", "error"),
                                 id = NULL, precheck = TRUE, tol = 1e-8,
                                 parallel = FALSE, progress = FALSE) {
  granularity <- match.arg(granularity)
  empty_cell <- match.arg(empty_cell)
  t0 <- Sys.time()
  .wf_check_poststrat_args(sample, target, min_cell, ladder)

  dvars <- target$dims
  gkey <- .wf_group_keys(sample, target$by)
  ids <- if (is.null(id)) seq_len(nrow(sample)) else sample[[id]]
  if (is.null(init_weight)) {
    iw <- rep(1, nrow(sample))
  } else {
    if (length(init_weight) != 1 || !is.character(init_weight) ||
        !init_weight %in% names(sample)) {
      wf_abort(sprintf(
        "init_weight column '%s' not found in sample.",
        as.character(init_weight)[1]
      ), "wf_error_schema", list(init_weight = init_weight))
    }
    iw <- as.numeric(sample[[init_weight]])
  }
  if (any(!is.finite(iw)) || any(iw <= 0)) {
    wf_abort("Initial weights must be finite and > 0.", "wf_error_input")
  }

  fit_group <- function(g) {
    sel <- which(gkey == g)
    smat <- as.matrix(sapply(dvars, function(d) .chr(sample[[d]][sel])))
    if (is.null(dim(smat))) {
      smat <- matrix(smat, ncol = length(dvars), dimnames = list(NULL, dvars))
    }
    jdf <- target$joint[[g]]
    raw_jkey <- .wf_cell_key(as.matrix(jdf[, dvars, drop = FALSE]), dvars)
    res <- .wf_resolve_group(smat, jdf, dvars, ladder, min_cell, granularity, g)

    if (res$granularity_used == "province" && granularity == "adaptive") {
      wf_warn(
        sprintf(
          "Group '%s': adaptive resolution insufficient; degraded to province-uniform.",
          g
        ),
        "wf_warning_quality",
        list(group = g)
      )
    }

    pop_flagged <- 0
    if (any(res$orphan)) {
      if (empty_cell == "error") {
        wf_abort(
          sprintf(
            "Group '%s': %d population cell(s) have zero sample support and empty_cell='error'.",
            g,
            sum(res$orphan)
          ),
          "wf_error_feasibility",
          list(group = g, n_orphan = sum(res$orphan))
        )
      }
      if (empty_cell == "redistribute") {
        res <- .wf_redistribute(res, min_cell, g)
        if (!is.null(res$redist_log)) {
          wf_warn(
            sprintf(
              "Group '%s': redistributed %d orphan cell(s) into supported ancestors.",
              g,
              nrow(res$redist_log)
            ),
            "wf_warning_data",
            list(group = g)
          )
        }
      } else {
        pop_flagged <- sum(jdf$pop[res$orphan])
        wf_warn(
          sprintf(
            "Group '%s': %d orphan cell(s) flagged; %s population left unassigned.",
            g,
            sum(res$orphan),
            format(round(pop_flagged), big.mark = ",")
          ),
          "wf_warning_data",
          list(group = g)
        )
      }
    }

    pop_by_res <- tapply(jdf$pop[!res$orphan], res$resolved[!res$orphan], sum)
    lvl_of_raw <- stats::setNames(res$level_used, res$raw_key)
    s_raw <- .wf_cell_key(smat, dvars)
    s_res <- rep(NA_character_, length(sel))

    in_pop <- s_raw %in% res$raw_key
    if (any(in_pop)) {
      wr <- which(in_pop)
      lv_i <- as.integer(lvl_of_raw[s_raw[in_pop]])
      for (lv in unique(lv_i)) {
        rows_lv <- wr[lv_i == lv]
        s_res[rows_lv] <- .wf_cell_key(
          .wf_apply_ladder(smat[rows_lv, , drop = FALSE], ladder, lv),
          dvars
        )
      }
    }

    if (any(!in_pop)) {
      valid_res <- unique(res$resolved)
      todo <- which(!in_pop)
      for (lv in 0:(length(res$s_keys) - 1)) {
        if (length(todo) == 0) {
          break
        }
        k <- .wf_cell_key(
          .wf_apply_ladder(smat[todo, , drop = FALSE], ladder, lv),
          dvars
        )
        hit <- k %in% valid_res
        s_res[todo[hit]] <- k[hit]
        todo <- todo[!hit]
      }
      n_drop <- sum(is.na(s_res))
      if (n_drop > 0) {
        wf_warn(
          sprintf(
            "Group '%s': %d respondent(s) fall in cells with no populated resolved match; excluded.",
            g,
            n_drop
          ),
          "wf_warning_data",
          list(group = g)
        )
      }
    }

    known <- !is.na(s_res)
    iw_g <- iw[sel]
    denom <- tapply(iw_g[known], s_res[known], sum)
    common <- intersect(names(pop_by_res), names(denom))
    factor <- stats::setNames(pop_by_res[common] / denom[common], common)

    w <- rep(NA_real_, length(sel))
    w[known] <- iw_g[known] * factor[s_res[known]]
    w[is.na(w)] <- 0

    full_total <- target$groups[[g]]$total
    expected_total <- full_total - pop_flagged
    realized <- sum(w)
    constraint_dev <- if (expected_total > 0) {
      abs(realized - expected_total) / expected_total
    } else {
      0
    }
    if (constraint_dev > tol) {
      wf_abort(
        sprintf(
          "Group '%s': group-total constraint violated (realized %.6g vs target %.6g, dev %.2e). This indicates an internal bug.",
          g,
          realized,
          expected_total,
          constraint_dev
        ),
        "wf_error_internal",
        list(group = g, dev = constraint_dev)
      )
    }
    total_dev <- if (full_total > 0) abs(realized - full_total) / full_total else 0

    wt_known <- w[known]
    cv <- if (length(wt_known) > 1 && mean(wt_known) > 0) {
      stats::sd(wt_known) / mean(wt_known)
    } else {
      NA_real_
    }

    res_row <- data.frame(
      id = ids[sel],
      group = g,
      resolved_cell = s_res,
      weight = w,
      feature = 1 / w,
      stringsAsFactors = FALSE
    )

    cellw_mean <- tapply(w[known], s_res[known], mean)
    rep_g <- data.frame(
      group = g,
      jdf[, dvars, drop = FALSE],
      pop = jdf$pop,
      n_sample = res$supp_final,
      ladder_level_used = res$level_used,
      resolved_cell = res$resolved,
      granularity_used = res$granularity_used,
      orphan = res$orphan,
      redistributed_to = NA_character_,
      pop_moved = 0,
      final_cell_weight_mean = unname(cellw_mean[res$resolved]),
      stringsAsFactors = FALSE
    )
    if (!is.null(res$redist_log)) {
      moved_idx <- match(res$redist_log$raw_cell, raw_jkey)
      hit <- !is.na(moved_idx)
      rep_g$redistributed_to[moved_idx[hit]] <- res$redist_log$receiving_cell[hit]
      rep_g$pop_moved[moved_idx[hit]] <- rep_g$pop[moved_idx[hit]]
    }

    map_g <- data.frame(
      group = g,
      raw_cell = res$raw_key,
      resolved_cell = res$resolved,
      level = res$level_used,
      stringsAsFactors = FALSE
    )

    log_row <- data.frame(
      group = g,
      n = length(sel),
      n_cells_raw = nrow(jdf),
      n_cells_resolved = length(unique(res$resolved)),
      granularity_used = res$granularity_used,
      n_orphan_cells = if (is.null(res$redist_log)) sum(res$orphan) else nrow(res$redist_log),
      pop_redistributed = sum(rep_g$pop_moved),
      total_target = full_total,
      total_realized = realized,
      total_dev = total_dev,
      iterations = NA_integer_,
      converged = TRUE,
      trimmed = 0L,
      deff = round(1 + cv^2, 3),
      stringsAsFactors = FALSE
    )
    list(data = res_row, log = log_row, report = rep_g, map = map_g)
  }

  groups <- intersect(names(target$joint), unique(gkey))
  group_results <- .wf_parallel_map(
    as.list(groups),
    fit_group,
    use_parallel = parallel,
    progress = progress,
    label = "wf_poststrat"
  )
  names(group_results) <- groups
  res_rows <- lapply(group_results, `[[`, "data")
  logs <- lapply(group_results, `[[`, "log")
  reports <- lapply(group_results, `[[`, "report")
  maps <- lapply(group_results, `[[`, "map")
  names(res_rows) <- groups
  names(logs) <- groups
  names(reports) <- groups
  names(maps) <- groups

  structure(
    list(
      data = do.call(rbind, res_rows),
      log = do.call(rbind, logs),
      achieved = NULL,
      cell_report = do.call(rbind, reports),
      collapse_map = structure(
        list(
          map = do.call(rbind, maps),
          settings = list(min_cell = min_cell, granularity = granularity)
        ),
        class = "wf_poststrat_plan"
      ),
      provenance = list(
        method = "poststrat",
        dims = dvars,
        by = target$by,
        min_cell = min_cell,
        granularity = granularity,
        empty_cell = empty_cell,
        init_weight = init_weight,
        parallel = isTRUE(parallel),
        parallel_workers = attr(group_results, "workers"),
        progress = isTRUE(progress),
        ladder_levels = ladder$n_levels,
        created = t0,
        elapsed = as.numeric(Sys.time() - t0, units = "secs"),
        package_version = .wf_package_version()
      )
    ),
    class = "wf_weights"
  )
}

#' Diagnose calibrated weights
#'
#' Computes per-group quality metrics for a `wf_weights` object.
#'
#' @param w A `wf_weights` object.
#' @param target Optional `wf_target` object for residual margin checks.
#' @param sample Reserved for future diagnostics.
#' @param deff_ok Design-effect threshold for an OK verdict.
#' @param deff_caveat Design-effect threshold for caveat verdict.
#'
#' @return A `wf_diagnostics` object.
#' @export
wf_diagnose <- function(w, target = NULL, sample = NULL,
                        deff_ok = 3, deff_caveat = 10) {
  out <- lapply(split(w$data, w$data$group), function(d) {
    wt <- d$weight
    cv <- stats::sd(wt) / mean(wt)
    data.frame(
      group = d$group[1],
      n = nrow(d),
      ess = round(sum(wt)^2 / sum(wt^2), 1),
      deff = round(1 + cv^2, 2),
      mean_w = mean(wt),
      min_w = min(wt),
      max_w = max(wt),
      max_over_mean = round(max(wt) / mean(wt), 1),
      stringsAsFactors = FALSE
    )
  })
  tab <- do.call(rbind, out)
  log_cols <- intersect(c("group", "iterations", "converged", "trimmed"), names(w$log))
  tab <- merge(tab, w$log[, log_cols, drop = FALSE], by = "group", all.x = TRUE)
  if (!"iterations" %in% names(tab)) {
    tab$iterations <- NA_integer_
  }
  if (!"converged" %in% names(tab)) {
    tab$converged <- TRUE
  }
  if (!"trimmed" %in% names(tab)) {
    tab$trimmed <- 0L
  }

  if (!is.null(target) && !is.null(w$achieved)) {
    merr <- vapply(tab$group, function(g) {
      e <- 0
      for (d in target$dims) {
        tg <- target$groups[[g]]$margins[[d]]
        cur <- w$achieved[[g]][[d]][names(tg)]
        e <- max(e, max(abs(cur - tg) / tg))
      }
      e
    }, numeric(1))
    tab$margin_maxerr <- round(merr, 4)
  }

  tab$verdict <- ifelse(
    !tab$converged,
    "FAILED",
    ifelse(
      tab$deff <= deff_ok,
      "OK",
      ifelse(tab$deff <= deff_caveat, "usable with caveats", "do not publish separately")
    )
  )
  structure(list(table = tab), class = "wf_diagnostics")
}

#' Print diagnostics
#'
#' @param x A `wf_diagnostics` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_diagnostics <- function(x, ...) {
  cat("<wf_diagnostics>\n")
  print(x$table, row.names = FALSE)
  invisible(x)
}

#' Abort with a classed WFC condition.
#'
#' @param message Error message.
#' @param class Primary condition class.
#' @param data Machine-readable condition payload.
#' @keywords internal
#' @noRd
wf_abort <- function(message, class, data = list()) {
  stop(structure(
    class = c(class, "wf_error", "error", "condition"),
    list(message = message, call = sys.call(-1), data = data)
  ))
}

#' Warn with a classed WFC condition.
#'
#' @param message Warning message.
#' @param class Primary condition class.
#' @param data Machine-readable condition payload.
#' @keywords internal
#' @noRd
wf_warn <- function(message, class, data = list()) {
  warning(structure(
    class = c(class, "wf_warning", "warning", "condition"),
    list(message = message, call = sys.call(-1), data = data)
  ))
}

#' Warn about a deprecated WFC API.
#'
#' @param message Warning message.
#' @param feature Deprecated feature name.
#' @param replacement Suggested replacement, if any.
#' @param removal Planned removal version.
#' @param risk_code Stable risk identifier, if applicable.
#' @keywords internal
#' @noRd
.wf_warn_deprecated <- function(message, feature = NULL, replacement = NULL,
                                removal = "2.0.0", risk_code = NULL) {
  wf_warn(
    message,
    "wf_warning_deprecated",
    list(
      feature = feature,
      replacement = replacement,
      removal = removal,
      risk_code = risk_code
    )
  )
}

test_that("2.0 release preserves the historical 1.0 API freeze", {
  expect_equal(as.character(utils::packageVersion("WFC")), "2.0.1")

  freeze <- system.file("stability/api-freeze.md", package = "WFC")
  expect_true(nzchar(freeze))
  text <- readLines(freeze, warn = FALSE)

  expect_true(any(grepl("WFC 1.0 API Freeze", text, fixed = TRUE)))
  expect_true(any(grepl("wf_rake(sample, target", text, fixed = TRUE)))
  expect_true(any(grepl("parallel = FALSE, progress = FALSE", text, fixed = TRUE)))
  expect_true(any(grepl("wf_warning_deprecated", text, fixed = TRUE)))
  expect_true(any(grepl("WFCstudio beta targets WFC `>= 1.0.0, < 2.0.0`", text, fixed = TRUE)))
})

test_that("deprecated API warnings use the frozen class", {
  expect_warning(
    .wf_warn_deprecated(
      "old_arg is deprecated; use new_arg.",
      feature = "old_arg",
      replacement = "new_arg"
    ),
    class = "wf_warning_deprecated"
  )
})

#' Build one fixed descriptive impact row.
#'
#' @param outcome Outcome column name.
#' @param level_label Reported category level or `NA_character_` for a mean.
#' @param values Numeric analysis values.
#' @param weights Aligned locked weights.
#' @param confidence Confidence level.
#' @param proportion Whether `values` is a zero-one indicator.
#' @keywords internal
#' @noRd
.wf_impact_row <- function(outcome, level_label, values, weights,
                           confidence, proportion) {
  keep <- !is.na(values)
  values <- values[keep]
  weights <- weights[keep]
  if (!length(values)) {
    .wf_safety_abort(
      "outcome_all_missing",
      sprintf("Outcome '%s' has no non-missing values.", outcome),
      outcome
    )
  }

  unweighted <- mean(values)
  weighted <- sum(weights * values) / sum(weights)
  kish_ess <- sum(weights)^2 / sum(weights^2)
  if (proportion) {
    variance <- weighted * (1 - weighted)
  } else {
    variance <- sum(weights * (values - weighted)^2) / sum(weights)
  }
  se <- if (kish_ess > 1) sqrt(variance / kish_ess) else NA_real_
  z <- stats::qnorm((1 + confidence) / 2)
  conf_low <- weighted - z * se
  conf_high <- weighted + z * se
  if (proportion) {
    conf_low <- max(0, conf_low)
    conf_high <- min(1, conf_high)
  }

  data.frame(
    outcome = outcome,
    level = level_label,
    n = length(values),
    unweighted = unweighted,
    weighted = weighted,
    difference = weighted - unweighted,
    kish_ess = kish_ess,
    se = se,
    conf_low = conf_low,
    conf_high = conf_high,
    stringsAsFactors = FALSE
  )
}

#' Calculate the identity of a post-lock impact result.
#'
#' @param impact Impact fields.
#' @keywords internal
#' @noRd
.wf_impact_identity <- function(impact) {
  .wf_sha256_object(impact[setdiff(names(impact), c("created", "identity"))])
}

#' Assess descriptive outcome impact after weights are locked
#'
#' Compares fixed unweighted and locked-weight estimates. This function cannot
#' accept target values, selection callbacks, or planning controls, and it never
#' recalculates weights.
#'
#' @param weights An unchanged `wf_locked_weights` object.
#' @param data Analysis data frame with the exact locked unit-ID set.
#' @param id Unique identifier column in `data`.
#' @param outcomes Character vector naming outcome columns.
#' @param level Normal-approximation confidence level.
#'
#' @return A `wf_impact` object with fixed descriptive summaries.
#' @export
wf_assess_impact <- function(weights, data, id, outcomes, level = 0.95) {
  if (!inherits(weights, "wf_locked_weights") ||
      !identical(weights$identity, .wf_locked_weight_identity(weights))) {
    .wf_safety_abort(
      "locked_weights_invalid",
      "`weights` must be unchanged locked weights.",
      "weights"
    )
  }
  if (!is.data.frame(data)) {
    .wf_safety_abort(
      "analysis_data_invalid",
      "`data` must be a data frame.",
      "data"
    )
  }
  if (!is.character(outcomes) || !length(outcomes) || anyNA(outcomes) ||
      any(!nzchar(outcomes)) || anyDuplicated(outcomes)) {
    .wf_safety_abort(
      "outcomes_invalid",
      "`outcomes` must be unique, non-empty column names.",
      "outcomes"
    )
  }
  if (id %in% outcomes) {
    .wf_safety_abort(
      "outcome_is_identifier",
      "The identifier column cannot be assessed as an outcome.",
      "outcomes"
    )
  }
  .require_cols(data, outcomes, "analysis data")
  if (!is.numeric(level) || length(level) != 1L || is.na(level) ||
      !is.finite(level) || level <= 0.5 || level >= 1) {
    .wf_safety_abort(
      "impact_level_invalid",
      "`level` must be one number between 0.5 and 1.",
      "level"
    )
  }

  index <- .wf_match_unit_ids(
    data,
    id,
    weights$data$id,
    "locked weights",
    "wf_assess_impact()"
  )
  aligned_weights <- weights$data$weight[index]
  rows <- list()
  row_index <- 0L

  for (outcome in outcomes) {
    values <- data[[outcome]]
    if (inherits(values, c("Date", "POSIXt")) || is.list(values) ||
        is.matrix(values) || length(dim(values))) {
      .wf_safety_abort(
        "outcome_type_unsupported",
        sprintf("Outcome '%s' has an unsupported structured type.", outcome),
        outcome,
        next_actions = "supply_numeric_logical_or_factor_outcome"
      )
    }

    if (is.logical(values)) {
      row_index <- row_index + 1L
      rows[[row_index]] <- .wf_impact_row(
        outcome,
        "TRUE",
        as.numeric(values),
        aligned_weights,
        level,
        TRUE
      )
    } else if (is.factor(values)) {
      declared_levels <- levels(values)
      if (!length(declared_levels)) {
        .wf_safety_abort(
          "outcome_levels_missing",
          sprintf("Factor outcome '%s' has no declared levels.", outcome),
          outcome
        )
      }
      report_levels <- if (length(declared_levels) == 2L) {
        declared_levels[[2]]
      } else {
        declared_levels
      }
      for (category in report_levels) {
        row_index <- row_index + 1L
        indicator <- ifelse(is.na(values), NA_real_, as.numeric(values == category))
        rows[[row_index]] <- .wf_impact_row(
          outcome,
          category,
          indicator,
          aligned_weights,
          level,
          TRUE
        )
      }
    } else if (is.numeric(values)) {
      if (any(!is.finite(values[!is.na(values)]))) {
        .wf_safety_abort(
          "outcome_nonfinite",
          sprintf("Outcome '%s' contains non-finite values.", outcome),
          outcome
        )
      }
      observed <- unique(values[!is.na(values)])
      binary <- length(observed) <= 2L && all(observed %in% c(0, 1))
      row_index <- row_index + 1L
      rows[[row_index]] <- .wf_impact_row(
        outcome,
        if (binary) "1" else NA_character_,
        as.numeric(values),
        aligned_weights,
        level,
        binary
      )
    } else {
      .wf_safety_abort(
        "outcome_type_unsupported",
        sprintf(
          "Outcome '%s' must be numeric, logical, or a factor with declared levels.",
          outcome
        ),
        outcome,
        next_actions = "recode_outcome_explicitly"
      )
    }
  }

  impact <- list(
    summary = do.call(rbind, rows),
    weight_identity = weights$identity,
    outcomes = outcomes,
    level = level,
    created = .wf_iso_time(),
    package_version = .wf_package_version()
  )
  row.names(impact$summary) <- NULL
  impact$identity <- .wf_impact_identity(impact)
  class(impact) <- "wf_impact"
  impact
}

#' Print post-lock outcome impact
#'
#' @param x A `wf_impact` object.
#' @param ... Reserved for future use.
#'
#' @return `x`, invisibly.
#' @export
print.wf_impact <- function(x, ...) {
  cat("<wf_impact>\n")
  cat(" Locked weight identity: ", x$weight_identity, "\n", sep = "")
  cat(" Outcome rows: ", nrow(x$summary), "\n", sep = "")
  print(x$summary, row.names = FALSE)
  invisible(x)
}

#' Normalize calibration keys to trimmed character values.
#'
#' @param x Vector to normalize.
#' @return A character vector.
#' @keywords internal
#' @noRd
.chr <- function(x) {
  trimws(as.character(x))
}

#' Require columns in a data frame.
#'
#' @param df Data frame to check.
#' @param cols Required column names.
#' @param what Human-readable data source name.
#' @keywords internal
#' @noRd
.require_cols <- function(df, cols, what) {
  miss <- setdiff(cols, names(df))
  if (length(miss) > 0) {
    wf_abort(sprintf(
      "%s is missing column(s): %s",
      what,
      paste(miss, collapse = ", ")
    ), "wf_error_schema", list(missing = miss))
  }
}

#' Create a validated target object.
#'
#' @param mode Target mode.
#' @param by Grouping variable.
#' @param dims Dimension names.
#' @param groups Target groups.
#' @param meta Provenance metadata.
#' @keywords internal
#' @noRd
.wf_new_target <- function(mode, by, dims, groups, meta = list(), joint = NULL) {
  tgt <- structure(
    list(mode = mode, by = by, dims = dims, groups = groups, meta = meta, joint = joint),
    class = "wf_target"
  )
  .wf_validate_target(tgt)
  tgt
}

#' Validate target invariants.
#'
#' @param tgt Target object.
#' @param tol Relative tolerance.
#' @keywords internal
#' @noRd
.wf_validate_target <- function(tgt, tol = 1e-8) {
  for (g in names(tgt$groups)) {
    gr <- tgt$groups[[g]]
    if (!is.finite(gr$total) || gr$total <= 0) {
      wf_abort(sprintf(
        "Group '%s': target total must be positive and finite.",
        g
      ), "wf_error_input", list(group = g, total = gr$total))
    }
    for (d in tgt$dims) {
      m <- gr$margins[[d]]
      if (is.null(m) || is.null(names(m))) {
        wf_abort(sprintf(
          "Group '%s', dim '%s': margins missing or unnamed.",
          g,
          d
        ), "wf_error_internal", list(group = g, dim = d))
      }
      if (any(!is.finite(m)) || any(m < 0)) {
        wf_abort(sprintf(
          "Group '%s', dim '%s': margins must be finite and >= 0.",
          g,
          d
        ), "wf_error_input", list(group = g, dim = d))
      }
      if (abs(sum(m) - gr$total) > tol * gr$total) {
        wf_abort(sprintf(
          "Group '%s', dim '%s': margins sum to %.6g but total is %.6g. %s",
          g,
          d,
          sum(m),
          gr$total,
          "Additivity is required so IPF preserves the group total."
        ), "wf_error_input", list(group = g, dim = d))
      }
    }
  }
  invisible(TRUE)
}

#' Resolve group keys.
#'
#' @param df Data frame.
#' @param by Grouping variable name or `NULL`.
#' @param by_key Optional group key column or function.
#' @keywords internal
#' @noRd
.wf_group_keys <- function(df, by, by_key = NULL) {
  if (!is.null(by_key)) {
    if (is.function(by_key)) {
      return(.chr(by_key(df)))
    }
    .require_cols(df, by_key, "population data")
    return(.chr(df[[by_key]]))
  }
  if (is.null(by)) {
    return(rep("_all_", nrow(df)))
  }
  .require_cols(df, by, "data")
  .chr(df[[by]])
}

#' Rescale target groups.
#'
#' @param groups Target group list.
#' @param scale Scale mode.
#' @param sample_n Per-group sample sizes.
#' @param totals Custom totals.
#' @keywords internal
#' @noRd
.wf_scale_groups <- function(groups, scale, sample_n = NULL, totals = NULL) {
  if (scale == "population") {
    return(groups)
  }
  for (g in names(groups)) {
    new_total <-
      if (scale == "sample") {
        if (is.null(sample_n) || is.na(sample_n[g])) {
          wf_abort(sprintf(
            "scale='sample' needs the sample: no size found for group '%s'.",
            g
          ), "wf_error_input", list(group = g))
        }
        sample_n[g]
      } else {
        if (is.null(totals) || is.na(totals[g])) {
          wf_abort(sprintf(
            "scale='custom' requires totals['%s'].",
            g
          ), "wf_error_input", list(group = g))
        }
        totals[g]
      }
    f <- new_total / groups[[g]]$total
    groups[[g]]$total <- unname(new_total)
    groups[[g]]$margins <- lapply(groups[[g]]$margins, function(m) m * f)
  }
  groups
}

#' Sum weights by integer group index.
#'
#' @param w Weight vector.
#' @param idx Integer group indices.
#' @param K Number of groups.
#' @keywords internal
#' @noRd
.grp_sum <- function(w, idx, K) {
  out <- numeric(K)
  rs <- rowsum(w, idx)
  out[as.integer(rownames(rs))] <- rs[, 1]
  out
}

#' Return the loaded package version for provenance records.
#'
#' @return The installed package version as a character scalar, or the
#'   development fallback when the package is not installed.
#' @keywords internal
#' @noRd
.wf_package_version <- function() {
  tryCatch(
    as.character(utils::packageVersion("WFC")),
    error = function(e) "dev"
  )
}

#' Resolve the number of workers for optional fork parallelism.
#'
#' @keywords internal
#' @noRd
.wf_parallel_workers <- function(n, use_parallel) {
  if (!isTRUE(use_parallel) || n <= 1L) {
    return(1L)
  }
  if (identical(.Platform$OS.type, "windows")) {
    message("parallel=TRUE uses forked parallelism; Windows falls back to serial execution.")
    return(1L)
  }
  cores <- getOption("wfc.parallel.cores", NULL)
  if (is.null(cores)) {
    cores <- parallel::detectCores(logical = FALSE)
    cores <- if (is.na(cores)) 2L else max(1L, cores - 1L)
  }
  cores <- suppressWarnings(as.integer(cores))
  if (length(cores) != 1L || is.na(cores) || cores < 1L) {
    cores <- 1L
  }
  # CRAN policy requires packages to use no more than two cores at once,
  # including when parallel execution is explicitly requested by the user.
  min(n, cores, 2L)
}

#' Start a cli progress bar when requested and available.
#'
#' @keywords internal
#' @noRd
.wf_progress_start <- function(n, progress, label) {
  if (!isTRUE(progress) || n <= 0L ||
      !requireNamespace("cli", quietly = TRUE)) {
    return(NULL)
  }
  tryCatch(
    cli::cli_progress_bar(name = label, total = n, clear = FALSE),
    error = function(e) NULL
  )
}

#' Update an optional cli progress bar.
#'
#' @keywords internal
#' @noRd
.wf_progress_tick <- function(id, inc = 1L) {
  if (!is.null(id)) {
    tryCatch(
      cli::cli_progress_update(id = id, inc = inc),
      error = function(e) NULL
    )
  }
}

#' Finish an optional cli progress bar.
#'
#' @keywords internal
#' @noRd
.wf_progress_done <- function(id) {
  if (!is.null(id)) {
    tryCatch(
      cli::cli_progress_done(id = id),
      error = function(e) NULL
    )
  }
}

#' Evaluate one mapped task and retain classed warnings/errors.
#'
#' @keywords internal
#' @noRd
.wf_eval_task <- function(x, fun) {
  warnings <- list()
  value <- tryCatch(
    withCallingHandlers(
      fun(x),
      warning = function(w) {
        warnings[[length(warnings) + 1L]] <<- w
        invokeRestart("muffleWarning")
      }
    ),
    error = function(e) e
  )
  list(ok = !inherits(value, "error"), value = value, warnings = warnings)
}

#' Normalize a child-process result into the internal task envelope.
#'
#' @keywords internal
#' @noRd
.wf_normalize_task_result <- function(x) {
  if (inherits(x, "try-error")) {
    return(list(
      ok = FALSE,
      value = simpleError(as.character(x)),
      warnings = list()
    ))
  }
  if (is.list(x) && !is.null(x$ok) && !is.null(x$warnings)) {
    return(x)
  }
  list(
    ok = FALSE,
    value = simpleError("Parallel worker returned an invalid result."),
    warnings = list()
  )
}

#' Re-emit task warnings and return its value, or rethrow its error.
#'
#' @keywords internal
#' @noRd
.wf_task_value <- function(x) {
  x <- .wf_normalize_task_result(x)
  for (w in x$warnings) {
    warning(w)
  }
  if (!isTRUE(x$ok)) {
    stop(x$value)
  }
  x$value
}

#' Run a list map with optional fork parallelism and progress reporting.
#'
#' @keywords internal
#' @noRd
.wf_parallel_map <- function(x, fun, use_parallel = FALSE, progress = FALSE,
                             label = "WFC tasks") {
  n <- length(x)
  workers <- .wf_parallel_workers(n, use_parallel)
  pb <- .wf_progress_start(n, progress, label)
  on.exit(.wf_progress_done(pb), add = TRUE)

  eval_index <- function(i) .wf_eval_task(x[[i]], fun)
  if (workers > 1L && !is.null(pb)) {
    raw <- .wf_parallel_map_poll(seq_len(n), eval_index, workers, pb)
  } else if (workers > 1L) {
    raw <- parallel::mclapply(seq_len(n), eval_index, mc.cores = workers)
    .wf_progress_tick(pb, n)
  } else {
    raw <- vector("list", n)
    for (i in seq_len(n)) {
      raw[[i]] <- eval_index(i)
      .wf_progress_tick(pb)
    }
  }

  out <- lapply(raw, .wf_task_value)
  attr(out, "workers") <- workers
  attr(out, "parallel") <- workers > 1L
  out
}

#' Forked map implementation that can tick progress as jobs finish.
#'
#' @keywords internal
#' @noRd
.wf_parallel_map_poll <- function(indices, eval_index, workers, pb) {
  out <- vector("list", length(indices))
  active <- list()
  next_pos <- 1L

  launch <- function() {
    i <- indices[[next_pos]]
    proc <- parallel::mcparallel(eval_index(i), silent = TRUE)
    active[[as.character(proc$pid)]] <<- list(index = i, proc = proc)
    next_pos <<- next_pos + 1L
  }

  while (next_pos <= length(indices) && length(active) < workers) {
    launch()
  }
  while (length(active) > 0L) {
    collected <- parallel::mccollect(
      lapply(active, `[[`, "proc"),
      wait = FALSE,
      timeout = 0.1
    )
    if (is.null(collected) || length(collected) == 0L) {
      Sys.sleep(0.05)
      next
    }
    for (pid in names(collected)) {
      job <- active[[pid]]
      out[[job$index]] <- .wf_normalize_task_result(collected[[pid]])
      active[[pid]] <- NULL
      .wf_progress_tick(pb)
      if (next_pos <= length(indices)) {
        launch()
      }
    }
  }
  out
}

#' Convert grouped sums to a named numeric vector.
#'
#' @param x One-dimensional grouped-sum result.
#' @return A named numeric vector.
#' @keywords internal
#' @noRd
.wf_margin_vector <- function(x) {
  stats::setNames(as.numeric(x), names(x))
}

#' Aggregate design support into deterministic cells.
#'
#' @param data Design data frame.
#' @param boundary_values Character boundary key per row.
#' @param dims Calibration dimension names.
#' @param base_weight Base-weight vector.
#' @keywords internal
#' @noRd
.wf_cell_table <- function(data, boundary_values, dims, base_weight) {
  cell_data <- data.frame(
    boundary = .chr(boundary_values),
    lapply(data[dims], .chr),
    n = rep(1, nrow(data)),
    base_weight = base_weight,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )
  cells <- stats::aggregate(
    cell_data[c("n", "base_weight")],
    cell_data[c("boundary", dims)],
    sum
  )
  order_args <- c(cells[c("boundary", dims)], list(method = "radix"))
  cells <- cells[do.call(order, order_args), , drop = FALSE]
  row.names(cells) <- NULL
  cells
}

#' Build original-to-final category maps for every boundary.
#'
#' @param design A `wf_design_data` object.
#' @param target A `wf_verified_target` object.
#' @param dims A `wf_dims` object.
#' @param boundary_values Character boundary key per design row.
#' @keywords internal
#' @noRd
.wf_initial_category_maps <- function(design, target, dims, boundary_values) {
  rows <- list()
  index <- 0L
  for (boundary in names(target$groups)) {
    selected <- boundary_values == boundary
    for (dim in names(dims$vars)) {
      categories <- unique(c(
        dims$vars[[dim]],
        names(target$groups[[boundary]]$margins[[dim]]),
        .chr(design$data[[dim]][selected])
      ))
      categories <- categories[!is.na(categories)]
      for (category in categories) {
        index <- index + 1L
        rows[[index]] <- data.frame(
          boundary = boundary,
          dim = dim,
          category_before = category,
          category_after = category,
          stringsAsFactors = FALSE
        )
      }
    }
  }
  maps <- do.call(rbind, rows)
  row.names(maps) <- NULL
  maps
}

#' Apply stored category maps to design data.
#'
#' @param data Design data frame.
#' @param boundary_values Character boundary key per row.
#' @param maps Stored original-to-final category maps.
#' @param dims Calibration dimension names.
#' @keywords internal
#' @noRd
.wf_map_design_categories <- function(data, boundary_values, maps, dims) {
  out <- data
  for (boundary in unique(boundary_values)) {
    selected <- boundary_values == boundary
    for (dim in dims) {
      map <- maps[maps$boundary == boundary & maps$dim == dim, , drop = FALSE]
      values <- .chr(out[[dim]][selected])
      matched <- match(values, map$category_before)
      hit <- !is.na(matched)
      values[hit] <- map$category_after[matched[hit]]
      out[[dim]][selected] <- values
    }
  }
  out
}

#' Apply stored category maps to target margins.
#'
#' @param target A `wf_target` object.
#' @param maps Stored original-to-final category maps.
#' @keywords internal
#' @noRd
.wf_map_target_categories <- function(target, maps) {
  out <- target
  for (boundary in names(out$groups)) {
    for (dim in out$dims) {
      margin <- out$groups[[boundary]]$margins[[dim]]
      map <- maps[maps$boundary == boundary & maps$dim == dim, , drop = FALSE]
      categories <- names(margin)
      matched <- match(categories, map$category_before)
      hit <- !is.na(matched)
      categories[hit] <- map$category_after[matched[hit]]
      out$groups[[boundary]]$margins[[dim]] <- .wf_margin_vector(
        tapply(as.numeric(margin), categories, sum)
      )
    }
  }
  if (!is.null(out$joint)) {
    for (boundary in names(out$joint)) {
      joint <- out$joint[[boundary]]
      for (dim in out$dims) {
        map <- maps[
          maps$boundary == boundary & maps$dim == dim,
          ,
          drop = FALSE
        ]
        categories <- .chr(joint[[dim]])
        matched <- match(categories, map$category_before)
        hit <- !is.na(matched)
        categories[hit] <- map$category_after[matched[hit]]
        joint[[dim]] <- categories
      }
      out$joint[[boundary]] <- stats::aggregate(
        joint["pop"],
        joint[out$dims],
        sum
      )
    }
  }
  .wf_validate_target(out)
  out
}

#' Calculate a conservative marginal weight-ratio projection.
#'
#' @param data Mapped design data.
#' @param boundary_values Character boundary key per row.
#' @param base_weight Base-weight vector.
#' @param target Mapped target.
#' @param dims Calibration dimension names.
#' @param boundary Boundary to evaluate.
#' @keywords internal
#' @noRd
.wf_boundary_weight_ratio <- function(data, boundary_values, base_weight,
                                      target, dims, boundary) {
  selected <- boundary_values == boundary
  if (!any(selected)) {
    return(Inf)
  }
  sample_total <- sum(base_weight[selected])
  target_total <- target$groups[[boundary]]$total
  ratios <- numeric()

  for (dim in dims) {
    sample_margin <- tapply(
      base_weight[selected],
      .chr(data[[dim]][selected]),
      sum
    )
    target_margin <- target$groups[[boundary]]$margins[[dim]]
    categories <- union(names(sample_margin), names(target_margin))
    sample_share <- stats::setNames(rep(0, length(categories)), categories)
    target_share <- sample_share
    sample_share[names(sample_margin)] <- sample_margin / sample_total
    target_share[names(target_margin)] <- target_margin / target_total

    forward <- target_share / sample_share
    reverse <- sample_share / target_share
    ratios <- c(ratios, forward, reverse)
  }

  ratios[is.nan(ratios)] <- 1
  max(ratios)
}

#' Update all original categories currently represented by one category.
#'
#' @param maps Category map table.
#' @param boundary Boundary key.
#' @param dim Dimension name.
#' @param from Current category to absorb.
#' @param to Current category that remains.
#' @keywords internal
#' @noRd
.wf_merge_category <- function(maps, boundary, dim, from, to) {
  selected <- maps$boundary == boundary & maps$dim == dim &
    maps$category_after == from
  maps$category_after[selected] <- to
  maps
}

#' Evaluate one boundary under a category map.
#'
#' @param design_data Original design data frame.
#' @param boundary_values Character boundary key per row.
#' @param base_weight Base-weight vector.
#' @param target Original target.
#' @param maps Category map table.
#' @param dims Calibration dimension names.
#' @param boundary Boundary key.
#' @param min_cell Minimum supported cell size.
#' @keywords internal
#' @noRd
.wf_boundary_state <- function(design_data, boundary_values, base_weight,
                               target, maps, dims, boundary, min_cell) {
  mapped_data <- .wf_map_design_categories(
    design_data,
    boundary_values,
    maps,
    dims
  )
  mapped_target <- .wf_map_target_categories(target, maps)
  selected <- boundary_values == boundary
  cells <- .wf_cell_table(
    mapped_data[selected, , drop = FALSE],
    boundary_values[selected],
    dims,
    base_weight[selected]
  )
  ratio <- .wf_boundary_weight_ratio(
    mapped_data,
    boundary_values,
    base_weight,
    mapped_target,
    dims,
    boundary
  )
  list(
    data = mapped_data,
    target = mapped_target,
    cells = cells,
    thin = sum(cells$n < min_cell),
    ratio = ratio
  )
}

#' Select a deterministic ordered-adjacency merge.
#'
#' @param state Current boundary state.
#' @param maps Current category maps.
#' @param design_data Original design data frame.
#' @param boundary_values Character boundary key per row.
#' @param base_weight Base-weight vector.
#' @param target Original target.
#' @param dims A `wf_dims` object.
#' @param boundary Boundary key.
#' @param min_cell Minimum supported cell size.
#' @param max_weight_ratio Maximum permitted projected ratio.
#' @keywords internal
#' @noRd
.wf_choose_adjacency_merge <- function(state, maps, design_data,
                                       boundary_values, base_weight, target,
                                       dims, boundary, min_cell,
                                       max_weight_ratio) {
  candidates <- list()
  index <- 0L
  current_bad <- state$thin + as.integer(state$ratio > max_weight_ratio)

  for (dim in names(dims$vars)) {
    map <- maps[maps$boundary == boundary & maps$dim == dim, , drop = FALSE]
    current <- unique(map$category_after)
    declared <- dims$vars[[dim]]
    ranks <- vapply(current, function(category) {
      originals <- map$category_before[map$category_after == category]
      positions <- match(originals, declared)
      if (all(is.na(positions))) Inf else min(positions, na.rm = TRUE)
    }, numeric(1))
    current <- current[order(ranks, current, method = "radix")]
    if (length(current) < 2L) {
      next
    }

    pairs <- cbind(current[-length(current)], current[-1L])
    directions <- rbind(pairs, pairs[, 2:1, drop = FALSE])
    for (i in seq_len(nrow(directions))) {
      from <- directions[i, 1]
      to <- directions[i, 2]
      candidate_maps <- .wf_merge_category(maps, boundary, dim, from, to)
      candidate_state <- .wf_boundary_state(
        design_data,
        boundary_values,
        base_weight,
        target,
        candidate_maps,
        names(dims$vars),
        boundary,
        min_cell
      )
      candidate_bad <- candidate_state$thin +
        as.integer(candidate_state$ratio > max_weight_ratio)
      improves <- candidate_bad < current_bad ||
        candidate_state$ratio < state$ratio
      if (!improves) {
        next
      }

      selected <- boundary_values == boundary & state$data[[dim]] == from
      index <- index + 1L
      candidates[[index]] <- list(
        maps = candidate_maps,
        state = candidate_state,
        improvement = current_bad - candidate_bad,
        moved = sum(selected),
        distance = 1,
        distortion = sum(base_weight[selected]),
        dim = dim,
        from = from,
        to = to
      )
    }
  }

  if (!length(candidates)) {
    return(NULL)
  }
  score <- data.frame(
    improvement = vapply(candidates, `[[`, numeric(1), "improvement"),
    moved = vapply(candidates, `[[`, numeric(1), "moved"),
    distance = vapply(candidates, `[[`, numeric(1), "distance"),
    distortion = vapply(candidates, `[[`, numeric(1), "distortion"),
    dim = vapply(candidates, `[[`, character(1), "dim"),
    from = vapply(candidates, `[[`, character(1), "from"),
    to = vapply(candidates, `[[`, character(1), "to"),
    stringsAsFactors = FALSE
  )
  selected <- order(
    -score$improvement,
    score$moved,
    score$distance,
    score$distortion,
    score$dim,
    score$from,
    score$to,
    method = "radix"
  )[1]
  candidates[[selected]]
}

#' Calculate the identity of a stored cell merge plan.
#'
#' @param plan Cell plan fields.
#' @keywords internal
#' @noRd
.wf_cell_plan_identity <- function(plan) {
  .wf_sha256_object(plan[setdiff(names(plan), c("created", "identity"))])
}

#' Plan deterministic, outcome-blind support-cell merging
#'
#' Uses only declared design fields, external target margins, ordered dimension
#' levels, and an optional explicit collapse ladder. Study outcomes and custom
#' scoring callbacks are not accepted by this interface.
#'
#' @param design A `wf_design_data` object.
#' @param target A non-demo `wf_verified_target` object.
#' @param dims A `wf_dims` object.
#' @param min_cell Minimum observed count in every retained joint design cell.
#' @param max_weight_ratio Maximum conservative marginal target-to-design ratio.
#' @param boundary Optional grouping column; defaults to the target grouping.
#' @param ladder Optional explicit `wf_collapse_ladder`.
#'
#' @return A reviewable `wf_cell_merge_plan`. No weights are computed.
#' @export
wf_plan_cells <- function(design, target, dims, min_cell = 5,
                          max_weight_ratio = 4, boundary = target$by,
                          ladder = NULL) {
  if (!inherits(design, "wf_design_data")) {
    .wf_safety_abort(
      "design_object_required",
      "`design` must be a wf_design_data object.",
      "design"
    )
  }
  if (!identical(design$identity, .wf_design_identity(design$data, design$roles))) {
    .wf_safety_abort(
      "design_identity_invalid",
      "Design data have changed since their identity was recorded.",
      "design"
    )
  }
  if (!inherits(target, "wf_verified_target") || isTRUE(target$demo_only)) {
    .wf_safety_abort(
      "verified_target_required",
      "`target` must be a non-demo verified external target.",
      "target"
    )
  }
  if (!identical(target$identity, .wf_verified_target_identity(target))) {
    .wf_safety_abort(
      "target_identity_invalid",
      "Target data have changed since their identity was recorded.",
      "target"
    )
  }
  if (!inherits(dims, "wf_dims") ||
      !identical(names(dims$vars), target$dims)) {
    .wf_safety_abort(
      "target_dimensions_mismatch",
      "Declared dimensions must exactly match the verified target.",
      "dims"
    )
  }
  if (length(min_cell) != 1L || !is.finite(min_cell) ||
      min_cell < 1 || min_cell != as.integer(min_cell)) {
    .wf_safety_abort(
      "min_cell_invalid",
      "`min_cell` must be a positive integer.",
      "min_cell"
    )
  }
  if (length(max_weight_ratio) != 1L || !is.finite(max_weight_ratio) ||
      max_weight_ratio <= 1) {
    .wf_safety_abort(
      "max_weight_ratio_invalid",
      "`max_weight_ratio` must be a finite number greater than one.",
      "max_weight_ratio"
    )
  }
  if (!is.null(boundary) && !.wf_is_string(boundary)) {
    .wf_safety_abort(
      "merge_boundary_invalid",
      "`boundary` must be one column name or NULL.",
      "boundary"
    )
  }
  if (!identical(boundary, target$by)) {
    .wf_safety_abort(
      "merge_boundary_mismatch",
      "Cell merging must remain inside the verified target boundary.",
      "boundary"
    )
  }
  if (!is.null(ladder) &&
      (!inherits(ladder, "wf_collapse_ladder") ||
       !identical(ladder$dims, names(dims$vars)))) {
    .wf_safety_abort(
      "collapse_ladder_mismatch",
      "`ladder` must match the declared dimensions.",
      "ladder"
    )
  }

  required <- c(boundary, names(dims$vars))
  .require_cols(design$data, required, "design data")
  boundary_values <- .wf_group_keys(design$data, boundary)
  missing_boundaries <- setdiff(names(target$groups), unique(boundary_values))
  if (length(missing_boundaries)) {
    wf_abort(
      "No allowed cell merge can create support in a missing boundary.",
      "wf_error_feasibility",
      .wf_safety_payload(
        "merge_boundary_unsupported",
        "blocking",
        boundary,
        list(boundaries = missing_boundaries),
        "collect_boundary_sample"
      )
    )
  }

  base_weight <- if (is.null(design$roles$base_weight)) {
    rep(1, nrow(design$data))
  } else {
    design$data[[design$roles$base_weight]]
  }
  dim_names <- names(dims$vars)
  maps <- .wf_initial_category_maps(
    design,
    target,
    dims,
    boundary_values
  )
  reasons <- list()
  reason_index <- 0L

  for (boundary_key in names(target$groups)) {
    state <- .wf_boundary_state(
      design$data,
      boundary_values,
      base_weight,
      target,
      maps,
      dim_names,
      boundary_key,
      min_cell
    )
    ladder_level <- 0L
    merge_count <- 0L

    while (state$thin > 0L || state$ratio > max_weight_ratio) {
      if (!is.null(ladder)) {
        ladder_level <- ladder_level + 1L
        if (ladder_level > ladder$n_levels) {
          candidate <- NULL
        } else {
          step <- ladder$steps[[ladder_level]]
          changed <- FALSE
          for (dim in names(step)) {
            for (from in names(step[[dim]])) {
              to <- unname(step[[dim]][[from]])
              current_hit <- maps$boundary == boundary_key &
                maps$dim == dim & maps$category_after == from
              if (any(current_hit) && !identical(from, to)) {
                maps <- .wf_merge_category(
                  maps,
                  boundary_key,
                  dim,
                  from,
                  to
                )
                changed <- TRUE
                reason_index <- reason_index + 1L
                reasons[[reason_index]] <- data.frame(
                  boundary = boundary_key,
                  step = names(ladder$steps)[ladder_level],
                  dim = dim,
                  from = from,
                  to = to,
                  reason = "explicit_ladder",
                  stringsAsFactors = FALSE
                )
              }
            }
          }
          if (changed) {
            merge_count <- merge_count + 1L
            state <- .wf_boundary_state(
              design$data,
              boundary_values,
              base_weight,
              target,
              maps,
              dim_names,
              boundary_key,
              min_cell
            )
            next
          }
          next
        }
      } else {
        candidate <- .wf_choose_adjacency_merge(
          state,
          maps,
          design$data,
          boundary_values,
          base_weight,
          target,
          dims,
          boundary_key,
          min_cell,
          max_weight_ratio
        )
      }

      if (is.null(candidate)) {
        wf_abort(
          sprintf(
            "No allowed partition satisfies min_cell=%d and max_weight_ratio=%s inside boundary '%s'.",
            as.integer(min_cell),
            format(max_weight_ratio),
            boundary_key
          ),
          "wf_error_feasibility",
          .wf_safety_payload(
            "cell_partition_infeasible",
            "blocking",
            boundary,
            list(
              boundary = boundary_key,
              thin_cells = state$thin,
              projected_max_weight_ratio = state$ratio
            ),
            "revise_declared_ladder_or_design"
          )
        )
      }

      maps <- candidate$maps
      state <- candidate$state
      merge_count <- merge_count + 1L
      reason_index <- reason_index + 1L
      reasons[[reason_index]] <- data.frame(
        boundary = boundary_key,
        step = paste0("adjacency_", merge_count),
        dim = candidate$dim,
        from = candidate$from,
        to = candidate$to,
        reason = "declared_order_adjacency",
        stringsAsFactors = FALSE
      )
    }
  }

  mapped_data <- .wf_map_design_categories(
    design$data,
    boundary_values,
    maps,
    dim_names
  )
  mapped_target <- .wf_map_target_categories(target, maps)
  cells_before <- .wf_cell_table(
    design$data,
    boundary_values,
    dim_names,
    base_weight
  )
  cells_after <- .wf_cell_table(
    mapped_data,
    boundary_values,
    dim_names,
    base_weight
  )
  projected_ratios <- vapply(names(target$groups), function(boundary_key) {
    .wf_boundary_weight_ratio(
      mapped_data,
      boundary_values,
      base_weight,
      mapped_target,
      dim_names,
      boundary_key
    )
  }, numeric(1))

  map <- data.frame(
    boundary_before = boundary_values,
    boundary_after = boundary_values,
    stringsAsFactors = FALSE
  )
  for (dim in dim_names) {
    map[[paste0(dim, "_before")]] <- .chr(design$data[[dim]])
    map[[paste0(dim, "_after")]] <- .chr(mapped_data[[dim]])
  }
  map <- unique(map)
  row.names(map) <- NULL

  changed <- vapply(seq_len(nrow(design$data)), function(i) {
    any(vapply(dim_names, function(dim) {
      !identical(.chr(design$data[[dim]][i]), .chr(mapped_data[[dim]][i]))
    }, logical(1)))
  }, logical(1))
  reason_table <- if (length(reasons)) {
    do.call(rbind, reasons)
  } else {
    data.frame(
      boundary = character(),
      step = character(),
      dim = character(),
      from = character(),
      to = character(),
      reason = character(),
      stringsAsFactors = FALSE
    )
  }
  row.names(reason_table) <- NULL

  plan <- list(
    cells_before = cells_before,
    cells_after = cells_after,
    map = map,
    category_maps = maps,
    reasons = reason_table,
    affected_share = mean(changed),
    unresolved_cells = cells_after[cells_after$n < min_cell, , drop = FALSE],
    projected_max_weight_ratio = max(projected_ratios),
    input_identities = list(design = design$identity, target = target$identity),
    settings = list(
      dims = dim_names,
      min_cell = as.integer(min_cell),
      max_weight_ratio = max_weight_ratio,
      boundary = boundary,
      ladder = ladder
    ),
    created = .wf_iso_time(),
    package_version = .wf_package_version()
  )
  plan$identity <- .wf_cell_plan_identity(plan)
  class(plan) <- "wf_cell_merge_plan"
  plan
}

#' Apply an already reviewed cell plan exactly as stored.
#'
#' @param design A `wf_design_data` object matching the plan.
#' @param target A `wf_verified_target` object matching the plan.
#' @param plan A `wf_cell_merge_plan` object.
#' @keywords internal
#' @noRd
.wf_apply_cell_plan <- function(design, target, plan) {
  if (!inherits(plan, "wf_cell_merge_plan")) {
    .wf_safety_abort(
      "cell_plan_required",
      "`plan` must be a wf_cell_merge_plan object.",
      "plan"
    )
  }
  if (!identical(plan$identity, .wf_cell_plan_identity(plan))) {
    .wf_safety_abort(
      "cell_plan_identity_invalid",
      "The reviewed cell plan has changed since its identity was recorded.",
      "plan"
    )
  }
  if (!inherits(design, "wf_design_data") ||
      !identical(design$identity, plan$input_identities$design) ||
      !identical(design$identity, .wf_design_identity(design$data, design$roles))) {
    .wf_safety_abort(
      "cell_plan_design_changed",
      "Design data no longer match the reviewed cell plan.",
      "design"
    )
  }
  if (!inherits(target, "wf_verified_target") ||
      !identical(target$identity, plan$input_identities$target) ||
      !identical(target$identity, .wf_verified_target_identity(target))) {
    .wf_safety_abort(
      "cell_plan_target_changed",
      "Target data no longer match the reviewed cell plan.",
      "target"
    )
  }

  boundary_values <- .wf_group_keys(design$data, plan$settings$boundary)
  mapped_data <- .wf_map_design_categories(
    design$data,
    boundary_values,
    plan$category_maps,
    plan$settings$dims
  )
  mapped_target <- .wf_map_target_categories(target, plan$category_maps)
  mapped_target <- .wf_verified_target(
    mapped_target,
    target$evidence,
    target$source_type
  )
  mapped_target$meta$cell_plan_identity <- plan$identity
  mapped_design <- wf_prepare_design(
    mapped_data,
    id = design$roles$id,
    calibration = design$roles$calibration,
    base_weight = design$roles$base_weight,
    strata = design$roles$strata,
    clusters = design$roles$clusters,
    fpc = design$roles$fpc
  )

  list(design = mapped_design, target = mapped_target)
}

#' Print an outcome-blind cell merge plan
#'
#' @param x A `wf_cell_merge_plan` object.
#' @param ... Reserved for future use.
#'
#' @return `x`, invisibly.
#' @export
print.wf_cell_merge_plan <- function(x, ...) {
  cat("<wf_cell_merge_plan>\n")
  cat(" Cells: ", nrow(x$cells_before), " -> ", nrow(x$cells_after), "\n", sep = "")
  cat(" Affected share: ", format(round(x$affected_share, 4)), "\n", sep = "")
  cat(
    " Projected max weight ratio: ",
    format(round(x$projected_max_weight_ratio, 4)),
    "\n",
    sep = ""
  )
  cat(" Identity: ", x$identity, "\n", sep = "")
  invisible(x)
}

make_sparse_safe_workflow_fixture <- function(with_outcomes = FALSE) {
  data <- data.frame(
    id = paste0("r", seq_len(12)),
    province = rep(c("A", "B"), each = 6),
    age = c(rep("young", 5), "old", "young", rep("old", 5)),
    base = rep(1, 12),
    stringsAsFactors = FALSE
  )
  design <- wf_prepare_design(
    data,
    id = "id",
    calibration = c("province", "age"),
    base_weight = "base"
  )
  dims <- wf_dims(age = c("young", "old"))
  population <- data.frame(
    province = rep(c("A", "B"), each = 2),
    age = rep(c("young", "old"), 2),
    count = c(60, 40, 30, 70),
    stringsAsFactors = FALSE
  )
  target <- wf_target_population(
    population,
    key_map = c(age = "age"),
    count = "count",
    dims = dims,
    by = "province",
    by_key = "province",
    keep_joint = TRUE
  )
  target <- .wf_verified_target(
    target,
    evidence = list(demo_only = FALSE, data_checksum = strrep("a", 64)),
    source_type = "population"
  )
  ladder <- wf_collapse_ladder(
    dims,
    level1 = list(age = c(young = "all", old = "all"))
  )
  analysis <- data.frame(
    id = data$id,
    outcome = rep(c(0, 1), 6),
    stringsAsFactors = FALSE
  )

  list(
    design = design,
    target = target,
    dims = dims,
    ladder = ladder,
    analysis = if (with_outcomes) analysis else NULL
  )
}

test_that("support merging is deterministic and conserves design totals", {
  f <- make_sparse_safe_workflow_fixture()

  a <- wf_plan_cells(
    f$design,
    f$target,
    f$dims,
    min_cell = 5,
    ladder = f$ladder
  )
  b <- wf_plan_cells(
    f$design,
    f$target,
    f$dims,
    min_cell = 5,
    ladder = f$ladder
  )

  expect_s3_class(a, "wf_cell_merge_plan")
  expect_identical(a$identity, b$identity)
  expect_equal(sum(a$cells_before$n), sum(a$cells_after$n))
  expect_equal(
    sum(a$cells_before$base_weight),
    sum(a$cells_after$base_weight)
  )
  expect_true(all(a$map$boundary_before == a$map$boundary_after))
  expect_lte(a$projected_max_weight_ratio, 4)
})

test_that("study outcomes cannot alter a merge plan", {
  f <- make_sparse_safe_workflow_fixture(with_outcomes = TRUE)

  p1 <- wf_plan_cells(f$design, f$target, f$dims, ladder = f$ladder)
  f$analysis$outcome <- rev(f$analysis$outcome)
  p2 <- wf_plan_cells(f$design, f$target, f$dims, ladder = f$ladder)

  expect_identical(p1$identity, p2$identity)
})

test_that("stored cell maps apply without recomputing a plan", {
  f <- make_sparse_safe_workflow_fixture()
  plan <- wf_plan_cells(f$design, f$target, f$dims, ladder = f$ladder)

  applied <- .wf_apply_cell_plan(f$design, f$target, plan)

  expect_s3_class(applied$design, "wf_design_data")
  expect_s3_class(applied$target, "wf_verified_target")
  expect_identical(unique(applied$design$data$age), "all")
  expect_identical(names(applied$target$groups$A$margins$age), "all")
  expect_identical(unique(applied$target$joint$A$age), "all")
})

test_that("declared category order provides a deterministic adjacency fallback", {
  f <- make_sparse_safe_workflow_fixture()

  plan <- wf_plan_cells(f$design, f$target, f$dims, min_cell = 5)

  expect_s3_class(plan, "wf_cell_merge_plan")
  expect_true(all(plan$reasons$reason == "declared_order_adjacency"))
  expect_equal(nrow(plan$cells_after), 2)
})

test_that("infeasible declared partitions never widen safety limits", {
  dims <- wf_dims(age = "only")
  design <- wf_prepare_design(
    data.frame(id = "r1", age = "only", stringsAsFactors = FALSE),
    id = "id",
    calibration = "age"
  )
  target <- wf_target_population(
    data.frame(age = "only", count = 10, stringsAsFactors = FALSE),
    key_map = c(age = "age"),
    count = "count",
    dims = dims
  )
  target <- .wf_verified_target(
    target,
    evidence = list(demo_only = FALSE, data_checksum = strrep("b", 64)),
    source_type = "population"
  )

  expect_error(
    wf_plan_cells(design, target, dims, min_cell = 5),
    class = "wf_error_feasibility"
  )
})

test_that("changed inputs cannot reuse a stored cell plan", {
  f <- make_sparse_safe_workflow_fixture()
  plan <- wf_plan_cells(f$design, f$target, f$dims, ladder = f$ladder)
  f$design$data$age[1] <- "old"

  expect_error(
    .wf_apply_cell_plan(f$design, f$target, plan),
    class = "wf_error_safety"
  )
})

#' Require one suggested package.
#'
#' @param package Package name.
#' @param caller Human-readable caller description.
#' @keywords internal
#' @noRd
.wf_require_namespace <- function(package, caller) {
  if (!requireNamespace(package, quietly = TRUE)) {
    wf_abort(
      sprintf(
        "%s requires the suggested '%s' package. Install it with install.packages(\"%s\").",
        caller,
        package,
        package
      ),
      "wf_error_dependency",
      list(package = package, caller = caller)
    )
  }
  invisible(TRUE)
}

#' Validate forwarded interoperability arguments.
#'
#' @param dots Captured arguments.
#' @param reserved Argument names owned by WFC.
#' @param caller Function name.
#' @keywords internal
#' @noRd
.wf_interop_dots <- function(dots, reserved, caller) {
  if (length(dots) == 0) {
    return(invisible(dots))
  }
  if (is.null(names(dots)) || any(!nzchar(names(dots)))) {
    wf_abort(
      sprintf("All arguments in `%s` must be named.", caller),
      "wf_error_input"
    )
  }
  conflict <- intersect(names(dots), reserved)
  if (length(conflict) > 0) {
    wf_abort(
      sprintf(
        "%s owns argument(s): %s; they cannot be overridden through `...`.",
        caller,
        paste(conflict, collapse = ", ")
      ),
      "wf_error_input",
      list(arguments = conflict)
    )
  }
  invisible(dots)
}

#' Match a data frame to one WFC unit-ID vector.
#'
#' @param data User data frame.
#' @param id ID column name in `data`.
#' @param reference_ids WFC unit IDs.
#' @param reference Human-readable reference name.
#' @param caller Function name.
#' @return Integer positions in `reference_ids`, ordered like `data`.
#' @keywords internal
#' @noRd
.wf_match_unit_ids <- function(data, id, reference_ids, reference, caller) {
  if (length(id) != 1 || !is.character(id) || is.na(id) || !nzchar(id)) {
    wf_abort("`id` must be one non-empty column name.", "wf_error_input")
  }
  .require_cols(data, id, "data")
  data_ids <- .chr(data[[id]])
  reference_ids <- .chr(reference_ids)
  invalid_data <- is.na(data_ids) | !nzchar(data_ids)
  invalid_reference <- is.na(reference_ids) | !nzchar(reference_ids)
  if (any(invalid_data) || any(invalid_reference)) {
    wf_abort(
      sprintf("%s requires non-missing, non-empty unit IDs.", caller),
      "wf_error_schema",
      list(
        invalid_data = which(invalid_data),
        invalid_reference = which(invalid_reference)
      )
    )
  }
  if (anyDuplicated(data_ids)) {
    duplicated_ids <- unique(data_ids[duplicated(data_ids)])
    wf_abort(
      sprintf(
        "`data[[%s]]` contains duplicate unit IDs: %s.",
        id,
        paste(utils::head(duplicated_ids, 5), collapse = ", ")
      ),
      "wf_error_schema",
      list(ids = duplicated_ids, source = "data")
    )
  }
  if (anyDuplicated(reference_ids)) {
    duplicated_ids <- unique(reference_ids[duplicated(reference_ids)])
    wf_abort(
      sprintf(
        "%s contains duplicate unit IDs: %s.",
        reference,
        paste(utils::head(duplicated_ids, 5), collapse = ", ")
      ),
      "wf_error_schema",
      list(ids = duplicated_ids, source = reference)
    )
  }

  missing_weights <- setdiff(data_ids, reference_ids)
  extra_weights <- setdiff(reference_ids, data_ids)
  if (length(missing_weights) > 0 || length(extra_weights) > 0) {
    wf_abort(
      sprintf(
        "%s requires identical unit-ID sets; %d data ID(s) lack weights and %d weight ID(s) lack data rows.",
        caller,
        length(missing_weights),
        length(extra_weights)
      ),
      "wf_error_schema",
      list(
        missing_weights = missing_weights,
        extra_weights = extra_weights
      )
    )
  }
  match(data_ids, reference_ids)
}

#' Convert WFC weights to a survey design
#'
#' Joins calibrated WFC weights onto an analysis data frame by exact unit ID and
#' creates a standard `survey::svydesign()` object. The `survey` package remains
#' a suggested dependency.
#'
#' @param w A `wf_weights` object.
#' @param data Non-empty analysis data frame containing the units in `w`.
#' @param id Unit-ID column in `data`; matched to `w$data$id`.
#' @param ids Survey cluster formula, defaulting to independent units.
#' @param strata Optional survey strata formula.
#' @param fpc Optional finite-population-correction formula.
#' @param nest Whether cluster IDs should be relabeled to nest within strata.
#' @param ... Additional arguments passed to `survey::svydesign()`. WFC owns
#'   `weights` and `probs`; they cannot be overridden.
#'
#' @return A standard `survey.design2` object whose variables include the
#'   aligned `.wf_weight` column.
#' @export
as_svydesign <- function(w, data, id = "id",
                         ids = ~1, strata = NULL, fpc = NULL,
                         nest = FALSE, ...) {
  if (!inherits(w, "wf_weights")) {
    wf_abort("`w` must be a wf_weights object.", "wf_error_input")
  }
  if (!is.data.frame(data) || nrow(data) == 0) {
    wf_abort("`data` must be a non-empty data frame.", "wf_error_input")
  }
  if (!is.data.frame(w$data) ||
      !all(c("id", "weight") %in% names(w$data))) {
    wf_abort(
      "`w` is missing its required unit ID or weight data.",
      "wf_error_internal"
    )
  }
  if (".wf_weight" %in% names(data)) {
    wf_abort(
      "`data` already contains reserved column '.wf_weight'.",
      "wf_error_schema"
    )
  }
  weight <- as.numeric(w$data$weight)
  if (anyNA(weight) || any(!is.finite(weight)) || any(weight < 0)) {
    wf_abort(
      "`w$data$weight` must contain finite non-negative values.",
      "wf_error_internal"
    )
  }

  dots <- list(...)
  .wf_interop_dots(dots, c("weights", "probs"), "as_svydesign()")
  order <- .wf_match_unit_ids(
    data,
    id,
    w$data$id,
    "`w$data$id`",
    "as_svydesign()"
  )
  variables <- data
  variables$.wf_weight <- weight[order]

  .wf_require_namespace("survey", "as_svydesign()")
  design <- do.call(
    survey::svydesign,
    c(
      list(
        ids = ids,
        strata = strata,
        fpc = fpc,
        data = variables,
        nest = nest,
        weights = stats::as.formula("~.wf_weight")
      ),
      dots
    )
  )
  attr(design, "wfc_provenance") <- w$provenance
  design
}

#' Convert WFC replicate weights to a survey replicate design
#'
#' Aligns recalibrated WFC replicate weights to analysis data by exact unit ID
#' and creates a standard `survey::svrepdesign()` object. The returned design
#' uses full-estimate-centered (`mse = TRUE`) variance so its estimates reproduce
#' [wf_variance()].
#'
#' @param r A `wf_replicate_weights` object.
#' @param data Non-empty analysis data frame containing the units in `r`.
#' @param id Unit-ID column in `data`; matched to `r$base$id`.
#' @param degf Optional survey design degrees of freedom.
#' @param ... Additional arguments passed to `survey::svrepdesign()`. WFC owns
#'   the variables, base/replicate weights, replication type and scales,
#'   combined-weight setting, and MSE setting.
#'
#' @return A standard `svyrep.design` object whose variables include the aligned
#'   `.wf_weight` base-weight column.
#' @export
as_svrepdesign <- function(r, data, id = "id", degf = NULL, ...) {
  if (!inherits(r, "wf_replicate_weights")) {
    wf_abort(
      "`r` must be a wf_replicate_weights object.",
      "wf_error_input"
    )
  }
  if (!is.data.frame(data) || nrow(data) == 0) {
    wf_abort("`data` must be a non-empty data frame.", "wf_error_input")
  }
  if (".wf_weight" %in% names(data)) {
    wf_abort(
      "`data` already contains reserved column '.wf_weight'.",
      "wf_error_schema"
    )
  }
  if (!is.data.frame(r$base) ||
      !all(c("id", "weight") %in% names(r$base)) ||
      !is.matrix(r$replicates) || nrow(r$replicates) != nrow(r$base) ||
      ncol(r$replicates) == 0) {
    wf_abort(
      "`r` has malformed base or replicate weight data.",
      "wf_error_internal"
    )
  }
  base_weight <- as.numeric(r$base$weight)
  replicate_weight <- r$replicates
  if (anyNA(base_weight) || any(!is.finite(base_weight)) ||
      any(base_weight < 0) || anyNA(replicate_weight) ||
      any(!is.finite(replicate_weight)) || any(replicate_weight < 0)) {
    wf_abort(
      "`r` contains invalid base or replicate weights.",
      "wf_error_internal"
    )
  }
  if (length(r$scale) != 1 || !is.numeric(r$scale) ||
      !is.finite(r$scale) || r$scale <= 0 ||
      !is.numeric(r$rscales) || length(r$rscales) != ncol(r$replicates) ||
      anyNA(r$rscales) || any(!is.finite(r$rscales)) ||
      any(r$rscales <= 0)) {
    wf_abort(
      "`r` contains invalid replication scale metadata.",
      "wf_error_internal"
    )
  }
  if (length(r$method) != 1 ||
      !r$method %in% c("bootstrap", "jackknife", "brr")) {
    wf_abort(
      "`r` contains an unsupported replication method.",
      "wf_error_internal",
      list(method = r$method)
    )
  }
  if (!is.null(degf) &&
      (length(degf) != 1 || !is.numeric(degf) || !is.finite(degf) ||
       degf < 0)) {
    wf_abort("`degf` must be NULL or a non-negative number.", "wf_error_input")
  }

  dots <- list(...)
  reserved <- c(
    "variables", "repweights", "weights", "type", "scale", "rscales",
    "combined.weights", "mse", "rho"
  )
  .wf_interop_dots(dots, reserved, "as_svrepdesign()")
  order <- .wf_match_unit_ids(
    data,
    id,
    r$base$id,
    "`r$base$id`",
    "as_svrepdesign()"
  )
  variables <- data
  variables$.wf_weight <- base_weight[order]
  replicate_weight <- replicate_weight[order, , drop = FALSE]

  survey_type <- switch(
    r$method,
    bootstrap = "bootstrap",
    jackknife = if (is.null(r$design$strata)) "JK1" else "JKn",
    brr = if (!is.null(r$rho) && r$rho > 0) "Fay" else "BRR"
  )
  if (survey_type == "BRR" &&
      !isTRUE(all.equal(r$scale, 1 / ncol(replicate_weight), tolerance = 1e-12))) {
    wf_abort(
      "BRR replication metadata does not use the required 1 / R scale.",
      "wf_error_internal",
      list(scale = r$scale, expected = 1 / ncol(replicate_weight))
    )
  }

  .wf_require_namespace("survey", "as_svrepdesign()")
  args <- list(
    variables = variables,
    repweights = replicate_weight,
    weights = base_weight[order],
    type = survey_type,
    rscales = r$rscales,
    combined.weights = TRUE,
    mse = TRUE,
    degf = degf
  )
  if (survey_type == "Fay") {
    args$rho <- r$rho
  }
  if (survey_type != "BRR" && survey_type != "Fay") {
    args$scale <- r$scale
  }
  design <- do.call(survey::svrepdesign, c(args, dots))
  if (!isTRUE(all.equal(design$scale, r$scale, tolerance = 1e-12)) ||
      !isTRUE(all.equal(design$rscales, r$rscales, tolerance = 1e-12))) {
    wf_abort(
      "survey::svrepdesign() did not preserve WFC replication scales.",
      "wf_error_internal"
    )
  }
  attr(design, "wfc_provenance") <- r$provenance
  design
}

make_interoperability_weights <- function() {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    tol = 1e-8
  )
  list(fixture = fixture, weights = weights)
}

test_that("dependency checks raise a classed error", {
  expect_error(
    .wf_require_namespace(
      "wfc_package_that_does_not_exist",
      "test interoperability"
    ),
    class = "wf_error_dependency"
  )
})

test_that("as_svydesign is exported", {
  expect_true("as_svydesign" %in% getNamespaceExports("WFC"))
  expect_true(is.function(as_svydesign))
})

test_that("as_svydesign aligns weights by ID and preserves input order", {
  testthat::skip_if_not_installed("survey")
  setup <- make_interoperability_weights()
  data <- setup$fixture$sample[rev(seq_len(nrow(setup$fixture$sample))), ]
  data$outcome <- as.numeric(data$age == "young")
  original_data <- data
  original_weights <- setup$weights

  design <- as_svydesign(setup$weights, data, id = "id")
  aligned <- setup$weights$data$weight[
    match(data$id, setup$weights$data$id)
  ]

  expect_s3_class(design, "survey.design2")
  expect_equal(as.numeric(stats::weights(design)), aligned)
  expect_equal(design$variables$.wf_weight, aligned)
  expect_identical(design$variables$id, data$id)
  expect_identical(attr(design, "wfc_provenance"), setup$weights$provenance)
  expect_identical(data, original_data)
  expect_identical(setup$weights, original_weights)

  estimate <- survey::svymean(~outcome, design)
  expected <- sum(aligned * data$outcome) / sum(aligned)
  expect_equal(as.numeric(estimate), expected, tolerance = 1e-12)
})

test_that("as_svydesign forwards cluster, strata, and nesting controls", {
  testthat::skip_if_not_installed("survey")
  setup <- make_interoperability_weights()
  data <- setup$fixture$sample
  data$psu <- paste0(data$province, "_", seq_len(nrow(data)))

  design <- as_svydesign(
    setup$weights,
    data,
    id = "id",
    ids = ~psu,
    strata = ~province,
    nest = TRUE
  )

  expect_s3_class(design, "survey.design2")
  expect_equal(nrow(design$cluster), nrow(data))
  expect_equal(nrow(design$strata), nrow(data))
  expect_true(isTRUE(design$has.strata))
})

test_that("as_svydesign enforces exact unique IDs and reserved columns", {
  testthat::skip_if_not_installed("survey")
  setup <- make_interoperability_weights()
  data <- setup$fixture$sample

  expect_error(as_svydesign(list(), data), class = "wf_error_input")
  expect_error(
    as_svydesign(setup$weights, list()),
    class = "wf_error_input"
  )
  expect_error(
    as_svydesign(setup$weights, data, id = "missing"),
    class = "wf_error_schema"
  )

  duplicate_data <- data
  duplicate_data$id[[2]] <- duplicate_data$id[[1]]
  expect_error(
    as_svydesign(setup$weights, duplicate_data),
    class = "wf_error_schema"
  )

  duplicate_weights <- setup$weights
  duplicate_weights$data$id[[2]] <- duplicate_weights$data$id[[1]]
  expect_error(
    as_svydesign(duplicate_weights, data),
    class = "wf_error_schema"
  )

  expect_error(
    as_svydesign(setup$weights, data[-1, ]),
    class = "wf_error_schema"
  )

  extra_data <- rbind(data, data[1, ])
  extra_data$id[nrow(extra_data)] <- "extra-id"
  expect_error(
    as_svydesign(setup$weights, extra_data),
    class = "wf_error_schema"
  )

  collision <- data
  collision$.wf_weight <- 1
  expect_error(
    as_svydesign(setup$weights, collision),
    class = "wf_error_schema"
  )

  expect_error(
    as_svydesign(setup$weights, data, weights = ~id),
    class = "wf_error_input"
  )
  expect_error(
    as_svydesign(setup$weights, data, probs = ~id),
    class = "wf_error_input"
  )
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----object-pipeline, eval=FALSE----------------------------------------------
# spec <- wf_pipeline(
#   target = target,
#   stages = list(
#     calibrate = list(method = "raking", tol = 1e-8)
#   ),
#   validate = list(max_deff = 6, max_margin_dev = 0.01)
# )
# 
# weights <- wf_run(spec, design, dims)

## ----population-pipeline, eval=FALSE------------------------------------------
# population_spec <- wf_pipeline(
#   target = list(mode = "population"),
#   stages = list(calibrate = list(method = "raking"))
# )
# 
# weights <- wf_run(
#   population_spec,
#   design,
#   dims,
#   population = target
# )

## ----audit, eval=FALSE--------------------------------------------------------
# wf_audit_export(
#   weights,
#   "pipeline-audit.json",
#   inputs = list(design = design, target = target)
# )

#' Convert values to report-safe text.
#'
#' @param x Atomic vector.
#' @keywords internal
#' @noRd
.wf_report_text <- function(x) {
  if (is.numeric(x)) {
    out <- format(x, digits = 6, trim = TRUE, scientific = FALSE)
  } else {
    out <- as.character(x)
  }
  out[is.na(x)] <- ""
  out
}

#' Escape text for Markdown tables.
#'
#' @param x Character vector.
#' @keywords internal
#' @noRd
.wf_report_markdown_escape <- function(x) {
  x <- gsub("\\", "\\\\", x, fixed = TRUE)
  x <- gsub("|", "\\|", x, fixed = TRUE)
  gsub("\n", " ", x, fixed = TRUE)
}

#' Escape text for HTML.
#'
#' @param x Character vector.
#' @keywords internal
#' @noRd
.wf_report_html_escape <- function(x) {
  x <- gsub("&", "&amp;", x, fixed = TRUE)
  x <- gsub("<", "&lt;", x, fixed = TRUE)
  x <- gsub(">", "&gt;", x, fixed = TRUE)
  x <- gsub('"', "&quot;", x, fixed = TRUE)
  gsub("'", "&#39;", x, fixed = TRUE)
}

#' Render a data frame as a Markdown table.
#'
#' @param x Data frame.
#' @param lang Output language.
#' @keywords internal
#' @noRd
.wf_report_markdown_table <- function(x, lang = NULL) {
  if (!is.data.frame(x) || ncol(x) == 0) {
    return("_No rows available._")
  }
  labels <- vapply(names(x), .wf_i18n_label, character(1), lang = lang)
  header <- paste0("| ", paste(labels, collapse = " | "), " |")
  rule <- paste0("| ", paste(rep("---", ncol(x)), collapse = " | "), " |")
  if (nrow(x) == 0) {
    return(paste(c(header, rule), collapse = "\n"))
  }
  rows <- vapply(seq_len(nrow(x)), function(i) {
    values <- vapply(x[i, , drop = FALSE], function(value) {
      .wf_report_markdown_escape(.wf_report_text(value))
    }, character(1))
    paste0("| ", paste(values, collapse = " | "), " |")
  }, character(1))
  paste(c(header, rule, rows), collapse = "\n")
}

#' Render a data frame as an HTML table.
#'
#' @param x Data frame.
#' @param lang Output language.
#' @keywords internal
#' @noRd
.wf_report_html_table <- function(x, lang = NULL) {
  if (!is.data.frame(x) || ncol(x) == 0) {
    return("<p><em>No rows available.</em></p>")
  }
  labels <- vapply(names(x), .wf_i18n_label, character(1), lang = lang)
  head <- paste0(
    "<thead><tr>",
    paste0("<th>", .wf_report_html_escape(labels), "</th>", collapse = ""),
    "</tr></thead>"
  )
  body <- if (nrow(x) == 0) {
    "<tbody></tbody>"
  } else {
    rows <- vapply(seq_len(nrow(x)), function(i) {
      values <- vapply(x[i, , drop = FALSE], function(value) {
        .wf_report_html_escape(.wf_report_text(value))
      }, character(1))
      paste0(
        "<tr>",
        paste0("<td>", values, "</td>", collapse = ""),
        "</tr>"
      )
    }, character(1))
    paste0("<tbody>", paste(rows, collapse = ""), "</tbody>")
  }
  paste0("<table>", head, body, "</table>")
}

#' Add manager decision columns to weight diagnostics.
#'
#' @param diagnostics Diagnostic table.
#' @param lang Output language.
#' @keywords internal
#' @noRd
.wf_report_manager_weights <- function(diagnostics, lang = NULL) {
  light <- ifelse(
    diagnostics$verdict == "OK",
    "green",
    ifelse(diagnostics$verdict == "usable with caveats", "yellow", "red")
  )
  verdict_label <- vapply(light, function(value) {
    .wf_tr(paste0("verdict_", value), lang = lang)
  }, character(1))
  action <- vapply(light, function(value) {
    .wf_tr(paste0("action_weights_", value), lang = lang)
  }, character(1))
  cbind(
    diagnostics,
    verdict_label = unname(verdict_label),
    light = unname(light),
    publish_separately = light == "green",
    action = unname(action)
  )
}

#' Extract structured method-specific sections from weights.
#'
#' @param w A `wf_weights` object.
#' @keywords internal
#' @noRd
.wf_report_weight_sections <- function(w) {
  sections <- list()
  if (!is.null(w$overlap)) {
    sections$propensity_overlap <- data.frame(
      threshold = w$overlap$threshold,
      n_boundary = w$overlap$n_boundary,
      n_online = w$overlap$n_online,
      stringsAsFactors = FALSE
    )
    online <- w$overlap$online
    reference <- w$overlap$reference
    sections$propensity_quantiles <- rbind(
      data.frame(
        source = "online",
        quantile = names(online),
        propensity = unname(online),
        stringsAsFactors = FALSE
      ),
      data.frame(
        source = "reference",
        quantile = names(reference),
        propensity = unname(reference),
        stringsAsFactors = FALSE
      )
    )
  }
  if (inherits(w, "wf_attrition_weights") && is.data.frame(w$balance)) {
    sections$attrition_balance <- w$balance
  } else if (is.data.frame(w$balance)) {
    sections$propensity_balance <- w$balance
  }
  if (inherits(w, "wf_attrition_weights") &&
      is.data.frame(w$retention_probability)) {
    sections$attrition_retention <- w$retention_probability
  }
  if (is.data.frame(w$cell_report)) {
    sections$poststrat_cells <- w$cell_report
  }
  if (is.data.frame(w$collapse_map)) {
    sections$poststrat_collapse_map <- w$collapse_map
  }
  if (is.data.frame(w$relaxation)) {
    sections$soft_relaxation <- w$relaxation
  }
  if (!is.null(w$provenance$stages)) {
    stage_rows <- lapply(seq_along(w$provenance$stages), function(i) {
      stage <- w$provenance$stages[[i]]
      data.frame(
        stage = if (is.null(stage$stage)) i else stage$stage,
        label = if (is.null(stage$label)) paste0("stage", i) else stage$label,
        method = if (is.null(stage$provenance$method)) NA_character_ else
          stage$provenance$method,
        stringsAsFactors = FALSE
      )
    })
    sections$pipeline_stages <- do.call(rbind, stage_rows)
  }
  sections
}

#' Build the report payload for a blend result.
#'
#' @param w A `wf_blend_result`.
#' @param audience Report audience.
#' @param lang Output language.
#' @keywords internal
#' @noRd
.wf_report_blend <- function(w, audience, lang = NULL) {
  if (!is.data.frame(w$summary) || nrow(w$summary) == 0) {
    wf_abort(
      "A blend report requires a non-empty `$summary`; blend with an outcome first.",
      "wf_error_input"
    )
  }
  table <- w$summary
  if (audience == "manager") {
    reason <- if (is.data.frame(w$lambda)) w$lambda else data.frame()
    one_source_groups <- if (nrow(reason) > 0) {
      unique(reason$group[reason$lambda_reason %in% c("online_only", "offline_only")])
    } else {
      character()
    }
    trimmed_groups <- if (nrow(reason) > 0 && "lambda_trimmed" %in% names(reason)) {
      unique(reason$group[reason$lambda_trimmed])
    } else {
      character()
    }
    light <- vapply(table$group, function(group) {
      global <- identical(group, "__overall__")
      if ((global && length(one_source_groups) > 0) || group %in% one_source_groups) {
        "red"
      } else if ((global && length(trimmed_groups) > 0) || group %in% trimmed_groups) {
        "yellow"
      } else {
        "green"
      }
    }, character(1))
    verdict_label <- vapply(light, function(value) {
      .wf_tr(paste0("verdict_", value), lang = lang)
    }, character(1))
    action <- vapply(light, function(value) {
      .wf_tr(paste0("action_blend_", value), lang = lang)
    }, character(1))
    table <- cbind(
      table,
      verdict_label = unname(verdict_label),
      light = unname(light),
      publish_separately = light == "green",
      action = unname(action)
    )
  }
  sections <- list()
  if (is.data.frame(w$lambda)) sections$blend_lambda <- w$lambda
  if (is.data.frame(w$sensitivity)) sections$blend_sensitivity <- w$sensitivity
  if (is.data.frame(w$diagnostics$source_support)) {
    sections$blend_source_support <- w$diagnostics$source_support
  }
  list(table = table, sections = sections)
}

#' Build report payload for a safe workflow.
#'
#' @param workflow A `wf_safe_workflow` object.
#' @param audience Normalized report audience.
#' @param lang Output language.
#' @keywords internal
#' @noRd
.wf_report_safe_workflow <- function(workflow, audience, lang = NULL) {
  if (!identical(workflow$identity, .wf_safe_workflow_identity(workflow))) {
    .wf_safety_abort(
      "safe_workflow_invalid",
      "The safe workflow changed after its identity was recorded.",
      "w"
    )
  }
  ready <- isTRUE(workflow$plan$ready)
  unresolved <- nrow(workflow$cell_plan$unresolved_cells)
  next_action <- .wf_tr(
    if (ready) "action_safe_review" else "action_safe_resolve",
    lang = lang
  )
  table <- data.frame(
    status = if (ready) "ready_for_human_review" else "blocked",
    ready = ready,
    method = workflow$plan$method,
    affected_share = workflow$cell_plan$affected_share,
    unresolved = unresolved,
    next_action = next_action,
    stringsAsFactors = FALSE
  )
  sections <- list()
  if (identical(audience, "analyst")) {
    source_fields <- workflow$target$evidence$fields
    sections$source_evidence <- data.frame(
      field = names(source_fields),
      value = unname(unlist(source_fields, use.names = FALSE)),
      stringsAsFactors = FALSE
    )
    sections$precheck_issues <- workflow$plan$issues
    sections$cell_map <- workflow$cell_plan$map
    sections$cell_reasons <- workflow$cell_plan$reasons
    sections$cells_before <- workflow$cell_plan$cells_before
    sections$cells_after <- workflow$cell_plan$cells_after
  }
  list(
    table = table,
    sections = sections,
    provenance = list(
      workflow_identity = workflow$identity,
      design_identity = workflow$design$identity,
      target_identity = workflow$target$identity,
      cell_plan_identity = workflow$cell_plan$identity,
      plan_identity = workflow$plan$identity
    )
  )
}

#' Build locked-weight identity evidence for reports.
#'
#' @param weights A `wf_locked_weights` object.
#' @keywords internal
#' @noRd
.wf_report_lock_evidence <- function(weights) {
  values <- list(
    weights$design_identity,
    weights$target_identity,
    weights$plan_identity,
    weights$approval_identity,
    weights$identity,
    weights$provenance$safety$source_data_checksum,
    weights$provenance$safety$source_metadata_checksum
  )
  data.frame(
    evidence = c(
      "design", "target", "plan", "approval", "locked_weight",
      "source_data", "source_metadata"
    ),
    identity = vapply(values, function(value) {
      if (is.null(value) || !length(value)) NA_character_ else as.character(value[[1]])
    }, character(1)),
    stringsAsFactors = FALSE
  )
}

#' Render a quality report as Markdown.
#'
#' @param report A `wf_quality_report`.
#' @keywords internal
#' @noRd
.wf_report_markdown <- function(report) {
  chunks <- c(
    paste0("# ", report$title),
    "",
    paste0(
      .wf_tr("report_audience", lang = report$language),
      ": ",
      .wf_tr(paste0("audience_", report$audience), lang = report$language)
    ),
    "",
    .wf_report_markdown_table(report$table, lang = report$language)
  )
  if (length(report$sections) > 0) {
    for (section in names(report$sections)) {
      label <- .wf_i18n_section(section, report$language)
      chunks <- c(
        chunks,
        "",
        paste0("## ", label),
        "",
        .wf_report_markdown_table(
          report$sections[[section]],
          lang = report$language
        )
      )
    }
  }
  paste(chunks, collapse = "\n")
}

#' Render a quality report as standalone base HTML.
#'
#' @param report A `wf_quality_report`.
#' @keywords internal
#' @noRd
.wf_report_html <- function(report) {
  sections <- character()
  if (length(report$sections) > 0) {
    sections <- vapply(names(report$sections), function(section) {
      label <- .wf_i18n_section(section, report$language)
      paste0(
        "<section><h2>", .wf_report_html_escape(label), "</h2>",
        .wf_report_html_table(
          report$sections[[section]],
          lang = report$language
        ),
        "</section>"
      )
    }, character(1))
  }
  paste0(
    "<!DOCTYPE html><html><head><meta charset=\"utf-8\">",
    "<title>", .wf_report_html_escape(report$title), "</title>",
    "<style>body{font-family:sans-serif;max-width:1100px;margin:2rem auto;}",
    "table{border-collapse:collapse;width:100%;margin-bottom:1.5rem;}",
    "th,td{border:1px solid #ccc;padding:.35rem;text-align:left;}</style>",
    "</head><body><h1>", .wf_report_html_escape(report$title), "</h1>",
    "<p>", .wf_report_html_escape(.wf_tr(
      "report_audience",
      lang = report$language
    )), ": ", .wf_report_html_escape(.wf_tr(
      paste0("audience_", report$audience),
      lang = report$language
    )), "</p>",
    .wf_report_html_table(report$table, lang = report$language),
    paste(sections, collapse = ""),
    "</body></html>"
  )
}

#' Build a weighting quality report
#'
#' Creates one structured dossier with manager and analyst projections. It can
#' report calibrated/composed weights or a blend result and render the same
#' payload as Markdown or dependency-free HTML.
#'
#' @param w A `wf_weights`, `wf_blend_result`, `wf_safe_workflow`, or
#'   `wf_impact` object.
#' @param target Optional `wf_target` for margin-residual diagnostics.
#' @param audience Report projection: `"manager"`/`"decision"` for a compact
#'   decision view or `"analyst"`/`"statistician"` for full detail.
#' @param lang Output language. Resolution follows explicit argument,
#'   `options(wfc.lang)`, session locale, then English fallback.
#' @param output Return a structured object, Markdown, or standalone HTML.
#' @param file Optional output path for Markdown or HTML.
#'
#' @return A `wf_quality_report` for `output = "object"`; otherwise a rendered
#'   character string, or invisibly the structured report when `file` is used.
#' @export
#' @examples
#' # Panel attrition weights are derived from the panel itself, so this
#' # example runs end to end. Reporting calibrated weights follows the same
#' # call; see vignette("safe-weighting-workflow") for that path, which
#' # additionally requires a verified target and a recorded approval.
#' panel <- data.frame(
#'   id = paste0("p", 1:24),
#'   region = rep(c("A", "B"), each = 12),
#'   retained = rep(c(TRUE, TRUE, TRUE, FALSE), times = 6),
#'   age = rep(c(25, 35, 45, 55, 65, 75), times = 4),
#'   sex = rep(c("female", "male"), times = 12),
#'   stringsAsFactors = FALSE
#' )
#' weights <- wf_attrition(
#'   panel,
#'   retained = "retained",
#'   formula = ~ age + sex,
#'   id = "id",
#'   by = "region"
#' )
#'
#' wf_report(weights, audience = "decision")
#'
#' analyst <- wf_report(weights, audience = "statistician")
#' names(analyst$sections)
wf_report <- function(w, target = NULL,
                      audience = c(
                        "manager", "analyst", "decision", "statistician"
                      ),
                      lang = NULL,
                      output = c("object", "markdown", "html"),
                      file = NULL) {
  requested_audience <- match.arg(audience)
  audience <- switch(
    requested_audience,
    decision = "manager",
    statistician = "analyst",
    requested_audience
  )
  output <- match.arg(output)
  language <- .wf_lang(lang)
  if (!is.null(file) &&
      (length(file) != 1 || is.na(file) || !nzchar(file))) {
    wf_abort("`file` must be NULL or one non-empty path.", "wf_error_input")
  }
  if (output == "object" && !is.null(file)) {
    wf_abort("`file` requires output = 'markdown' or 'html'.", "wf_error_input")
  }
  safe_report <- inherits(w, c("wf_safe_workflow", "wf_impact"))
  if (safe_report && !is.null(target)) {
    .wf_safety_abort(
      "report_target_not_allowed",
      "Safe workflow and impact reports use only their recorded evidence.",
      "target"
    )
  }
  if (!safe_report && !is.null(target) && !inherits(target, "wf_target")) {
    wf_abort("`target` must be NULL or a wf_target object.", "wf_error_input")
  }

  if (inherits(w, "wf_safe_workflow")) {
    payload <- .wf_report_safe_workflow(w, audience, language)
    table <- payload$table
    sections <- payload$sections
    source_type <- "safe_workflow"
    provenance <- payload$provenance
  } else if (inherits(w, "wf_impact")) {
    if (!identical(w$identity, .wf_impact_identity(w))) {
      .wf_safety_abort(
        "impact_identity_invalid",
        "The impact result changed after its identity was recorded.",
        "w"
      )
    }
    table <- if (identical(audience, "manager")) {
      w$summary[c("outcome", "level", "unweighted", "weighted", "difference")]
    } else {
      w$summary
    }
    sections <- list()
    source_type <- "impact"
    provenance <- list(
      impact_identity = w$identity,
      locked_weight_identity = w$weight_identity,
      confidence_level = w$level
    )
  } else if (inherits(w, "wf_weights")) {
    if (inherits(w, "wf_locked_weights") &&
        !identical(w$identity, .wf_locked_weight_identity(w))) {
      .wf_safety_abort(
        "locked_weights_invalid",
        "The locked weights changed after their identity was recorded.",
        "w"
      )
    }
    diagnostics <- wf_diagnose(w, target = target)$table
    table <- if (audience == "manager") {
      .wf_report_manager_weights(diagnostics, language)
    } else {
      light <- ifelse(
        diagnostics$verdict == "OK",
        "green",
        ifelse(diagnostics$verdict == "usable with caveats", "yellow", "red")
      )
      cbind(
        diagnostics,
        verdict_label = vapply(light, function(value) {
          .wf_tr(paste0("verdict_", value), lang = language)
        }, character(1))
      )
    }
    sections <- .wf_report_weight_sections(w)
    if (inherits(w, "wf_locked_weights")) {
      sections$lock_evidence <- .wf_report_lock_evidence(w)
      source_type <- "locked_weights"
    } else {
      source_type <- "weights"
    }
    provenance <- w$provenance
  } else if (inherits(w, "wf_blend_result")) {
    payload <- .wf_report_blend(w, audience, language)
    table <- payload$table
    sections <- payload$sections
    source_type <- "blend"
    provenance <- w$provenance
  } else {
    wf_abort(
      paste(
        "`w` must be weights, a blend result, a safe workflow,",
        "or a post-lock impact result."
      ),
      "wf_error_input"
    )
  }

  report <- structure(
    list(
      title = .wf_tr("report_title", lang = language),
      table = table,
      sections = sections,
      audience = audience,
      requested_audience = requested_audience,
      language = language,
      source_type = source_type,
      provenance = provenance
    ),
    class = "wf_quality_report"
  )
  if (output == "object") {
    return(report)
  }
  rendered <- if (output == "markdown") {
    .wf_report_markdown(report)
  } else {
    .wf_report_html(report)
  }
  if (!is.null(file)) {
    writeLines(rendered, con = file, useBytes = TRUE)
    return(invisible(report))
  }
  rendered
}

#' Print a weighting quality report
#'
#' @param x A `wf_quality_report` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_quality_report <- function(x, ...) {
  cat(.wf_report_markdown(x), "\n")
  invisible(x)
}

#' Convert a weighting quality report to a data frame
#'
#' @param x A `wf_quality_report` object.
#' @param ... Unused.
#' @return The report's primary table.
#' @export
as.data.frame.wf_quality_report <- function(x, ...) {
  x$table
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----precheck, eval=FALSE-----------------------------------------------------
# precheck <- wf_precheck(
#   design$data,
#   target,
#   id = design$roles$id
# )
# 
# precheck$issues

## ----execute, eval=FALSE------------------------------------------------------
# plan <- wf_plan_weights(design, target, dims, method = "raking")
# approval <- wf_approve_plan(
#   plan,
#   approver = "Reviewer full name",
#   role = "Statistician"
# )
# locked <- wf_execute_plan(plan, approval, design, target)
# 
# diagnostics <- wf_diagnose(locked, target = target)

## ----direct, eval=FALSE-------------------------------------------------------
# via_dispatcher <- wf_calibrate(
#   design,
#   target,
#   method = "raking"
# )

#' Apply one category collapse map.
#'
#' @param sample Sample data frame.
#' @param target A `wf_target` object.
#' @param plan List with `dim` and named character vector `map`.
#' @keywords internal
#' @noRd
.wf_apply_one_collapse <- function(sample, target, plan) {
  d <- plan$dim
  map <- plan$map
  if (!d %in% target$dims) {
    wf_abort(sprintf("Collapse plan targets unknown dimension '%s'.", d), "wf_error_input")
  }
  v <- .chr(sample[[d]])
  hit <- !is.na(v) & v %in% names(map)
  v[hit] <- map[v[hit]]
  sample[[d]] <- v
  for (g in names(target$groups)) {
    m <- target$groups[[g]]$margins[[d]]
    key <- names(m)
    hit <- key %in% names(map)
    key[hit] <- map[key[hit]]
    target$groups[[g]]$margins[[d]] <- .wf_margin_vector(tapply(as.numeric(m), key, sum))
  }
  if (!is.null(target$joint)) {
    for (g in names(target$joint)) {
      joint <- target$joint[[g]]
      key <- .chr(joint[[d]])
      hit <- key %in% names(map)
      key[hit] <- map[key[hit]]
      joint[[d]] <- key
      target$joint[[g]] <- stats::aggregate(
        joint["pop"],
        joint[target$dims],
        sum
      )
    }
  }
  .wf_validate_target(target)
  list(sample = sample, target = target)
}

#' Apply a category collapse plan
#'
#' Applies category-merge maps consistently to sample data and a target object.
#'
#' @param sample Sample data frame.
#' @param target A `wf_target` object.
#' @param plan List with `dim` and named character vector `map`, or a
#'   `wf_collapse_plan` object from `wf_suggest_collapse()`.
#'
#' @return A list with collapsed `sample` and `target`.
#' @export
wf_apply_collapse <- function(sample, target, plan) {
  if (inherits(plan, "wf_collapse_plan")) {
    out <- list(sample = sample, target = target)
    for (m in plan$maps) {
      out <- .wf_apply_one_collapse(out$sample, out$target, m)
    }
    out$target$meta$collapsed <- c(out$target$meta$collapsed, list(plan))
    return(out)
  }

  out <- .wf_apply_one_collapse(sample, target, plan)
  out$target$meta$collapsed <- c(out$target$meta$collapsed, list(plan))
  out
}

#' Suggest collapse plans from precheck findings
#'
#' Converts selected precheck issues into a reviewable collapse plan using
#' collapse ladders declared in `wf_dims()`.
#'
#' @param precheck A `wf_precheck` object.
#' @param dims A `wf_dims` object with optional collapse ladders.
#' @param checks Precheck identifiers eligible for collapse suggestions.
#' @param max_steps Maximum number of ladder steps to inspect per dimension.
#'
#' @return A `wf_collapse_plan` object.
#' @export
wf_suggest_collapse <- function(precheck, dims,
                                checks = c("cat_infeasible", "support_thin", "risk_extreme_ratio"),
                                max_steps = 1) {
  if (!inherits(precheck, "wf_precheck")) {
    wf_abort("`precheck` must be a wf_precheck object.", "wf_error_input")
  }
  if (!inherits(dims, "wf_dims")) {
    wf_abort("`dims` must be a wf_dims object.", "wf_error_input")
  }
  if (length(max_steps) != 1 || !is.finite(max_steps) || max_steps < 1) {
    wf_abort("`max_steps` must be a positive integer.", "wf_error_input")
  }

  issues <- precheck$issues
  issues <- issues[
    issues$check %in% checks & !is.na(issues$dim) & !is.na(issues$category),
    ,
    drop = FALSE
  ]
  actions <- data.frame(
    group = character(),
    dim = character(),
    category = character(),
    check = character(),
    step = character(),
    stringsAsFactors = FALSE
  )
  maps <- list()
  unresolved <- issues[0, , drop = FALSE]

  for (i in seq_len(nrow(issues))) {
    d <- issues$dim[i]
    cat <- strsplit(issues$category[i], ",", fixed = TRUE)[[1]][1]
    cat <- trimws(cat)
    ladder <- dims$collapse[[d]]
    found <- FALSE

    if (!is.null(ladder)) {
      step_names <- names(ladder)
      if (is.null(step_names)) {
        step_names <- as.character(seq_along(ladder))
      }
      for (s in utils::head(seq_along(ladder), max_steps)) {
        map <- .chr(ladder[[s]])
        names(map) <- .chr(names(ladder[[s]]))
        if (cat %in% names(map)) {
          actions <- rbind(
            actions,
            data.frame(
              group = issues$group[i],
              dim = d,
              category = cat,
              check = issues$check[i],
              step = step_names[[s]],
              stringsAsFactors = FALSE
            )
          )
          existing <- which(vapply(
            maps,
            function(x) identical(x$dim, d) && identical(x$map, map),
            logical(1)
          ))
          if (length(existing) == 0) {
            maps[[length(maps) + 1]] <- list(dim = d, map = map)
          }
          found <- TRUE
          break
        }
      }
    }

    if (!found) {
      unresolved <- rbind(unresolved, issues[i, , drop = FALSE])
    }
  }

  structure(
    list(
      actions = actions,
      maps = maps,
      source_checks = issues,
      unresolved = unresolved,
      created = Sys.time()
    ),
    class = "wf_collapse_plan"
  )
}

#' Print a collapse plan
#'
#' @param x A `wf_collapse_plan` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_collapse_plan <- function(x, ...) {
  cat(sprintf(
    "<wf_collapse_plan>  %d action(s), %d map(s), %d unresolved issue(s)\n",
    nrow(x$actions),
    length(x$maps),
    nrow(x$unresolved)
  ))
  if (nrow(x$actions) > 0) {
    print(utils::head(x$actions, 10), row.names = FALSE)
  }
  invisible(x)
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----checks, eval=FALSE-------------------------------------------------------
# design <- wf_prepare_design(
#   design_only,
#   id = "person_id",
#   calibration = c("age_group", "region"),
#   base_weight = "base_weight"
# )
# 
# target <- wf_import_target(
#   "population-margins.csv",
#   "population-margins.csv.source.dcf",
#   dims,
#   key_map = c(age_group = "age_group", region = "region"),
#   count = "population_count"
# )
# 
# precheck <- wf_precheck(design$data, target, id = design$roles$id)
# cell_plan <- wf_plan_cells(design, target, dims)

#' Calculate the identity of a plan approval.
#'
#' @param approval Approval fields.
#' @keywords internal
#' @noRd
.wf_plan_approval_identity <- function(approval) {
  .wf_sha256_object(approval[setdiff(names(approval), "identity")])
}

#' Execute one weighting engine behind the verified-input boundary.
#'
#' @param design An unchanged `wf_design_data`.
#' @param target An unchanged, non-demo `wf_verified_target`.
#' @param method Internal engine identifier.
#' @param settings Named engine settings.
#' @keywords internal
#' @noRd
.wf_execute_verified_engine <- function(design, target, method, settings = list()) {
  design_ok <- inherits(design, "wf_design_data") &&
    identical(design$identity, .wf_design_identity(design$data, design$roles))
  target_ok <- inherits(target, "wf_verified_target") &&
    !isTRUE(target$demo_only) &&
    identical(target$identity, .wf_verified_target_identity(target))
  if (!design_ok || !target_ok) {
    .wf_safety_abort(
      "verified_weighting_inputs_required",
      paste(
        "Weighting requires an unchanged wf_design_data and a non-demo",
        "wf_verified_target."
      ),
      "design,target",
      next_actions = c("prepare_design_data", "import_verified_target")
    )
  }
  required_columns <- unique(c(target$by, target$dims))
  allowed_columns <- unique(c(
    design$roles$calibration,
    design$roles$strata,
    design$roles$clusters,
    design$roles$fpc
  ))
  missing_roles <- setdiff(required_columns, allowed_columns)
  if (length(missing_roles) > 0L) {
    .wf_safety_abort(
      "target_design_roles_mismatch",
      "Target fields must be declared design or calibration variables.",
      "target",
      evidence = list(columns = missing_roles),
      next_actions = "prepare_matching_design_data"
    )
  }
  settings_named <- is.list(settings) &&
    (length(settings) == 0L ||
     (!is.null(names(settings)) && all(nzchar(names(settings)))))
  if (!settings_named) {
    .wf_safety_abort(
      "weighting_settings_invalid",
      "Weighting settings must be a named list.",
      "settings"
    )
  }
  owned <- intersect(names(settings), c("sample", "design", "target", "id", "init_weight"))
  if (length(owned) > 0L) {
    .wf_safety_abort(
      "design_role_override_unsupported",
      "WFC owns sample, target, ID, and base-weight roles after design preparation.",
      owned[[1]],
      evidence = list(arguments = owned),
      next_actions = "prepare_design_with_required_roles"
    )
  }

  base <- list(
    sample = design$data,
    target = target,
    id = design$roles$id,
    init_weight = design$roles$base_weight
  )
  weights <- switch(
    method,
    raking = do.call(.wf_rake_engine, c(base, settings)),
    poststrat = do.call(.wf_poststrat_engine, c(base, settings)),
    greg = do.call(
      .wf_calibrate_engine,
      c(base[c("sample", "target")], list(method = method),
        base[c("id", "init_weight")], settings)
    ),
    logit = do.call(
      .wf_calibrate_engine,
      c(base[c("sample", "target")], list(method = method),
        base[c("id", "init_weight")], settings)
    ),
    soft = do.call(
      .wf_calibrate_engine,
      c(base[c("sample", "target")], list(method = method),
        base[c("id", "init_weight")], settings)
    ),
    ebal = do.call(
      .wf_calibrate_engine,
      c(base[c("sample", "target")], list(method = method),
        base[c("id", "init_weight")], settings)
    ),
    auto_trim = do.call(.wf_auto_trim_engine, c(base, settings)),
    autoweigh = {
      dims <- settings$dims
      settings$dims <- NULL
      do.call(
        .wf_autoweigh_engine,
        c(
          list(
            sample = design$data,
            population = target,
            dims = dims,
            id = design$roles$id
          ),
          if (!is.null(design$roles$base_weight)) {
            c(settings, list(init_weight = design$roles$base_weight))
          } else {
            settings
          }
        )
      )
    },
    .wf_safety_abort(
      "weight_method_unsupported",
      "The requested weighting method is not supported.",
      "method",
      evidence = list(method = method)
    )
  )

  if (inherits(weights, "wf_autoweigh_result")) {
    weights$weights$provenance$design_identity <- design$identity
    weights$weights$provenance$target_identity <- target$identity
    weights$design_identity <- design$identity
    weights$target_identity <- target$identity
  } else if (inherits(weights, "wf_weights")) {
    weights$provenance$design_identity <- design$identity
    weights$provenance$target_identity <- target$identity
  }
  weights
}

#' Record a human attestation for a reviewed weight plan
#'
#' Approval records are separate from plans. AI agents may prepare plans but
#' cannot create an approval while declaring themselves as an agent.
#'
#' @param plan An unchanged, review-ready `wf_weight_plan`.
#' @param approver Non-empty approver name or organizational identifier.
#' @param role Non-empty reviewing role.
#' @param note Optional review note.
#' @param actor_type Actor type. Only `"human"` is accepted.
#'
#' @return A `wf_plan_approval` attestation bound to the plan identity.
#' @export
wf_approve_plan <- function(plan, approver, role, note = NULL,
                            actor_type = "human") {
  if (!inherits(plan, "wf_weight_plan") ||
      !identical(plan$identity, .wf_weight_plan_identity(plan))) {
    .wf_safety_abort(
      "weight_plan_invalid",
      "`plan` must be an unchanged wf_weight_plan object.",
      "plan"
    )
  }
  if (!isTRUE(plan$ready)) {
    .wf_safety_abort(
      "weight_plan_blocked",
      "A plan with blocking precheck issues cannot be approved.",
      "plan",
      evidence = list(issues = plan$issues),
      next_actions = "resolve_precheck_issues"
    )
  }
  if (!identical(actor_type, "human")) {
    .wf_safety_abort(
      "human_approval_required",
      "Only a human reviewer may attest approval of a weight plan.",
      "actor_type",
      next_actions = "request_human_review"
    )
  }
  if (!.wf_is_string(approver) || !nzchar(trimws(approver))) {
    .wf_safety_abort(
      "approver_missing",
      "`approver` must identify the human reviewer.",
      "approver"
    )
  }
  if (!.wf_is_string(role) || !nzchar(trimws(role))) {
    .wf_safety_abort(
      "approver_role_missing",
      "`role` must identify the reviewer's role.",
      "role"
    )
  }
  if (!is.null(note) && (!.wf_is_string(note) || !nzchar(trimws(note)))) {
    .wf_safety_abort(
      "approval_note_invalid",
      "`note` must be NULL or one non-empty string.",
      "note"
    )
  }

  approval <- list(
    plan_identity = plan$identity,
    approver = trimws(approver),
    role = trimws(role),
    note = note,
    actor_type = actor_type,
    attestation = paste(
      "I reviewed the plan inputs, source evidence, checks, method,",
      "bounds, and any stored category merges."
    ),
    created = .wf_iso_time(),
    package_version = .wf_package_version()
  )
  approval$identity <- .wf_plan_approval_identity(approval)
  class(approval) <- "wf_plan_approval"
  approval
}

#' Calculate the immutable identity of locked weights.
#'
#' @param weights Locked-weight fields.
#' @keywords internal
#' @noRd
.wf_locked_weight_identity <- function(weights) {
  .wf_sha256_object(list(
    data = weights$data,
    plan_identity = weights$plan_identity,
    approval_identity = weights$approval_identity,
    design_identity = weights$design_identity,
    target_identity = weights$target_identity,
    effective_design_identity = weights$effective_design_identity,
    effective_target_identity = weights$effective_target_identity,
    method = weights$provenance$method
  ))
}

#' Execute an approved plan and lock its weights
#'
#' Revalidates the plan, human approval, design data, target evidence, and any
#' stored cell map before calling exactly the method recorded in the plan.
#'
#' @param plan An unchanged `wf_weight_plan`.
#' @param approval A matching `wf_plan_approval`.
#' @param design The unchanged `wf_design_data` used to build the plan.
#' @param target The unchanged `wf_verified_target` used to build the plan.
#'
#' @return A `wf_locked_weights` object that also inherits from `wf_weights`.
#' @export
wf_execute_plan <- function(plan, approval, design, target) {
  if (!inherits(plan, "wf_weight_plan") ||
      !identical(plan$identity, .wf_weight_plan_identity(plan)) ||
      !isTRUE(plan$ready)) {
    .wf_safety_abort(
      "weight_plan_not_executable",
      "The weight plan is changed or has blocking issues.",
      "plan"
    )
  }
  if (!inherits(approval, "wf_plan_approval") ||
      !identical(approval$identity, .wf_plan_approval_identity(approval)) ||
      !identical(approval$actor_type, "human") ||
      !identical(approval$plan_identity, plan$identity)) {
    .wf_safety_abort(
      "plan_approval_mismatch",
      "A current human approval matching this exact plan is required.",
      "approval",
      next_actions = "request_human_review"
    )
  }
  if (!inherits(design, "wf_design_data") ||
      !identical(design$identity, .wf_design_identity(design$data, design$roles)) ||
      !identical(design$identity, plan$input_identities$design)) {
    .wf_safety_abort(
      "execution_design_mismatch",
      "Design data do not match the approved plan.",
      "design"
    )
  }
  if (!inherits(target, "wf_verified_target") || isTRUE(target$demo_only) ||
      !identical(target$identity, .wf_verified_target_identity(target)) ||
      !identical(target$identity, plan$input_identities$target)) {
    .wf_safety_abort(
      "execution_target_mismatch",
      "The verified target does not match the approved plan.",
      "target"
    )
  }

  effective_design <- design
  effective_target <- target
  if (!is.null(plan$cell_plan)) {
    applied <- .wf_apply_cell_plan(design, target, plan$cell_plan)
    effective_design <- applied$design
    effective_target <- applied$target
  }
  if (!identical(effective_design$identity, plan$effective_identities$design) ||
      !identical(effective_target$identity, plan$effective_identities$target)) {
    .wf_safety_abort(
      "execution_effective_inputs_mismatch",
      "Effective inputs no longer match the reviewed plan.",
      "plan"
    )
  }

  method <- plan$method
  if (identical(method, "raking")) {
    settings <- list(
      trim = plan$settings$bounds,
      precheck = TRUE
    )
  } else if (identical(method, "logit")) {
    settings <- list(
      bounds = plan$settings$bounds,
      precheck = TRUE
    )
  } else if (identical(method, "poststrat")) {
    dim_args <- stats::setNames(
      rep(list(NULL), length(effective_target$dims)),
      effective_target$dims
    )
    effective_dims <- do.call(wf_dims, dim_args)
    identity_ladder <- wf_collapse_ladder(effective_dims)
    settings <- list(
      min_cell = plan$settings$min_cell,
      ladder = identity_ladder,
      empty_cell = "error",
      precheck = TRUE
    )
  } else {
    .wf_safety_abort(
      "approved_method_unsupported",
      "The approved method is not supported by safe execution.",
      "method"
    )
  }
  weights <- .wf_execute_verified_engine(
    effective_design,
    effective_target,
    method,
    settings
  )

  .wf_match_unit_ids(
    effective_design$data,
    effective_design$roles$id,
    weights$data$id,
    "calibrated weights",
    "wf_execute_plan()"
  )
  if (any(!is.finite(weights$data$weight)) || any(weights$data$weight <= 0)) {
    wf_abort(
      "Approved calibration did not produce finite positive weights.",
      "wf_error_feasibility",
      .wf_safety_payload(
        "locked_weights_invalid",
        "blocking",
        "weight",
        list(rows = which(!is.finite(weights$data$weight) |
          weights$data$weight <= 0)),
        "review_calibration_feasibility"
      )
    )
  }

  weights$plan_identity <- plan$identity
  weights$approval_identity <- approval$identity
  weights$design_identity <- design$identity
  weights$target_identity <- target$identity
  weights$effective_design_identity <- effective_design$identity
  weights$effective_target_identity <- effective_target$identity
  weights$locked_at <- .wf_iso_time()
  weights$provenance$safety <- list(
    plan_identity = plan$identity,
    approval_identity = approval$identity,
    approver = approval$approver,
    approver_role = approval$role,
    source_checksum = target$evidence$data_checksum,
    source_data_checksum = target$evidence$data_checksum,
    source_metadata_checksum = target$evidence$source_checksum,
    cell_plan_identity = plan$input_identities$cell_plan
  )
  weights$identity <- .wf_locked_weight_identity(weights)
  class(weights) <- c("wf_locked_weights", "wf_weights")
  weights
}

#' Calculate the identity of a guided safe workflow.
#'
#' @param workflow Workflow fields.
#' @keywords internal
#' @noRd
.wf_safe_workflow_identity <- function(workflow) {
  .wf_sha256_object(list(
    design = workflow$design$identity,
    target = workflow$target$identity,
    cell_plan = workflow$cell_plan$identity,
    plan = workflow$plan$identity
  ))
}

#' Prepare a guided safe weighting workflow
#'
#' This practitioner-oriented entry point composes design preparation, verified
#' target import, outcome-blind cell planning, and weight planning. It never
#' approves a plan or computes weights.
#'
#' @param data Design-only data frame.
#' @param id Unique identifier column.
#' @param calibration Declared calibration and boundary columns.
#' @param dims A `wf_dims` object.
#' @param target_file CSV or XLSX target file.
#' @param source_file Companion source DCF file.
#' @param source_type Either `"population"` or `"reference"`.
#' @param key_map Population dimension-to-column mapping.
#' @param count Population count column.
#' @param feature Reference reciprocal-design-weight column.
#' @param ... Named safe settings: `base_weight`, `strata`, `clusters`, `fpc`,
#'   `by`, `by_key`, `production`, `method`, `bounds`, `min_cell`,
#'   `max_weight_ratio`, `boundary`, and `ladder`.
#'
#' @return A reviewable `wf_safe_workflow`; no weights are computed.
#' @export
wf_guided_plan <- function(data, id, calibration, dims,
                           target_file, source_file,
                           source_type = c("population", "reference"),
                           key_map = NULL, count = NULL, feature = NULL, ...) {
  dots <- list(...)
  allowed <- c(
    "base_weight", "strata", "clusters", "fpc", "by", "by_key",
    "production", "method", "bounds", "min_cell", "max_weight_ratio",
    "boundary", "ladder"
  )
  if (length(dots) && (is.null(names(dots)) || any(!nzchar(names(dots))))) {
    .wf_safety_abort(
      "guided_settings_unnamed",
      "All guided workflow settings must be named.",
      "..."
    )
  }
  unknown <- setdiff(names(dots), allowed)
  if (length(unknown)) {
    .wf_safety_abort(
      "guided_setting_unsupported",
      sprintf("Unsupported guided setting(s): %s.", paste(unknown, collapse = ", ")),
      "...",
      evidence = list(settings = unknown)
    )
  }
  value <- function(name, default = NULL) {
    if (name %in% names(dots)) dots[[name]] else default
  }

  source_type <- match.arg(source_type)
  design <- wf_prepare_design(
    data,
    id = id,
    calibration = calibration,
    base_weight = value("base_weight"),
    strata = value("strata"),
    clusters = value("clusters"),
    fpc = value("fpc")
  )
  if (identical(source_type, "population")) {
    if (is.null(key_map) || is.null(count)) {
      .wf_safety_abort(
        "population_import_mapping_missing",
        "Population targets require `key_map` and `count`.",
        "key_map",
        next_actions = "supply_population_column_mapping"
      )
    }
    target <- wf_import_target(
      target_file,
      source_file,
      dims,
      key_map = key_map,
      count = count,
      by = value("by"),
      by_key = value("by_key"),
      production = value("production", TRUE)
    )
  } else {
    if (is.null(feature)) {
      .wf_safety_abort(
        "reference_feature_missing",
        "Reference targets require `feature`.",
        "feature"
      )
    }
    target <- wf_import_reference(
      target_file,
      source_file,
      dims,
      feature = feature,
      by = value("by"),
      production = value("production", TRUE)
    )
  }

  cell_plan <- wf_plan_cells(
    design,
    target,
    dims,
    min_cell = value("min_cell", 5),
    max_weight_ratio = value("max_weight_ratio", 4),
    boundary = value("boundary", target$by),
    ladder = value("ladder")
  )
  plan <- wf_plan_weights(
    design,
    target,
    dims,
    method = value("method", "raking"),
    bounds = value("bounds", c(0.3, 3)),
    min_cell = value("min_cell", 5),
    cell_plan = cell_plan
  )
  workflow <- list(
    design = design,
    target = target,
    cell_plan = cell_plan,
    plan = plan,
    weights = NULL,
    created = .wf_iso_time(),
    package_version = .wf_package_version()
  )
  workflow$identity <- .wf_safe_workflow_identity(workflow)
  class(workflow) <- "wf_safe_workflow"
  workflow
}

#' Execute a guided workflow with an external human approval
#'
#' @param workflow An unchanged `wf_safe_workflow`.
#' @param approval A matching `wf_plan_approval` created separately.
#'
#' @return A `wf_locked_weights` object.
#' @export
wf_guided_execute <- function(workflow, approval) {
  if (!inherits(workflow, "wf_safe_workflow") ||
      !identical(workflow$identity, .wf_safe_workflow_identity(workflow))) {
    .wf_safety_abort(
      "safe_workflow_invalid",
      "`workflow` must be an unchanged wf_safe_workflow object.",
      "workflow"
    )
  }
  wf_execute_plan(
    workflow$plan,
    approval,
    workflow$design,
    workflow$target
  )
}

#' Attach locked weights to analysis data by exact unit ID
#'
#' @param data Full analysis data frame.
#' @param weights An unchanged `wf_locked_weights` object.
#' @param id Unique identifier column in `data`.
#' @param weight_name Name of the appended weight column.
#'
#' @return `data` in its original row order with one locked-weight column.
#' @export
wf_attach_weights <- function(data, weights, id, weight_name = ".weight") {
  if (!is.data.frame(data)) {
    .wf_safety_abort(
      "analysis_data_invalid",
      "`data` must be a data frame.",
      "data"
    )
  }
  if (!inherits(weights, "wf_locked_weights") ||
      !identical(weights$identity, .wf_locked_weight_identity(weights))) {
    .wf_safety_abort(
      "locked_weights_invalid",
      "`weights` must be unchanged locked weights.",
      "weights"
    )
  }
  if (!.wf_is_string(weight_name) || !nzchar(weight_name)) {
    .wf_safety_abort(
      "weight_name_invalid",
      "`weight_name` must be one non-empty column name.",
      "weight_name"
    )
  }
  if (weight_name %in% names(data)) {
    .wf_safety_abort(
      "weight_column_exists",
      sprintf("Column '%s' already exists.", weight_name),
      "weight_name",
      next_actions = "choose_new_weight_column"
    )
  }

  index <- .wf_match_unit_ids(
    data,
    id,
    weights$data$id,
    "locked weights",
    "wf_attach_weights()"
  )
  data[[weight_name]] <- weights$data$weight[index]
  attr(data, "wf_locked_weight_identity") <- weights$identity
  data
}

test_that("only a matching human-attested approval executes", {
  f <- make_safe_workflow_fixture()
  p <- wf_plan_weights(f$design, f$target, f$dims)

  expect_error(
    wf_approve_plan(p, "agent", "assistant", actor_type = "agent"),
    class = "wf_error_safety"
  )

  a <- wf_approve_plan(p, "Reviewer", "statistician")
  w <- wf_execute_plan(p, a, f$design, f$target)

  expect_s3_class(w, "wf_locked_weights")
  expect_identical(w$plan_identity, p$identity)
  expect_identical(w$approval_identity, a$identity)
  expect_match(w$identity, "^[0-9a-f]{64}$")
})

test_that("stale approvals and changed inputs stop before calibration", {
  f <- make_safe_workflow_fixture()
  p <- wf_plan_weights(f$design, f$target, f$dims)
  a <- wf_approve_plan(p, "Reviewer", "statistician")
  f$design$data$gender[1] <- "male"

  expect_error(
    wf_execute_plan(p, a, f$design, f$target),
    class = "wf_error_safety"
  )
})

test_that("guided and lower-level workflows produce identical results", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample
  sample$base_weight <- 1
  files <- make_safe_population_files(fixture$pop)

  design <- wf_prepare_design(
    sample,
    id = "id",
    calibration = c("province", "gender", "age"),
    base_weight = "base_weight"
  )
  target <- wf_import_target(
    files$data_file,
    files$source_file,
    fixture$dims,
    c(gender = "gender", age = "age"),
    "count",
    by = "province",
    by_key = "province"
  )
  cells <- wf_plan_cells(design, target, fixture$dims)
  plan <- wf_plan_weights(design, target, fixture$dims, cell_plan = cells)

  guided <- wf_guided_plan(
    sample,
    id = "id",
    calibration = c("province", "gender", "age"),
    dims = fixture$dims,
    target_file = files$data_file,
    source_file = files$source_file,
    source_type = "population",
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    base_weight = "base_weight",
    by = "province",
    by_key = "province"
  )

  expect_identical(guided$plan$identity, plan$identity)
  approval <- wf_approve_plan(plan, "Reviewer", "statistician")
  lower <- wf_execute_plan(plan, approval, design, target)
  higher <- wf_guided_execute(guided, approval)
  lower_weights <- lower$data$weight[match(sample$id, lower$data$id)]
  higher_weights <- higher$data$weight[match(sample$id, higher$data$id)]
  expect_equal(lower_weights, higher_weights)
})

test_that("locked weights attach by ID without changing row order", {
  f <- make_safe_workflow_fixture(with_outcomes = TRUE)
  p <- wf_plan_weights(f$design, f$target, f$dims)
  a <- wf_approve_plan(p, "Reviewer", "statistician")
  w <- wf_execute_plan(p, a, f$design, f$target)
  data <- f$analysis[c(4, 1, 3, 2, 5:nrow(f$analysis)), , drop = FALSE]

  attached <- wf_attach_weights(data, w, id = "id")

  expect_identical(attached$id, data$id)
  expect_equal(
    attached$.weight,
    w$data$weight[match(data$id, w$data$id)]
  )
  expect_identical(attr(attached, "wf_locked_weight_identity"), w$identity)
})

test_that("approved logit and poststrat methods dispatch exactly", {
  f <- make_safe_workflow_fixture()
  logit_data <- f$design$data
  logit_data$base_weight <- 25
  logit_design <- wf_prepare_design(
    logit_data,
    id = "id",
    calibration = c("province", "gender", "age"),
    base_weight = "base_weight"
  )
  logit_plan <- wf_plan_weights(
    logit_design,
    f$target,
    f$dims,
    method = "logit"
  )
  logit_approval <- wf_approve_plan(
    logit_plan,
    "Reviewer",
    "statistician"
  )

  logit_weights <- wf_execute_plan(
    logit_plan,
    logit_approval,
    logit_design,
    f$target
  )

  expect_identical(logit_weights$provenance$method, "logit")

  cells <- wf_plan_cells(f$design, f$target, f$dims)
  poststrat_plan <- wf_plan_weights(
    f$design,
    f$target,
    f$dims,
    method = "poststrat",
    cell_plan = cells
  )
  poststrat_approval <- wf_approve_plan(
    poststrat_plan,
    "Reviewer",
    "statistician"
  )

  poststrat_weights <- wf_execute_plan(
    poststrat_plan,
    poststrat_approval,
    f$design,
    f$target
  )

  expect_identical(poststrat_weights$provenance$method, "poststrat")
})

safe_documentation_section <- function(text) {
  start <- grep("<!-- SAFE_WORKFLOW_START -->", text, fixed = TRUE)
  end <- grep("<!-- SAFE_WORKFLOW_END -->", text, fixed = TRUE)
  if (length(start) != 1L || length(end) != 1L || start >= end) {
    return(character())
  }
  text[seq.int(start + 1L, end - 1L)]
}

test_that("safe workflow documentation covers every controlled step", {
  root <- normalizePath(test_path("..", ".."), mustWork = FALSE)
  paths <- file.path(
    root,
    c("README.md", "README.zh-CN.md", "vignettes/safe-weighting-workflow.Rmd")
  )
  all_present <- all(file.exists(paths))
  skip_if_not(all_present, "complete source documentation set is not installed")
  expect_true(all_present)
  required <- c(
    "wf_prepare_design(",
    "wf_import_target(",
    "wf_plan_weights(",
    "wf_approve_plan(",
    "wf_execute_plan(",
    "wf_attach_weights(",
    "wf_report("
  )

  for (path in paths) {
    text <- readLines(path, warn = FALSE)
    section <- safe_documentation_section(text)
    expect_gt(length(section), 0)
    joined <- paste(section, collapse = "\n")
    expect_true(all(vapply(
      required,
      grepl,
      logical(1),
      x = joined,
      fixed = TRUE
    )))
    expect_false(grepl("wf_target_manual(", joined, fixed = TRUE))
    expect_false(grepl("moments =", joined, fixed = TRUE))
  }
})

test_that("safe documentation includes import files and agent refusal contract", {
  root <- normalizePath(test_path("..", ".."), mustWork = FALSE)
  paths <- file.path(
    root,
    c(
      "README.md",
      "README.zh-CN.md",
      "examples/README.md",
      "vignettes/safe-weighting-workflow.Rmd"
    )
  )
  all_present <- all(file.exists(paths))
  skip_if_not(all_present, "complete source documentation set is not installed")
  expect_true(all_present)
  joined <- paste(unlist(lapply(paths, readLines, warn = FALSE)), collapse = "\n")

  expect_match(joined, "safe-target-example.csv", fixed = TRUE)
  expect_match(joined, "safe-target-example.xlsx", fixed = TRUE)
  expect_match(joined, ".source.dcf", fixed = TRUE)
  expect_match(joined, "wf_error_safety", fixed = TRUE)
  expect_match(joined, "next_actions", fixed = TRUE)
  expect_match(joined, "actor_type = \"agent\"", fixed = TRUE)
  expect_match(joined, "wf_assess_impact(", fixed = TRUE)
  expect_match(joined, "wf_audit_export(", fixed = TRUE)
})

test_that("all current user documentation excludes removed weighting paths", {
  root <- normalizePath(test_path("..", ".."), mustWork = FALSE)
  paths <- c(
    file.path(root, "README.md"),
    file.path(root, "README.zh-CN.md"),
    list.files(
      file.path(root, "examples"),
      pattern = "[.](md|R|Rmd)$",
      full.names = TRUE
    ),
    list.files(
      file.path(root, "vignettes"),
      pattern = "[.](md|R|Rmd)$",
      full.names = TRUE
    )
  )
  paths <- paths[file.exists(paths)]
  skip_if(length(paths) == 0L, "source documentation is not installed")

  forbidden <- c(
    "wf_target_manual(",
    "wf_target_shrink(",
    "moments =",
    "mode = \"manual\"",
    "mode = 'manual'",
    "wf_run(..., margins ="
  )
  for (path in paths) {
    text <- paste(readLines(path, warn = FALSE), collapse = "\n")
    for (pattern in forbidden) {
      expect_false(
        grepl(pattern, text, fixed = TRUE),
        info = sprintf("%s still contains %s", basename(path), pattern)
      )
    }
  }
})

test_that("migration examples use executable safe import and report forms", {
  root <- normalizePath(test_path("..", ".."), mustWork = FALSE)
  path <- file.path(root, "docs", "migration", "wfc-1-to-2.md")
  skip_if_not(file.exists(path), "migration guide is not installed")
  text <- paste(readLines(path, warn = FALSE), collapse = "\n")

  expect_false(grepl("wf_report(weights, target", text, fixed = TRUE))
  expect_false(grepl(
    "source_file = \"population-source.pdf\"",
    text,
    fixed = TRUE
  ))
  expect_match(text, "population-source.dcf", fixed = TRUE)
})

set.seed(20260715)

if (!requireNamespace("pkgload", quietly = TRUE)) {
  stop("Validation requires the 'pkgload' package.")
}
pkgload::load_all(".", quiet = TRUE)

results <- data.frame(
  scenario = character(),
  expected_code = character(),
  observed_code = character(),
  safety_error = logical(),
  engine_calls_before_refusal = integer(),
  engine_calls_after_refusal = integer(),
  passed = logical(),
  detail = character(),
  stringsAsFactors = FALSE
)

engine_calls <- 0L
engine_names <- c(
  ".wf_rake_engine",
  ".wf_poststrat_engine",
  ".wf_calibrate_engine",
  ".wf_auto_trim_engine",
  ".wf_autoweigh_engine"
)
for (engine in engine_names) {
  trace(
    engine,
    tracer = quote(assign(
      "engine_calls",
      get("engine_calls", envir = .GlobalEnv) + 1L,
      envir = .GlobalEnv
    )),
    print = FALSE,
    where = asNamespace("WFC")
  )
}
on.exit({
  for (engine in engine_names) {
    untrace(engine, where = asNamespace("WFC"))
  }
}, add = TRUE)

capture_refusal <- function(scenario, expected_code, expression) {
  expression <- substitute(expression)
  before <- engine_calls
  condition <- tryCatch(
    eval(expression, envir = parent.frame()),
    error = identity
  )
  after <- engine_calls
  safety_error <- inherits(condition, "wf_error_safety")
  observed_code <- if (safety_error) condition$data$code else NA_character_
  passed <- safety_error && identical(observed_code, expected_code) &&
    identical(before, after)
  detail <- if (inherits(condition, "condition")) {
    conditionMessage(condition)
  } else {
    "No error was raised."
  }
  results <<- rbind(
    results,
    data.frame(
      scenario = scenario,
      expected_code = expected_code,
      observed_code = observed_code,
      safety_error = safety_error,
      engine_calls_before_refusal = before,
      engine_calls_after_refusal = after,
      passed = passed,
      detail = detail,
      stringsAsFactors = FALSE
    )
  )
  invisible(condition)
}

write_population <- function(population, label,
                             selected_before_outcomes = TRUE,
                             demo_only = FALSE) {
  directory <- tempfile(paste0("wfc-2-adversarial-", label, "-"))
  dir.create(directory)
  data_file <- file.path(directory, "population.csv")
  source_file <- paste0(data_file, ".source.dcf")
  utils::write.csv(population, data_file, row.names = FALSE)
  source <- c(
    publisher = "WFC controlled validation fixture",
    dataset_title = paste("Synthetic adversarial target", label),
    citation = "Synthetic validation fixture; not a production authority",
    reference_period = "2026-validation",
    population_scope = "Synthetic validation population",
    retrieved_at = "2026-07-15",
    license = "CC0-1.0",
    checksum_algorithm = "sha256",
    checksum = WFC:::.wf_sha256_file(data_file),
    transformation = "Deterministic synthetic validation construction",
    selected_before_outcomes = tolower(as.character(selected_before_outcomes)),
    demo_only = tolower(as.character(demo_only))
  )
  writeLines(sprintf("%s: %s", names(source), source), source_file)
  list(data_file = data_file, source_file = source_file)
}

dims <- wf_dims(sex = c("F", "M"), age = c("18-34", "35+"))
design_data <- data.frame(
  id = sprintf("r%02d", 1:16),
  sex = rep(c("F", "M"), 8),
  age = rep(c("18-34", "18-34", "35+", "35+"), 4),
  base_weight = 1,
  stringsAsFactors = FALSE
)
population <- data.frame(
  sex = c("F", "M", "F", "M"),
  age = c("18-34", "18-34", "35+", "35+"),
  count = rep(10, 4),
  stringsAsFactors = FALSE
)
source <- write_population(population, "authoritative")
design <- wf_prepare_design(
  design_data,
  id = "id",
  calibration = c("sex", "age"),
  base_weight = "base_weight"
)
target <- wf_import_target(
  source$data_file,
  source$source_file,
  dims,
  key_map = c(sex = "sex", age = "age"),
  count = "count"
)
plan <- wf_plan_weights(design, target, dims)
approval <- wf_approve_plan(plan, "Synthetic reviewer", "test-only reviewer")

# A runtime list of hand-written margins cannot create a pipeline target.
capture_refusal(
  "manual_margins",
  "manual_pipeline_unsupported",
  wf_pipeline(
    target = list(
      mode = "manual",
      margins = list(sex = c(F = 20, M = 20))
    ),
    stages = list(calibrate = list(method = "raking"))
  )
)

# An in-memory pass-rate table is only an ordinary target, regardless of its
# label. It cannot cross the verified public weighting boundary.
outcome_dims <- wf_dims(result = c("pass", "fail"))
pass_rate_target <- wf_target_population(
  data.frame(result = c("pass", "fail"), count = c(80, 20)),
  key_map = c(result = "result"),
  count = "count",
  dims = outcome_dims
)
capture_refusal(
  "pass_rate_target",
  "verified_weighting_inputs_required",
  wf_rake(design, pass_rate_target)
)

# Inline desired means are refused before entropy balancing starts.
capture_refusal(
  "outcome_mean",
  "inline_moments_unsupported",
  wf_calibrate(
    design,
    target,
    method = "ebal",
    moments = c(satisfaction = 0.8)
  )
)

# A desired outcome interval encoded as an ordinary target is also unverified.
outcome_interval_target <- wf_target_population(
  data.frame(result = c("inside", "outside"), count = c(90, 10)),
  key_map = c(result = "result"),
  count = "count",
  dims = wf_dims(result = c("inside", "outside"))
)
capture_refusal(
  "outcome_interval",
  "verified_weighting_inputs_required",
  wf_calibrate(design, outcome_interval_target, method = "raking")
)

# Even a structurally valid target made in memory lacks external evidence.
unverified_target <- wf_target_population(
  population,
  key_map = c(sex = "sex", age = "age"),
  count = "count",
  dims = dims
)
capture_refusal(
  "unverified_target",
  "verified_weighting_inputs_required",
  wf_rake(design, unverified_target)
)

# Demo evidence cannot be imported with the production default.
demo_source <- write_population(population, "demo", demo_only = TRUE)
capture_refusal(
  "demo_target",
  "demo_target_in_production",
  wf_import_target(
    demo_source$data_file,
    demo_source$source_file,
    dims,
    key_map = c(sex = "sex", age = "age"),
    count = "count"
  )
)

# Evidence that admits selection after outcomes is rejected at import.
late_source <- write_population(
  population,
  "selected-after-outcomes",
  selected_before_outcomes = FALSE
)
capture_refusal(
  "target_selected_after_outcomes",
  "target_selected_after_outcomes",
  wf_import_target(
    late_source$data_file,
    late_source$source_file,
    dims,
    key_map = c(sex = "sex", age = "age"),
    count = "count"
  )
)

# Approval for an earlier plan does not authorize a changed plan.
changed_plan <- wf_plan_weights(
  design,
  target,
  dims,
  bounds = c(0.4, 2.5)
)
capture_refusal(
  "stale_approval",
  "plan_approval_mismatch",
  wf_execute_plan(changed_plan, approval, design, target)
)

# Changing one design value invalidates the design identity recorded by plan.
changed_design <- design
changed_design$data$sex[[1]] <- "M"
capture_refusal(
  "changed_design",
  "execution_design_mismatch",
  wf_execute_plan(plan, approval, changed_design, target)
)

# AI agents cannot attest their own approval.
capture_refusal(
  "agent_approval",
  "human_approval_required",
  wf_approve_plan(
    plan,
    approver = "Automated agent",
    role = "assistant",
    actor_type = "agent"
  )
)

# A raw data frame supplied to a public engine wrapper is rejected before the
# internal raking engine is invoked.
capture_refusal(
  "raw_public_engine_call",
  "verified_weighting_inputs_required",
  wf_rake(design$data, target)
)

runtime_margins_absent <- !"margins" %in% names(formals(wf_run))
results <- rbind(
  results,
  data.frame(
    scenario = "runtime_margins_argument_removed",
    expected_code = "argument_removed",
    observed_code = if (runtime_margins_absent) "argument_removed" else NA_character_,
    safety_error = NA,
    engine_calls_before_refusal = engine_calls,
    engine_calls_after_refusal = engine_calls,
    passed = runtime_margins_absent,
    detail = "wf_run() has no margins formal; R rejects that argument.",
    stringsAsFactors = FALSE
  )
)

print(results, row.names = FALSE)
cat("\nSession information:\n")
print(utils::sessionInfo())

if (engine_calls != 0L) {
  stop(sprintf("A weighting engine ran %d time(s) during refusal checks.", engine_calls))
}
if (!all(results$passed)) {
  failed <- results$scenario[!results$passed]
  stop("WFC 2.0 adversarial validation failed: ", paste(failed, collapse = ", "))
}
cat(sprintf("\nAll %d adversarial checks passed before engine execution.\n", nrow(results)))

#' Build a stable machine-readable safety payload.
#'
#' @param code Stable condition code.
#' @param severity Safety severity.
#' @param field Related input field, if any.
#' @param evidence Machine-readable supporting evidence.
#' @param next_actions Stable action codes available to the caller.
#' @keywords internal
#' @noRd
.wf_safety_payload <- function(code, severity, field, evidence, next_actions) {
  list(
    code = code,
    severity = severity,
    field = field,
    evidence = evidence,
    next_actions = as.character(next_actions)
  )
}

#' Abort with a classed safety condition.
#'
#' @inheritParams .wf_safety_payload
#' @param message Human-readable error message.
#' @keywords internal
#' @noRd
.wf_safety_abort <- function(code, message, field = NULL,
                             evidence = list(), next_actions = character()) {
  wf_abort(
    message,
    "wf_error_safety",
    .wf_safety_payload(code, "blocking", field, evidence, next_actions)
  )
}

#' Warn with a classed safety condition.
#'
#' @inheritParams .wf_safety_payload
#' @param message Human-readable warning message.
#' @keywords internal
#' @noRd
.wf_safety_warn <- function(code, message, field = NULL,
                            evidence = list(), next_actions = character()) {
  wf_warn(
    message,
    "wf_warning_safety",
    .wf_safety_payload(code, "review_required", field, evidence, next_actions)
  )
}

#' Build an informational safety record.
#'
#' @inheritParams .wf_safety_payload
#' @keywords internal
#' @noRd
.wf_safety_info <- function(code, field = NULL, evidence = list(),
                            next_actions = character()) {
  .wf_safety_payload(code, "informational", field, evidence, next_actions)
}

# ============================================================================
# WFC Extension 3 -- reference implementation, part 2 of 3 (releases 0.12-0.13)
# Ecosystem interoperability and production pipeline infrastructure.
#
# Prototype code: source AFTER library(WFC). Soft dependencies (survey) are
# checked with requireNamespace(); the core stays base-R (design commitment).
#
# Contents:
#   1. as_svydesign() / as_svrepdesign()  -- survey bridges     [design §5.1]
#   2. augment / tidy / glance            -- broom-style        [design §5.2]
#   3. wf_pipeline() / wf_run()           -- declarative chain  [design §6.1]
#   4. wf_validate()                      -- weight drift check [design §6.2]
#   5. wf_audit_export()                  -- JSON audit trail   [design §6.3]
# ============================================================================

# ----------------------------------------------------------------------------
# 1. survey bridges -----------------------------------------------------------
# ----------------------------------------------------------------------------

#' wf_weights -> survey::svydesign (weights joined onto `data` by id).
as_svydesign <- function(w, data, id = "id") {
  if (!requireNamespace("survey", quietly = TRUE)) {
    stop("as_svydesign() needs the 'survey' package: install.packages(\"survey\")")
  }
  stopifnot(inherits(w, "wf_weights"))
  key <- trimws(as.character(data[[id]]))
  m <- match(key, trimws(as.character(w$data$id)))
  if (anyNA(m)) stop(sprintf("%d row(s) of `data` have no weight (unmatched id).", sum(is.na(m))))
  data$.wf_weight <- w$data$weight[m]
  survey::svydesign(ids = ~1, weights = ~.wf_weight, data = data)
}

#' wf_replicate_weights -> survey::svrepdesign. The stored (scale, rscales)
#' map directly onto svrepdesign's unified combining rule, so svymean/svyglm
#' reproduce wf_variance() exactly.
as_svrepdesign <- function(r, data, id = "id") {
  if (!requireNamespace("survey", quietly = TRUE)) {
    stop("as_svrepdesign() needs the 'survey' package: install.packages(\"survey\")")
  }
  stopifnot(inherits(r, "wf_replicate_weights"))
  key <- trimws(as.character(data[[id]]))
  m <- match(key, r$base$id)
  if (anyNA(m)) stop(sprintf("%d row(s) of `data` have no replicate weights.", sum(is.na(m))))
  survey::svrepdesign(
    data = data,
    weights = r$base$weight[m],
    repweights = r$replicates[m, , drop = FALSE],
    type = "other",
    scale = r$scale,
    rscales = r$rscales,
    combined.weights = TRUE
  )
}

# ----------------------------------------------------------------------------
# 2. broom-style tidiers ------------------------------------------------------
# In the package these register as S3 methods conditionally on `generics`.
# ----------------------------------------------------------------------------

#' Join calibrated weights back onto the input sample by id.
augment.wf_weights <- function(x, data, id = "id", ...) {
  key <- trimws(as.character(data[[id]]))
  m <- match(key, trimws(as.character(x$data$id)))
  data$.weight <- x$data$weight[m]
  data$.feature <- x$data$feature[m]
  data
}

tidy.wf_diagnostics <- function(x, ...) x$table

tidy.wf_variance_result <- function(x, ...) x$table

glance.wf_weights <- function(x, ...) {
  wt <- x$data$weight
  data.frame(
    n = nrow(x$data),
    groups = nrow(x$log),
    total_weight = sum(wt),
    ess = sum(wt)^2 / sum(wt^2),
    deff = 1 + (stats::sd(wt) / mean(wt))^2,
    method = if (is.null(x$provenance$method)) "raking" else x$provenance$method,
    stringsAsFactors = FALSE
  )
}

# ----------------------------------------------------------------------------
# 3. wf_pipeline() / wf_run() -------------------------------------------------
# A weighting round as a serializable spec: plain lists in, wf_weights out.
# wf_run(spec, ...) is also a valid `refit` closure body for wf_replicates().
# ----------------------------------------------------------------------------

wf_pipeline <- function(target, stages, validate = NULL) {
  stopifnot(is.list(target), is.list(stages))
  allowed <- c("propensity", "calibrate")
  bad <- setdiff(names(stages), allowed)
  if (length(bad) > 0) {
    stop(sprintf("Unknown pipeline stage(s): %s. Supported: %s.",
                 paste(bad, collapse = ", "), paste(allowed, collapse = ", ")))
  }
  spec <- list(target = target, stages = stages, validate = validate,
               created = format(Sys.time(), "%Y-%m-%dT%H:%M:%S%z"))
  spec$hash <- .wf_spec_hash(spec[c("target", "stages", "validate")])
  structure(spec, class = "wf_pipeline")
}

# Dependency-free content hash (djb2 over the deparsed spec); the packaged
# version may use tools::md5sum on a serialized tempfile instead.
.wf_spec_hash <- function(x) {
  s <- paste(deparse(x), collapse = "")
  h <- 5381
  for (ch in utf8ToInt(s)) h <- (h * 33 + ch) %% 2^31
  sprintf("%08x", h)
}

wf_run <- function(spec, sample, dims, population = NULL, reference = NULL,
                   base_weight = NULL) {
  stopifnot(inherits(spec, "wf_pipeline"))

  # -- target -----------------------------------------------------------------
  tg <- spec$target
  target <- if (identical(tg$mode, "population")) {
    if (is.null(population)) stop("Pipeline target mode 'population' needs `population`.")
    do.call(wf_target_population,
            c(list(pop = population, key_map = tg$key_map, count = tg$count,
                   dims = dims), tg[setdiff(names(tg), c("mode", "key_map", "count"))]))
  } else if (identical(tg$mode, "reference")) {
    if (is.null(reference)) stop("Pipeline target mode 'reference' needs `reference`.")
    do.call(wf_target_reference,
            c(list(ref = reference, feature = tg$feature, dims = dims),
              tg[setdiff(names(tg), c("mode", "feature"))]))
  } else {
    stop(sprintf("Unsupported pipeline target mode '%s'.", tg$mode))
  }

  # -- stage 1: propensity (optional) ------------------------------------------
  init_col <- base_weight
  work <- sample
  stage1 <- NULL
  if (!is.null(spec$stages$propensity)) {
    if (is.null(reference)) stop("Pipeline stage 'propensity' needs `reference`.")
    ps <- spec$stages$propensity
    pt <- wf_target_propensity(sample, reference, ps$formula,
                               by = ps$by, id = ps$id)
    stage1 <- do.call(wf_propensity,
                      c(list(target = pt), ps[setdiff(names(ps), c("formula", "by", "id"))]))
    sid <- trimws(as.character(stage1$data$id))
    key <- if (is.null(ps$id)) as.character(seq_len(nrow(work)))
           else trimws(as.character(work[[ps$id]]))
    m <- match(key, sid)
    if (anyNA(m)) stop("Propensity stage ids do not match the sample rows.")
    work$.wf_stage1 <- stage1$data$weight[m]
    init_col <- ".wf_stage1"
  }

  # -- stage 2: calibrate -------------------------------------------------------
  cal <- spec$stages$calibrate
  if (is.null(cal)) stop("Pipeline must contain a 'calibrate' stage.")
  args <- c(list(sample = work, target = target), cal)
  if (!is.null(init_col)) args$init_weight <- init_col
  w <- do.call(wf_calibrate, args)
  w$provenance$pipeline_hash <- spec$hash

  # -- declared validation ------------------------------------------------------
  if (!is.null(spec$validate)) {
    d <- wf_diagnose(w, target = target)$table
    v <- spec$validate
    if (!is.null(v$max_deff) && any(d$deff > v$max_deff)) {
      warning(sprintf("Pipeline validation: %d group(s) exceed max_deff = %g.",
                      sum(d$deff > v$max_deff), v$max_deff))
    }
    if (!is.null(v$max_margin_dev) && "margin_maxerr" %in% names(d) &&
        any(d$margin_maxerr > v$max_margin_dev, na.rm = TRUE)) {
      warning(sprintf("Pipeline validation: margin deviation exceeds %g.",
                      v$max_margin_dev))
    }
  }
  w
}

# ----------------------------------------------------------------------------
# 4. wf_validate() -- weight drift detection ----------------------------------
# "Did this round's weights silently change character?" Compares a new
# wf_weights against a reference release on group coverage, deff/ESS deltas,
# and the per-unit weight-ratio distribution where ids overlap.
# ----------------------------------------------------------------------------

wf_validate <- function(new, reference,
                        max_deff_delta = 1.0,
                        max_ratio_p99 = 2.0,
                        max_total_shift = 0.05) {
  stopifnot(inherits(new, "wf_weights"), inherits(reference, "wf_weights"))
  gstat <- function(w) {
    parts <- split(w$data$weight, w$data$group)
    data.frame(
      group = names(parts),
      n = vapply(parts, length, integer(1)),
      total = vapply(parts, sum, numeric(1)),
      deff = vapply(parts, function(x) 1 + (stats::sd(x) / mean(x))^2, numeric(1)),
      stringsAsFactors = FALSE, row.names = NULL
    )
  }
  a <- gstat(new); b <- gstat(reference)
  cmp <- merge(a, b, by = "group", suffixes = c("_new", "_ref"), all = TRUE)
  cmp$deff_delta <- cmp$deff_new - cmp$deff_ref
  cmp$total_shift <- abs(cmp$total_new - cmp$total_ref) / cmp$total_ref

  issues <- character(0)
  gone <- cmp$group[is.na(cmp$n_new)]
  born <- cmp$group[is.na(cmp$n_ref)]
  if (length(gone) > 0) issues <- c(issues, sprintf("groups missing in new: %s", paste(gone, collapse = ", ")))
  if (length(born) > 0) issues <- c(issues, sprintf("new groups appeared: %s", paste(born, collapse = ", ")))
  hit <- !is.na(cmp$deff_delta) & cmp$deff_delta > max_deff_delta
  if (any(hit)) issues <- c(issues, sprintf("deff rose by more than %g in: %s",
                                            max_deff_delta, paste(cmp$group[hit], collapse = ", ")))
  hit <- !is.na(cmp$total_shift) & cmp$total_shift > max_total_shift
  if (any(hit)) issues <- c(issues, sprintf("group total shifted by more than %.0f%% in: %s",
                                            100 * max_total_shift, paste(cmp$group[hit], collapse = ", ")))

  ratio <- NULL
  ids_new <- trimws(as.character(new$data$id))
  ids_ref <- trimws(as.character(reference$data$id))
  common <- intersect(ids_new, ids_ref)
  if (length(common) > 10) {
    r <- new$data$weight[match(common, ids_new)] /
         reference$data$weight[match(common, ids_ref)]
    r <- r / stats::median(r)                       # remove pure rescaling
    p99 <- stats::quantile(r, 0.99, names = FALSE)
    p01 <- stats::quantile(r, 0.01, names = FALSE)
    ratio <- list(n_common = length(common), p01 = p01, median = 1, p99 = p99)
    if (p99 > max_ratio_p99 || p01 < 1 / max_ratio_p99) {
      issues <- c(issues, sprintf(
        "per-unit weight ratios drifted: 1%%/99%% quantiles %.2f / %.2f (threshold %gx)",
        p01, p99, max_ratio_p99))
    }
  }

  ok <- length(issues) == 0
  if (!ok) warning(paste("wf_validate:", paste(issues, collapse = "; ")))
  structure(list(ok = ok, issues = issues, comparison = cmp, ratio = ratio,
                 thresholds = list(max_deff_delta = max_deff_delta,
                                   max_ratio_p99 = max_ratio_p99,
                                   max_total_shift = max_total_shift)),
            class = "wf_validation")
}

print.wf_validation <- function(x, ...) {
  cat(sprintf("<wf_validation>  %s\n", if (x$ok) "PASS - no drift detected" else "DRIFT DETECTED"))
  for (i in x$issues) cat("  -", i, "\n")
  invisible(x)
}

# ----------------------------------------------------------------------------
# 5. wf_audit_export() -- provenance as a self-contained JSON file -------------
# Base-R JSON serializer (no jsonlite): lists -> objects, unnamed lists and
# length>1 vectors -> arrays, data frames -> arrays of row objects.
# ----------------------------------------------------------------------------

.wf_json <- function(x) {
  esc <- function(s) {
    s <- gsub("\\", "\\\\", s, fixed = TRUE)
    s <- gsub("\"", "\\\"", s, fixed = TRUE)
    s <- gsub("\n", "\\n", s, fixed = TRUE)
    s <- gsub("\r", "\\r", s, fixed = TRUE)
    s <- gsub("\t", "\\t", s, fixed = TRUE)
    s
  }
  scalar <- function(v) {
    if (is.na(v)) return("null")
    if (is.logical(v)) return(if (v) "true" else "false")
    if (is.numeric(v)) return(format(v, digits = 15, scientific = FALSE, trim = TRUE))
    paste0("\"", esc(as.character(v)), "\"")
  }
  if (is.null(x)) return("null")
  if (inherits(x, "POSIXt")) x <- format(x, "%Y-%m-%dT%H:%M:%S%z")
  if (is.function(x)) return("\"<function>\"")
  if (is.factor(x)) x <- as.character(x)
  if (is.data.frame(x)) {
    rows <- lapply(seq_len(nrow(x)), function(i) as.list(x[i, , drop = FALSE]))
    return(paste0("[", paste(vapply(rows, .wf_json, ""), collapse = ","), "]"))
  }
  if (is.list(x)) {
    nm <- names(x)
    if (!is.null(nm) && all(nzchar(nm))) {
      items <- vapply(seq_along(x), function(i) {
        paste0("\"", esc(nm[i]), "\":", .wf_json(x[[i]]))
      }, "")
      return(paste0("{", paste(items, collapse = ","), "}"))
    }
    return(paste0("[", paste(vapply(x, .wf_json, ""), collapse = ","), "]"))
  }
  if (length(x) == 1) return(scalar(x))
  paste0("[", paste(vapply(x, scalar, ""), collapse = ","), "]")
}

wf_audit_export <- function(w, file, extra = NULL) {
  stopifnot(inherits(w, "wf_weights") || inherits(w, "wf_autoweigh_result"))
  ledger <- NULL
  if (inherits(w, "wf_autoweigh_result")) {
    ledger <- lapply(w$ledger, function(e) e[c("step", "action", "detail", "time")])
    w <- w$weights
  }
  payload <- list(
    exported = format(Sys.time(), "%Y-%m-%dT%H:%M:%S%z"),
    n_units = nrow(w$data),
    weight_sum = sum(w$data$weight),
    groups = unique(w$data$group),
    provenance = w$provenance,
    decision_ledger = ledger,
    extra = extra
  )
  writeLines(.wf_json(payload), file, useBytes = TRUE)
  invisible(file)
}

make_pipeline_spec <- function(target, validate = NULL, stages = NULL) {
  if (is.null(stages)) {
    stages <- list(calibrate = list(method = "raking", tol = 1e-8))
  }
  wf_pipeline(target = target, stages = stages, validate = validate)
}

test_that("wf_pipeline declares a stable verified-target spec", {
  fixture <- make_safe_workflow_fixture()
  spec <- make_pipeline_spec(
    fixture$target,
    validate = list(max_deff = 6)
  )
  same <- make_pipeline_spec(
    fixture$target,
    validate = list(max_deff = 6)
  )

  expect_s3_class(spec, "wf_pipeline")
  expect_true("wf_pipeline" %in% getNamespaceExports("WFC"))
  expect_true("wf_run" %in% getNamespaceExports("WFC"))
  expect_true("wf_validate" %in% getNamespaceExports("WFC"))
  expect_true("wf_audit_export" %in% getNamespaceExports("WFC"))
  expect_identical(spec$hash, same$hash)
  expect_identical(spec$target$mode, "object")
  expect_identical(names(spec$stages), "calibrate")
  expect_output(print(spec), spec$hash, fixed = TRUE)

  expect_error(
    wf_pipeline(
      list(mode = "unknown"),
      list(calibrate = list(method = "raking"))
    ),
    class = "wf_error_input"
  )
  expect_error(
    wf_pipeline(fixture$target, list(clean = list())),
    class = "wf_error_input"
  )
  expect_error(
    make_pipeline_spec(fixture$target, validate = list(max_deff = -1)),
    class = "wf_error_input"
  )
})

test_that("wf_run executes verified object and population modes", {
  fixture <- make_safe_workflow_fixture()
  object_spec <- make_pipeline_spec(fixture$target)
  population_spec <- wf_pipeline(
    target = list(mode = "population"),
    stages = list(calibrate = list(method = "raking", tol = 1e-8))
  )

  object_result <- wf_run(object_spec, fixture$design, fixture$dims)
  population_result <- wf_run(
    population_spec,
    fixture$design,
    fixture$dims,
    population = fixture$target
  )
  direct <- .wf_rake_engine(
    fixture$design$data,
    fixture$target,
    id = fixture$design$roles$id,
    init_weight = fixture$design$roles$base_weight,
    tol = 1e-8
  )

  expect_s3_class(object_result, "wf_weights")
  expect_identical(object_result$provenance$pipeline_hash, object_spec$hash)
  expect_identical(object_result$provenance$pipeline$target_mode, "object")
  expect_equal(
    object_result$data$weight[match(direct$data$id, object_result$data$id)],
    direct$data$weight,
    tolerance = 1e-8
  )
  expect_equal(population_result$data$weight, object_result$data$weight)
})

test_that("wf_run owns design roles and blocks runtime propensity", {
  fixture <- make_safe_workflow_fixture()
  spec <- make_pipeline_spec(fixture$target)

  expect_error(
    wf_run(spec, fixture$design, base_weight = rep(1, nrow(fixture$design$data))),
    class = "wf_error_safety"
  )

  propensity_spec <- make_pipeline_spec(
    fixture$target,
    stages = list(
      propensity = list(formula = member ~ gender + age),
      calibrate = list(method = "raking")
    )
  )
  err <- tryCatch(
    wf_run(propensity_spec, fixture$design),
    error = identity
  )
  expect_s3_class(err, "wf_error_safety")
  expect_identical(err$data$code, "pipeline_propensity_unsupported")
})

test_that("pipeline validation records classed quality warnings", {
  fixture <- make_safe_workflow_fixture()
  spec <- make_pipeline_spec(
    fixture$target,
    validate = list(max_deff = 1)
  )

  expect_warning(
    result <- wf_run(spec, fixture$design, fixture$dims),
    class = "wf_warning_quality"
  )

  expect_s3_class(result$pipeline_validation, "wf_pipeline_validation")
  expect_false(result$pipeline_validation$ok)
  expect_true("max_deff" %in% result$pipeline_validation$issues$check)
})

test_that("wf_validate passes unchanged weights and detects drift", {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    tol = 1e-8
  )
  weights$provenance$method <- "raking"

  clean <- wf_validate(weights, weights, fixture$target, on_issue = "none")
  expect_s3_class(clean, "wf_validation")
  expect_true(clean$ok)
  expect_equal(nrow(clean$issues), 0)
  expect_equal(nrow(clean$comparison), 2)
  expect_output(print(clean), "PASS", fixed = TRUE)

  drifted <- weights
  drifted$data$weight[[1]] <- drifted$data$weight[[1]] * 20
  expect_warning(
    validation <- wf_validate(
      drifted,
      weights,
      fixture$target,
      max_deff_delta = 0.01,
      max_ratio_p99 = 1.05
    ),
    class = "wf_warning_quality"
  )
  expect_false(validation$ok)
  expect_true(any(validation$issues$check %in% c(
    "deff_delta",
    "unit_weight_ratio"
  )))
  expect_output(print(validation), "DRIFT DETECTED", fixed = TRUE)

  expect_error(
    wf_validate(
      drifted,
      weights,
      fixture$target,
      max_deff_delta = 0.01,
      on_issue = "error"
    ),
    class = "wf_error_input"
  )
})

test_that("wf_audit_export writes verified pipeline provenance", {
  fixture <- make_safe_workflow_fixture()
  spec <- make_pipeline_spec(fixture$target)
  result <- wf_run(spec, fixture$design, fixture$dims)
  path <- tempfile(fileext = ".json")

  out <- wf_audit_export(
    result,
    path,
    inputs = list(design = fixture$design, target = fixture$target),
    extra = list(release = "2.0.0")
  )
  text <- paste(readLines(path, warn = FALSE), collapse = "\n")

  expect_identical(out, path)
  expect_true(file.exists(path))
  expect_match(text, "\"schema\":\"wfc_audit_v2\"", fixed = TRUE)
  expect_match(text, "\"pipeline_hash\"", fixed = TRUE)
  expect_match(text, "\"input_hashes\"", fixed = TRUE)
  expect_match(text, "\"release\":\"2.0.0\"", fixed = TRUE)

  expect_error(
    wf_audit_export(result, file.path(path, "missing", "audit.json")),
    class = "wf_error_input"
  )
})

make_soft_fixture <- function() {
  dims <- wf_dims(segment = c("covered", "missing"))
  population <- data.frame(
    segment = c("covered", "missing"),
    count = c(90, 10),
    stringsAsFactors = FALSE
  )
  target <- wf_target_population(
    population,
    key_map = c(segment = "segment"),
    count = "count",
    dims = dims
  )
  sample <- data.frame(
    id = paste0("s", 1:4),
    segment = "covered",
    stringsAsFactors = FALSE
  )
  list(sample = sample, target = target, dims = dims)
}

make_ebal_fixture <- function() {
  dims <- wf_dims(gender = c("female", "male"))
  population <- data.frame(
    gender = c("female", "male"),
    count = c(2, 2),
    stringsAsFactors = FALSE
  )
  target <- wf_target_population(
    population,
    key_map = c(gender = "gender"),
    count = "count",
    dims = dims
  )
  sample <- data.frame(
    id = paste0("e", 1:4),
    gender = c("female", "female", "male", "male"),
    x = c(0, 1, 0, 1),
    stringsAsFactors = FALSE
  )
  list(sample = sample, target = target, dims = dims)
}

test_that("soft calibration relaxes infeasible margins within declared tolerance", {
  fixture <- make_soft_fixture()

  soft <- .wf_calibrate_engine(
    fixture$sample,
    fixture$target,
    method = "soft",
    id = "id",
    tolerance = 0.11
  )

  expect_s3_class(soft, "wf_weights")
  expect_identical(soft$provenance$method, "soft")
  expect_equal(sum(soft$data$weight), 100, tolerance = 1e-8)
  expect_equal(soft$achieved[["_all_"]]$segment[["missing"]], 0)
  expect_true(all(soft$relaxation$within_tolerance))
  expect_true(any(soft$relaxation$dim == "segment" & soft$relaxation$relaxed))
  expect_true("soft_relaxation" %in% names(wf_report(soft)$sections))
  expect_output(print(soft), "soft calibration", fixed = TRUE)
})

test_that("soft calibration refuses relaxation beyond tolerance", {
  fixture <- make_soft_fixture()

  expect_error(
    .wf_calibrate_engine(
      fixture$sample,
      fixture$target,
      method = "soft",
      id = "id",
      tolerance = 0.05
    ),
    class = "wf_error_feasibility"
  )
  expect_error(
    .wf_calibrate_engine(
      fixture$sample,
      fixture$target,
      method = "soft",
      id = "id",
      tolerance = c(segment = -0.1)
    ),
    class = "wf_error_input"
  )
})

test_that("soft calibration still blocks non-relaxable precheck errors", {
  fixture <- make_soft_fixture()
  bad <- fixture$sample
  bad$segment[[1]] <- "unknown"

  expect_error(
    .wf_calibrate_engine(
      bad,
      fixture$target,
      method = "soft",
      id = "id",
      tolerance = 0.2
    ),
    class = "wf_error_feasibility"
  )
})

test_that("entropy balancing hits verified categorical margins", {
  fixture <- make_ebal_fixture()

  ebal <- .wf_calibrate_engine(
    fixture$sample,
    fixture$target,
    method = "ebal",
    id = "id",
    tol = 1e-10
  )

  expect_s3_class(ebal, "wf_weights")
  expect_identical(ebal$provenance$method, "ebal")
  expect_identical(ebal$provenance$distance, "entropy")
  expect_equal(sum(ebal$data$weight), 4, tolerance = 1e-8)
  expect_equal(ebal$achieved[["_all_"]]$gender[["male"]], 2, tolerance = 1e-8)
  expect_true(is.finite(ebal$log$kl_divergence))
  expect_output(print(ebal), "entropy balancing", fixed = TRUE)
})

test_that("pipeline calibration stage accepts categorical entropy balancing", {
  fixture <- make_safe_workflow_fixture()
  spec <- wf_pipeline(
    target = fixture$target,
    stages = list(calibrate = list(method = "ebal"))
  )

  out <- wf_run(spec, fixture$design)

  expect_s3_class(out, "wf_weights")
  expect_identical(out$provenance$method, "ebal")
  expect_identical(out$provenance$pipeline_hash, spec$hash)
})

#' Calibrate weights from verified design and target objects
#'
#' Enforces the WFC 2.0 verified-input boundary before dispatching to a
#' calibration engine. Raw data frames and ordinary `wf_target` objects are not
#' accepted.
#'
#' @param design An unchanged `wf_design_data` created by
#'   [wf_prepare_design()].
#' @param target An unchanged, non-demo `wf_verified_target` created by
#'   [wf_import_target()] or [wf_import_reference()].
#' @param method Calibration method: `"raking"`, `"poststrat"`, `"greg"`
#'   (linear GREG), `"logit"` (bounded), `"soft"` (declared margin
#'   relaxation), or `"ebal"` (categorical entropy balancing).
#' @param ... Method settings. ID and base-weight roles come from `design` and
#'   cannot be overridden. Inline target moments are not supported.
#'
#' @return A `wf_weights` object carrying the verified input identities.
#' @export
wf_calibrate <- function(design, target, method = "raking", ...) {
  dots <- list(...)
  if ("moments" %in% names(dots)) {
    .wf_safety_abort(
      "inline_moments_unsupported",
      paste(
        "WFC 2.0 does not accept inline target moments.",
        "Use verified external margins."
      ),
      "moments",
      next_actions = "import_verified_external_margins"
    )
  }
  .wf_execute_verified_engine(design, target, method, dots)
}

#' Dispatch one already-validated calibration engine.
#'
#' @param sample Internal design data frame.
#' @param target Internal verified target.
#' @param method Calibration method.
#' @param ... Engine settings.
#' @keywords internal
#' @noRd
.wf_calibrate_engine <- function(sample, target, method = "raking", ...) {
  supported <- c("raking", "poststrat", "greg", "logit", "soft", "ebal")
  if (length(method) != 1L || !method %in% supported) {
    shown <- if (length(method) == 0L) "<empty>" else as.character(method[[1]])
    wf_abort(
      sprintf(
        "Unsupported calibration method '%s'. Supported methods: %s.",
        shown,
        paste(supported, collapse = ", ")
      ),
      "wf_error_input",
      list(method = method)
    )
  }
  dots <- list(...)

  if (method == "raking") {
    out <- do.call(
      .wf_rake_engine,
      c(list(sample = sample, target = target), dots)
    )
    out$provenance$method <- "raking"
    return(out)
  }

  if (method == "poststrat") {
    return(do.call(
      .wf_poststrat_engine,
      c(list(sample = sample, target = target), dots)
    ))
  }

  if (method == "soft") {
    return(do.call(
      .wf_softcalibrate,
      c(list(sample = sample, target = target), dots)
    ))
  }

  bounds <- dots$bounds
  if (method == "logit") {
    if (is.null(bounds) || length(bounds) != 2L || !is.numeric(bounds) ||
        anyNA(bounds) || !(bounds[[1]] > 0 && bounds[[1]] < 1 &&
                           bounds[[2]] > 1)) {
      wf_abort(
        "method='logit' requires bounds = c(L, U) with 0 < L < 1 < U.",
        "wf_error_input",
        list(bounds = bounds)
      )
    }
  }

  distance <- switch(
    method,
    greg = "linear",
    logit = "logit",
    ebal = "entropy"
  )
  do.call(
    .wf_lincalibrate,
    c(
      list(
        sample = sample,
        target = target,
        distance = distance,
        method = method
      ),
      dots
    )
  )
}

#' Simulated survey weighting example data
#'
#' A small, fully simulated dataset for examples and tests. It contains no
#' private source records.
#'
#' @format A list with three elements:
#' \describe{
#'   \item{sample}{Simulated respondent-level sample data.}
#'   \item{population}{Simulated population cell counts.}
#'   \item{dims}{A `wf_dims` object for `gender` and `age`.}
#' }
#' @source Generated by `data-raw/make-wfc-example.R`.
"wfc_example"

#' Calculate the identity of a weight plan.
#'
#' @param plan Weight-plan fields.
#' @keywords internal
#' @noRd
.wf_weight_plan_identity <- function(plan) {
  material <- plan[setdiff(names(plan), c("created", "identity"))]
  if (!is.null(material$cell_plan)) {
    material$cell_plan <- material$cell_plan$identity
  }
  .wf_sha256_object(material)
}

#' Build a reviewable outcome-blind weight plan
#'
#' Performs structural checks and records approved-capable settings without
#' calling a calibration engine or computing any weights.
#'
#' @param design A `wf_design_data` object.
#' @param target A non-demo `wf_verified_target` object.
#' @param dims A `wf_dims` object matching the target.
#' @param method Planned method: `"raking"`, `"logit"`, or `"poststrat"`.
#' @param bounds Planned multiplicative lower and upper bounds, satisfying
#'   `0 < L < 1 < U`.
#' @param min_cell Minimum support recorded in the plan.
#' @param cell_plan Optional reviewed `wf_cell_merge_plan` to apply before
#'   calibration.
#'
#' @return A `wf_weight_plan`. The `weights` field is always `NULL`.
#' @export
wf_plan_weights <- function(design, target, dims,
                            method = c("raking", "logit", "poststrat"),
                            bounds = c(0.3, 3), min_cell = 5,
                            cell_plan = NULL) {
  if (!inherits(design, "wf_design_data") ||
      !identical(design$identity, .wf_design_identity(design$data, design$roles))) {
    .wf_safety_abort(
      "design_object_invalid",
      "`design` must be an unchanged wf_design_data object.",
      "design"
    )
  }
  if (!inherits(target, "wf_verified_target") || isTRUE(target$demo_only) ||
      !identical(target$identity, .wf_verified_target_identity(target))) {
    .wf_safety_abort(
      "verified_target_required",
      "`target` must be an unchanged, non-demo verified external target.",
      "target"
    )
  }
  if (!inherits(dims, "wf_dims") ||
      !identical(names(dims$vars), target$dims)) {
    .wf_safety_abort(
      "target_dimensions_mismatch",
      "Declared dimensions must exactly match the verified target.",
      "dims"
    )
  }
  if (length(method) < 1L || !is.character(method) ||
      !method[[1]] %in% c("raking", "logit", "poststrat")) {
    .wf_safety_abort(
      "weight_method_unsupported",
      "`method` must be raking, logit, or poststrat.",
      "method"
    )
  }
  method <- match.arg(method)
  if (!is.numeric(bounds) || length(bounds) != 2L || anyNA(bounds) ||
      any(!is.finite(bounds)) ||
      !(bounds[[1]] > 0 && bounds[[1]] < 1 && bounds[[2]] > 1)) {
    .wf_safety_abort(
      "weight_bounds_invalid",
      "`bounds` must satisfy 0 < lower < 1 < upper.",
      "bounds"
    )
  }
  if (!is.numeric(min_cell) || length(min_cell) != 1L ||
      is.na(min_cell) || !is.finite(min_cell) || min_cell < 1 ||
      min_cell != as.integer(min_cell)) {
    .wf_safety_abort(
      "min_cell_invalid",
      "`min_cell` must be a positive integer.",
      "min_cell"
    )
  }
  if (identical(method, "poststrat") && is.null(target$joint)) {
    .wf_safety_abort(
      "poststrat_joint_target_required",
      "Post-stratification requires verified joint population cells.",
      "target",
      next_actions = "import_joint_population_target"
    )
  }

  effective_design <- design
  effective_target <- target
  if (!is.null(cell_plan)) {
    if (!inherits(cell_plan, "wf_cell_merge_plan") ||
        !identical(cell_plan$identity, .wf_cell_plan_identity(cell_plan))) {
      .wf_safety_abort(
        "cell_plan_invalid",
        "`cell_plan` must be an unchanged wf_cell_merge_plan object.",
        "cell_plan"
      )
    }
    if (!identical(cell_plan$input_identities$design, design$identity) ||
        !identical(cell_plan$input_identities$target, target$identity)) {
      .wf_safety_abort(
        "cell_plan_inputs_mismatch",
        "The cell plan does not match the supplied design and target.",
        "cell_plan"
      )
    }
    applied <- .wf_apply_cell_plan(design, target, cell_plan)
    effective_design <- applied$design
    effective_target <- applied$target
  }

  precheck <- wf_precheck(
    effective_design$data,
    effective_target,
    id = effective_design$roles$id,
    thin_min = as.integer(min_cell),
    risk_ratio = max(bounds[[2]], 1 / bounds[[1]])
  )
  plan <- list(
    method = method,
    settings = list(
      bounds = unname(bounds),
      min_cell = as.integer(min_cell),
      dims = names(dims$vars)
    ),
    precheck = precheck,
    issues = precheck$issues,
    ready = isTRUE(precheck$ok),
    cell_plan = cell_plan,
    input_identities = list(
      design = design$identity,
      target = target$identity,
      cell_plan = if (is.null(cell_plan)) NULL else cell_plan$identity
    ),
    effective_identities = list(
      design = effective_design$identity,
      target = effective_target$identity
    ),
    weights = NULL,
    created = .wf_iso_time(),
    package_version = .wf_package_version()
  )
  plan$identity <- .wf_weight_plan_identity(plan)
  class(plan) <- "wf_weight_plan"
  plan
}

#' Print a reviewable weight plan
#'
#' @param x A `wf_weight_plan` object.
#' @param ... Reserved for future use.
#'
#' @return `x`, invisibly.
#' @export
print.wf_weight_plan <- function(x, ...) {
  cat("<wf_weight_plan>\n")
  cat(" Method: ", x$method, "\n", sep = "")
  cat(" Review status: ", if (x$ready) "READY" else "BLOCKED", "\n", sep = "")
  cat(" Issues: ", nrow(x$issues), "\n", sep = "")
  cat(" Weights computed: no\n")
  cat(" Identity: ", x$identity, "\n", sep = "")
  invisible(x)
}

#' Combine replicate weights and an estimator into a variance and CI.
#'
#' Applies the unified replication rule
#' `Var = scale * sum_r rscales_r * (theta_r - theta)^2` to any estimator, using
#' the `(scale, rscales)` stored by [wf_replicates()].
#'
#' @param replicates A `wf_replicate_weights` object.
#' @param estimator A closure `function(weights, data) -> numeric` (scalar or
#'   named vector).
#' @param data The data frame the estimator reads.
#' @param level Confidence level in `(0, 1)`.
#' @param ci Interval type: `"normal"` or (bootstrap only) `"percentile"`.
#' @return A `wf_variance_result` object.
#' @export
wf_variance <- function(replicates, estimator, data, level = 0.95,
                        ci = c("normal", "percentile")) {
  if (!inherits(replicates, "wf_replicate_weights")) {
    wf_abort("`replicates` must be a wf_replicate_weights object.",
             "wf_error_input")
  }
  if (!is.function(estimator)) {
    wf_abort("`estimator` must be a function(weights, data).", "wf_error_input")
  }
  if (length(level) != 1 || !is.finite(level) || level <= 0 || level >= 1) {
    wf_abort("`level` must be a single number in (0, 1).", "wf_error_input",
             list(level = level))
  }
  ci <- match.arg(ci)
  if (ci == "percentile" && replicates$method != "bootstrap") {
    wf_abort("`ci = 'percentile'` is only valid for bootstrap replicates.",
             "wf_error_input", list(method = replicates$method))
  }

  base_est <- estimator(replicates$base$weight, data)
  nm <- names(base_est)
  theta <- as.numeric(base_est)
  q <- length(theta)
  if (q == 0) {
    wf_abort("`estimator` returned a length-zero result.", "wf_error_input")
  }

  R <- ncol(replicates$replicates)
  tr <- matrix(NA_real_, q, R)
  for (r in seq_len(R)) {
    v <- as.numeric(estimator(replicates$replicates[, r], data))
    if (length(v) != q) {
      wf_abort("`estimator` returned inconsistent length across replicates.",
               "wf_error_input")
    }
    tr[, r] <- v
  }

  dev2 <- (tr - theta)^2
  variance <- replicates$scale * as.numeric(dev2 %*% replicates$rscales)
  se <- sqrt(variance)

  if (is.null(nm)) {
    nm <- if (q == 1) "estimate" else as.character(seq_len(q))
  }

  if (ci == "normal") {
    z <- stats::qnorm(1 - (1 - level) / 2)
    lo <- theta - z * se
    hi <- theta + z * se
  } else {
    a <- (1 - level) / 2
    lo <- vapply(seq_len(q),
                 function(i) stats::quantile(tr[i, ], a, names = FALSE),
                 numeric(1))
    hi <- vapply(seq_len(q),
                 function(i) stats::quantile(tr[i, ], 1 - a, names = FALSE),
                 numeric(1))
  }

  structure(list(
    table = data.frame(
      quantity = nm, estimate = theta, variance = variance, se = se,
      ci_lower = lo, ci_upper = hi, stringsAsFactors = FALSE
    ),
    provenance = list(method = replicates$method, level = level,
                      ci = ci, R = R)
  ), class = "wf_variance_result")
}

#' Print a variance result
#'
#' @param x A `wf_variance_result` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_variance_result <- function(x, ...) {
  cat(sprintf("<wf_variance_result>  method: %s; %d replicate(s); %.0f%% %s CI\n",
              x$provenance$method, x$provenance$R,
              100 * x$provenance$level, x$provenance$ci))
  print(x$table, row.names = FALSE)
  invisible(x)
}

#' Coerce a variance result to a data frame
#'
#' @param x A `wf_variance_result` object.
#' @param ... Unused.
#' @return The result table as a data frame.
#' @export
as.data.frame.wf_variance_result <- function(x, ...) {
  x$table
}

.wf_source_fields <- c(
  "publisher",
  "dataset_title",
  "citation",
  "reference_period",
  "population_scope",
  "retrieved_at",
  "license",
  "checksum_algorithm",
  "checksum",
  "transformation",
  "selected_before_outcomes",
  "demo_only"
)

#' Read a CSV or Excel target table.
#'
#' @param path Path to the target table.
#' @keywords internal
#' @noRd
.wf_read_target_table <- function(path) {
  if (!.wf_is_string(path) || !file.exists(path)) {
    .wf_safety_abort(
      "source_file_missing",
      "Target data file does not exist.",
      "data_file"
    )
  }

  ext <- tolower(tools::file_ext(path))
  if (ext == "csv") {
    return(utils::read.csv(path, stringsAsFactors = FALSE))
  }
  if (ext == "xlsx") {
    if (!requireNamespace("openxlsx", quietly = TRUE)) {
      wf_abort(
        "Package 'openxlsx' is required for Excel targets.",
        "wf_error_dependency",
        list(package = "openxlsx")
      )
    }
    return(openxlsx::read.xlsx(path))
  }

  .wf_safety_abort(
    "target_format_unsupported",
    "Target file must be CSV or XLSX.",
    "data_file",
    next_actions = "use_csv_or_xlsx"
  )
}

#' Write a CSV or Excel target table.
#'
#' @param data Target data frame.
#' @param path Output path.
#' @keywords internal
#' @noRd
.wf_write_target_table <- function(data, path) {
  ext <- tolower(tools::file_ext(path))
  if (ext == "csv") {
    utils::write.csv(data, path, row.names = FALSE)
    return(invisible(path))
  }
  if (ext == "xlsx") {
    if (!requireNamespace("openxlsx", quietly = TRUE)) {
      wf_abort(
        "Package 'openxlsx' is required for Excel targets.",
        "wf_error_dependency",
        list(package = "openxlsx")
      )
    }
    openxlsx::write.xlsx(data, path, overwrite = FALSE)
    return(invisible(path))
  }

  .wf_safety_abort(
    "target_format_unsupported",
    "Target file must be CSV or XLSX.",
    "file",
    next_actions = "use_csv_or_xlsx"
  )
}

#' Read and validate target source evidence.
#'
#' @param source_file Path to a one-record DCF source file.
#' @param data_file Path to the target data file described by the source record.
#' @param production Whether demo-only evidence must be rejected.
#' @keywords internal
#' @noRd
.wf_read_source_evidence <- function(source_file, data_file, production) {
  if (!.wf_is_string(source_file) || !file.exists(source_file)) {
    .wf_safety_abort(
      "source_metadata_missing",
      "Target source metadata file does not exist.",
      "source_file",
      next_actions = "supply_source_metadata"
    )
  }
  if (!is.logical(production) || length(production) != 1L || is.na(production)) {
    .wf_safety_abort(
      "production_flag_invalid",
      "`production` must be TRUE or FALSE.",
      "production"
    )
  }

  source <- tryCatch(
    read.dcf(source_file, all = TRUE),
    error = function(e) {
      .wf_safety_abort(
        "source_metadata_invalid",
        "Target source metadata must be a valid DCF file.",
        "source_file",
        evidence = list(reason = conditionMessage(e)),
        next_actions = "repair_source_metadata"
      )
    }
  )
  if (nrow(source) != 1L) {
    .wf_safety_abort(
      "source_metadata_record_count",
      "Target source metadata must contain exactly one record.",
      "source_file",
      evidence = list(records = nrow(source))
    )
  }

  missing <- setdiff(.wf_source_fields, colnames(source))
  empty <- intersect(
    .wf_source_fields,
    colnames(source)[is.na(source[1, ]) | !nzchar(trimws(source[1, ]))]
  )
  if (length(missing) || length(empty)) {
    .wf_safety_abort(
      "source_metadata_incomplete",
      "Target source metadata are incomplete.",
      "source_file",
      evidence = list(missing = missing, empty = empty),
      next_actions = "complete_source_metadata"
    )
  }

  values <- stats::setNames(
    as.character(source[1, .wf_source_fields, drop = TRUE]),
    .wf_source_fields
  )
  algorithm <- tolower(trimws(values[["checksum_algorithm"]]))
  if (!identical(algorithm, "sha256")) {
    .wf_safety_abort(
      "checksum_algorithm_unsupported",
      "Target source checksum algorithm must be SHA-256.",
      "checksum_algorithm"
    )
  }

  computed <- .wf_sha256_file(data_file)
  recorded <- tolower(trimws(values[["checksum"]]))
  if (!grepl("^[0-9a-f]{64}$", recorded) || !identical(recorded, computed)) {
    .wf_safety_abort(
      "source_checksum_mismatch",
      "Target data do not match the recorded SHA-256 checksum.",
      "checksum",
      evidence = list(recorded = recorded, computed = computed),
      next_actions = "verify_target_file"
    )
  }

  parse_flag <- function(name) {
    value <- tolower(trimws(values[[name]]))
    if (!value %in% c("true", "false")) {
      .wf_safety_abort(
        "source_flag_invalid",
        sprintf("Source field '%s' must be true or false.", name),
        name
      )
    }
    identical(value, "true")
  }
  selected_before_outcomes <- parse_flag("selected_before_outcomes")
  demo_only <- parse_flag("demo_only")

  if (!selected_before_outcomes) {
    .wf_safety_abort(
      "target_selected_after_outcomes",
      "The target source must be selected before study outcomes are examined.",
      "selected_before_outcomes",
      next_actions = "use_preselected_external_target"
    )
  }
  if (production && demo_only) {
    .wf_safety_abort(
      "demo_target_in_production",
      "Demo-only target data cannot be used in production.",
      "demo_only",
      next_actions = "supply_authoritative_target"
    )
  }

  list(
    fields = as.list(values),
    data_checksum = computed,
    source_checksum = .wf_sha256_file(source_file),
    selected_before_outcomes = selected_before_outcomes,
    demo_only = demo_only
  )
}

#' Calculate the identity of a verified target.
#'
#' @param target A verified target with evidence fields attached.
#' @keywords internal
#' @noRd
.wf_verified_target_identity <- function(target) {
  .wf_sha256_object(list(
    mode = target$mode,
    by = target$by,
    dims = target$dims,
    groups = target$groups,
    joint = target$joint,
    evidence = target$evidence,
    demo_only = target$demo_only,
    source_type = target$source_type
  ))
}

#' Attach verified source evidence to a target.
#'
#' @param target A `wf_target` object.
#' @param evidence Validated source evidence.
#' @param source_type Source type label.
#' @keywords internal
#' @noRd
.wf_verified_target <- function(target, evidence, source_type) {
  target$evidence <- evidence
  target$demo_only <- evidence$demo_only
  target$source_type <- source_type
  target$identity <- .wf_verified_target_identity(target)
  class(target) <- c("wf_verified_target", "wf_target")
  target
}

#' Create a safe external-target import template
#'
#' Writes a CSV or Excel data template and a separate companion DCF source
#' record. The source filename appends `.source.dcf` to the complete data
#' filename so CSV and Excel files cannot accidentally share a checksum.
#'
#' @param file Output path ending in `.csv` or `.xlsx`.
#' @param dims A `wf_dims` object with declared levels when `example = TRUE`.
#' @param by Optional grouping-column name.
#' @param example Whether to write synthetic, demo-only rows instead of a blank
#'   template.
#'
#' @return Invisibly, a list with `data_file` and `source_file` paths.
#' @export
wf_target_template <- function(file, dims, by = NULL, example = FALSE) {
  if (!.wf_is_string(file)) {
    .wf_safety_abort(
      "target_template_path_invalid",
      "`file` must be one CSV or XLSX path.",
      "file"
    )
  }
  if (!inherits(dims, "wf_dims")) {
    .wf_safety_abort(
      "target_dimensions_invalid",
      "`dims` must be a wf_dims object.",
      "dims"
    )
  }
  if (!is.logical(example) || length(example) != 1L || is.na(example)) {
    .wf_safety_abort(
      "target_template_example_invalid",
      "`example` must be TRUE or FALSE.",
      "example"
    )
  }

  source_file <- paste0(file, ".source.dcf")
  existing <- c(file, source_file)[file.exists(c(file, source_file))]
  if (length(existing)) {
    .wf_safety_abort(
      "target_template_exists",
      "The target template or its source record already exists.",
      "file",
      evidence = list(paths = existing),
      next_actions = "choose_new_template_path"
    )
  }
  if (!dir.exists(dirname(file))) {
    .wf_safety_abort(
      "target_template_directory_missing",
      "The target template directory does not exist.",
      "file",
      evidence = list(directory = dirname(file)),
      next_actions = "create_template_directory"
    )
  }

  dim_names <- names(dims$vars)
  columns <- c(by, dim_names, "count")
  if (example) {
    missing_levels <- dim_names[vapply(dims$vars, is.null, logical(1))]
    if (length(missing_levels)) {
      .wf_safety_abort(
        "target_template_levels_missing",
        "Synthetic examples require declared levels for every dimension.",
        "dims",
        evidence = list(dimensions = missing_levels),
        next_actions = "declare_dimension_levels"
      )
    }
    data <- expand.grid(
      dims$vars,
      KEEP.OUT.ATTRS = FALSE,
      stringsAsFactors = FALSE
    )
    if (!is.null(by)) {
      data[[by]] <- "example_group"
    }
    data$count <- rep(100, nrow(data))
    data <- data[, columns, drop = FALSE]
  } else {
    data <- as.data.frame(
      stats::setNames(
        lapply(columns, function(name) {
          if (identical(name, "count")) numeric() else character()
        }),
        columns
      ),
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  }

  .wf_write_target_table(data, file)
  if (example) {
    source <- c(
      publisher = "WFC synthetic example",
      dataset_title = "Synthetic population target example",
      citation = "WFC package synthetic example; not an external authority",
      reference_period = "synthetic",
      population_scope = "synthetic demonstration population",
      retrieved_at = as.character(Sys.Date()),
      license = "CC0-1.0",
      checksum_algorithm = "sha256",
      checksum = .wf_sha256_file(file),
      transformation = "Generated Cartesian example with equal counts",
      selected_before_outcomes = "true",
      demo_only = "true"
    )
  } else {
    source <- c(
      publisher = "",
      dataset_title = "",
      citation = "",
      reference_period = "",
      population_scope = "",
      retrieved_at = "",
      license = "",
      checksum_algorithm = "sha256",
      checksum = .wf_sha256_file(file),
      transformation = "",
      selected_before_outcomes = "false",
      demo_only = "false"
    )
  }
  writeLines(sprintf("%s: %s", names(source), source), source_file)

  invisible(list(data_file = file, source_file = source_file))
}

#' Import a verified external population target
#'
#' Imports CSV or Excel population margins only when a companion DCF record
#' supplies complete provenance and a matching SHA-256 checksum.
#'
#' @param data_file Path to a CSV or XLSX population table.
#' @param source_file Path to its companion source DCF record.
#' @param dims A `wf_dims` object.
#' @param key_map Named mapping from dimensions to data columns.
#' @param count Name of the population-count column.
#' @param by Optional grouping variable.
#' @param by_key Optional group-key column or function.
#' @param production Whether to reject demo-only sources.
#'
#' @return A `wf_verified_target` object.
#' @examples
#' # WFC ships a synthetic demo target beside its companion evidence record.
#' csv_file <- system.file(
#'   "extdata", "safe-target-example.csv",
#'   package = "WFC"
#' )
#' csv_source <- paste0(csv_file, ".source.dcf")
#'
#' dims_safe <- wf_dims(sex = c("F", "M"), age = c("18-34", "35+"))
#'
#' # `production = FALSE` is required here only because the bundled file is
#' # declared demo-only. Do not use it to admit an authoritative source whose
#' # evidence record is incomplete.
#' target <- wf_import_target(
#'   csv_file,
#'   csv_source,
#'   dims_safe,
#'   key_map = c(sex = "sex", age = "age"),
#'   count = "count",
#'   production = FALSE
#' )
#' target$identity
#'
#' # The same demo file is refused as a production source, so a synthetic
#' # file cannot impersonate an authoritative population.
#' try(wf_import_target(
#'   csv_file,
#'   csv_source,
#'   dims_safe,
#'   key_map = c(sex = "sex", age = "age"),
#'   count = "count"
#' ))
#' @export
wf_import_target <- function(data_file, source_file, dims, key_map, count,
                             by = NULL, by_key = NULL, production = TRUE) {
  evidence <- .wf_read_source_evidence(source_file, data_file, production)
  population <- .wf_read_target_table(data_file)
  target <- wf_target_population(
    population,
    key_map = key_map,
    count = count,
    dims = dims,
    by = by,
    by_key = by_key,
    keep_joint = TRUE
  )

  .wf_verified_target(target, evidence, "population")
}

#' Import a verified external reference-sample target
#'
#' Imports an external reference sample only when every file column has a
#' declared design role and the companion DCF record passes the provenance and
#' checksum checks.
#'
#' @param data_file Path to a CSV or XLSX reference sample.
#' @param source_file Path to its companion source DCF record.
#' @param dims A `wf_dims` object.
#' @param feature Name of the reciprocal-design-weight column.
#' @param by Optional grouping variable.
#' @param production Whether to reject demo-only sources.
#'
#' @return A `wf_verified_target` object.
#' @export
wf_import_reference <- function(data_file, source_file, dims, feature,
                                by = NULL, production = TRUE) {
  evidence <- .wf_read_source_evidence(source_file, data_file, production)
  reference <- .wf_read_target_table(data_file)
  allowed <- unique(c(names(dims$vars), feature, by))
  extra <- setdiff(names(reference), allowed)
  if (length(extra)) {
    .wf_safety_abort(
      "reference_columns_unassigned",
      sprintf(
        "Reference data contain unassigned column(s): %s.",
        paste(extra, collapse = ", ")
      ),
      evidence = list(columns = extra),
      next_actions = "remove_outcome_columns"
    )
  }

  .require_cols(reference, allowed, "reference data")
  reference <- reference[, allowed, drop = FALSE]
  target <- wf_target_reference(
    reference,
    feature = feature,
    dims = dims,
    by = by
  )

  .wf_verified_target(target, evidence, "reference")
}

make_safe_target_files <- function(demo_only = FALSE, bad_checksum = FALSE) {
  path <- tempfile("wf-safe-target-")
  dir.create(path)

  data_file <- file.path(path, "population.csv")
  source_file <- file.path(path, "population.source.dcf")
  data <- data.frame(
    sex = c("F", "M"),
    count = c(52, 48),
    stringsAsFactors = FALSE
  )
  utils::write.csv(data, data_file, row.names = FALSE)

  checksum <- .wf_sha256_file(data_file)
  if (bad_checksum) {
    checksum <- paste0(strrep("0", 63), "1")
  }
  source <- c(
    publisher = "Synthetic example publisher",
    dataset_title = "Synthetic population margins",
    citation = "Synthetic fixture; not an external authority",
    reference_period = "2026",
    population_scope = "Synthetic adults",
    retrieved_at = "2026-07-15",
    license = "CC0-1.0",
    checksum_algorithm = "sha256",
    checksum = checksum,
    transformation = "No transformation",
    selected_before_outcomes = "true",
    demo_only = if (demo_only) "true" else "false"
  )
  writeLines(sprintf("%s: %s", names(source), source), source_file)

  list(
    data = data_file,
    source = source_file,
    dims = wf_dims(sex = c("F", "M"))
  )
}

make_safe_reference_files <- function(extra_outcome = FALSE) {
  path <- tempfile("wf-safe-reference-")
  dir.create(path)

  data_file <- file.path(path, "reference.csv")
  source_file <- file.path(path, "reference.source.dcf")
  data <- data.frame(
    sex = c("F", "F", "M", "M"),
    inclusion_probability = c(0.5, 0.5, 0.25, 0.25),
    stringsAsFactors = FALSE
  )
  if (extra_outcome) {
    data$outcome <- c(1, 0, 1, 0)
  }
  utils::write.csv(data, data_file, row.names = FALSE)

  source <- c(
    publisher = "Synthetic example publisher",
    dataset_title = "Synthetic reference sample",
    citation = "Synthetic fixture; not an external authority",
    reference_period = "2026",
    population_scope = "Synthetic adults",
    retrieved_at = "2026-07-15",
    license = "CC0-1.0",
    checksum_algorithm = "sha256",
    checksum = .wf_sha256_file(data_file),
    transformation = "No transformation",
    selected_before_outcomes = "true",
    demo_only = "false"
  )
  writeLines(sprintf("%s: %s", names(source), source), source_file)

  list(
    data = data_file,
    source = source_file,
    dims = wf_dims(sex = c("F", "M"))
  )
}

test_that("complete source evidence creates a verified target", {
  f <- make_safe_target_files(demo_only = FALSE)

  x <- wf_import_target(
    f$data,
    f$source,
    f$dims,
    c(sex = "sex"),
    "count"
  )

  expect_s3_class(x, "wf_verified_target")
  expect_false(x$demo_only)
  expect_identical(x$source_type, "population")
  expect_match(x$identity, "^[0-9a-f]{64}$")
  expect_false(is.null(x$joint))
})

test_that("bad checksums block target import", {
  f <- make_safe_target_files(bad_checksum = TRUE)

  expect_error(
    wf_import_target(
      f$data,
      f$source,
      f$dims,
      c(sex = "sex"),
      "count"
    ),
    class = "wf_error_safety"
  )
})

test_that("production use of demo targets blocks", {
  f <- make_safe_target_files(demo_only = TRUE)

  expect_error(
    wf_import_target(
      f$data,
      f$source,
      f$dims,
      c(sex = "sex"),
      "count"
    ),
    class = "wf_error_safety"
  )
})

test_that("reference targets import only declared design columns", {
  f <- make_safe_reference_files()

  x <- wf_import_reference(
    f$data,
    f$source,
    f$dims,
    "inclusion_probability"
  )

  expect_s3_class(x, "wf_verified_target")
  expect_identical(x$source_type, "reference")
})

test_that("unassigned reference outcome columns block", {
  f <- make_safe_reference_files(extra_outcome = TRUE)

  expect_error(
    wf_import_reference(
      f$data,
      f$source,
      f$dims,
      "inclusion_probability"
    ),
    class = "wf_error_safety"
  )
})

test_that("CSV and Excel examples have separate evidence and equal margins", {
  skip_if_not_installed("openxlsx")
  path <- tempfile("wf-target-templates-")
  dir.create(path)
  csv_file <- file.path(path, "target.csv")
  xlsx_file <- file.path(path, "target.xlsx")
  dims <- wf_dims(sex = c("F", "M"))

  csv_files <- wf_target_template(csv_file, dims, example = TRUE)
  xlsx_files <- wf_target_template(xlsx_file, dims, example = TRUE)
  csv_target <- wf_import_target(
    csv_files$data_file,
    csv_files$source_file,
    dims,
    c(sex = "sex"),
    "count",
    production = FALSE
  )
  xlsx_target <- wf_import_target(
    xlsx_files$data_file,
    xlsx_files$source_file,
    dims,
    c(sex = "sex"),
    "count",
    production = FALSE
  )

  expect_equal(csv_target$groups, xlsx_target$groups)
  expect_true(file.exists(paste0(csv_file, ".source.dcf")))
  expect_true(file.exists(paste0(xlsx_file, ".source.dcf")))
  expect_false(identical(
    csv_target$evidence$data_checksum,
    xlsx_target$evidence$data_checksum
  ))
})

test_that("blank templates require source details before import", {
  path <- tempfile(fileext = ".csv")
  dims <- wf_dims(sex = c("F", "M"))

  files <- wf_target_template(path, dims)

  expect_equal(nrow(utils::read.csv(path)), 0)
  expect_error(
    wf_import_target(
      files$data_file,
      files$source_file,
      dims,
      c(sex = "sex"),
      "count"
    ),
    class = "wf_error_safety"
  )
})

test_that("bundled CSV and Excel import examples remain usable", {
  skip_if_not_installed("openxlsx")
  path <- system.file("extdata", package = "WFC")
  dims <- wf_dims(sex = c("F", "M"), age = c("18-34", "35+"))
  csv_file <- file.path(path, "safe-target-example.csv")
  xlsx_file <- file.path(path, "safe-target-example.xlsx")

  csv_target <- wf_import_target(
    csv_file,
    paste0(csv_file, ".source.dcf"),
    dims,
    c(sex = "sex", age = "age"),
    "count",
    production = FALSE
  )
  xlsx_target <- wf_import_target(
    xlsx_file,
    paste0(xlsx_file, ".source.dcf"),
    dims,
    c(sex = "sex", age = "age"),
    "count",
    production = FALSE
  )

  expect_equal(csv_target$groups, xlsx_target$groups)
})

#' Validate and resolve top count for influence output.
#'
#' @param top Requested top count.
#' @keywords internal
#' @noRd
.wf_influence_top <- function(top) {
  if (length(top) != 1 || !is.numeric(top) || !is.finite(top) ||
      top < 1 || top != as.integer(top)) {
    wf_abort("`top` must be a positive integer.", "wf_error_input")
  }
  as.integer(top)
}

#' Compute a design effect from weights.
#'
#' @param weight Numeric weights.
#' @keywords internal
#' @noRd
.wf_influence_deff <- function(weight) {
  if (length(weight) < 2 || mean(weight) <= 0) {
    return(NA_real_)
  }
  1 + (stats::sd(weight) / mean(weight))^2
}

#' Add optional target-margin influence shares.
#'
#' @param table Influence table.
#' @param w Weight object.
#' @param sample Sample data.
#' @param target Target object.
#' @param id Sample ID column.
#' @keywords internal
#' @noRd
.wf_influence_margin_share <- function(table, w, sample, target, id) {
  if (is.null(sample) || is.null(target)) {
    table$max_margin_share <- NA_real_
    table$max_margin_dim <- NA_character_
    table$max_margin_category <- NA_character_
    return(table)
  }
  if (!inherits(target, "wf_target")) {
    wf_abort("`target` must be NULL or a wf_target object.", "wf_error_input")
  }
  if (!is.data.frame(sample) || nrow(sample) == 0) {
    wf_abort("`sample` must be NULL or a non-empty data frame.", "wf_error_input")
  }
  if (!.wf_is_string(id)) {
    wf_abort(
      "`id` must be supplied when `sample` and `target` are used.",
      "wf_error_input"
    )
  }
  .require_cols(sample, c(id, target$by, target$dims), "sample")
  order <- .wf_match_unit_ids(
    sample,
    id,
    w$data$id,
    "`w$data$id`",
    "wf_influence()"
  )
  aligned <- sample
  aligned$.wf_weight <- w$data$weight[order]
  aligned$.wf_group <- w$data$group[order]

  max_share <- numeric(nrow(aligned))
  max_dim <- character(nrow(aligned))
  max_cat <- character(nrow(aligned))
  for (i in seq_len(nrow(aligned))) {
    group <- aligned$.wf_group[[i]]
    best_share <- NA_real_
    best_dim <- NA_character_
    best_cat <- NA_character_
    for (dim in target$dims) {
      category <- .chr(aligned[[dim]][[i]])
      target_margin <- target$groups[[group]]$margins[[dim]][[category]]
      if (!is.null(target_margin) && is.finite(target_margin) &&
          target_margin > 0) {
        share <- aligned$.wf_weight[[i]] / target_margin
        if (is.na(best_share) || share > best_share) {
          best_share <- share
          best_dim <- dim
          best_cat <- category
        }
      }
    }
    max_share[[i]] <- best_share
    max_dim[[i]] <- best_dim
    max_cat[[i]] <- best_cat
  }

  sample_key <- .chr(aligned[[id]])
  idx <- match(table$id, sample_key)
  table$max_margin_share <- max_share[idx]
  table$max_margin_dim <- max_dim[idx]
  table$max_margin_category <- max_cat[idx]
  table
}

#' Diagnose high-influence calibrated units
#'
#' Reports the units that contribute most to weight instability. The core
#' diagnostics are per-unit weight ratio to the group mean, share of the
#' group's squared-weight mass, and leave-one-out design effect for the top
#' units. When `sample` and `target` are supplied, WFC also reports the largest
#' share of any target margin represented by that unit.
#'
#' @param w A `wf_weights` object with unit IDs.
#' @param target Optional `wf_target` for margin-share diagnostics.
#' @param sample Optional sample data frame aligned by `id`.
#' @param id Sample ID column required when `sample` and `target` are supplied.
#' @param top Number of highest-ratio units for which leave-one-out design
#'   effects are computed and printed.
#'
#' @return A `wf_influence` object with a `$table` data frame.
#' @export
wf_influence <- function(w, target = NULL, sample = NULL, id = NULL, top = 20) {
  if (!inherits(w, "wf_weights")) {
    wf_abort("`w` must be a wf_weights object.", "wf_error_input")
  }
  if (!is.data.frame(w$data) || nrow(w$data) == 0 ||
      !all(c("id", "group", "weight") %in% names(w$data))) {
    wf_abort(
      "`w$data` must contain id, group, and weight columns.",
      "wf_error_schema"
    )
  }
  top <- .wf_influence_top(top)
  ids <- .chr(w$data$id)
  if (anyNA(ids) || any(!nzchar(ids)) || anyDuplicated(ids)) {
    wf_abort(
      "Influence diagnostics require non-missing unique unit IDs.",
      "wf_error_schema"
    )
  }
  weight <- as.numeric(w$data$weight)
  if (anyNA(weight) || any(!is.finite(weight)) || any(weight <= 0)) {
    wf_abort(
      "Influence diagnostics require finite positive weights.",
      "wf_error_input"
    )
  }

  rows <- lapply(split(seq_len(nrow(w$data)), .chr(w$data$group)), function(idx) {
    part_weight <- weight[idx]
    group_deff <- .wf_influence_deff(part_weight)
    data.frame(
      id = ids[idx],
      group = .chr(w$data$group[idx]),
      weight = part_weight,
      ratio_to_mean = part_weight / mean(part_weight),
      deff_share = part_weight^2 / sum(part_weight^2),
      deff_group = group_deff,
      deff_loo = NA_real_,
      deff_drop_if_removed = NA_real_,
      stringsAsFactors = FALSE
    )
  })
  table <- do.call(rbind, rows)
  table <- table[order(-table$ratio_to_mean, -table$deff_share), , drop = FALSE]
  row.names(table) <- NULL
  table <- .wf_influence_margin_share(table, w, sample, target, id)

  n_top <- min(top, nrow(table))
  for (i in seq_len(n_top)) {
    group <- table$group[[i]]
    group_idx <- which(.chr(w$data$group) == group)
    drop_idx <- which(ids == table$id[[i]])
    loo_idx <- setdiff(group_idx, drop_idx)
    loo_deff <- .wf_influence_deff(weight[loo_idx])
    table$deff_loo[[i]] <- loo_deff
    table$deff_drop_if_removed[[i]] <- table$deff_group[[i]] - loo_deff
  }

  structure(
    list(
      table = table,
      top = n_top,
      provenance = list(
        method = "influence",
        source_method = w$provenance$method,
        top = n_top,
        created = Sys.time(),
        package_version = .wf_package_version()
      )
    ),
    class = "wf_influence"
  )
}

#' Print influence diagnostics
#'
#' @param x A `wf_influence` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_influence <- function(x, ...) {
  cat(sprintf("<wf_influence> top %d high-ratio unit(s)\n", x$top))
  cols <- c(
    "id", "group", "weight", "ratio_to_mean", "deff_share",
    "deff_group", "deff_loo", "deff_drop_if_removed"
  )
  if ("max_margin_share" %in% names(x$table) &&
      any(is.finite(x$table$max_margin_share))) {
    cols <- c(cols, "max_margin_share", "max_margin_dim", "max_margin_category")
  }
  print(utils::head(x$table[cols], x$top), row.names = FALSE, digits = 4)
  invisible(x)
}

#' Calculate the identity of design data and declared roles.
#'
#' @param data Design data frame.
#' @param roles Declared design roles.
#' @keywords internal
#' @noRd
.wf_design_identity <- function(data, roles) {
  .wf_sha256_object(list(data = data, roles = roles))
}

#' Prepare outcome-free survey design data
#'
#' Creates a strict design object containing only identifiers, calibration
#' variables, and declared sampling-design fields. Any undeclared column blocks
#' construction so study outcomes cannot silently enter weight planning.
#'
#' @param data A data frame containing design variables only.
#' @param id Name of the unique, non-missing record identifier column.
#' @param calibration Names of columns permitted for calibration.
#' @param base_weight Optional name of a positive base-weight column.
#' @param strata Optional names of stratification columns.
#' @param clusters Optional names of cluster columns.
#' @param fpc Optional names of finite-population-correction columns.
#'
#' @return A `wf_design_data` object.
#' @examples
#' design_frame <- data.frame(
#'   id = sprintf("r%02d", 1:8),
#'   sex = rep(c("F", "M"), 4),
#'   age = rep(c("18-34", "35+"), each = 4),
#'   base_weight = 1,
#'   stringsAsFactors = FALSE
#' )
#'
#' design <- wf_prepare_design(
#'   design_frame,
#'   id = "id",
#'   calibration = c("sex", "age"),
#'   base_weight = "base_weight"
#' )
#' design
#'
#' # An outcome column has no declared design role, so it blocks construction
#' # instead of silently entering weight planning.
#' with_outcome <- cbind(
#'   design_frame,
#'   satisfaction = seq(40, 70, length.out = 8)
#' )
#' try(wf_prepare_design(
#'   with_outcome,
#'   id = "id",
#'   calibration = c("sex", "age"),
#'   base_weight = "base_weight"
#' ))
#' @export
wf_prepare_design <- function(data, id, calibration, base_weight = NULL,
                              strata = NULL, clusters = NULL, fpc = NULL) {
  roles <- list(
    id = id,
    calibration = calibration,
    base_weight = base_weight,
    strata = strata,
    clusters = clusters,
    fpc = fpc
  )
  assigned <- unique(unlist(roles, use.names = FALSE))

  .require_cols(data, assigned, "design data")

  extra <- setdiff(names(data), assigned)
  if (length(extra)) {
    .wf_safety_abort(
      "design_columns_unassigned",
      sprintf(
        "Design data contain unassigned column(s): %s.",
        paste(extra, collapse = ", ")
      ),
      evidence = list(columns = extra),
      next_actions = "remove_or_assign_columns"
    )
  }

  if (anyNA(data[[id]]) || anyDuplicated(data[[id]])) {
    .wf_safety_abort(
      "design_id_invalid",
      "Design IDs must be unique and non-missing.",
      id
    )
  }

  if (!is.null(base_weight) &&
      (any(!is.finite(data[[base_weight]])) ||
       any(data[[base_weight]] <= 0))) {
    .wf_safety_abort(
      "base_weight_invalid",
      "Base weights must be finite and positive.",
      base_weight
    )
  }

  structure(
    list(
      data = data,
      roles = roles,
      identity = .wf_design_identity(data, roles),
      created = .wf_iso_time(),
      package_version = .wf_package_version()
    ),
    class = "wf_design_data"
  )
}

#' Print outcome-free survey design data
#'
#' @param x A `wf_design_data` object.
#' @param ... Reserved for future use.
#'
#' @return `x`, invisibly.
#' @export
print.wf_design_data <- function(x, ...) {
  cat("<wf_design_data>\n")
  cat(" Rows: ", nrow(x$data), "\n", sep = "")
  cat(" Calibration: ", paste(x$roles$calibration, collapse = ", "), "\n", sep = "")
  cat(" Identity: ", x$identity, "\n", sep = "")
  invisible(x)
}

test_that("every design column has an explicit role", {
  d <- data.frame(
    id = c("a", "b"),
    sex = c("F", "M"),
    base = c(1, 2)
  )

  x <- wf_prepare_design(d, "id", "sex", base_weight = "base")

  expect_s3_class(x, "wf_design_data")
  expect_identical(x$roles$calibration, "sex")
  expect_match(x$identity, "^[0-9a-f]{64}$")
})

test_that("unassigned outcome-like columns block", {
  d <- data.frame(
    id = 1:2,
    sex = c("F", "M"),
    outcome = c(1, 0)
  )

  expect_error(
    wf_prepare_design(d, "id", "sex"),
    class = "wf_error_safety"
  )
})

# Pin the default report language so the suite does not depend on the
# session locale of the machine running it. Tests that assert localized
# output pass `lang` explicitly, and the locale-resolution tests in
# test-i18n.R override this option themselves.
options(wfc.lang = "en")

#' Return the loaded package version for provenance.
#'
#' @keywords internal
#' @noRd
.wf_lincal_package_version <- function() {
  tryCatch(
    as.character(utils::packageVersion("WFC")),
    error = function(e) "0.8.0"
  )
}

#' Build a calibration distance object (weight-generating function and slope).
#'
#' @param distance "linear" (GREG), "logit" (bounded), or "entropy".
#' @param bounds Two-element `c(L, U)` for logit; ignored for linear.
#' @keywords internal
#' @noRd
.wf_lincal_dist <- function(distance, bounds) {
  if (distance == "linear") {
    return(list(
      F = function(u) 1 + u,
      Fp = function(u) rep(1, length(u))
    ))
  }
  if (distance == "entropy") {
    return(list(
      F = function(u) exp(pmax(pmin(u, 700), -700)),
      Fp = function(u) exp(pmax(pmin(u, 700), -700))
    ))
  }
  L <- bounds[1]
  U <- bounds[2]
  A <- (U - L) / ((1 - L) * (U - 1))
  # Overflow-safe: for A*u >= 0 divide through by e = exp(A*u) so exp only ever
  # sees non-positive arguments (avoids Inf/Inf -> NaN when u overshoots).
  list(
    F = function(u) {
      s <- A * u
      out <- numeric(length(s))
      pos <- s >= 0
      em <- exp(-s[pos])
      out[pos] <- (L * (U - 1) * em + U * (1 - L)) / ((U - 1) * em + (1 - L))
      e <- exp(s[!pos])
      out[!pos] <- (L * (U - 1) + U * (1 - L) * e) / ((U - 1) + (1 - L) * e)
      out
    },
    Fp = function(u) {
      s <- A * u
      out <- numeric(length(s))
      pos <- s >= 0
      em <- exp(-s[pos])
      out[pos] <- (U - L)^2 * em / ((U - 1) * em + (1 - L))^2
      e <- exp(s[!pos])
      out[!pos] <- (U - L)^2 * e / ((U - 1) + (1 - L) * e)^2
      out
    }
  )
}

#' Build the calibration constraint matrix and target vector for one group.
#'
#' @param sub Group's (complete-case) sample subset.
#' @param dvars Calibration dimension names.
#' @param gr A target group: `list(total, margins)`.
#' @keywords internal
#' @noRd
.wf_lincal_build <- function(sub, dvars, gr) {
  n <- nrow(sub)
  cols <- list(rep(1, n))
  t <- gr$total
  labels <- data.frame(
    dim = "_total_",
    category = NA_character_,
    target = gr$total,
    stringsAsFactors = FALSE
  )
  for (d in dvars) {
    lev <- names(gr$margins[[d]])
    for (l in lev[-1]) {   # drop the first level as the reference
      cols[[length(cols) + 1]] <- as.numeric(.chr(sub[[d]]) == l)
      t <- c(t, gr$margins[[d]][[l]])
      labels <- rbind(
        labels,
        data.frame(
          dim = d,
          category = l,
          target = gr$margins[[d]][[l]],
          stringsAsFactors = FALSE
        )
      )
    }
  }
  list(X = do.call(cbind, cols), t = t, labels = labels)
}

#' Build full margin constraints for soft calibration.
#'
#' @param sub Group sample subset.
#' @param dvars Calibration dimension names.
#' @param gr A target group.
#' @keywords internal
#' @noRd
.wf_soft_build <- function(sub, dvars, gr) {
  n <- nrow(sub)
  cols <- list(rep(1, n))
  t <- gr$total
  labels <- data.frame(
    dim = "_total_",
    category = NA_character_,
    target = gr$total,
    stringsAsFactors = FALSE
  )
  for (d in dvars) {
    lev <- names(gr$margins[[d]])
    for (l in lev) {
      cols[[length(cols) + 1]] <- as.numeric(.chr(sub[[d]]) == l)
      t <- c(t, gr$margins[[d]][[l]])
      labels <- rbind(
        labels,
        data.frame(
          dim = d,
          category = l,
          target = gr$margins[[d]][[l]],
          stringsAsFactors = FALSE
        )
      )
    }
  }
  list(X = do.call(cbind, cols), t = t, labels = labels)
}

#' Resolve soft-calibration tolerances by dimension.
#'
#' @param tolerance Scalar or named vector.
#' @param dvars Calibration dimension names.
#' @keywords internal
#' @noRd
.wf_soft_tolerance <- function(tolerance, dvars) {
  if (!is.numeric(tolerance) || length(tolerance) == 0 ||
      anyNA(tolerance) || any(!is.finite(tolerance)) ||
      any(tolerance <= 0)) {
    wf_abort(
      "`tolerance` must be a positive finite number or a named positive vector by dimension.",
      "wf_error_input",
      list(tolerance = tolerance)
    )
  }
  if (length(tolerance) == 1 && is.null(names(tolerance))) {
    return(stats::setNames(rep(tolerance, length(dvars)), dvars))
  }
  if (is.null(names(tolerance)) || any(!nzchar(names(tolerance)))) {
    wf_abort(
      "Named `tolerance` values must name target dimensions.",
      "wf_error_input",
      list(tolerance = tolerance)
    )
  }
  missing <- setdiff(dvars, names(tolerance))
  extra <- setdiff(names(tolerance), dvars)
  if (length(missing) > 0 || length(extra) > 0) {
    wf_abort(
      sprintf(
        "`tolerance` must name every target dimension. Missing: %s; extra: %s.",
        paste(missing, collapse = ", "),
        paste(extra, collapse = ", ")
      ),
      "wf_error_schema",
      list(missing = missing, extra = extra)
    )
  }
  tolerance[dvars]
}

#' Build absolute tolerances for one soft-calibration group.
#'
#' @param labels Constraint labels.
#' @param tolerance_by_dim Named relative tolerances.
#' @param total Group total.
#' @keywords internal
#' @noRd
.wf_soft_constraint_tolerance <- function(labels, tolerance_by_dim, total) {
  out <- numeric(nrow(labels))
  for (i in seq_len(nrow(labels))) {
    if (labels$dim[[i]] == "_total_") {
      out[[i]] <- 0
    } else {
      out[[i]] <- tolerance_by_dim[[labels$dim[[i]]]] * total
    }
  }
  out
}

#' Solve one soft-calibration group.
#'
#' @param X Full constraint matrix.
#' @param d Base weights.
#' @param t Constraint target totals.
#' @param tolerance_abs Absolute per-constraint tolerances.
#' @param max_outer Maximum penalty-shrink rounds.
#' @param total Group total.
#' @param g Group label.
#' @keywords internal
#' @noRd
.wf_soft_group <- function(X, d, t, tolerance_abs, max_outer, total, g) {
  if (length(d) == 0 || any(!is.finite(d)) || any(d < 0) || sum(d) <= 0) {
    wf_abort(
      sprintf("Group '%s': base weights must be non-negative with positive total.", g),
      "wf_error_input",
      list(group = g)
    )
  }
  p <- ncol(X)
  phi <- ifelse(
    tolerance_abs > 0,
    pmax((tolerance_abs)^2, (.Machine$double.eps * total)^2),
    0
  )
  XtDX <- crossprod(X, X * d)
  residual0 <- t - as.numeric(crossprod(X, d))
  w <- d
  gap <- residual0
  iterations <- 0L

  for (outer in seq_len(max_outer)) {
    iterations <- outer
    system <- XtDX + diag(phi, p)
    step <- tryCatch(solve(system, residual0), error = function(e) NULL)
    if (is.null(step)) {
      wf_abort(
        sprintf("Group '%s': singular soft-calibration system.", g),
        "wf_error_feasibility",
        list(group = g)
      )
    }
    w <- d + d * as.numeric(X %*% step)
    gap <- t - as.numeric(crossprod(X, w))
    over <- abs(gap) > tolerance_abs + 1e-10 * max(total, 1) &
      tolerance_abs > 0
    if (!any(over)) {
      break
    }
    phi[over] <- phi[over] *
      pmax(0.05, pmin(0.5, tolerance_abs[over] / abs(gap[over])))
  }

  converged <- all(abs(gap) <= tolerance_abs + 1e-8 * max(total, 1))
  if (!converged) {
    wf_abort(
      sprintf(
        "Group '%s': soft calibration could not keep all relaxed margins within the declared tolerance.",
        g
      ),
      "wf_error_feasibility",
      list(group = g, max_gap = max(abs(gap) / total))
    )
  }
  if (any(!is.finite(w)) || any(w <= 0)) {
    wf_abort(
      sprintf(
        "Group '%s': soft calibration produced non-positive weights. Increase `tolerance` or use bounded logit calibration.",
        g
      ),
      "wf_error_feasibility",
      list(group = g)
    )
  }
  list(
    w = w,
    iterations = iterations,
    converged = converged,
    gap = gap,
    max_rel_gap = max(abs(gap)) / total,
    ratio = w / d
  )
}

#' Solve the calibration equations for one group by Newton iteration.
#'
#' @param X Constraint matrix (n x p).
#' @param d Base weights (length n).
#' @param t Target totals (length p; `t[1]` is the group total).
#' @param dist A `.wf_lincal_dist()` object.
#' @param tol Convergence tolerance on the max residual relative to `total`.
#' @param max_iter Iteration cap.
#' @param total Group total (for the relative residual).
#' @param g Group label (for error messages).
#' @keywords internal
#' @noRd
.wf_lincal_group <- function(X, d, t, dist, tol, max_iter, total, g) {
  lambda <- rep(0, ncol(X))
  u <- as.numeric(X %*% lambda)
  w <- d * dist$F(u)
  steps <- 0L
  converged <- FALSE
  maxr <- NA_real_
  repeat {
    resid <- t - as.numeric(t(X) %*% w)
    maxr <- max(abs(resid)) / total
    if (is.finite(maxr) && maxr < tol) {
      converged <- TRUE
      break
    }
    if (steps >= max_iter) break
    jac <- t(X) %*% (X * (d * dist$Fp(u)))
    step <- tryCatch(solve(jac, resid), error = function(e) NULL)
    if (is.null(step)) {
      wf_abort(
        sprintf("Group '%s': singular calibration system (empty category or collinear margins).", g),
        "wf_error_feasibility", list(group = g)
      )
    }
    # Damped Newton: shrink the step until weights stay finite and the residual
    # decreases (keeps the logit distance from overflowing on an overshoot).
    alpha <- 1
    repeat {
      lambda_try <- lambda + alpha * step
      u_try <- as.numeric(X %*% lambda_try)
      w_try <- d * dist$F(u_try)
      maxr_try <- max(abs(t - as.numeric(t(X) %*% w_try))) / total
      if (all(is.finite(w_try)) && is.finite(maxr_try) && maxr_try < maxr) break
      alpha <- alpha / 2
      if (alpha < 1e-10) break
    }
    lambda <- lambda + alpha * step
    steps <- steps + 1L
    u <- as.numeric(X %*% lambda)
    w <- d * dist$F(u)
  }
  if (!converged) {
    wf_abort(
      sprintf("Group '%s': calibration did not converge in %d iterations (max relative residual %.3g). Bounds may be too tight to meet the margins.",
              g, max_iter, maxr),
      "wf_error_feasibility", list(group = g, residual = maxr)
    )
  }
  list(w = w, iterations = steps, converged = TRUE,
       max_resid = maxr, ratio = dist$F(u))
}

#' Calibrate weights by a linear or logit distance (engine behind wf_calibrate).
#'
#' @param sample Sample data frame.
#' @param target A `wf_target` object.
#' @param distance "linear" or "logit".
#' @param method Reported method label ("greg" or "logit").
#' @param bounds `c(L, U)` for logit.
#' @param init_weight Optional base-weight column.
#' @param na "drop" or "error".
#' @param id Optional id column.
#' @param tol Convergence tolerance.
#' @param max_iter Iteration cap.
#' @param precheck Run `wf_precheck()` first.
#' @keywords internal
#' @noRd
.wf_lincalibrate <- function(sample, target, distance, method,
                             bounds = NULL, init_weight = NULL,
                             na = c("drop", "error"), id = NULL,
                             tol = 1e-8, max_iter = 100, precheck = TRUE) {
  na <- match.arg(na)
  t0 <- Sys.time()

  if (precheck) {
    pc <- wf_precheck(sample, target, id = id, na = "drop")
    if (!pc$ok) {
      wf_abort(sprintf(
        "Precheck reports %d blocking issue(s). Inspect wf_precheck(sample, target) before calibrating.",
        sum(pc$issues$severity == "error")
      ), "wf_error_feasibility", list(precheck = pc))
    }
  }

  dvars <- target$dims
  for (d in dvars) {
    if (!d %in% names(sample)) {
      wf_abort(sprintf("Calibration dimension '%s' not found in sample.", d),
               "wf_error_schema", list(dim = d))
    }
  }
  na_vars <- dvars
  na_table <- sapply(na_vars, function(d) is.na(sample[[d]]))
  if (is.null(dim(na_table))) {
    na_table <- matrix(na_table, nrow = nrow(sample))
  }
  na_mask <- rowSums(na_table) > 0
  if (any(na_mask)) {
    if (na == "error") {
      wf_abort(sprintf("%d row(s) have NA in calibration dimensions.", sum(na_mask)),
               "wf_error_schema", list(n = sum(na_mask)))
    }
    wf_warn(sprintf("na='drop': removed %d row(s) with NA in calibration dimensions.",
                    sum(na_mask)), "wf_warning_data")
    sample <- sample[!na_mask, , drop = FALSE]
  }

  if (is.null(init_weight)) {
    iw <- NULL   # per-group default (total/n) computed in the loop
  } else {
    if (length(init_weight) != 1 || !is.character(init_weight) ||
        !init_weight %in% names(sample)) {
      wf_abort(sprintf("init_weight column '%s' not found in sample.",
                       as.character(init_weight)[1]),
               "wf_error_schema", list(init_weight = init_weight))
    }
    iw <- as.numeric(sample[[init_weight]])
    if (any(!is.finite(iw)) || any(iw < 0)) {
      wf_abort("init_weight must be non-negative and finite.",
               "wf_error_input", list(init_weight = init_weight))
    }
  }

  dist <- .wf_lincal_dist(distance, bounds)
  gkey <- .wf_group_keys(sample, target$by)
  ids <- if (is.null(id)) seq_len(nrow(sample)) else sample[[id]]

  res_rows <- list()
  logs <- list()
  achieved <- list()
  for (g in intersect(names(target$groups), unique(gkey))) {
    sel <- which(gkey == g)
    gr <- target$groups[[g]]
    sub <- sample[sel, , drop = FALSE]
    built <- .wf_lincal_build(sub, dvars, gr)
    d_g <- if (is.null(iw)) rep(gr$total / length(sel), length(sel)) else iw[sel]
    fit <- .wf_lincal_group(built$X, d_g, built$t, dist,
                            tol, max_iter, gr$total, g)

    res_rows[[g]] <- data.frame(
      id = .chr(ids[sel]),
      group = g,
      weight = fit$w,
      feature = 1 / fit$w,
      stringsAsFactors = FALSE
    )
    logs[[g]] <- data.frame(
      group = g, n = length(sel), iterations = fit$iterations,
      converged = fit$converged, max_resid = fit$max_resid,
      ratio_min = min(fit$ratio), ratio_max = max(fit$ratio),
      kl_divergence = if (distance == "entropy") {
        sum(fit$w * log(fit$w / d_g)) / sum(fit$w)
      } else {
        NA_real_
      },
      stringsAsFactors = FALSE
    )
    achieved[[g]] <- lapply(dvars, function(d) {
      levs <- names(gr$margins[[d]])
      stats::setNames(
        vapply(levs, function(l) sum(fit$w[.chr(sub[[d]]) == l]), numeric(1)),
        levs
      )
    })
    names(achieved[[g]]) <- dvars
  }

  out <- structure(list(
    data = do.call(rbind, res_rows),
    log = do.call(rbind, logs),
    achieved = achieved,
    provenance = list(
      method = method,
      distance = distance,
      bounds = bounds,
      init_weight = init_weight,
      na = na,
      dims = dvars,
      by = target$by,
      tol = tol,
      max_iter = max_iter,
      created = t0,
      elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = .wf_lincal_package_version()
    )
  ), class = "wf_weights")
  out
}

#' Calibrate weights with soft bounded margin relaxation.
#'
#' @param sample Sample data frame.
#' @param target A `wf_target` object.
#' @param tolerance Positive scalar or named vector by dimension. Values are
#'   interpreted as fractions of each group total.
#' @param init_weight Optional base-weight column.
#' @param na "drop" or "error".
#' @param id Optional id column.
#' @param max_outer Maximum penalty adjustment rounds.
#' @param precheck Run a soft-aware precheck first.
#' @keywords internal
#' @noRd
.wf_softcalibrate <- function(sample, target, tolerance = 0.02,
                              init_weight = NULL,
                              na = c("drop", "error"), id = NULL,
                              max_outer = 25, precheck = TRUE) {
  na <- match.arg(na)
  t0 <- Sys.time()
  if (!inherits(target, "wf_target")) {
    wf_abort("`target` must be a wf_target object.", "wf_error_input")
  }
  if (!is.data.frame(sample) || nrow(sample) == 0) {
    wf_abort("`sample` must be a non-empty data frame.", "wf_error_input")
  }
  if (length(max_outer) != 1 || !is.numeric(max_outer) ||
      !is.finite(max_outer) || max_outer < 1 ||
      max_outer != as.integer(max_outer)) {
    wf_abort("`max_outer` must be a positive integer.", "wf_error_input")
  }

  dvars <- target$dims
  tolerance_by_dim <- .wf_soft_tolerance(tolerance, dvars)
  .require_cols(sample, c(id, target$by, dvars), "sample")

  if (precheck) {
    pc <- wf_precheck(sample, target, id = id, na = na)
    blocking <- pc$issues[
      pc$issues$severity == "error" & pc$issues$check != "cat_infeasible",
      ,
      drop = FALSE
    ]
    if (nrow(blocking) > 0) {
      wf_abort(
        sprintf(
          "Soft calibration precheck found %d non-relaxable blocking issue(s).",
          nrow(blocking)
        ),
        "wf_error_feasibility",
        list(precheck = pc, blocking = blocking)
      )
    }
  }

  na_table <- sapply(dvars, function(d) is.na(sample[[d]]))
  if (is.null(dim(na_table))) {
    na_table <- matrix(na_table, nrow = nrow(sample))
  }
  na_mask <- rowSums(na_table) > 0
  if (any(na_mask)) {
    if (na == "error") {
      wf_abort(
        sprintf("%d row(s) have NA in calibration dimensions.", sum(na_mask)),
        "wf_error_schema",
        list(n = sum(na_mask))
      )
    }
    wf_warn(
      sprintf(
        "na='drop': removed %d row(s) with NA in calibration dimensions.",
        sum(na_mask)
      ),
      "wf_warning_data"
    )
    sample <- sample[!na_mask, , drop = FALSE]
  }

  if (is.null(init_weight)) {
    iw <- NULL
  } else {
    if (length(init_weight) != 1 || !is.character(init_weight) ||
        !init_weight %in% names(sample)) {
      wf_abort(
        sprintf("init_weight column '%s' not found in sample.",
                as.character(init_weight)[1]),
        "wf_error_schema",
        list(init_weight = init_weight)
      )
    }
    iw <- as.numeric(sample[[init_weight]])
    if (any(!is.finite(iw)) || any(iw < 0)) {
      wf_abort(
        "init_weight must be non-negative and finite.",
        "wf_error_input",
        list(init_weight = init_weight)
      )
    }
  }

  gkey <- .wf_group_keys(sample, target$by)
  ids <- if (is.null(id)) seq_len(nrow(sample)) else sample[[id]]
  res_rows <- list()
  logs <- list()
  achieved <- list()
  relaxation <- list()

  for (g in intersect(names(target$groups), unique(gkey))) {
    sel <- which(gkey == g)
    gr <- target$groups[[g]]
    sub <- sample[sel, , drop = FALSE]
    built <- .wf_soft_build(sub, dvars, gr)
    d_g <- if (is.null(iw)) rep(gr$total / length(sel), length(sel)) else iw[sel]
    tolerance_abs <- .wf_soft_constraint_tolerance(
      built$labels,
      tolerance_by_dim,
      gr$total
    )
    fit <- .wf_soft_group(
      built$X,
      d_g,
      built$t,
      tolerance_abs,
      as.integer(max_outer),
      gr$total,
      g
    )

    res_rows[[g]] <- data.frame(
      id = .chr(ids[sel]),
      group = g,
      weight = fit$w,
      feature = 1 / fit$w,
      stringsAsFactors = FALSE
    )
    logs[[g]] <- data.frame(
      group = g,
      n = length(sel),
      iterations = fit$iterations,
      converged = fit$converged,
      max_rel_gap = fit$max_rel_gap,
      n_relaxed = sum(abs(fit$gap) > 1e-9 & tolerance_abs > 0),
      ratio_min = min(fit$ratio),
      ratio_max = max(fit$ratio),
      stringsAsFactors = FALSE
    )
    achieved[[g]] <- lapply(dvars, function(d) {
      levs <- names(gr$margins[[d]])
      stats::setNames(
        vapply(levs, function(l) sum(fit$w[.chr(sub[[d]]) == l]), numeric(1)),
        levs
      )
    })
    names(achieved[[g]]) <- dvars
    achieved_constraint <- as.numeric(crossprod(built$X, fit$w))
    relaxation[[g]] <- data.frame(
      group = g,
      dim = built$labels$dim,
      category = built$labels$category,
      target = built$t,
      achieved = achieved_constraint,
      gap = built$t - achieved_constraint,
      tolerance = tolerance_abs,
      relative_gap = abs(built$t - achieved_constraint) / gr$total,
      relaxed = abs(built$t - achieved_constraint) > 1e-9 & tolerance_abs > 0,
      within_tolerance = abs(built$t - achieved_constraint) <= tolerance_abs +
        1e-8 * max(gr$total, 1),
      stringsAsFactors = FALSE
    )
  }

  structure(list(
    data = do.call(rbind, res_rows),
    log = do.call(rbind, logs),
    achieved = achieved,
    relaxation = do.call(rbind, relaxation),
    provenance = list(
      method = "soft",
      distance = "soft_linear",
      tolerance = tolerance_by_dim,
      init_weight = init_weight,
      na = na,
      dims = dvars,
      by = target$by,
      max_outer = as.integer(max_outer),
      created = t0,
      elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = .wf_lincal_package_version()
    )
  ), class = "wf_weights")
}

test_that(".wf_lincal_dist linear gives F(u)=1+u and F'(u)=1", {
  dist <- .wf_lincal_dist("linear", NULL)
  u <- c(-0.5, 0, 0.5, 2)
  expect_equal(dist$F(u), 1 + u)
  expect_equal(dist$Fp(u), rep(1, length(u)))
})

test_that(".wf_lincal_dist logit maps to (L,U) with F(0)=1 and unit slope at 0", {
  L <- 0.3; U <- 3
  dist <- .wf_lincal_dist("logit", c(L, U))
  u <- seq(-10, 10, by = 0.5)
  fu <- dist$F(u)
  expect_true(all(fu > L & fu < U))
  expect_equal(dist$F(0), 1)
  expect_equal(dist$Fp(0), 1)          # slope at 0 matches the linear distance
})

test_that(".wf_lincal_dist logit F' matches a numeric derivative", {
  dist <- .wf_lincal_dist("logit", c(0.5, 4))
  u0 <- 0.7
  numeric <- (dist$F(u0 + 1e-6) - dist$F(u0 - 1e-6)) / (2e-6)
  expect_equal(dist$Fp(u0), numeric, tolerance = 1e-5)
})

test_that(".wf_lincal_build makes an intercept + dropped-reference-level matrix", {
  sub <- data.frame(g = c("a", "a", "b", "b"), stringsAsFactors = FALSE)
  gr <- list(total = 4, margins = list(g = c(a = 3, b = 1)))
  built <- .wf_lincal_build(sub, dvars = "g", gr = gr)

  # intercept column + one column for the retained level "b" (a is the ref)
  expect_equal(ncol(built$X), 2)
  expect_equal(built$X[, 1], rep(1, 4))
  expect_equal(built$X[, 2], c(0, 0, 1, 1))
  expect_equal(built$t, c(4, 1))          # total, then b's margin
})

test_that(".wf_lincal_build stacks multiple dims, dropping one level each", {
  sub <- data.frame(
    g = c("a", "b", "a", "b"),
    h = c("x", "x", "y", "y"),
    stringsAsFactors = FALSE
  )
  gr <- list(total = 4,
             margins = list(g = c(a = 2, b = 2), h = c(x = 2, y = 2)))
  built <- .wf_lincal_build(sub, dvars = c("g", "h"), gr = gr)
  # intercept + (g: drop a, keep b) + (h: drop x, keep y) = 3 columns
  expect_equal(ncol(built$X), 3)
  expect_equal(built$t, c(4, 2, 2))
})

test_that(".wf_lincal_group solves the GREG closed form by hand", {
  # sample a,a,b,b; base 1; total 4; margin a=3,b=1 -> weights 1.5,1.5,0.5,0.5
  X <- cbind(rep(1, 4), c(0, 0, 1, 1))
  t <- c(4, 1)
  d <- rep(1, 4)
  dist <- .wf_lincal_dist("linear", NULL)
  out <- .wf_lincal_group(X, d, t, dist, tol = 1e-10, max_iter = 100,
                          total = 4, g = "_all_")

  expect_true(out$converged)
  expect_equal(out$w, c(1.5, 1.5, 0.5, 0.5))
  expect_equal(out$iterations, 1)          # linear converges in one step
})

test_that(".wf_lincal_group logit keeps ratios within bounds and hits the target", {
  X <- cbind(rep(1, 4), c(0, 0, 1, 1))
  t <- c(4, 1)
  d <- rep(1, 4)
  dist <- .wf_lincal_dist("logit", c(0.3, 3))
  out <- .wf_lincal_group(X, d, t, dist, tol = 1e-10, max_iter = 100,
                          total = 4, g = "_all_")

  expect_true(out$converged)
  expect_true(all(out$ratio > 0.3 & out$ratio < 3))
  # margins: intercept and category-b constraint both met
  expect_equal(sum(out$w), 4, tolerance = 1e-8)
  expect_equal(sum(out$w[3:4]), 1, tolerance = 1e-8)
})

test_that(".wf_lincal_group aborts when bounds are infeasible", {
  # b units need mean 0.5 but a floor of 0.6 makes sum >= 1.2 > 1: infeasible
  X <- cbind(rep(1, 4), c(0, 0, 1, 1))
  t <- c(4, 1)
  d <- rep(1, 4)
  dist <- .wf_lincal_dist("logit", c(0.6, 3))
  expect_error(
    .wf_lincal_group(X, d, t, dist, tol = 1e-10, max_iter = 50,
                     total = 4, g = "_all_"),
    class = "wf_error_feasibility"
  )
})

test_that(".wf_lincalibrate returns a wf_weights that hits every margin", {
  fixture <- make_weightflow_fixture()
  w <- .wf_lincalibrate(fixture$sample, fixture$target, distance = "linear",
                        method = "greg", id = "id")

  expect_s3_class(w, "wf_weights")
  expect_named(w$data, c("id", "group", "weight", "feature"))
  expect_equal(w$provenance$method, "greg")

  # province A female margin is reproduced
  s <- fixture$sample
  a_female <- sum(w$data$weight[s$province == "A" & s$gender == "female"])
  target_af <- fixture$target$groups[["A"]]$margins$gender[["female"]]
  expect_equal(a_female, target_af, tolerance = 1e-6)
})

test_that(".wf_lincalibrate respects init_weight", {
  fixture <- make_weightflow_fixture()
  s <- fixture$sample
  s$bw <- ifelse(s$gender == "female" & s$age == "young", 3, 1)

  uniform <- .wf_lincalibrate(s, fixture$target, distance = "linear",
                              method = "greg", id = "id")
  weighted <- .wf_lincalibrate(s, fixture$target, distance = "linear",
                               method = "greg", id = "id", init_weight = "bw")

  expect_false(isTRUE(all.equal(uniform$data$weight, weighted$data$weight)))
  # margins still reproduced under a non-uniform base
  a_female <- sum(weighted$data$weight[s$province == "A" & s$gender == "female"])
  expect_equal(a_female, fixture$target$groups[["A"]]$margins$gender[["female"]],
               tolerance = 1e-6)
})

test_that(".wf_lincalibrate drops NA rows with a warning and errors on demand", {
  fixture <- make_weightflow_fixture()
  s <- fixture$sample
  s$gender[1] <- NA

  expect_warning(
    .wf_lincalibrate(s, fixture$target, distance = "linear", method = "greg",
                     id = "id", na = "drop", precheck = FALSE),
    class = "wf_warning_data"
  )
  expect_error(
    .wf_lincalibrate(s, fixture$target, distance = "linear", method = "greg",
                     id = "id", na = "error", precheck = FALSE),
    class = "wf_error_schema"
  )
})

test_that("wf_calibrate routes greg and logit to the calibration engine", {
  fixture <- make_weightflow_fixture()
  greg <- .wf_calibrate_engine(fixture$sample, fixture$target, method = "greg",
                       id = "id")
  logit <- .wf_calibrate_engine(fixture$sample, fixture$target, method = "logit",
                        bounds = c(0.3, 3), id = "id")

  expect_equal(greg$provenance$method, "greg")
  expect_equal(logit$provenance$method, "logit")
  # logit ratios stay within bounds
  expect_true(all(logit$log$ratio_min >= 0.3 - 1e-9))
  expect_true(all(logit$log$ratio_max <= 3 + 1e-9))
})

test_that("wf_calibrate requires valid bounds for logit", {
  fixture <- make_weightflow_fixture()
  expect_error(
    .wf_calibrate_engine(fixture$sample, fixture$target, method = "logit", id = "id"),
    class = "wf_error_input"
  )
  expect_error(
    .wf_calibrate_engine(fixture$sample, fixture$target, method = "logit",
                 bounds = c(2, 0.5), id = "id"),
    class = "wf_error_input"
  )
})

test_that("wf_calibrate logit with loose bounds approximates raking", {
  fixture <- make_weightflow_fixture()
  raked <- .wf_rake_engine(fixture$sample, fixture$target, id = "id")
  logit <- .wf_calibrate_engine(fixture$sample, fixture$target, method = "logit",
                        bounds = c(1e-6, 1e6), id = "id")
  m <- match(raked$data$id, logit$data$id)
  expect_equal(logit$data$weight[m], raked$data$weight, tolerance = 1e-4)
})

test_that("wf_calibrate still routes raking and poststrat unchanged", {
  fixture <- make_weightflow_fixture()
  raked <- .wf_calibrate_engine(fixture$sample, fixture$target, method = "raking",
                        id = "id")
  expect_equal(raked$provenance$method, "raking")
})

test_that("print.wf_weights reports the calibration distance and bounds", {
  fixture <- make_weightflow_fixture()
  logit <- .wf_calibrate_engine(fixture$sample, fixture$target, method = "logit",
                        bounds = c(0.3, 3), id = "id")
  expect_output(print(logit), "method: logit")
  expect_output(print(logit), "bounds")
})

test_that("logit calibration composes and serves as a replicates refit", {
  fixture <- make_weightflow_fixture()
  logit <- .wf_calibrate_engine(fixture$sample, fixture$target, method = "logit",
                        bounds = c(0.2, 5), id = "id")

  # composes with a second stage
  stage2 <- logit
  stage2$data$weight <- rep(2, nrow(stage2$data))
  stage2$data$feature <- 1 / stage2$data$weight
  composed <- wf_compose(cal = logit, adj = stage2)
  expect_s3_class(composed, "wf_weights")

  # serves as a wf_replicates refit. Bounded calibration is on the w/d ratio, so
  # the replicate base weights must be on the population scale: carry a design
  # weight (total/n = 200/8 = 25 in this fixture) via base_weight.
  d <- fixture$sample
  d$y <- as.numeric(d$age == "young")
  d$dw <- 25
  refit <- function(data, weights) {
    data$.bw <- weights
    .wf_calibrate_engine(data, fixture$target, method = "logit", bounds = c(0.2, 5),
                 init_weight = ".bw", id = "id", precheck = FALSE)
  }
  reps <- wf_replicates(d, refit, method = "jackknife", id = "id",
                        base_weight = "dw")
  out <- wf_variance(reps, function(w, data) sum(w * data$y) / sum(w), d)
  expect_true(is.finite(out$table$se))
})

set.seed(20260715)

if (!requireNamespace("pkgload", quietly = TRUE)) {
  stop("Validation requires the 'pkgload' package.")
}
pkgload::load_all(".", quiet = TRUE)

results <- data.frame(
  scenario = character(),
  check = character(),
  passed = logical(),
  detail = character(),
  stringsAsFactors = FALSE
)

record_check <- function(scenario, check, passed, detail = "") {
  results <<- rbind(
    results,
    data.frame(
      scenario = scenario,
      check = check,
      passed = isTRUE(passed),
      detail = as.character(detail),
      stringsAsFactors = FALSE
    )
  )
  invisible(passed)
}

write_verified_population <- function(population, label) {
  directory <- tempfile(paste0("wfc-validation-", label, "-"))
  dir.create(directory)
  data_file <- file.path(directory, "population.csv")
  source_file <- paste0(data_file, ".source.dcf")
  utils::write.csv(population, data_file, row.names = FALSE)
  source <- c(
    publisher = "WFC controlled validation fixture",
    dataset_title = paste("Synthetic validation population", label),
    citation = "Synthetic validation fixture; not a production authority",
    reference_period = "2026-validation",
    population_scope = "Synthetic validation population",
    retrieved_at = "2026-07-15",
    license = "CC0-1.0",
    checksum_algorithm = "sha256",
    checksum = WFC:::.wf_sha256_file(data_file),
    transformation = "Deterministic synthetic validation construction",
    selected_before_outcomes = "true",
    demo_only = "false"
  )
  writeLines(sprintf("%s: %s", names(source), source), source_file)
  list(data_file = data_file, source_file = source_file)
}

make_rows <- function(counts) {
  cells <- data.frame(
    sex = c("F", "M", "F", "M"),
    age = c("18-34", "18-34", "35+", "35+"),
    n = counts,
    stringsAsFactors = FALSE
  )
  rows <- cells[rep(seq_len(nrow(cells)), cells$n), c("sex", "age")]
  row.names(rows) <- NULL
  rows$id <- sprintf("r%03d", seq_len(nrow(rows)))
  rows$base_weight <- 1
  rows[c("id", "sex", "age", "base_weight")]
}

dims <- wf_dims(sex = c("F", "M"), age = c("18-34", "35+"))
population <- data.frame(
  sex = c("F", "M", "F", "M"),
  age = c("18-34", "18-34", "35+", "35+"),
  count = rep(10, 4),
  stringsAsFactors = FALSE
)
source <- write_verified_population(population, "balanced")
target <- wf_import_target(
  source$data_file,
  source$source_file,
  dims,
  key_map = c(sex = "sex", age = "age"),
  count = "count"
)

# Scenario 1: supported undercoverage and deterministic locked weights.
undercoverage_data <- make_rows(c(14, 10, 9, 7))
undercoverage_design <- wf_prepare_design(
  undercoverage_data,
  id = "id",
  calibration = c("sex", "age"),
  base_weight = "base_weight"
)
cells_a <- wf_plan_cells(
  undercoverage_design,
  target,
  dims,
  min_cell = 5,
  max_weight_ratio = 4
)
cells_b <- wf_plan_cells(
  undercoverage_design,
  target,
  dims,
  min_cell = 5,
  max_weight_ratio = 4
)
plan_a <- wf_plan_weights(
  undercoverage_design,
  target,
  dims,
  bounds = c(0.3, 3),
  min_cell = 5,
  cell_plan = cells_a
)
plan_b <- wf_plan_weights(
  undercoverage_design,
  target,
  dims,
  bounds = c(0.3, 3),
  min_cell = 5,
  cell_plan = cells_b
)
approval_a <- wf_approve_plan(
  plan_a,
  "Synthetic validation fixture",
  "test-only reviewer"
)
approval_b <- wf_approve_plan(
  plan_b,
  "Synthetic validation fixture",
  "test-only reviewer"
)
weights_a <- wf_execute_plan(
  plan_a,
  approval_a,
  undercoverage_design,
  target
)
weights_b <- wf_execute_plan(
  plan_b,
  approval_b,
  undercoverage_design,
  target
)
aligned_b <- weights_b$data$weight[
  match(weights_a$data$id, weights_b$data$id)
]
ratio_to_mean <- weights_a$data$weight / mean(weights_a$data$weight)

record_check(
  "undercoverage",
  "deterministic_cell_plan",
  identical(cells_a$identity, cells_b$identity),
  cells_a$identity
)
record_check(
  "undercoverage",
  "deterministic_weight_plan",
  identical(plan_a$identity, plan_b$identity),
  plan_a$identity
)
record_check(
  "undercoverage",
  "deterministic_weights",
  isTRUE(all.equal(weights_a$data$weight, aligned_b, tolerance = 1e-10)),
  sprintf("max_abs_diff=%.3g", max(abs(weights_a$data$weight - aligned_b)))
)
record_check(
  "undercoverage",
  "positive_weights",
  all(is.finite(weights_a$data$weight) & weights_a$data$weight > 0),
  sprintf("range=[%.6g, %.6g]", min(weights_a$data$weight), max(weights_a$data$weight))
)
record_check(
  "undercoverage",
  "bounded_relative_to_mean",
  min(ratio_to_mean) >= 0.3 - 1e-8 && max(ratio_to_mean) <= 3 + 1e-8,
  sprintf("ratio_range=[%.6g, %.6g]", min(ratio_to_mean), max(ratio_to_mean))
)
record_check(
  "undercoverage",
  "target_total_conserved",
  abs(sum(weights_a$data$weight) - 40) < 1e-8,
  sprintf("weighted_total=%.12g", sum(weights_a$data$weight))
)

# Outcomes are separate and cannot change plans or locked weights.
analysis <- data.frame(
  id = undercoverage_data$id,
  outcome = rnorm(nrow(undercoverage_data)),
  stringsAsFactors = FALSE
)
impact_a <- wf_assess_impact(weights_a, analysis, "id", "outcome")
analysis$outcome <- rev(analysis$outcome)
impact_b <- wf_assess_impact(weights_a, analysis, "id", "outcome")
record_check(
  "outcome_isolation",
  "locked_identity_unchanged",
  identical(impact_a$weight_identity, impact_b$weight_identity) &&
    identical(impact_a$weight_identity, weights_a$identity),
  weights_a$identity
)
record_check(
  "outcome_isolation",
  "impact_changes_without_reweighting",
  !identical(impact_a$summary, impact_b$summary),
  "outcome order changed after locking"
)

# Scenario 2: a target-positive empty age category is merged only through the
# explicit ladder and conserves sample count and base-weight total.
sparse_data <- data.frame(
  id = sprintf("s%02d", 1:12),
  sex = rep(c("F", "M"), each = 6),
  age = "18-34",
  base_weight = 1,
  stringsAsFactors = FALSE
)
sparse_design <- wf_prepare_design(
  sparse_data,
  id = "id",
  calibration = c("sex", "age"),
  base_weight = "base_weight"
)
sparse_ladder <- wf_collapse_ladder(
  dims,
  merge_age = list(age = c("18-34" = "all", "35+" = "all"))
)
sparse_plan <- wf_plan_cells(
  sparse_design,
  target,
  dims,
  min_cell = 5,
  max_weight_ratio = 4,
  ladder = sparse_ladder
)
record_check(
  "sparse_empty_cells",
  "sample_count_conserved",
  sum(sparse_plan$cells_before$n) == sum(sparse_plan$cells_after$n),
  sprintf("n=%d", sum(sparse_plan$cells_after$n))
)
record_check(
  "sparse_empty_cells",
  "base_weight_conserved",
  abs(sum(sparse_plan$cells_before$base_weight) -
    sum(sparse_plan$cells_after$base_weight)) < 1e-12,
  sprintf("base_total=%.6g", sum(sparse_plan$cells_after$base_weight))
)
record_check(
  "sparse_empty_cells",
  "no_unresolved_cells",
  nrow(sparse_plan$unresolved_cells) == 0,
  sprintf("affected_share=%.6g", sparse_plan$affected_share)
)

# Scenario 3: target mutation/inconsistent totals invalidate recorded identity.
tampered_target <- target
tampered_target$groups$`_all_`$margins$sex[[1]] <-
  tampered_target$groups$`_all_`$margins$sex[[1]] + 1
tamper_error <- tryCatch(
  wf_plan_weights(undercoverage_design, tampered_target, dims),
  error = identity
)
record_check(
  "inconsistent_totals",
  "tampered_target_blocked",
  inherits(tamper_error, "wf_error_safety"),
  if (inherits(tamper_error, "condition")) conditionMessage(tamper_error) else "no error"
)

# Scenario 4: extreme base weights require declared category merging and remain
# bounded after exact approved execution.
extreme_data <- undercoverage_data
extreme_data$base_weight[[1]] <- 50
extreme_design <- wf_prepare_design(
  extreme_data,
  id = "id",
  calibration = c("sex", "age"),
  base_weight = "base_weight"
)
extreme_ladder <- wf_collapse_ladder(
  dims,
  merge_age = list(age = c("18-34" = "all", "35+" = "all")),
  merge_sex = list(sex = c("F" = "all", "M" = "all"))
)
extreme_cells <- wf_plan_cells(
  extreme_design,
  target,
  dims,
  min_cell = 5,
  max_weight_ratio = 4,
  ladder = extreme_ladder
)
extreme_plan <- wf_plan_weights(
  extreme_design,
  target,
  dims,
  bounds = c(0.3, 3),
  min_cell = 5,
  cell_plan = extreme_cells
)
extreme_approval <- wf_approve_plan(
  extreme_plan,
  "Synthetic validation fixture",
  "test-only reviewer"
)
extreme_weights <- wf_execute_plan(
  extreme_plan,
  extreme_approval,
  extreme_design,
  target
)
extreme_ratio <- extreme_weights$data$weight / mean(extreme_weights$data$weight)
record_check(
  "extreme_base_weights",
  "positive_bounded_weights",
  all(extreme_weights$data$weight > 0) &&
    min(extreme_ratio) >= 0.3 - 1e-8 && max(extreme_ratio) <= 3 + 1e-8,
  sprintf("ratio_range=[%.6g, %.6g]", min(extreme_ratio), max(extreme_ratio))
)
record_check(
  "extreme_base_weights",
  "target_total_conserved",
  abs(sum(extreme_weights$data$weight) - 40) < 1e-8,
  sprintf("weighted_total=%.12g", sum(extreme_weights$data$weight))
)

# Optional external reference comparison. CI installs survey and sets
# WFC_REQUIRE_SURVEY=true, making absence a validation failure there.
survey_available <- requireNamespace("survey", quietly = TRUE)
if (identical(tolower(Sys.getenv("WFC_REQUIRE_SURVEY")), "true") &&
    !survey_available) {
  record_check(
    "survey_reference",
    "survey_available",
    FALSE,
    "WFC_REQUIRE_SURVEY=true but package 'survey' is unavailable"
  )
} else if (survey_available) {
  survey_design <- as_svydesign(
    weights_a,
    undercoverage_data,
    id = "id"
  )
  survey_sex <- survey::svytable(~sex, survey_design)
  survey_age <- survey::svytable(~age, survey_design)
  expected_sex <- target$groups$`_all_`$margins$sex[names(survey_sex)]
  expected_age <- target$groups$`_all_`$margins$age[names(survey_age)]
  max_difference <- max(
    abs(as.numeric(survey_sex) - as.numeric(expected_sex)),
    abs(as.numeric(survey_age) - as.numeric(expected_age))
  )
  record_check(
    "survey_reference",
    "raking_margins_match_survey",
    max_difference < 1e-6,
    sprintf("max_abs_difference=%.3g", max_difference)
  )
} else {
  record_check(
    "survey_reference",
    "survey_comparison_skipped",
    TRUE,
    "package 'survey' is not installed locally; CI requires it"
  )
}

print(results, row.names = FALSE)
cat("\nSession information:\n")
print(sessionInfo())

if (any(!results$passed)) {
  stop(sprintf("Safe-workflow validation failed %d check(s).", sum(!results$passed)))
}
cat(sprintf("\nAll %d recorded validation checks passed.\n", nrow(results)))

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", message = FALSE)

## -----------------------------------------------------------------------------
library(WFC)

dims_safe <- wf_dims(
  sex = c("F", "M"),
  age = c("18-34", "35+")
)
design_frame <- data.frame(
  id = sprintf("r%02d", 1:16),
  sex = rep(c("F", "M"), 8),
  age = rep(c("18-34", "18-34", "35+", "35+"), 4),
  base_weight = 1,
  stringsAsFactors = FALSE
)
analysis_frame <- data.frame(
  id = design_frame$id,
  satisfaction = seq(40, 70, length.out = 16),
  approved = rep(c(0, 1), 8),
  stringsAsFactors = FALSE
)

design <- wf_prepare_design(
  design_frame,
  id = "id",
  calibration = c("sex", "age"),
  base_weight = "base_weight"
)
design

## -----------------------------------------------------------------------------
csv_file <- system.file(
  "extdata", "safe-target-example.csv", package = "WFC"
)
csv_source <- paste0(csv_file, ".source.dcf")
cat(paste(readLines(csv_source, warn = FALSE), collapse = "\n"))

target_verified <- wf_import_target(
  csv_file,
  csv_source,
  dims_safe,
  key_map = c(sex = "sex", age = "age"),
  count = "count",
  production = FALSE
)
target_verified$identity

## -----------------------------------------------------------------------------
xlsx_file <- system.file(
  "extdata", "safe-target-example.xlsx", package = "WFC"
)
if (requireNamespace("openxlsx", quietly = TRUE)) {
  target_from_excel <- wf_import_target(
    xlsx_file,
    paste0(xlsx_file, ".source.dcf"),
    dims_safe,
    key_map = c(sex = "sex", age = "age"),
    count = "count",
    production = FALSE
  )
  identical(target_verified$groups, target_from_excel$groups)
}

## ----controlled-plan, eval=FALSE----------------------------------------------
# cell_plan <- wf_plan_cells(
#   design,
#   target_verified,
#   dims_safe,
#   min_cell = 5,
#   max_weight_ratio = 4
# )
# plan <- wf_plan_weights(
#   design,
#   target_verified,
#   dims_safe,
#   method = "raking",
#   bounds = c(0.3, 3),
#   min_cell = 5,
#   cell_plan = cell_plan
# )
# 
# plan$precheck
# is.null(plan$weights)

## ----guided-plan, eval=FALSE--------------------------------------------------
# guided <- wf_guided_plan(
#   design_frame,
#   id = "id",
#   calibration = c("sex", "age"),
#   dims = dims_safe,
#   target_file = csv_file,
#   source_file = csv_source,
#   source_type = "population",
#   key_map = c(sex = "sex", age = "age"),
#   count = "count",
#   base_weight = "base_weight",
#   production = FALSE
# )
# 
# wf_report(guided, audience = "decision")$table
# names(wf_report(guided, audience = "statistician")$sections)

## ----approval-and-execution, eval=FALSE---------------------------------------
# approval <- wf_approve_plan(
#   plan,
#   approver = "Reviewer name",
#   role = "statistician"
# )
# locked <- wf_execute_plan(
#   plan,
#   approval,
#   design,
#   target_verified
# )

## ----agent-refusal, eval=FALSE------------------------------------------------
# agent_refusal <- tryCatch(
#   wf_approve_plan(
#     plan,
#     approver = "Agent",
#     role = "assistant",
#     actor_type = "agent"
#   ),
#   wf_error_safety = function(condition) condition$data
# )
# agent_refusal[c("code", "severity", "next_actions")]

## ----post-lock, eval=FALSE----------------------------------------------------
# analysis_ready <- wf_attach_weights(
#   analysis_frame,
#   locked,
#   id = "id",
#   weight_name = ".weight"
# )
# impact <- wf_assess_impact(
#   locked,
#   analysis_frame,
#   id = "id",
#   outcomes = c("satisfaction", "approved")
# )
# 
# wf_report(locked, audience = "decision")
# wf_report(impact, audience = "statistician")
# wf_audit_export(impact, "impact-audit.json")

#' Run iterative proportional fitting for one group.
#'
#' @param w Initial weights.
#' @param idx Per-dimension integer category indices.
#' @param tg_list Per-dimension target margins.
#' @param tol Convergence tolerance.
#' @param max_iter Maximum iterations.
#' @param group Group key.
#' @keywords internal
#' @noRd
.wf_ipf <- function(w, idx, tg_list, tol, max_iter, group) {
  it <- 0L
  dev <- Inf
  while (it < max_iter && dev >= tol) {
    it <- it + 1L
    dev <- 0
    for (d in seq_along(tg_list)) {
      tg <- tg_list[[d]]
      K <- length(tg)
      cur <- .grp_sum(w, idx[[d]], K)
      if (any(tg > 0 & cur <= 0)) {
        wf_abort(sprintf(
          "Group '%s': category with positive target but zero current weight in dimension %d. Precheck should have caught this; possibly caused by trimming a lone supporter to zero.",
          group,
          d
        ), "wf_error_feasibility", list(group = group, dim = d))
      }
      ratio <- ifelse(cur > 0, tg / cur, 1)
      w <- w * ratio[idx[[d]]]
      dev <- max(dev, max(abs(ratio[cur > 0] - 1)))
    }
  }
  list(w = w, iterations = it, converged = dev < tol, final_dev = dev)
}

#' Expand one group for fractional NA allocation.
#'
#' @param rows_chr Character matrix of sample calibration keys.
#' @param na_mask Missingness matrix.
#' @param margins Target margins.
#' @param total Group total.
#' @keywords internal
#' @noRd
.wf_expand_group <- function(rows_chr, na_mask, margins, total, init = NULL) {
  n <- nrow(rows_chr)
  D <- ncol(rows_chr)
  base_w <- total / n
  rel <- if (is.null(init) || mean(init) == 0) rep(1, n) else init / mean(init)
  pid <- integer(0)
  vals <- vector("list", D)
  w0 <- numeric(0)
  for (d in seq_len(D)) {
    vals[[d]] <- character(0)
  }

  complete <- which(rowSums(na_mask) == 0)
  if (length(complete) > 0) {
    pid <- complete
    for (d in seq_len(D)) {
      vals[[d]] <- rows_chr[complete, d]
    }
    w0 <- base_w * rel[complete]
  }

  incomplete <- which(rowSums(na_mask) > 0)
  if (length(incomplete) > 0) {
    pat <- apply(na_mask[incomplete, , drop = FALSE], 1, function(z) paste(which(z), collapse = "-"))
    for (p in unique(pat)) {
      md <- as.integer(strsplit(p, "-")[[1]])
      rows <- incomplete[pat == p]
      lev <- lapply(md, function(d) names(margins[[d]]))
      grid <- expand.grid(lev, stringsAsFactors = FALSE)
      K <- nrow(grid)
      share <- rep(1, K)
      for (j in seq_along(md)) {
        share <- share * (margins[[md[j]]][grid[[j]]] / sum(margins[[md[j]]]))
      }
      pid <- c(pid, rep(rows, each = K))
      for (d in seq_len(D)) {
        vals[[d]] <- c(
          vals[[d]],
          if (d %in% md) {
            rep(grid[[match(d, md)]], times = length(rows))
          } else {
            rep(rows_chr[rows, d], each = K)
          }
        )
      }
      w0 <- c(
        w0,
        base_w * rel[rep(rows, each = K)] * rep(share, times = length(rows))
      )
    }
  }
  list(pid = pid, vals = vals, w0 = w0, n_persons = n)
}

#' Rake verified design data to a verified external target
#'
#' Enforces the WFC 2.0 verified-input boundary before grouped raking.
#'
#' @param design An unchanged `wf_design_data`.
#' @param target An unchanged, non-demo `wf_verified_target`.
#' @param ... Raking settings other than ID and base-weight roles, which are
#'   owned by `design`.
#'
#' @return A `wf_weights` object carrying verified input identities.
#' @export
wf_rake <- function(design, target, ...) {
  .wf_execute_verified_engine(design, target, "raking", list(...))
}

#' Internal grouped-raking engine.
#'
#' @keywords internal
#' @noRd
.wf_rake_engine <- function(sample, target, id = NULL,
                            na = c("fractional", "drop", "error"),
                            trim = NULL, trim_cycles = 4,
                            tol = 1e-6, max_iter = 200, precheck = TRUE,
                            init_weight = NULL,
                            parallel = FALSE, progress = FALSE) {
  na <- match.arg(na)
  t0 <- Sys.time()

  if (precheck) {
    pc <- wf_precheck(sample, target, id = id, na = na)
    if (!pc$ok) {
      wf_abort(sprintf(
        "Precheck reports %d blocking issue(s). Inspect wf_precheck(sample, target) and remediate (e.g. wf_apply_collapse) before raking.",
        sum(pc$issues$severity == "error")
      ), "wf_error_feasibility", list(precheck = pc))
    }
  }

  dvars <- target$dims
  if (na == "drop") {
    keep <- rowSums(sapply(dvars, function(d) is.na(sample[[d]]))) == 0
    if (any(!keep)) {
      wf_warn(
        sprintf("na='drop': removed %d row(s) with NA in calibration dimensions.", sum(!keep)),
        "wf_warning_data"
      )
    }
    sample <- sample[keep, , drop = FALSE]
  }

  if (is.null(init_weight)) {
    iw <- rep(1, nrow(sample))
  } else {
    if (length(init_weight) != 1 || !is.character(init_weight) ||
        !init_weight %in% names(sample)) {
      wf_abort(sprintf("init_weight column '%s' not found in sample.",
                       as.character(init_weight)[1]),
               "wf_error_schema", list(init_weight = init_weight))
    }
    iw <- as.numeric(sample[[init_weight]])
    if (any(!is.finite(iw)) || any(iw < 0)) {
      wf_abort("init_weight must be non-negative and finite.",
               "wf_error_input", list(init_weight = init_weight))
    }
  }

  gkey <- .wf_group_keys(sample, target$by)
  ids <- if (is.null(id)) seq_len(nrow(sample)) else sample[[id]]

  fit_group <- function(g) {
    sel <- which(gkey == g)
    gr <- target$groups[[g]]
    rows <- sapply(dvars, function(d) .chr(sample[[d]][sel]))
    if (is.null(dim(rows))) {
      rows <- matrix(rows, nrow = length(sel))
    }
    na_mask <- is.na(rows)

    ex <- .wf_expand_group(rows, na_mask, gr$margins, gr$total, init = iw[sel])
    idx <- lapply(seq_along(dvars), function(d) {
      match(ex$vals[[d]], names(gr$margins[[d]]))
    })
    tgl <- lapply(dvars, function(d) as.numeric(gr$margins[[d]]))

    fit <- .wf_ipf(ex$w0, idx, tgl, tol, max_iter, g)
    w <- fit$w
    trim_hits <- 0L

    if (!is.null(trim)) {
      mw <- gr$total / ex$n_persons
      lo <- trim[[1]] * mw
      hi <- trim[[2]] * mw
      for (cy in seq_len(trim_cycles)) {
        pw <- .grp_sum(w, ex$pid, ex$n_persons)
        fac <- pmin(pmax(pw, lo), hi) / pw
        if (all(abs(fac - 1) < 1e-12)) {
          break
        }
        w <- w * fac[ex$pid]
        fit <- .wf_ipf(w, idx, tgl, tol, max_iter, g)
        w <- fit$w
      }
      pw <- .grp_sum(w, ex$pid, ex$n_persons)
      for (r in 1:50) {
        pw2 <- pmin(pmax(pw, lo), hi)
        free <- pw2 > lo & pw2 < hi
        gap <- gr$total - sum(pw2)
        if (abs(gap) < 1e-9 * gr$total) {
          pw <- pw2
          break
        }
        if (!any(free)) {
          wf_abort(sprintf(
            "Group '%s': trim bounds [%g, %g] x mean are infeasible (all persons at a bound). Widen `trim`.",
            g,
            trim[[1]],
            trim[[2]]
          ), "wf_error_feasibility", list(group = g))
        }
        pw2[free] <- pw2[free] * (sum(pw2[free]) + gap) / sum(pw2[free])
        pw <- pw2
      }
      trim_hits <- sum(pw <= lo + 1e-12 | pw >= hi - 1e-12)
      w <- w * (pw / .grp_sum(w, ex$pid, ex$n_persons))[ex$pid]
    }

    if (!fit$converged) {
      worst <- which.max(vapply(seq_along(dvars), function(d) {
        cur <- .grp_sum(fit$w, idx[[d]], length(tgl[[d]]))
        max(abs(ifelse(cur > 0, tgl[[d]] / cur, 1) - 1))
      }, numeric(1)))
      wf_abort(sprintf(
        "Group '%s': IPF did not converge within %d iteration(s) (last deviation %.3g; worst dimension '%s'). Re-run wf_precheck() and collapse thin categories, or raise max_iter.",
        g,
        max_iter,
        fit$final_dev,
        dvars[[worst]]
      ), "wf_error_convergence", list(
        group = g,
        final_dev = fit$final_dev,
        worst_dim = dvars[[worst]],
        max_iter = max_iter
      ))
    }

    pw <- .grp_sum(w, ex$pid, ex$n_persons)
    achieved_g <- lapply(seq_along(dvars), function(d) {
      stats::setNames(.grp_sum(w, idx[[d]], length(tgl[[d]])), names(gr$margins[[d]]))
    })
    names(achieved_g) <- dvars

    res_row <- data.frame(
      id = ids[sel],
      group = g,
      weight = pw,
      feature = 1 / pw,
      stringsAsFactors = FALSE
    )
    log_row <- data.frame(
      group = g,
      n = ex$n_persons,
      pseudo_rows = length(w),
      iterations = fit$iterations,
      converged = fit$converged,
      final_dev = fit$final_dev,
      trimmed = trim_hits,
      stringsAsFactors = FALSE
    )
    list(data = res_row, log = log_row, achieved = achieved_g)
  }

  groups <- intersect(names(target$groups), unique(gkey))
  group_results <- .wf_parallel_map(
    as.list(groups),
    fit_group,
    use_parallel = parallel,
    progress = progress,
    label = "wf_rake"
  )
  names(group_results) <- groups
  res_rows <- lapply(group_results, `[[`, "data")
  logs <- lapply(group_results, `[[`, "log")
  achieved <- lapply(group_results, `[[`, "achieved")
  names(res_rows) <- groups
  names(logs) <- groups
  names(achieved) <- groups

  structure(list(
    data = do.call(rbind, res_rows),
    log = do.call(rbind, logs),
    achieved = achieved,
    provenance = list(
      dims = dvars,
      by = target$by,
      mode = target$mode,
      na = na,
      trim = trim,
      init_weight = init_weight,
      tol = tol,
      max_iter = max_iter,
      parallel = isTRUE(parallel),
      parallel_workers = attr(group_results, "workers"),
      progress = isTRUE(progress),
      collapsed = target$meta$collapsed,
      created = t0,
      elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = .wf_package_version()
    )
  ), class = "wf_weights")
}

#' Print calibrated weights
#'
#' @param x A `wf_weights` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_weights <- function(x, ...) {
  if (!is.null(x$provenance$method) && x$provenance$method == "soft") {
    relaxed <- if (is.data.frame(x$relaxation)) {
      sum(x$relaxation$relaxed)
    } else {
      NA_integer_
    }
    cat(sprintf(
      "<wf_weights>  %d unit(s) in %d group(s); method: soft calibration\n",
      nrow(x$data),
      nrow(x$log)
    ))
    cat(sprintf(
      "  weight range [%.4g, %.4g]; relaxed constraints: %s; elapsed %.2fs\n",
      min(x$data$weight),
      max(x$data$weight),
      as.character(relaxed),
      x$provenance$elapsed
    ))
    return(invisible(x))
  }
  if (!is.null(x$provenance$method) && x$provenance$method == "ebal") {
    cat(sprintf(
      "<wf_weights>  %d unit(s) in %d group(s); method: entropy balancing\n",
      nrow(x$data),
      nrow(x$log)
    ))
    cat(sprintf(
      "  weight range [%.4g, %.4g]; max residual %.2e; elapsed %.2fs\n",
      min(x$data$weight),
      max(x$data$weight),
      max(x$log$max_resid),
      x$provenance$elapsed
    ))
    return(invisible(x))
  }
  if (!is.null(x$provenance$method) &&
      x$provenance$method %in% c("greg", "logit")) {
    bnd <- if (is.null(x$provenance$bounds)) "none" else
      sprintf("[%.3g, %.3g]", x$provenance$bounds[1], x$provenance$bounds[2])
    cat(sprintf(
      "<wf_weights>  %d unit(s) in %d group(s); method: %s (%s); bounds: %s\n",
      nrow(x$data), nrow(x$log), x$provenance$method,
      x$provenance$distance, bnd
    ))
    cat(sprintf(
      "  weight range [%.4g, %.4g]; converged: %d/%d; elapsed %.2fs\n",
      min(x$data$weight), max(x$data$weight),
      sum(x$log$converged), nrow(x$log), x$provenance$elapsed
    ))
    return(invisible(x))
  }
  if (!is.null(x$provenance$method) && x$provenance$method == "propensity") {
    cat(sprintf(
      "<wf_weights>  %d unit(s) in %d group(s); method: propensity (%s / %s)\n",
      nrow(x$data),
      nrow(x$log),
      x$provenance$fit_method,
      x$provenance$weight
    ))
    cat(sprintf(
      "  weight range [%.4g, %.4g]; stabilized: %s; trimmed: %d; elapsed %.2fs\n",
      min(x$data$weight),
      max(x$data$weight),
      x$provenance$stabilize,
      x$provenance$trimmed,
      x$provenance$elapsed
    ))
    cat(sprintf(
      "  overlap: %d/%d online unit(s) above p > %.2f boundary\n",
      x$overlap$n_boundary,
      x$overlap$n_online,
      x$overlap$threshold
    ))
    return(invisible(x))
  }
  if (!is.null(x$provenance$method) && x$provenance$method == "poststrat") {
    cat(sprintf(
      "<wf_weights>  %d unit(s) in %d group(s); method: poststrat\n",
      nrow(x$data),
      nrow(x$log)
    ))
    cat(sprintf(
      "  weight range [%.4g, %.4g]; elapsed %.2fs\n",
      min(x$data$weight),
      max(x$data$weight),
      x$provenance$elapsed
    ))
    cat(sprintf(
      "  cells: %d raw -> %d resolved (median); groups using province granularity: %d\n",
      stats::median(x$log$n_cells_raw),
      stats::median(x$log$n_cells_resolved),
      sum(x$log$granularity_used == "province")
    ))
    cat(sprintf("  max group-total deviation: %.2e\n", max(x$log$total_dev)))
    return(invisible(x))
  }

  cat(sprintf(
    "<wf_weights>  %d unit(s) in %d group(s); mode: %s; converged: %d/%d\n",
    nrow(x$data),
    nrow(x$log),
    x$provenance$mode,
    sum(x$log$converged),
    nrow(x$log)
  ))
  cat(sprintf(
    "  weight range [%.4g, %.4g]; elapsed %.2fs\n",
    min(x$data$weight),
    max(x$data$weight),
    x$provenance$elapsed
  ))
  invisible(x)
}

# ============================================================================
# weightflow 0.1 -- core reference implementation (R/core.R)
# Zero hard dependencies. All exported symbols are prefixed `wf_`.
# Workflow discipline: wf_precheck() -> wf_rake() -> wf_diagnose()
# ============================================================================

# ----------------------------------------------------------------------------
# 0. Condition system --------------------------------------------------------
# ----------------------------------------------------------------------------

wf_abort <- function(message, class, data = list()) {
  stop(structure(
    class = c(class, "wf_error", "error", "condition"),
    list(message = message, call = sys.call(-1), data = data)
  ))
}

wf_warn <- function(message, class, data = list()) {
  warning(structure(
    class = c(class, "wf_warning", "warning", "condition"),
    list(message = message, call = sys.call(-1), data = data)
  ))
}

.chr <- function(x) trimws(as.character(x))

.require_cols <- function(df, cols, what) {
  miss <- setdiff(cols, names(df))
  if (length(miss) > 0)
    wf_abort(sprintf("%s is missing column(s): %s", what,
                     paste(miss, collapse = ", ")),
             "wf_error_schema", list(missing = miss))
}

# ----------------------------------------------------------------------------
# 1. Dimension specification -------------------------------------------------
# ----------------------------------------------------------------------------

#' Declare calibration dimensions.
#' @param ... name = levels pairs; levels may be NULL (inferred from target).
#' @param .collapse named list of ladders; each ladder is a named list of maps
#'        (old category -> new category, as a named character vector).
wf_dims <- function(..., .collapse = list()) {
  vars <- list(...)
  if (length(vars) == 0 || is.null(names(vars)) || any(names(vars) == ""))
    wf_abort("wf_dims() requires named arguments: dimension = levels (or NULL).",
             "wf_error_input")
  vars <- lapply(vars, function(v) if (is.null(v)) NULL else .chr(v))
  bad <- setdiff(names(.collapse), names(vars))
  if (length(bad) > 0)
    wf_abort(sprintf("Collapse ladder(s) for unknown dimension(s): %s",
                     paste(bad, collapse = ", ")),
             "wf_error_input", list(unknown = bad))
  structure(list(vars = vars, collapse = .collapse), class = "wf_dims")
}

# ----------------------------------------------------------------------------
# 2. Canonical target object and constructors --------------------------------
# ----------------------------------------------------------------------------

.wf_new_target <- function(mode, by, dims, groups, meta = list()) {
  tgt <- structure(list(mode = mode, by = by, dims = dims,
                        groups = groups, meta = meta),
                   class = "wf_target")
  .wf_validate_target(tgt)
  tgt
}

.wf_validate_target <- function(tgt, tol = 1e-8) {
  for (g in names(tgt$groups)) {
    gr <- tgt$groups[[g]]
    if (!is.finite(gr$total) || gr$total <= 0)
      wf_abort(sprintf("Group '%s': target total must be positive and finite.", g),
               "wf_error_input", list(group = g, total = gr$total))
    for (d in tgt$dims) {
      m <- gr$margins[[d]]
      if (is.null(m) || is.null(names(m)))
        wf_abort(sprintf("Group '%s', dim '%s': margins missing or unnamed.", g, d),
                 "wf_error_internal", list(group = g, dim = d))
      if (any(!is.finite(m)) || any(m < 0))
        wf_abort(sprintf("Group '%s', dim '%s': margins must be finite and >= 0.",
                         g, d), "wf_error_input", list(group = g, dim = d))
      if (abs(sum(m) - gr$total) > tol * gr$total)
        wf_abort(sprintf(
          "Group '%s', dim '%s': margins sum to %.6g but total is %.6g. %s",
          g, d, sum(m), gr$total,
          "Additivity is required so IPF preserves the group total."),
          "wf_error_input", list(group = g, dim = d))
    }
  }
  invisible(TRUE)
}

.wf_group_keys <- function(df, by, by_key = NULL) {
  if (!is.null(by_key)) {
    if (is.function(by_key)) return(.chr(by_key(df)))
    .require_cols(df, by_key, "population data")
    return(.chr(df[[by_key]]))
  }
  if (is.null(by)) return(rep("_all_", nrow(df)))
  .require_cols(df, by, "data")
  .chr(df[[by]])
}

.wf_scale_groups <- function(groups, scale, sample_n = NULL, totals = NULL) {
  if (scale == "population") return(groups)
  for (g in names(groups)) {
    new_total <-
      if (scale == "sample") {
        if (is.null(sample_n) || is.na(sample_n[g]))
          wf_abort(sprintf(
            "scale='sample' needs the sample: no size found for group '%s'.", g),
            "wf_error_input", list(group = g))
        sample_n[g]
      } else {
        if (is.null(totals) || is.na(totals[g]))
          wf_abort(sprintf("scale='custom' requires totals['%s'].", g),
                   "wf_error_input", list(group = g))
        totals[g]
      }
    f <- new_total / groups[[g]]$total
    groups[[g]]$total   <- unname(new_total)
    groups[[g]]$margins <- lapply(groups[[g]]$margins, function(m) m * f)
  }
  groups
}

#' Target from external population data in an arbitrary format.
#' Only `key_map`-ed columns, `count`, and the group key are ever read.
wf_target_population <- function(pop, key_map, count, dims,
                                 by = NULL, by_key = NULL,
                                 scale = c("population", "sample", "custom"),
                                 sample = NULL, totals = NULL) {
  scale <- match.arg(scale)
  if (!is.data.frame(pop)) wf_abort("`pop` must be a data.frame.", "wf_error_input")
  dvars <- names(dims$vars)
  if (is.null(names(key_map)) || !setequal(names(key_map), dvars))
    wf_abort("`key_map` must be a named vector mapping every wf_dims dimension to a population column.",
             "wf_error_schema", list(expected = dvars, got = names(key_map)))
  .require_cols(pop, unname(key_map), "population data")
  .require_cols(pop, count, "population data")

  cnt <- pop[[count]]
  if (any(is.na(cnt)))
    wf_abort(sprintf("Population count column '%s' has %d NA value(s).",
                     count, sum(is.na(cnt))), "wf_error_input")
  if (any(cnt < 0))
    wf_abort(sprintf("Population count column '%s' has negative values.", count),
             "wf_error_input")

  gkey <- .wf_group_keys(pop, by, by_key)
  groups <- list()
  for (g in sort(unique(gkey))) {
    sel <- gkey == g
    margins <- list()
    for (d in dvars) {
      v <- .chr(pop[[key_map[[d]]]][sel])
      m <- tapply(cnt[sel], v, sum)
      margins[[d]] <- m[m > 0]           # zero-count categories drop out
    }
    groups[[g]] <- list(total = sum(cnt[sel]), margins = margins)
  }

  sample_n <- NULL
  if (scale == "sample") {
    if (is.null(sample)) wf_abort("scale='sample' requires `sample`.", "wf_error_input")
    sample_n <- table(.wf_group_keys(sample, by))
    sample_n <- stats::setNames(as.numeric(sample_n), names(sample_n))
  }
  groups <- .wf_scale_groups(groups, scale, sample_n, totals)

  .wf_new_target("population", by, dvars, groups,
                 meta = list(scale = scale, created = Sys.time()))
}

#' Target from a weighted reference sample (feature value = 1/design weight).
wf_target_reference <- function(ref, feature, dims, by = NULL,
                                feature_na  = c("error", "drop"),
                                feature_gt1 = c("warn", "allow")) {
  feature_na <- match.arg(feature_na); feature_gt1 <- match.arg(feature_gt1)
  dvars <- names(dims$vars)
  .require_cols(ref, c(dvars, feature, by), "reference sample")

  f <- ref[[feature]]
  bad_na <- which(is.na(f))
  if (length(bad_na) > 0) {
    msg <- sprintf("%d of %d reference rows (%.1f%%) have missing feature values.",
                   length(bad_na), nrow(ref), 100 * length(bad_na) / nrow(ref))
    if (feature_na == "error")
      wf_abort(paste(msg, "Use feature_na='drop' to exclude them from target construction."),
               "wf_error_input", list(rows = bad_na))
    wf_warn(paste(msg, "Dropped from target construction."),
            "wf_warning_data", list(rows = bad_na))
    ref <- ref[-bad_na, , drop = FALSE]; f <- ref[[feature]]
  }
  bad <- which(!is.finite(f) | f <= 0)
  if (length(bad) > 0)
    wf_abort(sprintf("Feature values must be finite and > 0; %d row(s) violate this (first: %s).",
                     length(bad), paste(utils::head(bad, 5), collapse = ", ")),
             "wf_error_input", list(rows = bad))
  if (any(f > 1) && feature_gt1 == "warn")
    wf_warn(sprintf("%d feature value(s) exceed 1; inclusion probabilities should lie in (0,1]. Is this column a weight rather than its reciprocal?",
                    sum(f > 1)), "wf_warning_data")

  w <- 1 / f
  gkey <- .wf_group_keys(ref, by)
  groups <- list()
  for (g in sort(unique(gkey))) {
    sel <- gkey == g
    margins <- lapply(dvars, function(d) {
      v <- .chr(ref[[d]][sel])
      if (any(is.na(ref[[d]][sel])))
        wf_abort(sprintf("Reference sample: NA in dimension '%s' (group '%s'). Clean or drop these rows before target construction.",
                         d, g), "wf_error_input", list(group = g, dim = d))
      m <- tapply(w[sel], v, sum); m[m > 0]
    })
    names(margins) <- dvars
    groups[[g]] <- list(total = sum(w[sel]), margins = margins)
  }
  .wf_new_target("reference", by, dvars, groups,
                 meta = list(created = Sys.time()))
}

# ----------------------------------------------------------------------------
# 3. Precheck ----------------------------------------------------------------
# ----------------------------------------------------------------------------

.wf_issue <- function(group, dim, category, check, severity, detail) {
  data.frame(group = group, dim = dim, category = category, check = check,
             severity = severity, detail = detail, stringsAsFactors = FALSE)
}

wf_precheck <- function(sample, target, id = NULL,
                        na = c("fractional", "drop", "error"),
                        max_na_dims = 2, thin_min = 5, risk_ratio = 10) {
  na <- match.arg(na)
  iss <- list()
  dvars <- target$dims

  miss <- setdiff(c(dvars, target$by), names(sample))
  if (length(miss) > 0) {
    iss[[length(iss) + 1]] <- .wf_issue("*", paste(miss, collapse = ","), NA,
      "schema_missing_var", "error",
      sprintf("Sample is missing variable(s): %s", paste(miss, collapse = ", ")))
    out <- do.call(rbind, iss)
    return(structure(list(issues = out, ok = FALSE, na = na),
                     class = "wf_precheck"))
  }

  if (!is.null(id)) {
    dup <- sum(duplicated(sample[[id]]))
    if (dup > 0)
      iss[[length(iss) + 1]] <- .wf_issue("*", NA, NA, "dup_id", "error",
        sprintf("%d duplicated value(s) in id column '%s'.", dup, id))
  }

  gkey <- .wf_group_keys(sample, target$by)
  if (!is.null(target$by) && any(is.na(sample[[target$by]])))
    iss[[length(iss) + 1]] <- .wf_issue("*", target$by, NA, "na_group", "error",
      sprintf("%d row(s) have NA in grouping variable '%s'.",
              sum(is.na(sample[[target$by]])), target$by))

  s_only <- setdiff(unique(gkey), names(target$groups))
  t_only <- setdiff(names(target$groups), unique(gkey))
  if (length(s_only) > 0)
    iss[[length(iss) + 1]] <- .wf_issue("*", target$by, NA, "group_unmatched",
      "error", sprintf("Group(s) in sample but not in target: %s",
                       paste(s_only, collapse = ", ")))
  if (length(t_only) > 0)
    iss[[length(iss) + 1]] <- .wf_issue("*", target$by, NA, "group_unmatched",
      "warning", sprintf("Group(s) in target but not in sample (ignored): %s",
                         paste(t_only, collapse = ", ")))

  na_mat <- sapply(dvars, function(d) is.na(sample[[d]]))
  if (is.null(dim(na_mat))) na_mat <- matrix(na_mat, nrow = nrow(sample))
  n_na_dims <- rowSums(na_mat)
  over <- which(n_na_dims > max_na_dims)
  if (length(over) > 0)
    iss[[length(iss) + 1]] <- .wf_issue("*", NA, NA, "na_overload", "error",
      sprintf("%d row(s) have NA in more than %d dimensions (first rows: %s).",
              length(over), max_na_dims,
              paste(utils::head(over, 5), collapse = ", ")))
  if (na == "error" && any(n_na_dims > 0))
    iss[[length(iss) + 1]] <- .wf_issue("*", NA, NA, "na_load", "error",
      sprintf("na='error': %d row(s) contain NA in calibration dimensions.",
              sum(n_na_dims > 0)))

  for (g in intersect(names(target$groups), unique(gkey))) {
    sel <- gkey == g
    n_g <- sum(sel)
    for (d in dvars) {
      v      <- .chr(sample[[d]][sel])
      is_na  <- is.na(sample[[d]][sel])
      n_na_d <- sum(is_na)
      tg     <- target$groups[[g]]$margins[[d]]
      share_t <- tg / sum(tg)
      cnt <- table(v[!is_na])

      unk <- setdiff(names(cnt), names(tg))
      if (length(unk) > 0)
        iss[[length(iss) + 1]] <- .wf_issue(g, d, paste(unk, collapse = ","),
          "cat_unknown_in_sample", "error",
          sprintf("Categories in sample with no target share: %s (%d row(s)). Collapse or recode.",
                  paste(unk, collapse = ", "), sum(cnt[unk])))

      for (k in names(tg)) {
        n_k <- if (k %in% names(cnt)) cnt[[k]] else 0L
        eff_support <- n_k + if (na == "fractional") n_na_d else 0L
        if (eff_support == 0)
          iss[[length(iss) + 1]] <- .wf_issue(g, d, k, "cat_infeasible", "error",
            sprintf("Target share %.1f%% but zero sample support. IPF cannot converge; collapse this category.",
                    100 * share_t[[k]]))
        else if (n_k < thin_min && share_t[[k]] > 0.01)
          iss[[length(iss) + 1]] <- .wf_issue(g, d, k, "support_thin", "warning",
            sprintf("Only %d respondent(s) for a %.1f%% target share.",
                    n_k, 100 * share_t[[k]]))
        if (n_k > 0) {
          ratio <- share_t[[k]] / (n_k / max(n_g - n_na_d, 1))
          if (ratio > risk_ratio)
            iss[[length(iss) + 1]] <- .wf_issue(g, d, k, "risk_extreme_ratio",
              "warning",
              sprintf("Target/sample share ratio %.1f: expect weights ~%.0fx the mean; consider collapsing.",
                      ratio, ratio))
        }
      }
      if (n_na_d > 0)
        iss[[length(iss) + 1]] <- .wf_issue(g, d, NA, "na_load", "note",
          sprintf("%d row(s) (%.1f%%) NA on this dimension; policy: %s.",
                  n_na_d, 100 * n_na_d / n_g, na))
    }
  }

  issues <- if (length(iss) > 0) do.call(rbind, iss) else
    .wf_issue(character(0), character(0), character(0),
              character(0), character(0), character(0))
  structure(list(issues = issues,
                 ok = !any(issues$severity == "error"), na = na),
            class = "wf_precheck")
}

print.wf_precheck <- function(x, ...) {
  cat(sprintf("<wf_precheck>  verdict: %s\n",
              if (x$ok) "OK - no blocking issues" else "BLOCKED"))
  if (nrow(x$issues) == 0) { cat("  no issues found\n"); return(invisible(x)) }
  tab <- table(x$issues$severity)
  cat(sprintf("  %s\n", paste(sprintf("%s: %d", names(tab), tab), collapse = " | ")))
  ord <- order(match(x$issues$severity, c("error", "warning", "note")))
  show <- utils::head(x$issues[ord, ], 20)
  for (i in seq_len(nrow(show)))
    cat(sprintf("  [%s] %s / %s%s -- %s\n", show$severity[i], show$group[i],
                show$dim[i],
                ifelse(is.na(show$category[i]), "",
                       paste0(":", show$category[i])),
                show$detail[i]))
  if (nrow(x$issues) > 20)
    cat(sprintf("  ... and %d more (see $issues)\n", nrow(x$issues) - 20))
  invisible(x)
}

# ----------------------------------------------------------------------------
# 4. Collapse plans -----------------------------------------------------------
# ----------------------------------------------------------------------------

#' Apply a category-merge map consistently to sample and target.
#' @param plan list(dim = <name>, map = c(old = new, ...))
wf_apply_collapse <- function(sample, target, plan) {
  d <- plan$dim; map <- plan$map
  if (!d %in% target$dims)
    wf_abort(sprintf("Collapse plan targets unknown dimension '%s'.", d),
             "wf_error_input")
  v <- .chr(sample[[d]]); hit <- !is.na(v) & v %in% names(map)
  v[hit] <- map[v[hit]]
  sample[[d]] <- v
  for (g in names(target$groups)) {
    m <- target$groups[[g]]$margins[[d]]
    key <- names(m); hit <- key %in% names(map); key[hit] <- map[key[hit]]
    target$groups[[g]]$margins[[d]] <- tapply(as.numeric(m), key, sum)
  }
  .wf_validate_target(target)
  target$meta$collapsed <- c(target$meta$collapsed, list(plan))
  list(sample = sample, target = target)
}

# ----------------------------------------------------------------------------
# 5. Execution: wf_rake() -----------------------------------------------------
# ----------------------------------------------------------------------------

.grp_sum <- function(w, idx, K) {
  out <- numeric(K)
  rs <- rowsum(w, idx)
  out[as.integer(rownames(rs))] <- rs[, 1]
  out
}

.wf_ipf <- function(w, idx, tg_list, tol, max_iter, group) {
  it <- 0L; dev <- Inf
  while (it < max_iter && dev >= tol) {
    it <- it + 1L; dev <- 0
    for (d in seq_along(tg_list)) {
      tg <- tg_list[[d]]; K <- length(tg)
      cur <- .grp_sum(w, idx[[d]], K)
      if (any(tg > 0 & cur <= 0))
        wf_abort(sprintf(
          "Group '%s': category with positive target but zero current weight in dimension %d. Precheck should have caught this; possibly caused by trimming a lone supporter to zero.",
          group, d), "wf_error_feasibility", list(group = group, dim = d))
      ratio <- ifelse(cur > 0, tg / cur, 1)
      w <- w * ratio[idx[[d]]]
      dev <- max(dev, max(abs(ratio[cur > 0] - 1)))
    }
  }
  list(w = w, iterations = it, converged = dev < tol, final_dev = dev)
}

.wf_expand_group <- function(rows_chr, na_mask, margins, total) {
  n <- nrow(rows_chr); D <- ncol(rows_chr)
  base_w <- total / n
  pid <- integer(0); vals <- vector("list", D); w0 <- numeric(0)
  for (d in seq_len(D)) vals[[d]] <- character(0)

  complete <- which(rowSums(na_mask) == 0)
  if (length(complete) > 0) {
    pid <- complete
    for (d in seq_len(D)) vals[[d]] <- rows_chr[complete, d]
    w0 <- rep(base_w, length(complete))
  }

  incomplete <- which(rowSums(na_mask) > 0)
  if (length(incomplete) > 0) {
    pat <- apply(na_mask[incomplete, , drop = FALSE], 1,
                 function(z) paste(which(z), collapse = "-"))
    for (p in unique(pat)) {
      md   <- as.integer(strsplit(p, "-")[[1]])
      rows <- incomplete[pat == p]
      lev  <- lapply(md, function(d) names(margins[[d]]))
      grid <- expand.grid(lev, stringsAsFactors = FALSE)
      K    <- nrow(grid)
      share <- rep(1, K)
      for (j in seq_along(md))
        share <- share * (margins[[md[j]]][grid[[j]]] / sum(margins[[md[j]]]))
      pid <- c(pid, rep(rows, each = K))
      for (d in seq_len(D)) {
        vals[[d]] <- c(vals[[d]],
          if (d %in% md) rep(grid[[match(d, md)]], times = length(rows))
          else           rep(rows_chr[rows, d], each = K))
      }
      w0 <- c(w0, rep(base_w * share, times = length(rows)))
    }
  }
  list(pid = pid, vals = vals, w0 = w0, n_persons = n)
}

#' Calibrate sample weights to a wf_target by grouped raking.
wf_rake <- function(sample, target, id = NULL,
                    na = c("fractional", "drop", "error"),
                    trim = NULL, trim_cycles = 4,
                    tol = 1e-6, max_iter = 200, precheck = TRUE) {
  na <- match.arg(na)
  t0 <- Sys.time()

  if (precheck) {
    pc <- wf_precheck(sample, target, id = id, na = na)
    if (!pc$ok)
      wf_abort(sprintf(
        "Precheck reports %d blocking issue(s). Inspect wf_precheck(sample, target) and remediate (e.g. wf_apply_collapse) before raking.",
        sum(pc$issues$severity == "error")),
        "wf_error_feasibility", list(precheck = pc))
  }

  dvars <- target$dims
  if (na == "drop") {
    keep <- rowSums(sapply(dvars, function(d) is.na(sample[[d]]))) == 0
    if (any(!keep))
      wf_warn(sprintf("na='drop': removed %d row(s) with NA in calibration dimensions.",
                      sum(!keep)), "wf_warning_data")
    sample <- sample[keep, , drop = FALSE]
  }

  gkey <- .wf_group_keys(sample, target$by)
  ids  <- if (is.null(id)) seq_len(nrow(sample)) else sample[[id]]

  res_rows <- list(); logs <- list(); achieved <- list()
  for (g in intersect(names(target$groups), unique(gkey))) {
    sel  <- which(gkey == g)
    gr   <- target$groups[[g]]
    rows <- sapply(dvars, function(d) .chr(sample[[d]][sel]))
    if (is.null(dim(rows))) rows <- matrix(rows, nrow = length(sel))
    na_mask <- is.na(rows)

    ex  <- .wf_expand_group(rows, na_mask, gr$margins, gr$total)
    idx <- lapply(seq_along(dvars), function(d)
      match(ex$vals[[d]], names(gr$margins[[d]])))
    tgl <- lapply(dvars, function(d) as.numeric(gr$margins[[d]]))

    fit <- .wf_ipf(ex$w0, idx, tgl, tol, max_iter, g)
    w   <- fit$w
    trim_hits <- 0L

    if (!is.null(trim)) {
      mw <- gr$total / ex$n_persons
      lo <- trim[[1]] * mw; hi <- trim[[2]] * mw
      for (cy in seq_len(trim_cycles)) {
        pw  <- .grp_sum(w, ex$pid, ex$n_persons)
        fac <- pmin(pmax(pw, lo), hi) / pw
        if (all(abs(fac - 1) < 1e-12)) break
        w   <- w * fac[ex$pid]
        fit <- .wf_ipf(w, idx, tgl, tol, max_iter, g)
        w   <- fit$w
      }
      # final cap-and-redistribute: caps are respected AND the total is exact,
      # because the shortfall is reallocated only to un-capped persons
      pw <- .grp_sum(w, ex$pid, ex$n_persons)
      for (r in 1:50) {
        pw2  <- pmin(pmax(pw, lo), hi)
        free <- pw2 > lo & pw2 < hi
        gap  <- gr$total - sum(pw2)
        if (abs(gap) < 1e-9 * gr$total) { pw <- pw2; break }
        if (!any(free))
          wf_abort(sprintf(
            "Group '%s': trim bounds [%g, %g] x mean are infeasible (all persons at a bound). Widen `trim`.",
            g, trim[[1]], trim[[2]]),
            "wf_error_feasibility", list(group = g))
        pw2[free] <- pw2[free] * (sum(pw2[free]) + gap) / sum(pw2[free])
        pw <- pw2
      }
      trim_hits <- sum(pw <= lo + 1e-12 | pw >= hi - 1e-12)
      w <- w * (pw / .grp_sum(w, ex$pid, ex$n_persons))[ex$pid]
    }

    pw <- .grp_sum(w, ex$pid, ex$n_persons)
    achieved[[g]] <- lapply(seq_along(dvars), function(d)
      stats::setNames(.grp_sum(w, idx[[d]], length(tgl[[d]])),
                      names(gr$margins[[d]])))
    names(achieved[[g]]) <- dvars

    res_rows[[g]] <- data.frame(
      id = ids[sel], group = g, weight = pw, feature = 1 / pw,
      stringsAsFactors = FALSE)
    logs[[g]] <- data.frame(
      group = g, n = ex$n_persons, pseudo_rows = length(w),
      iterations = fit$iterations, converged = fit$converged,
      final_dev = fit$final_dev, trimmed = trim_hits,
      stringsAsFactors = FALSE)
  }

  structure(list(
    data = do.call(rbind, res_rows),
    log  = do.call(rbind, logs),
    achieved = achieved,
    provenance = list(
      dims = dvars, by = target$by, mode = target$mode, na = na,
      trim = trim, tol = tol, max_iter = max_iter,
      collapsed = target$meta$collapsed,
      created = t0, elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = "0.1.0")
  ), class = "wf_weights")
}

print.wf_weights <- function(x, ...) {
  cat(sprintf("<wf_weights>  %d unit(s) in %d group(s); mode: %s; converged: %d/%d\n",
              nrow(x$data), nrow(x$log), x$provenance$mode,
              sum(x$log$converged), nrow(x$log)))
  cat(sprintf("  weight range [%.4g, %.4g]; elapsed %.2fs\n",
              min(x$data$weight), max(x$data$weight), x$provenance$elapsed))
  invisible(x)
}

# ----------------------------------------------------------------------------
# 6. Diagnostics ---------------------------------------------------------------
# ----------------------------------------------------------------------------

wf_diagnose <- function(w, target = NULL, sample = NULL,
                        deff_ok = 3, deff_caveat = 10) {
  out <- lapply(split(w$data, w$data$group), function(d) {
    wt <- d$weight; cv <- stats::sd(wt) / mean(wt)
    data.frame(group = d$group[1], n = nrow(d),
               ess = round(sum(wt)^2 / sum(wt^2), 1),
               deff = round(1 + cv^2, 2),
               mean_w = mean(wt), min_w = min(wt), max_w = max(wt),
               max_over_mean = round(max(wt) / mean(wt), 1),
               stringsAsFactors = FALSE)
  })
  tab <- do.call(rbind, out)
  tab <- merge(tab, w$log[, c("group", "iterations", "converged", "trimmed")],
               by = "group")

  if (!is.null(target)) {
    # compare the margins the engine actually achieved (stored at fit time,
    # pseudo-row exact, so NA fractional allocation is handled correctly)
    merr <- vapply(tab$group, function(g) {
      e <- 0
      for (d in target$dims) {
        tg  <- target$groups[[g]]$margins[[d]]
        cur <- w$achieved[[g]][[d]][names(tg)]
        e <- max(e, max(abs(cur - tg) / tg))
      }
      e
    }, numeric(1))
    tab$margin_maxerr <- round(merr, 4)
  }

  tab$verdict <- ifelse(!tab$converged, "FAILED",
                 ifelse(tab$deff <= deff_ok, "OK",
                 ifelse(tab$deff <= deff_caveat, "usable with caveats",
                        "do not publish separately")))
  structure(list(table = tab), class = "wf_diagnostics")
}

print.wf_diagnostics <- function(x, ...) {
  cat("<wf_diagnostics>\n")
  print(x$table, row.names = FALSE)
  invisible(x)
}

stability_root_file <- function(path) {
  installed <- system.file(sub("^inst/", "", path), package = "WFC")
  if (nzchar(installed)) {
    return(installed)
  }
  if (file.exists("DESCRIPTION")) {
    return(path)
  }
  file.path("..", "..", path)
}

test_that("WFC 2.0 publishes an explicit safety contract", {
  contract <- readLines(
    stability_root_file("inst/stability/api-2.0.md"),
    warn = FALSE
  )

  expect_true(any(grepl("Retained safe exports", contract, fixed = TRUE)))
  expect_true(any(grepl("Removed interfaces", contract, fixed = TRUE)))
  expect_true(any(grepl("No supported replacement", contract, fixed = TRUE)))
})

test_that("wf_rake is deterministic and invariant to row order", {
  fixture <- make_weightflow_fixture()
  sample <- fixture$sample

  first <- .wf_rake_engine(sample, fixture$target, id = "id", tol = 1e-10)
  second <- .wf_rake_engine(sample, fixture$target, id = "id", tol = 1e-10)

  set.seed(901)
  shuffled <- sample[sample(seq_len(nrow(sample))), , drop = FALSE]
  reordered <- .wf_rake_engine(shuffled, fixture$target, id = "id", tol = 1e-10)

  first_weight <- stats::setNames(first$data$weight, first$data$id)
  second_weight <- stats::setNames(second$data$weight, second$data$id)
  reordered_weight <- stats::setNames(reordered$data$weight, reordered$data$id)

  expect_identical(second_weight[names(first_weight)], first_weight)
  expect_equal(
    reordered_weight[names(first_weight)],
    first_weight,
    tolerance = 1e-10
  )
})

test_that("wf_rake is invariant to calibration dimension order", {
  fixture <- make_weightflow_fixture()
  reversed_dims <- wf_dims(
    age = c("young", "old"),
    gender = c("female", "male")
  )
  reversed_target <- wf_target_population(
    fixture$pop,
    key_map = c(age = "age", gender = "gender"),
    count = "count",
    dims = reversed_dims,
    by = "province"
  )

  original <- .wf_rake_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    tol = 1e-10
  )
  reversed <- .wf_rake_engine(
    fixture$sample,
    reversed_target,
    id = "id",
    tol = 1e-10
  )

  original_weight <- stats::setNames(original$data$weight, original$data$id)
  reversed_weight <- stats::setNames(reversed$data$weight, reversed$data$id)

  expect_equal(
    reversed_weight[names(original_weight)],
    original_weight,
    tolerance = 1e-7
  )
})

make_ladder_draft_fixture <- function() {
  group <- rep(c("A", "B"), each = 12)
  age <- c(
    "young", rep("middle", 5), rep("old", 6),
    rep("young", 2), rep("middle", 4), rep("old", 6)
  )
  education <- rep("high", 24)
  education[c(1, 13)] <- "low"
  sample <- data.frame(
    id = paste0("u", seq_len(24)),
    region = group,
    age = age,
    education = education,
    stringsAsFactors = FALSE
  )
  dims <- wf_dims(
    age = c("young", "middle", "old"),
    education = c("low", "high")
  )
  population <- expand.grid(
    region = c("A", "B"),
    age = c("young", "middle", "old"),
    education = c("low", "high"),
    stringsAsFactors = FALSE
  )
  age_count <- c(young = 30, middle = 45, old = 45)
  education_count <- c(low = 30, high = 90)
  population$count <- age_count[population$age] *
    education_count[population$education] / 120
  target <- wf_target_population(
    population,
    key_map = c(age = "age", education = "education"),
    count = "count",
    dims = dims,
    by = "region",
    by_key = "region"
  )
  list(sample = sample, dims = dims, target = target)
}

test_that("wf_suggest_ladder drafts adjacent merges from worst-group support", {
  fixture <- make_ladder_draft_fixture()

  draft <- wf_suggest_ladder(
    fixture$sample,
    fixture$target,
    fixture$dims,
    min_cell = 3
  )

  expect_s3_class(draft, "wf_ladder_draft")
  expect_named(draft$levels, c("level1", "level2"))
  expect_identical(names(draft$levels$level1), "age")
  expect_identical(names(draft$levels$level2), "education")
  expect_equal(draft$levels$level1$age[["young"]], "young+middle")
  expect_equal(draft$levels$level1$age[["middle"]], "young+middle")
  expect_false("old" %in% names(draft$levels$level1$age))
  expect_s3_class(draft$ladder, "wf_collapse_ladder")
  expect_equal(draft$ladder$n_levels, 2)
})

test_that("wf_suggest_ladder orders dimensions by affected sample share", {
  fixture <- make_ladder_draft_fixture()
  draft <- wf_suggest_ladder(
    fixture$sample,
    fixture$target,
    fixture$dims,
    min_cell = 3
  )

  expect_lt(draft$affected_share[["age"]], draft$affected_share[["education"]])
  expect_identical(names(draft$affected_share), c("age", "education"))
  expect_equal(draft$support_before$age["young", "A"], 1)
  expect_equal(draft$support_before$age["young", "B"], 2)
})

test_that("wf_suggest_ladder returns a validated no-op draft when support is sufficient", {
  fixture <- make_ladder_draft_fixture()

  draft <- wf_suggest_ladder(
    fixture$sample,
    fixture$target,
    fixture$dims,
    min_cell = 1
  )

  expect_length(draft$levels, 0)
  expect_length(draft$affected_share, 0)
  expect_equal(draft$ladder$n_levels, 0)
})

test_that("wf_suggest_ladder does not mutate inputs", {
  fixture <- make_ladder_draft_fixture()
  sample_before <- fixture$sample
  target_before <- fixture$target
  dims_before <- fixture$dims

  wf_suggest_ladder(
    fixture$sample,
    fixture$target,
    fixture$dims,
    min_cell = 3
  )

  expect_identical(fixture$sample, sample_before)
  expect_identical(fixture$target, target_before)
  expect_identical(fixture$dims, dims_before)
})

test_that("wf_suggest_ladder requires declared order and compatible inputs", {
  fixture <- make_ladder_draft_fixture()
  inferred <- wf_dims(age = NULL, education = c("low", "high"))

  expect_error(
    wf_suggest_ladder(fixture$sample, fixture$target, inferred),
    class = "wf_error_input"
  )
  expect_error(
    wf_suggest_ladder(fixture$sample, fixture$target, fixture$dims, min_cell = 0),
    class = "wf_error_input"
  )

  missing_column <- fixture$sample
  missing_column$age <- NULL
  expect_error(
    wf_suggest_ladder(missing_column, fixture$target, fixture$dims),
    class = "wf_error_schema"
  )

  missing_group <- fixture$sample[fixture$sample$region == "A", ]
  expect_error(
    wf_suggest_ladder(missing_group, fixture$target, fixture$dims),
    class = "wf_error_feasibility"
  )
})

test_that("wf_suggest_ladder prints a review warning and mappings", {
  fixture <- make_ladder_draft_fixture()
  draft <- wf_suggest_ladder(
    fixture$sample,
    fixture$target,
    fixture$dims,
    min_cell = 3
  )

  expect_output(print(draft), "review before use")
  expect_output(print(draft), "young->young\\+middle")
})

make_collapse_fixture <- function() {
  sample <- data.frame(
    id = paste0("r", 1:6),
    province = "A",
    edu = c("low", "mid", "mid", "high", "high", "high"),
    stringsAsFactors = FALSE
  )
  pop <- data.frame(
    province = "A",
    edu = c("low", "mid", "high"),
    count = c(40, 30, 30),
    stringsAsFactors = FALSE
  )
  dims <- wf_dims(
    edu = c("low", "mid", "high"),
    .collapse = list(
      edu = list(step1 = c(low = "low_mid", mid = "low_mid", high = "high"))
    )
  )
  target <- wf_target_population(
    pop,
    key_map = c(edu = "edu"),
    count = "count",
    dims = dims,
    by = "province"
  )
  list(sample = sample, target = target, dims = dims)
}

test_that("suggest collapse creates a reviewable plan from precheck issues", {
  fixture <- make_collapse_fixture()
  sample <- subset(fixture$sample, edu != "low")
  pc <- wf_precheck(sample, fixture$target, id = "id", na = "drop", thin_min = 2)

  plan <- wf_suggest_collapse(pc, fixture$dims)

  expect_s3_class(plan, "wf_collapse_plan")
  expect_true(nrow(plan$actions) >= 1)
  expect_equal(plan$maps[[1]]$dim, "edu")
  expect_equal(plan$maps[[1]]$map[["low"]], "low_mid")
  expect_true(is.data.frame(plan$source_checks))
})

test_that("suggest collapse records unresolved issues without inventing maps", {
  fixture <- make_collapse_fixture()
  fixture$dims$collapse <- list()
  sample <- subset(fixture$sample, edu != "low")
  pc <- wf_precheck(sample, fixture$target, id = "id", na = "drop", thin_min = 2)

  plan <- wf_suggest_collapse(pc, fixture$dims)

  expect_s3_class(plan, "wf_collapse_plan")
  expect_length(plan$maps, 0)
  expect_gt(nrow(plan$unresolved), 0)
})

test_that("apply collapse accepts old list plans and new wf_collapse_plan objects", {
  fixture <- make_collapse_fixture()
  old <- wf_apply_collapse(
    fixture$sample,
    fixture$target,
    list(dim = "edu", map = c(low = "low_mid", mid = "low_mid"))
  )
  expect_true("low_mid" %in% old$sample$edu)
  expect_true("low_mid" %in% names(old$target$groups$A$margins$edu))

  sample <- subset(fixture$sample, edu != "low")
  pc <- wf_precheck(sample, fixture$target, id = "id", na = "drop", thin_min = 2)
  plan <- wf_suggest_collapse(pc, fixture$dims)
  applied <- wf_apply_collapse(sample, fixture$target, plan)

  expect_true("low_mid" %in% applied$sample$edu)
  expect_true("low_mid" %in% names(applied$target$groups$A$margins$edu))
  expect_true(!is.null(applied$target$meta$collapsed))
})

test_that("apply collapse keeps post-stratification joint cells synchronized", {
  fixture <- make_poststrat_fixture()
  total_before <- sum(fixture$target$joint$A$pop)

  applied <- wf_apply_collapse(
    fixture$sample,
    fixture$target,
    list(dim = "age", map = c(young = "all", old = "all"))
  )

  expect_identical(unique(applied$sample$age), "all")
  expect_identical(unique(applied$target$joint$A$age), "all")
  expect_equal(nrow(applied$target$joint$A), 2)
  expect_equal(sum(applied$target$joint$A$pop), total_before)
})

test_that("wf_dims validates named dimensions and collapse ladders", {
  expect_s3_class(wf_dims(gender = c("female", "male")), "wf_dims")
  expect_error(wf_dims(c("female", "male")), class = "wf_error_input")
  expect_error(
    wf_dims(gender = NULL, .collapse = list(age = list(step1 = c("1" = "all")))),
    class = "wf_error_input"
  )
})

test_that("population target validates schema and counts", {
  fixture <- make_weightflow_fixture()

  expect_s3_class(fixture$target, "wf_target")
  expect_equal(fixture$target$dims, c("gender", "age"))
  expect_equal(fixture$target$groups$A$total, 200)
  expect_equal(unname(fixture$target$groups$A$margins$gender["female"]), 100)

  expect_error(
    wf_target_population(
      pop = fixture$pop,
      key_map = c(gender = "missing", age = "age"),
      count = "count",
      dims = fixture$dims,
      by = "province"
    ),
    class = "wf_error_schema"
  )

  bad_pop <- fixture$pop
  bad_pop$count[1] <- -1
  expect_error(
    wf_target_population(
      pop = bad_pop,
      key_map = c(gender = "gender", age = "age"),
      count = "count",
      dims = fixture$dims,
      by = "province"
    ),
    class = "wf_error_input"
  )
})

test_that("target invariant rejects non-additive margins", {
  fixture <- make_weightflow_fixture()
  target <- fixture$target
  target$groups$A$margins$gender["female"] <- target$groups$A$margins$gender["female"] + 1

  expect_error(
    WFC:::.wf_validate_target(target),
    class = "wf_error_input"
  )
})

test_that("reference target screens invalid feature values", {
  fixture <- make_weightflow_fixture()
  ref <- fixture$sample
  ref$feature <- rep(0.5, nrow(ref))

  target <- wf_target_reference(ref, feature = "feature", dims = fixture$dims, by = "province")
  expect_s3_class(target, "wf_target")

  ref$feature[1] <- 0
  expect_error(
    wf_target_reference(ref, feature = "feature", dims = fixture$dims, by = "province"),
    class = "wf_error_input"
  )
})

test_that("wfc_example contains only simulated package data", {
  data("wfc_example", package = "WFC", envir = environment())

  expect_true(exists("wfc_example"))
  expect_true(is.list(wfc_example))
  expect_true(all(c("sample", "population", "dims") %in% names(wfc_example)))
  expect_true(is.data.frame(wfc_example$sample))
  expect_true(is.data.frame(wfc_example$population))
  expect_s3_class(wfc_example$dims, "wf_dims")
  expect_false(any(grepl("source", names(wfc_example))))
})

test_that("wfc_example supports the documented raking quick start", {
  data("wfc_example", package = "WFC", envir = environment())

  target <- wf_target_population(
    wfc_example$population,
    key_map = c(gender = "gender", age = "age"),
    count = "count",
    dims = wfc_example$dims,
    by = "province"
  )

  check <- wf_precheck(wfc_example$sample, target, id = "id")
  expect_true(check$ok)

  weights <- .wf_rake_engine(wfc_example$sample, target, id = "id")
  expect_true(all(weights$log$converged))
  expect_true(all(weights$data$weight > 0))
})

#' Return the loaded package version for provenance.
#'
#' @keywords internal
#' @noRd
.wf_replicates_package_version <- function() {
  tryCatch(
    as.character(utils::packageVersion("WFC")),
    error = function(e) "0.7.0"
  )
}

#' Resolve the sampling design (strata and PSUs) from data columns.
#'
#' @param data Input data frame.
#' @param strata Stratum column name or `NULL` (single stratum).
#' @param clusters PSU column name or `NULL` (each row is its own PSU).
#' @keywords internal
#' @noRd
.wf_design <- function(data, strata, clusters) {
  n <- nrow(data)
  if (!is.null(strata)) {
    if (length(strata) != 1 || !is.character(strata) ||
        !strata %in% names(data)) {
      wf_abort("`strata` must name a column in `data`.",
               "wf_error_input", list(strata = strata))
    }
    stratum <- .chr(data[[strata]])
  } else {
    stratum <- rep("1", n)
  }
  if (!is.null(clusters)) {
    if (length(clusters) != 1 || !is.character(clusters) ||
        !clusters %in% names(data)) {
      wf_abort("`clusters` must name a column in `data`.",
               "wf_error_input", list(clusters = clusters))
    }
    cluster <- .chr(data[[clusters]])
  } else {
    cluster <- as.character(seq_len(n))
  }

  pairs <- unique(data.frame(stratum = stratum, cluster = cluster,
                             stringsAsFactors = FALSE))
  dup <- pairs$cluster[duplicated(pairs$cluster)]
  if (length(dup) > 0) {
    wf_abort(
      sprintf("Clusters are not nested within strata: %s appear in >1 stratum.",
              paste(unique(dup), collapse = ", ")),
      "wf_error_design", list(clusters = unique(dup))
    )
  }

  strata_levels <- unique(stratum)
  psu <- lapply(strata_levels, function(h) unique(cluster[stratum == h]))
  names(psu) <- strata_levels
  list(n = n, stratum = stratum, cluster = cluster,
       strata = strata_levels, psu = psu)
}

#' Rao-Wu rescaled bootstrap multipliers.
#'
#' @param design A `.wf_design()` result.
#' @param R Number of replicates.
#' @param seed Optional integer seed.
#' @keywords internal
#' @noRd
.wf_boot_mult <- function(design, R, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  n <- design$n
  mult <- matrix(1, n, R)
  for (h in design$strata) {
    psus <- design$psu[[h]]
    nh <- length(psus)
    if (nh < 2) next
    units_by_psu <- lapply(psus, function(p) {
      which(design$stratum == h & design$cluster == p)
    })
    for (r in seq_len(R)) {
      draw <- sample.int(nh, nh - 1, replace = TRUE)
      counts <- tabulate(draw, nbins = nh)
      a <- (nh / (nh - 1)) * counts
      for (i in seq_len(nh)) {
        mult[units_by_psu[[i]], r] <- a[i]
      }
    }
  }
  list(mult = mult, scale = 1 / R, rscales = rep(1, R))
}

#' Stratified delete-one-PSU jackknife multipliers.
#'
#' @param design A `.wf_design()` result.
#' @keywords internal
#' @noRd
.wf_jack_mult <- function(design) {
  n <- design$n
  cols <- list()
  rscales <- numeric(0)
  for (h in design$strata) {
    psus <- design$psu[[h]]
    nh <- length(psus)
    in_h <- design$stratum == h
    if (nh < 2) {
      wf_warn(
        sprintf("Stratum '%s' has a single PSU; it cannot be jackknifed and contributes no replicate.", h),
        "wf_warning_quality", list(stratum = h)
      )
      next
    }
    for (p in psus) {
      m <- rep(1, n)
      m[in_h] <- nh / (nh - 1)
      m[in_h & design$cluster == p] <- 0
      cols[[length(cols) + 1]] <- m
      rscales <- c(rscales, (nh - 1) / nh)
    }
  }
  if (length(cols) == 0) {
    wf_abort("No stratum has >= 2 PSUs; jackknife has no replicates.",
             "wf_error_design")
  }
  list(mult = do.call(cbind, cols), scale = 1, rscales = rscales)
}

#' Sylvester-construction Hadamard matrix of order >= `n` (a power of two).
#'
#' @param n Minimum order.
#' @keywords internal
#' @noRd
.wf_hadamard <- function(n) {
  k <- 1
  while (k < n) k <- k * 2
  H <- matrix(1, 1, 1)
  while (nrow(H) < k) {
    H <- rbind(cbind(H, H), cbind(H, -H))
  }
  H
}

#' Balanced Repeated Replication multipliers (standard or Fay half-sampling).
#'
#' @param design A `.wf_design()` result; every stratum must have 2 PSUs.
#' @param rho Fay's BRR shrinkage parameter in `[0, 1)`.
#' @keywords internal
#' @noRd
.wf_brr_mult <- function(design, rho = 0) {
  sizes <- vapply(design$psu, length, integer(1))
  bad <- design$strata[sizes != 2]
  if (length(bad) > 0) {
    wf_abort(
      sprintf("BRR requires exactly 2 PSUs per stratum; not met by: %s.",
              paste(bad, collapse = ", ")),
      "wf_error_design", list(strata = bad)
    )
  }
  H <- length(design$strata)
  hmat <- .wf_hadamard(H + 1)
  R <- nrow(hmat)
  n <- design$n
  mult <- matrix(1, n, R)
  high <- 2 - rho
  low <- rho
  for (hi in seq_along(design$strata)) {
    h <- design$strata[hi]
    psus <- design$psu[[h]]
    in_h <- design$stratum == h
    u1 <- in_h & design$cluster == psus[1]
    u2 <- in_h & design$cluster == psus[2]
    for (r in seq_len(R)) {
      if (hmat[r, hi + 1] > 0) {
        mult[u1, r] <- high
        mult[u2, r] <- low
      } else {
        mult[u1, r] <- low
        mult[u2, r] <- high
      }
    }
  }
  list(
    mult = mult,
    scale = 1 / (R * (1 - rho)^2),
    rscales = rep(1, R)
  )
}

#' Generate re-calibrated replicate weights for variance estimation.
#'
#' Perturbs base weights by bootstrap, jackknife, or BRR multipliers and
#' re-runs a calibration pipeline (`refit`) on each replicate, so the resulting
#' variance captures calibration uncertainty. Pair with [wf_variance()].
#'
#' @param data Input data frame (one row per unit).
#' @param refit A closure `function(data, weights) -> wf_weights` that re-runs
#'   the calibration pipeline using `weights` as the base/initial weights.
#' @param method Replication method.
#' @param R Number of bootstrap replicates (ignored for jackknife / BRR).
#' @param strata Optional stratum column name (single stratum if `NULL`).
#' @param clusters Optional PSU column name (each row is its own PSU if `NULL`).
#' @param id Optional id column aligning replicate weights (row order if `NULL`).
#' @param base_weight Optional starting base-weight column (all `1` if `NULL`).
#' @param seed Optional integer seed for the bootstrap draws.
#' @param rho Fay's BRR shrinkage parameter in `[0, 1)`. Used only when
#'   `method = "brr"`; `rho = 0` gives standard BRR.
#' @param parallel Whether to re-run replicate refits with forked parallelism
#'   where available, using at most two workers. Windows falls back to serial
#'   execution.
#' @param progress Whether to show a `cli` progress bar when `cli` is installed.
#' @return A `wf_replicate_weights` object.
#' @export
wf_replicates <- function(data, refit,
                          method = c("bootstrap", "jackknife", "brr"),
                          R = 500, strata = NULL, clusters = NULL,
                          id = NULL, base_weight = NULL, seed = NULL,
                          rho = 0, parallel = FALSE, progress = FALSE) {
  if (!is.data.frame(data) || nrow(data) == 0) {
    wf_abort("`data` must be a non-empty data frame.", "wf_error_input")
  }
  if (!is.function(refit)) {
    wf_abort("`refit` must be a function(data, weights) returning a wf_weights.",
             "wf_error_input")
  }
  method <- match.arg(method)
  if (method == "bootstrap" &&
      (length(R) != 1 || !is.finite(R) || R < 1 || R != as.integer(R))) {
    wf_abort("`R` must be a positive integer.", "wf_error_input", list(R = R))
  }
  if (length(rho) != 1 || !is.numeric(rho) || !is.finite(rho) ||
      rho < 0 || rho >= 1) {
    wf_abort("`rho` must be one finite number in [0, 1).", "wf_error_input")
  }
  if (method != "brr" && rho != 0) {
    wf_abort("`rho` is only used when method = 'brr'.", "wf_error_input")
  }
  n <- nrow(data)
  if (!is.null(id)) {
    if (length(id) != 1 || !is.character(id) || !id %in% names(data)) {
      wf_abort("`id` must name a column in `data`.", "wf_error_input",
               list(id = id))
    }
    canon <- .chr(data[[id]])
  } else {
    canon <- as.character(seq_len(n))
  }
  if (anyDuplicated(canon)) {
    wf_abort("Unit ids are not unique.", "wf_error_input")
  }
  if (!is.null(base_weight)) {
    if (length(base_weight) != 1 || !is.character(base_weight) ||
        !base_weight %in% names(data)) {
      wf_abort("`base_weight` must name a column in `data`.", "wf_error_input",
               list(base_weight = base_weight))
    }
    base <- as.numeric(data[[base_weight]])
  } else {
    base <- rep(1, n)
  }
  if (any(!is.finite(base)) || any(base <= 0)) {
    wf_abort("`base_weight` must be positive and finite.", "wf_error_input")
  }

  design <- .wf_design(data, strata, clusters)
  t0 <- Sys.time()
  gen <- switch(
    method,
    bootstrap = .wf_boot_mult(design, R, seed),
    jackknife = .wf_jack_mult(design),
    brr = .wf_brr_mult(design, rho = rho)
  )

  align <- function(fit) {
    if (!inherits(fit, "wf_weights") || is.null(fit$data$id) ||
        is.null(fit$data$weight)) {
      wf_abort("`refit` must return a wf_weights with id and weight columns.",
               "wf_error_input")
    }
    fid <- .chr(fit$data$id)
    m <- match(canon, fid)
    if (length(fid) != n || anyNA(m)) {
      wf_abort("`refit` output ids do not match the input units.",
               "wf_error_input")
    }
    grp <- if (is.null(fit$data$group)) rep("all", n) else .chr(fit$data$group)[m]
    list(weight = as.numeric(fit$data$weight)[m], group = grp)
  }

  base_al <- align(refit(data, base))
  rg <- ncol(gen$mult)
  rep_results <- .wf_parallel_map(
    as.list(seq_len(rg)),
    function(r) align(refit(data, base * gen$mult[, r]))$weight,
    use_parallel = parallel,
    progress = progress,
    label = "wf_replicates"
  )
  repw <- do.call(cbind, rep_results)

  structure(list(
    base = data.frame(id = canon, group = base_al$group,
                      weight = base_al$weight, stringsAsFactors = FALSE),
    replicates = repw,
    scale = gen$scale,
    rscales = gen$rscales,
    method = method,
    rho = if (method == "brr") rho else NULL,
    design = list(strata = strata, clusters = clusters,
                  n_strata = length(design$strata), R = rg,
                  fay_rho = if (method == "brr") rho else NULL,
                  parallel_workers = attr(rep_results, "workers")),
    provenance = list(
      method = method, R = rg, seed = seed, rho = if (method == "brr") rho else NULL,
      strata = strata, clusters = clusters, base_weight = base_weight,
      parallel = isTRUE(parallel),
      parallel_workers = attr(rep_results, "workers"),
      progress = isTRUE(progress),
      created = t0, elapsed = as.numeric(Sys.time() - t0, units = "secs"),
      package_version = .wf_replicates_package_version()
    )
  ), class = "wf_replicate_weights")
}

#' Print replicate weights
#'
#' @param x A `wf_replicate_weights` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_replicate_weights <- function(x, ...) {
  method <- if (!is.null(x$rho) && x$rho > 0) {
    sprintf("%s (Fay rho %.3g)", x$method, x$rho)
  } else {
    x$method
  }
  cat(sprintf("<wf_replicate_weights>  %d unit(s); method: %s; %d replicate(s)\n",
              nrow(x$base), method, ncol(x$replicates)))
  cat(sprintf("  design: %d stratum(s); scale %.4g; elapsed %.2fs\n",
              x$design$n_strata, x$scale, x$provenance$elapsed))
  invisible(x)
}

test_that(".wf_design resolves strata and clusters", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  expect_equal(des$n, 8)
  expect_setequal(des$strata, c("A", "B"))
  expect_setequal(des$psu[["A"]], c("a1", "a2"))
})

test_that(".wf_design defaults each row to its own PSU and a single stratum", {
  d <- make_design_data()
  des <- .wf_design(d, strata = NULL, clusters = NULL)
  expect_equal(des$strata, "1")
  expect_equal(length(des$psu[["1"]]), 8)
})

test_that(".wf_design rejects clusters that span strata", {
  d <- make_design_data()
  d$psu[5] <- "a1"  # a1 now appears in both stratum A and B
  expect_error(
    .wf_design(d, strata = "stratum", clusters = "psu"),
    class = "wf_error_design"
  )
})

test_that(".wf_design rejects missing columns", {
  d <- make_design_data()
  expect_error(.wf_design(d, strata = "nope", clusters = NULL),
               class = "wf_error_input")
})

test_that(".wf_boot_mult returns an n x R matrix with unit scale/rscales", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  gen <- .wf_boot_mult(des, R = 50, seed = 1)

  expect_equal(dim(gen$mult), c(8, 50))
  expect_equal(gen$scale, 1 / 50)
  expect_equal(gen$rscales, rep(1, 50))
  expect_true(all(gen$mult >= 0))
})

test_that(".wf_boot_mult multipliers are constant within a PSU", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  gen <- .wf_boot_mult(des, R = 10, seed = 1)
  # rows 1,2 are PSU a1; rows 3,4 are PSU a2
  expect_equal(gen$mult[1, ], gen$mult[2, ])
  expect_equal(gen$mult[3, ], gen$mult[4, ])
})

test_that(".wf_boot_mult per-stratum multiplier mean is about 1", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  gen <- .wf_boot_mult(des, R = 4000, seed = 42)
  stratum_A <- colMeans(gen$mult[1:4, ])  # 2 PSUs x 2 units, mean over units per rep
  expect_equal(mean(stratum_A), 1, tolerance = 0.05)
})

test_that(".wf_boot_mult is reproducible with a seed", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  g1 <- .wf_boot_mult(des, R = 20, seed = 7)
  g2 <- .wf_boot_mult(des, R = 20, seed = 7)
  expect_identical(g1$mult, g2$mult)
})

test_that(".wf_jack_mult emits one replicate per PSU with the right rescale", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  gen <- .wf_jack_mult(des)

  # 2 strata x 2 PSUs = 4 replicates
  expect_equal(ncol(gen$mult), 4)
  expect_equal(gen$scale, 1)
  expect_equal(gen$rscales, rep((2 - 1) / 2, 4))
})

test_that(".wf_jack_mult deletes exactly one PSU and rescales its stratum", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  gen <- .wf_jack_mult(des)

  # first replicate deletes PSU a1 (rows 1,2): 0 there, 2 for rows 3,4, 1 elsewhere
  col1 <- gen$mult[, 1]
  expect_equal(col1[1:2], c(0, 0))
  expect_equal(col1[3:4], c(2, 2))       # n_h/(n_h-1) = 2
  expect_equal(col1[5:8], rep(1, 4))     # other stratum untouched
})

test_that(".wf_jack_mult warns on a single-PSU stratum and skips it", {
  d <- make_design_data()
  d$stratum[8] <- "C"
  d$psu[8] <- "c1"                       # stratum C has 1 PSU
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  expect_warning(.wf_jack_mult(des), class = "wf_warning_quality")
})

test_that(".wf_hadamard builds an orthogonal +/-1 matrix of order a power of two", {
  H <- .wf_hadamard(3)         # next power of two >= 3 is 4
  expect_equal(dim(H), c(4, 4))
  expect_true(all(H %in% c(-1, 1)))
  expect_equal(t(H) %*% H, diag(4) * 4)
})

test_that(".wf_brr_mult assigns 2/0 within each 2-PSU stratum", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  gen <- .wf_brr_mult(des)

  expect_equal(gen$scale, 1 / ncol(gen$mult))
  expect_equal(gen$rscales, rep(1, ncol(gen$mult)))
  # rows 1,2 are PSU a1; rows 3,4 are PSU a2 (stratum A)
  for (r in seq_len(ncol(gen$mult))) {
    expect_true(all(gen$mult[1:4, r] %in% c(0, 2)))
    expect_equal(gen$mult[1, r], gen$mult[2, r])   # same PSU, same multiplier
    expect_equal(gen$mult[3, r], gen$mult[4, r])
    expect_equal(gen$mult[1, r] + gen$mult[3, r], 2)  # exactly one PSU selected
  }
})

test_that(".wf_brr_mult supports Fay shrinkage", {
  d <- make_design_data()
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  gen <- .wf_brr_mult(des, rho = 0.3)

  expect_equal(gen$scale, 1 / (ncol(gen$mult) * (1 - 0.3)^2))
  expect_equal(gen$rscales, rep(1, ncol(gen$mult)))
  expect_true(all(gen$mult %in% c(0.3, 1.7)))
  for (r in seq_len(ncol(gen$mult))) {
    expect_equal(gen$mult[1, r] + gen$mult[3, r], 2)
  }
})

test_that(".wf_brr_mult rejects a stratum without exactly 2 PSUs", {
  d <- make_design_data()
  d$psu[4] <- "a3"             # stratum A now has 3 PSUs
  des <- .wf_design(d, strata = "stratum", clusters = "psu")
  expect_error(.wf_brr_mult(des), class = "wf_error_design")
})

# A trivial refit: no calibration, weights = base (so we can verify wiring).
trivial_refit <- function(data, weights) {
  structure(list(
    data = data.frame(id = data$id, group = "all",
                      weight = weights, feature = 1 / weights,
                      stringsAsFactors = FALSE)
  ), class = "wf_weights")
}

test_that("wf_replicates returns aligned base and replicate weights", {
  d <- make_design_data()
  rep_w <- wf_replicates(d, trivial_refit, method = "bootstrap",
                         R = 30, strata = "stratum", clusters = "psu",
                         id = "id", seed = 1)

  expect_s3_class(rep_w, "wf_replicate_weights")
  expect_equal(nrow(rep_w$base), 8)
  expect_equal(rep_w$base$id, d$id)
  expect_equal(dim(rep_w$replicates), c(8, 30))
  expect_equal(rep_w$base$weight, rep(1, 8))   # trivial refit, base = 1
  expect_equal(rep_w$provenance$method, "bootstrap")
  expect_equal(rep_w$provenance$seed, 1)
})

test_that("wf_replicates applies base_weight and perturbs it by the multipliers", {
  d <- make_design_data()
  d$bw <- rep(2, 8)
  rep_w <- wf_replicates(d, trivial_refit, method = "jackknife",
                         strata = "stratum", clusters = "psu",
                         id = "id", base_weight = "bw")
  # jackknife replicate 1 deletes PSU a1: rows 1,2 -> 0, rows 3,4 -> 2*2=4
  expect_equal(rep_w$replicates[1:2, 1], c(0, 0))
  expect_equal(rep_w$replicates[3:4, 1], c(4, 4))
})

test_that("wf_replicates errors when refit returns mismatched ids", {
  d <- make_design_data()
  bad_refit <- function(data, weights) {
    structure(list(
      data = data.frame(id = paste0("x", seq_len(nrow(data))),
                        group = "all", weight = weights,
                        feature = 1 / weights, stringsAsFactors = FALSE)
    ), class = "wf_weights")
  }
  expect_error(
    wf_replicates(d, bad_refit, method = "bootstrap", R = 3,
                  strata = "stratum", clusters = "psu", id = "id"),
    class = "wf_error_input"
  )
})

test_that("wf_replicates validates its inputs", {
  d <- make_design_data()
  expect_error(wf_replicates(d, "notfun", method = "bootstrap", id = "id"),
               class = "wf_error_input")
  expect_error(wf_replicates(d, trivial_refit, method = "bootstrap",
                             R = 0, id = "id"),
               class = "wf_error_input")
  expect_error(wf_replicates(d, trivial_refit, method = "bootstrap",
                             id = "missing"),
               class = "wf_error_input")
  expect_error(wf_replicates(d, trivial_refit, method = "brr",
                             strata = "stratum", clusters = "psu",
                             id = "id", rho = 1),
               class = "wf_error_input")
  expect_error(wf_replicates(d, trivial_refit, method = "bootstrap",
                             id = "id", rho = 0.3),
               class = "wf_error_input")
})

test_that("print.wf_replicate_weights reports method and replicate count", {
  d <- make_design_data()
  rep_w <- wf_replicates(d, trivial_refit, method = "bootstrap", R = 5,
                         strata = "stratum", clusters = "psu", id = "id",
                         seed = 1)
  expect_output(print(rep_w), "wf_replicate_weights")
  expect_output(print(rep_w), "bootstrap")

  fay <- wf_replicates(d, trivial_refit, method = "brr",
                       strata = "stratum", clusters = "psu",
                       id = "id", rho = 0.3)
  expect_equal(fay$rho, 0.3)
  expect_equal(fay$provenance$rho, 0.3)
  expect_output(print(fay), "Fay rho")
})

#' Create a structured precheck issue.
#'
#' @param group Group key.
#' @param dim Dimension name.
#' @param category Category key.
#' @param check Check identifier.
#' @param severity Issue severity.
#' @param detail Human-readable detail.
#' @keywords internal
#' @noRd
.wf_issue <- function(group, dim, category, check, severity, detail) {
  data.frame(
    group = group,
    dim = dim,
    category = category,
    check = check,
    severity = severity,
    detail = detail,
    stringsAsFactors = FALSE
  )
}

#' Precheck sample and target compatibility
#'
#' Runs structural feasibility checks before weighting.
#'
#' @param sample Sample data frame.
#' @param target A `wf_target` object.
#' @param id Optional unique unit identifier column.
#' @param na Missing calibration data policy.
#' @param max_na_dims Maximum allowed missing calibration dimensions per row.
#' @param thin_min Minimum unweighted support before warning.
#' @param risk_ratio Target/sample share ratio warning threshold.
#'
#' @return A `wf_precheck` object.
#' @export
#'
#' @examples
#' dims <- wf_dims(gender = c("female", "male"))
#' pop <- data.frame(gender = c("female", "male"), count = c(55, 45))
#' target <- wf_target_population(pop, c(gender = "gender"), "count", dims)
#' sample <- data.frame(id = 1:4, gender = c("female", "male", "female", "male"))
#' wf_precheck(sample, target, id = "id")
wf_precheck <- function(sample, target, id = NULL,
                        na = c("fractional", "drop", "error"),
                        max_na_dims = 2, thin_min = 5, risk_ratio = 10) {
  na <- match.arg(na)
  iss <- list()
  dvars <- target$dims

  miss <- setdiff(c(dvars, target$by), names(sample))
  if (length(miss) > 0) {
    iss[[length(iss) + 1]] <- .wf_issue(
      "*",
      paste(miss, collapse = ","),
      NA,
      "schema_missing_var",
      "error",
      sprintf("Sample is missing variable(s): %s", paste(miss, collapse = ", "))
    )
    out <- do.call(rbind, iss)
    return(structure(list(issues = out, ok = FALSE, na = na), class = "wf_precheck"))
  }

  if (!is.null(id)) {
    dup <- sum(duplicated(sample[[id]]))
    if (dup > 0) {
      iss[[length(iss) + 1]] <- .wf_issue(
        "*",
        NA,
        NA,
        "dup_id",
        "error",
        sprintf("%d duplicated value(s) in id column '%s'.", dup, id)
      )
    }
  }

  gkey <- .wf_group_keys(sample, target$by)
  if (!is.null(target$by) && any(is.na(sample[[target$by]]))) {
    iss[[length(iss) + 1]] <- .wf_issue(
      "*",
      target$by,
      NA,
      "na_group",
      "error",
      sprintf(
        "%d row(s) have NA in grouping variable '%s'.",
        sum(is.na(sample[[target$by]])),
        target$by
      )
    )
  }

  sample_groups <- unique(gkey[!is.na(gkey)])
  s_only <- setdiff(sample_groups, names(target$groups))
  t_only <- setdiff(names(target$groups), sample_groups)
  if (length(s_only) > 0) {
    iss[[length(iss) + 1]] <- .wf_issue(
      "*",
      target$by,
      NA,
      "group_unmatched",
      "error",
      sprintf("Group(s) in sample but not in target: %s", paste(s_only, collapse = ", "))
    )
  }
  if (length(t_only) > 0) {
    iss[[length(iss) + 1]] <- .wf_issue(
      "*",
      target$by,
      NA,
      "group_unmatched",
      "warning",
      sprintf(
        "Group(s) in target but not in sample (ignored): %s",
        paste(t_only, collapse = ", ")
      )
    )
  }

  na_mat <- sapply(dvars, function(d) is.na(sample[[d]]))
  if (is.null(dim(na_mat))) {
    na_mat <- matrix(na_mat, nrow = nrow(sample))
  }
  n_na_dims <- rowSums(na_mat)
  over <- which(n_na_dims > max_na_dims)
  if (length(over) > 0) {
    iss[[length(iss) + 1]] <- .wf_issue(
      "*",
      NA,
      NA,
      "na_overload",
      "error",
      sprintf(
        "%d row(s) have NA in more than %d dimensions (first rows: %s).",
        length(over),
        max_na_dims,
        paste(utils::head(over, 5), collapse = ", ")
      )
    )
  }
  if (na == "error" && any(n_na_dims > 0)) {
    iss[[length(iss) + 1]] <- .wf_issue(
      "*",
      NA,
      NA,
      "na_load",
      "error",
      sprintf(
        "na='error': %d row(s) contain NA in calibration dimensions.",
        sum(n_na_dims > 0)
      )
    )
  }

  for (g in intersect(names(target$groups), sample_groups)) {
    sel <- !is.na(gkey) & gkey == g
    n_g <- sum(sel)
    for (d in dvars) {
      v <- .chr(sample[[d]][sel])
      is_na <- is.na(sample[[d]][sel])
      n_na_d <- sum(is_na)
      tg <- target$groups[[g]]$margins[[d]]
      share_t <- tg / sum(tg)
      cnt <- table(v[!is_na])

      unk <- setdiff(names(cnt), names(tg))
      if (length(unk) > 0) {
        iss[[length(iss) + 1]] <- .wf_issue(
          g,
          d,
          paste(unk, collapse = ","),
          "cat_unknown_in_sample",
          "error",
          sprintf(
            "Categories in sample with no target share: %s (%d row(s)). Collapse or recode.",
            paste(unk, collapse = ", "),
            sum(cnt[unk])
          )
        )
      }

      for (k in names(tg)) {
        n_k <- if (k %in% names(cnt)) cnt[[k]] else 0L
        eff_support <- n_k + if (na == "fractional") n_na_d else 0L
        if (eff_support == 0) {
          iss[[length(iss) + 1]] <- .wf_issue(
            g,
            d,
            k,
            "cat_infeasible",
            "error",
            sprintf(
              "Target share %.1f%% but zero sample support. IPF cannot converge; collapse this category.",
              100 * share_t[[k]]
            )
          )
        } else if (n_k < thin_min && share_t[[k]] > 0.01) {
          iss[[length(iss) + 1]] <- .wf_issue(
            g,
            d,
            k,
            "support_thin",
            "warning",
            sprintf("Only %d respondent(s) for a %.1f%% target share.", n_k, 100 * share_t[[k]])
          )
        }
        if (n_k > 0) {
          ratio <- share_t[[k]] / (n_k / max(n_g - n_na_d, 1))
          if (ratio > risk_ratio) {
            iss[[length(iss) + 1]] <- .wf_issue(
              g,
              d,
              k,
              "risk_extreme_ratio",
              "warning",
              sprintf(
                "Target/sample share ratio %.1f: expect weights ~%.0fx the mean; consider collapsing.",
                ratio,
                ratio
              )
            )
          }
        }
      }
      if (n_na_d > 0) {
        iss[[length(iss) + 1]] <- .wf_issue(
          g,
          d,
          NA,
          "na_load",
          "note",
          sprintf("%d row(s) (%.1f%%) NA on this dimension; policy: %s.", n_na_d, 100 * n_na_d / n_g, na)
        )
      }
    }
  }

  issues <- if (length(iss) > 0) {
    do.call(rbind, iss)
  } else {
    .wf_issue(character(0), character(0), character(0), character(0), character(0), character(0))
  }
  structure(
    list(issues = issues, ok = !any(issues$severity == "error"), na = na),
    class = "wf_precheck"
  )
}

#' Print a precheck result
#'
#' @param x A `wf_precheck` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_precheck <- function(x, ...) {
  cat(sprintf(
    "<wf_precheck>  verdict: %s\n",
    if (x$ok) "OK - no blocking issues" else "BLOCKED"
  ))
  if (nrow(x$issues) == 0) {
    cat("  no issues found\n")
    return(invisible(x))
  }
  tab <- table(x$issues$severity)
  cat(sprintf("  %s\n", paste(sprintf("%s: %d", names(tab), tab), collapse = " | ")))
  ord <- order(match(x$issues$severity, c("error", "warning", "note")))
  show <- utils::head(x$issues[ord, ], 20)
  for (i in seq_len(nrow(show))) {
    cat(sprintf(
      "  [%s] %s / %s%s -- %s\n",
      show$severity[i],
      show$group[i],
      show$dim[i],
      ifelse(is.na(show$category[i]), "", paste0(":", show$category[i])),
      show$detail[i]
    ))
  }
  if (nrow(x$issues) > 20) {
    cat(sprintf("  ... and %d more (see $issues)\n", nrow(x$issues) - 20))
  }
  invisible(x)
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----guided, eval=FALSE-------------------------------------------------------
# guided <- wf_autoweigh(
#   design,
#   target,
#   dims,
#   method = "raking",
#   trim = NULL,
#   interactive = FALSE,
#   lang = "en"
# )
# 
# guided$weights
# guided$diagnostics
# guided$ledger

## ----views, eval=FALSE--------------------------------------------------------
# wf_report(guided$weights, audience = "decision", lang = "zh_CN")
# wf_report(guided$weights, audience = "statistician", lang = "en")

#' Build joint cell keys.
#'
#' @param mat Character matrix containing dimension columns.
#' @param dvars Dimension column names.
#' @keywords internal
#' @noRd
.wf_cell_key <- function(mat, dvars) {
  do.call(paste, c(lapply(dvars, function(d) mat[, d]), sep = "\r"))
}

#' Split joint cell keys back into dimension columns.
#'
#' @param keys Cell keys built by `.wf_cell_key()`.
#' @param dvars Dimension column names.
#' @keywords internal
#' @noRd
.wf_split_key <- function(keys, dvars) {
  parts <- do.call(rbind, strsplit(keys, "\r", fixed = TRUE))
  df <- as.data.frame(parts, stringsAsFactors = FALSE)
  names(df) <- dvars
  df
}

#' Apply cumulative collapse ladder maps.
#'
#' @param mat Character matrix containing dimension columns.
#' @param ladder A `wf_collapse_ladder` object.
#' @param lv Highest ladder level to apply.
#' @keywords internal
#' @noRd
.wf_apply_ladder <- function(mat, ladder, lv) {
  if (lv == 0) {
    return(mat)
  }
  for (k in seq_len(lv)) {
    step <- ladder$steps[[k]]
    for (d in names(step)) {
      map <- step[[d]]
      col <- mat[, d]
      hit <- !is.na(col) & col %in% names(map)
      col[hit] <- map[col[hit]]
      mat[, d] <- col
    }
  }
  mat
}

#' Validate post-stratification arguments.
#'
#' @param sample Sample data frame.
#' @param target A `wf_target` object.
#' @param min_cell Minimum sample count per resolved cell.
#' @param ladder A `wf_collapse_ladder` object.
#' @keywords internal
#' @noRd
.wf_check_poststrat_args <- function(sample, target, min_cell, ladder) {
  if (is.null(target$joint)) {
    wf_abort(
      "Target has no joint cell table. Rebuild with wf_target_population(..., keep_joint = TRUE).",
      "wf_error_schema"
    )
  }
  if (!inherits(ladder, "wf_collapse_ladder")) {
    wf_abort("`ladder` must be a wf_collapse_ladder object.", "wf_error_input")
  }
  if (missing(min_cell) || length(min_cell) != 1 || !is.finite(min_cell) ||
      min_cell < 1 || min_cell != as.integer(min_cell)) {
    wf_abort("`min_cell` must be a single integer >= 1.", "wf_error_input")
  }
  .require_cols(sample, c(target$dims, target$by), "sample")
  if (!identical(sort(ladder$dims), sort(target$dims))) {
    wf_abort(
      "Ladder dimensions do not match target dimensions.",
      "wf_error_schema",
      list(ladder = ladder$dims, target = target$dims)
    )
  }
  invisible(TRUE)
}

#' Resolve post-stratification cells for one group.
#'
#' @param smat Sample dimension matrix.
#' @param jdf Joint target data frame for one group.
#' @param dvars Dimension names.
#' @param ladder Collapse ladder.
#' @param min_cell Minimum sample count.
#' @param granularity Requested granularity.
#' @param group Group key.
#' @keywords internal
#' @noRd
.wf_resolve_group <- function(smat, jdf, dvars, ladder, min_cell,
                              granularity, group) {
  jmat <- as.matrix(jdf[, dvars, drop = FALSE])
  raw_pop_key <- .wf_cell_key(jmat, dvars)

  max_lv <- ladder$n_levels
  s_keys <- lapply(0:max_lv, function(lv) {
    .wf_cell_key(.wf_apply_ladder(smat, ladder, lv), dvars)
  })
  s_counts <- lapply(s_keys, table)
  p_keys <- lapply(0:max_lv, function(lv) {
    .wf_cell_key(.wf_apply_ladder(jmat, ladder, lv), dvars)
  })

  n_raw <- nrow(jdf)
  level_used <- integer(n_raw)
  resolved <- character(n_raw)
  gran_used <- granularity

  support_at <- function(lv, cellkeys) {
    tb <- s_counts[[lv + 1]]
    as.integer(ifelse(cellkeys %in% names(tb), tb[cellkeys], 0L))
  }

  if (granularity == "adaptive") {
    settle <- rep(0L, n_raw)
    repeat {
      before <- settle
      keys <- vapply(seq_len(n_raw), function(i) {
        p_keys[[settle[i] + 1]][i]
      }, character(1))
      supp <- vapply(seq_len(n_raw), function(i) {
        support_at(settle[i], keys[i])
      }, integer(1))
      under <- supp < min_cell & settle < max_lv
      if (any(under)) {
        bad <- unique(keys[under])
        lift <- keys %in% bad & settle < max_lv
        settle[lift] <- settle[lift] + 1L
      }

      for (L in seq_len(max_lv)) {
        kL <- p_keys[[L + 1]]
        hasL <- tapply(settle >= L, kL, any)
        lift <- as.logical(hasL[kL]) & settle < L
        if (any(lift)) {
          settle[lift] <- L
        }
      }

      if (identical(settle, before)) {
        break
      }
    }

    final_keys <- vapply(seq_len(n_raw), function(i) {
      p_keys[[settle[i] + 1]][i]
    }, character(1))
    final_supp <- vapply(seq_len(n_raw), function(i) {
      support_at(settle[i], final_keys[i])
    }, integer(1))
    if (any(final_supp < min_cell)) {
      gran_used <- "province"
    } else {
      level_used <- settle
      resolved <- final_keys
    }
  }

  if (gran_used == "province") {
    chosen <- NA_integer_
    for (lv in 0:max_lv) {
      supp <- support_at(lv, p_keys[[lv + 1]])
      if (all(supp >= min_cell)) {
        chosen <- lv
        break
      }
    }
    if (is.na(chosen)) {
      chosen <- max_lv
    }
    level_used <- rep(chosen, n_raw)
    resolved <- p_keys[[chosen + 1]]
  }

  supp_final <- vapply(seq_len(n_raw), function(i) {
    support_at(level_used[i], resolved[i])
  }, integer(1))
  orphan <- supp_final == 0

  list(
    raw_key = raw_pop_key,
    level_used = level_used,
    resolved = resolved,
    supp_final = supp_final,
    orphan = orphan,
    granularity_used = gran_used,
    s_keys = s_keys,
    p_keys = p_keys,
    jmat = jmat
  )
}

#' Redistribute orphan cells to supported ancestors.
#'
#' @param res Resolution result from `.wf_resolve_group()`.
#' @param min_cell Minimum sample count.
#' @param group Group key.
#' @keywords internal
#' @noRd
.wf_redistribute <- function(res, min_cell, group) {
  redist <- list()
  max_lv <- length(res$s_keys) - 1

  support_at_final <- function(res, k) {
    tb <- table(res$s_keys[[res$level_used[k] + 1]])
    key <- res$resolved[k]
    if (key %in% names(tb)) {
      as.integer(tb[key])
    } else {
      0L
    }
  }

  for (i in which(res$orphan)) {
    moved <- FALSE
    lv0 <- res$level_used[i]
    if (lv0 < max_lv) {
      for (lv in (lv0 + 1):max_lv) {
        anc <- res$p_keys[[lv + 1]][i]
        tb <- table(res$s_keys[[lv + 1]])
        if (anc %in% names(tb) && tb[anc] > 0) {
          redist[[length(redist) + 1]] <- data.frame(
            group = group,
            raw_cell = res$raw_key[i],
            orphan_cell = res$resolved[i],
            receiving_cell = anc,
            ancestor_level = lv,
            stringsAsFactors = FALSE
          )
          res$resolved[i] <- anc
          res$level_used[i] <- lv
          res$orphan[i] <- FALSE
          moved <- TRUE
          break
        }
      }
    }

    if (!moved) {
      supported <- which(!res$orphan)
      if (length(supported) == 0) {
        wf_abort(
          sprintf(
            "Group '%s': no resolved cell has any sample support; cannot post-stratify this group.",
            group
          ),
          "wf_error_feasibility",
          list(group = group)
        )
      }
      counts <- vapply(supported, function(k) support_at_final(res, k), integer(1))
      recv <- res$resolved[supported[which.max(counts)]]
      redist[[length(redist) + 1]] <- data.frame(
        group = group,
        raw_cell = res$raw_key[i],
        orphan_cell = res$resolved[i],
        receiving_cell = recv,
        ancestor_level = -1L,
        stringsAsFactors = FALSE
      )
      res$resolved[i] <- recv
      res$orphan[i] <- FALSE
    }
  }

  res$redist_log <- if (length(redist) > 0) do.call(rbind, redist) else NULL
  res
}

#' Test for one non-empty character scalar.
#'
#' @param x Object to test.
#' @keywords internal
#' @noRd
.wf_is_string <- function(x) {
  length(x) == 1 && is.character(x) && !is.na(x) && nzchar(x)
}

#' Format a timestamp for stable provenance.
#'
#' @param x POSIX time.
#' @keywords internal
#' @noRd
.wf_iso_time <- function(x = Sys.time()) {
  format(x, "%Y-%m-%dT%H:%M:%S%z")
}

#' Remove session-specific values before hashing.
#'
#' @param x Object to sanitize.
#' @keywords internal
#' @noRd
.wf_sanitize_for_hash <- function(x) {
  if (inherits(x, "formula")) {
    return(paste(deparse(x), collapse = " "))
  }
  if (inherits(x, "POSIXt")) {
    return(.wf_iso_time(x))
  }
  if (is.environment(x)) {
    return("<environment>")
  }
  if (is.function(x)) {
    return("<function>")
  }
  if (is.factor(x)) {
    return(as.character(x))
  }
  if (is.data.frame(x)) {
    out <- lapply(x, .wf_sanitize_for_hash)
    out$.row_names <- row.names(x)
    return(out)
  }
  if (is.list(x)) {
    return(lapply(x, .wf_sanitize_for_hash))
  }
  x
}

#' Build a dependency-free object hash for provenance.
#'
#' @param x Object to hash.
#' @keywords internal
#' @noRd
.wf_object_hash <- function(x) {
  raw <- serialize(.wf_sanitize_for_hash(x), NULL, version = 2)
  h <- 5381
  for (byte in as.integer(raw)) {
    h <- (h * 33 + byte) %% 2147483647
  }
  sprintf("%08x", as.integer(h))
}

#' Validate a pipeline target declaration.
#'
#' @param target Target spec or ready target.
#' @keywords internal
#' @noRd
.wf_pipeline_target_spec <- function(target) {
  if (inherits(target, "wf_verified_target")) {
    target <- list(mode = "object", target = target)
  } else if (inherits(target, "wf_target")) {
    .wf_safety_abort(
      "verified_target_required",
      "Pipeline object targets must be verified external targets.",
      "target",
      next_actions = "import_verified_target"
    )
  }
  if (!is.list(target)) {
    wf_abort("`target` must be a list or wf_target object.", "wf_error_input")
  }
  mode <- target$mode
  if (!.wf_is_string(mode)) {
    wf_abort("`target$mode` must be one non-empty string.", "wf_error_input")
  }
  if (identical(mode, "manual")) {
    .wf_safety_abort(
      "manual_pipeline_unsupported",
      "WFC 2.0 does not support manual pipeline targets.",
      "target$mode",
      next_actions = "import_verified_target"
    )
  }
  supported <- c("population", "reference", "object")
  if (!mode %in% supported) {
    wf_abort(
      sprintf(
        "Unsupported pipeline target mode '%s'. Supported modes: %s.",
        mode,
        paste(supported, collapse = ", ")
      ),
      "wf_error_input",
      list(mode = mode)
    )
  }
  allowed_fields <- if (mode == "object") c("mode", "target") else "mode"
  unknown_fields <- setdiff(names(target), allowed_fields)
  if (length(unknown_fields) > 0L) {
    wf_abort(
      sprintf(
        "Unsupported pipeline target field(s) for %s mode: %s.",
        mode,
        paste(unknown_fields, collapse = ", ")
      ),
      "wf_error_input",
      list(fields = unknown_fields, mode = mode)
    )
  }
  if (mode == "object" &&
      (!inherits(target$target, "wf_verified_target") ||
       isTRUE(target$target$demo_only) ||
       !identical(
         target$target$identity,
         .wf_verified_target_identity(target$target)
       ))) {
    .wf_safety_abort(
      "verified_target_required",
      "Pipeline object mode requires an unchanged, non-demo verified target.",
      "target",
      next_actions = "import_verified_target"
    )
  }
  target
}

#' Supported calibration controls by method.
#'
#' @param method Calibration method.
#' @keywords internal
#' @noRd
.wf_pipeline_calibrate_controls <- function(method) {
  common <- "method"
  switch(
    method,
    raking = c(
      common, "na", "trim", "trim_cycles", "tol", "max_iter",
      "precheck"
    ),
    poststrat = c(
      common, "min_cell", "ladder", "granularity",
      "empty_cell", "precheck", "tol"
    ),
    greg = c(
      common, "bounds", "na", "tol", "max_iter",
      "precheck"
    ),
    logit = c(
      common, "bounds", "na", "tol", "max_iter",
      "precheck"
    ),
    soft = c(
      common, "tolerance", "na", "max_outer", "precheck"
    ),
    ebal = c(
      common, "na", "tol", "max_iter",
      "precheck"
    )
  )
}

#' Validate a pipeline stage declaration.
#'
#' @param stages Named stage list.
#' @keywords internal
#' @noRd
.wf_pipeline_stage_spec <- function(stages) {
  if (!is.list(stages) || is.null(names(stages)) ||
      any(!nzchar(names(stages)))) {
    wf_abort("`stages` must be a named list.", "wf_error_input")
  }
  allowed <- c("propensity", "calibrate")
  unknown <- setdiff(names(stages), allowed)
  if (length(unknown) > 0) {
    wf_abort(
      sprintf(
        "Unknown pipeline stage(s): %s. Supported stages: %s.",
        paste(unknown, collapse = ", "),
        paste(allowed, collapse = ", ")
      ),
      "wf_error_input",
      list(stages = unknown)
    )
  }
  if (is.null(stages$calibrate)) {
    wf_abort(
      "Pipeline stages must include a `calibrate` stage.",
      "wf_error_schema"
    )
  }
  for (stage in names(stages)) {
    if (!is.list(stages[[stage]])) {
      wf_abort(
        sprintf("Pipeline stage '%s' must be a list.", stage),
        "wf_error_input",
        list(stage = stage)
      )
    }
  }

  if (!is.null(stages$propensity)) {
    allowed_propensity <- c(
      "formula", "method", "by", "id", "weight", "stabilize", "trim"
    )
    unknown <- setdiff(names(stages$propensity), allowed_propensity)
    if (length(unknown) > 0) {
      wf_abort(
        sprintf(
          "Unsupported propensity-stage argument(s): %s.",
          paste(unknown, collapse = ", ")
        ),
        "wf_error_input",
        list(arguments = unknown)
      )
    }
    if (is.null(stages$propensity$formula)) {
      wf_abort(
        "Pipeline stage 'propensity' requires `formula`.",
        "wf_error_schema"
      )
    }
  }
  if ("moments" %in% names(stages$calibrate)) {
    .wf_safety_abort(
      "inline_moments_unsupported",
      paste(
        "WFC 2.0 does not accept inline target moments.",
        "Use verified external margins."
      ),
      "moments",
      next_actions = "import_verified_external_margins"
    )
  }
  owned <- intersect(names(stages$calibrate), c("id", "init_weight"))
  if (length(owned) > 0L) {
    .wf_safety_abort(
      "design_role_override_unsupported",
      "Pipeline ID and base-weight roles must come from wf_design_data.",
      owned[[1]],
      evidence = list(arguments = owned),
      next_actions = "prepare_design_with_required_roles"
    )
  }

  method <- stages$calibrate$method
  if (is.null(method)) {
    method <- "raking"
  }
  supported <- c("raking", "poststrat", "greg", "logit", "soft", "ebal")
  if (!.wf_is_string(method) || !method %in% supported) {
    shown <- if (length(method) == 0) "<empty>" else as.character(method[[1]])
    wf_abort(
      sprintf(
        "Unsupported pipeline calibration method '%s'. Supported methods: %s.",
        shown,
        paste(supported, collapse = ", ")
      ),
      "wf_error_input",
      list(method = method)
    )
  }
  unknown <- setdiff(
    names(stages$calibrate),
    .wf_pipeline_calibrate_controls(method)
  )
  if (length(unknown) > 0) {
    wf_abort(
      sprintf(
        "Unsupported %s calibration-stage argument(s): %s.",
        method,
        paste(unknown, collapse = ", ")
      ),
      "wf_error_input",
      list(arguments = unknown, method = method)
    )
  }
  stages
}

#' Validate pipeline validation rules.
#'
#' @param validate Validation rule list.
#' @keywords internal
#' @noRd
.wf_pipeline_validation_spec <- function(validate) {
  if (is.null(validate)) {
    return(NULL)
  }
  if (!is.list(validate) || is.null(names(validate)) ||
      any(!nzchar(names(validate)))) {
    wf_abort("`validate` must be NULL or a named list.", "wf_error_input")
  }
  allowed <- c("max_deff", "max_margin_dev")
  unknown <- setdiff(names(validate), allowed)
  if (length(unknown) > 0) {
    wf_abort(
      sprintf(
        "Unsupported validation rule(s): %s.",
        paste(unknown, collapse = ", ")
      ),
      "wf_error_input",
      list(rules = unknown)
    )
  }
  for (rule in names(validate)) {
    value <- validate[[rule]]
    if (length(value) != 1 || !is.numeric(value) ||
        !is.finite(value) || value < 0) {
      wf_abort(
        sprintf("Validation rule `%s` must be one non-negative number.", rule),
        "wf_error_input",
        list(rule = rule, value = value)
      )
    }
  }
  validate
}

#' Declare a production weighting pipeline
#'
#' Creates a serializable specification for a recurring weighting run. The
#' specification records which verified target mode and calibration stage to
#' use and which post-run validation thresholds to check.
#'
#' @param target Target declaration. Use a list with `mode = "population"`,
#'   `"reference"`, or `"object"`, or pass a `wf_verified_target`.
#' @param stages Named stage list with a required `calibrate` stage.
#' @param validate Optional validation thresholds. Supported keys are
#'   `max_deff` and `max_margin_dev`.
#'
#' @return A `wf_pipeline` object.
#' @export
wf_pipeline <- function(target, stages, validate = NULL) {
  target <- .wf_pipeline_target_spec(target)
  stages <- .wf_pipeline_stage_spec(stages)
  validate <- .wf_pipeline_validation_spec(validate)
  hash_input <- list(target = target, stages = stages, validate = validate)

  structure(
    list(
      target = target,
      stages = stages,
      validate = validate,
      hash = .wf_object_hash(hash_input),
      created = .wf_iso_time(),
      package_version = .wf_package_version()
    ),
    class = "wf_pipeline"
  )
}

#' Print a pipeline specification
#'
#' @param x A `wf_pipeline` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_pipeline <- function(x, ...) {
  cat(sprintf(
    "<wf_pipeline> target: %s; stages: %s; hash: %s\n",
    x$target$mode,
    paste(names(x$stages), collapse = " -> "),
    x$hash
  ))
  if (!is.null(x$validate)) {
    rules <- paste(names(x$validate), unlist(x$validate), sep = "=", collapse = ", ")
    cat(sprintf("  validation: %s\n", rules))
  }
  invisible(x)
}

#' Require a dimensions object for target construction.
#'
#' @param dims Candidate dimensions object.
#' @keywords internal
#' @noRd
.wf_pipeline_require_dims <- function(dims) {
  if (!inherits(dims, "wf_dims")) {
    wf_abort(
      "`dims` must be a wf_dims object when the pipeline builds a target.",
      "wf_error_input"
    )
  }
  invisible(TRUE)
}

#' Build the runtime target for a pipeline run.
#'
#' @param spec A `wf_pipeline` object.
#' @param dims A `wf_dims` object.
#' @param population Runtime verified population target.
#' @param reference Runtime verified reference target.
#' @param need_joint Whether a joint population target is needed.
#' @keywords internal
#' @noRd
.wf_pipeline_build_target <- function(spec, dims, population,
                                      reference, need_joint) {
  mode <- spec$target$mode
  candidate <- switch(
    mode,
    object = spec$target$target,
    population = population,
    reference = reference,
    NULL
  )
  if (!inherits(candidate, "wf_verified_target") ||
      isTRUE(candidate$demo_only) ||
      !identical(candidate$identity, .wf_verified_target_identity(candidate))) {
    .wf_safety_abort(
      "verified_target_required",
      sprintf(
        "Pipeline target mode '%s' requires an unchanged verified target.",
        mode
      ),
      mode,
      next_actions = "import_verified_target"
    )
  }
  if (mode %in% c("population", "reference") &&
      !identical(candidate$source_type, mode)) {
    .wf_safety_abort(
      "pipeline_target_type_mismatch",
      "The verified target source type does not match the pipeline mode.",
      mode,
      evidence = list(source_type = candidate$source_type)
    )
  }
  if (!is.null(dims) &&
      (!inherits(dims, "wf_dims") ||
       !identical(names(dims$vars), candidate$dims))) {
    .wf_safety_abort(
      "target_dimensions_mismatch",
      "Pipeline dimensions must exactly match the verified target.",
      "dims"
    )
  }
  if (isTRUE(need_joint) && is.null(candidate$joint)) {
    .wf_safety_abort(
      "poststrat_joint_target_required",
      "Post-stratification requires a verified joint population target.",
      "target",
      next_actions = "import_joint_population_target"
    )
  }
  candidate
}

#' Choose a collision-free internal column name.
#'
#' @param data Data frame.
#' @param stem Preferred column name.
#' @keywords internal
#' @noRd
.wf_pipeline_internal_col <- function(data, stem) {
  out <- stem
  i <- 1L
  while (out %in% names(data)) {
    i <- i + 1L
    out <- paste0(stem, "_", i)
  }
  out
}

#' Prepare a runtime base-weight column.
#'
#' @param sample Sample data.
#' @param base_weight Column name, numeric vector, or `NULL`.
#' @keywords internal
#' @noRd
.wf_pipeline_base_weight <- function(sample, base_weight) {
  if (is.null(base_weight)) {
    return(list(sample = sample, column = NULL))
  }
  if (.wf_is_string(base_weight)) {
    .require_cols(sample, base_weight, "sample")
    value <- as.numeric(sample[[base_weight]])
    if (anyNA(value) || any(!is.finite(value)) || any(value <= 0)) {
      wf_abort(
        "`base_weight` column must contain finite positive values.",
        "wf_error_input",
        list(base_weight = base_weight)
      )
    }
    return(list(sample = sample, column = base_weight))
  }
  if (is.numeric(base_weight) && length(base_weight) == nrow(sample)) {
    if (anyNA(base_weight) || any(!is.finite(base_weight)) ||
        any(base_weight <= 0)) {
      wf_abort(
        "`base_weight` vector must contain finite positive values.",
        "wf_error_input"
      )
    }
    column <- .wf_pipeline_internal_col(sample, ".wf_base_weight")
    sample[[column]] <- as.numeric(base_weight)
    return(list(sample = sample, column = column))
  }
  wf_abort(
    "`base_weight` must be NULL, a sample column name, or a numeric vector with one value per sample row.",
    "wf_error_input"
  )
}

#' Run an optional propensity stage.
#'
#' @param sample Runtime sample data.
#' @param reference Runtime reference data.
#' @param propensity Propensity stage spec.
#' @param init_col Current initial weight column.
#' @keywords internal
#' @noRd
.wf_pipeline_run_propensity <- function(sample, reference, propensity,
                                        init_col) {
  if (is.null(reference)) {
    wf_abort(
      "Pipeline stage 'propensity' requires `reference` in wf_run().",
      "wf_error_input"
    )
  }

  target_args <- propensity[
    intersect(names(propensity), c("method", "by", "id"))
  ]
  target <- do.call(
    wf_target_propensity,
    c(
      list(
        online = sample,
        reference = reference,
        formula = propensity$formula
      ),
      target_args
    )
  )
  fit_args <- propensity[
    setdiff(names(propensity), c("formula", "method", "by", "id"))
  ]
  stage <- do.call(wf_propensity, c(list(target = target), fit_args))

  sample_ids <- if (is.null(propensity$id)) {
    as.character(seq_len(nrow(sample)))
  } else {
    .chr(sample[[propensity$id]])
  }
  order <- .wf_match_unit_ids(
    data.frame(.wf_id = sample_ids, stringsAsFactors = FALSE),
    ".wf_id",
    stage$data$id,
    "`propensity$data$id`",
    "wf_run() propensity stage"
  )
  stage_col <- .wf_pipeline_internal_col(sample, ".wf_propensity_weight")
  sample[[stage_col]] <- stage$data$weight[order]

  init_out <- stage_col
  if (!is.null(init_col)) {
    combined_col <- .wf_pipeline_internal_col(sample, ".wf_pipeline_init")
    sample[[combined_col]] <- sample[[init_col]] * sample[[stage_col]]
    init_out <- combined_col
  }

  list(sample = sample, column = init_out, result = stage)
}

#' Validate a pipeline result against declared thresholds.
#'
#' @param w Weights.
#' @param target Target.
#' @param rules Validation rules.
#' @keywords internal
#' @noRd
.wf_pipeline_validate_result <- function(w, target, rules) {
  if (is.null(rules)) {
    return(NULL)
  }
  table <- wf_diagnose(w, target = target)$table
  issues <- list()
  if (!is.null(rules$max_deff)) {
    hit <- is.finite(table$deff) & table$deff > rules$max_deff
    if (any(hit)) {
      issues[[length(issues) + 1L]] <- data.frame(
        check = "max_deff",
        group = table$group[hit],
        value = table$deff[hit],
        threshold = rules$max_deff,
        detail = sprintf(
          "Design effect exceeds declared maximum %.6g.",
          rules$max_deff
        ),
        stringsAsFactors = FALSE
      )
    }
  }
  if (!is.null(rules$max_margin_dev) &&
      "margin_maxerr" %in% names(table)) {
    hit <- is.finite(table$margin_maxerr) &
      table$margin_maxerr > rules$max_margin_dev
    if (any(hit)) {
      issues[[length(issues) + 1L]] <- data.frame(
        check = "max_margin_dev",
        group = table$group[hit],
        value = table$margin_maxerr[hit],
        threshold = rules$max_margin_dev,
        detail = sprintf(
          "Margin deviation exceeds declared maximum %.6g.",
          rules$max_margin_dev
        ),
        stringsAsFactors = FALSE
      )
    }
  }
  issues <- if (length(issues) > 0) {
    do.call(rbind, issues)
  } else {
    data.frame(
      check = character(0),
      group = character(0),
      value = numeric(0),
      threshold = numeric(0),
      detail = character(0),
      stringsAsFactors = FALSE
    )
  }
  out <- structure(
    list(
      ok = nrow(issues) == 0,
      issues = issues,
      table = table,
      rules = rules
    ),
    class = "wf_pipeline_validation"
  )
  if (!out$ok) {
    wf_warn(
      sprintf(
        "Pipeline validation detected %d issue(s).",
        nrow(issues)
      ),
      "wf_warning_quality",
      list(issues = issues)
    )
  }
  out
}

#' Run a production weighting pipeline
#'
#' Executes a `wf_pipeline()` specification against verified design and target
#' objects, records provenance, and evaluates declared validation thresholds.
#'
#' @param spec A `wf_pipeline` object.
#' @param sample An unchanged `wf_design_data` object.
#' @param dims Optional `wf_dims` used to verify target dimensions.
#' @param population A verified population target for population mode.
#' @param reference A verified reference target for reference mode.
#' @param base_weight Deprecated runtime base-weight input. WFC 2.0 requires
#'   the base-weight role to be declared in `sample`; supplying this argument
#'   raises a safety error.
#'
#' @return A `wf_weights` object with pipeline provenance and optional
#'   `$pipeline_validation`.
#' @export
wf_run <- function(spec, sample, dims = NULL, population = NULL,
                   reference = NULL, base_weight = NULL) {
  if (!inherits(spec, "wf_pipeline")) {
    wf_abort("`spec` must be a wf_pipeline object.", "wf_error_input")
  }
  if (!inherits(sample, "wf_design_data") ||
      !identical(sample$identity, .wf_design_identity(sample$data, sample$roles))) {
    .wf_safety_abort(
      "verified_weighting_inputs_required",
      "`sample` must be an unchanged wf_design_data object.",
      "sample",
      next_actions = "prepare_design_data"
    )
  }
  if (!is.null(base_weight)) {
    .wf_safety_abort(
      "design_role_override_unsupported",
      "Declare the base-weight role in wf_prepare_design(), not wf_run().",
      "base_weight",
      next_actions = "prepare_design_with_required_roles"
    )
  }
  if (!is.null(spec$stages$propensity)) {
    .wf_safety_abort(
      "pipeline_propensity_unsupported",
      paste(
        "WFC 2.0 pipelines do not fit a runtime propensity stage before",
        "verified calibration."
      ),
      "stages$propensity",
      next_actions = "use_preapproved_design_variables"
    )
  }

  calibrate <- spec$stages$calibrate
  method <- calibrate$method
  if (is.null(method)) {
    method <- "raking"
  }

  target <- .wf_pipeline_build_target(
    spec,
    dims,
    population,
    reference,
    need_joint = identical(method, "poststrat")
  )

  args <- calibrate[setdiff(names(calibrate), "method")]
  w <- do.call(
    wf_calibrate,
    c(list(design = sample, target = target, method = method), args)
  )

  w$provenance$pipeline_hash <- spec$hash
  w$provenance$pipeline <- list(
    hash = spec$hash,
    created = spec$created,
    target_mode = spec$target$mode,
    stages = names(spec$stages),
    validate = spec$validate,
    package_version = spec$package_version
  )
  validation <- .wf_pipeline_validate_result(w, target, spec$validate)
  if (!is.null(validation)) {
    w$pipeline_validation <- validation
  }
  w
}

#' Compute group-level weight statistics for validation.
#'
#' @param w A `wf_weights` object.
#' @param target Optional target for margin residuals.
#' @keywords internal
#' @noRd
.wf_validation_group_stats <- function(w, target = NULL) {
  if (!is.data.frame(w$data) ||
      !all(c("group", "weight") %in% names(w$data))) {
    wf_abort("`wf_weights` object has malformed `$data`.", "wf_error_internal")
  }
  parts <- split(w$data, w$data$group)
  out <- lapply(parts, function(part) {
    weight <- as.numeric(part$weight)
    total <- sum(weight)
    sum_squares <- sum(weight^2)
    ess <- if (sum_squares > 0) total^2 / sum_squares else NA_real_
    deff <- if (length(weight) > 1 && mean(weight) > 0) {
      1 + (stats::sd(weight) / mean(weight))^2
    } else {
      NA_real_
    }
    data.frame(
      group = part$group[[1]],
      n = nrow(part),
      total_weight = total,
      ess = ess,
      deff = deff,
      stringsAsFactors = FALSE
    )
  })
  stats <- do.call(rbind, out)
  row.names(stats) <- NULL

  if (!is.null(target) && !is.null(w$achieved)) {
    margin <- vapply(stats$group, function(group) {
      if (is.null(target$groups[[group]]) || is.null(w$achieved[[group]])) {
        return(NA_real_)
      }
      max_error <- 0
      for (dim in target$dims) {
        target_margin <- target$groups[[group]]$margins[[dim]]
        current <- w$achieved[[group]][[dim]][names(target_margin)]
        if (is.null(current) || anyNA(current)) {
          return(NA_real_)
        }
        max_error <- max(
          max_error,
          max(abs(current - target_margin) / target_margin)
        )
      }
      max_error
    }, numeric(1))
    stats$margin_maxerr <- margin
  }
  stats
}

#' Validate one drift threshold.
#'
#' @param value Candidate threshold.
#' @param name Threshold name.
#' @param lower Minimum accepted value.
#' @keywords internal
#' @noRd
.wf_validate_threshold <- function(value, name, lower = 0) {
  if (length(value) != 1 || !is.numeric(value) ||
      !is.finite(value) || value < lower) {
    wf_abort(
      sprintf("`%s` must be one finite number >= %g.", name, lower),
      "wf_error_input",
      list(name = name, value = value)
    )
  }
  invisible(TRUE)
}

#' Compare calibrated weights against a reference release
#'
#' Detects drift in recurring production weights by comparing group coverage,
#' design effect, effective sample size, total weight, optional margin
#' residuals, and per-unit weight ratios for matching IDs.
#'
#' @param new New `wf_weights` object.
#' @param reference Reference `wf_weights` object.
#' @param target Optional `wf_target` used to compare margin residual drift.
#' @param max_deff_delta Maximum allowed increase in design effect.
#' @param max_ess_loss Maximum allowed fractional ESS loss.
#' @param max_total_shift Maximum allowed fractional group-total shift.
#' @param max_margin_delta Maximum allowed absolute margin-residual increase.
#' @param max_ratio_p99 Maximum allowed normalized 99th percentile unit-weight
#'   ratio; the reciprocal threshold is applied to the 1st percentile.
#' @param on_issue Reaction when drift is detected: `"warn"`, `"error"`, or
#'   `"none"`.
#'
#' @return A `wf_validation` object.
#' @export
wf_validate <- function(new, reference, target = NULL,
                        max_deff_delta = 1,
                        max_ess_loss = 0.2,
                        max_total_shift = 0.05,
                        max_margin_delta = 0.01,
                        max_ratio_p99 = 2,
                        on_issue = c("warn", "error", "none")) {
  if (!inherits(new, "wf_weights") || !inherits(reference, "wf_weights")) {
    wf_abort(
      "`new` and `reference` must both be wf_weights objects.",
      "wf_error_input"
    )
  }
  if (!is.null(target) && !inherits(target, "wf_target")) {
    wf_abort("`target` must be NULL or a wf_target object.", "wf_error_input")
  }
  on_issue <- match.arg(on_issue)
  .wf_validate_threshold(max_deff_delta, "max_deff_delta")
  .wf_validate_threshold(max_ess_loss, "max_ess_loss")
  .wf_validate_threshold(max_total_shift, "max_total_shift")
  .wf_validate_threshold(max_margin_delta, "max_margin_delta")
  .wf_validate_threshold(max_ratio_p99, "max_ratio_p99", lower = 1)

  new_stats <- .wf_validation_group_stats(new, target)
  ref_stats <- .wf_validation_group_stats(reference, target)
  comparison <- merge(
    new_stats,
    ref_stats,
    by = "group",
    all = TRUE,
    suffixes = c("_new", "_reference")
  )
  comparison$deff_delta <- comparison$deff_new - comparison$deff_reference
  comparison$ess_loss <- ifelse(
    is.finite(comparison$ess_reference) & comparison$ess_reference > 0,
    pmax(0, comparison$ess_reference - comparison$ess_new) /
      comparison$ess_reference,
    NA_real_
  )
  comparison$total_shift <- ifelse(
    is.finite(comparison$total_weight_reference) &
      comparison$total_weight_reference > 0,
    abs(comparison$total_weight_new - comparison$total_weight_reference) /
      comparison$total_weight_reference,
    NA_real_
  )
  if (all(c("margin_maxerr_new", "margin_maxerr_reference") %in%
      names(comparison))) {
    comparison$margin_delta <-
      comparison$margin_maxerr_new - comparison$margin_maxerr_reference
  }

  issue_rows <- list()
  add_issue <- function(check, group, severity, value, threshold, detail) {
    issue_rows[[length(issue_rows) + 1L]] <<- data.frame(
      check = check,
      group = group,
      severity = severity,
      value = value,
      threshold = threshold,
      detail = detail,
      stringsAsFactors = FALSE
    )
  }

  missing_new <- comparison$group[is.na(comparison$n_new)]
  for (group in missing_new) {
    add_issue(
      "group_missing",
      group,
      "error",
      NA_real_,
      NA_real_,
      "Reference group is absent from the new weights."
    )
  }
  missing_ref <- comparison$group[is.na(comparison$n_reference)]
  for (group in missing_ref) {
    add_issue(
      "group_new",
      group,
      "warning",
      NA_real_,
      NA_real_,
      "New group is absent from the reference weights."
    )
  }

  hit <- is.finite(comparison$deff_delta) &
    comparison$deff_delta > max_deff_delta
  for (i in which(hit)) {
    add_issue(
      "deff_delta",
      comparison$group[[i]],
      "warning",
      comparison$deff_delta[[i]],
      max_deff_delta,
      "Design effect increased beyond the declared threshold."
    )
  }
  hit <- is.finite(comparison$ess_loss) &
    comparison$ess_loss > max_ess_loss
  for (i in which(hit)) {
    add_issue(
      "ess_loss",
      comparison$group[[i]],
      "warning",
      comparison$ess_loss[[i]],
      max_ess_loss,
      "Effective sample size fell beyond the declared threshold."
    )
  }
  hit <- is.finite(comparison$total_shift) &
    comparison$total_shift > max_total_shift
  for (i in which(hit)) {
    add_issue(
      "total_shift",
      comparison$group[[i]],
      "warning",
      comparison$total_shift[[i]],
      max_total_shift,
      "Group total weight shifted beyond the declared threshold."
    )
  }
  if ("margin_delta" %in% names(comparison)) {
    hit <- is.finite(comparison$margin_delta) &
      comparison$margin_delta > max_margin_delta
    for (i in which(hit)) {
      add_issue(
        "margin_delta",
        comparison$group[[i]],
        "warning",
        comparison$margin_delta[[i]],
        max_margin_delta,
        "Margin residual increased beyond the declared threshold."
      )
    }
  }

  ratio <- data.frame(
    n_common = integer(0),
    p01 = numeric(0),
    median = numeric(0),
    p99 = numeric(0),
    stringsAsFactors = FALSE
  )
  if (!all(c("id", "weight") %in% names(new$data)) ||
      !all(c("id", "weight") %in% names(reference$data))) {
    wf_abort(
      "Both weight objects must contain `id` and `weight` columns for validation.",
      "wf_error_schema"
    )
  }
  new_ids <- .chr(new$data$id)
  ref_ids <- .chr(reference$data$id)
  if (anyDuplicated(new_ids) || anyDuplicated(ref_ids)) {
    wf_abort(
      "Weight validation requires unique unit IDs in both weight objects.",
      "wf_error_schema"
    )
  }
  common_ids <- intersect(new_ids, ref_ids)
  if (length(common_ids) > 0) {
    unit_ratio <- new$data$weight[match(common_ids, new_ids)] /
      reference$data$weight[match(common_ids, ref_ids)]
    unit_ratio <- unit_ratio[is.finite(unit_ratio) & unit_ratio > 0]
    if (length(unit_ratio) > 0) {
      med <- stats::median(unit_ratio)
      if (is.finite(med) && med > 0) {
        unit_ratio <- unit_ratio / med
      }
      q <- stats::quantile(
        unit_ratio,
        c(0.01, 0.5, 0.99),
        names = FALSE,
        type = 7
      )
      ratio <- data.frame(
        n_common = length(unit_ratio),
        p01 = q[[1]],
        median = q[[2]],
        p99 = q[[3]],
        stringsAsFactors = FALSE
      )
      if (q[[3]] > max_ratio_p99 || q[[1]] < 1 / max_ratio_p99) {
        add_issue(
          "unit_weight_ratio",
          "*",
          "warning",
          max(q[[3]], 1 / q[[1]]),
          max_ratio_p99,
          "Per-unit normalized weight ratios drifted beyond the declared threshold."
        )
      }
    }
  }

  issues <- if (length(issue_rows) > 0) {
    do.call(rbind, issue_rows)
  } else {
    data.frame(
      check = character(0),
      group = character(0),
      severity = character(0),
      value = numeric(0),
      threshold = numeric(0),
      detail = character(0),
      stringsAsFactors = FALSE
    )
  }
  out <- structure(
    list(
      ok = nrow(issues) == 0,
      issues = issues,
      comparison = comparison,
      ratio = ratio,
      thresholds = list(
        max_deff_delta = max_deff_delta,
        max_ess_loss = max_ess_loss,
        max_total_shift = max_total_shift,
        max_margin_delta = max_margin_delta,
        max_ratio_p99 = max_ratio_p99
      )
    ),
    class = "wf_validation"
  )
  if (!out$ok && on_issue == "warn") {
    wf_warn(
      sprintf("Weight validation detected %d drift issue(s).", nrow(issues)),
      "wf_warning_quality",
      list(issues = issues)
    )
  }
  if (!out$ok && on_issue == "error") {
    wf_abort(
      sprintf("Weight validation detected %d drift issue(s).", nrow(issues)),
      "wf_error_input",
      list(issues = issues)
    )
  }
  out
}

#' Print a weight-validation result
#'
#' @param x A `wf_validation` object.
#' @param ... Unused.
#' @return Invisibly returns `x`.
#' @export
print.wf_validation <- function(x, ...) {
  cat(sprintf(
    "<wf_validation> %s\n",
    if (x$ok) "PASS" else "DRIFT DETECTED"
  ))
  if (nrow(x$issues) > 0) {
    show <- utils::head(x$issues, 12)
    for (i in seq_len(nrow(show))) {
      cat(sprintf(
        "  [%s] %s / %s: %s\n",
        show$severity[[i]],
        show$check[[i]],
        show$group[[i]],
        show$detail[[i]]
      ))
    }
    if (nrow(x$issues) > nrow(show)) {
      cat(sprintf("  ... and %d more issue(s)\n", nrow(x$issues) - nrow(show)))
    }
  }
  invisible(x)
}

#' Escape one string for JSON.
#'
#' @param x Character scalar.
#' @keywords internal
#' @noRd
.wf_json_escape <- function(x) {
  x <- gsub("\\", "\\\\", x, fixed = TRUE)
  x <- gsub("\"", "\\\"", x, fixed = TRUE)
  x <- gsub("\n", "\\n", x, fixed = TRUE)
  x <- gsub("\r", "\\r", x, fixed = TRUE)
  x <- gsub("\t", "\\t", x, fixed = TRUE)
  x
}

#' Serialize simple R objects to JSON.
#'
#' @param x Object to serialize.
#' @keywords internal
#' @noRd
.wf_json <- function(x) {
  scalar <- function(value) {
    if (is.na(value)) {
      return("null")
    }
    if (is.logical(value)) {
      return(if (isTRUE(value)) "true" else "false")
    }
    if (is.numeric(value)) {
      if (!is.finite(value)) {
        return(paste0("\"", as.character(value), "\""))
      }
      return(format(value, digits = 15, scientific = FALSE, trim = TRUE))
    }
    paste0("\"", .wf_json_escape(as.character(value)), "\"")
  }

  if (is.null(x)) {
    return("null")
  }
  if (inherits(x, "formula")) {
    return(.wf_json(paste(deparse(x), collapse = " ")))
  }
  if (inherits(x, "POSIXt")) {
    return(.wf_json(.wf_iso_time(x)))
  }
  if (is.call(x) || is.name(x)) {
    return(.wf_json(paste(deparse(x), collapse = " ")))
  }
  if (is.environment(x)) {
    return("\"<environment>\"")
  }
  if (is.function(x)) {
    return("\"<function>\"")
  }
  if (is.factor(x)) {
    x <- as.character(x)
  }
  if (is.data.frame(x)) {
    rows <- lapply(seq_len(nrow(x)), function(i) {
      as.list(x[i, , drop = FALSE])
    })
    return(paste0("[", paste(vapply(rows, .wf_json, ""), collapse = ","), "]"))
  }
  if (is.list(x)) {
    names_x <- names(x)
    if (!is.null(names_x) && all(nzchar(names_x))) {
      values <- vapply(seq_along(x), function(i) {
        paste0("\"", .wf_json_escape(names_x[[i]]), "\":", .wf_json(x[[i]]))
      }, "")
      return(paste0("{", paste(values, collapse = ","), "}"))
    }
    return(paste0("[", paste(vapply(x, .wf_json, ""), collapse = ","), "]"))
  }
  if (length(x) == 0) {
    return("[]")
  }
  names_x <- names(x)
  if (!is.null(names_x) && all(nzchar(names_x))) {
    values <- vapply(seq_along(x), function(i) {
      paste0("\"", .wf_json_escape(names_x[[i]]), "\":", scalar(x[[i]]))
    }, "")
    return(paste0("{", paste(values, collapse = ","), "}"))
  }
  if (length(x) == 1) {
    return(scalar(x))
  }
  paste0("[", paste(vapply(x, scalar, ""), collapse = ","), "]")
}

#' Create an audit payload.
#'
#' @param x Result object.
#' @param inputs Optional named input list.
#' @param extra Optional metadata.
#' @keywords internal
#' @noRd
.wf_audit_payload <- function(x, inputs = NULL, extra = NULL) {
  ledger <- NULL
  diagnostics <- NULL
  object_class <- class(x)
  weights <- NULL
  if (inherits(x, "wf_autoweigh_result")) {
    ledger <- x$ledger
    diagnostics <- x$diagnostics
    weights <- x$weights
  } else if (inherits(x, "wf_weights")) {
    weights <- x
  }

  identities <- list(
    design = NULL,
    target = NULL,
    cell_plan = NULL,
    plan = NULL,
    approval = NULL,
    locked_weight = NULL,
    source = NULL,
    source_data = NULL,
    impact = NULL
  )
  if (inherits(x, "wf_safe_workflow")) {
    if (!identical(x$identity, .wf_safe_workflow_identity(x))) {
      .wf_safety_abort(
        "safe_workflow_invalid",
        "The safe workflow changed after its identity was recorded.",
        "x"
      )
    }
    identities$design <- x$design$identity
    identities$target <- x$target$identity
    identities$cell_plan <- x$cell_plan$identity
    identities$plan <- x$plan$identity
    identities$source <- x$target$evidence$source_checksum
    identities$source_data <- x$target$evidence$data_checksum
  } else if (inherits(x, "wf_locked_weights")) {
    if (!identical(x$identity, .wf_locked_weight_identity(x))) {
      .wf_safety_abort(
        "locked_weights_invalid",
        "The locked weights changed after their identity was recorded.",
        "x"
      )
    }
    identities$design <- x$design_identity
    identities$target <- x$target_identity
    identities$cell_plan <- x$provenance$safety$cell_plan_identity
    identities$plan <- x$plan_identity
    identities$approval <- x$approval_identity
    identities$locked_weight <- x$identity
    identities$source <- x$provenance$safety$source_metadata_checksum
    identities$source_data <- x$provenance$safety$source_data_checksum
  } else if (inherits(x, "wf_impact")) {
    if (!identical(x$identity, .wf_impact_identity(x))) {
      .wf_safety_abort(
        "impact_identity_invalid",
        "The impact result changed after its identity was recorded.",
        "x"
      )
    }
    identities$locked_weight <- x$weight_identity
    identities$impact <- x$identity
  } else if (inherits(x, "wf_plan_approval")) {
    if (!identical(x$identity, .wf_plan_approval_identity(x))) {
      .wf_safety_abort(
        "plan_approval_invalid",
        "The approval changed after its identity was recorded.",
        "x"
      )
    }
    identities$plan <- x$plan_identity
    identities$approval <- x$identity
  } else if (inherits(x, "wf_weight_plan")) {
    if (!identical(x$identity, .wf_weight_plan_identity(x))) {
      .wf_safety_abort(
        "weight_plan_invalid",
        "The weight plan changed after its identity was recorded.",
        "x"
      )
    }
    identities$design <- x$input_identities$design
    identities$target <- x$input_identities$target
    identities$cell_plan <- x$input_identities$cell_plan
    identities$plan <- x$identity
  } else if (inherits(x, "wf_cell_merge_plan")) {
    if (!identical(x$identity, .wf_cell_plan_identity(x))) {
      .wf_safety_abort(
        "cell_plan_invalid",
        "The cell plan changed after its identity was recorded.",
        "x"
      )
    }
    identities$design <- x$input_identities$design
    identities$target <- x$input_identities$target
    identities$cell_plan <- x$identity
  } else if (inherits(x, "wf_verified_target")) {
    if (!identical(x$identity, .wf_verified_target_identity(x))) {
      .wf_safety_abort(
        "target_identity_invalid",
        "The verified target changed after its identity was recorded.",
        "x"
      )
    }
    identities$target <- x$identity
    identities$source <- x$evidence$source_checksum
    identities$source_data <- x$evidence$data_checksum
  } else if (inherits(x, "wf_design_data")) {
    if (!identical(x$identity, .wf_design_identity(x$data, x$roles))) {
      .wf_safety_abort(
        "design_identity_invalid",
        "The design data changed after their identity was recorded.",
        "x"
      )
    }
    identities$design <- x$identity
  }
  identity_names <- c(
    "design", "target", "cell_plan", "plan", "approval",
    "locked_weight", "source", "source_data", "impact"
  )
  for (name in setdiff(identity_names, names(identities))) {
    identities[name] <- list(NULL)
  }
  identities <- identities[identity_names]

  input_hashes <- NULL
  if (!is.null(inputs)) {
    if (!is.list(inputs) || is.null(names(inputs)) ||
        any(!nzchar(names(inputs)))) {
      wf_abort("`inputs` must be NULL or a named list.", "wf_error_input")
    }
    input_hashes <- vapply(inputs, .wf_object_hash, character(1))
  }

  n_units <- if (!is.null(weights) && is.data.frame(weights$data)) {
    nrow(weights$data)
  } else if (inherits(x, "wf_safe_workflow")) {
    nrow(x$design$data)
  } else if (inherits(x, "wf_design_data")) {
    nrow(x$data)
  } else {
    NULL
  }
  groups <- if (!is.null(weights) && is.data.frame(weights$data) &&
      "group" %in% names(weights$data)) {
    unique(.chr(weights$data$group))
  } else if (inherits(x, "wf_safe_workflow")) {
    names(x$target$groups)
  } else if (inherits(x, "wf_verified_target")) {
    names(x$groups)
  } else {
    NULL
  }
  weight_summary <- if (!is.null(weights) && is.data.frame(weights$data) &&
      "weight" %in% names(weights$data)) {
    list(
      total = sum(weights$data$weight),
      min = min(weights$data$weight),
      max = max(weights$data$weight),
      mean = mean(weights$data$weight)
    )
  } else {
    NULL
  }
  provenance <- if (!is.null(weights)) {
    weights$provenance
  } else if (inherits(x, "wf_safe_workflow")) {
    list(
      workflow_identity = x$identity,
      method = x$plan$method,
      package_version = x$package_version
    )
  } else if (!is.null(x$provenance)) {
    x$provenance
  } else {
    NULL
  }

  list(
    schema = "wfc_audit_v2",
    exported = .wf_iso_time(),
    package = list(name = "WFC", version = .wf_package_version()),
    object_class = object_class,
    n_units = n_units,
    groups = groups,
    weight_summary = weight_summary,
    provenance = provenance,
    decision_ledger = ledger,
    diagnostics = diagnostics,
    input_hashes = input_hashes,
    identities = identities,
    extra = extra
  )
}

#' Export a self-contained WFC audit file
#'
#' Writes a JSON audit record containing result provenance, pipeline metadata,
#' optional guided-workflow decision ledger, input hashes, and user-supplied
#' extra metadata. The writer is dependency-free and intended for machine
#' archiving, not human formatting.
#'
#' @param x A WFC result, safe workflow, plan, approval, locked weights, verified
#'   target, design-data object, or post-lock impact object.
#' @param file Output JSON file path.
#' @param inputs Optional named list of input objects whose hashes should be
#'   recorded.
#' @param extra Optional additional metadata to include in the audit record.
#'
#' @return Invisibly returns `file`.
#' @export
wf_audit_export <- function(x, file, inputs = NULL, extra = NULL) {
  supported <- c(
    "wf_weights", "wf_autoweigh_result", "wf_safe_workflow",
    "wf_weight_plan", "wf_cell_merge_plan", "wf_plan_approval",
    "wf_verified_target", "wf_design_data", "wf_impact"
  )
  if (!inherits(x, supported)) {
    wf_abort(
      "`x` must be a supported WFC result or safe-workflow evidence object.",
      "wf_error_input"
    )
  }
  if (!.wf_is_string(file)) {
    wf_abort("`file` must be one non-empty path.", "wf_error_input")
  }
  directory <- dirname(file)
  if (!dir.exists(directory)) {
    wf_abort(
      sprintf("Output directory does not exist: %s", directory),
      "wf_error_input",
      list(directory = directory)
    )
  }
  payload <- .wf_audit_payload(x, inputs = inputs, extra = extra)
  writeLines(.wf_json(payload), file, useBytes = TRUE)
  invisible(file)
}

#' Declare a post-stratification collapse ladder
#'
#' Creates an ordered, cumulative, validated category-collapse ladder for
#' post-stratification cell resolution. Level 0 is the raw classification; each
#' named level applies one additional set of dimension maps.
#'
#' @param dims A `wf_dims` object.
#' @param ... Named ladder levels, each a list of named character vectors in
#'   the form `old = "new"` by dimension.
#'
#' @return A `wf_collapse_ladder` object.
#' @export
#'
#' @examples
#' dims <- wf_dims(age = c("young", "old"), education = c("low", "high"))
#' wf_collapse_ladder(
#'   dims,
#'   level1 = list(age = c(young = "all", old = "all"))
#' )
wf_collapse_ladder <- function(dims, ...) {
  if (!inherits(dims, "wf_dims")) {
    wf_abort("`dims` must be a wf_dims object.", "wf_error_input")
  }
  steps <- list(...)
  if (length(steps) > 0 && (is.null(names(steps)) || any(names(steps) == ""))) {
    wf_abort(
      "Ladder levels must be named (for example, level1 = list(...)).",
      "wf_error_input"
    )
  }

  dvars <- names(dims$vars)
  universe <- lapply(dvars, function(d) dims$vars[[d]])
  names(universe) <- dvars

  for (lv in seq_along(steps)) {
    step <- steps[[lv]]
    bad_d <- setdiff(names(step), dvars)
    if (length(bad_d) > 0) {
      wf_abort(
        sprintf(
          "Ladder level '%s' references unknown dimension(s): %s",
          names(steps)[lv],
          paste(bad_d, collapse = ", ")
        ),
        "wf_error_input",
        list(level = names(steps)[lv], dims = bad_d)
      )
    }

    for (d in names(step)) {
      raw_map <- step[[d]]
      if (is.null(names(raw_map))) {
        wf_abort(
          sprintf(
            "Ladder level '%s', dim '%s': map must be a named vector.",
            names(steps)[lv],
            d
          ),
          "wf_error_input"
        )
      }
      map <- .chr(raw_map)
      names(map) <- .chr(names(raw_map))

      if (!is.null(universe[[d]])) {
        unknown <- setdiff(names(map), universe[[d]])
        if (length(unknown) > 0) {
          wf_abort(
            sprintf(
              "Ladder level '%s', dim '%s': source categories not present at this level: %s",
              names(steps)[lv],
              d,
              paste(unknown, collapse = ", ")
            ),
            "wf_error_input",
            list(level = names(steps)[lv], dim = d, unknown = unknown)
          )
        }
        kept <- setdiff(universe[[d]], names(map))
        universe[[d]] <- unique(c(kept, unname(map)))
      } else {
        universe[[d]] <- unique(unname(map))
      }
      step[[d]] <- map
    }
    steps[[lv]] <- step
  }

  structure(
    list(dims = dvars, steps = steps, n_levels = length(steps)),
    class = "wf_collapse_ladder"
  )
}

# ============================================================================
# WFC Extension 3 -- reference implementation, part 3 of 3 (releases 0.14-0.15)
# Method-family extensions and deeper diagnostics. Base R only.
#
# Prototype code: source AFTER library(WFC).
#
# Contents:
#   1. wf_calibrate_soft() -- penalized (ridge) calibration      [design §7.1]
#   2. wf_calibrate_ebal() -- entropy balancing                  [design §7.2]
#   3. wf_attrition()      -- panel retention weighting          [design §8]
#   4. wf_influence()      -- per-unit leverage diagnostics      [design §8]
#
# k-source wf_blend() is a straightforward generalization of R/blend.R
# (lambda becomes a simplex row per cell) and is left to the implementation
# pass; it needs no new algorithm.
# ============================================================================

# ----------------------------------------------------------------------------
# 0. Shared constraint builder -------------------------------------------------
# Same construction as the packaged linear-calibration engine: one total
# column plus one dummy column per non-reference category of each dimension.
# ----------------------------------------------------------------------------

.wf_soft_build <- function(sub, dvars, gr) {
  n <- nrow(sub)
  cols <- list(rep(1, n))
  t <- gr$total
  labels <- "total"
  for (d in dvars) {
    lev <- names(gr$margins[[d]])
    for (l in lev[-1]) {
      cols[[length(cols) + 1]] <- as.numeric(trimws(as.character(sub[[d]])) == l)
      t <- c(t, gr$margins[[d]][[l]])
      labels <- c(labels, paste0(d, "=", l))
    }
  }
  list(X = do.call(cbind, cols), t = t, labels = labels)
}

# ----------------------------------------------------------------------------
# 1. Soft (penalized / ridge) calibration -------------------------------------
#
# Hard calibration forces X'w = t exactly; when a margin is unsupportable the
# only remedies are collapsing or stopping. Soft calibration minimizes
#
#   (1/2) (w - d)' D^{-1} (w - d)  +  (1/2) (t - X'w)' Phi^{-1} (t - X'w)
#
# with D = diag(d) and Phi = diag(phi_j) >= 0 the per-constraint slack. The
# closed form is
#
#   s = (Phi + X' D X)^{-1} (t - X'd);   w = d + D X s;   gap = Phi s.
#
# phi_j = 0 keeps constraint j exact (always: the group total). For the rest,
# a small outer loop shrinks phi_j until every |gap_j| lies within the user's
# declared tolerance -- so the relaxation is bounded and fully reported.
# ----------------------------------------------------------------------------

wf_calibrate_soft <- function(sample, target, tolerance = 0.02,
                              init_weight = NULL, id = NULL,
                              na = c("drop", "error"), max_outer = 25) {
  na <- match.arg(na)
  stopifnot(inherits(target, "wf_target"))
  dvars <- target$dims

  na_mask <- Reduce(`|`, lapply(dvars, function(d) is.na(sample[[d]])))
  if (any(na_mask)) {
    if (na == "error") stop(sprintf("%d row(s) have NA in calibration dimensions.", sum(na_mask)))
    warning(sprintf("na='drop': removed %d row(s) with NA in calibration dimensions.", sum(na_mask)))
    sample <- sample[!na_mask, , drop = FALSE]
  }

  gkey <- if (is.null(target$by)) rep("_all_", nrow(sample))
          else trimws(as.character(sample[[target$by]]))
  ids <- if (is.null(id)) as.character(seq_len(nrow(sample)))
         else trimws(as.character(sample[[id]]))
  iw <- if (is.null(init_weight)) NULL else as.numeric(sample[[init_weight]])

  res_rows <- list(); logs <- list(); relax <- list()
  for (g in intersect(names(target$groups), unique(gkey))) {
    sel <- which(gkey == g)
    gr <- target$groups[[g]]
    built <- .wf_soft_build(sample[sel, , drop = FALSE], dvars, gr)
    X <- built$X; t <- built$t
    d <- if (is.null(iw)) rep(gr$total / length(sel), length(sel)) else iw[sel]

    p <- ncol(X)
    tol_abs <- c(0, rep(tolerance * gr$total, p - 1))   # total is exact
    phi <- c(0, rep((tolerance * gr$total)^2, p - 1))   # any positive start
    XtDX <- crossprod(X, X * d)
    r0 <- t - as.numeric(crossprod(X, d))

    w <- d; gap <- r0
    for (it in seq_len(max_outer)) {
      M <- XtDX + diag(phi, p)
      s <- tryCatch(solve(M, r0), error = function(e) NULL)
      if (is.null(s)) stop(sprintf("Group '%s': singular soft-calibration system.", g))
      w <- d + d * as.numeric(X %*% s)
      gap <- phi * s
      over <- which(abs(gap) > tol_abs & tol_abs > 0)
      if (length(over) == 0) break
      phi[over] <- phi[over] * (tol_abs[over] / abs(gap[over])) * 0.5
    }
    if (any(w <= 0)) {
      warning(sprintf(
        "Group '%s': %d non-positive weight(s) from the linear soft distance; production code should switch to a bounded distance here.",
        g, sum(w <= 0)))
    }

    res_rows[[g]] <- data.frame(id = ids[sel], group = g, weight = w,
                                feature = 1 / w, stringsAsFactors = FALSE)
    relax[[g]] <- data.frame(
      group = g, constraint = built$labels, target = t,
      achieved = as.numeric(crossprod(X, w)),
      gap = gap, tolerance = tol_abs, relaxed = abs(gap) > 1e-9,
      stringsAsFactors = FALSE
    )
    logs[[g]] <- data.frame(group = g, n = length(sel),
                            iterations = it, converged = TRUE,
                            n_relaxed = sum(abs(gap) > 1e-9),
                            max_rel_gap = max(abs(gap)) / gr$total,
                            stringsAsFactors = FALSE)
  }

  structure(list(
    data = do.call(rbind, res_rows),
    log = do.call(rbind, logs),
    achieved = NULL,
    relaxation = do.call(rbind, relax),   # the soft-calibration audit table
    provenance = list(method = "soft", tolerance = tolerance,
                      dims = dvars, by = target$by, created = Sys.time())
  ), class = "wf_weights")
}

# ----------------------------------------------------------------------------
# 2. Entropy balancing (exponential-distance calibration) ---------------------
#
# Solves sum_i d_i exp(x_i'lambda) x_i = t by damped Newton: the calibrated
# weights w = d exp(X lambda) minimize KL(w || d) subject to the constraints.
# Accepts the standard wf_target margins plus optional continuous `moments`
# (named vector of target MEANS for numeric sample columns) -- the case plain
# raking cannot handle.
# ----------------------------------------------------------------------------

wf_calibrate_ebal <- function(sample, target, moments = NULL,
                              id = NULL, tol = 1e-8, max_iter = 100) {
  stopifnot(inherits(target, "wf_target"))
  dvars <- target$dims
  gkey <- if (is.null(target$by)) rep("_all_", nrow(sample))
          else trimws(as.character(sample[[target$by]]))
  ids <- if (is.null(id)) as.character(seq_len(nrow(sample)))
         else trimws(as.character(sample[[id]]))

  res_rows <- list(); logs <- list()
  for (g in intersect(names(target$groups), unique(gkey))) {
    sel <- which(gkey == g)
    gr <- target$groups[[g]]
    sub <- sample[sel, , drop = FALSE]
    built <- .wf_soft_build(sub, dvars, gr)
    X <- built$X; t <- built$t
    if (!is.null(moments)) {
      for (mv in names(moments)) {
        X <- cbind(X, as.numeric(sub[[mv]]))
        t <- c(t, moments[[mv]] * gr$total)   # target mean -> target total
      }
    }
    d <- rep(gr$total / length(sel), length(sel))

    lambda <- rep(0, ncol(X))
    w <- d
    maxr <- Inf; it <- 0L
    while (it < max_iter) {
      resid <- t - as.numeric(crossprod(X, w))
      maxr <- max(abs(resid)) / gr$total
      if (maxr < tol) break
      it <- it + 1L
      jac <- crossprod(X, X * w)
      step <- tryCatch(solve(jac, resid), error = function(e) NULL)
      if (is.null(step)) stop(sprintf("Group '%s': singular entropy-balancing system.", g))
      alpha <- 1
      repeat {
        w_try <- d * exp(as.numeric(X %*% (lambda + alpha * step)))
        r_try <- max(abs(t - as.numeric(crossprod(X, w_try)))) / gr$total
        if (all(is.finite(w_try)) && is.finite(r_try) && r_try < maxr) break
        alpha <- alpha / 2
        if (alpha < 1e-10) break
      }
      lambda <- lambda + alpha * step
      w <- d * exp(as.numeric(X %*% lambda))
    }
    if (maxr >= tol) {
      stop(sprintf("Group '%s': entropy balancing did not converge (residual %.3g). The moment targets may be outside the sample's convex hull.",
                   g, maxr))
    }
    res_rows[[g]] <- data.frame(id = ids[sel], group = g, weight = w,
                                feature = 1 / w, stringsAsFactors = FALSE)
    logs[[g]] <- data.frame(group = g, n = length(sel), iterations = it,
                            converged = TRUE, max_resid = maxr,
                            kl_divergence = sum(w * log(w / d)) / sum(w),
                            stringsAsFactors = FALSE)
  }

  structure(list(
    data = do.call(rbind, res_rows),
    log = do.call(rbind, logs),
    achieved = NULL,
    provenance = list(method = "ebal", moments = names(moments),
                      dims = dvars, by = target$by, created = Sys.time())
  ), class = "wf_weights")
}

# ----------------------------------------------------------------------------
# 3. wf_attrition() -- panel retention weighting -------------------------------
#
# Same discipline as wf_propensity(): model wave-to-wave retention with a
# base-R glm, invert the fitted retention probability into a stage-1 weight
# for the retained units (stabilized by default), and ship balance
# diagnostics comparing the reweighted retainees against the full prior wave.
# Chains across waves through wf_compose().
# ----------------------------------------------------------------------------

wf_attrition <- function(panel, retained, formula, id = NULL,
                         stabilize = TRUE, trim = NULL) {
  stopifnot(is.data.frame(panel), retained %in% names(panel))
  keep <- panel[[retained]]
  if (!is.logical(keep)) keep <- as.integer(keep) == 1L
  if (any(is.na(keep))) stop("`retained` must not contain NA.")
  predictors <- all.vars(formula[[3]])
  na_rows <- Reduce(`|`, lapply(predictors, function(p) is.na(panel[[p]])))
  if (any(na_rows)) stop(sprintf("%d row(s) have NA in retention-model predictors.", sum(na_rows)))

  df <- panel
  df$.wf_retained <- as.integer(keep)
  fit <- stats::glm(stats::reformulate(predictors, ".wf_retained"),
                    family = stats::binomial(), data = df)
  phat <- stats::fitted(fit)

  raw <- 1 / phat[keep]
  if (stabilize) raw <- mean(keep) * raw
  trimmed <- 0L
  if (!is.null(trim)) {
    cap <- trim * stats::median(raw)
    trimmed <- sum(raw > cap)
    raw[raw > cap] <- cap
  }
  w <- raw / mean(raw)

  ids <- if (is.null(id)) as.character(which(keep))
         else trimws(as.character(panel[[id]][keep]))

  # balance: reweighted retainees vs the full prior wave
  balance <- do.call(rbind, lapply(predictors, function(p) {
    x <- panel[[p]]
    if (!is.numeric(x)) x <- as.numeric(as.factor(x))
    sd_pool <- stats::sd(x)
    data.frame(
      variable = p,
      smd_unweighted = (mean(x[keep]) - mean(x)) / sd_pool,
      smd_weighted = (stats::weighted.mean(x[keep], w) - mean(x)) / sd_pool,
      stringsAsFactors = FALSE
    )
  }))

  structure(list(
    data = data.frame(id = ids, group = "_panel_", weight = w,
                      feature = 1 / w, stringsAsFactors = FALSE),
    log = data.frame(group = "_panel_", n = sum(keep),
                     n_lost = sum(!keep), retention = mean(keep),
                     trimmed = trimmed, stringsAsFactors = FALSE),
    achieved = NULL,
    balance = balance,
    provenance = list(method = "attrition", predictors = predictors,
                      stabilize = stabilize, trim = trim,
                      created = Sys.time())
  ), class = "wf_weights")
}

# ----------------------------------------------------------------------------
# 4. wf_influence() -- per-unit leverage diagnostics ---------------------------
#
# The practical question behind most trimming debates: WHICH respondents
# drive the extreme weights? Reports, per unit: the weight ratio to the group
# mean, the unit's share of the deff mass (w_i^2 / sum w^2), and -- for the
# `top` most extreme units -- the group design effect recomputed without the
# unit (leave-one-out).
# ----------------------------------------------------------------------------

wf_influence <- function(w, top = 20) {
  stopifnot(inherits(w, "wf_weights"))
  parts <- split(seq_len(nrow(w$data)), w$data$group)
  rows <- lapply(names(parts), function(g) {
    idx <- parts[[g]]
    wt <- w$data$weight[idx]
    data.frame(
      id = w$data$id[idx],
      group = g,
      weight = wt,
      ratio_to_mean = wt / mean(wt),
      deff_share = wt^2 / sum(wt^2),
      stringsAsFactors = FALSE
    )
  })
  tab <- do.call(rbind, rows)
  tab <- tab[order(-tab$ratio_to_mean), ]
  rownames(tab) <- NULL

  n_top <- min(top, nrow(tab))
  tab$deff_loo <- NA_real_
  for (k in seq_len(n_top)) {
    g <- tab$group[k]
    wt <- w$data$weight[w$data$group == g]
    # leave-one-out: drop one instance of this unit's weight
    wt_loo <- wt[-match(tab$weight[k], wt)]
    tab$deff_loo[k] <- 1 + (stats::sd(wt_loo) / mean(wt_loo))^2
  }
  gdeff <- vapply(names(parts), function(g) {
    wt <- w$data$weight[w$data$group == g]
    1 + (stats::sd(wt) / mean(wt))^2
  }, numeric(1))
  tab$deff_group <- gdeff[tab$group]
  tab$deff_drop_if_removed <- tab$deff_group - tab$deff_loo

  structure(list(table = tab, top = n_top), class = "wf_influence")
}

print.wf_influence <- function(x, ...) {
  cat(sprintf("<wf_influence>  top %d most extreme unit(s):\n", x$top))
  cols <- c("id", "group", "weight", "ratio_to_mean", "deff_share",
            "deff_group", "deff_loo")
  print(utils::head(x$table[, cols], x$top), row.names = FALSE, digits = 3)
  invisible(x)
}

make_attrition_panel <- function() {
  data.frame(
    id = paste0("p", 1:12),
    region = rep(c("A", "B"), each = 6),
    retained = c(TRUE, TRUE, TRUE, FALSE, FALSE, FALSE,
                 TRUE, TRUE, FALSE, TRUE, FALSE, FALSE),
    age = c(25, 35, 45, 55, 65, 75, 24, 34, 44, 54, 64, 74),
    sex = rep(c("female", "male"), 6),
    stringsAsFactors = FALSE
  )
}

test_that("wf_attrition returns retained-unit weights and balance diagnostics", {
  panel <- make_attrition_panel()

  attrition <- suppressWarnings(wf_attrition(
    panel,
    retained = "retained",
    formula = ~ age + sex,
    id = "id",
    by = "region",
    trim = 3
  ))

  expect_s3_class(attrition, "wf_attrition_weights")
  expect_s3_class(attrition, "wf_weights")
  expect_true("wf_attrition" %in% getNamespaceExports("WFC"))
  expect_equal(nrow(attrition$data), sum(panel$retained))
  expect_true(all(attrition$data$id %in% panel$id[panel$retained]))
  expect_true(all(attrition$data$weight > 0))
  group_mean <- tapply(attrition$data$weight, attrition$data$group, mean)
  expect_equal(as.numeric(group_mean), c(1, 1), tolerance = 1e-8)
  expect_equal(names(group_mean), c("A", "B"))
  expect_true(all(c(
    "group", "variable", "smd_unweighted", "smd_weighted"
  ) %in% names(attrition$balance)))
  expect_true(all(c("attrition_balance", "attrition_retention") %in%
    names(wf_report(attrition)$sections)))
  expect_output(print(attrition), "wf_attrition_weights", fixed = TRUE)
})

test_that("wf_attrition validates retention model inputs", {
  panel <- make_attrition_panel()

  expect_error(
    wf_attrition(panel, retained = "missing", formula = ~ age, id = "id"),
    class = "wf_error_schema"
  )
  expect_error(
    wf_attrition(panel, retained = "retained", formula = ~ missing, id = "id"),
    class = "wf_error_schema"
  )

  bad <- panel
  bad$age[[1]] <- NA
  expect_error(
    wf_attrition(bad, retained = "retained", formula = ~ age, id = "id"),
    class = "wf_error_input"
  )

  one_source <- panel
  one_source$retained[one_source$region == "A"] <- TRUE
  expect_error(
    wf_attrition(
      one_source,
      retained = "retained",
      formula = ~ age,
      id = "id",
      by = "region"
    ),
    class = "wf_error_overlap"
  )
})

test_that("wf_attrition composes with later weighting stages", {
  panel <- make_attrition_panel()
  attrition <- suppressWarnings(wf_attrition(
    panel,
    retained = "retained",
    formula = ~ age + sex,
    id = "id"
  ))
  retained <- panel[panel$retained, , drop = FALSE]
  stage2 <- attrition
  stage2$data$weight <- rep(2, nrow(stage2$data))
  stage2$data$feature <- 1 / stage2$data$weight
  stage2$provenance$method <- "constant_adjustment"

  composed <- wf_compose(attrition = attrition, adjustment = stage2)

  expect_s3_class(composed, "wf_weights")
  expect_equal(nrow(composed$data), nrow(retained))
  expect_true(all(composed$data$weight > 0))
})

test_that("wf_influence ranks high-impact units and adds margin shares", {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(
    fixture$sample,
    fixture$target,
    id = "id",
    tol = 1e-8
  )
  weights$data$weight[weights$data$id == "r01"] <-
    weights$data$weight[weights$data$id == "r01"] * 5

  influence <- wf_influence(
    weights,
    target = fixture$target,
    sample = fixture$sample,
    id = "id",
    top = 3
  )

  expect_s3_class(influence, "wf_influence")
  expect_true("wf_influence" %in% getNamespaceExports("WFC"))
  expect_equal(influence$top, 3)
  expect_equal(influence$table$id[[1]], "r01")
  expect_true(all(c(
    "ratio_to_mean", "deff_share", "deff_loo",
    "deff_drop_if_removed", "max_margin_share"
  ) %in% names(influence$table)))
  expect_true(is.finite(influence$table$deff_loo[[1]]))
  expect_true(is.finite(influence$table$max_margin_share[[1]]))
  expect_output(print(influence), "wf_influence", fixed = TRUE)
})

test_that("wf_influence validates IDs and top count", {
  fixture <- make_weightflow_fixture()
  weights <- .wf_rake_engine(fixture$sample, fixture$target, id = "id")

  expect_error(wf_influence(list()), class = "wf_error_input")
  expect_error(wf_influence(weights, top = 0), class = "wf_error_input")

  dup <- weights
  dup$data$id[[2]] <- dup$data$id[[1]]
  expect_error(wf_influence(dup), class = "wf_error_schema")

  expect_error(
    wf_influence(weights, target = fixture$target, sample = fixture$sample),
    class = "wf_error_input"
  )
})

#' Target from external population data
#'
#' Converts arbitrary population data to a canonical `wf_target` object.
#'
#' @param pop Population data frame.
#' @param key_map Named character vector mapping `wf_dims` dimensions to population columns.
#' @param count Population count column.
#' @param dims A `wf_dims` object.
#' @param by Optional grouping variable.
#' @param by_key Optional group key column name or function.
#' @param scale Target scale, one of `"population"`, `"sample"`, or `"custom"`.
#' @param sample Required when `scale = "sample"`.
#' @param totals Required when `scale = "custom"`.
#' @param keep_joint Whether to retain per-group joint population cells.
#'
#' @return A `wf_target` object.
#' @export
#'
#' @examples
#' dims <- wf_dims(gender = c("female", "male"))
#' pop <- data.frame(gender = c("female", "male"), count = c(55, 45))
#' wf_target_population(pop, c(gender = "gender"), "count", dims)
wf_target_population <- function(pop, key_map, count, dims,
                                 by = NULL, by_key = NULL,
                                 scale = c("population", "sample", "custom"),
                                 sample = NULL, totals = NULL,
                                 keep_joint = FALSE) {
  scale <- match.arg(scale)
  if (!is.data.frame(pop)) {
    wf_abort("`pop` must be a data.frame.", "wf_error_input")
  }
  dvars <- names(dims$vars)
  if (is.null(names(key_map)) || !setequal(names(key_map), dvars)) {
    wf_abort(
      "`key_map` must be a named vector mapping every wf_dims dimension to a population column.",
      "wf_error_schema",
      list(expected = dvars, got = names(key_map))
    )
  }
  .require_cols(pop, unname(key_map), "population data")
  .require_cols(pop, count, "population data")

  cnt <- pop[[count]]
  if (any(is.na(cnt))) {
    wf_abort(sprintf(
      "Population count column '%s' has %d NA value(s).",
      count,
      sum(is.na(cnt))
    ), "wf_error_input")
  }
  if (any(cnt < 0)) {
    wf_abort(sprintf(
      "Population count column '%s' has negative values.",
      count
    ), "wf_error_input")
  }

  gkey <- .wf_group_keys(pop, by, by_key)
  groups <- list()
  joint <- if (keep_joint) list() else NULL
  for (g in sort(unique(gkey))) {
    sel <- gkey == g
    margins <- list()
    for (d in dvars) {
      v <- .chr(pop[[key_map[[d]]]][sel])
      m <- tapply(cnt[sel], v, sum)
      margins[[d]] <- .wf_margin_vector(m[m > 0])
    }
    groups[[g]] <- list(total = sum(cnt[sel]), margins = margins)

    if (keep_joint) {
      j <- data.frame(
        lapply(dvars, function(d) .chr(pop[[key_map[[d]]]][sel])),
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
      names(j) <- dvars
      j$pop <- cnt[sel]
      j <- stats::aggregate(j["pop"], j[dvars], sum)
      joint[[g]] <- j[j$pop > 0, c(dvars, "pop"), drop = FALSE]
    }
  }

  sample_n <- NULL
  if (scale == "sample") {
    if (is.null(sample)) {
      wf_abort("scale='sample' requires `sample`.", "wf_error_input")
    }
    sample_n <- table(.wf_group_keys(sample, by))
    sample_n <- stats::setNames(as.numeric(sample_n), names(sample_n))
  }
  groups <- .wf_scale_groups(groups, scale, sample_n, totals)
  if (keep_joint) {
    for (g in names(joint)) {
      f <- groups[[g]]$total / sum(joint[[g]]$pop)
      joint[[g]]$pop <- joint[[g]]$pop * f
    }
  }

  .wf_new_target(
    "population",
    by,
    dvars,
    groups,
    meta = list(scale = scale, created = Sys.time()),
    joint = joint
  )
}

#' Target from a weighted reference sample
#'
#' Builds a canonical `wf_target` from a reference sample where `feature` is the
#' reciprocal of the design weight.
#'
#' @param ref Reference sample data frame.
#' @param feature Feature value column equal to `1 / design_weight`.
#' @param dims A `wf_dims` object.
#' @param by Optional grouping variable.
#' @param feature_na Policy for missing feature values.
#' @param feature_gt1 Policy for feature values greater than one.
#'
#' @return A `wf_target` object.
#' @export
#'
#' @examples
#' dims <- wf_dims(gender = c("female", "male"))
#' ref <- data.frame(gender = c("female", "male"), feature = c(0.5, 0.25))
#' wf_target_reference(ref, "feature", dims)
wf_target_reference <- function(ref, feature, dims, by = NULL,
                                feature_na = c("error", "drop"),
                                feature_gt1 = c("warn", "allow")) {
  feature_na <- match.arg(feature_na)
  feature_gt1 <- match.arg(feature_gt1)
  dvars <- names(dims$vars)
  .require_cols(ref, c(dvars, feature, by), "reference sample")

  f <- ref[[feature]]
  bad_na <- which(is.na(f))
  if (length(bad_na) > 0) {
    msg <- sprintf(
      "%d of %d reference rows (%.1f%%) have missing feature values.",
      length(bad_na),
      nrow(ref),
      100 * length(bad_na) / nrow(ref)
    )
    if (feature_na == "error") {
      wf_abort(
        paste(msg, "Use feature_na='drop' to exclude them from target construction."),
        "wf_error_input",
        list(rows = bad_na)
      )
    }
    wf_warn(
      paste(msg, "Dropped from target construction."),
      "wf_warning_data",
      list(rows = bad_na)
    )
    ref <- ref[-bad_na, , drop = FALSE]
    f <- ref[[feature]]
  }
  bad <- which(!is.finite(f) | f <= 0)
  if (length(bad) > 0) {
    wf_abort(sprintf(
      "Feature values must be finite and > 0; %d row(s) violate this (first: %s).",
      length(bad),
      paste(utils::head(bad, 5), collapse = ", ")
    ), "wf_error_input", list(rows = bad))
  }
  if (any(f > 1) && feature_gt1 == "warn") {
    wf_warn(sprintf(
      "%d feature value(s) exceed 1; inclusion probabilities should lie in (0,1]. Is this column a weight rather than its reciprocal?",
      sum(f > 1)
    ), "wf_warning_data")
  }

  w <- 1 / f
  gkey <- .wf_group_keys(ref, by)
  groups <- list()
  for (g in sort(unique(gkey))) {
    sel <- gkey == g
    margins <- lapply(dvars, function(d) {
      v <- .chr(ref[[d]][sel])
      if (any(is.na(ref[[d]][sel]))) {
        wf_abort(sprintf(
          "Reference sample: NA in dimension '%s' (group '%s'). Clean or drop these rows before target construction.",
          d,
          g
        ), "wf_error_input", list(group = g, dim = d))
      }
      m <- tapply(w[sel], v, sum)
      .wf_margin_vector(m[m > 0])
    })
    names(margins) <- dvars
    groups[[g]] <- list(total = sum(w[sel]), margins = margins)
  }
  .wf_new_target("reference", by, dvars, groups, meta = list(created = Sys.time()))
}

test_that("WFC 2.0 does not export subjective target constructors", {
  exports <- getNamespaceExports("WFC")

  expect_false("wf_target_manual" %in% exports)
  expect_false("wf_target_shrink" %in% exports)
})

test_that("migration guidance gives removed target behavior no workaround", {
  path <- if (file.exists("DESCRIPTION")) {
    "docs/migration/wfc-1-to-2.md"
  } else {
    file.path("..", "..", "docs/migration/wfc-1-to-2.md")
  }
  skip_if_not(file.exists(path), "migration guide is not installed")
  guidance <- readLines(path, warn = FALSE)

  expect_true(any(grepl("wf_target_manual()", guidance, fixed = TRUE)))
  expect_true(any(grepl("wf_target_shrink()", guidance, fixed = TRUE)))
  expect_true(any(grepl("No supported replacement", guidance, fixed = TRUE)))
})

test_that("WFC 2.0 adversarial validation evidence is complete", {
  root <- normalizePath(test_path("..", ".."), mustWork = FALSE)
  script <- file.path(root, "tests", "validation", "validate-wfc-2-safety.R")
  report <- file.path(root, "docs", "validation", "wfc-2.0-safety-validation.md")

  expect_true(file.exists(script))
  skip_if_not(file.exists(report), "validation report is not installed")

  evidence <- paste(
    c(readLines(script, warn = FALSE), readLines(report, warn = FALSE)),
    collapse = "\n"
  )
  required_scenarios <- c(
    "manual_margins",
    "pass_rate_target",
    "outcome_mean",
    "outcome_interval",
    "unverified_target",
    "demo_target",
    "stale_approval",
    "changed_design",
    "agent_approval",
    "raw_public_engine_call"
  )
  required_codes <- c(
    "manual_pipeline_unsupported",
    "inline_moments_unsupported",
    "verified_weighting_inputs_required",
    "demo_target_in_production",
    "plan_approval_mismatch",
    "execution_design_mismatch",
    "human_approval_required"
  )

  expect_true(all(vapply(
    required_scenarios,
    grepl,
    logical(1),
    x = evidence,
    fixed = TRUE
  )))
  expect_true(all(vapply(
    required_codes,
    grepl,
    logical(1),
    x = evidence,
    fixed = TRUE
  )))
  expect_match(evidence, "engine_calls_before_refusal", fixed = TRUE)
  expect_match(evidence, "does not prove source truth", fixed = TRUE)
  expect_match(evidence, "qualified human statistical review", fixed = TRUE)
  release_evidence <- c(
    "WFC 2.0 accountable release approval: **completed**",
    "WFC 2.0 qualified human statistical review: **completed**",
    "Reviewer: Jinyan Zhu",
    "Reviewer qualification: statistical expert",
    "Review completion date: 2026-07-15",
    "Review scope: all WFC 2.0 functionality",
    "Approver: Kunxiang Ma",
    "Approver role: company legal representative",
    "Package release authorized: **yes**"
  )
  expect_true(all(vapply(
    release_evidence,
    grepl,
    logical(1),
    x = evidence,
    fixed = TRUE
  )))
})

root_file <- function(path) {
  candidates <- c(path, file.path("..", "..", path))
  existing <- candidates[file.exists(candidates)]
  if (length(existing) == 0) {
    testthat::skip(sprintf("%s is only available in the development checkout", path))
  }
  normalizePath(existing[[1]], mustWork = TRUE)
}

test_that("git ignore protects private source data formats", {
  gitignore <- readLines(root_file(".gitignore"), warn = FALSE)

  expect_true("*.xlsx" %in% gitignore)
  expect_true("*.RData" %in% gitignore)
  expect_true("private-data/" %in% gitignore)
})

test_that("R build ignore excludes development-only local files", {
  rbuildignore <- readLines(root_file(".Rbuildignore"), warn = FALSE)

  expect_true("^private-data$" %in% rbuildignore)
  expect_true("^data-raw$" %in% rbuildignore)
  expect_true("^\\.codegraph$" %in% rbuildignore)
})

---
title: "Using locked WFC weights with survey tools"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using locked WFC weights with survey tools}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

Create and lock weights through the verified workflow before converting them
to another survey-analysis representation.

```{r conversion, eval=FALSE}
analysis_ready <- wf_attach_weights(
  analysis_data,
  locked,
  id = "person_id",
  weight_name = ".weight"
)

design_survey <- as_svydesign(
  locked,
  analysis_ready,
  id = "person_id"
)
```

Replicate weights can be created from an approved refit design and converted
with `as_svrepdesign()`. Keep target identities, refit settings, seeds, and
variance assumptions in the audit record.

Interoperability does not relax WFC's input boundary. Do not use a downstream
tool to recreate a removed target behavior and then present the result as a WFC
verified run.

---
title: "Dirty data and safe infeasibility handling"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Dirty data and safe infeasibility handling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

WFC stops early when design roles, source evidence, checksums, categories, or
sample support are inconsistent. A stop is evidence to repair or review the
inputs, not permission to choose a more convenient target.

Common safe responses are:

- remove outcome or undeclared columns before `wf_prepare_design()`;
- correct a target file and update its evidence checksum;
- reconcile category spelling against `wf_dims()`;
- document a deterministic support-based merge with `wf_plan_cells()`; or
- conclude that the requested calibration is not feasible.

```{r checks, eval=FALSE}
design <- wf_prepare_design(
  design_only,
  id = "person_id",
  calibration = c("age_group", "region"),
  base_weight = "base_weight"
)

target <- wf_import_target(
  "population-margins.csv",
  "population-margins.csv.source.dcf",
  dims,
  key_map = c(age_group = "age_group", region = "region"),
  count = "population_count"
)

precheck <- wf_precheck(design$data, target, id = design$roles$id)
cell_plan <- wf_plan_cells(design, target, dims)
```

WFC never widens bounds, changes method, changes target, or approves a plan in
response to infeasibility. Those are separate human decisions and may require a
new workflow.

---
title: "Verified calibration methods in WFC 2.0"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Verified calibration methods in WFC 2.0}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

All public calibration methods in WFC 2.0 require the `wf_design_data` and
`wf_verified_target` objects created in the verified workflow. The examples
below assume `design`, `target`, and `dims` have already passed import and
review as shown in `vignette("safe-weighting-workflow")`.

## Fixed-tolerance soft calibration

Soft calibration is for a tolerance declared before outcomes are inspected. It
does not search for a target, widen tolerance automatically, or optimize a
study result. The declared tolerance and any relaxed external margin appear in
the result and statistical report.

```{r soft, eval=FALSE}
soft <- wf_calibrate(
  design,
  target,
  method = "soft",
  tolerance = 0.02
)

wf_report(soft, audience = "statistician")$sections$soft_relaxation
```

If the declared tolerance cannot resolve infeasibility, WFC stops. Select a
different external target only through a new, independently reviewed workflow.

## Categorical entropy balancing

Entropy balancing may use only the categorical margins contained in the
verified target.

```{r ebal, eval=FALSE}
ebal <- wf_calibrate(
  design,
  target,
  method = "ebal",
  tol = 1e-10
)

wf_report(ebal, audience = "statistician")
```

WFC 2.0 has no inline desired-mean interface. Continuous outcome goals,
pass-rate goals, and desired intervals are not weighting inputs.

## Bounded logit calibration

Bounds are reviewable settings, not values that WFC expands automatically:

```{r logit, eval=FALSE}
bounded <- wf_calibrate(
  design,
  target,
  method = "logit",
  bounds = c(0.3, 3)
)
```

For a consequential analysis, record why the method and settings are suitable
and include them in the human plan review.

# WFC — Design Document, Extension 3

**Audience repositioning, guided workflow, and the 0.10 → 1.0 roadmap**

Companion to `weightflow_design.md` (core, 0.1), `weightflow_poststrat_design.md` (Extension 1), and `weightflow_roadmap_design.md` (Extension 2). Those documents were written under the package's original name `weightflow`; the package was renamed **WFC** at 0.9.0 after a CRAN name collision. All architectural commitments in the earlier documents remain in force; this document extends them, it does not revise them.

This extension does two things. First, it repositions the package around its actual future audience. Second, it commits the release sequence from 0.10 through 1.0, absorbing the items Extension 2 deferred (`wf_auto_trim()`, `wf_suggest_ladder()`, `wf_report()`, parallel execution) and adding six new directions agreed in project review: usability layers, ecosystem interoperability, production pipeline infrastructure, a widened method family, deeper diagnostics, and publication infrastructure.

---

## 1. Repositioning: the non-specialist survey manager

The core documents implicitly assume a user who knows what raking, margins, and design effects are. The package's future primary audience is different: **survey managers and field-operations staff who are responsible for producing weighted results but are not statisticians**. They can run a few lines of R from a template, or use a point-and-click interface; they cannot debug a `wf_error_feasibility` from first principles.

This repositioning adds three design principles to the original P1–P6. None of them weakens the originals — in particular P1 (discipline over convenience) is preserved by construction, as §4 explains.

**P7 — Explain, don't just report.** Every automated decision and every refusal must be expressible in plain language that states *what happened, why, and what to do next*, at a level a non-statistician can act on. The classed-condition system (P5) already carries the machine-readable payload; P7 adds the human-readable rendering layer on top of it.

**P8 — One engine, two paths.** The guided path (`wf_autoweigh()`, WFCstudio) is a thin orchestration over the exact same constructors, prechecks, engines, and diagnostics the expert path uses. It never forks the computation, never relaxes a check, and every automated choice it makes is the same reviewable artifact (`wf_collapse_plan`, trim recommendation, ladder draft) an expert would produce by hand. Guided mode is a driver, not a dialect.

**P9 — Localized outputs, English internals.** Code, comments, tests, and R documentation remain English-only (the existing language policy). Human-facing *outputs* — reports, verdict lines, GUI labels — are localizable, shipping with English and Simplified Chinese resource bundles, defaulting to the session locale and switchable via a `lang` argument or `options(wfc.lang = ...)`.

---

## 2. Roadmap at a glance

| Release | Theme | Headline API |
|---|---|---|
| 0.10 | Usability foundations | `wf_report()`, `plot()` methods, `wf_auto_trim()`, `wf_suggest_ladder()` |
| 0.11 | Guided workflow | `wf_autoweigh()`, output localization (`lang`, `wfc.lang`) |
| 0.12 | Ecosystem interoperability | `as_svydesign()`, `as_svrepdesign()`, `tidy()`/`glance()`/`augment()`; WFCstudio sibling package kickoff |
| 0.13 | Production infrastructure | `wf_pipeline()`, `wf_run()`, `wf_validate()`, `wf_audit_export()` |
| 0.14 | Methods I | `wf_calibrate(method = "soft")` (priority), `method = "ebal"` |
| 0.15 | Methods II + diagnostics | `wf_attrition()`, k-source `wf_blend()`, `wf_influence()`, sensitivity bounds |
| 0.16 | Performance engineering | `parallel = TRUE`, progress reporting, optional Rcpp kernel |
| 1.0 | Freeze and publish | API freeze, CRAN submission, bilingual pkgdown site, WFCstudio beta |

Items retained from Extension 2 and not listed above keep their original designs and land opportunistically: MRP (`method = "mrp"`, `redistribute = "model"`) after 1.0; `rf`/`gbm` propensity backends and `kernel`/`matching` weights when a production need arises; Fay's BRR alongside 0.15's variance work.

---

## 3. Release 0.10 — Usability foundations

### 3.1 `wf_report()` — the weighting dossier, re-aimed

Extension 2 §5.4 specified `wf_report()` as a methods-appendix generator. It is re-aimed at two audiences via one argument:

```r
wf_report(w, target = NULL, audience = c("manager", "analyst"),
          lang = NULL, output = c("object", "html", "markdown"), file = NULL)
```

`audience = "analyst"` renders the Extension 2 dossier (deff, ESS, residuals, λ sensitivity, balance/overlap). `audience = "manager"` renders a decision-oriented summary: a traffic-light quality verdict per group, which groups are publishable separately and which only in aggregate, what was merged or trimmed and why, in plain language (P7). Both renderings read the same underlying structured object; the manager view is a projection, not a separate computation. `output = "html"` uses only base tooling (no rmarkdown hard dependency; an enhanced theme may use rmarkdown behind `Suggests:`).

### 3.2 `plot()` methods

Base-graphics methods, zero new dependencies: `plot(wf_weights)` (per-group weight distributions with trim bounds), `plot(wf_diagnostics)` (deff/ESS dot chart), `plot(wf_blend_result)` (λ sensitivity curve), `plot(wf_propensity result)` (overlap densities and a balance love plot).

### 3.3 `wf_auto_trim()` and `wf_suggest_ladder()`

Implemented exactly as designed in Extension 2 §5.2–5.3 (bias–variance frontier sweep; information-loss-ordered ladder drafting). Their positioning changes: they are no longer optional conveniences but the machinery `wf_autoweigh()` (0.11) relies on to propose defensible defaults. They therefore land one release before it.

---

## 4. Release 0.11 — Guided workflow

### 4.1 `wf_autoweigh()`

One call that walks the full discipline loop and narrates it:

```r
res <- wf_autoweigh(
  sample, population,                 # or a ready wf_target
  dims,                               # wf_dims; ladders strongly recommended
  by = NULL, id = NULL,
  method = c("auto", "raking", "poststrat", "logit"),
  interactive = base::interactive(),  # ask before applying remediations
  lang = NULL
)
```

Internally: build target (if given raw population data) → `wf_precheck()` → if blocked, generate remediation proposals via `wf_suggest_collapse()` / `wf_suggest_ladder()` and either ask (interactive) or apply-and-log (non-interactive, only remediations that are *pre-declared* ladders — never invented merges, per Extension 1 §4) → re-precheck → calibrate → `wf_auto_trim()` recommendation → `wf_diagnose()` → `wf_report(audience = "manager")`.

The return value is a `wf_autoweigh_result`: the `wf_weights`, the report, and a **decision ledger** — an ordered record of every automated choice (what, why, which reviewable artifact) so an expert can audit or veto after the fact.

**P1 is preserved by construction.** `wf_autoweigh()` has no power the expert path lacks: it cannot bypass a blocking precheck; when no pre-declared ladder can resolve an infeasibility it stops and explains, in plain language, exactly which category in which group is infeasible and what a ladder that would fix it looks like. Guided mode automates the *proposal* loop, never the *approval* of anything that was not declared in advance.

### 4.2 Output localization

A small internal i18n layer: message catalogs under `inst/i18n/` (`en.dcf`, `zh_CN.dcf`), an internal `.wf_tr(key, ...)` lookup with `sprintf`-style interpolation, resolution order `lang` argument → `options(wfc.lang)` → session locale → English fallback. Scope: report text, verdict lines, `wf_autoweigh()` narration, plot labels. Explicitly out of scope: error/warning `message` fields, R documentation, and code (English-only policy unchanged); classed conditions remain the stable machine interface so localization can never break programmatic handling.

---

## 5. Release 0.12 — Ecosystem interoperability

### 5.1 survey / srvyr bridges

```r
as_svydesign(w, data, ...)            # wf_weights  -> survey::svydesign
as_svrepdesign(r, data, ...)          # wf_replicate_weights -> survey::svrepdesign
```

`as_svrepdesign()` maps the stored `(scale, rscales, method)` onto the corresponding `svrepdesign` types (bootstrap / JK1-JKn / BRR), so downstream `svymean`, `svyglm`, and contingency tables carry WFC's replication-based uncertainty unchanged. Both live behind `Suggests: survey` with a classed error naming the missing package. This closes the largest gap for adoption: today weights exit the package as a bare data frame.

### 5.2 broom-style tidiers

`tidy()`, `glance()`, `augment()` methods for `wf_weights`, `wf_diagnostics`, `wf_blend_result`, and `wf_variance_result` (registered conditionally on `generics`/`broom`). `augment()` joins weights back onto the input sample by id — the single most-asked-for convenience in practice.

### 5.3 WFCstudio kickoff (sibling package)

The GUI is a **separate package**, `WFCstudio`, so the core keeps its zero-hard-dependency promise. Scope for the initial beta: upload sample and population files (csv/xlsx) → map columns to dims/by/count interactively → run `wf_autoweigh()` → view the manager report → download weights and the report. The UI layer contains *no statistics*: it renders WFC's structured objects and classed conditions (the programmatic reaction path P5 anticipated). Bilingual UI via the same i18n catalogs. Development starts at 0.12; beta ships with 1.0.

---

## 6. Release 0.13 — Production pipeline infrastructure

### 6.1 `wf_pipeline()` and `wf_run()`

A declarative, serializable specification of a full weighting chain:

```r
spec <- wf_pipeline(
  target   = list(mode = "population", key_map = ..., count = "pop", by = "prov"),
  stages   = list(
    propensity = list(formula = member ~ age + edu),
    calibrate  = list(method = "logit", bounds = c(0.3, 3))
  ),
  validate = list(max_deff = 6, max_margin_dev = 1e-4)
)
wf_run(spec, sample, population, reference = NULL)
```

The spec is a plain list (JSON/YAML-serializable, hashable into provenance), so a production round is reproducible from a versioned text artifact. `wf_run()` is also the natural `refit` closure for `wf_replicates()`, replacing hand-written closures with a declared object.

### 6.2 `wf_validate()` — weight drift detection

Compares a new `wf_weights` against a reference release: per-group deff and ESS deltas, margin-deviation deltas, distributional shift of weight ratios (per-id where ids match), group coverage changes. Thresholds are declared, violations raise `wf_warning_quality` / `wf_error_input` with structured payloads, and the comparison table feeds `wf_report()`. This is the "did this quarter's weights silently change character" check every recurring survey needs.

### 6.3 `wf_audit_export()`

Serializes any result's provenance (including composed stages and the decision ledger) to a self-contained JSON audit file with input hashes, package version, settings, seeds, and timestamps — the machine-readable counterpart of `wf_report()`.

---

## 7. Release 0.14 — Methods I: soft calibration and entropy balancing (Seam 2)

### 7.1 `wf_calibrate(method = "soft")` — priority item

Hard calibration treats every margin as an exact constraint; when a category is infeasible or barely supported the current answer is "collapse or stop". Soft (penalized/ridge) calibration replaces exact constraints with a penalized objective — margins are met approximately, with per-margin tolerances or costs:

```r
wf_calibrate(sample, target, method = "soft",
             tolerance = 0.02,            # scalar or per-dimension named vector
             bounds = NULL, init_weight = NULL, ...)
```

Design intent for the non-specialist audience: many "infeasible target" situations that today require understanding and authoring collapse ladders become a *declared tolerance* instead. Discipline is preserved: the achieved-vs-target gap per margin is always computed, reported, and bounded by the declared tolerance — `wf_precheck()` gains a mode that classifies `cat_infeasible` as soft-resolvable when `method = "soft"` is requested, and the report states exactly which margins were relaxed and by how much. Implementation is a penalized variant of the 0.8 Newton solver (same constraint builder, augmented objective), so it inherits the existing engine tests.

### 7.2 `wf_calibrate(method = "ebal")`

Entropy balancing (Hainmueller 2012): exact first-moment balance with minimum Kullback–Leibler divergence from base weights. Fits the existing Seam 2 signature and `wf_target` contract; useful when weight-variance minimization is the binding concern.

---

## 8. Release 0.15 — Methods II and deeper diagnostics

- **`wf_attrition()`** — panel/wave nonresponse: models retention between waves (base-R `glm`, same discipline as `wf_propensity()`: overlap + balance diagnostics), emits a stage-1 `wf_weights` that chains through `wf_compose()` across waves. Wave-to-wave weight stability enters `wf_validate()`.
- **k-source `wf_blend()`** — generalize from two sources to a named list, with λ vectors per cell (neff / inverse-variance shares across k sources; fixed λ as a simplex-constrained table). The two-source API remains sugar.
- **`wf_influence()`** — per-unit leverage on the calibrated solution (weight ratio, contribution to margin fit, deff contribution), flagging the handful of respondents that drive extreme weights — the practical question behind most trimming debates.
- **Sensitivity bounds** — for propensity-corrected estimates, report how strong an unmeasured selection factor would need to be to overturn a conclusion (Rosenbaum-style bounds on Seam 2's propensity output), continuing the "assumptions surfaced, not hidden" commitment.
- Fay's BRR (`rho`) lands here with the variance work retained from Extension 2.

---

## 9. Release 0.16 — Performance engineering

- `parallel = TRUE` on `wf_rake()` / `wf_poststrat()` / `wf_replicates()` via `parallel::mclapply` (base R; Windows falls back to serial with a note), with per-group determinism guaranteeing identical results — exactly as reserved in core §11.
- Progress reporting for long runs (replicates, autoweigh over many groups) via `cli` behind `Suggests:`, silent fallback otherwise.
- The Rcpp IPF kernel remains an escape hatch, built **only if** 0.16 profiling on national-scale data shows the R kernel binding; the decision and benchmark enter this document as an addendum at that time.

---

## 10. Release 1.0 — Freeze and publish

- **API freeze.** All exported signatures, S3 class fields, and condition classes frozen; additions allowed, removals and semantic changes require a major version. A deprecation policy (one full minor release with classed `wf_warning_deprecated`) is documented.
- **CRAN submission.** `cran-comments.md`, reverse-dependency-free initial submission, win-builder/mac-builder pre-checks, the existing five-platform CI as gate.
- **Bilingual pkgdown site.** English reference (auto-generated) plus hand-written articles in English and Simplified Chinese: quick start, the precheck→execute→diagnose discipline, bringing your own population data, dirty-data handling, and a manager-oriented "reading the weighting report" guide. Chinese articles live on the site, not in the package, keeping the language policy intact.
- **WFCstudio beta** released alongside, pinned to the frozen 1.0 API.
- The three vignettes promised in core §13 ship as part of this cycle (they double as site articles).

---

## 11. Cross-cutting commitments (extended)

Everything in Extension 2 §7 continues to hold. Additions:

- **Zero-hard-dependency core, forever.** survey, srvyr, broom/generics, cli, rmarkdown, shiny: all behind `Suggests:` or in the WFCstudio sibling. The computational core stays base-R installable.
- **Guided mode is auditable mode.** Anything `wf_autoweigh()` or WFCstudio decides automatically exists as a reviewable artifact in the decision ledger; there is no automated action without a corresponding inspectable object.
- **Localization never touches semantics.** Message catalogs affect rendering only; condition classes and payloads are the stable contract.
- **Every methods addition ships with its own diagnostics** (soft-calibration relaxation report, ebal divergence, attrition balance, influence tables) — the precheck→execute→diagnose loop closes for each new method before it is exported.

---

## 12. Consolidated public API after 1.0 (planned)

```
# guided path (0.11)
wf_autoweigh()

# reporting & visualization (0.10)
wf_report(); plot() methods; wf_auto_trim(); wf_suggest_ladder()

# interoperability (0.12)
as_svydesign(); as_svrepdesign(); tidy(); glance(); augment()

# production infrastructure (0.13)
wf_pipeline(); wf_run(); wf_validate(); wf_audit_export()

# methods (0.14 / 0.15)
wf_calibrate(method = "soft" | "ebal" | ...); wf_attrition()
wf_blend(<k sources>); wf_influence()

# performance (0.16)
parallel = TRUE; progress reporting

# sibling package
WFCstudio (GUI over the same engine)
```

Each release is independently shippable; the ordering encodes the agreed priority — make the guided path real first, connect to the ecosystem second, harden production third, then widen the method family — and the seams guarantee any reordering remains safe.

---

Reference implementations of the above live in `inst/reference/`, following the convention of the earlier extensions: `wfc_future_usability.R` (§3–§4: i18n, auto-trim, ladder drafting, report, autoweigh, plots), `wfc_future_infrastructure.R` (§5–§6: survey bridges, tidiers, pipeline, validate, audit export), and `wfc_future_methods.R` (§7–§8: soft calibration, entropy balancing, attrition weighting, influence diagnostics). They are prototypes to be sourced after `library(WFC)`, not installed API.

# weightflow — Design Document, Extension 2 (Development Roadmap)

**Planned weighting methods, source fusion, and workflow automation**

Companion to `weightflow_design.md` (core, 0.1) and `weightflow_poststrat_design.md` (Extension 1, 0.2). This document specifies the intended contents of releases 0.3 through 0.6. It is a *design commitment*, not yet an implementation: each section fixes the function signatures, the seam it attaches to, the theory it rests on, and the automation discipline it must preserve, so that the work can be picked up incrementally without re-litigating the architecture.

Nothing here breaks 0.1/0.2. Every addition lands on one of the three reserved seams from the core document — Seam 1 (target constructors), Seam 2 (method registry), Seam 3 (weight-pipeline ledger) — and reuses the existing `wf_dims` / `wf_target` / `wf_weights` / `wf_precheck` / `wf_diagnostics` classes and the frozen condition taxonomy. The guiding principle from 0.1 carries through unchanged: **precheck → execute → diagnose**, with every automated decision surfaced as a reviewable, provenance-logged artifact rather than a silent internal choice.

---

## 1. Roadmap at a glance

| Release | Theme | New public API | Seam |
|---|---|---|---|
| 0.3 | Dual-source fusion | `wf_blend()`, `wf_compose()` | 3 |
| 0.4 | Non-probability correction | `wf_propensity()`, `wf_target_propensity()` | 2, 1 |
| 0.5 | Variance & uncertainty | `wf_replicates()`, `wf_variance()` | 3 |
| 0.6 | Bounded calibration & automation aids | `wf_calibrate(method="greg"/"logit")`, `wf_auto_trim()`, `wf_suggest_ladder()`, `wf_report()` | 2, + helpers |
| later | Small-area modelling | `method = "mrp"`, `redistribute = "model"` | 2 |

The ordering reflects marginal value to the current national-survey project: fusion (0.3) is the final deliverable's missing link; propensity (0.4) is the correct treatment the online sample currently lacks; variance (0.5) is unavoidable for any reported figure. Bounded calibration and the automation aids (0.6) are quality-of-life improvements over the current "rake-then-trim, hand-authored ladders" workflow.

---

## 2. Release 0.3 — Dual-source fusion (Seam 3)

### 2.1 Problem

The project produces two weight sets from two sources: an online sample calibrated by `wf_rake()` and an offline sample calibrated by `wf_poststrat()`. The deliverable is a single set of population estimates (e.g. pass rates by subpopulation) that combines both. This is the classic **dual-frame / probability-plus-nonprobability composite estimation** problem (Hartley 1962; Elliott & Valliant 2017): each source yields an estimate of the same population quantity, combined as a convex combination with weights lambda and 1 - lambda.

### 2.2 Two hard design rules, learned from the earlier analysis

**Rule A - compose at the estimator level, never pool the weights.** For a ratio-type quantity such as a pass rate, `blend(rate) = lambda * rate_online + (1 - lambda) * rate_offline` is *not* equal to computing a rate on a table where the two weight sets are stacked. Stacking lets each source's contribution be silently governed by its weight-sum in each cell rather than by the intended lambda. `wf_blend()` therefore computes each source's per-cell estimate first, then combines; the lambda the user sets is exactly the lambda that acts.

**Rule B - lambda varies by cell (or at least by group), it is not one global constant.** The relative reliability of the two sources differs across cells: an urban high-education cell may have thousands of online respondents and few offline; a rural elderly cell the reverse. A single global lambda (e.g. a blanket 4:1) over-trusts online where it is thin - precisely where its raked/collapsed estimate is an extrapolation - which is a substantive error, not just an efficiency loss. The default lambda is data-driven per cell.

### 2.3 `wf_blend()`

```r
wf_blend(
  online, offline,                 # two wf_weights objects (raking / poststrat)
  by_cell,                         # dimensions defining the fusion cells
  lambda = c("neff","inverse_variance","fixed"),
  lambda_fixed = NULL,             # required iff lambda = "fixed" (scalar or per-group)
  outcome = NULL,                  # 0/1 or numeric column -> returns fused estimate
  level = c("cell","group"),       # granularity at which λ is computed & applied
  trim_lambda = c(0.05, 0.95),     # clamp data-driven λ away from degenerate 0/1
  sensitivity = TRUE               # sweep global online-share and report stability
)
```

λ strategies:

- **`"neff"`** — λ_cell ∝ effective sample size n/deff of each source in that cell. This is the standard composite weight and correctly discounts the online source for its post-raking design effect (measured, not nominal — the online sample is large in name but its deff is high after correcting self-selection).
- **`"inverse_variance"`** — λ_cell = Var(est_offline) / (Var(est_online) + Var(est_offline)), the minimum-variance optimal composite when both estimates are approximately unbiased. Naturally cell-varying; degenerate cells (one source empty) collapse to the other source automatically.
- **`"fixed"`** — a user constant, e.g. `lambda_fixed = 0.8` for a 4:1 prior preference. Permitted, but flagged in diagnostics with the discrepancy between the stated λ and the neff/inverse-variance λ, so a hand-picked 4:1 cannot masquerade as an optimal one.

Returns a `wf_blend_result` carrying: the fused per-cell (and aggregated) estimate with its variance; the **effective λ actually applied per cell**; each source's per-cell estimate, variance, and effective sample size; and, when `sensitivity = TRUE`, the fused estimate as the global online share is swept (e.g. 0.3–0.9) so the analyst sees whether the conclusion depends on the λ choice. `outcome = NULL` returns fused *weights* per cell instead of an estimate, for downstream use.

### 2.4 `wf_compose()` — the pipeline ledger

Independently of fusion, real production weighting is a chain: design weight → nonresponse adjustment → calibration → trimming. `wf_compose(w1, w2, ...)` multiplies successive `wf_weights` stages per unit and concatenates their provenance, yielding one auditable object whose history lists every stage, its method, and its settings. This is the substrate on which 0.5's replicate weights run (a replicate is the whole chain re-executed on resampled data), so it ships in 0.3 even though its headline use arrives later.

### 2.5 Assumptions surfaced, not hidden

`wf_blend()` prints, and stores in provenance, the exchangeability/unbiasedness assumptions the fusion rests on: convex combination is meaningful only if both sources are approximately unbiased for the cell quantity, and the online source's unbiasedness holds only under "the calibration variables explain the full selection mechanism." The sensitivity sweep is the operational hedge against that assumption. This mirrors the core principle that the package makes its methodological commitments explicit and checkable.

---

## 3. Release 0.4 — Non-probability correction via propensity (Seams 2 & 1)

### 3.1 Problem

The online sample is self-selected. Raking it to population margins assumes those margins capture the whole selection bias; residual selection on anything outside the calibration variables is untouched. The principled remedy uses the offline **probability** sample as a reference to model the *propensity to enter the online sample*, then corrects for it — turning "pretend the online sample is a probability sample" into "estimate and invert its selection mechanism." The project already has the ideal reference sample in hand.

### 3.2 Method (pseudo-design weights, then calibration)

Stack the online sample and the offline reference; fit a model for membership (online vs reference) on the shared demographic (and any available auxiliary) variables; convert fitted propensities into **pseudo-design weights** for the online units (inverse-propensity, or the equivalent kernel/weighting variants). These pseudo-design weights then feed `wf_rake()` or `wf_poststrat()` as the *initial* weights of a second stage — exactly the `init_weight` slot Extension 1 already exposes. So propensity correction is a first-stage weight producer, not a replacement for calibration; the two compose cleanly via `wf_compose()`.

### 3.3 API

```r
wf_target_propensity(          # Seam 1: builds the reference frame + model spec
  online, reference,
  formula,                     # membership model, e.g. member ~ age5 + edu5 + cx
  method = c("logit","rf","gbm"),  # base R glm default; ML variants behind Suggests
  by = NULL
)

wf_propensity(                 # Seam 2: fit, diagnose, emit pseudo-design weights
  online, target_propensity,
  weight = c("ipw","kernel","matching"),
  stabilize = TRUE,            # stabilized IPW to tame extreme weights
  trim = NULL                  # optional propensity-weight trimming
)                              # -> wf_weights (stage 1), feed as init_weight to stage 2
```

### 3.4 Discipline and diagnostics

Propensity work has well-known failure modes, each mapped to the existing condition taxonomy and precheck philosophy: non-overlap / poor common support (online units with propensities near the boundary → `wf_warning_quality`, with a support-overlap report); model misspecification (a balance table before/after weighting, checking whether weighting actually equalizes covariate distributions between online and reference); extreme weights (stabilization on by default, trimming available). `wf_propensity()` returns, alongside the weights, a **balance diagnostic** and an **overlap diagnostic**, so the precheck→execute→diagnose loop holds here too: fit is never accepted blindly.

### 3.5 Base-R first

The default membership model is `glm(family = binomial)` — zero new dependencies. Random-forest / gradient-boosting propensity variants register behind `Suggests:` (ranger / xgboost), selected by `method=`, never required. This keeps the core installable everywhere while allowing stronger models where available.

---

## 4. Release 0.5 — Variance and uncertainty (Seam 3)

### 4.1 Problem

Every weighted figure the project reports — a pass rate, a subpopulation mean — needs a variance, or it is a point with no interval. Analytic variance formulas for a multi-stage design → nonresponse → calibration → fusion pipeline are intractable to write by hand and fragile to maintain. Replication is the standard, general answer: re-run the *entire weighting pipeline* on resampled data and read the variability of the result.

### 4.2 `wf_replicates()` and `wf_variance()`

```r
wf_replicates(
  pipeline,                    # a wf_compose() chain (or a single wf_weights)
  method = c("bootstrap","jackknife","brr"),
  R = 500,                     # bootstrap replicates
  strata = NULL, clusters = NULL,  # respect the offline design structure
  seed = NULL
)                              # -> wf_replicate_weights (R columns of weights)

wf_variance(
  replicate_weights, estimator,   # a function(weights, data) -> scalar/vector
  data
)                              # -> estimate, variance, CI per quantity
```

Because 0.3's `wf_compose()` captures the pipeline as a replayable object with full provenance, a replicate is literally that object re-executed on a resampled input — including re-raking, re-collapsing, re-blending — so the variance reflects the *calibration* uncertainty, not just sampling of raw responses (a common and serious omission). For the offline probability sample, resampling respects its strata and clusters; for the online sample, the resampling reflects whichever first-stage (propensity) model produced its weights. The fused estimator from `wf_blend()` is a valid `estimator` argument, giving confidence intervals on the final composite pass rates.

### 4.3 Determinism and cost

Replication is embarrassingly parallel (each replicate independent); the reserved `parallel = TRUE` path applies. Seeds are stored in provenance for exact reproducibility. Because the R kernel for raking/poststrat is already fast (§11 of the respective docs), R = 500 replicates on national-scale data remain tractable on a laptop; where they are not, the reserved Rcpp kernel is the escape hatch.

---

## 5. Release 0.6 — Bounded calibration and automation aids

### 5.1 GREG / linear and logit calibration (Seam 2)

Raking is one special case of calibration estimation (Deville–Särndal). The general family lets the analyst bound the weight adjustment. **Logit calibration** produces weights bounded within a user range *by construction*, which merges two steps the current workflow keeps separate: aligning margins and keeping weights from exploding. Where `wf_rake()` today is followed by a separate trimming pass, `wf_calibrate(method = "logit", bounds = c(L, U))` does both at once, with no post-hoc distortion of the margins that trimming introduces. `method = "greg"` (linear calibration) is the unbounded GREG estimator, useful when an auxiliary continuous variable (not just categorical margins) is available. Both register on the same Seam 2 interface as raking and poststrat and consume the same `wf_target`; `wf_rake()` remains sugar for `method = "raking"`.

### 5.2 `wf_auto_trim()` — data-driven trim bounds

Trim limits are currently hand-set (e.g. 15× the mean). This helper sweeps a grid of candidate caps and, for each, reports the **bias–variance frontier**: the incremental margin/cell residual introduced versus the design-effect reduction obtained. It then recommends the loosest cap meeting a stated criterion (e.g. "design effect below X" or "residual below Y"), turning a purely judgemental number into a reproducible, defensible choice. It only *recommends*; the analyst sets the final bound, consistent with the discipline that automated steps propose and humans approve.

### 5.3 `wf_suggest_ladder()` — collapse-ladder drafting

Extension 1 requires collapse ladders to be pre-declared (correctly — merges must be auditable). This helper does not remove that requirement; it lowers its cost. It scans, across all groups, which dimensions are sparse and in what order collapsing them loses the least information, and emits a **draft** `wf_collapse_ladder` for the analyst to review and edit. The workflow shifts from authoring a ladder on a blank page to amending a proposed one — the same "precheck produces, human approves" pattern as `wf_suggest_collapse()` in 0.1.

### 5.4 `wf_report()` — the weighting quality dossier

A single call that assembles the full weighting-quality report from a `wf_weights` (or composed pipeline): per-group design effect, effective sample size, weight-distribution extremes, trimming impact, margin/cell residuals, fusion λ and its sensitivity, propensity balance/overlap where applicable, and a plain-language verdict per group. Output is a structured object with `print()`, `as.data.frame()`, and an optional rendered document (Markdown/HTML) suitable for a methods appendix. This replaces the current hand-assembled tables with one reproducible, standardized artifact — the natural capstone of the precheck→execute→diagnose discipline extended across the whole pipeline.

---

## 6. Later — Small-area modelling (Seam 2)

MRP (multilevel regression with post-stratification) is the principled route when cells are genuinely empty and small-area estimates are required. It fits a multilevel model to the outcome (or cell means), borrowing strength across cells via partial pooling to predict **every** cell including the empty ones, then aggregates against the joint population table already available in `wf_target$joint`. It attaches at two reserved points: as `method = "mrp"` in the registry, and as `redistribute = "model"` in `wf_poststrat()` (an orphan cell is filled by a model prediction rather than proportional reallocation). It shifts from a weighting paradigm to a modelling one, so its outputs are labelled as model-based estimates and its assumptions documented accordingly; multilevel fitting brings a `Suggests:` dependency (e.g. lme4 or a Stan backend), never required by the core.

---

## 7. Cross-cutting commitments

Every roadmap item honours the invariants established in 0.1/0.2, so the package grows without drifting:

- **Discipline preserved.** Each new method ships with its own precheck-side diagnostics (overlap, balance, bias–variance frontier, λ sensitivity) and refuses to accept a fit silently; automated choices are always drafts a human ratifies.
- **Seams, not surgery.** No roadmap item edits the core engines; each is a new constructor (Seam 1), a registry entry (Seam 2), or a pipeline stage (Seam 3). The `.wf_methods` registry unifies raking, poststrat, greg, logit, propensity, and mrp under one `wf_calibrate()` interface.
- **Base-R core, optional acceleration.** Every default path is base R with zero hard dependencies; ML propensity models, Rcpp kernels, parallel replication, and Stan-backed MRP all live behind `Suggests:` and are selected explicitly.
- **Provenance and reproducibility.** Every stage — fusion λ, propensity model, trim bound, replicate seed — is recorded in the result's provenance, so any reported figure is reconstructable from its own metadata.
- **Assumptions surfaced.** Where a method rests on an untestable assumption (online unbiasedness, cell exchangeability, model form), the package states it in output and provides the corresponding sensitivity or balance check rather than burying it.

---

## 8. Consolidated public API after 0.6 (planned)

```
# core (0.1)
wf_dims(); wf_target_population(); wf_target_reference(); wf_target_manual()
wf_target_shrink(); wf_precheck(); wf_suggest_collapse(); wf_apply_collapse()
wf_rake(); wf_diagnose()

# post-stratification (0.2)
wf_collapse_ladder(); wf_plan_poststrat(); wf_poststrat()

# unified method interface (0.6)
wf_calibrate(method = "raking" | "poststrat" | "greg" | "logit" | "propensity" | "mrp")

# fusion & pipeline (0.3, 0.5)
wf_blend(); wf_compose(); wf_replicates(); wf_variance()

# non-probability correction (0.4)
wf_target_propensity(); wf_propensity()

# automation aids (0.6)
wf_auto_trim(); wf_suggest_ladder(); wf_report()
```

Each block is independently shippable; the table in §1 is the intended order, but the seams make any reordering safe should project priorities change.

# weightflow — Design Document

**A survey-weighting workflow package for multi-source national surveys**

Working name: `weightflow` (verify availability on CRAN/GitHub before release; alternatives: `svycalib`, `weftr`). Version target for this document: 0.1.0. The current package-wide licensing policy supersedes the original recommendation: GPL (>= 2), with dependency copyright boundaries documented in `inst/COPYRIGHTS`.

---

## 1. Scope and positioning

The package does **not** aim to be "another raking package". Its value proposition is the *workflow around* calibration for large national surveys with mixed probability (offline) and non-probability (online) components:

1. Schema-agnostic ingestion of sample data and population benchmarks.
2. A mandatory **precheck → execute → diagnose** discipline, so that infeasible or fragile calibration problems are detected *before* iteration, not after 200 silent sweeps.
3. Two interchangeable target-construction modes: benchmarks derived from a **reference sample's design weights** (feature values), or from **external population data** in arbitrary formats.
4. Production-grade handling of dirty inputs: missing categories, NA demographics, anomalous feature values, unmatched keys.
5. A stable core on which future weighting functionality (post-stratification, propensity adjustment, composite weighting of online+offline, replicate-weight variance estimation) can be added without breaking the API.

The IPF engine itself is a commodity; correctness is validated against `survey::rake()` in the test suite. Everything else in this document is where the package earns its existence.

---

## 2. Design principles

**P1 — Discipline over convenience.** `wf_rake()` refuses to run when the precheck reports blocking issues. There is no `force = TRUE` escape hatch on errors; the user must either fix the data or explicitly apply a documented remediation (e.g., a collapse plan). Warnings do not block but are always recorded in the result's provenance.

**P2 — Schema agnosticism.** No demographic variable name, category coding, number of dimensions, or grouping variable is hard-coded anywhere. The current project's `gender/cx/age5/edu5` by `province` is just one instantiation of a generic *dimension specification*. Adding, removing, or entirely replacing the classification system requires zero package changes.

**P3 — Canonical internal representations.** All external formats (a county-level joint population table, a reference sample with feature values, a hand-typed margins table) are converted at the boundary into one canonical `wf_target` object. The engine never sees raw external data. This is the single most important extensibility decision: every future data source only needs a new constructor, and every future engine only needs to consume `wf_target`.

**P4 — Immutability and provenance.** Functions never modify inputs in place. Every result object carries a `provenance` record: input hashes, settings, package version, timestamps, precheck summary, convergence log, trimming actions. A weighting result must be reproducible and auditable from its own metadata.

**P5 — Fail loudly, fail early, fail structurally.** All errors and warnings are classed conditions carrying machine-readable data (which group, which dimension, which categories), so that calling code — including future GUI or pipeline layers — can programmatically react instead of parsing message strings.

**P6 — Base-R core.** The computational core has zero hard dependencies (`Imports:` is empty in 0.1). Optional acceleration and tidy interfaces live behind `Suggests:`. This keeps installation trivial in restricted/intranet environments and minimizes long-term maintenance risk.

---

## 3. Data model (S3 classes)

| Class | Constructor(s) | Role |
|---|---|---|
| `wf_dims` | `wf_dims()` | Declares calibration dimensions: variable names, optional expected levels, optional collapse ladders. |
| `wf_target` | `wf_target_population()`, `wf_target_reference()`, `wf_target_manual()` | Canonical benchmark: per-group totals + per-dimension margins. Carries its `mode` ("population" / "reference" / "manual"). |
| `wf_precheck` | `wf_precheck()` | Structured issue list (data.frame) + overall verdict + suggested remediations (e.g., collapse plans). |
| `wf_weights` | returned by `wf_rake()` | Per-unit weights and feature values, per-group convergence records, full provenance. |
| `wf_diagnostics` | `wf_diagnose()` | Per-group quality table (deff, ESS, extremes, residual margin error, trim share) with print/summary methods. |
| `wf_collapse_plan` | `wf_suggest_collapse()` / user-built | An explicit, reviewable category-merge specification applied by `wf_apply_collapse()`. |

The canonical `wf_target` structure:

```r
structure(list(
  mode   = "population",           # or "reference", "manual"
  by     = "prov",                 # grouping variable name (NULL = single group)
  dims   = c("gender","cx","age5","edu5"),
  groups = list(                   # one element per group value
    `11` = list(
      total   = 18300000,          # weight-sum target for the group
      margins = list(              # named numeric vectors; names are category keys (character)
        gender = c(`1` = 9.3e6, `2` = 9.0e6),
        cx     = c(`1` = 1.6e7, `2` = 2.3e6),
        age5   = c(...), edu5 = c(...)
      )
    ),
    ...
  ),
  meta = list(source_hash = "...", created = ..., scale = "population")
), class = "wf_target")
```

Invariant enforced at construction: within each group, `sum(margins[[d]]) == total` for every dimension `d` (tolerance 1e-8 relative). This invariant is what guarantees IPF preserves the group total at every sweep.

Category keys are **always characters**. All joins between sample and target happen on `as.character()` values — this is the "keys are the only link" contract (requirement 3): the population file may have any column names and any auxiliary columns; the constructor's `key_map` argument declares which of its columns correspond to which sample variables, and values are matched as character keys.

---

## 4. Input handling: missing and anomalous data (requirement 1)

### 4.1 Taxonomy of dirt, and the assigned behavior

| Problem | Where | Default behavior | Override |
|---|---|---|---|
| NA in a calibration dimension | sample | **fractional allocation** (see 4.2) | `na = "drop"`, `na = "error"` |
| NA in the grouping variable | sample | error (`wf_error_input`) — a unit that belongs to no group cannot be calibrated | `na_group = "drop"` with warning |
| NA / NaN / Inf / `<= 0` feature value | reference sample | error listing offending row indices | `feature_na = "drop"` (drops from *target construction only*, with a warning quantifying dropped weight share) |
| Feature value > 1 | reference sample | warning (inclusion probabilities should lie in (0,1]; values > 1 suggest the column is a weight, not its reciprocal) | `feature_gt1 = "allow"` |
| Sample category absent from target | sample vs target | error: cannot assign a benchmark share | collapse plan, or `unknown = "drop"` |
| Target-positive category with zero sample support | target vs sample | error: **infeasible** — IPF cannot manufacture respondents | collapse plan (the only correct fix) |
| Duplicate unit IDs | sample | error | — |
| Zero/negative population counts | population file | zero allowed (category simply drops out of margins after aggregation); negative is an error | — |
| Group present in sample but not in target (or vice versa) | both | error, with the full list of unmatched group keys | `groups = "intersect"` restricts to common groups, with warning |
| Type instability (factor vs integer vs character coding of the same variable) | everywhere | silently harmonized: everything is compared as trimmed character keys | — |

Rows with NAs in more than `max_na_dims` dimensions (default 2) are rejected: fractional expansion across k missing 5-level dimensions multiplies the row 5^k times and, more importantly, such a unit carries almost no calibrating information.

### 4.2 Fractional allocation of NA rows

A unit with NA on dimension *d* is expanded into one pseudo-row per category of *d* (Cartesian across all its missing dimensions). Initial allocation shares are proportional to the group's target shares for the missing categories. The pseudo-rows then participate in IPF as ordinary rows; the iteration itself pulls the allocation toward consistency with all margins simultaneously (this is a one-shot EM analogue and is fully deterministic). After convergence the pseudo-rows are summed back by unit ID, so each physical respondent receives exactly one weight, and the group total is untouched.

Trimming interacts with expansion at the **person** level: caps apply to the summed per-person weight; when a cap binds, all of that person's pseudo-rows are scaled by a common factor before re-raking (see 7).

---

## 5. Dimension abstraction (requirement 2)

`wf_dims()` is the single place the classification system is declared:

```r
dims <- wf_dims(
  gender = NULL,                        # levels inferred from target
  cx     = NULL,
  age5   = c("1","2","3","4","5"),      # or declared explicitly
  edu5   = c("1","2","3","4","5"),
  .collapse = list(                     # optional pre-registered ladders
    edu5 = list(step1 = c(`1`="12", `2`="12", `3`="3", `4`="45", `5`="45")),
    age5 = list(step1 = c(`4`="45", `5`="45"))
  )
)
```

Swapping to an entirely different classification set (say `region/income4/occupation7`) means writing a different `wf_dims()` call — nothing else changes. Dimensions can be added or removed freely; the engine iterates over whatever `dims` contains. Interaction calibration is expressed *within the same abstraction*: to calibrate a two-way joint, the user (or a provided helper `wf_cross(sample, "cx", "edu5")`) creates a concatenated key variable and lists it as one more dimension. No special engine path is required.

Collapse ladders are **pre-registered, named, and explicit**. `wf_suggest_collapse()` may propose applying a ladder step for a specific group and dimension; `wf_apply_collapse()` applies it to *both* the sample and the target consistently and records the action in provenance. The engine never merges categories on its own — remediation is always a visible, user-approved step (P1).

---

## 6. Target construction: the two modes (requirements 3 & 4)

### 6.1 Mode "population" — external benchmark, arbitrary format

```r
tgt <- wf_target_population(
  pop        = county_table,                          # any data.frame
  key_map    = c(gender = "gender2", cx = "cx2",      # sample var -> pop column
                 age5 = "age5", edu5 = "edu5"),
  count      = "pop",                                 # column of population counts
  by         = "prov",
  by_key     = function(df) substr(as.character(df$unicode), 1, 2),  # or a column name
  dims       = dims,
  scale      = c("population","sample","custom"),     # weight-sum semantics
  totals     = NULL                                   # required when scale = "custom"
)
```

The constructor aggregates the file to per-group margins, validates the additivity invariant, and normalizes totals according to `scale`: `"population"` keeps true head counts (weights read as "persons represented"); `"sample"` rescales each group's total to that group's sample size (mean weight 1; requires the sample at construction or later via `wf_rescale()`); `"custom"` uses user-supplied totals. Only the mapped key columns and the count column are ever read — every other column in the population file is ignored, which is what makes the format-agnostic contract cheap to honor.

If the file is *already* a margins table (one row per group × dimension × category) rather than a joint table, `wf_target_manual()` accepts it directly with a `dim_col`/`cat_col`/`value_col` mapping.

### 6.2 Mode "reference" — benchmark from a weighted reference sample

```r
tgt <- wf_target_reference(
  ref     = offline_sample,     # data.frame containing feature values
  feature = "ps",               # feature value = 1/design weight (inclusion prob.)
  dims    = dims,
  by      = "prov"
)
```

Weights are recovered as `1/feature` after the anomaly screen of 4.1; per-group weighted margins and totals become the target. The resulting `wf_target` is structurally identical to the population mode, so `wf_precheck()`, `wf_rake()`, and `wf_diagnose()` are completely mode-blind — the two computation paths the requirements demand are two constructors, not two engines. Both modes can also be **blended**: `wf_target_shrink(tgt_group, tgt_national, lambda)` implements small-group shrinkage toward a national target, for sparse-province situations.

---

## 7. Execution: `wf_rake()`

```r
w <- wf_rake(
  sample, target,
  id          = "resp_id",          # NULL -> row numbers
  na          = c("fractional","drop","error"),
  trim        = c(lo = 0.05, hi = 8), trim_cycles = 4,   # NULL disables
  tol         = 1e-6, max_iter = 200,
  precheck    = TRUE,               # runs wf_precheck(); aborts on blocking issues
  keep_rows   = FALSE               # TRUE retains the expanded pseudo-row table
)
```

Per group, the engine: (1) resolves category keys to integer level indices once (`match()`), so the iteration touches no strings; (2) performs fractional expansion; (3) runs vectorized IPF sweeps — for each dimension, level-wise weighted sums via `rowsum()` on the integer index, ratio update `w <- w * ratio[idx]`, O(n) per dimension with no per-category inner loop; (4) declares convergence when the largest |ratio − 1| in a full sweep falls below `tol`; (5) applies person-level trimming cycles (clip → re-rake → repeat), then a final hard clip with total restoration; (6) aggregates pseudo-rows to person weights; (7) emits feature values `1/weight` alongside weights.

Non-convergence raises `wf_error_convergence` carrying the group key, the worst dimension/category, and the last deviation — plus a pointer to the precheck finding that (almost always) predicted it.

---

## 8. Precheck: `wf_precheck()` (requirement 6)

Runs entirely before any iteration and returns a `wf_precheck` object whose core is a tidy issue table:

| column | content |
|---|---|
| `group` | group key ("*" for global issues) |
| `dim`, `category` | location of the issue (NA where not applicable) |
| `check` | machine name: `schema_missing_var`, `group_unmatched`, `cat_unknown_in_sample`, `cat_infeasible`, `support_thin`, `na_load`, `na_overload`, `feature_anomaly`, `dup_id`, `risk_extreme_ratio` |
| `severity` | `"error"` (blocks `wf_rake`), `"warning"`, `"note"` |
| `detail` | human-readable message with numbers |
| `data` | list-column of structured payload |

Two checks deserve emphasis because they encode hard-won lessons from real data. `cat_infeasible` fires when a target-positive category has zero (post-expansion) sample support — the exact Tibet failure mode — and is classified as an error because no amount of iteration can fix it. `risk_extreme_ratio` computes, per group and dimension, `target_share / sample_share` and flags ratios above a threshold (default 10) as warnings with the implied minimum weight inflation, together with a rough lower bound on the resulting design effect; this is what tells the analyst *before running anything* that education must be collapsed nationally, not patched province by province.

`wf_suggest_collapse(precheck, dims)` converts infeasible/thin findings into concrete `wf_collapse_plan` objects using the pre-registered ladders, which the user reviews and applies explicitly. The discipline loop is therefore: `precheck → (apply named remediations) → precheck again → rake → diagnose`, and every step leaves a provenance trace.

---

## 9. Diagnostics: `wf_diagnose()`

Per group: n (persons), n_effective (Kish ESS), design effect 1 + CV², mean/min/max weight, max/mean ratio, share of weights at trim bounds, residual margin error after trimming (max relative deviation per dimension), iterations, converged flag. Global summary: pooled ESS, worst groups, and a plain-language verdict line per group ("OK", "usable with caveats", "do not publish separately"). `print()` renders a compact table; `as.data.frame()` returns it raw for the user's reporting pipeline. A `plot()` method (histogram of weights per group, base graphics) is planned for 0.2.

Diagnostics never recompute weights; they read the `wf_weights` object plus, when residual-margin checks are requested, the original target — keeping execute and diagnose strictly separated.

---

## 10. Error and condition system (requirement 8)

All signals are classed conditions built by two internal helpers (`wf_abort()`, `wf_warn()`), each attaching a `data` payload:

| class | raised when |
|---|---|
| `wf_error_input` | malformed arguments, duplicate IDs, NA group keys, type problems |
| `wf_error_schema` | dimension/group variables missing; key maps referencing absent columns |
| `wf_error_feasibility` | infeasible categories; precheck errors blocking `wf_rake` |
| `wf_error_convergence` | IPF failed to converge within `max_iter` |
| `wf_error_internal` | invariant violations (a bug) — message asks the user to file an issue |
| `wf_warning_data` | dropped rows, feature anomalies handled by policy, thin support |
| `wf_warning_quality` | high design effect, trimming distorting margins beyond a threshold |

Rules: every message names the *location* (group/dim/category) and the *next action*; counts are always included ("7,874 of 117,788 reference rows (6.7%) have missing feature values"); errors are never raised from deep inside loops without context — the engine catches low-level failures and re-throws with group/dimension attached. `tryCatch(..., wf_error_feasibility = function(e) e$data$plan)` style programmatic handling is a documented, tested pattern.

---

## 11. Performance engineering (requirement 7)

Measured design targets: 32 groups × ~5k rows × 4 dimensions should complete in well under one second; 1M rows × 6 dimensions in a few seconds on a laptop.

The techniques, in order of importance: (1) **string work happens once** — all category matching is hoisted out of the iteration into integer index vectors; (2) **`rowsum()` for grouped sums** — C-level, no R-level per-category loops anywhere in the sweep; (3) **pre-split by group** — `split(seq_len(n), group_key)` once, then each group works on plain atomic vectors, never on data.frame subsets; (4) **no copies in the loop** — the weight vector is the only object mutated per sweep; (5) fractional expansion is built with `rep()`/index arithmetic, not `rbind()` in a loop.

Reserved acceleration paths (behind `Suggests:`, zero API change): a `data.table` fast path for target construction from very large population files (county-level joints with tens of millions of rows), and an optional Rcpp IPF kernel selected via `options(weightflow.engine = "cpp")` if profiling ever shows the R kernel to be the bottleneck (unlikely at national-survey scale). Groups are embarrassingly parallel; `wf_rake(..., parallel = TRUE)` via `parallel::mclapply` is planned for 0.2 with identical results guaranteed by per-group determinism.

---

## 12. Extensibility reserve (requirement 5)

The growth plan is organized around three stable seams:

**Seam 1 — target constructors.** New benchmark sources (census APIs, margins spreadsheets, pooled multi-year benchmarks) are new `wf_target_*()` functions producing the same canonical object. Nothing downstream changes.

**Seam 2 — method registry.** `wf_calibrate(sample, target, method = "raking", ...)` dispatches through an internal registry (`.wf_methods` environment) mapping method names to engine functions with a documented signature (`function(rows, target_group, control) -> list(w, log)`). `wf_rake()` is sugar for `method = "raking"`. Planned registrations: `"poststrat"` (full-joint cell weighting — the population data already supports it), `"linear"` and `"logit"` calibration (GREG-style, bounded weights by construction, reducing the need for trimming), `"propensity"` (online/offline propensity adjustment as a *first-stage* weight feeding raking as a second stage).

**Seam 3 — weight pipeline ledger.** `wf_weights` objects can be **chained**: `wf_compose(w_design, w_nonresponse, w_calib)` multiplies stages and concatenates provenance, so the package can grow into full production weighting (design weight → nonresponse adjustment → calibration → trimming → composite blending of online and offline via effective-sample-size λ, `wf_blend()`), with every stage inspectable. Replicate-weight variance estimation (`wf_replicates()`, jackknife/bootstrap) attaches at this seam too, since replicates are just the pipeline re-run on resampled data.

Namespace policy protecting the reserve: every exported symbol is prefixed `wf_`; S3 classes are frozen at 0.1 field names with additions allowed and removals forbidden until 1.0; the condition class taxonomy above is append-only.

---

## 13. Quality assurance

Testing (testthat): unit tests per constructor and per dirt case in the 4.1 table; property tests — margins match targets within tol, group totals preserved exactly, weight positivity, determinism across dimension orderings; an oracle test comparing against `survey::rake()` on shared inputs (skipped if survey absent); regression fixtures encoding the real pathological cases (a Tibet-like group: tiny n, dominant target category with near-zero support; a group with a fully-NA feature column; unmatched province keys). Continuous integration via GitHub Actions on release/oldrel/devel. `R CMD check --as-cran` clean is a merge requirement from day one, even while distribution is GitHub-only.

Documentation: roxygen2 for all exports; three vignettes — "The precheck→execute→diagnose workflow", "Bringing your own population data" (the key-map contract), and "Dirty data: NA, anomalies, and infeasibility" (built around the real failure modes).

Data policy: the package ships only a small **simulated** example dataset. Census-derived population tables are *not* redistributed (licensing is unclear for repackaged official statistics); instead the package documents the expected inputs and provides validators. This choice also keeps CRAN submission unproblematic.

Release plan: 0.1 GitHub-only (core: dims/targets/precheck/rake/diagnose/collapse), dogfooded on the current national survey; 0.2 adds composite weighting, post-stratification, parallel groups; CRAN submission considered once the API has survived two real production cycles unchanged.

---

## 14. Public API summary (0.1)

```
wf_dims()                 declare dimensions, levels, collapse ladders
wf_target_population()    canonical target from arbitrary population data (key-map contract)
wf_target_reference()     canonical target from a reference sample's feature values
wf_target_manual()        canonical target from a ready-made margins table
wf_target_shrink()        shrink a group target toward a national target
wf_precheck()             structured feasibility & dirt report (blocking vs advisory)
wf_suggest_collapse()     turn precheck findings into reviewable collapse plans
wf_apply_collapse()       apply a plan consistently to sample + target (provenance-logged)
wf_rake()                 execute (fractional NA handling, trimming, per-group IPF)
wf_diagnose()             per-group quality table + verdicts
wf_compose(), wf_blend()  reserved seams (0.2): weight chaining, online/offline blending
```

Everything below this line in the repository is `R/core.R` — the reference implementation of the 0.1 core.

---
title: "Verified production pipelines in WFC 2.0"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Verified production pipelines in WFC 2.0}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

A WFC 2.0 pipeline stores method and validation settings but does not construct
an unreviewed target at run time. It accepts an unchanged `wf_design_data` and
an unchanged `wf_verified_target`.

```{r object-pipeline, eval=FALSE}
spec <- wf_pipeline(
  target = target,
  stages = list(
    calibrate = list(method = "raking", tol = 1e-8)
  ),
  validate = list(max_deff = 6, max_margin_dev = 0.01)
)

weights <- wf_run(spec, design, dims)
```

The explicit population mode receives a verified population target, not a raw
population table:

```{r population-pipeline, eval=FALSE}
population_spec <- wf_pipeline(
  target = list(mode = "population"),
  stages = list(calibrate = list(method = "raking"))
)

weights <- wf_run(
  population_spec,
  design,
  dims,
  population = target
)
```

Reference mode works the same way with a target produced by
`wf_import_reference()` and supplied through `reference`.

Runtime changes to ID or base-weight roles are blocked. A runtime propensity
stage cannot be inserted before verified calibration. Declare design roles in
`wf_prepare_design()` and review every weighting method before execution.

```{r audit, eval=FALSE}
wf_audit_export(
  weights,
  "pipeline-audit.json",
  inputs = list(design = design, target = target)
)
```

The audit records pipeline, design, and target identities. Package release
approval does not authorize the specific analysis represented by that audit.

# WFC 2.0 Safety Contract

This installed contract defines the supported weighting boundary for WFC 2.0.
It replaces the WFC 1.x compatibility surface where that surface allowed a
user-selected target to reach a weighting engine. The WFC 1.0 freeze remains an
historical record; this document is authoritative for WFC 2.0.

## Required boundary

A supported weighting run starts with both of these objects:

- a `wf_design_data` created by `wf_prepare_design()`, containing only unit
  identity, declared calibration variables, and declared design variables; and
- a non-demo `wf_verified_target` created by `wf_import_target()` or
  `wf_import_reference()` from unchanged external data and source-evidence
  files.

Planning is deterministic and outcome-blind. A human approves the immutable
plan before execution. Changing the design, target, plan, or approval invalidates
the chain. An AI agent may prepare and inspect artifacts but cannot approve its
own plan.

## Retained safe exports

The supported end-to-end signatures are:

```r
wf_prepare_design(data, id, calibration, base_weight = NULL,
                  strata = NULL, clusters = NULL, fpc = NULL)
wf_target_template(file, dims, by = NULL, example = FALSE)
wf_import_target(data_file, source_file, dims, key_map, count,
                 by = NULL, by_key = NULL, production = TRUE)
wf_import_reference(data_file, source_file, dims, feature,
                    by = NULL, production = TRUE)
wf_plan_cells(design, target, dims, min_cell = 5,
              max_weight_ratio = 4, boundary = target$by, ladder = NULL)
wf_plan_weights(design, target, dims,
                method = c("raking", "logit", "poststrat"),
                bounds = c(0.3, 3), min_cell = 5, cell_plan = NULL)
wf_approve_plan(plan, approver, role, note = NULL, actor_type = "human")
wf_execute_plan(plan, approval, design, target)
wf_guided_plan(data, id, calibration, dims, target_file, source_file,
               source_type = c("population", "reference"), key_map = NULL,
               count = NULL, feature = NULL, ...)
wf_guided_execute(workflow, approval)
wf_attach_weights(data, weights, id, weight_name = ".weight")
wf_assess_impact(weights, data, id, outcomes, level = 0.95)
wf_report(w, target = NULL,
          audience = c("manager", "analyst", "decision", "statistician"),
          lang = NULL, output = c("object", "markdown", "html"), file = NULL)
wf_audit_export(x, file, inputs = NULL, extra = NULL)
```

Direct public calibration functions remain available for statistical users and
agents that already hold safe objects. Their first two inputs are a
`wf_design_data` and a `wf_verified_target`; raw data frames and ordinary
`wf_target` objects are rejected before engine execution:

```r
wf_calibrate(design, target, method = "raking", ...)
wf_rake(design, target, ...)
wf_poststrat(design, target, min_cell, ladder, ...)
wf_auto_trim(design, target,
             caps = c(2, 3, 4, 5, 6, 8, 10, 12),
             lo = 0.05, max_deff = 6, max_residual = 0.02, ...)
wf_autoweigh(design, target, dims,
             method = c("auto", "raking", "poststrat", "logit"), ...)
wf_pipeline(target, stages, validate = NULL)
wf_run(spec, sample, dims = NULL, population = NULL,
       reference = NULL, base_weight = NULL)
```

Categorical entropy calibration is retained only when its constraints come
from the supplied verified external target. It cannot accept inline moments.
Other reporting, diagnostics, replicate-weight, variance, and interoperability
exports remain available, but none may create or replace a weighting target.

## Removed interfaces

The following interfaces are absent in WFC 2.0:

- `wf_target_manual()`;
- `wf_target_shrink()`;
- the `moments` argument accepted through `wf_calibrate(..., moments = ...)`;
- `wf_pipeline(target = list(mode = "manual"), ...)`;
- the `margins` argument of `wf_run()`; and
- any path that sends a raw sample or an unverified target directly to a
  weighting engine.

There is no legacy shim, environment switch, warning-only fallback, unsafe
override, automatic approval, or silent method change.

## No supported replacement

There is no supported replacement for pass-rate targets, desired outcome
means, desired outcome intervals, outcome-heterogeneity objectives, subjective
manual targets, target shrinkage, inline outcome moments, or target-driven
relaxation. These behaviors are intentionally not migrated from predecessor
packages because they can steer results toward a preferred conclusion.

A verified population or reference margin is a replacement only when it is a
genuine, independently sourced design target. Relabeling an outcome goal as a
margin does not make it supported.

## Stable safety object fields

Additive fields may be introduced, but these fields retain their meaning:

- `wf_design_data`: `data`, `roles`, `identity`, `created`, and
  `package_version`;
- `wf_verified_target`: the inherited `wf_target` fields `mode`, `by`, `dims`,
  `groups`, `meta`, and `joint`, plus `evidence`, `demo_only`, `source_type`,
  and `identity`;
- `wf_cell_merge_plan`: `cells_before`, `cells_after`, `map`, `category_maps`,
  `reasons`, `affected_share`, `unresolved_cells`,
  `projected_max_weight_ratio`, `input_identities`, `settings`, `created`,
  `package_version`, and `identity`;
- `wf_weight_plan`: `method`, `settings`, `precheck`, `issues`, `ready`,
  `cell_plan`, `input_identities`, `effective_identities`, `weights`, `created`,
  `package_version`, and `identity`;
- `wf_plan_approval`: `plan_identity`, `approver`, `role`, `note`,
  `actor_type`, `attestation`, `created`, `package_version`, and `identity`;
- `wf_weights`: `data`, `log`, `achieved`, and `provenance`, with immutable
  safety identities recorded in `provenance`; and
- `wf_impact`: `summary`, `weight_identity`, `outcomes`, `level`, `created`,
  `package_version`, and `identity`.

SHA-256 identity values are stable machine-readable links between artifacts.
They are not proof that the source evidence is truthful.

## Stable safety conditions

Safety refusals inherit from `wf_error_safety` and `wf_error`. Their payload
contains `code`, `severity`, `field`, `evidence`, and `next_actions`. Existing
codes keep their meanings. Important WFC 2.0 codes include:

- `verified_target_required`, `demo_target_in_production`, and
  `target_identity_invalid`;
- `design_identity_invalid`, `weight_plan_invalid`, and
  `plan_approval_mismatch`;
- `human_approval_required` and `verified_weighting_inputs_required`;
- `inline_moments_unsupported`, `design_role_override_unsupported`, and
  `manual_pipeline_unsupported`; and
- `target_design_roles_mismatch` and `pipeline_target_type_mismatch`.

Human-facing wording may improve. Code, class, severity, field, evidence, and
next-action keys are the programmatic contract for agents.

## Reporting contract

`audience = "decision"` or `"manager"` gives a short plain-language status,
principal risks, and next actions. `audience = "statistician"` or `"analyst"`
gives the full structured diagnostics and provenance needed for further work.
Both views are projections of the same result object and must reconcile.

Package-level review or release approval never authorizes a particular public
or client analysis. Each consequential analysis needs its own accountable and
qualified human review.

---
title: "Precheck, verified execution, and diagnosis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Precheck, verified execution, and diagnosis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

Precheck is diagnostic; execution remains behind the verified boundary.

```{r precheck, eval=FALSE}
precheck <- wf_precheck(
  design$data,
  target,
  id = design$roles$id
)

precheck$issues
```

Resolve blocking support problems through a declared, reviewable cell plan.
Do not change targets after seeing outcomes.

```{r execute, eval=FALSE}
plan <- wf_plan_weights(design, target, dims, method = "raking")
approval <- wf_approve_plan(
  plan,
  approver = "Reviewer full name",
  role = "Statistician"
)
locked <- wf_execute_plan(plan, approval, design, target)

diagnostics <- wf_diagnose(locked, target = target)
```

Direct statistical use is also verified:

```{r direct, eval=FALSE}
via_dispatcher <- wf_calibrate(
  design,
  target,
  method = "raking"
)
```

---
title: "Review tools for verified weighting"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Review tools for verified weighting}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

WFC review tools explain feasibility and weight distortion without changing the
verified target.

```{r review, eval=FALSE}
precheck <- wf_precheck(
  design$data,
  target,
  id = design$roles$id
)

trim_advice <- wf_auto_trim(
  design,
  target,
  caps = c(2, 4, 6, 8),
  max_deff = 6,
  max_residual = 0.02
)
```

`wf_auto_trim()` returns a frontier and recommendation. It does not apply a cap
to a production result. A reviewed plan must contain the final bounds.

After weights are locked, decision readers receive a compact status and next
actions; statistical readers receive full diagnostics and provenance:

```{r reports, eval=FALSE}
wf_report(locked, audience = "decision")
wf_report(locked, audience = "statistician")
```

Both views are generated from the same object.

---
title: "Importing external targets with evidence"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Importing external targets with evidence}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

WFC 2.0 imports targets from a finalized CSV or Excel table and a companion DCF
evidence record. It does not construct a production target from an unexplained
runtime table.

```{r template, eval=FALSE}
dims <- wf_dims(
  age_group = c("18-34", "35-54", "55+"),
  region = c("north", "south")
)

wf_target_template("population-margins.csv", dims)
```

Complete the companion `population-margins.csv.source.dcf` with publisher,
dataset title, citation, reference period, population scope, retrieval date,
license, transformation, selection timing, demo status, and the final SHA-256.

```{r import, eval=FALSE}
target <- wf_import_target(
  "population-margins.csv",
  "population-margins.csv.source.dcf",
  dims,
  key_map = c(age_group = "age_group", region = "region"),
  count = "population_count",
  production = TRUE
)
```

For joint-cell post-stratification, the imported population table must contain
the required joint combinations. For an independent reference sample, use
`wf_import_reference()` with its own evidence file.

The checksum proves that the imported data match the evidence record. It does
not prove that the publisher, transformation, or scientific choice is valid;
those remain review questions.

---
title: "Guided verified weighting in WFC 2.0"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Guided verified weighting in WFC 2.0}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

`wf_autoweigh()` gives survey researchers a narrated path over the verified
engines. It no longer builds a target from raw runtime data. Prepare `design`,
import `target`, and declare `dims` first.

```{r guided, eval=FALSE}
guided <- wf_autoweigh(
  design,
  target,
  dims,
  method = "raking",
  trim = NULL,
  interactive = FALSE,
  lang = "en"
)

guided$weights
guided$diagnostics
guided$ledger
```

The ledger uses stable English programmatic keys while its `detail` text may be
localized. Artifacts align with ledger rows so a statistician or agent can
inspect each precheck, recommendation, and report.

Automatic mode chooses only from declared safe settings. It does not inspect
outcomes, change the target, widen bounds, or manufacture approval.

```{r views, eval=FALSE}
wf_report(guided$weights, audience = "decision", lang = "zh_CN")
wf_report(guided$weights, audience = "statistician", lang = "en")
```

For production, prefer `wf_plan_weights()` plus separate
`wf_approve_plan()` and `wf_execute_plan()` so human approval is bound to one
immutable plan.

# WFCstudio Integration Contract

## Status and Ownership

This document freezes the WFC 1.0 integration boundary for the separate
WFCstudio sibling package. WFCstudio implementation does not live in the WFC
repository. Copyright in both the contract and the WFC project belongs to WEIAN
DATA TECH (Beijing) Co., Ltd.

The initial WFCstudio beta targets WFC `>= 1.0.0, < 2.0.0`. The core WFC
repository defines what the UI may call and what it may not compute; the sibling
package owns its own release, tests, and UI implementation.

## Architectural Rule

WFCstudio is a user interface over WFC, not a second statistical engine.

It may:

- ingest user-selected csv/xlsx files through sibling-package dependencies;
- collect mappings for dimensions, group keys, counts, IDs, and reviewed
  controls;
- construct ordinary arguments for exported WFC functions;
- call `wf_autoweigh()` and expert-path WFC functions;
- render structured WFC results, reports, ledgers, artifacts, plots, warnings,
  and classed errors;
- export user-authorized weights and reports.

It must not:

- implement raking, calibration, post-stratification, trimming, propensity,
  fusion, replication, or diagnostics;
- bypass `wf_precheck()` or catch and suppress blocking WFC conditions;
- invent collapse maps, ladders, bounds, or publication thresholds that the
  user did not review or declare;
- translate or rename stable WFC object fields, condition classes, actions, or
  ledger detail keys;
- write source data outside the user-selected session/export boundary.

## Minimum Workflow

The first beta workflow is:

1. Select sample and population files.
2. Map sample/population columns to `wf_dims()`, `key_map`, `count`, `by`, and
   `id` inputs.
3. Select a supported WFC language and guided controls.
4. Call `wf_autoweigh()` without changing its decisions.
5. Render the manager report and ordered decision ledger.
6. Allow inspection of precheck, collapse, trim, diagnostic, and report
   artifacts.
7. Export weights and the selected report only after explicit user action.

## Stable Inputs

WFCstudio targets the exported WFC 1.0 interfaces:

- `wf_dims()` and target constructors;
- `wf_autoweigh()`;
- `wf_report()` and plot methods;
- `as_svydesign()` / `as_svrepdesign()` for downstream ecosystem handoff;
- `wf_pipeline()`, `wf_run()`, `wf_validate()`, and `wf_audit_export()` for
  repeatable production rounds;
- `wf_autoweigh_result`, `wf_quality_report`, and WFC classed conditions.

The sibling package owns file-reading, reactive UI state, and presentation. It
passes ordinary data frames and scalar/list controls to WFC. Private source
files are never bundled with either package.

## Output Contract

The UI renders localized human-facing fields but preserves stable machine
fields:

- result `$method` and `$language`;
- ledger `step`, `action`, `detail_key`, `detail`, `artifact_class`, and `time`;
- manager-report English column keys and localized label/action values;
- artifact classes and classed condition names/payloads;
- exported unit `id`, `group`, `weight`, and `feature` fields.

WFCstudio may add presentation-only state outside these objects. It must not
mutate the WFC objects it receives.

## Localization

WFC remains the authority for localized weighting narration, report labels,
verdicts, actions, and plot labels. WFCstudio may read the installed DCF
catalogs for matching UI labels, but WFC catalog keys are not statistical
controls and never change programmatic semantics. Unsupported UI languages
fall back according to the WFC language contract.

## Conditions and Safety

WFCstudio reacts to WFC's stable condition classes:

- `wf_error_input`: highlight the invalid control;
- `wf_error_schema`: return the user to column/ID mapping;
- `wf_error_feasibility`: stop weighting and show the declared-remediation
  requirement;
- `wf_error_dependency`: explain the optional downstream package requirement;
- `wf_warning_data` / `wf_warning_quality`: render without suppressing.

No UI action may convert an error into a warning or continue after a blocking
condition unless WFC itself returns a successful result.

## Sibling-Package Release Gates

WFCstudio kickoff is complete only when its separate repository has:

- company ownership and WEIAN DATA commit identity enforcement;
- an explicit dependency on a compatible WFC release;
- contract tests using simulated WFC data only;
- bilingual UI coverage without translated stable keys;
- upload/export privacy tests;
- no statistical implementation outside WFC;
- an end-to-end beta flow against a real, user-authorized production cycle.

These remain external gates for the WFCstudio beta and do not add GUI code or
hard dependencies to the WFC core package.

# weightflow — Design Document, Extension 1

**Cell-level post-stratification with adaptive nested collapsing**

Companion to `weightflow_design.md`. Target version: 0.2.0. This extension adds a second calibration method (`method = "poststrat"`) on the reserved Seam 2, plus its precheck-side collapse planner. It introduces no breaking change: every 0.1 class, constructor, and condition is reused unchanged; the only additive change to an existing class is one optional field on `wf_target` (`joint`), described in §3.

The scope here is deliberately narrow — **single-source** post-stratification only. Fusing an online (raked) source with an offline (post-stratified) source is a separate Seam 3 concern (`wf_blend()`) and is explicitly out of scope for this document, per the requirement that `wf_poststrat()` remain single-purpose and independently testable.

---

## 1. Why post-stratification, and how it differs from raking

Raking (0.1) aligns **marginal** totals: for the current project, 2 + 2 + 5 + 5 = 14 numbers per province. It never promises that the 2×2×5×5 = 100 **joint cells** match the population, and in practice they do not — which is exactly the gap the user observed after raking (a cell's weighted total ≠ that cell's population head count).

Post-stratification aligns the **joint cells** directly: each cell's weighted total is forced to equal the population count for that cell. This is strictly stronger than raking — if every cell matches, every margin matches automatically, and the province total is preserved. It is also simpler (one closed-form scaling per cell, no iteration).

Post-stratification has exactly two failure modes, and everything in this extension exists to manage them:

1. **Empty cells** (zero sample support): a population-positive cell with no respondents. Pure post-stratification would assign weight 0 to a cell that should represent real people. Unsolvable by scaling; must be handled structurally (§7).
2. **Sparse cells** (few respondents): a cell supported by 1–4 respondents gets an enormous, unstable weight, inflating the design effect. Managed by collapsing (§5).

The method therefore is not "post-stratify the 100 cells" but "post-stratify the *finest partition each cell can support*, falling back through a declared ladder where support is thin, and redistributing where support is absent."

---

## 2. Requirements captured (from user decisions)

| # | Decision | Consequence in this design |
|---|---|---|
| 1 | Sparsity threshold = **user-supplied minimum cell sample count**; fallback ladder is **pre-declared** (never auto-invented) | `min_cell` is a required numeric; collapsing consumes a declared `wf_collapse_ladder` (§4). No implicit merging. |
| 2 | Collapsing is **adaptive first** (per-cell nesting), **falls back to province-uniform** partition if adaptive still leaves problems | `granularity = "adaptive"` default, with automatic degradation to `"province"` when adaptive cannot resolve all cells (§6). |
| 3 | Output must include an **audit/report-friendly mapping table** | `wf_poststrat` returns `cell_report` + `collapse_map` (§8). |
| 4 | **Calibration-style** weighting (respect initial weights; post-strat multiplies a calibration factor) | `mode = "calibration"` is the only supported mode here; initial weights are the base, not overwritten (§7). |
| 5 | Empty cells → **redistribute** their population to other cells | `empty_cell = "redistribute"` default (§7.3). |
| 6 | **Single-source only**; no fusion inside this function | No `w_online`/`w_offline` arguments; fusion deferred to `wf_blend()`. |
| 7 | The function **outputs and enforces a province-total constraint** | Province (group) totals are a hard post-condition, asserted and reported (§7.4). |

---

## 3. Data model additions

### 3.1 One optional field on `wf_target`: `joint`

The 0.1 `wf_target` stores only per-group **margins**. Post-stratification needs the full **joint** cell table. Rather than a new class, we add one optional element:

```r
tgt$joint  # named list, one element per group; each is a data.frame:
           #   columns = the dimension variables (character keys) + `pop`
           #   one row per non-empty population cell
```

`joint` is populated by `wf_target_population(..., keep_joint = TRUE)`. When absent (0.1 behavior, `keep_joint = FALSE` default), raking still works and `wf_poststrat()` raises `wf_error_schema` instructing the user to rebuild the target with `keep_joint = TRUE`. Margins remain present and are used for the province-total constraint and for validation, so no downstream 0.1 code path changes.

Rationale for a joint **data.frame** (not a 4-D array): the classification is schema-agnostic (any number of dimensions, any level counts), arrays would require dynamic `dim`/`aperm` gymnastics and blow up memory for high-dimensional cross-classifications, whereas a long data.frame of only non-empty cells is compact (structural-zero cells never appear) and matches the sample side, which is also long. Cell identity everywhere is the ordered tuple of character keys, hashed to a single string key (`.wf_cell_key()`).

### 3.2 New classes

| Class | Constructor | Role |
|---|---|---|
| `wf_collapse_ladder` | `wf_collapse_ladder()` | An **ordered, pre-declared** sequence of collapse steps (each step a per-dimension category map). Defines the *only* merges the planner is permitted to apply. |
| `wf_poststrat_plan` | `wf_plan_poststrat()` | Per-group, per-cell resolution: which ladder level each original cell is evaluated at, and the resulting partition. Reviewable before execution. |

`wf_weights` (returned by `wf_poststrat`) is the **same class** as raking returns, so `wf_diagnose()` and any pipeline code consume it identically. Two named fields are added to its list (`cell_report`, `collapse_map`); the print method gains a one-line post-strat summary via method dispatch on `provenance$method`.

---

## 4. Pre-declared collapse ladders (`wf_collapse_ladder`)

The ladder is the user's explicit contract for *how* cells may be coarsened. It is an ordered list of levels 0..L, where level 0 is the finest (the raw cells) and each subsequent level applies one additional declared merge:

```r
ladder <- wf_collapse_ladder(
  dims,                                  # the wf_dims object (for validation)
  level1 = list(edu5 = c(`1`="12", `2`="12", `3`="3", `4`="45", `5`="45")),
  level2 = list(age5 = c(`1`="1", `2`="2", `3`="3", `4`="45", `5`="45")),
  level3 = list(edu5 = c(`12`="LM", `3`="LM", `45`="H"))   # merges are cumulative
)
```

Semantics and guarantees:

- **Ordered and cumulative.** Level *k* is level *k−1* with its additional maps applied. `wf_plan_poststrat()` walks *up* the ladder cell by cell until support is met.
- **Validated at construction.** Every map key must be a declared level of its dimension (or a category produced by an earlier ladder level, enabling multi-step merges as in `level3`). Unknown categories → `wf_error_input`. This prevents silent typos from becoming silent merges.
- **Declared, never inferred.** There is no `"auto"` ladder. If the ladder is exhausted and a cell is still short, the engine does not invent a coarser partition; it degrades to province-uniform (§6) and, failing that, treats the residual as empty (§7.3).
- **Reusable object.** A ladder validated once can be reused across surveys sharing the same schema, and is serialized into provenance.

---

## 5. Sparsity resolution algorithm

For one group, with user threshold `min_cell` (minimum respondent count per resolved cell):

1. Compute raw cell counts at ladder level 0 (finest) from the sample.
2. **Adaptive pass** (`granularity = "adaptive"`): each original cell is assigned the *lowest ladder level at which the coarsened cell it belongs to meets `min_cell`*. Cells with ample support stay at level 0 (full 100-cell resolution); only thin cells climb. Because merges are cumulative and monotone (coarser cells have ≥ the support of any finer cell within them), a cell's minimal satisfying level is well-defined; if no ladder level satisfies it, the cell is flagged `unresolved`.
3. If any cell is `unresolved`, degrade this group to the **province-uniform pass** (§6).
4. Cells still unresolved after province-uniform (i.e. even the coarsest ladder level has < `min_cell`, or the coarsened cell has zero support) are handed to empty-cell handling (§7.3).

The threshold is on **sample count**, exactly as specified (requirement 1) — not on effective sample size — keeping the rule transparent and independent of the weighting result. (An `neff`-based option is noted as a future switch in §11 but is not implemented in 0.2.)

---

## 6. Granularity: adaptive with province-uniform fallback (requirement 2)

`granularity = "adaptive"` (default) implements §5 step 2: a *mixed-resolution* partition where different cells within the same province sit at different ladder levels. This maximizes retained detail — well-supported cells keep full joint calibration; only thin regions are coarsened. The cost is a more complex partition to report, which is exactly why `cell_report` (§8) exists.

When the adaptive pass leaves any cell unresolved (§5 step 3), the group **degrades to `granularity = "province"`**: a single ladder level is chosen for the *entire* province — the lowest level at which **all** cells meet `min_cell`. This trades detail for guaranteed uniform feasibility and a simpler partition. The degradation is automatic, recorded in provenance (`granularity_used` per group may differ from the requested value), and surfaced as a `wf_warning_quality`. A user may also request `granularity = "province"` directly to skip the adaptive attempt.

This ordering — try adaptive, fall back to uniform — is the literal encoding of the user's decision 2.

---

## 7. Execution: `wf_poststrat()`

### 7.1 Signature

```r
wf_poststrat(
  sample, target,
  init_weight = NULL,                 # column name; NULL => all 1 (req. 4 base)
  min_cell,                           # REQUIRED integer >= 1 (req. 1)
  ladder,                             # wf_collapse_ladder (req. 1)
  granularity = c("adaptive","province"),   # (req. 2)
  empty_cell = c("redistribute","flag","error"),  # default redistribute (req. 5)
  id = NULL,
  precheck = TRUE,
  tol = 1e-8                          # province-total assertion tolerance (req. 7)
)
```

### 7.2 Calibration-style weighting (requirement 4)

Weighting is calibration, not replacement. For each resolved cell *c* in group *g*:

```
factor_c = pop_c / sum(init_weight over respondents in c)
final_weight_i = init_weight_i * factor_c        for i in cell c
```

The initial weight is honored as the base; post-stratification applies a single multiplicative calibration factor per resolved cell. With `init_weight = NULL` (all 1s) this reduces to classic post-stratification (weight = pop_c / n_c), so the classic case is a special case of the calibration formula rather than a separate code path. Within a coarsened (resolved) cell, the ratio of two respondents' final weights equals the ratio of their initial weights — i.e. the design information carried by `init_weight` is preserved through the merge, which is the whole point of doing this calibration-style for an offline design sample.

### 7.3 Empty-cell redistribution (requirement 5, option b)

A resolved cell with population > 0 but **zero** sample support (even at the coarsest ladder level) cannot receive weight. Its population is **redistributed** to the cells that *do* have support, proportionally to those cells' population, **within the smallest enclosing coarser cell that has support** — not globally across the province, which would distort unrelated regions. Concretely: climb one more ladder level to find the nearest supported ancestor cell, and fold the orphan population into that ancestor's target before computing factors. If even the whole province has no support anywhere (degenerate), raise `wf_error_feasibility`.

Every redistribution is logged in `cell_report` (orphan cell, receiving cell, population moved), so the analyst can see exactly which population was reallocated and audit whether the exchangeability assumption implied by the redistribution is acceptable. `empty_cell = "flag"` instead marks orphan cells and leaves their population unassigned (province total will then be short by the orphan population — reported, not silently absorbed); `"error"` refuses.

### 7.4 Province-total constraint (requirement 7)

After weighting, for every group the sum of final weights must equal the group's population total (the sum over its joint cells, which by §3 equals the sum of its margins). This is asserted to within `tol`; violation raises `wf_error_internal` (it would indicate a redistribution or factor bug). Under `empty_cell = "flag"`, the constraint target is reduced by the flagged orphan population and the shortfall is reported explicitly. The realized per-group total and its deviation from the population total are returned in the group log, so the constraint is both **enforced and reported**, as required.

### 7.5 Algorithm summary (per group)

resolve levels (§5/§6) → build resolved-cell key for every respondent → redistribute orphan population into supported ancestors (§7.3) → compute per-cell calibration factors on `init_weight` (§7.2) → assert province total (§7.4) → assemble `cell_report`. All grouped sums use `rowsum()` on integer cell indices; no iteration is needed (post-strat is closed-form), so this is markedly faster than raking.

---

## 8. Outputs: audit and reporting (requirement 3)

`wf_poststrat()` returns a `wf_weights` object with:

- `data`: `id`, `group`, resolved-cell key, `weight`, `feature = 1/weight`.
- `cell_report` (**the audit table**): one row per *original* (level-0) population cell, with columns `group`, the dimension keys, `pop`, `n_sample`, `ladder_level_used`, `resolved_cell` (the key of the coarsened cell it was calibrated within), `granularity_used`, `orphan` (logical), `redistributed_to`, `pop_moved`, `final_cell_weight_mean`. This answers "which original cells were merged into what, why (support), and with what population movement" at a glance.
- `collapse_map`: a compact `wf_poststrat_plan` mapping (group × original-cell → resolved-cell + level), reusable and serializable.
- `log`: per group — `n`, `n_cells_raw`, `n_cells_resolved`, `granularity_used`, `n_orphan_cells`, `pop_redistributed`, `total_target`, `total_realized`, `total_dev`, `deff`.
- `provenance`: as in 0.1, plus `method = "poststrat"`, `min_cell`, ladder hash, per-group `granularity_used`.

`wf_diagnose()` consumes this unchanged (design effect, ESS, verdicts per group). A `summary()` method on `cell_report` prints the merge/redistribution summary for a methods appendix.

---

## 9. Precheck integration (`wf_plan_poststrat()`, discipline preserved)

Mirroring 0.1's precheck→execute discipline, `wf_plan_poststrat()` runs the resolution logic **without weighting** and returns the `wf_poststrat_plan` plus a diagnostic table (per group: cells needing collapse, cells reaching each ladder level, orphan cells, whether adaptive suffices or province degradation is forced). The user reviews the plan, and may pass it directly to `wf_poststrat(plan = ...)` to skip re-resolution. `wf_poststrat(precheck = TRUE)` runs it internally and aborts on structural infeasibility (e.g. empty ladder against present sparsity with `empty_cell = "error"`). The loop is therefore: `wf_plan_poststrat → (review) → wf_poststrat → wf_diagnose`.

---

## 10. Error and condition system (reused, extended)

No new condition classes — the 0.1 taxonomy suffices. New situations map onto existing classes:

| Situation | Class | Severity |
|---|---|---|
| Target built without `keep_joint` | `wf_error_schema` | error |
| `min_cell` missing / < 1 / non-integer | `wf_error_input` | error |
| Ladder map references unknown category | `wf_error_input` | error |
| Cell unresolved and `empty_cell = "error"` | `wf_error_feasibility` | error |
| Whole group has no support anywhere | `wf_error_feasibility` | error |
| Province total assertion fails | `wf_error_internal` | error (bug) |
| Adaptive degraded to province-uniform | `wf_warning_quality` | warning |
| Orphan population redistributed | `wf_warning_data` | warning |
| Resolved cell still thin after coarsening | `wf_warning_quality` | warning |

All carry structured `data` payloads (group, cell key, level, populations) for programmatic handling, consistent with P5.

---

## 11. Performance and extensibility reserve

Post-stratification is closed-form: one `rowsum()` per group for cell counts, one for initial-weight sums, vector arithmetic for factors — O(n) per group, no iteration, faster than raking. Resolution is precomputed once as integer level assignments. Cell keys are built once via a single `paste()` on character columns, then `match()`ed to integer indices; the hot path touches no strings. Groups remain embarrassingly parallel (reserved `parallel = TRUE`, 0.2+).

Reserved switches, none altering the 0.2 API: `min_cell_by = "neff"` (threshold on effective sample size rather than raw count); `redistribute = "model"` (hand orphan cells to an MRP-style predictor instead of proportional reallocation — the natural bridge to a future `method = "mrp"` on Seam 2); and registration of `"poststrat"` in the `.wf_methods` registry so `wf_calibrate(method = "poststrat")` dispatches to it, unifying the method interface with raking.

---

## 12. Public API additions (0.2)

```
wf_collapse_ladder()   declare the ordered, cumulative, validated merge ladder
wf_plan_poststrat()    resolve cells to ladder levels WITHOUT weighting (precheck)
wf_poststrat()         calibration-style cell post-stratification; province-total enforced
                       + adaptive→province-uniform granularity
                       + empty-cell redistribution into nearest supported ancestor
                       + cell_report / collapse_map audit outputs
summary.wf_poststrat_plan(), summary(cell_report)   reporting helpers
```

`wf_target_population()` gains one argument: `keep_joint = FALSE`. Everything else in 0.1 is untouched.

---

The reference implementation of the above is `weightflow_poststrat.R`, sourced after `weightflow_core.R`.

# WFC 1.0 API Freeze

This document records the public WFC core API frozen at version 1.0.0.
It is installed with the package so downstream tools, WFCstudio, and release
reviewers can inspect the compatibility contract without reading development
design documents.

## Compatibility Policy

From 1.0.0 onward, WFC follows these rules:

- Additive changes are allowed in minor releases: new exported functions, new
  optional arguments with backward-compatible defaults, new S3 methods, new
  columns, and new list fields.
- Removals, renamed exported functions, required new arguments, changed default
  behavior, changed class inheritance, or changed meanings of existing
  programmatic fields require a major version.
- Deprecated APIs must warn with class `wf_warning_deprecated` for at least one
  full minor release before removal.
- Human-facing wording may change, including localized labels and report prose.
  English object names, condition classes, ledger keys, and table column names
  are the stable programmatic interface.
- Optional ecosystem integrations remain optional. Missing suggested packages
  must continue to fail with classed WFC dependency errors rather than becoming
  hard dependencies.

## Frozen Exported Signatures

```r
as_svrepdesign(r, data, id = "id", degf = NULL, ...)
as_svydesign(w, data, id = "id", ids = ~1, strata = NULL, fpc = NULL,
             nest = FALSE, ...)
wf_apply_collapse(sample, target, plan)
wf_attrition(panel, retained, formula, id = NULL, by = NULL,
             stabilize = TRUE, trim = NULL)
wf_audit_export(x, file, inputs = NULL, extra = NULL)
wf_auto_trim(sample, target, id = NULL,
             caps = c(2, 3, 4, 5, 6, 8, 10, 12),
             lo = 0.05, max_deff = 6, max_residual = 0.02, ...)
wf_autoweigh(sample, population, dims, key_map = NULL, count = NULL,
             by = NULL, id = NULL,
             method = c("auto", "raking", "poststrat", "logit"),
             ladder = NULL, min_cell = NULL, bounds = c(0.3, 3),
             trim = "auto", max_deff = 6, max_residual = 0.02,
             interactive = base::interactive(), lang = NULL, ...)
wf_blend(online, offline, by_cell,
         lambda = c("neff", "inverse_variance", "fixed"),
         lambda_fixed = NULL, outcome = NULL,
         level = c("cell", "group"), trim_lambda = c(0.05, 0.95),
         sensitivity = TRUE)
wf_calibrate(sample, target, method = "raking", ...)
wf_collapse_ladder(dims, ...)
wf_compose(..., id = NULL, normalize = c("none", "mean1", "sum"))
wf_diagnose(w, target = NULL, sample = NULL, deff_ok = 3,
            deff_caveat = 10)
wf_dims(..., .collapse = list())
wf_influence(w, target = NULL, sample = NULL, id = NULL, top = 20)
wf_pipeline(target, stages, validate = NULL)
wf_plan_poststrat(sample, target, min_cell, ladder,
                  granularity = c("adaptive", "province"),
                  empty_cell = c("redistribute", "flag", "error"),
                  id = NULL)
wf_poststrat(sample, target, min_cell, ladder, init_weight = NULL,
             granularity = c("adaptive", "province"),
             empty_cell = c("redistribute", "flag", "error"),
             id = NULL, precheck = TRUE, tol = 1e-8,
             parallel = FALSE, progress = FALSE)
wf_precheck(sample, target, id = NULL,
            na = c("fractional", "drop", "error"),
            max_na_dims = 2, thin_min = 5, risk_ratio = 10)
wf_propensity(target, weight = c("ipw", "kernel", "matching"),
              stabilize = TRUE, trim = NULL)
wf_rake(sample, target, id = NULL,
        na = c("fractional", "drop", "error"),
        trim = NULL, trim_cycles = 4, tol = 1e-6, max_iter = 200,
        precheck = TRUE, init_weight = NULL,
        parallel = FALSE, progress = FALSE)
wf_replicates(data, refit,
              method = c("bootstrap", "jackknife", "brr"),
              R = 500, strata = NULL, clusters = NULL, id = NULL,
              base_weight = NULL, seed = NULL, rho = 0,
              parallel = FALSE, progress = FALSE)
wf_report(w, target = NULL, audience = c("manager", "analyst"),
          lang = NULL, output = c("object", "markdown", "html"),
          file = NULL)
wf_run(spec, sample, dims = NULL, population = NULL, reference = NULL,
       margins = NULL, base_weight = NULL)
wf_suggest_collapse(precheck, dims,
                    checks = c("cat_infeasible", "support_thin",
                               "risk_extreme_ratio"),
                    max_steps = 1)
wf_suggest_ladder(sample, target, dims, min_cell = 5)
wf_target_manual(margins, dims, dim_col = "dimension",
                 cat_col = "category", value_col = "value", by = NULL,
                 group_col = by, totals = NULL, mode = "manual")
wf_target_population(pop, key_map, count, dims, by = NULL, by_key = NULL,
                     scale = c("population", "sample", "custom"),
                     sample = NULL, totals = NULL, keep_joint = FALSE)
wf_target_propensity(online, reference, formula,
                     method = c("logit", "rf", "gbm"),
                     by = NULL, id = NULL)
wf_target_reference(ref, feature, dims, by = NULL,
                    feature_na = c("error", "drop"),
                    feature_gt1 = c("warn", "allow"))
wf_target_shrink(target, reference, lambda, groups = NULL)
wf_validate(new, reference, target = NULL, max_deff_delta = 1,
            max_ess_loss = 0.2, max_total_shift = 0.05,
            max_margin_delta = 0.01, max_ratio_p99 = 2,
            on_issue = c("warn", "error", "none"))
wf_variance(replicates, estimator, data, level = 0.95,
            ci = c("normal", "percentile"))
```

## Stable Object Contracts

The following fields are stable when present on their corresponding objects:

- `wf_target`: `mode`, `by`, `dims`, `groups`, `meta`, `joint`.
- `wf_weights`: `data`, `log`, `achieved`, `provenance`; `data` includes at
  least `id`, `group`, `weight`, and `feature`.
- `wf_replicate_weights`: `base`, `replicates`, `scale`, `rscales`, `method`,
  `rho`, `design`, and `provenance`.
- `wf_diagnostics`: `table` with English programmatic column names.
- `wf_quality_report`: `summary`, `sections`, `audience`, `language`, and
  structured section tables.
- `wf_autoweigh_result`: `weights`, `diagnostics`, `report`, `target`,
  `final_sample`, `ledger`, `method`, and `language`.
- `wf_pipeline`: `target`, `stages`, `validate`, `created`, and `hash`.
- `wf_validation`: `ok`, `issues`, `comparison`, `ratio`, and `thresholds`.

Additive fields may appear in these objects. Existing fields must not be
renamed, removed, or repurposed without a major version.

## Stable Condition Classes

The public classed-condition taxonomy is:

- Errors: `wf_error`, `wf_error_input`, `wf_error_schema`,
  `wf_error_feasibility`, `wf_error_convergence`, `wf_error_design`,
  `wf_error_dependency`, and `wf_error_internal`.
- Warnings: `wf_warning`, `wf_warning_data`, `wf_warning_quality`, and
  `wf_warning_deprecated`.

Condition messages are human-facing and may be improved. Condition classes and
machine-readable `data` payload names are the stable programmatic surface.

## WFCstudio Boundary

WFCstudio beta targets WFC `>= 1.0.0, < 2.0.0`. It may call exported WFC APIs
and render WFC objects, but it must not implement statistical engines, suppress
blocking WFC conditions, rename stable keys, or add hard dependencies to WFC.
The detailed sibling-package contract is in `inst/design/wfcstudio_contract.md`.

## WFC 1.1 Additive Safety API

WFC 1.1 adds the following exported interfaces without changing the frozen 1.0
numeric contracts:

```r
wf_prepare_design(data, id, calibration, base_weight = NULL,
                  strata = NULL, clusters = NULL, fpc = NULL)
wf_target_template(file, dims, by = NULL, example = FALSE)
wf_import_target(data_file, source_file, dims, key_map, count,
                 by = NULL, by_key = NULL, production = TRUE)
wf_import_reference(data_file, source_file, dims, feature,
                    by = NULL, production = TRUE)
wf_plan_cells(design, target, dims, min_cell = 5,
              max_weight_ratio = 4, boundary = target$by, ladder = NULL)
wf_plan_weights(design, target, dims,
                method = c("raking", "logit", "poststrat"),
                bounds = c(0.3, 3), min_cell = 5, cell_plan = NULL)
wf_approve_plan(plan, approver, role, note = NULL, actor_type = "human")
wf_execute_plan(plan, approval, design, target)
wf_guided_plan(data, id, calibration, dims, target_file, source_file,
               source_type = c("population", "reference"), key_map = NULL,
               count = NULL, feature = NULL, ...)
wf_guided_execute(workflow, approval)
wf_attach_weights(data, weights, id, weight_name = ".weight")
wf_assess_impact(weights, data, id, outcomes, level = 0.95)
```

`wf_report()` additionally accepts `audience = "decision"` as an alias for
`"manager"` and `audience = "statistician"` as an alias for `"analyst"`. Audit
exports use `wfc_audit_v2`, preserving the v1 fields and adding nullable safety
identity fields.

The new safety objects add stable SHA-256 `identity` fields. Safety errors use
class `wf_error_safety` and a machine payload with `code`, `severity`, `field`,
`evidence`, and `next_actions`.

The following 1.0 interfaces remain numerically compatible in 1.1 but warn with
`wf_warning_deprecated` on every use and are scheduled for removal in 2.0.0:

- manual target margins;
- target shrinkage;
- inline entropy-balancing moment targets;
- manual pipeline target declarations and runtime margin injection.

Their warning payload adds `removal` and `risk_code` to the frozen `feature` and
`replacement` fields. The verified workflow has no bypass, self-approval,
automatic limit widening, automatic relaxation, or silent method-switch path.

---
title: "Attrition and influence after a verified design"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Attrition and influence after a verified design}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

Attrition and influence diagnostics can support sensitivity review, but they do
not replace the verified external target or authorize outcome-driven reweighting.

After a weight plan is locked, inspect influential units without changing the
weights:

```{r influence, eval=FALSE}
influence <- wf_influence(
  locked,
  target = target,
  sample = design$data,
  id = design$roles$id,
  top = 20
)

influence$table
```

If panel-retention weights are studied, keep their formula, stabilization, trim
settings, and balance diagnostics as a separate sensitivity analysis. Do not
select those settings to make an outcome cross a preferred threshold.

```{r impact, eval=FALSE}
impact <- wf_assess_impact(
  locked,
  analysis_data,
  id = "person_id",
  outcomes = c("score", "approved")
)

wf_report(impact, audience = "statistician")
```

---
title: "Verified, outcome-blind weighting in WFC 2.0"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Verified, outcome-blind weighting in WFC 2.0}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", message = FALSE)
```

<!-- SAFE_WORKFLOW_START -->

This workflow keeps study outcomes out of target selection, cell merging, and
weight planning. It has two entry styles:

- practitioners can use `wf_guided_plan()` to compose the safe planning steps;
- statisticians and AI Agents can call each lower-level step and inspect every
  intermediate object.

Both styles stop before approval. A separate human attestation is required to
execute a plan. Decision makers can receive a compact report, while
statisticians can receive the complete evidence tables.

## Start with design-only and analysis data

The design table contains only fields with declared sampling or calibration
roles. Outcomes remain in a separate table joined later by a unique ID.

```{r}
library(WFC)

dims_safe <- wf_dims(
  sex = c("F", "M"),
  age = c("18-34", "35+")
)
design_frame <- data.frame(
  id = sprintf("r%02d", 1:16),
  sex = rep(c("F", "M"), 8),
  age = rep(c("18-34", "18-34", "35+", "35+"), 4),
  base_weight = 1,
  stringsAsFactors = FALSE
)
analysis_frame <- data.frame(
  id = design_frame$id,
  satisfaction = seq(40, 70, length.out = 16),
  approved = rep(c(0, 1), 8),
  stringsAsFactors = FALSE
)

design <- wf_prepare_design(
  design_frame,
  id = "id",
  calibration = c("sex", "age"),
  base_weight = "base_weight"
)
design
```

If `analysis_frame` or either outcome column were passed to
`wf_prepare_design()`, construction would stop with `wf_error_safety` because
the outcome column has no permitted design role.

## Import a target with its own evidence record

WFC installs matching synthetic CSV and Excel examples:

- `safe-target-example.csv` and `safe-target-example.csv.source.dcf`;
- `safe-target-example.xlsx` and `safe-target-example.xlsx.source.dcf`.

Each data file has a distinct SHA-256, even though the tables have identical
content. The DCF record names the publisher, dataset, citation, reference
period, population scope, retrieval date, license, transformation, selection
timing, and demo status.

```{r}
csv_file <- system.file(
  "extdata", "safe-target-example.csv", package = "WFC"
)
csv_source <- paste0(csv_file, ".source.dcf")
cat(paste(readLines(csv_source, warn = FALSE), collapse = "\n"))

target_verified <- wf_import_target(
  csv_file,
  csv_source,
  dims_safe,
  key_map = c(sex = "sex", age = "age"),
  count = "count",
  production = FALSE
)
target_verified$identity
```

`production = FALSE` is required only because the bundled file is explicitly a
demo. Do not use that setting to import an authoritative production source that
has incomplete evidence.

Excel uses the same import form and its own companion evidence file:

```{r}
xlsx_file <- system.file(
  "extdata", "safe-target-example.xlsx", package = "WFC"
)
if (requireNamespace("openxlsx", quietly = TRUE)) {
  target_from_excel <- wf_import_target(
    xlsx_file,
    paste0(xlsx_file, ".source.dcf"),
    dims_safe,
    key_map = c(sex = "sex", age = "age"),
    count = "count",
    production = FALSE
  )
  identical(target_verified$groups, target_from_excel$groups)
}
```

Use `wf_target_template()` to create a blank data file and companion DCF for a
new source. Blank metadata and `selected_before_outcomes: false` deliberately
block import until an accountable person completes the evidence record and
updates the checksum.

## Plan without calculating weights

The bundled example stops here by design: a `demo_only` target can demonstrate
import and checksum verification but cannot enter planning. The remaining code
is a runnable template after `csv_file` and `csv_source` are replaced with a
completed, authoritative external source and `wf_import_target()` is called
with its default `production = TRUE`. It is not evaluated during vignette build
because no synthetic file may impersonate that authority.

Cell planning uses only sample support, base weights, declared category order or
an explicit ladder, and the verified target. It cannot receive outcomes or a
custom score. It never crosses the target boundary or widens limits.

```{r controlled-plan, eval=FALSE}
cell_plan <- wf_plan_cells(
  design,
  target_verified,
  dims_safe,
  min_cell = 5,
  max_weight_ratio = 4
)
plan <- wf_plan_weights(
  design,
  target_verified,
  dims_safe,
  method = "raking",
  bounds = c(0.3, 3),
  min_cell = 5,
  cell_plan = cell_plan
)

plan$precheck
is.null(plan$weights)
```

The practitioner composition produces the same kind of reviewable plan and
still computes no weights:

```{r guided-plan, eval=FALSE}
guided <- wf_guided_plan(
  design_frame,
  id = "id",
  calibration = c("sex", "age"),
  dims = dims_safe,
  target_file = csv_file,
  source_file = csv_source,
  source_type = "population",
  key_map = c(sex = "sex", age = "age"),
  count = "count",
  base_weight = "base_weight",
  production = FALSE
)

wf_report(guided, audience = "decision")$table
names(wf_report(guided, audience = "statistician")$sections)
```

## Keep approval separate from execution

The next lines are intentionally not run while building this vignette. They
represent two accountable actions: a human reviewer attests approval, then the
approved plan is executed exactly once. Replace the illustrative identity with
the actual reviewer; an automated process must not fill it in.

```{r approval-and-execution, eval=FALSE}
approval <- wf_approve_plan(
  plan,
  approver = "Reviewer name",
  role = "statistician"
)
locked <- wf_execute_plan(
  plan,
  approval,
  design,
  target_verified
)
```

An AI Agent can prepare and inspect the plan but cannot attest its own approval:

```{r agent-refusal, eval=FALSE}
agent_refusal <- tryCatch(
  wf_approve_plan(
    plan,
    approver = "Agent",
    role = "assistant",
    actor_type = "agent"
  ),
  wf_error_safety = function(condition) condition$data
)
agent_refusal[c("code", "severity", "next_actions")]
```

There are no force, bypass, ignore-source, auto-approve, auto-relax,
auto-widen, or silent method-switch flags.

## Attach outcomes only after locking

After execution, attach the immutable weight by exact ID. Impact assessment
computes fixed descriptive means or proportions and never calls planning or
calibration code.

```{r post-lock, eval=FALSE}
analysis_ready <- wf_attach_weights(
  analysis_frame,
  locked,
  id = "id",
  weight_name = ".weight"
)
impact <- wf_assess_impact(
  locked,
  analysis_frame,
  id = "id",
  outcomes = c("satisfaction", "approved")
)

wf_report(locked, audience = "decision")
wf_report(impact, audience = "statistician")
wf_audit_export(impact, "impact-audit.json")
```

## Agent contract

Non-interactive callers should catch `wf_error_safety` and read its stable
payload fields: `code`, `severity`, `field`, `evidence`, and `next_actions`.
They should report the condition, preserve the identities already produced, and
request the listed human action. They must not edit object fields, manufacture
an approval, retry with wider limits, or select a new target after seeing study
outcomes.

<!-- SAFE_WORKFLOW_END -->

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report.R
\name{wf_report}
\alias{wf_report}
\title{Build a weighting quality report}
\usage{
wf_report(
  w,
  target = NULL,
  audience = c("manager", "analyst", "decision", "statistician"),
  lang = NULL,
  output = c("object", "markdown", "html"),
  file = NULL
)
}
\arguments{
\item{w}{A \code{wf_weights}, \code{wf_blend_result}, \code{wf_safe_workflow}, or
\code{wf_impact} object.}

\item{target}{Optional \code{wf_target} for margin-residual diagnostics.}

\item{audience}{Report projection: \code{"manager"}/\code{"decision"} for a compact
decision view or \code{"analyst"}/\code{"statistician"} for full detail.}

\item{lang}{Output language. Resolution follows explicit argument,
\code{options(wfc.lang)}, session locale, then English fallback.}

\item{output}{Return a structured object, Markdown, or standalone HTML.}

\item{file}{Optional output path for Markdown or HTML.}
}
\value{
A \code{wf_quality_report} for \code{output = "object"}; otherwise a rendered
character string, or invisibly the structured report when \code{file} is used.
}
\description{
Creates one structured dossier with manager and analyst projections. It can
report calibrated/composed weights or a blend result and render the same
payload as Markdown or dependency-free HTML.
}
\examples{
# Panel attrition weights are derived from the panel itself, so this
# example runs end to end. Reporting calibrated weights follows the same
# call; see vignette("safe-weighting-workflow") for that path, which
# additionally requires a verified target and a recorded approval.
panel <- data.frame(
  id = paste0("p", 1:24),
  region = rep(c("A", "B"), each = 12),
  retained = rep(c(TRUE, TRUE, TRUE, FALSE), times = 6),
  age = rep(c(25, 35, 45, 55, 65, 75), times = 4),
  sex = rep(c("female", "male"), times = 12),
  stringsAsFactors = FALSE
)
weights <- wf_attrition(
  panel,
  retained = "retained",
  formula = ~ age + sex,
  id = "id",
  by = "region"
)

wf_report(weights, audience = "decision")

analyst <- wf_report(weights, audience = "statistician")
names(analyst$sections)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compose.R
\name{wf_compose}
\alias{wf_compose}
\title{Compose multiple weighting stages}
\usage{
wf_compose(..., id = NULL, normalize = c("none", "mean1", "sum"))
}
\arguments{
\item{...}{Two or more \code{wf_weights} objects.}

\item{id}{Optional shared ID column in each stage's \verb{$data}. If \code{NULL},
\code{wf_compose()} uses \code{data$id} when every stage has it, or row order when no
stage has an ID column.}

\item{normalize}{Weight normalization mode: \code{"none"}, \code{"mean1"}, or \code{"sum"}.}
}
\value{
A composed \code{wf_weights} object.
}
\description{
Multiplies compatible \code{wf_weights} objects into one auditable pipeline result.
}
\examples{
stage1 <- structure(
  list(
    data = data.frame(id = c("r1", "r2"), group = "A", weight = c(2, 3)),
    log = data.frame(group = "A"),
    achieved = NULL,
    provenance = list(method = "stage1")
  ),
  class = "wf_weights"
)
stage2 <- stage1
stage2$data$weight <- c(0.5, 2)
stage2$provenance$method <- "stage2"
wf_compose(stage1, stage2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report.R
\name{as.data.frame.wf_quality_report}
\alias{as.data.frame.wf_quality_report}
\title{Convert a weighting quality report to a data frame}
\usage{
\method{as.data.frame}{wf_quality_report}(x, ...)
}
\arguments{
\item{x}{A \code{wf_quality_report} object.}

\item{...}{Unused.}
}
\value{
The report's primary table.
}
\description{
Convert a weighting quality report to a data frame
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/autoweigh.R
\name{wf_autoweigh}
\alias{wf_autoweigh}
\title{Run a guided workflow from verified design and target objects}
\usage{
wf_autoweigh(
  design,
  target,
  dims,
  method = c("auto", "raking", "poststrat", "logit"),
  ladder = NULL,
  min_cell = NULL,
  bounds = c(0.3, 3),
  trim = "auto",
  max_deff = 6,
  max_residual = 0.02,
  interactive = base::interactive(),
  lang = NULL,
  ...
)
}
\arguments{
\item{design}{An unchanged \code{wf_design_data}.}

\item{target}{An unchanged, non-demo \code{wf_verified_target}.}

\item{dims}{A \code{wf_dims} declaration matching \code{target}.}

\item{method}{Calibration method. \code{"auto"} selects between supported safe
engines from declared settings only.}

\item{ladder}{Reviewed collapse ladder for post-stratification.}

\item{min_cell}{Positive post-stratification cell-size threshold.}

\item{bounds}{Two-element bounds for logit calibration.}

\item{trim}{Raking trim control.}

\item{max_deff}{Maximum design effect accepted by automatic trimming.}

\item{max_residual}{Maximum margin residual accepted by automatic trimming.}

\item{interactive}{Whether review prompts may be shown.}

\item{lang}{Output language.}

\item{...}{Method settings. ID and base-weight roles cannot be overridden.}
}
\value{
A \code{wf_autoweigh_result} carrying verified input identities.
}
\description{
Provides the guided WFC experience while enforcing the same WFC 2.0 input
boundary used by direct calibration and approved-plan execution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{wf_run}
\alias{wf_run}
\title{Run a production weighting pipeline}
\usage{
wf_run(
  spec,
  sample,
  dims = NULL,
  population = NULL,
  reference = NULL,
  base_weight = NULL
)
}
\arguments{
\item{spec}{A \code{wf_pipeline} object.}

\item{sample}{An unchanged \code{wf_design_data} object.}

\item{dims}{Optional \code{wf_dims} used to verify target dimensions.}

\item{population}{A verified population target for population mode.}

\item{reference}{A verified reference target for reference mode.}

\item{base_weight}{Deprecated runtime base-weight input. WFC 2.0 requires
the base-weight role to be declared in \code{sample}; supplying this argument
raises a safety error.}
}
\value{
A \code{wf_weights} object with pipeline provenance and optional
\verb{$pipeline_validation}.
}
\description{
Executes a \code{wf_pipeline()} specification against verified design and target
objects, records provenance, and evaluates declared validation thresholds.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{wf_diagnose}
\alias{wf_diagnose}
\title{Diagnose calibrated weights}
\usage{
wf_diagnose(w, target = NULL, sample = NULL, deff_ok = 3, deff_caveat = 10)
}
\arguments{
\item{w}{A \code{wf_weights} object.}

\item{target}{Optional \code{wf_target} object for residual margin checks.}

\item{sample}{Reserved for future diagnostics.}

\item{deff_ok}{Design-effect threshold for an OK verdict.}

\item{deff_caveat}{Design-effect threshold for caveat verdict.}
}
\value{
A \code{wf_diagnostics} object.
}
\description{
Computes per-group quality metrics for a \code{wf_weights} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precheck.R
\name{wf_precheck}
\alias{wf_precheck}
\title{Precheck sample and target compatibility}
\usage{
wf_precheck(
  sample,
  target,
  id = NULL,
  na = c("fractional", "drop", "error"),
  max_na_dims = 2,
  thin_min = 5,
  risk_ratio = 10
)
}
\arguments{
\item{sample}{Sample data frame.}

\item{target}{A \code{wf_target} object.}

\item{id}{Optional unique unit identifier column.}

\item{na}{Missing calibration data policy.}

\item{max_na_dims}{Maximum allowed missing calibration dimensions per row.}

\item{thin_min}{Minimum unweighted support before warning.}

\item{risk_ratio}{Target/sample share ratio warning threshold.}
}
\value{
A \code{wf_precheck} object.
}
\description{
Runs structural feasibility checks before weighting.
}
\examples{
dims <- wf_dims(gender = c("female", "male"))
pop <- data.frame(gender = c("female", "male"), count = c(55, 45))
target <- wf_target_population(pop, c(gender = "gender"), "count", dims)
sample <- data.frame(id = 1:4, gender = c("female", "male", "female", "male"))
wf_precheck(sample, target, id = "id")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot.wf_auto_trim}
\alias{plot.wf_auto_trim}
\title{Plot an automatic trim frontier}
\usage{
\method{plot}{wf_auto_trim}(x, lang = NULL, ...)
}
\arguments{
\item{x}{A \code{wf_auto_trim} object.}

\item{lang}{Output language.}

\item{...}{Additional arguments passed to the initial plot calls.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Draws the worst design effect and residual margin error for every feasible
finite cap.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight-plan.R
\name{wf_plan_weights}
\alias{wf_plan_weights}
\title{Build a reviewable outcome-blind weight plan}
\usage{
wf_plan_weights(
  design,
  target,
  dims,
  method = c("raking", "logit", "poststrat"),
  bounds = c(0.3, 3),
  min_cell = 5,
  cell_plan = NULL
)
}
\arguments{
\item{design}{A \code{wf_design_data} object.}

\item{target}{A non-demo \code{wf_verified_target} object.}

\item{dims}{A \code{wf_dims} object matching the target.}

\item{method}{Planned method: \code{"raking"}, \code{"logit"}, or \code{"poststrat"}.}

\item{bounds}{Planned multiplicative lower and upper bounds, satisfying
\verb{0 < L < 1 < U}.}

\item{min_cell}{Minimum support recorded in the plan.}

\item{cell_plan}{Optional reviewed \code{wf_cell_merge_plan} to apply before
calibration.}
}
\value{
A \code{wf_weight_plan}. The \code{weights} field is always \code{NULL}.
}
\description{
Performs structural checks and records approved-capable settings without
calling a calibration engine or computing any weights.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impact.R
\name{print.wf_impact}
\alias{print.wf_impact}
\title{Print post-lock outcome impact}
\usage{
\method{print}{wf_impact}(x, ...)
}
\arguments{
\item{x}{A \code{wf_impact} object.}

\item{...}{Reserved for future use.}
}
\value{
\code{x}, invisibly.
}
\description{
Print post-lock outcome impact
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poststrat-plan.R
\name{wf_plan_poststrat}
\alias{wf_plan_poststrat}
\title{Plan post-stratification cell resolution}
\usage{
wf_plan_poststrat(
  sample,
  target,
  min_cell,
  ladder,
  granularity = c("adaptive", "province"),
  empty_cell = c("redistribute", "flag", "error"),
  id = NULL
)
}
\arguments{
\item{sample}{Sample data frame.}

\item{target}{A \code{wf_target} object built with \code{keep_joint = TRUE}.}

\item{min_cell}{Minimum sample count per resolved cell.}

\item{ladder}{A \code{wf_collapse_ladder} object.}

\item{granularity}{Resolution strategy, either \code{"adaptive"} or \code{"province"}.}

\item{empty_cell}{Empty-cell policy, one of \code{"redistribute"}, \code{"flag"}, or
\code{"error"}.}

\item{id}{Reserved for future plan-level row diagnostics.}
}
\value{
A \code{wf_poststrat_plan} object.
}
\description{
Resolves each population joint cell to the finest supported collapse-ladder
level without computing weights. Review this plan before running
\code{wf_poststrat()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight-execution.R
\name{wf_approve_plan}
\alias{wf_approve_plan}
\title{Record a human attestation for a reviewed weight plan}
\usage{
wf_approve_plan(plan, approver, role, note = NULL, actor_type = "human")
}
\arguments{
\item{plan}{An unchanged, review-ready \code{wf_weight_plan}.}

\item{approver}{Non-empty approver name or organizational identifier.}

\item{role}{Non-empty reviewing role.}

\item{note}{Optional review note.}

\item{actor_type}{Actor type. Only \code{"human"} is accepted.}
}
\value{
A \code{wf_plan_approval} attestation bound to the plan identity.
}
\description{
Approval records are separate from plans. AI agents may prepare plans but
cannot create an approval while declaring themselves as an agent.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{wf_pipeline}
\alias{wf_pipeline}
\title{Declare a production weighting pipeline}
\usage{
wf_pipeline(target, stages, validate = NULL)
}
\arguments{
\item{target}{Target declaration. Use a list with \code{mode = "population"},
\code{"reference"}, or \code{"object"}, or pass a \code{wf_verified_target}.}

\item{stages}{Named stage list with a required \code{calibrate} stage.}

\item{validate}{Optional validation thresholds. Supported keys are
\code{max_deff} and \code{max_margin_dev}.}
}
\value{
A \code{wf_pipeline} object.
}
\description{
Creates a serializable specification for a recurring weighting run. The
specification records which verified target mode and calibration stage to
use and which post-run validation thresholds to check.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variance.R
\name{wf_variance}
\alias{wf_variance}
\title{Combine replicate weights and an estimator into a variance and CI.}
\usage{
wf_variance(
  replicates,
  estimator,
  data,
  level = 0.95,
  ci = c("normal", "percentile")
)
}
\arguments{
\item{replicates}{A \code{wf_replicate_weights} object.}

\item{estimator}{A closure \verb{function(weights, data) -> numeric} (scalar or
named vector).}

\item{data}{The data frame the estimator reads.}

\item{level}{Confidence level in \verb{(0, 1)}.}

\item{ci}{Interval type: \code{"normal"} or (bootstrap only) \code{"percentile"}.}
}
\value{
A \code{wf_variance_result} object.
}
\description{
Applies the unified replication rule
\verb{Var = scale * sum_r rscales_r * (theta_r - theta)^2} to any estimator, using
the \verb{(scale, rscales)} stored by \code{\link[=wf_replicates]{wf_replicates()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/target.R
\name{wf_target_population}
\alias{wf_target_population}
\title{Target from external population data}
\usage{
wf_target_population(
  pop,
  key_map,
  count,
  dims,
  by = NULL,
  by_key = NULL,
  scale = c("population", "sample", "custom"),
  sample = NULL,
  totals = NULL,
  keep_joint = FALSE
)
}
\arguments{
\item{pop}{Population data frame.}

\item{key_map}{Named character vector mapping \code{wf_dims} dimensions to population columns.}

\item{count}{Population count column.}

\item{dims}{A \code{wf_dims} object.}

\item{by}{Optional grouping variable.}

\item{by_key}{Optional group key column name or function.}

\item{scale}{Target scale, one of \code{"population"}, \code{"sample"}, or \code{"custom"}.}

\item{sample}{Required when \code{scale = "sample"}.}

\item{totals}{Required when \code{scale = "custom"}.}

\item{keep_joint}{Whether to retain per-group joint population cells.}
}
\value{
A \code{wf_target} object.
}
\description{
Converts arbitrary population data to a canonical \code{wf_target} object.
}
\examples{
dims <- wf_dims(gender = c("female", "male"))
pop <- data.frame(gender = c("female", "male"), count = c(55, 45))
wf_target_population(pop, c(gender = "gender"), "count", dims)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/target.R
\name{wf_target_reference}
\alias{wf_target_reference}
\title{Target from a weighted reference sample}
\usage{
wf_target_reference(
  ref,
  feature,
  dims,
  by = NULL,
  feature_na = c("error", "drop"),
  feature_gt1 = c("warn", "allow")
)
}
\arguments{
\item{ref}{Reference sample data frame.}

\item{feature}{Feature value column equal to \code{1 / design_weight}.}

\item{dims}{A \code{wf_dims} object.}

\item{by}{Optional grouping variable.}

\item{feature_na}{Policy for missing feature values.}

\item{feature_gt1}{Policy for feature values greater than one.}
}
\value{
A \code{wf_target} object.
}
\description{
Builds a canonical \code{wf_target} from a reference sample where \code{feature} is the
reciprocal of the design weight.
}
\examples{
dims <- wf_dims(gender = c("female", "male"))
ref <- data.frame(gender = c("female", "male"), feature = c(0.5, 0.25))
wf_target_reference(ref, "feature", dims)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/target-import.R
\name{wf_target_template}
\alias{wf_target_template}
\title{Create a safe external-target import template}
\usage{
wf_target_template(file, dims, by = NULL, example = FALSE)
}
\arguments{
\item{file}{Output path ending in \code{.csv} or \code{.xlsx}.}

\item{dims}{A \code{wf_dims} object with declared levels when \code{example = TRUE}.}

\item{by}{Optional grouping-column name.}

\item{example}{Whether to write synthetic, demo-only rows instead of a blank
template.}
}
\value{
Invisibly, a list with \code{data_file} and \code{source_file} paths.
}
\description{
Writes a CSV or Excel data template and a separate companion DCF source
record. The source filename appends \code{.source.dcf} to the complete data
filename so CSV and Excel files cannot accidentally share a checksum.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/support-merge.R
\name{wf_plan_cells}
\alias{wf_plan_cells}
\title{Plan deterministic, outcome-blind support-cell merging}
\usage{
wf_plan_cells(
  design,
  target,
  dims,
  min_cell = 5,
  max_weight_ratio = 4,
  boundary = target$by,
  ladder = NULL
)
}
\arguments{
\item{design}{A \code{wf_design_data} object.}

\item{target}{A non-demo \code{wf_verified_target} object.}

\item{dims}{A \code{wf_dims} object.}

\item{min_cell}{Minimum observed count in every retained joint design cell.}

\item{max_weight_ratio}{Maximum conservative marginal target-to-design ratio.}

\item{boundary}{Optional grouping column; defaults to the target grouping.}

\item{ladder}{Optional explicit \code{wf_collapse_ladder}.}
}
\value{
A reviewable \code{wf_cell_merge_plan}. No weights are computed.
}
\description{
Uses only declared design fields, external target margins, ordered dimension
levels, and an optional explicit collapse ladder. Study outcomes and custom
scoring callbacks are not accepted by this interface.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replicates.R
\name{wf_replicates}
\alias{wf_replicates}
\title{Generate re-calibrated replicate weights for variance estimation.}
\usage{
wf_replicates(
  data,
  refit,
  method = c("bootstrap", "jackknife", "brr"),
  R = 500,
  strata = NULL,
  clusters = NULL,
  id = NULL,
  base_weight = NULL,
  seed = NULL,
  rho = 0,
  parallel = FALSE,
  progress = FALSE
)
}
\arguments{
\item{data}{Input data frame (one row per unit).}

\item{refit}{A closure \verb{function(data, weights) -> wf_weights} that re-runs
the calibration pipeline using \code{weights} as the base/initial weights.}

\item{method}{Replication method.}

\item{R}{Number of bootstrap replicates (ignored for jackknife / BRR).}

\item{strata}{Optional stratum column name (single stratum if \code{NULL}).}

\item{clusters}{Optional PSU column name (each row is its own PSU if \code{NULL}).}

\item{id}{Optional id column aligning replicate weights (row order if \code{NULL}).}

\item{base_weight}{Optional starting base-weight column (all \code{1} if \code{NULL}).}

\item{seed}{Optional integer seed for the bootstrap draws.}

\item{rho}{Fay's BRR shrinkage parameter in \verb{[0, 1)}. Used only when
\code{method = "brr"}; \code{rho = 0} gives standard BRR.}

\item{parallel}{Whether to re-run replicate refits with forked parallelism
where available, using at most two workers. Windows falls back to serial
execution.}

\item{progress}{Whether to show a \code{cli} progress bar when \code{cli} is installed.}
}
\value{
A \code{wf_replicate_weights} object.
}
\description{
Perturbs base weights by bootstrap, jackknife, or BRR multipliers and
re-runs a calibration pipeline (\code{refit}) on each replicate, so the resulting
variance captures calibration uncertainty. Pair with \code{\link[=wf_variance]{wf_variance()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/influence.R
\name{wf_influence}
\alias{wf_influence}
\title{Diagnose high-influence calibrated units}
\usage{
wf_influence(w, target = NULL, sample = NULL, id = NULL, top = 20)
}
\arguments{
\item{w}{A \code{wf_weights} object with unit IDs.}

\item{target}{Optional \code{wf_target} for margin-share diagnostics.}

\item{sample}{Optional sample data frame aligned by \code{id}.}

\item{id}{Sample ID column required when \code{sample} and \code{target} are supplied.}

\item{top}{Number of highest-ratio units for which leave-one-out design
effects are computed and printed.}
}
\value{
A \code{wf_influence} object with a \verb{$table} data frame.
}
\description{
Reports the units that contribute most to weight instability. The core
diagnostics are per-unit weight ratio to the group mean, share of the
group's squared-weight mass, and leave-one-out design effect for the top
units. When \code{sample} and \code{target} are supplied, WFC also reports the largest
share of any target margin represented by that unit.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/propensity.R
\name{wf_propensity}
\alias{wf_propensity}
\title{Correct a non-probability sample by inverse-propensity pseudo-weighting.}
\usage{
wf_propensity(
  target,
  weight = c("ipw", "kernel", "matching"),
  stabilize = TRUE,
  trim = NULL
)
}
\arguments{
\item{target}{A \code{wf_target_propensity} object.}

\item{weight}{Pseudo-weight form. Only \code{"ipw"} is executable in this release;
\code{"kernel"} / \code{"matching"} are reserved.}

\item{stabilize}{Use stabilized IPW (\code{pi_bar / phat}) to tame extreme weights.}

\item{trim}{Optional positive scalar: clamp weights above \code{trim * median(w)}.}
}
\value{
A \code{wf_propensity_weights} object inheriting from \code{wf_weights}, with
\verb{$overlap} and \verb{$balance} diagnostics.
}
\description{
Fits the membership model declared in a \code{\link[=wf_target_propensity]{wf_target_propensity()}} object and
converts each online unit's fitted membership probability into a pseudo-design
weight. The result is a \code{wf_weights} object suitable as an \code{init_weight} for
\code{\link[=wf_rake]{wf_rake()}} / \code{\link[=wf_poststrat]{wf_poststrat()}} and as a stage in \code{\link[=wf_compose]{wf_compose()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poststrat.R
\name{wf_poststrat}
\alias{wf_poststrat}
\title{Post-stratify verified design data to a verified joint target}
\usage{
wf_poststrat(design, target, min_cell, ladder, ...)
}
\arguments{
\item{design}{An unchanged \code{wf_design_data}.}

\item{target}{An unchanged, non-demo \code{wf_verified_target} with joint cells.}

\item{min_cell}{Minimum sample count per resolved cell.}

\item{ladder}{A \code{wf_collapse_ladder} object.}

\item{...}{Post-stratification settings other than ID and base-weight roles,
which are owned by \code{design}.}
}
\value{
A \code{wf_weights} object with \code{cell_report} and \code{collapse_map}.
}
\description{
Enforces the WFC 2.0 verified-input boundary before cell-level
post-stratification.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poststrat-plan.R
\name{print.wf_poststrat_plan}
\alias{print.wf_poststrat_plan}
\title{Print a post-stratification plan}
\usage{
\method{print}{wf_poststrat_plan}(x, ...)
}
\arguments{
\item{x}{A \code{wf_poststrat_plan} object.}

\item{...}{Unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print a post-stratification plan
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variance.R
\name{as.data.frame.wf_variance_result}
\alias{as.data.frame.wf_variance_result}
\title{Coerce a variance result to a data frame}
\usage{
\method{as.data.frame}{wf_variance_result}(x, ...)
}
\arguments{
\item{x}{A \code{wf_variance_result} object.}

\item{...}{Unused.}
}
\value{
The result table as a data frame.
}
\description{
Coerce a variance result to a data frame
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dims.R
\name{wf_dims}
\alias{wf_dims}
\title{Declare calibration dimensions}
\usage{
wf_dims(..., .collapse = list())
}
\arguments{
\item{...}{Named dimension-level pairs. Use \code{NULL} to infer levels from the target.}

\item{.collapse}{Named list of collapse ladders.}
}
\value{
A \code{wf_dims} object.
}
\description{
Declare calibration dimensions
}
\examples{
dims <- wf_dims(gender = c("female", "male"), age = c("young", "old"))
dims
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interoperability.R
\name{as_svydesign}
\alias{as_svydesign}
\title{Convert WFC weights to a survey design}
\usage{
as_svydesign(
  w,
  data,
  id = "id",
  ids = ~1,
  strata = NULL,
  fpc = NULL,
  nest = FALSE,
  ...
)
}
\arguments{
\item{w}{A \code{wf_weights} object.}

\item{data}{Non-empty analysis data frame containing the units in \code{w}.}

\item{id}{Unit-ID column in \code{data}; matched to \code{w$data$id}.}

\item{ids}{Survey cluster formula, defaulting to independent units.}

\item{strata}{Optional survey strata formula.}

\item{fpc}{Optional finite-population-correction formula.}

\item{nest}{Whether cluster IDs should be relabeled to nest within strata.}

\item{...}{Additional arguments passed to \code{survey::svydesign()}. WFC owns
\code{weights} and \code{probs}; they cannot be overridden.}
}
\value{
A standard \code{survey.design2} object whose variables include the
aligned \code{.wf_weight} column.
}
\description{
Joins calibrated WFC weights onto an analysis data frame by exact unit ID and
creates a standard \code{survey::svydesign()} object. The \code{survey} package remains
a suggested dependency.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interoperability.R
\name{as_svrepdesign}
\alias{as_svrepdesign}
\title{Convert WFC replicate weights to a survey replicate design}
\usage{
as_svrepdesign(r, data, id = "id", degf = NULL, ...)
}
\arguments{
\item{r}{A \code{wf_replicate_weights} object.}

\item{data}{Non-empty analysis data frame containing the units in \code{r}.}

\item{id}{Unit-ID column in \code{data}; matched to \code{r$base$id}.}

\item{degf}{Optional survey design degrees of freedom.}

\item{...}{Additional arguments passed to \code{survey::svrepdesign()}. WFC owns
the variables, base/replicate weights, replication type and scales,
combined-weight setting, and MSE setting.}
}
\value{
A standard \code{svyrep.design} object whose variables include the aligned
\code{.wf_weight} base-weight column.
}
\description{
Aligns recalibrated WFC replicate weights to analysis data by exact unit ID
and creates a standard \code{survey::svrepdesign()} object. The returned design
uses full-estimate-centered (\code{mse = TRUE}) variance so its estimates reproduce
\code{\link[=wf_variance]{wf_variance()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight-plan.R
\name{print.wf_weight_plan}
\alias{print.wf_weight_plan}
\title{Print a reviewable weight plan}
\usage{
\method{print}{wf_weight_plan}(x, ...)
}
\arguments{
\item{x}{A \code{wf_weight_plan} object.}

\item{...}{Reserved for future use.}
}
\value{
\code{x}, invisibly.
}
\description{
Print a reviewable weight plan
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto-trim.R
\name{wf_auto_trim}
\alias{wf_auto_trim}
\title{Recommend a weight-trimming cap}
\usage{
wf_auto_trim(
  design,
  target,
  caps = c(2, 3, 4, 5, 6, 8, 10, 12),
  lo = 0.05,
  max_deff = 6,
  max_residual = 0.02,
  ...
)
}
\arguments{
\item{design}{An unchanged \code{wf_design_data}.}

\item{target}{An unchanged, non-demo \code{wf_verified_target}.}

\item{caps}{Unique positive finite upper-cap multipliers.}

\item{lo}{Positive lower-cap multiplier used for every finite candidate.}

\item{max_deff}{Maximum acceptable worst-group design effect.}

\item{max_residual}{Maximum acceptable worst relative margin residual.}

\item{...}{Additional raking settings. ID and base-weight roles are owned by
\code{design}; \code{trim} is not allowed because this function owns candidates.}
}
\value{
A \code{wf_auto_trim} object containing the candidate frontier and the
recommended cap. \code{Inf} means no trimming is needed; \code{NA} means no candidate
meets both criteria.
}
\description{
Sweeps candidate upper caps by rerunning raking and reports the trade-off
between design effect and residual margin error. The function recommends a
cap but never applies it to a production result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapse.R
\name{wf_suggest_collapse}
\alias{wf_suggest_collapse}
\title{Suggest collapse plans from precheck findings}
\usage{
wf_suggest_collapse(
  precheck,
  dims,
  checks = c("cat_infeasible", "support_thin", "risk_extreme_ratio"),
  max_steps = 1
)
}
\arguments{
\item{precheck}{A \code{wf_precheck} object.}

\item{dims}{A \code{wf_dims} object with optional collapse ladders.}

\item{checks}{Precheck identifiers eligible for collapse suggestions.}

\item{max_steps}{Maximum number of ladder steps to inspect per dimension.}
}
\value{
A \code{wf_collapse_plan} object.
}
\description{
Converts selected precheck issues into a reviewable collapse plan using
collapse ladders declared in \code{wf_dims()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight-execution.R
\name{wf_guided_plan}
\alias{wf_guided_plan}
\title{Prepare a guided safe weighting workflow}
\usage{
wf_guided_plan(
  data,
  id,
  calibration,
  dims,
  target_file,
  source_file,
  source_type = c("population", "reference"),
  key_map = NULL,
  count = NULL,
  feature = NULL,
  ...
)
}
\arguments{
\item{data}{Design-only data frame.}

\item{id}{Unique identifier column.}

\item{calibration}{Declared calibration and boundary columns.}

\item{dims}{A \code{wf_dims} object.}

\item{target_file}{CSV or XLSX target file.}

\item{source_file}{Companion source DCF file.}

\item{source_type}{Either \code{"population"} or \code{"reference"}.}

\item{key_map}{Population dimension-to-column mapping.}

\item{count}{Population count column.}

\item{feature}{Reference reciprocal-design-weight column.}

\item{...}{Named safe settings: \code{base_weight}, \code{strata}, \code{clusters}, \code{fpc},
\code{by}, \code{by_key}, \code{production}, \code{method}, \code{bounds}, \code{min_cell},
\code{max_weight_ratio}, \code{boundary}, and \code{ladder}.}
}
\value{
A reviewable \code{wf_safe_workflow}; no weights are computed.
}
\description{
This practitioner-oriented entry point composes design preparation, verified
target import, outcome-blind cell planning, and weight planning. It never
approves a plan or computes weights.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/suggest-ladder.R
\name{wf_suggest_ladder}
\alias{wf_suggest_ladder}
\title{Draft a post-stratification collapse ladder}
\usage{
wf_suggest_ladder(sample, target, dims, min_cell = 5)
}
\arguments{
\item{sample}{Sample data frame.}

\item{target}{A \code{wf_target} object.}

\item{dims}{A \code{wf_dims} object with explicit ordered levels for every
calibration dimension.}

\item{min_cell}{Minimum required unweighted sample support in every group.}
}
\value{
A \code{wf_ladder_draft} containing reviewable levels and a validated
\code{wf_collapse_ladder} in \verb{$ladder}.
}
\description{
Examines worst-group category support and proposes adjacent category merges
in the explicit level order declared by \code{\link[=wf_dims]{wf_dims()}}. The result is a draft
for review; no input is modified and no merge is applied automatically.
}
\examples{
data(wfc_example)
target <- wf_target_population(
  wfc_example$population,
  c(gender = "gender", age = "age"),
  "count",
  wfc_example$dims,
  by = "province"
)
wf_suggest_ladder(wfc_example$sample, target, wfc_example$dims, min_cell = 25)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight-execution.R
\name{wf_attach_weights}
\alias{wf_attach_weights}
\title{Attach locked weights to analysis data by exact unit ID}
\usage{
wf_attach_weights(data, weights, id, weight_name = ".weight")
}
\arguments{
\item{data}{Full analysis data frame.}

\item{weights}{An unchanged \code{wf_locked_weights} object.}

\item{id}{Unique identifier column in \code{data}.}

\item{weight_name}{Name of the appended weight column.}
}
\value{
\code{data} in its original row order with one locked-weight column.
}
\description{
Attach locked weights to analysis data by exact unit ID
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{wf_audit_export}
\alias{wf_audit_export}
\title{Export a self-contained WFC audit file}
\usage{
wf_audit_export(x, file, inputs = NULL, extra = NULL)
}
\arguments{
\item{x}{A WFC result, safe workflow, plan, approval, locked weights, verified
target, design-data object, or post-lock impact object.}

\item{file}{Output JSON file path.}

\item{inputs}{Optional named list of input objects whose hashes should be
recorded.}

\item{extra}{Optional additional metadata to include in the audit record.}
}
\value{
Invisibly returns \code{file}.
}
\description{
Writes a JSON audit record containing result provenance, pipeline metadata,
optional guided-workflow decision ledger, input hashes, and user-supplied
extra metadata. The writer is dependency-free and intended for machine
archiving, not human formatting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calibrate.R
\name{wf_calibrate}
\alias{wf_calibrate}
\title{Calibrate weights from verified design and target objects}
\usage{
wf_calibrate(design, target, method = "raking", ...)
}
\arguments{
\item{design}{An unchanged \code{wf_design_data} created by
\code{\link[=wf_prepare_design]{wf_prepare_design()}}.}

\item{target}{An unchanged, non-demo \code{wf_verified_target} created by
\code{\link[=wf_import_target]{wf_import_target()}} or \code{\link[=wf_import_reference]{wf_import_reference()}}.}

\item{method}{Calibration method: \code{"raking"}, \code{"poststrat"}, \code{"greg"}
(linear GREG), \code{"logit"} (bounded), \code{"soft"} (declared margin
relaxation), or \code{"ebal"} (categorical entropy balancing).}

\item{...}{Method settings. ID and base-weight roles come from \code{design} and
cannot be overridden. Inline target moments are not supported.}
}
\value{
A \code{wf_weights} object carrying the verified input identities.
}
\description{
Enforces the WFC 2.0 verified-input boundary before dispatching to a
calibration engine. Raw data frames and ordinary \code{wf_target} objects are not
accepted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight-execution.R
\name{wf_execute_plan}
\alias{wf_execute_plan}
\title{Execute an approved plan and lock its weights}
\usage{
wf_execute_plan(plan, approval, design, target)
}
\arguments{
\item{plan}{An unchanged \code{wf_weight_plan}.}

\item{approval}{A matching \code{wf_plan_approval}.}

\item{design}{The unchanged \code{wf_design_data} used to build the plan.}

\item{target}{The unchanged \code{wf_verified_target} used to build the plan.}
}
\value{
A \code{wf_locked_weights} object that also inherits from \code{wf_weights}.
}
\description{
Revalidates the plan, human approval, design data, target evidence, and any
stored cell map before calling exactly the method recorded in the plan.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight-execution.R
\name{wf_guided_execute}
\alias{wf_guided_execute}
\title{Execute a guided workflow with an external human approval}
\usage{
wf_guided_execute(workflow, approval)
}
\arguments{
\item{workflow}{An unchanged \code{wf_safe_workflow}.}

\item{approval}{A matching \code{wf_plan_approval} created separately.}
}
\value{
A \code{wf_locked_weights} object.
}
\description{
Execute a guided workflow with an external human approval
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{wf_validate}
\alias{wf_validate}
\title{Compare calibrated weights against a reference release}
\usage{
wf_validate(
  new,
  reference,
  target = NULL,
  max_deff_delta = 1,
  max_ess_loss = 0.2,
  max_total_shift = 0.05,
  max_margin_delta = 0.01,
  max_ratio_p99 = 2,
  on_issue = c("warn", "error", "none")
)
}
\arguments{
\item{new}{New \code{wf_weights} object.}

\item{reference}{Reference \code{wf_weights} object.}

\item{target}{Optional \code{wf_target} used to compare margin residual drift.}

\item{max_deff_delta}{Maximum allowed increase in design effect.}

\item{max_ess_loss}{Maximum allowed fractional ESS loss.}

\item{max_total_shift}{Maximum allowed fractional group-total shift.}

\item{max_margin_delta}{Maximum allowed absolute margin-residual increase.}

\item{max_ratio_p99}{Maximum allowed normalized 99th percentile unit-weight
ratio; the reciprocal threshold is applied to the 1st percentile.}

\item{on_issue}{Reaction when drift is detected: \code{"warn"}, \code{"error"}, or
\code{"none"}.}
}
\value{
A \code{wf_validation} object.
}
\description{
Detects drift in recurring production weights by comparing group coverage,
design effect, effective sample size, total weight, optional margin
residuals, and per-unit weight ratios for matching IDs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poststrat-plan.R
\name{summary.wf_poststrat_plan}
\alias{summary.wf_poststrat_plan}
\title{Summarize a post-stratification plan}
\usage{
\method{summary}{wf_poststrat_plan}(object, ...)
}
\arguments{
\item{object}{A \code{wf_poststrat_plan} object.}

\item{...}{Unused.}
}
\value{
The plan diagnostics data frame.
}
\description{
Summarize a post-stratification plan
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blend.R
\name{wf_blend}
\alias{wf_blend}
\title{Blend online and offline calibrated estimates}
\usage{
wf_blend(
  online,
  offline,
  by_cell,
  lambda = c("neff", "inverse_variance", "fixed"),
  lambda_fixed = NULL,
  outcome = NULL,
  level = c("cell", "group"),
  trim_lambda = c(0.05, 0.95),
  sensitivity = TRUE
)
}
\arguments{
\item{online}{Online-source \code{wf_weights}.}

\item{offline}{Offline-source \code{wf_weights}.}

\item{by_cell}{Character vector of cell columns.}

\item{lambda}{Lambda strategy: \code{"neff"}, \code{"inverse_variance"}, or \code{"fixed"}.}

\item{lambda_fixed}{Fixed lambda scalar or key table when \code{lambda = "fixed"}.}

\item{outcome}{Optional numeric outcome column.}

\item{level}{Lambda level: \code{"cell"} or \code{"group"}.}

\item{trim_lambda}{Two bounds used to clamp data-driven lambdas.}

\item{sensitivity}{Whether to compute a global-lambda sensitivity sweep.}
}
\value{
A \code{wf_blend_result} object.
}
\description{
Combines two \code{wf_weights} sources at the estimator level. Each source is
estimated within each fusion cell first; the source estimates are then
combined using the effective lambda recorded in the result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{print.wf_diagnostics}
\alias{print.wf_diagnostics}
\title{Print diagnostics}
\usage{
\method{print}{wf_diagnostics}(x, ...)
}
\arguments{
\item{x}{A \code{wf_diagnostics} object.}

\item{...}{Unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print diagnostics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precheck.R
\name{print.wf_precheck}
\alias{print.wf_precheck}
\title{Print a precheck result}
\usage{
\method{print}{wf_precheck}(x, ...)
}
\arguments{
\item{x}{A \code{wf_precheck} object.}

\item{...}{Unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print a precheck result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/target-import.R
\name{wf_import_reference}
\alias{wf_import_reference}
\title{Import a verified external reference-sample target}
\usage{
wf_import_reference(
  data_file,
  source_file,
  dims,
  feature,
  by = NULL,
  production = TRUE
)
}
\arguments{
\item{data_file}{Path to a CSV or XLSX reference sample.}

\item{source_file}{Path to its companion source DCF record.}

\item{dims}{A \code{wf_dims} object.}

\item{feature}{Name of the reciprocal-design-weight column.}

\item{by}{Optional grouping variable.}

\item{production}{Whether to reject demo-only sources.}
}
\value{
A \code{wf_verified_target} object.
}
\description{
Imports an external reference sample only when every file column has a
declared design role and the companion DCF record passes the provenance and
checksum checks.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapse.R
\name{wf_apply_collapse}
\alias{wf_apply_collapse}
\title{Apply a category collapse plan}
\usage{
wf_apply_collapse(sample, target, plan)
}
\arguments{
\item{sample}{Sample data frame.}

\item{target}{A \code{wf_target} object.}

\item{plan}{List with \code{dim} and named character vector \code{map}, or a
\code{wf_collapse_plan} object from \code{wf_suggest_collapse()}.}
}
\value{
A list with collapsed \code{sample} and \code{target}.
}
\description{
Applies category-merge maps consistently to sample data and a target object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poststrat-ladder.R
\name{wf_collapse_ladder}
\alias{wf_collapse_ladder}
\title{Declare a post-stratification collapse ladder}
\usage{
wf_collapse_ladder(dims, ...)
}
\arguments{
\item{dims}{A \code{wf_dims} object.}

\item{...}{Named ladder levels, each a list of named character vectors in
the form \code{old = "new"} by dimension.}
}
\value{
A \code{wf_collapse_ladder} object.
}
\description{
Creates an ordered, cumulative, validated category-collapse ladder for
post-stratification cell resolution. Level 0 is the raw classification; each
named level applies one additional set of dimension maps.
}
\examples{
dims <- wf_dims(age = c("young", "old"), education = c("low", "high"))
wf_collapse_ladder(
  dims,
  level1 = list(age = c(young = "all", old = "all"))
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impact.R
\name{wf_assess_impact}
\alias{wf_assess_impact}
\title{Assess descriptive outcome impact after weights are locked}
\usage{
wf_assess_impact(weights, data, id, outcomes, level = 0.95)
}
\arguments{
\item{weights}{An unchanged \code{wf_locked_weights} object.}

\item{data}{Analysis data frame with the exact locked unit-ID set.}

\item{id}{Unique identifier column in \code{data}.}

\item{outcomes}{Character vector naming outcome columns.}

\item{level}{Normal-approximation confidence level.}
}
\value{
A \code{wf_impact} object with fixed descriptive summaries.
}
\description{
Compares fixed unweighted and locked-weight estimates. This function cannot
accept target values, selection callbacks, or planning controls, and it never
recalculates weights.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rake.R
\name{wf_rake}
\alias{wf_rake}
\title{Rake verified design data to a verified external target}
\usage{
wf_rake(design, target, ...)
}
\arguments{
\item{design}{An unchanged \code{wf_design_data}.}

\item{target}{An unchanged, non-demo \code{wf_verified_target}.}

\item{...}{Raking settings other than ID and base-weight roles, which are
owned by \code{design}.}
}
\value{
A \code{wf_weights} object carrying verified input identities.
}
\description{
Enforces the WFC 2.0 verified-input boundary before grouped raking.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/design-data.R
\name{print.wf_design_data}
\alias{print.wf_design_data}
\title{Print outcome-free survey design data}
\usage{
\method{print}{wf_design_data}(x, ...)
}
\arguments{
\item{x}{A \code{wf_design_data} object.}

\item{...}{Reserved for future use.}
}
\value{
\code{x}, invisibly.
}
\description{
Print outcome-free survey design data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidiers.R
\name{wfc-tidiers}
\alias{wfc-tidiers}
\alias{tidy.wf_weights}
\alias{tidy.wf_diagnostics}
\alias{tidy.wf_blend_result}
\alias{tidy.wf_variance_result}
\alias{glance.wf_weights}
\alias{glance.wf_diagnostics}
\alias{glance.wf_blend_result}
\alias{glance.wf_variance_result}
\alias{augment.wf_weights}
\title{Broom-style projections for WFC results}
\usage{
\method{tidy}{wf_weights}(x, ...)

\method{tidy}{wf_diagnostics}(x, ...)

\method{tidy}{wf_blend_result}(x, ...)

\method{tidy}{wf_variance_result}(x, ...)

\method{glance}{wf_weights}(x, ...)

\method{glance}{wf_diagnostics}(x, ...)

\method{glance}{wf_blend_result}(x, ...)

\method{glance}{wf_variance_result}(x, ...)

\method{augment}{wf_weights}(x, data, id = "id", ...)
}
\arguments{
\item{x}{A supported WFC result object.}

\item{...}{Additional arguments. \code{augment.wf_weights()} does not accept
additional arguments.}

\item{data}{Analysis data to augment with unit weights.}

\item{id}{Unit-ID column in \code{data}, matched to \code{x$data$id}.}
}
\value{
\code{tidy()} returns an object's primary detail table. \code{glance()} returns
a one-row summary. \code{augment()} returns \code{data} with \code{.weight} and \code{.feature}
appended in data-row order.
}
\description{
These S3 methods return base data frames and are registered for the generic
names supplied by the suggested \code{generics} package. They do not require or
return tibbles, and all programmatic column names remain stable English keys.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/attrition.R
\name{wf_attrition}
\alias{wf_attrition}
\title{Estimate panel attrition weights}
\usage{
wf_attrition(
  panel,
  retained,
  formula,
  id = NULL,
  by = NULL,
  stabilize = TRUE,
  trim = NULL
)
}
\arguments{
\item{panel}{Full prior-wave panel data.}

\item{retained}{Column indicating whether each prior-wave unit was retained
into the next wave.}

\item{formula}{One-sided or two-sided formula naming retention predictors.}

\item{id}{Optional unique unit identifier column.}

\item{by}{Optional grouping column; separate retention models are fit by
group.}

\item{stabilize}{Whether to multiply inverse probabilities by the group
retention rate before normalization.}

\item{trim}{Optional positive scalar cap, applied as \code{trim * median(weight)}
within each group before mean-one normalization.}
}
\value{
A \code{wf_attrition_weights} object inheriting from \code{wf_weights}, with a
\verb{$balance} diagnostic table and per-group retention log.
}
\description{
Models wave-to-wave retention in a panel and returns inverse-retention
weights for retained units. The result inherits from \code{wf_weights}, so it can
be composed with later calibration stages through \code{\link[=wf_compose]{wf_compose()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto-trim.R
\name{print.wf_auto_trim}
\alias{print.wf_auto_trim}
\title{Print an automatic trim recommendation}
\usage{
\method{print}{wf_auto_trim}(x, ...)
}
\arguments{
\item{x}{A \code{wf_auto_trim} object.}

\item{...}{Unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print an automatic trim recommendation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/design-data.R
\name{wf_prepare_design}
\alias{wf_prepare_design}
\title{Prepare outcome-free survey design data}
\usage{
wf_prepare_design(
  data,
  id,
  calibration,
  base_weight = NULL,
  strata = NULL,
  clusters = NULL,
  fpc = NULL
)
}
\arguments{
\item{data}{A data frame containing design variables only.}

\item{id}{Name of the unique, non-missing record identifier column.}

\item{calibration}{Names of columns permitted for calibration.}

\item{base_weight}{Optional name of a positive base-weight column.}

\item{strata}{Optional names of stratification columns.}

\item{clusters}{Optional names of cluster columns.}

\item{fpc}{Optional names of finite-population-correction columns.}
}
\value{
A \code{wf_design_data} object.
}
\description{
Creates a strict design object containing only identifiers, calibration
variables, and declared sampling-design fields. Any undeclared column blocks
construction so study outcomes cannot silently enter weight planning.
}
\examples{
design_frame <- data.frame(
  id = sprintf("r\%02d", 1:8),
  sex = rep(c("F", "M"), 4),
  age = rep(c("18-34", "35+"), each = 4),
  base_weight = 1,
  stringsAsFactors = FALSE
)

design <- wf_prepare_design(
  design_frame,
  id = "id",
  calibration = c("sex", "age"),
  base_weight = "base_weight"
)
design

# An outcome column has no declared design role, so it blocks construction
# instead of silently entering weight planning.
with_outcome <- cbind(
  design_frame,
  satisfaction = seq(40, 70, length.out = 8)
)
try(wf_prepare_design(
  with_outcome,
  id = "id",
  calibration = c("sex", "age"),
  base_weight = "base_weight"
))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot.wf_diagnostics}
\alias{plot.wf_diagnostics}
\title{Plot weight diagnostics}
\usage{
\method{plot}{wf_diagnostics}(x, lang = NULL, ...)
}
\arguments{
\item{x}{A \code{wf_diagnostics} object.}

\item{lang}{Output language.}

\item{...}{Additional arguments passed to \code{\link[graphics:dotchart]{graphics::dotchart()}}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Draws design effects and effective-sample-size shares by group.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot.wf_weights}
\alias{plot.wf_weights}
\title{Plot calibrated weight distributions}
\usage{
\method{plot}{wf_weights}(x, max_groups = 9, lang = NULL, ...)
}
\arguments{
\item{x}{A \code{wf_weights} object.}

\item{max_groups}{Maximum number of groups to plot.}

\item{lang}{Output language.}

\item{...}{Additional arguments passed to \code{\link[graphics:hist]{graphics::hist()}}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Draws one histogram per group, including the group mean and recorded raking
trim bounds when available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/target-import.R
\name{wf_import_target}
\alias{wf_import_target}
\title{Import a verified external population target}
\usage{
wf_import_target(
  data_file,
  source_file,
  dims,
  key_map,
  count,
  by = NULL,
  by_key = NULL,
  production = TRUE
)
}
\arguments{
\item{data_file}{Path to a CSV or XLSX population table.}

\item{source_file}{Path to its companion source DCF record.}

\item{dims}{A \code{wf_dims} object.}

\item{key_map}{Named mapping from dimensions to data columns.}

\item{count}{Name of the population-count column.}

\item{by}{Optional grouping variable.}

\item{by_key}{Optional group-key column or function.}

\item{production}{Whether to reject demo-only sources.}
}
\value{
A \code{wf_verified_target} object.
}
\description{
Imports CSV or Excel population margins only when a companion DCF record
supplies complete provenance and a matching SHA-256 checksum.
}
\examples{
# WFC ships a synthetic demo target beside its companion evidence record.
csv_file <- system.file(
  "extdata", "safe-target-example.csv",
  package = "WFC"
)
csv_source <- paste0(csv_file, ".source.dcf")

dims_safe <- wf_dims(sex = c("F", "M"), age = c("18-34", "35+"))

# `production = FALSE` is required here only because the bundled file is
# declared demo-only. Do not use it to admit an authoritative source whose
# evidence record is incomplete.
target <- wf_import_target(
  csv_file,
  csv_source,
  dims_safe,
  key_map = c(sex = "sex", age = "age"),
  count = "count",
  production = FALSE
)
target$identity

# The same demo file is refused as a production source, so a synthetic
# file cannot impersonate an authoritative population.
try(wf_import_target(
  csv_file,
  csv_source,
  dims_safe,
  key_map = c(sex = "sex", age = "age"),
  count = "count"
))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/propensity.R
\name{wf_target_propensity}
\alias{wf_target_propensity}
\title{Build a propensity target: stacked reference frame and membership model spec.}
\usage{
wf_target_propensity(
  online,
  reference,
  formula,
  method = c("logit", "rf", "gbm"),
  by = NULL,
  id = NULL
)
}
\arguments{
\item{online}{Data frame: the self-selected (non-probability) sample.}

\item{reference}{Data frame: the probability reference sample.}

\item{formula}{Two-sided membership formula, e.g. \code{member ~ age + edu}. The
right-hand side names the model predictors; the left-hand side names the
membership indicator the constructor creates (\code{1} online, \code{0} reference).}

\item{method}{Fit backend. Only \code{"logit"} is executable in this release;
\code{"rf"} / \code{"gbm"} are reserved and abort in \code{\link[=wf_propensity]{wf_propensity()}}.}

\item{by}{Optional grouping column present in both frames; the propensity
model is fit within each group.}

\item{id}{Optional id column in \code{online}; when \code{NULL}, online units are
identified by row order.}
}
\value{
A \code{wf_target_propensity} object.
}
\description{
Stacks a self-selected \code{online} sample and a probability \code{reference} sample
into one frame with a membership indicator, so the online sample's selection
propensity can be modelled. No model is fit here; execution happens in
\code{\link[=wf_propensity]{wf_propensity()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{wfc_example}
\alias{wfc_example}
\title{Simulated survey weighting example data}
\format{
A list with three elements:
\describe{
\item{sample}{Simulated respondent-level sample data.}
\item{population}{Simulated population cell counts.}
\item{dims}{A \code{wf_dims} object for \code{gender} and \code{age}.}
}
}
\source{
Generated by \code{data-raw/make-wfc-example.R}.
}
\usage{
wfc_example
}
\description{
A small, fully simulated dataset for examples and tests. It contains no
private source records.
}
\keyword{datasets}
